#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "ltlc.h"
#include "tarjan.h"

static int debugdepth = 1;
static int snapshot_number = 1;

IFLTLDEBUG(
           static unsigned int number_of_found_sccs = 0;
           static unsigned int *stored_scc_length = 0;
           static ltlatom ***stored_sccs = 0;
           static ltlatom **stored_scc_entry = 0;

           static void print_prolog_stored_sccs(FILE *file, ltl_formula *f); 
           );

void ltl_print_stats(ltl_formula *f);

int ltldeb_get() {
  return debugdepth;
}

void ltldeb_set(int depth) {
  debugdepth = depth;
}

void ltldeb_inc() {
  debugdepth++;
}

void ltldeb_dec() {
  debugdepth--;
  if (debugdepth < 0) {
    printf("Warning: Debugging level < 0!\n");
  }
}

void ltlmsg(const char *msg, ...) {
  va_list ap;
  int i;
  fputs("LTL-C:", stdout);
  for (i=0; i<debugdepth; i++) {
    putchar(' ');
  }
  va_start(ap, msg);
  vfprintf(stdout, msg, ap);
  va_end(ap);
  fflush(stdout);
}

void ltl_printnode(ltlnode *node) {
  SP_atom atom;
  int arity;
  switch (node->type) {
  case int_node: printf("%i", node->id.id); break;
  case term_node:
    SP_get_functor(node->id.ref, &atom, &arity);
    printf("%s/%i", SP_string_from_atom(atom), arity);
    break;
  default: printf("**invalid**"); break;
  }
}

void ltl_printatom(ltlatom *atom) {
    printf("%d", atom->tarjan_number);
    printf(" ( "); ltl_printnode(atom->node); printf(" )");
}

void ltl_printnode_with_info(const char *info, ltlnode *node) {
    printf("%s",info);
    ltl_printnode(node);
    printf("\n");
}

void ltl_printterm_with_info(const char *info, SP_term_ref term) {
  SP_atom functor;
  int arity;
  SP_get_functor(term, &functor, &arity);
  const char* name = SP_string_from_atom(functor);
  printf("%s: %s\n", info, name);
}

static int trans_in_path(ltlatom *from, ltlatom *to, found_scc_type *found_scc) {
  if (found_scc) {
    int i;
    ltlatom **path = found_scc->path;
    for (i=1; i<found_scc->number_of_path_elements; i++) {
      if (from==path[i-1] && to==path[i])
        return 1;
    }
    if (found_scc->number_of_path_elements>0
        && from==path[i-1] 
        && to == found_scc->scc[0])
      return 1;
  }
  return 0;
}

static int atom_in_path(ltlatom *atom, found_scc_type *found_scc) {
  int i;
  if (found_scc) {
    for (i=0; i<found_scc->number_of_path_elements; i++) {
      if (atom==found_scc->path[i]) return 1;
    }
  }
  return 0;
}

static int atom_in_scc(ltlatom *atom, found_scc_type *found_scc) {
  int i;
  if (found_scc) {
    for (i=0; i<found_scc->number_of_scc_atoms; i++) {
      if (atom==found_scc->scc[i]) return 1;
    }
  }
  return 0;
}

static int trans_in_scc(ltlatom *from, ltlatom *to, found_scc_type *found_scc) {
  return atom_in_scc(from, found_scc) && atom_in_scc(to, found_scc);
}

void ltl_print_stats(ltl_formula *f) {
  ltlatom *atom;
  ltlnode *node;
  bitfield *evaluation;
  int anum;
  int i = 0;
  ltlmsg("Statistics:\n");
  ltldeb_inc();
  while (i<f->number_of_atoms) {
    get_atom(&f->astore, i, &atom, &evaluation);
    node = atom->node;
    anum = node->number_of_atoms;
    ltlmsg("Node ");ltl_printnode(node);printf(":\n");ltldeb_inc();
    ltlmsg("atoms %u - %u\n", node->first_atom_id, node->first_atom_id + anum - 1);
    ltldeb_dec();
    i += anum;
  }
  ltldeb_dec();
}

void print_atom_pointer(ltl_formula *f) {
  ltlatom *atom;
  ltlatom *trans;
  bitfield *evaluation;
  int i,j;
  ltlmsg("Atom pointer:\n");
  ltldeb_inc();
  for(i=0; i<f->number_of_atoms; i++) {
    get_atom(&f->astore, i, &atom, &evaluation);
    ltlmsg("atom %i: %p of node %p\n", i, atom, atom->node);
    ltldeb_inc();
    for (j=0; j<atom->number_of_transitions; j++) {
      trans = get_atom_trans(f, atom->first_transition_id + j);
      ltlmsg("-> %p\n", trans);
    }
    ltldeb_dec();
  }
  ltldeb_dec();
}

void ltl_snapshot(ltl_formula *f) {
  ltlmsg("take snapshot %i\n", snapshot_number);
  ltldeb_inc();
  ltl_print_graph(f, 0, snapshot_number++);
  ltldeb_dec();
}

static void ltl_print_prolog_bitfield(FILE *file, bitfield *field, int size) {
  int i;
  fputs("[",file);
  for (i=0; i<size; i++) {
    if (i>0) {
      fputs(",", file);
    }
    fputs(getbit(field,i)?"1":"0", file);
  }
  fputs("]",file);
}

static void ltl_print_prolog_atomstore(FILE *file, ltl_formula *f) {
  unsigned int i,j,id;
  bitfield *ev;

  ltlatom *atom;
  fprintf(file, "/* atom(AtomId,NodeId,Evaluations,Number of transitions) */\n");
  fprintf(file, "/* atrans(AtomId,ATransId,SuccAtomId) */\n");
  for (i=0; i<f->number_of_atoms; i++) {
    get_atom(&f->astore, i, &atom, &ev);
    fprintf(file, "atom(%u,", i);
    print_prolog_node(file, atom->node);
    fputs(",", file);
    ltl_print_prolog_bitfield(file,ev,f->closure_size);
    fprintf(file, ",%i).\n", atom->number_of_transitions);
    if (atom->number_of_transitions > 0) {
      for (j=0; j<atom->number_of_transitions; j++) {
        id = find_atom_id(f, get_atom_trans(f, atom->first_transition_id + j));
        fprintf(file, "atrans(%u,%u,%u).\n", i, atom->first_transition_id + j, id);
      }
    }
  }
}

static void ltl_print_gf_stack(FILE *file, ltl_formula *f,
                               ltlarray *bf_nodes, unsigned int bf_size) {
  loopitem *item;
  unsigned int i,atomid;
  fprintf(file, "/* bfstack(Index,AtomId,NodeId,PreviousIndex,Number of Encounters,[Used Untils Bitfield, Until Goals Bitfield]) */\n");
  for (i=0; i<bf_size; i++) {
    item = (loopitem*) get_array_element(bf_nodes, i);
    atomid = find_atom_id(f, item->atom);
    fprintf(file, "bfstack(%u,%u,%u,%u,%u,[", i, atomid, item->atom->node->id.id,item->previous, item->number_of_encounters);
    ltl_print_prolog_bitfield(file,used_untils_of_loopitem(f,item),f->number_of_untils);
    fputs(",", file);
    ltl_print_prolog_bitfield(file,until_goals_of_loopitem(item),f->number_of_untils);
    fputs(");\n",file);
  }
}

static void ltl_print_prolog_scc(FILE *file, ltl_formula *f, found_scc_type *found_scc) {
  int i;
  unsigned int id;
  fputs("/* path(List of AtomIds) */\n", file);
  fputs("path([", file);
  for (i=0; i<found_scc->number_of_path_elements; i++) {
    if (i>0) {
      fputs(",", file);
    }
    id = find_atom_id(f, found_scc->path[i]);
    fprintf(file, "%u", id);
  }
  fputs("]).\n", file);

  fputs("/* scc_entry(AtomId) */\n", file);
  id = find_atom_id(f, found_scc->scc_entry);
  fprintf(file, "scc_entry(%u).\n", id);

  fputs("/* scc(AtomId) */\n", file);
  for (i=0; i<found_scc->number_of_scc_atoms; i++) {
    id = find_atom_id(f, found_scc->scc[i]);
    fprintf(file, "scc(%u).\n", id);
  }
}

static void ltl_print_prolog_pform_ref(FILE *file, ltl_pform_ref *ref) {
  fprintf(file, "%s(%i)", ref->signum ? "pos" : "neg", ref->ref);
}

static void ltl_print_prolog_formula(FILE *file, ltl_formula *f) {
  int i;
  ltl_part_formula *pf;
  fputs("formula([", file);
  for (i=0; i<f->closure_size; i++) {
    if (i>0) {
      fputs(",", file);
    }
    pf = &f->closure[i];
    switch(pf->op) {
    case ltlap: fputs("ap", file); break;
    case ltltp: fputs("tp", file); break;
    case ltlnext: fprintf(file, "next(%i)", pf->args.case_next); break;
    case ltlnextnot: fprintf(file, "next_not(%i)", pf->args.case_nextnot); break;
    case ltlyesterday: fprintf(file, "yesterday(%i)", pf->args.case_yesterday); break;
    case ltlyesterdaynot: fprintf(file, "yesterday_not(%i)", pf->args.case_yesterdaynot); break;
    case ltlor: fputs("or(", file); 
      ltl_print_prolog_pform_ref(file, &pf->args.case_or.a); fputs(",", file);
      ltl_print_prolog_pform_ref(file, &pf->args.case_or.b); fputs(")", file); break;
    case ltluntil: fputs("until(", file);
      ltl_print_prolog_pform_ref(file, &pf->args.case_until.inv); fputs(",", file);
      ltl_print_prolog_pform_ref(file, &pf->args.case_until.goal);
      fprintf(file, ",%i)", pf->args.case_until.next); break;
    case ltlsince: fputs("since(", file);
      ltl_print_prolog_pform_ref(file, &pf->args.case_since.inv); fputs(",", file);
      ltl_print_prolog_pform_ref(file, &pf->args.case_since.goal);
      fprintf(file, ",%i)", pf->args.case_since.yest); break;
    default: fputs("unknown", file); break;
    }
  }
  fputs("]).\n", file);

  fputs("main_formula(", file);
  ltl_print_prolog_pform_ref(file, &f->mainref);
  fputs(").\n", file);
}

void ltl_print_prolog_graph(ltl_formula *f, found_scc_type *found_scc,
                            ltlarray *bf_nodes, unsigned int bf_size) {
  // printf("printing to ltl_graphdump.pl\n");
  FILE *file = fopen("ltl_graphdump.pl", "w");
  if(file) {
	#ifdef LTLSTORELOWLINK
	  fputs(":- discontiguous atom/6, atrans/2, scc/2, scc_entry/2.\n", file);
	  fputs("atom(Id,Node,Eval,NrOut,DFS,Lowlink) :- atom(Id,Node,Eval,NrOut,_).\n", file);
	#else
	  fputs(":- discontiguous atom/5, atrans/2, scc/2, scc_entry/2.\n", file);
	#endif

	  /* formula */
	  ltl_print_prolog_formula(file, f);
	  /* nodes */
	  print_prolog_nodestore(file, &f->nstore);
	  /* atoms */
	  ltl_print_prolog_atomstore(file, f);
	  /* scc */
	  if (found_scc) {
		ltl_print_prolog_scc(file, f, found_scc);
	  }
	  if (bf_size>0) {
		ltl_print_gf_stack(file,f,bf_nodes,bf_size);
	  }
  }
  fclose(file);
}

void ltl_print_graph(ltl_formula *f, found_scc_type *found_scc, int number) {
  unsigned int newnode,i,nodeindex,atomline,atomid2;
  int j;
  ltlatom *atom,*atom2;
  ltlnode *node,*oldnode;
  bitfield *bitfield;
  char buf[100];

  if (number >= 0) {
    snprintf(buf, 100, "debug_ltl_atoms.%i.dot", number);
  } else {
    strcpy(buf, "debug_ltl_atoms.dot");
  }
  // printf("printing to debug_ltl_atoms.dot\n");
  FILE *dot = fopen(buf,"w");
  atomline = 0;

  fprintf(dot, "digraph ltltableau {\n");

  /* write all atoms, clustered by nodes */
  oldnode = 0;
  nodeindex = 0;
  for (i=0; i<f->number_of_atoms; i++) {
    get_atom(&f->astore, i, &atom, &bitfield);    
    if (atom->node == 0) {
      fprintf(stderr, "node is NULL!!\n");
      exit(1);
    }
    node = atom->node;
    newnode = node!=oldnode;
    if (newnode) {
      if (oldnode!=0) fprintf(dot, "\n  }\n");
      atomline=0;
      fprintf(dot, "  subgraph cluster%u {\n", nodeindex);
      if (node->type==int_node) {
        fprintf(dot, "    label=\"%i(", node->id.id);
      } else {
        SP_atom atom;
        int arity;
        SP_get_functor(node->id.ref, &atom, &arity);
        fprintf(dot, "    label=\"%s(", SP_string_from_atom(atom));
      }
      fshow_bitfield(dot, bitfield, 0, f->number_of_aps);
      fprintf(dot, ")\"");
    }
    fprintf(dot, "\n    atom%u[label=\"%i -- %i(", i, i, atom->tarjan_number);
    if (f->number_of_yesterdays > 0) {
      fprintf(dot, "Y");
      fshow_bitfield(dot, bitfield, f->index_yesterdays, f->number_of_nexts);
    }
    if (f->number_of_nexts > 0) {
      fprintf(dot, " N");
      fshow_bitfield(dot, bitfield, f->index_nexts, f->number_of_nexts);
    }
    if (f->number_of_generics > 0) {
      fprintf(dot, " G");
      fshow_bitfield(dot, bitfield, f->index_generics, f->number_of_generics);
    }
    #ifdef LTLSTORELOWLINK
    fprintf(dot, ", %i", atom->tarjan_lowlink);
    #endif
    fprintf(dot, ")\"");
    if (atom->number_of_transitions < 0) {
      fprintf(dot, " style=filled,color=grey");
    } else if (atom_in_scc(atom, found_scc)) {
      fprintf(dot, " color=red");
    } else if (atom_in_path(atom, found_scc)) {
      fprintf(dot, " color=blue");
    }
    fprintf(dot, "];");
    
    if (newnode) {
      nodeindex++;
    }
    atomline++;
    oldnode = node;
  }
  fprintf(dot, "\n  }\n");

  /* write all transitions */
  for (i=0; i<f->number_of_atoms; i++) {
    /* ltlmsg("write trans. of atom %u\n", i); */
    get_atom(&f->astore, i, &atom, &bitfield);
    if (atom==0) printf("atom is NULL.\n");
    for (j=0; j<atom->number_of_transitions; j++) {
      /* ltlmsg("write trans from %u to %i\n", i, atom->first_transition_id + j); */
      atom2 = get_atom_trans(f, atom->first_transition_id + j);
      atomid2 = find_atom_id(f, atom2);
      fprintf(dot, "  atom%u->atom%u[", i, atomid2);
      if (trans_in_scc(atom, atom2, found_scc)) {
        fprintf(dot, "color=red");
      } else if (trans_in_path(atom, atom2, found_scc)) {
        fprintf(dot, "color=blue");
      }
      fprintf(dot, "];\n");
    }
  }

  fprintf(dot, "}\n");
  fclose(dot);
}

static void v_ltl_raise(const char *format, va_list ap) {
  int size = strlen(format) + 500;
  char *buffer = (char*) ltl_malloc(sizeof(char) * size);
  SP_term_ref arg,exception;
  SP_atom atom;

  vsnprintf(buffer, size, format, ap);

  arg = SP_new_term_ref();
  exception = SP_new_term_ref();
  atom = SP_atom_from_string(buffer);
  if (!atom) {
    /* SP_put_list_codes(arg, SP_new_term_ref(), buffer); */
    atom = SP_atom_from_string("**encoding problems");
  }
  SP_put_atom(arg, atom);
  SP_cons_functor(exception, SP_atom_from_string("ltlcexception"), 1, arg);
  SP_raise_exception(exception);

  ltl_free(buffer);
}

int ltl_raise(const char *format, ...) {
  va_list ap;
  va_start(ap, format);
  v_ltl_raise(format, ap);
  va_end(ap);
  return 1;
}

int ltl_assert(int pred, const char *format, ...) {
  va_list ap;
  if (!pred) {
    va_start(ap, format);
    v_ltl_raise(format, ap);
    va_end(ap);
  }
  return pred;
}

unsigned int find_atom_id(ltl_formula *f, ltlatom *atom) {
  unsigned int atomid;
  ltlatom *found;
  ltlnode *node = atom->node;
  bitfield *bitfield;
  int i;
  if (!node) ltlmsg("node is NULL!\n");
  for (i=0; i<node->number_of_atoms; i++) {
    
    atomid = node->first_atom_id + i;
    get_atom(&f->astore, atomid, &found, &bitfield);
    if (found==atom) return atomid;
  }
  printf("Warning: id of atom not found!\n");
  return 0;
}

void debug_print_trace(ltl_formula *f, modelitem *result,int resultlength,int is_deadlock) {
  int i;
  unsigned int id;
  ltlatom *atom;
  ltlnode *node;
  printf("Result AtomId: ");
  for (i=0; i<resultlength; i++) {
    atom = result[i].atom;
    id = find_atom_id(f, atom);
    printf("%5i",id);
  }
  printf("\n       NodeId: ");
  for (i=0; i<resultlength; i++) {
    node = (result[i].atom)->node;
    if (node->type == int_node) {
      printf("%5i",node->id.id);
    } else {
      printf(" ??? ");
    }
  }
  printf("\n");
}

IFLTLDEBUG(
static void print_prolog_stored_scc(FILE *file, ltl_formula *f, unsigned int nr) {
  unsigned int i;
  unsigned int size = stored_scc_length[nr];
  ltlatom **scc = stored_sccs[nr];
  fprintf(file, "scc_entry(%u,%u).\n", nr, find_atom_id(f, stored_scc_entry[nr]));
  for (i=0; i<size; i++) {
    fprintf(file, "scc(%u,%u).\n", nr, find_atom_id(f, scc[i]));
  }
}

static void print_prolog_stored_sccs(FILE *file, ltl_formula *f) {
  unsigned int i;
  for (i=0; i<number_of_found_sccs; i++) {
    print_prolog_stored_scc(file, f, i);
  }
}

void ltldebug_store_scc(tarjandata *t, unsigned int size) {
  const unsigned int nr = number_of_found_sccs;
  stored_scc_length = (unsigned int *) realloc(stored_scc_length,(nr+1)*sizeof(unsigned int));
  stored_sccs = (ltlatom ***) realloc(stored_sccs,(nr+1)*sizeof(ltlatom **));
  stored_scc_entry = (ltlatom **) realloc(stored_scc_entry,(nr+1)*sizeof(ltlatom *));

  stored_scc_length[nr] = size;
  stored_sccs[nr] = save_stack(t,size,stored_scc_entry+nr,0,0);
  
  number_of_found_sccs += 1;
}
           );
