% (c) 2018-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(regexp, [regexp_init/0,
  regexp_match/2,
  is_regexp/1,
  regexp_replace/4,
  regexp_search_first/3,
  regexp_search_first_detailed/4,
  regexp_search_all/3]).


/* the interface to a heap global variable stored and maintained in C */


foreign(cpp_regex_match,cpp_regex_match(+string,+string,+integer,[-integer])).
foreign(cpp_regex_replace,cpp_regex_replace(+string,+string,+string,[-atom])).
foreign(cpp_regex_search_all_str,cpp_regex_search_all_str(+string,+string,[-term])).
foreign(cpp_regex_search_first_str,cpp_regex_search_first_str(+string,+string,[-atom])).
foreign(cpp_regex_search_first,cpp_regex_search_first(+string,+integer,+string,[-term])).
foreign_resource(regexp,
  [
   cpp_regex_match,
   cpp_regex_replace,
   cpp_regex_search_first_str,
   cpp_regex_search_first,
   cpp_regex_search_all_str
  ]).

:- use_module(probsrc(error_manager)).

% match an atom agains a regular expression in ECMA syntax also represented as an atom
regexp_match(String,Pattern) :-
   atom(String), atom(Pattern),
   regexp_init,
   !,
   cpp_regex_match(String,Pattern,1,Res),
   %format('String ~w for Pattern ~w: Result=~w~n',[String,Pattern,Res]),
   (Res=1 -> true
    ; Res>1
     -> add_error(regexp,'Illegal Regular Expression Pattern (use ECMAScript syntax): ',Pattern),
      fail).
regexp_match(String,Pattern) :-
   format(user_error,'Illegal call: ',regepx_match(String,Pattern)),
   fail.

% check if an atom is a valid regular expression pattern
is_regexp(Pattern) :- atom(Pattern),
   regexp_init,
   !,
   cpp_regex_match('',Pattern,0,Res), % TO DO: will still print error messages
   Res<2. % 0 or 1 means pattern was ok
is_regexp(Pattern) :-
   format(user_error,'Illegal call: ',is_regexp(Pattern)),
   fail.


% match an atom agains a regular expression in ECMA syntax also represented as an atom
regexp_replace(String,Pattern,ReplStr,Result) :-
   atom(String), atom(Pattern), atom(ReplStr), regexp_init,
   !,
   cpp_regex_replace(String,Pattern,ReplStr,Result).
   %,format('Replace ~w for Pattern ~w with ~w: Result=~w~n',[String,Pattern,ReplStr,Result]).
regexp_replace(String,Pattern,ReplStr,Result) :-
   format(user_error,'Illegal call: ',regexp_replace(String,Pattern,ReplStr,Result)),
   fail.

% search for a first match of a pattern; returns empty string if no match found
regexp_search_first(String,Pattern,Result) :-
   atom(String), atom(Pattern), regexp_init,
   !,
   cpp_regex_search_first_str(String,Pattern,Result).
   %,format('Search ~w for Pattern ~w: Result=~w~n',[String,Pattern,Result]).
regexp_search_first(String,Pattern,Result) :-
   format(user_error,'Illegal call: ',regexp_search_first(String,Pattern,Result)),
   fail.

% search for a first match of a pattern;
% returns either match(Pos,Len,[MatchAtom|SubMatches]) or no-match
regexp_search_first_detailed(String,FromPosition,Pattern,Result) :-
   atom(String), atom(Pattern), FromPosition>0,
   regexp_init,
   !,
   Offset is FromPosition-1,
   cpp_regex_search_first(String,Offset,Pattern,Result).
   %, format('Search ~w for Pattern ~w: Result=~w~n',[String,Pattern,Result]).
regexp_search_first_detailed(String,From,Pattern,Result) :-
   format(user_error,'Illegal call: ',regexp_search_first_detailed(String,From,Pattern,Result)),
   fail.

:- use_module(library(lists),[reverse/2]).
% search for a first match of a pattern; returns empty string if no match found
regexp_search_all(String,Pattern,Result) :-
   atom(String), atom(Pattern), regexp_init,
   !,
   cpp_regex_search_all_str(String,Pattern,CRes),
   reverse(CRes,Result).
   %,format('Search ~w for Pattern ~w: Result=~w~n',[String,Pattern,Result]).
regexp_search_all(String,Pattern,Result) :-
   format(user_error,'Illegal call: ',regexp_search_first(String,Pattern,Result)),
   fail.

:- dynamic loaded/0.

regexp_init :-
  (loaded -> true
   ;  assert(loaded),
      %assert_dir,
      %assert(user:library_directory(prob_home('.'))),
      %print(loading_foreign_resource(myheap)),nl,
      safe_load_foreign_resource(regexp)
   ).

assert_dir :- (user:library_directory('.') -> true ; assert(user:library_directory('.'))).

safe_load_foreign_resource(R) :-
   E=error(existence_error(_,_),_),
   % print(loading_foreign_resource(R)),nl,
   on_exception(E,
                load_foreign_resource(library(R)),
                (format(user_error,'~n! Could not load ~w library!~n! Check that it is available in the lib folder of ProB.~n~n',[R]),throw(E))).

