#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "shm_common.h"
#include "hashtrie.h"

#include "transition_trie.h"


/*  In order to generate a trace, i.e. a path from the root node to an
 *  arbitrary state, we need to know one of the state's predecessors;
 *  otherwise, we have to re-calculate the entire state space.
 * 
 *  The idea is to store a mapping #to -> #from.
 *  We could use another hash trie for it. However, the prefix tree would
 *  be exactly the same.
 * 
 *  Hence, we use our existing hash trie to look up the index where #to
 *  is stored and allocate a block of memory ourselves; at this index
 *  we store #from (and maybe an identifier for the event).
 * 
 *  To generate a trace to a state, we look up the state in this allocated
 *  block via the provided prefix tree. Then we get this state's predecessor.
 *  If we repeat this, we will find predecessors until we hit the root node,
 *  which has no predecessor.
 *
 *  For this to work, we need to make sure we avoid loops. One way to do so is
 *  that we store the first predecessor to a state that we encounter.
 */


// FIXME:
// is it a problem if the proxy has not put the hash code yet?
// it might prove useful to send this information to another component
// that blocks until it finds a hash code


static void ensure_allocated(transition_chunks * transitions,
                             const int64_t index,
                             const size_t elements_per_digests_block) {
    while (elements_per_digests_block * transitions->size < index) {
        transitions->size++;
        transitions->chunks = realloc(transitions->chunks, sizeof(struct transition_chunks*) * transitions->size);
        transitions->chunks[transitions->size-1] = (transition_chunk) calloc(elements_per_digests_block, sizeof(transition));
    }
}


static void store_at_index(transition_chunks * transitions,
                           const int64_t index,
                           char * const hash,
                           const size_t elements_per_digests_block) {
    ensure_allocated(transitions, index, elements_per_digests_block);

    size_t block;
    size_t offset;
    get_pos_tuple(elements_per_digests_block, index, &block, &offset);

    transition_chunk c = transitions->chunks[block];
    // only put the first transition we encounter to avoid loops
    if (c[offset].digest == NULL) {
        c[offset].digest = hash;
    }
}

// TODO: also store transition?
void put_predecessor(transition_chunks *transitions,
                     const hashtrie t,
                     char const * const hash_from,
                     char const * const hash_to) {
    int64_t index_to = lookup_index(t, hash_to);
    int64_t index_from = lookup_index(t, hash_from);
    char * const digest_from = lookup_digest(t.shm, index_from);
    store_at_index(transitions, index_to, digest_from, t.shm->elements_per_digests_block);
}




static char const * const lookup_predecessor(const hashtrie t,
                         transition_chunks const * const transitions,
                         char const * const hash_to) {
    int64_t index = lookup_index(t, hash_to);
    size_t block;
    size_t offset;
    get_pos_tuple(t.shm->elements_per_digests_block, index, &block, &offset);
    
    return transitions->chunks[block][offset].digest;
}


static hash_list *h_generate_trace(const hashtrie t,
                                   transition_chunks const * const transitions,
                                   char const * const hash_to,
                                   hash_list *next) {
    if (hash_to == NULL) {
        return next;
    }

    char const * const via = lookup_predecessor(t, transitions, hash_to);

    hash_list *l = (hash_list*) malloc(sizeof(hash_list));
    memcpy(l->hash, hash_to, HASH_LEN);
    l->next = next;

    return h_generate_trace(t, transitions, via, l);
}


hash_list *generate_trace(const hashtrie t,
                          transition_chunks const * const transitions,
                          char const * const hash_to) {
    return h_generate_trace(t, transitions, hash_to, NULL);
}

void teardown_hash_list(hash_list **l) {
    assert(l);
    assert(*l);

    while (*l != NULL) {
        hash_list *old = *l;
        *l = old->next;
        free(old);
    }
}
