#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/queue.h"
#include "../common/hash.h"
#include "../common/shm_cln.h"
#include "../common/msg_headers.h"
#include "../common/magic_constants.h"

#include "../common/assert_helper.h"

#include "mid.h"
#include "mpull.h"
#include "mstats.h"
#include "mwork.h"

#include "../common/hashtrie.h"

#include <unistd.h>

#define HASHTRIE_ID 0

// TCP port list:
// PORT              Proxy  Master
// port_start + 0    PUSH   PULL
// port_start + 1    SUB    PUB
// port_start + 2    REQ    REP
// port_start + 3    -      PUB
// port_start + 4    PUB    -

zchunk_t *alloc_term(char *term, int len) {
    return zchunk_new(term, len);
}


size_t published_message_count = 0;

int safe_to_terminate(size_t max_states, size_t checked, size_t known) {
    if (checked == known) {
        return 1;
    }

    if (max_states > 0 && max_states <= checked) {
        return 1;
    }
   
    return 0;
}

int master_main(int port_start,
                int max_states,
                int strict_mode,
                zchunk_t *model,
                zchunk_t *fwp,
                char *master_ip,
                char *logfile,
                char *tmpdir,
                int hash_cycle_msecs) {
    (void) tmpdir;
    int fd_logfile = open(logfile, O_CREAT | O_RDWR | O_TRUNC, 0666);
    assert_esf(fd_logfile != -1, "Opening logfile failed. Filename was %s", logfile);

    dprintf(fd_logfile, "I am %s", master_ip);
    zsys_set_io_threads(2);

    zsys_set_sndhwm(200000); // seems to be a bad idea?
    zsys_set_rcvhwm(0);

    const int PORT_PULL = port_start;
    const int PORT_PUB  = port_start + 1;
    const int PORT_REP  = port_start + 2;
    const int PORT_PUB_HASHES = port_start + 3;

    zsock_t *pull = master_setup_pull(PORT_PULL);
    zsock_t *pub_cmds = master_setup_pub (PORT_PUB);
    zsock_t *rep  = master_id_setup  (PORT_REP);
    zsock_t *pub_hashes = master_setup_pub (PORT_PUB_HASHES);

    assert(zsock_rcvhwm(pull) == 0);
    //assert(zsock_sndhwm(pub) == 0);


    struct id_map proxy_ips = {0};
    mStats stats = {0};
    struct worker_queuesizes wqs = {};
    wQueue *q = init_queue(0);


    char *fwp_term = malloc(zchunk_size(fwp));
    memcpy(fwp_term, (char*) zchunk_data(fwp), zchunk_size(fwp));

    char fwp_hash[HASH_LEN];
    hash(fwp_term, zchunk_size(fwp), fwp_hash);

    enqueue(q, fwp_term, zchunk_size(fwp), fwp_hash);

    FILE *msg_log = fopen("msgs.log", "w");
    zsys_set_logstream(msg_log);


    /* FIXME: this should not be needed anymore after reworking the
     *        shared memory code, hopefully */
    /* idea: make sure shared memory is set up and initialised before
     *       we try to access to it.
     *       if the proxy has the same IP as the master, it will use a
     *       special header in its ID request.
     *       we serve ID requests until we see that special header.
     *       then we know that it is safe to access to the shared memory.
     *       after this, business as usual and set up of the reactor.
     *       by this, we would also know when it is safe to access the trie
     *       to gather the known hashcodes to send them to new machines. */

    while (master_h_id(rep, master_ip, PORT_PUB_HASHES, &proxy_ips, model, 0, (hashtrie) {0}) != P_ID_REQ_TRIE_UP) {
    }



    struct semap_env senv = open_semaphores(HASHTRIE_ID);
    hashtrie trie = get_trie(senv, HASHTRIE_KEY_START);



    zpoller_t *poller_pull = zpoller_new(pull, NULL);
    zpoller_t *poller_rep  = zpoller_new(rep, NULL);

    int work_sender_active = 1;
    zlist_t *hash_accumulator = zlist_new();
    int64_t last_time = zclock_mono();
    while (!zsys_interrupted) {
        if (zpoller_wait(poller_pull, 1)) { // it is okay to poll on this one for a while
            int rc = master_h_pull(pull, pub_cmds, pub_hashes, strict_mode, &stats, &proxy_ips, fd_logfile, &wqs, max_states, trie, msg_log, hash_accumulator);
            if (rc == -1) {
                break;
            }
        }
        if (zpoller_wait(poller_rep, 0)) {
            master_h_id(rep, master_ip, PORT_PUB_HASHES, &proxy_ips, model, 1, trie);
        }

        if (work_sender_active) {
            work_sender_active = work_sender_no_loop(&stats, &proxy_ips, q, fd_logfile);
        }
        int64_t tmp = zclock_mono();
        if (tmp - last_time >= hash_cycle_msecs) {
            if (zlist_size(hash_accumulator)) {
                send_hashes(pub_hashes, hash_accumulator, fd_logfile);
            }
            last_time = tmp;
        }
    }
    dprintf(fd_logfile, "[DEBUG] quit main loop\n");
    fflush(NULL);
    send_hashes(pub_hashes, hash_accumulator, fd_logfile);

    dprintf(fd_logfile, "[DEBUG] sent hashes\n");
    fflush(NULL);
    zlist_destroy(&hash_accumulator);

    puts("done");
    puts("blocking until all messages are processed by proxy...");
    dprintf(fd_logfile, "[DEBUG] checked: %ld, known: %ld\n", trie_get_checked_states_count(trie), trie_count_elements(trie));
    fflush(NULL);

    while (!safe_to_terminate(max_states, trie_get_checked_states_count(trie), trie_count_elements(trie))) {
        // take care of dangling hash messages etc.
        if (zpoller_wait(poller_pull, 1)) {
            int header;
            zmsg_t *msg;
            int rc = zsock_recv(pull, "im", &header, &msg);
            assert(rc == 0);
            dprintf(fd_logfile, "[DEBUG] checked: %ld, known: %ld\n", trie_get_checked_states_count(trie), trie_count_elements(trie));
            zmsg_print(msg); // TODO: probably refactor the result handling here
            fflush(NULL);
            if (header == P_IS_HASH_MSG) {
                rc = zsock_send(pub_hashes, "im", CTRL_ADD_HASHES, msg);
                assert(rc == 0);
            } else if (header == P_RESULT_MSG) {
                zmsg_print(msg); // TODO: probably refactor the result handling here
            }

        }
    }
    shutdown_everything(pub_cmds);

    fclose(msg_log);

    zpoller_destroy(&poller_pull);
    zpoller_destroy(&poller_rep);

    size_t checked_states = trie_get_checked_states_count(trie);
    size_t known_states = trie_count_elements(trie);
    dprintf(fd_logfile, "{:checked %lu, :known %lu}\n", checked_states, known_states);
    printf("{:checked %lu, :known %lu}\n", checked_states, known_states);

    teardown_id_map(&proxy_ips);
    teardown_worker_queuesizes(&wqs);
    teardown_queue(&q);
    teardown_hashtrie(&trie, 0, HASHTRIE_ID);

    zchunk_destroy(&model);
    zchunk_destroy(&fwp);

    zsock_destroy(&pull);
    zsock_destroy(&pub_hashes);
    zsock_destroy(&pub_cmds);
    zsock_destroy(&rep);
    close(fd_logfile);

    return 0;
}

