#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/msg_headers.h"
#include "../proxy/pstats.h"


static void test_update_queuesizes(void) {
    struct worker_queues wqs = {0};

    int dirty = update_qs(&wqs, 1, 0, 100);
    assert(dirty);
    assert(wqs.queuesizes != NULL);
    assert(wqs.len == 1);
    assert(wqs.empty_workers == 1);
    assert(wqs.balanced_workers == 0);
    assert(wqs.full_workers == 0);
    assert(wqs.in_transfer == 0);

    dirty = update_qs(&wqs, 1, 0, 100);
    assert(dirty == 0); // nothing changed
    assert(wqs.queuesizes != NULL);
    assert(wqs.len == 1);
    assert(wqs.empty_workers == 1);
    assert(wqs.balanced_workers == 0);
    assert(wqs.full_workers == 0);
    assert(wqs.in_transfer == 0);

    dirty = update_qs(&wqs, 1, 1, 100);
    assert(dirty); // worker is now balanced
    assert(wqs.queuesizes != NULL);
    assert(wqs.len == 1);
    assert(wqs.empty_workers == 0);
    assert(wqs.balanced_workers == 1);
    assert(wqs.full_workers == 0);
    assert(wqs.in_transfer == 0);

    dirty = update_qs(&wqs, 4, 2000, 100);
    assert(dirty); // added workers
    assert(wqs.queuesizes != NULL);
    assert(wqs.len == 4);
    assert(wqs.empty_workers == 2);
    assert(wqs.balanced_workers == 1);
    assert(wqs.full_workers == 1);
    assert(wqs.in_transfer == 0);

    free(wqs.queuesizes);
    printf(".");
}

static void test_create_stats_balanced(void) {
    struct worker_queues wqs = {0};
    size_t arr[] =  {0, 1, 123, 42};
    wqs.queuesizes = arr; 
    wqs.len = 3;
    wqs.empty_workers = 0;
    wqs.balanced_workers = 3;
    wqs.full_workers = 0;


    zmsg_t *msg = create_stats_msg(666, &wqs, 5000, 0);
    assert(zmsg_size(msg) == 2);

    char *header_s = zmsg_popstr(msg);
    int header;
    int rc = sscanf(header_s, "%d", &header);
    assert(rc == 1);
    char *id = zmsg_popstr(msg);
    zmsg_destroy(&msg);

    assert(header == PSTATS_HAS_PMS);
    assert(0 == strcmp(id, "666"));
    free(header_s);
    free(id);

    printf(".");
}

static void test_create_stats_proxy_balances(void) {
    struct worker_queues wqs = {0};
    size_t arr[] =  {0, 1, 2000, 0};
    wqs.queuesizes = arr; 
    wqs.len = 3;
    wqs.empty_workers = 1;
    wqs.balanced_workers = 1;
    wqs.full_workers = 1;


    zmsg_t *msg = create_stats_msg(666, &wqs, 5000, 0);
    assert(zmsg_size(msg) == 2);

    char *header_s = zmsg_popstr(msg);
    int header;
    int rc = sscanf(header_s, "%d", &header);
    assert(rc == 1);
    char *id = zmsg_popstr(msg);
    zmsg_destroy(&msg);

    assert(header == PSTATS_HAS_PMS);
    assert(0 == strcmp(id, "666"));
    free(header_s);
    free(id);

    printf(".");
}

static void test_create_stats_has_empty(void) {
    struct worker_queues wqs = {0};
    size_t arr[] =  {0, 1, 150, 0};
    wqs.queuesizes = arr; 
    wqs.len = 3;
    wqs.empty_workers = 1;
    wqs.balanced_workers = 2;
    wqs.full_workers = 0;
    wqs.in_transfer = 0;


    zmsg_t *msg = create_stats_msg(666, &wqs, 5000, 0);
    assert(zmsg_size(msg) == 4);

    char *header_s = zmsg_popstr(msg);
    int header;
    int rc = sscanf(header_s, "%d", &header);
    assert(rc == 1);
    char *id = zmsg_popstr(msg);
    char *amount_s = zmsg_popstr(msg);
    char *port_empty_str = zmsg_popstr(msg);
    zmsg_destroy(&msg);

    assert(header == PSTATS_HAS_EMPTY);
    assert(0 == strcmp(id, "666"));
    assert(0 == strcmp(amount_s, "1"));
    assert(0 == strcmp(port_empty_str, "5013"));
    free(port_empty_str);
    free(header_s);
    free(amount_s);
    free(id);


    printf(".");
}

static void test_create_stats_has_full(void) {
    struct worker_queues wqs = {0};
    size_t arr[] =  {0, 1, 1500, 2345};
    wqs.queuesizes = arr; 
    wqs.len = 3;
    wqs.empty_workers = 0;
    wqs.balanced_workers = 1;
    wqs.full_workers = 2;


    zmsg_t *msg = create_stats_msg(666, &wqs, 5000, 0);
    assert(zmsg_size(msg) == 3);

    char *header_s = zmsg_popstr(msg);
    int header;
    int rc = sscanf(header_s, "%d", &header);
    assert(rc == 1);
    char *id = zmsg_popstr(msg);
    char *full_str = zmsg_popstr(msg);
    zmsg_destroy(&msg);
    
    assert(header == PSTATS_HAS_FULL);
    assert(0 == strcmp(id, "666"));
    assert(0 == strcmp(full_str, "2"));
    free(header_s);
    free(full_str);
    free(id);

    printf(".");
}

void test_transfer_masking(void) {
    struct worker_queues wqs = {0};
    size_t arr[] = {0, 1, 2, 3, 4};
    wqs.queuesizes = arr;
    wqs.balanced_workers = 4;
    wqs.len = 4;

    const int thresh = 100;
    assert(wqs.in_transfer == 0);
    assert(wqs.balanced_workers == 4);
    int dirty = mark_transfer_begin(&wqs, 1, thresh);
    assert(wqs.in_transfer == 1);
    assert(wqs.balanced_workers == 3);
    dirty = mark_transfer_begin(&wqs, 2, thresh);
    assert(dirty);
    assert(wqs.in_transfer == 2);
    assert(wqs.balanced_workers == 2);
    assert(arr[0] == 0);
    assert(arr[1] == (1 | ((size_t) 1 << 63)));
    assert(arr[2] == (2 | ((size_t) 1 << 63)));
    assert(arr[3] == 3);
    assert(arr[4] == 4);

    dirty = mark_transfer_end(&wqs, 1, thresh);
    assert(dirty);
    assert(wqs.in_transfer == 1);
    assert(wqs.balanced_workers == 3);
    assert(arr[0] == 0);
    assert(arr[1] == 1);
    assert(arr[2] == (2 | ((size_t) 1 << 63)));
    assert(arr[3] == 3);
    assert(arr[4] == 4);

    // idempotency
    // FIXME: this should not happen in regular mode but the signal MIGHT get
    //        send when a full queue is moved to anywhere else. 
    //        work should be accepted and a signal sent anyway.
    //        this should have no impact though...
    dirty = mark_transfer_end(&wqs, 1, thresh);
    assert(dirty == 0);
    assert(wqs.in_transfer == 1);
    assert(wqs.balanced_workers == 3);
    assert(arr[0] == 0);
    assert(arr[1] == 1);
    assert(arr[2] == (2 | ((size_t) 1 << 63)));
    assert(arr[3] == 3);
    assert(arr[4] == 4);

    dirty = mark_transfer_end(&wqs, 2, thresh);
    assert(dirty);
    assert(wqs.in_transfer == 0);
    assert(wqs.balanced_workers == 4);
    assert(arr[0] == 0);
    assert(arr[1] == 1);
    assert(arr[2] == 2);
    assert(arr[3] == 3);
    assert(arr[4] == 4);
    printf(".");
}


static void test_get_max_queue(void) {
    pStats wqs = {0};
    wqs.len = 4;

    size_t queuesizes[] = {1, 0, 0, 0, 0};
    wqs.queuesizes=queuesizes;
    int rc = get_max_not_transfer_queue(&wqs);
    assert(rc == -1);
    rc = get_max_queue(&wqs);
    assert(rc == -1);


    size_t queuesizes2[] = {1, 1, 0, 0, 0};
    wqs.queuesizes=queuesizes2;
    rc = get_max_not_transfer_queue(&wqs);
    assert(rc == 1);
    rc = get_max_queue(&wqs);
    assert(rc == 1);


    size_t queuesizes3[] = {1, 1, 4, 3, 0};
    wqs.queuesizes=queuesizes3;
    rc = get_max_not_transfer_queue(&wqs);
    assert(rc == 2);
    rc = get_max_queue(&wqs);
    assert(rc == 2);


    size_t queuesizes4[] = {1, 1, 4, ((size_t) 1 << 63) | 3, 0};
    wqs.queuesizes=queuesizes4;
    rc = get_max_not_transfer_queue(&wqs);
    assert(rc == 2);
    rc = get_max_queue(&wqs);
    assert(rc == 2);


    size_t queuesizes5[] = {1, 1, 2, ((size_t) 1 << 63) | 3, 0};
    wqs.queuesizes=queuesizes5;
    rc = get_max_not_transfer_queue(&wqs);
    assert(rc == 2);
    rc = get_max_queue(&wqs);
    assert(rc == 3);


    size_t queuesizes6[] = {1, 1, 4, 3, 5};
    wqs.queuesizes=queuesizes6;
    rc = get_max_not_transfer_queue(&wqs);
    assert(rc == 4);
    rc = get_max_queue(&wqs);
    assert(rc == 4);

    printf(".");
}


static void test_proxy_can_balance(void) {
    pStats wqs = {0};
    wqs.len = 4;

    size_t queuesizes[] = {1, 0, 0, 0, 0};
    wqs.queuesizes=queuesizes;
    int rc = proxy_can_balance(&wqs, 100);
    assert(rc == 0);
    

    size_t queuesizes2[] = {1000, 0, 1, 0, 1000};
    wqs.queuesizes=queuesizes2;
    rc = proxy_can_balance(&wqs, 100);
    assert(rc == 1);

    size_t queuesizes3[] = {1000, 0, 0, 2000, 1000};
    wqs.queuesizes=queuesizes3;
    rc = proxy_can_balance(&wqs, 100);
    assert(rc == 2);

    size_t queuesizes4[] = {1000, 100, 0, 2000, 1000};
    wqs.queuesizes=queuesizes4;
    rc = proxy_can_balance(&wqs, 100);
    assert(rc == 1);

    size_t queuesizes5[] = {1000, 100, 1, 2000, 1000};
    wqs.queuesizes=queuesizes5;
    rc = proxy_can_balance(&wqs, 100);
    assert(rc == 0);


    printf(".");
}

static void test_get_empty_queue(void) {
    pStats wqs = {0};
    wqs.len = 4;

    size_t queuesizes[] = {1, 0, 0, 0, 0};
    wqs.queuesizes=queuesizes;
    wqs.empty_workers = 4;
    int rc = get_empty_queue(&wqs);
    assert(rc == 1);

    size_t queuesizes2[] = {1, 2, 0, 1, 0};
    wqs.empty_workers = 2;
    wqs.queuesizes=queuesizes2;
    rc = get_empty_queue(&wqs);
    assert(rc == 2);

    size_t queuesizes3[] = {1, 2, ((size_t) 1 << 63), 1, 0};
    wqs.empty_workers = 1;
    wqs.queuesizes=queuesizes3;
    rc = get_empty_queue(&wqs);
    assert(rc == 4);

    printf(".");
}

int main(int argc, char *argv[]) {
    test_update_queuesizes();
    test_create_stats_balanced();
    test_create_stats_proxy_balances();
    test_create_stats_has_empty();
    test_create_stats_has_full();
    test_get_max_queue();
    test_proxy_can_balance();
    test_get_empty_queue();

    test_transfer_masking();
    puts("");
    return 0;
}

