% (c) 2014-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)
:- module(worker, [start_worker/5,load/1]).

:- use_module(library(fastrw),[ fast_buf_read/2,fast_buf_write/3 ]).
:- use_module(library(lists)).

:- use_module(library(process)).

:- use_module('../../../src/module_information.pl').
:- use_module(probsrc(error_manager),[add_error/3]).
:- use_module('../zmq.pl', [init_zmq/0,
                            setup_worker/5, work_reactor/1, teardown_worker/0,
                            new_results_message/1, start_timer/0, stop_timer/0,
                            msg_put_str/2, put_succ/4, add_stats/1, send_and_destroy_msg/1]).

:- module_info(group,experimental).
:- module_info(description,'This is the interface to C code for distributed model checking.').
:- module_info(revision,'$Rev: 9609 $').
:- module_info(lastchanged,'$LastChangedDate: 2011-11-18 10:43:02 +0100 (Fr, 18 Nov 2011) $').


% The worker gets some initialisation code, e.g., a B model and the instruction what to check.
load(Init) :-
    fast_buf_read(Initialisation,Init),
    callback_load(Module,Pred),
    if(call(Module:Pred,Initialisation),true,
       (add_error(zmq_worker,'Loading spec failed: ',Module:Pred),fail)).


:-dynamic assertion_state/1.
:-dynamic assertion_task/2.


% The worker gets a Workpackage. It must produce a list of computation results
% (which are being forwarded to the master) and a list of successor Workpackages.
% The successor workpackages are sent to this predicate at a later point in time.
% It is not guaranteed that it will be sent to this process, so they must contain all
% required information to process the workpackage.
% It can be assumed that all processes that might receive this workpackage executed
% the same initialisation code.
%process(_,[some_result(4), another_result(x)], [foo(baz), doh(dude)]).


:- use_module(library(process)).
:- use_module(probsrc(translate)).
:- public process/3.
process(assertion,result([], []), S) :- findall(check(N),assertion_task(N,_),S).


process(check(N),R,[]) :- %process_id(PID),
                          assertion_task(N,A),
                          check_assertion(A,R).

process(root,R,S) :- process(state(root),R,S).
%process(cbc_test_case(Depth,State,Trace),R,S) :- ....
process(state(State),result(ResString, StateAtom),Successors) :- !,  %print(treating(State)),nl,flush_output,
 % statistics(walltime,[T1,_]),
 % statistics(runtime,[T3,_]),
  % prolog_flag(gc,G),print(gc(G)),nl,
   specfile:prepare_state_for_specfile_trans(State,PreparedState), % will perform unpacking of constants just once, for example
   (debug_mode->count(X),assert(state(X,State)),inc;true),
   (specfile:state_corresponds_to_initialised_b_machine(PreparedState,S2) ->
      bmachine:b_get_invariant_from_machine(Invariant),
      ( b_interpreter:b_test_boolean_expression_for_ground_state(Invariant,[],S2)
        % TO DO: add time_out, specialized invariants, assertions, ...
       -> ResString = [] %, print(invariant_ok),nl
        ;  ResString = ['invariant_violated'], translate:translate_bstate(State, StateAtom)
      )
  %  ,print(checked_inv(Res)),nl
    ; ResString = []  %, print(not_initialised(PreparedState)),nl
   ),

   findall(state(NewState),
           compute_successors(PreparedState,_ActionName,NewState),
           S),
  (debug_mode->dump_trans(PreparedState);true),
 %  print(successors(S)),nl,
%statistics(walltime,[T2,_]),
%statistics(runtime,[T4,_]),
%TW is T2 - T1,
%TR is T4 - T3,

%(TW > 10 -> ( print(time(TW,TR)),nl,statistics,print(x),nl);true),
    Successors = S.
process(max_reached,Res,[]) :- !,
   (succeed_max:max_reached -> Res=true ; Res=false).
process(X,_,S) :- print(unhandled(X)),nl, S=[].


check_assertion(A,result(R, StateAtom)) :- assertion_state(State),
    (b_interpreter:b_test_boolean_expression_for_ground_state(A,[],State)
      -> (R = [], StateAtom = [])
      ;  (mk_assertion_violation(A,R),
         translate:translate_bstate(State, StateAtom))).


mk_assertion_violation(Term, Atom) :-
    translate:translate_bexpression(Term,Atom).

compute_successors(CurState,ActionName,NewState) :-
   specfile:specfile_possible_trans_name(CurState,ActionName),
   specfile:specfile_trans(CurState,ActionName,_Act,NewState,_TransInfo,_Residue).
     % TO DO: check Residue, add time-out, deadlock, ...

% dump transitons for debugging
dump_trans(S) :- findall(trans(ActionName,NewState),
                         compute_successors(S,ActionName,NewState), List),
                 dump_trans(List,S).
dump_trans([],_).
dump_trans([trans(Action,H)|T],SS) :- assert(trans(Action,SS, H)), dump_trans(T,SS).

:- dynamic callback_load/2.

%:- dynamic is_initialised/0.
:- dynamic count/1.
:- dynamic state/2.
:- dynamic trans/3.
:- dynamic debug_mode/0.
%:- dynamic local_mode/0.

:- public set_debug/0.
set_debug :- assert(debug_mode).
%set_local :- assert(local_mode).

count(0).
inc :- count(X), retractall(count(_)), X1 is X + 1, assert(count(X1)).

%:- dynamic ids/1.
%ids(1).
%next_id :- ids(X), retractall(ids(_)), X1 is X + 1, assert(ids(X1)).


:- use_module(probsrc(state_space_exploration_modes),[depth_breadth_first_mode/1]).
:- meta_predicate start_worker(+,+,+,+,1).
start_worker(Port,ProxyID,Logfile,TmpDir,LoadCallbackModule:LoadCallback) :-
    % Host is not needed since we connect to a proxy at localhost
    init_zmq,
    retractall(callback_load(_,_)), assert(callback_load(LoadCallbackModule,LoadCallback)),
    depth_breadth_first_mode(MODE),
    (debug:debug_mode(off) -> Level = 20 ;
                              debug:debug_level(Level)),
    setup_worker(ProxyID, Port, MODE, Logfile, TmpDir), !,
    my_reactor_loop,
    teardown_worker.

my_reactor_loop :-
    work_reactor(X), !,
    (X =\= 0 -> print(work_reactor_retval(X)),nl,flush_output
              ; my_reactor_loop).


%:- dynamic queue/2.
%:- public local_zmq_check/1.
%local_zmq_check(N) :-
%  queue(N, Workpackage),
%  process(Workpackage, R, S),
%  add_result(R),
%  add_successor(S).

:- public zmq_check/3.
zmq_check(WPRead, HashMsg, FD_Logfile) :-
  %print(yo),nl,flush_output,
  start_timer,
  %print(start_timer),nl,flush_output,
  fast_buf_read(Workpackage, WPRead),
  %print(processing),nl,flush_output,
  process(Workpackage, Result, SuccessorList),
  %print(res(Result)),nl,flush_output,
  %print(succs(SuccessorList)),nl,flush_output,
  stop_timer,

  (Result = result(ResAtom, StateAtom)),
  put_successors(SuccessorList, HashMsg, FD_Logfile),
  put_results(ResAtom, StateAtom),

  add_stats(HashMsg),
  send_and_destroy_msg(HashMsg).

put_results([], _StateAtom). % no results -> don't create a message
put_results([H|T], StateAtom) :-
    new_results_message(Msg),
    put_results2([H|T], Msg),
    msg_put_str(Msg, StateAtom),
    send_and_destroy_msg(Msg).
put_results2([], _).
put_results2([H|T], Msg) :-
    msg_put_str(Msg, H),
    put_results(T, Msg).

/* TODO: fix symmetry reduction (see below for details) */
put_successors([], _Msg, _FD_Logfile).
put_successors([H|T], Msg, FD_Logfile) :-
    fast_buf_write(H, Len, Addr),
    put_succ(Msg, Addr, Len, FD_Logfile), % TODO: additional information?
    put_successors(T, Msg, FD_Logfile).

/* old code that supported symmetry reduction
:- use_module(probsrc(symmetry_marker),[compute_marker_for_state/2]).
add_successor([]).
add_successor([H|T]) :-
  (debug_mode->inc_transitions; true),
%  (local_mode->ids(X), next_id, assert(queue(X, H)) ; X = 0),
  H = state(S),
  (preferences:get_preference(symmetry_mode,hash) ->
    symmetry_marker:compute_marker_for_state(S,M)
    ; M = H),
  fast_buf_write(M,Len2,Addr2),
  hashchar(Addr2, Len2, Sha),

  fast_buf_write(H,Len1,Addr1),
  put_successor(Addr1,Len1, Sha, []),
  add_successor(T).

:- dynamic count_transitions/1.
count_transitions(0).
inc_transitions :- count_transitions(X), retractall(count_transitions(_)), X1 is X + 1, assert(count_transitions(X1)).
*/

print_substat(Atom) :-
  statistics(Atom, Val),
  print(':'), print(Atom), print(' '), print(Val), print(' ').

:- public print_stats/1.
print_stats(WorkerId) :-
  garbage_collect,
  print('{'),
  print(':workerid '), print(WorkerId), print(' '),
  process_id(PID),
  print(':PID '), print(PID), print(' '),
  print_substat(memory_used),
  print_substat(memory_free),
  print_substat(choice_used),
  print_substat(gc_count),
  print_substat(agc_count),
  print_substat(global_stack_used),
  print_substat(local_stack_used),
  print_substat(trail_used),
  print_substat(atoms_used),
  print('}'),
  nl, flush_output.

:- use_module(probsrc(tools_printing),[print_dynamic_pred/3]).
:- public dump/0.
dump :-
  portray_clause(start(0)), print('.'),nl,
  (succeed_max:max_reached -> print('/* max_reached */'),nl ; true),
  portray_clause((prop(S,'='(Var,Val)) :- state(S,L), member(bind(Var,Val),L))), print('.'),nl,
  print_dynamic_pred(worker,trans,3),
  print_dynamic_pred(worker,state,2).
