% (c) 2017-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)
:- module(alloy2b, [load_alloy_ast_prolog_file/1,
                    load_alloy_model/2,
                    load_alloy_model/3,
                    translate_alloy_model/2]).

:- use_module(probsrc(debug)).
:- use_module(library(lists)).
:- use_module(library(plunit)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(translate)).
:- use_module(probsrc(preferences), [get_preference/2,temporary_set_preference/2]).
:- use_module(library(file_systems)).
:- use_module(probsrc(error_manager)).
:- use_module(library(sets), [subtract/3]).
:- use_module(probsrc(specfile), [set_animation_minor_mode/1]).
:- use_module(library(ordsets), [ord_intersection/3,ord_union/3,ord_subtract/3]).

:- dynamic natural_type/1, integer_type/1, sig_field_id/2, enumerated_set/1, singleton_set/1, total_function/2, ordered_signature/1, command_counter/1, extending_signatures/1, is_sub_signature_of/3, is_seq/1, current_command/1, maxseq/1, artificial_parent_type/2.
:- volatile natural_type/1, integer_type/1, sig_field_id/2, enumerated_set/1, singleton_set/1, total_function/2, ordered_signature/1, command_counter/1, extending_signatures/1, is_sub_signature_of/3, is_seq/1, current_command/1, maxseq/1, artificial_parent_type/2.

:- meta_predicate map_translate(3, -, -, -).

command_counter(0).

default_max_seq(7).

% An automated translation from Alloy to classical B.
% The Alloy abstract syntax tree is translated to an untyped B AST as supported by ProB.
% Afterwards, the untyped AST can be typechecked and loaded by ProB.
% The used positions are from the Alloy parser and thus refer to the Alloy model.
% Usage:
% load_alloy_model/2: input is an Alloy model represented as a Prolog term as provided by the Kotlin translation; the model is translated to an untyped B AST and loaded by ProB
% translate_alloy_model/2: input is an Alloy model represented as a Prolog term as provided by the Kotlin translation, output is an untyped B AST
% Run tests:
% - launch ProB Tcl/Tk or ProB CLI
% - 'use_module(probsrc('alloy2b/alloy2b')).'
% - 'error_manager:reset_errors , plunit:run_tests(full_machine_translation).'.
% - run a single test: 'error_manager:reset_errors , plunit:run_tests(full_machine_translation:directctrl).'
% - 'bmachine:full_b_machine(L) , translate:print_machine(L)'.
% Further tests are integrated in probsrc(testcases).
%
% KNOWN ISSUES / TO DOs
% ----------------------
% - nested quantifiers: the usage of facts for singleton_set is error prone, and cannot deal with
% nested quantifiers with variable clashes, in particular when in the inner quantifier the
% identifier is not a singleton set
% - nested multiplicity annotations for non-binary relations (Joshua: is this possible in Alloy?)
% - variable capture due to use of DEFINITIONS
% - modulo currently only works for positive numbers

load_alloy_ast_prolog_file(File) :-
    \+ file_exists(File, read),
    !,
    add_error(load_alloy_ast_prolog_file, 'Alloy AST file has not been created:', [File]),
    fail.
load_alloy_ast_prolog_file(File) :-
    formatsilent('Opening Alloy AST file: ~w~n', [File]),
    open(File, read, Stream),
    safe_read_term(Stream, AlloyTerm),
    close(Stream),
    load_alloy_model(AlloyTerm, 'alloytranslation').

safe_read_term(Stream, Term) :-
    on_exception(error(E,_), read(Stream, Term),
                 (   add_internal_error('Exception while reading Prolog term from stream:', E),
                     fail
                 )).

assert_integer_type_from_preference :-
    get_preference(alloy_use_implementable_integers, Pref),
    retractall(integer_type(_)),
    retractall(natural_type(_)),
    (   Pref == true
    ->  asserta(integer_type('INT')),
        asserta(natural_type('NAT'))
    ;   asserta(integer_type('INTEGER')),
        asserta(natural_type('NATURAL'))
    ).

% TO DO: fix translation of Marc's machines

% TO DO: do we need to translate the 'expect' keyword for commands? is it inlined by the alloy parser?

load_alloy_model(AlloyTerm, AlloyFile) :-
    % use the first run or check command by default for models that use sequences, otherwise all commands are translated
    load_alloy_model(0, AlloyTerm, AlloyFile).

load_alloy_model(CurrentCommand, AlloyTerm, AlloyFile) :-
    assert_integer_type_from_preference,
    integer(CurrentCommand),
    reset_errors,
    debug_format(19, 'Translating AST to B.~n', []),
    retract_state,              % in case things not properly reseted or initialised
    assert(current_command(CurrentCommand)),
    % tools_printing:nested_print_term(AlloyTerm,10),
    translate_alloy_model(AlloyTerm, UntypedBAst),
    !,
    UntypedBAst = machine(generated(none,AbstractMachine)), % like @generated pragma
    b_load_machine_from_term(AlloyFile, complete_machine('alloytranslation',[AbstractMachine],[AlloyFile])),
    set_animation_minor_mode(alloy).
load_alloy_model(CurrentCommand, _, _) :-
    \+ integer(CurrentCommand),
    add_error(load_alloy_model, 'Current command has to be an integer:', CurrentCommand),
    fail.
load_alloy_model(_, AlloyTerm, _) :-
    add_error(load_alloy_model, 'Translating Alloy AST failed:', AlloyTerm),
    fail.

process_options(Options) :-
    (   \+member(sequences:_, Options)
    ;   \+member(parent_types:_, Options)
    ),
    !,
    add_error('Alloy2B parser failed to analyze types or check for sequences (be sure Alloy2B parser library is up-to-date).',[]),
    fail.
process_options(Options) :-
    process_options_safe(Options).

process_options_safe(Options) :-
    select(sequences:B, Options, RestOptions),
    (   B == true,
        !
    ;   retractall(current_command(_))),
    process_options_safe(RestOptions).
process_options_safe(Options) :-
    select(parent_types:SetOfTypeSets, Options, RestOptions),
    !,
    assert_artificial_parent_types(SetOfTypeSets),
    process_options_safe(RestOptions).
process_options_safe(_).

% Given a list of list of signature names. We need to introduce a parent type in B for each list of signatures.
assert_artificial_parent_types(SetOfTypeSets) :-
    assert_artificial_parent_types(0, SetOfTypeSets).

assert_artificial_parent_types(_, []).
assert_artificial_parent_types(C, [SetOfTypes|T]) :-
    gen_identifier(C, "_P", ParentId),
    C1 is C + 1,
    assert_parent_id_for_types(SetOfTypes, ParentId),
    assert_artificial_parent_types(C1, T).

assert_parent_id_for_types([], _).
assert_parent_id_for_types([SignatureName|T], ParentId) :-
    asserta(artificial_parent_type(SignatureName,ParentId)),
    assert_parent_id_for_types(T, ParentId).

% Log the current module to define unique field names.
:- dynamic module_in_scope_but_root/2, current_module_in_scope/1.
:- volatile module_in_scope_but_root/2, current_module_in_scope/1.

% Translate an Alloy model and all its imports (excluding some utility libraries we provide manual implementations for).
translate_alloy_model(alloy(RootModule,Models), UntypedBAst) :-
    retractall(translate:atelierb_mode(_)),
    select(alloy_model((RootModule,Alias),Facts,Assertions,Commands,Functions,Signatures,OSigNames,Options), Models, RestModels),
    !,
    translate_alloy_models(alloy_model((RootModule,Alias),Facts,Assertions,Commands,Functions,Signatures,OSigNames,Options), RestModels, UntypedBAst).

translate_alloy_models(alloy_model((RootModule,RootAlias),Facts,Assertions,Commands,Functions,Signatures,OSigNames,Options), RestModels, UntypedBAst) :-
    empty_machine_acc(MAcc),
    findall(ModuleName,(member(alloy_model((ModuleName,Alias),_,_,_,_,_,_,_),RestModels), asserta(module_in_scope_but_root(ModuleName,Alias))),_),
    translate_alloy_models(MAcc, alloy_model((RootModule,RootAlias),Facts,Assertions,Commands,Functions,Signatures,OSigNames,Options), RestModels, UntypedBAst).
translate_alloy_models(MAcc, AlloyModel, RestModels, UntypedBAst) :-
    translate_single_alloy_model(AlloyModel, MAcc, NewMAcc),
    (   RestModels == []
    ->      debug_println(19, build_machine_ast),
            build_machine_ast(NewMAcc, UntypedBAst),
            assert_atelierb_mode,
            % tools_printing:nested_print_term(UntypedBAst,20),
            (   debug_mode(off) ->
                true
            ;   nl,
                (   translate:print_raw_machine_terms(UntypedBAst)
                ;   debug_println(19, 'translate:print_raw_machine_terms/1 failed.')),
                nl
            ),
            retract_state,!
    ;   RestModels = [ImportedModel|T],
        translate_alloy_models(NewMAcc, ImportedModel, T, UntypedBAst)).

skip_module(ModuleName) :-
    supported_default_module(ModuleName); unsupported_module(ModuleName).

assert_atelierb_mode :-
    get_preference(alloy_translation_for_proof, Proof),
    Proof == true,
    !,
    temporary_set_preference(translate_print_typing_infos, true),
    asserta(translate:atelierb_mode(native)).
assert_atelierb_mode.

% Alloy utility modules we provide manual implementations for.
supported_default_module('util''boolean').
supported_default_module('util''integer').
supported_default_module('util''natural').
supported_default_module('util''ordering').
supported_default_module('util''relation').
supported_default_module('util''sequence').

unsupported_module('util''sequniv'). % uses univ type in function args; we probably do not need this module as it provides the same functions as util/sequence
unsupported_module('util''seqrel'). % TO DO: ?
unsupported_module('util''time'). % TO DO: ?
% following modules are translated using the tool: util/graph, util/ternary

% Give unique names for identifiers of imported modules (prepend module name).
% Identifier names in the root module are not extended.
assert_current_module_in_scope(ModuleName, Alias) :-
    module_in_scope_but_root(ModuleName, Alias),
    retractall(current_module_in_scope(_)),
    asserta(current_module_in_scope(ModuleName)).
assert_current_module_in_scope(_, _).

get_seq_scope_from_command([Command|_], CommandId, CmdMaxSeq) :-
    arg(7,Command,Id),
    Id==index(CommandId),
    arg(6,Command,maxseq(CmdMaxSeq)),
    !.
get_seq_scope_from_command([_|T], CommandId, CmdMaxSeq) :-
    get_seq_scope_from_command(T, CommandId, CmdMaxSeq).

translate_single_alloy_model(alloy_model((ModuleName,_),_,_,_,_,_,_,_), MAcc, MAcc) :-
    skip_module(ModuleName),
    !.
translate_single_alloy_model(alloy_model((ModuleName,Alias),facts(Facts),assertions(_Assertions),commands(Commands),functions(Functions),
                                  signatures(Signatures),ordered_signatures(OSigNames),Options), MAcc, NewMAcc) :-
    % singleton sets are asserted at runtime using singleton_set/1
    % accumulate all translations, afterwards we build the untyped machine ast
    assert_current_module_in_scope(ModuleName, Alias),
    process_options(Options),
    (   current_command(CommandId)
    ->  get_seq_scope_from_command(Commands, CommandId, CmdMaxSeq)
    ;   true
    ),
    maplist(assert_ordered_sig_name, OSigNames),
    % assertions are inlined in run or check commands and not translated in the properties
    extract_common_scopes(Commands, GS, ExactScopes, UpBoundScopes, BitWidth, CommonMaxSeq, Con),
    (   CommonMaxSeq = maxseq(MaxSeq)
    ;   CmdMaxSeq \== -1,
        MaxSeq = CmdMaxSeq
    ;   % default for current command only
        default_max_seq(MaxSeq)
    ),
    !,
    asserta(maxseq(MaxSeq)), % we need max seq scope for signature fields
    debug_println(19, translating_signatures),
    maplist(preprocess_signature, Signatures), % assert some signature names as singleton
    define_artifical_parent_types_if_necessary(MAcc,MAcc1),
    map_translate(translate_signature, MAcc1, Signatures, MAcc2),
    debug_format(19, 'Common scopes to all commands : global=~w, exact=~w, bw=~w, ms=~w, conflicts=~w~n', [GS,ExactScopes,BitWidth,MaxSeq,Con]),
    (   number(GS),
        GS > 0 ->
        preferences:temporary_set_preference(globalsets_fdrange, GS) % TO DO: improve
    ;   true
    ),
    get_preference(alloy_scopeless_translation,Scopeless),
    (   Scopeless == true
    ->  MAcc3 = MAcc2
    ;   try_set_bitwidth(BitWidth),
        maplist(generate_exact_scope_predicate, ExactScopes, ScopePreds1),
        maplist(generate_leq_scope_predicate, UpBoundScopes, ScopePreds2),
        append(ScopePreds1, ScopePreds2, ScopePreds),
        extend_machine_with_conjuncts(properties, ScopePreds, MAcc2, MAcc3)
    ),
    (   CommonMaxSeq \== none
    ->  % we can translate all commands at once if all have a common seq scope (otherwise only one at a time is translated)
        possibly_retract_current_command,
        (   Scopeless == true,
            MAcc4 = MAcc3,
            !
        ;   translate_seq_scopes(MaxSeq, SeqScopesPred),
            extend_machine_acc(properties, MAcc3, SeqScopesPred, MAcc4))
    ;   MAcc4 = MAcc3
    ),
    debug_println(19, translating_commands),
    map_translate(translate_command, MAcc4, Commands, MAcc5),
    debug_println(19, translating_functions),
    map_translate(translate_function, MAcc5, Functions, MAcc6),
    debug_println(19, translating_facts),
    map_translate(translate_fact, MAcc6, Facts, MAcc7),
    !,
    finalize_extending_signatures(MAcc7, NewMAcc).

% Introduce parent types in B for Alloy signatures without a common parent type but univ that interact with each other.
define_artifical_parent_types_if_necessary(MAcc, NewMAcc) :-
    findall(ParentId, artificial_parent_type(_, ParentId), TParents),
    remove_dups(TParents, Parents),
    define_artifical_parent_types(MAcc, Parents, NewMAcc).

define_artifical_parent_types(MAcc, [], MAcc).
define_artifical_parent_types(MAcc, [ParentId|T], NewMAcc) :-
    ParentId = identifier(_,ParentName),
    findall(SigName, artificial_parent_type(SigName,ParentId), DSigNames),
    DSigNames \== [],
    remove_dups(DSigNames, SigNames),
    maplist(translate_e_p, SigNames, TSigNames),
        % should be: _P0 = S1 \/ .. \/ Sn
        % ProB currently does not set the deferred set size of P0 in cbc mode if it is defined within an operation's precondition
        % (ProB uses the preference 'deferred_setsize' which then might be a contradiction)
        %join_untyped_ast_nodes(union, TSigNames, Union),
        %ParentDomain = equal(none, ParentId, Union),
    % TO DO: use above code if this issue is solved, otherwise stick to this translation which is just less performant in case we do not constrain _P0
    %        if we constrain _P0, e.g., using id(_P0), this could be an error as _P0 possibly contains more elements than necessary
    % Note: if preference 'deferred_setsize' is smaller than n, this is still a contradiction
    % S1 \/ .. \/ Sn <: P0
    join_untyped_ast_nodes(union, TSigNames, Union),
    ParentDomain = subset(none, Union, ParentId),
    extend_machine_acc(properties, MAcc, ParentDomain, MAcc1),
    % S1 /\ S2 /= {} & .. & Sn-1 /\ Sn /= {}
    disjoint_set_of_signatures(TSigNames, DisjointConstraint),
    extend_machine_acc(properties, MAcc1, DisjointConstraint, MAcc2),
    extend_machine_acc(sets, MAcc2, deferred_set(none,ParentName), MAcc3),
    define_artifical_parent_types(MAcc3, T, NewMAcc).

disjoint_set_of_signatures(TSigNames, DisjointConstraint) :-
    get_signature_pairs(TSigNames, SigPairs),
    disjoint_set_of_signatures_from_pairs(SigPairs, [], DisjointList),
    join_untyped_ast_nodes(conjunct, DisjointList, DisjointConstraint).

disjoint_set_of_signatures_from_pairs([], Acc, Acc).
disjoint_set_of_signatures_from_pairs([(S1,S2)|T], Acc, DisjointList) :-
    Disjoint = equal(none,intersection(none,S1,S2),empty_set(none)),
    disjoint_set_of_signatures_from_pairs(T, [Disjoint|Acc], DisjointList).

get_signature_pairs(TSigNames, SigPairs) :-
    findall((S1,S2), (nth0(I1, TSigNames, S1), nth0(I2, TSigNames, S2), I1 < I2), SigPairs).

possibly_retract_current_command :-
    % only a single command in proof mode
    get_preference(alloy_translation_for_proof, Proof),
    Proof == true,
    !.
possibly_retract_current_command :-
    retractall(current_command(_)).

map_translate(_, MAcc, [], MAcc).
map_translate(TypePred, MAcc, [Part|T], Res) :-
    call(TypePred, MAcc, Part, NewMAcc),
    map_translate(TypePred, NewMAcc, T, Res).

try_set_bitwidth(BitWidth) :-
    BitWidth = bitwidth(BW),
    number(BW),
    BW > 0,
    !,
    MaxInt is integer(2**(BW-1))-1,
    MinInt is integer(-(2**(BW-1))),
    preferences:temporary_set_preference(maxint, MaxInt),
    preferences:temporary_set_preference(minint, MinInt).
try_set_bitwidth(_).

% things that need to be done before translating the signatures, e.g., assert singleton set info
preprocess_signature(signature(Name,_Fields,_Facts,Options,_)) :-
    % assert signatures for singleton checks
    assert_singleton_set(Options, Name),
    assert_enumerated_set(Options, Name).

% translate signature declaration sig Name {Fields}
translate_signature(MAcc, signature(Name,Fields,Facts,Options,pos(Col,Row)), NewMAcc) :-
    extend_machine_acc(signatures, MAcc, [Name], MAcc1),
    translate_signature_aux(Name, Options, pos(Col,Row), MAcc1, MAcc2),
    translate_signature_fields(Fields, Name, MAcc2, MAcc3),
    translate_e_p(Name, TName),
    map_translate(translate_signature_fact(TName), MAcc3, Facts, NewMAcc).

translate_signature_aux(Name, Options, Pos, MAcc, NewMAcc) :-
    % ordered signatures are defined as distinct sets of integer when translating a command
    member(ordered, Options),
    !,
    define_ordered_signature_functions(Pos, MAcc, Name, MAcc1),
    extend_machine_acc(signatures, MAcc1, [Name], MAcc2),
    translate_e_p(Name, TID),
    integer_type(IntOrInteger),
    extend_machine_acc(properties, MAcc2,
                       member(none,TID,pow_subset(none,integer_set(none,IntOrInteger))), MAcc3),
    define_sig_as_set_or_constant_aux(constants, MAcc3, Name, Options, Pos, NewMAcc).
translate_signature_aux(Name, Options, Pos, MAcc, NewMAcc) :-
    define_sig_as_set_or_constant(MAcc, Name, Options, Pos, NewMAcc).

define_ordered_signature_functions(POS, MAcc, Sig, NewMAcc) :-
    translate_pos(POS, BPOS),
    gen_identifier(0, "_c_", TIDX),
    TIDS = identifier(BPOS,s),
    translate_e_p(Sig, TSig),
    MemberX = member(BPOS,TIDX,TSig),
    % next_Sig(s)  == IF succ[s] <: Sig THEN succ[s] ELSE {} END
    atom_concat(next_, Sig, NextName),
    SuccessorImg = image(BPOS,successor(BPOS),TIDS),
    PredecessorImg = image(BPOS,predecessor(BPOS),TIDS),
    extend_machine_acc(definitions, MAcc,
                       expression_definition(BPOS,NextName,[TIDS],if_then_else(BPOS,subset(BPOS,SuccessorImg,TSig),SuccessorImg,empty_set(BPOS))), MAcc1),
    % nexts_Sig(s) == {x|x:Sig & min({x} \/ s) /= x}
    atom_concat(nexts_, Sig, NextsName),
    NextsBody = conjunct(BPOS,MemberX,not_equal(BPOS,min(BPOS,union(BPOS,set_extension(BPOS,[TIDX]),TIDS)),TIDX)),
    extend_machine_acc(definitions, MAcc1,
                       expression_definition(BPOS,NextsName,[TIDS],comprehension_set(BPOS,[TIDX],NextsBody)), MAcc2),
    atom_concat(prev_, Sig, PrevName),
    % prev_Sig(s)  == IF pred[s] <: Sig THEN pred[s] ELSE {} END
    extend_machine_acc(definitions, MAcc2,
                       expression_definition(BPOS,PrevName,[TIDS],if_then_else(BPOS,subset(BPOS,PredecessorImg,TSig),PredecessorImg,empty_set(BPOS))), MAcc3),
    % prevs_Sig(s) == {x|x:Sig & max({x} \/ s) /= x}
    atom_concat(prevs_, Sig, PrevsName),
    PrevsBody = conjunct(BPOS,MemberX,not_equal(BPOS,max(BPOS,union(BPOS,set_extension(BPOS,[TIDX]),TIDS)),TIDX)),
    extend_machine_acc(definitions, MAcc3,
                       expression_definition(BPOS,PrevsName,[TIDS],comprehension_set(BPOS,[TIDX],PrevsBody)), NewMAcc).

translate_signature_fields(Fields, SignatureName) -->
    { translate_e_p(SignatureName, TSignatureName) },
    translate_signature_fields_aux(Fields, SignatureName, TSignatureName).

translate_signature_fields_aux([], _, _) -->
    [].
translate_signature_fields_aux([Field|T], SignatureName, TSignatureName) -->
    translate_signature_field(SignatureName, TSignatureName, Field),
    translate_signature_fields_aux(T, SignatureName, TSignatureName).

translate_signature_field(SignatureName, TSignatureName, Field, MAcc, NewMAcc) :-
    translate_signature_field_aux(SignatureName, TSignatureName, Field, TField, MAcc, MAcc1),
    extend_machine_acc(properties, MAcc1, TField, NewMAcc).

translate_signature_field_aux(SignatureName, TSignatureName,
                              field(FieldID,Expr,_FieldType,Options,POS), TFieldPred, MAcc, NewMAcc) :-
    FieldID = identifier(FieldName,_,_),
    asserta(sig_field_id(FieldName, SignatureName)),
    translate_e_p(FieldID, TFieldID),
    extend_machine_acc(constants, MAcc, TFieldID, MAcc1),
    translate_pos(POS, BPOS),
    disjoint_field_declaration(TSignatureName, BPOS, Options, MAcc1, TFieldID, NewMAcc),
    translate_field_decl(BPOS, SignatureName, TSignatureName, Expr, FieldID, TFieldID, TFieldPred),
    !.

translate_function_field(field(FieldName,Expr,type(_Type,_Arity),_Options,POS), TField) :-
    % (Functor = function -> assert_singleton_set(FieldName) ; true) , % NO LONGER NEEDED
    translate_e_p(FieldName, TTFieldName),
    strip_singleton_set_from_this(TTFieldName, TFieldName), % necessary for nested definition calls with 'this'
    translate_pos(POS, BPOS),
    fun_or_pred_decl_special_cases(Expr, TFieldName, BPOS, TField, _IsSingleton),
    !.
% TO DO: !! what if IsSingleton is set !!
disjoint_field_declaration(TSignatureName, TPos, Options, MAcc, TFieldName, NewMAcc) :-
    member(disj, Options),
    !,
    % all a, b: S | a != b implies no a.f & b.f
    IdA = identifier(TPos,a),
    IdB = identifier(TPos,b),
    ImplLhs = conjunct(TPos,conjunct(TPos,member(TPos,IdA,TSignatureName),member(TPos,IdB,TSignatureName)),
                       not_equal(TPos,IdA,IdB)),
    ImplRhs = equal(TPos,intersection(TPos,image(TPos,TFieldName,set_extension(none,[IdA])),
                                      image(TPos,TFieldName,set_extension(none,[IdB]))),empty_set(TPos)),
    Disjoint = forall(TPos,[IdA,IdB],implication(TPos,ImplLhs,ImplRhs)),
    extend_machine_acc(properties, MAcc, Disjoint, NewMAcc).
disjoint_field_declaration(_, _, _, MAcc, _, MAcc).


is_disjoint_quantifier(Fields) :-
    Fields = [F|_],
    is_disjoint_quantifier_field(F). % assumption: all or no field is disjoint
is_disjoint_quantifier_field(field(_FieldName,_Expr,type(_Type,_Arity),Options,_POS)) :-
    member(disj, Options).

translate_quantifier_field(field(FieldName,Expr,Type,_Options,POS), TField, NName) :-
    translate_pos(POS, BPOS),
    fun_or_pred_decl_special_cases(Expr, TFieldName, BPOS, TField, IsSingleton),
    !,
    (   IsSingleton == singleton ->
        FieldName = identifier(Name,_,_),
        prepend_identifier_of_module_import(Name, Type, NName),
        assert_singleton_set(NName)
    ;   NName = alloy2b_none
    ),
    translate_e_p(FieldName, TFieldName).


translate_fact(MAcc, fact(Expr,_Pos), NewMAcc) :-
    translate_e_p(Expr, TExpr),
    extend_machine_acc(properties, MAcc, TExpr, NewMAcc).

%translate_assertion(MAcc, fact(Expr,_Pos), NewMAcc) :-
%    translate_e_p(Expr, TExpr),
%    extend_machine_acc(assertions, MAcc, TExpr, NewMAcc).

% Signature facts: Identifiers may refer to the signature and are joined with 'this'.
% We then need a universal quantification using 'this'.
translate_signature_fact(TSignatureName, MAcc, Expr, NewMAcc) :-
    alloy_expr_contains_join(Expr),
    translate_e_p(Expr, TExpr), % we quantify in such a way that 'this' is a singleton_set_id; 'this' is always considered a singleton set id
    ThisID = identifier(none,'this'),
    TImplication = implication(none,subset(none,set_extension(none,[ThisID]),TSignatureName),TExpr),
    FORALL = forall(none,[ThisID],TImplication),
    extend_machine_acc(properties, MAcc, FORALL, NewMAcc).
translate_signature_fact(_TSignatureName, MAcc, Expr, NewMAcc) :-
    translate_e_p(Expr, TExpr),
    extend_machine_acc(properties, MAcc, TExpr, NewMAcc).

% TO DO: write generic AST traversal predicate
alloy_expr_contains_join(AndOr) :-
    AndOr =.. [Functor,ListOfAsts,_Pos],
    !,
    member(Functor, [and,or]),
    member(Ast, ListOfAsts),
    alloy_expr_contains_join(Ast),
    !.
alloy_expr_contains_join(identifier('this',_,_)).
% alloy_expr_contains_join_with_this(join(Arg1,Arg2,_Type,_Pos)) :-
% Arg1 = identifier('this',_,_) ; Arg2 = identifier('this',_,_).
alloy_expr_contains_join(Expr) :-
    Expr =.. [_Functor,Arg1,Arg2,_Type,_Pos],
    !,
    (   alloy_expr_contains_join(Arg1) ->
        true
    ;   alloy_expr_contains_join(Arg2)
    ).
alloy_expr_contains_join(Expr) :-
    Expr =.. [_Functor,Arg,_Type,_Pos],
    alloy_expr_contains_join(Arg).
alloy_expr_contains_join(Expr) :-
    Expr =.. [Functor,_Params,Fields,Body,_Type,_POS],
    member(Functor, [all,no,some,one,lone,comprehension]),
    (   alloy_expr_contains_join(Body) ->
        true
    ;   member(F, Fields),
        alloy_expr_contains_join(F) ->
        true
    ).

% translate fun NAME Decls { Body }  or pred Name Decls { Body }
translate_function(MAcc, FunctionOrPredicate, NewMAcc) :-
    FunctionOrPredicate =.. [Functor,Name,Params,Decls,Body,POS],
    alloy_to_b_operator(Functor, BFunctor),
    maplist(translate_e_p, Params, TParams),
    % identifier like 'this' are singleton sets, parameters have to be identifiers
    maplist(strip_singleton_set, TParams, TTParams),
    maplist(translate_function_field, Decls, TDecls),
    translate_e_p(Body, TBody),
    translate_pos(POS, BPOS),
    (   Functor == function
    ->
        (   TDecls == []
        ->
            TResult = TBody     % no need to create set comprehension
        ;                       % create {temp | TDecls & temp : TBody }
            append(TDecls, [member(none,identifier(none,'_fun_res_'),TBody)], BehaviorList),
            join_untyped_ast_nodes(conjunct, BehaviorList, Behavior),
            TResult = comprehension_set(BPOS,[identifier(none,'_fun_res_')],Behavior)
        ),
        UAst =.. [BFunctor,BPOS,Name,TTParams,TResult]
    ;   append(TDecls, [TBody], BehaviorList),
        join_untyped_ast_nodes(conjunct, BehaviorList, Behavior),
        UAst =.. [BFunctor,BPOS,Name,TTParams,Behavior]
    ),
    extend_machine_acc(definitions, MAcc, UAst, NewMAcc).

% extract scopes common to all commands
extract_common_scopes([Cmd|T], GlobalScopes, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, Conflicts) :-
    get_scopes(Cmd, GlobalScopes1, ExactScopes1, UpBound1, BitWidth1, MaxSeq1),
    !,
    extract_common_aux(T, GlobalScopes1, ExactScopes1, UpBound1, BitWidth1, MaxSeq1, [],
                       GlobalScopes, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, Conflicts).
extract_common_scopes(_, -1, [], [], none, maxseq(MaxSeq), []) :-
    default_max_seq(MaxSeq).
extract_common_aux([], GS, ES, UpS, BW, MS, C, GS, ES, UpS, BW, maxseq(MaxSeq), C) :-
    % no seq scope is set at all so we use the default scope as common scope
    MS == maxseq(-1),
    !,
    default_max_seq(MaxSeq).
extract_common_aux([], GS, ES, UpS, BW, MS, C, GS, ES, UpS, BW, none, C) :-
    MS == none.
extract_common_aux([], GS, ES, UpS, BW, MS, C, GS, ES, UpS, BW, MS, C).
extract_common_aux([Cmd|T], GlobalScope0, ExactScopes0, UpBoundScopes0, BitWidth0, MaxSeq0, Conflicts0,
                   GlobalScopes, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, Conflicts) :-
    get_scopes(Cmd, GlobalScope1, ExactScopes1, UpBoundScopes1, BitWidth1, MaxSeq1),
    (   GlobalScope0 = GlobalScope1 ->
        GlobalScope2 = GlobalScope0
    ;   GlobalScope2 = none
    ),
    (   BitWidth0 = BitWidth1 ->
        BitWidth2 = BitWidth1
    ;   BitWidth2 = none
    ),
    (   MaxSeq0 = MaxSeq1 ->
        MaxSeq2 = MaxSeq1
    ;   MaxSeq2 = none
    ),
    join_scopes(ExactScopes0, ExactScopes1, ExactScopes2, Conflicts1),
    ord_union(Conflicts0, Conflicts1, Conflicts2),
    join_scopes(UpBoundScopes0, UpBoundScopes1, UpBoundScopes2, Conflicts3),
    ord_union(Conflicts2, Conflicts3, Conflicts4),
    !,
    extract_common_aux(T, GlobalScope2, ExactScopes2, UpBoundScopes2, BitWidth2, MaxSeq2, Conflicts4,
                       GlobalScopes, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, Conflicts).

join_scopes(S1, S2, Inter, Conflicts) :-
    sort(S1, SS1),
    sort(S2, SS2),
    ord_intersection(SS1, SS2, Inter),
    ord_union(SS1, SS2, Union),
    ord_subtract(Union, Inter, Conflicts).

get_scopes(Command, GlobalScope, ExactScopes, UpBoundScopes, BitWidth, MaxSeq) :-
    Command =.. [_Functor,_Body,global_scope(GlobalScope),
                 exact_scopes(ExactScopes),upper_bound_scopes(UpBoundScopes),BitWidth,MaxSeq|_],
    !,
    debug_format(19, 'Command scopes: global=~w, exact=~w, upbound=~w,bitwidth=~w~n,maxseq=~w~n',
                 [GlobalScope,ExactScopes,UpBoundScopes,BitWidth,MaxSeq]).
get_scopes(Command, -1, [], [], none, none) :-
    add_error(alloy2b, 'Cannot process Alloy Command:', Command).

translate_command(MAcc, Command, NewMAcc) :-
    Command =.. [Functor,Body,GlobalScope,ExactScopes,UpBoundScopes,BitWidth,MaxSeq,index(CommandId),POS],
    % if a model does not use sequences or all commands have a common scope and we are not in proof mode, we translate all commands
    (   \+ current_command(_)
    ;   current_command(CurrentId),
        CurrentId == CommandId
    ),
    !,
    translate_command_aux(MAcc, Functor, Body, GlobalScope, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, POS, NewMAcc).
translate_command(MAcc, Command, MAcc) :-
    % only translate the current command if using sequences (either selected by the user or the first one by default)
    Command =.. [_,_,_,_,_,_,_,index(CommandId),_],
    current_command(CurrentId),
    CurrentId =\= CommandId,
    !.
translate_command(MAcc, Command, MAcc) :-
    add_error('Unknown Alloy command (be sure Alloy2B parser library is up-to-date):', Command).

translate_command_aux(MAcc, Functor, Body, GlobalScope, ExactScopes, UpBoundScopes, BitWidth, _MaxSeqT, POS, NewMAcc) :-
    debug_println(19, translating_command(Functor,GlobalScope,ExactScopes)),
    (   Functor = check
    ;   Functor = run
    ),
    % we need all signature names to define the global scope
    get_signature_names_from_machine_acc(MAcc, SignatureNames),
    translate_pos(POS, BPOS),
    get_preference(alloy_scopeless_translation, Scopeless),
    (   Scopeless == true,
        TScopesPred = truth(BPOS),
        !
    ;   debug_println(19, translate_scopes(SignatureNames)),
        maxseq(MaxSeq),
        translate_scopes(SignatureNames, GlobalScope, ExactScopes, UpBoundScopes, BitWidth, maxseq(MaxSeq), TScopesPred)
    ),
    TScopeBody = conjunct(BPOS,TScopesPred,TBody),
    get_preference(alloy_translation_for_proof, Proof),
    (   Proof == true
    ->  remove_negations_of_check_command(POS, Functor, Body, NBody),
        translate_e_p(NBody, TBody),
        extend_machine_acc(assertions, MAcc, TScopeBody, NewMAcc)
    ;   translate_e_p(Body, TBody),
        get_command_counter_atom(CommandCounter),
        atom_concat(Functor, CommandCounter, OperationName),
        Operation = operation(BPOS,identifier(BPOS,OperationName),[],[],precondition(BPOS,TScopeBody,skip(none))),
        extend_machine_acc(operations, MAcc, Operation, NewMAcc)
    ).

% We need to remove the negation of assertions stated in a check command when
% translating to the assertions to prove in AterlierB.
remove_negations_of_check_command(POS,check, Body, NBody) :-
    Body =.. [and,List,APos],
    remove_negations_of_check_command_list(POS, List, NList),
    NBody =.. [and,NList,APos].
remove_negations_of_check_command(_, run, Body, Body).

remove_negations_of_check_command_list(_, [], []).
remove_negations_of_check_command_list(pos(_, CommandCol), [AlloyTerm|T], NList) :-
    get_alloy_position(AlloyTerm, APOS),
    APOS = pos(_, Col),
    Col >= CommandCol,
    % filter those assertions stated in the command (do not remove negations of inlined assertions)
    !,
    remove_negation_alloy_term(AlloyTerm, NewAlloyTerm),
    remove_negations_of_check_command_list(pos(_, CommandCol), T, NT),
    NList = [NewAlloyTerm|NT].
remove_negations_of_check_command_list(POS, [AlloyTerm|T], [AlloyTerm|NT]) :-
    remove_negations_of_check_command_list(POS, T, NT).

remove_negation_alloy_term(not(AlloyTerm,_,_), AlloyTerm) :-
    !.
remove_negation_alloy_term(AlloyTerm, AlloyTerm).

% global scope, exact scopes and bitwidth
% We do not need to set the bitwidth since we have real integers in B.
% Note: If only one command is defined in the Alloy model we could set min and max int of ProB in the definitions.
translate_scopes(SignatureNames, global_scope(GlobalScope), exact_scopes(ExactScopes), upper_bound_scopes(UpBoundScopes), _BitWidth, maxseq(MaxSeq),
                 conjunct(none,SeqScopePred,conjunct(none,TScopes,TGlobalScopes))) :-
    % translate exact and upper bound scopes first
    translate_exact_and_upper_bound_scopes(0, ExactScopes, UpBoundScopes, EUSignatureNames, TScopes, OrdSigCounter1),
    % if present, define the global scope for the remaining signatures
    translate_global_scope(OrdSigCounter1, GlobalScope, SignatureNames, EUSignatureNames, TGlobalScopes, _),
    translate_seq_scopes(MaxSeq, SeqScopePred).

translate_seq_scopes(MaxSeq, ScopePred) :-
    findall((Seq,MaxSeq), is_seq(Seq), Seqs),
    maplist(generate_leq_scope_predicate, Seqs, ScopePreds),
    join_predicates(conjunct, ScopePreds, ScopePred).

% no exact scopes defined
translate_exact_and_upper_bound_scopes(OrdSigCounter, [], [], ExactSignatureNames, TGlobalScopes, OrdSigCounter) :-
    !,
    ExactSignatureNames = [],
    TGlobalScopes = truth(none).
translate_exact_and_upper_bound_scopes(OrdSigCounter, ExactScopes, UpBoundScopes, SignatureNames, TScopes, NewOrdSigCounter) :-
    translate_exact_or_upper_bound_scopes(upper, OrdSigCounter, UpBoundScopes, [], UpperBoundSignatureNames, [], TempTUpScopes, OrdSigCounter1),
    translate_exact_or_upper_bound_scopes(exact, OrdSigCounter1, ExactScopes, [], ExactSignatureNames, [], TempTExactScopes, NewOrdSigCounter),
    append(TempTUpScopes, TempTExactScopes, TempTScopes),
    append(ExactSignatureNames, UpperBoundSignatureNames, SignatureNames),
    join_predicates(conjunct, TempTScopes, TScopes).

translate_exact_or_upper_bound_scopes(_, OrdSigCounter, [], NamesAcc, NamesAcc, TScopeAcc, TScopeAcc, OrdSigCounter).
translate_exact_or_upper_bound_scopes(ScopeType, OrdSigCounter, [(SignatureName,Scope)|T], NameAcc, ExactSignatureNames, TScopeAcc, TempTExactScopes, NewOrdSigCounter) :-
    is_ordered_signature(SignatureName),
    !,
    define_ordered_signature_as_integer_set(OrdSigCounter, SignatureName, Scope,
                                            TScope, TNewOrdSigCounter),
    translate_exact_or_upper_bound_scopes(ScopeType, TNewOrdSigCounter, T, [SignatureName|NameAcc], ExactSignatureNames, [TScope|TScopeAcc], TempTExactScopes, NewOrdSigCounter).
translate_exact_or_upper_bound_scopes(exact, OrdSigCounter, [(SignatureName,Scope)|T], NameAcc, ExactSignatureNames, TScopeAcc, TempTExactScopes, NewOrdSigCounter) :-
    generate_exact_scope_predicate((SignatureName,Scope), TScope),
    translate_exact_or_upper_bound_scopes(exact, OrdSigCounter, T, [SignatureName|NameAcc], ExactSignatureNames, [TScope|TScopeAcc], TempTExactScopes, NewOrdSigCounter).
translate_exact_or_upper_bound_scopes(upper, OrdSigCounter, [(SignatureName,Scope)|T], NameAcc, ExactSignatureNames, TScopeAcc, TempTExactScopes, NewOrdSigCounter) :-
    generate_leq_scope_predicate((SignatureName,Scope), TScope),
    translate_exact_or_upper_bound_scopes(upper, OrdSigCounter, T, [SignatureName|NameAcc], ExactSignatureNames, [TScope|TScopeAcc], TempTExactScopes, NewOrdSigCounter).

generate_exact_scope_predicate((SignatureName,Scope), TScopePred) :-
    translate_e_p(SignatureName, TSignatureName),
    % debug_format(19,'Exact scope for SignatureName=~w with scope ~w~n',[SignatureName,Scope]),
    TScopePred = equal(none,card(none,TSignatureName),integer(none,Scope)).
generate_leq_scope_predicate((SignatureName,Scope), TScopePred) :-
    translate_e_p(SignatureName, TSignatureName),
    TScopePred = less_equal(none,card(none,TSignatureName),integer(none,Scope)).

define_ordered_signatures_as_integer_set(OrdSigCounter, [], _, [], OrdSigCounter).
define_ordered_signatures_as_integer_set(OrdSigCounter, [TSignatureName|T], Scope, [TScope|TScopes], NewOrdSigCounter) :-
    define_ordered_signature_as_integer_set(OrdSigCounter, TSignatureName, Scope, TScope, TNewOrdSigCounter),
    define_ordered_signatures_as_integer_set(TNewOrdSigCounter, T, Scope, TScopes, NewOrdSigCounter).

define_ordered_signature_as_integer_set(OrdSigCounter, TSignatureName, Scope, TScope, NewOrdSigCounter) :-
    NewOrdSigCounter is OrdSigCounter+Scope,
    NewOrdSigCounter1 is NewOrdSigCounter-1,
    TScope = equal(none,identifier(none,TSignatureName),interval(none,integer(none,OrdSigCounter),integer(none,NewOrdSigCounter1))).

translate_global_scope(OrdSigCounter, -1, _, _, TGlobalScopes, OrdSigCounter) :-
    !,
    TGlobalScopes = truth(none).                          % no global scope defined
translate_global_scope(OrdSigCounter, GlobalScope, SignatureNames, ExactSignatureNames, TGlobalScopes, NewOrdSigCounter) :-
    split_ordered_unordered_signatures(SignatureNames, ExactSignatureNames, OrderedSignatureNames, RestSignatureNames),
    translate_global_scope_aux(RestSignatureNames, GlobalScope, [], TempTGlobalScopes),
    join_predicates(conjunct, TempTGlobalScopes, TTGlobalScopes),
    define_ordered_signatures_as_integer_set(OrdSigCounter, OrderedSignatureNames, GlobalScope, TOrderedScopes, NewOrdSigCounter),
    join_predicates(conjunct, [TTGlobalScopes|TOrderedScopes], TGlobalScopes).

translate_global_scope_aux([], _, Acc, Acc).
translate_global_scope_aux([SignatureName|T], GlobalScope, Acc, TGlobalScopes) :-
    translate_e_p(SignatureName, TSignatureName),
    TScope = less_equal(none,card(none,TSignatureName),integer(none,GlobalScope)),
    translate_global_scope_aux(T, GlobalScope, [TScope|Acc], TGlobalScopes).
translate_global_scope_aux([_|T], GlobalScope, Acc, TGlobalScopes) :-
    translate_global_scope_aux(T, GlobalScope, Acc, TGlobalScopes).

split_ordered_unordered_signatures(SNames, ExactSNames, OSNames, Rest) :-
    split_ordered_unordered_signatures(SNames, [], [], TOSNames, TRest),
    % remove exact signatures which are defined separately
    subtract(TOSNames, ExactSNames, OSNames),
    subtract(TRest, ExactSNames, Rest).

split_ordered_unordered_signatures([], OAcc, RAcc, OAcc, RAcc).
split_ordered_unordered_signatures([SName|T], OAcc, RAcc, OSNames, Rest) :-
    is_ordered_signature(SName),
    \+ member(SName, OAcc),
    !,
    split_ordered_unordered_signatures(T, [SName|OAcc], RAcc, OSNames, Rest).
split_ordered_unordered_signatures([SName|T], OAcc, RAcc, OSNames, Rest) :-
    \+ is_ordered_signature(SName),
    \+ member(SName, RAcc),
    !,
    split_ordered_unordered_signatures(T, OAcc, [SName|RAcc], OSNames, Rest).
split_ordered_unordered_signatures([_|T], OAcc, RAcc, OSNames, Rest) :-
    split_ordered_unordered_signatures(T, OAcc, RAcc, OSNames, Rest).

% signature in (subset)
define_sig_as_set_or_constant(MAcc, Name, Options, POS, NewMAcc) :-
    memberchk(subset(Parents), Options),
    !,
    % maplist(assert_extending_signature(Name),Parents) , % subset signatures do not need to be distinct
    translate_pos(POS, BPOS),
    define_sig_as_set_or_constant_aux(constants, MAcc, Name, Options, BPOS, MAcc1),
    % TO DO: consider several parents -> we need the universe type
    Parents = [Parent|_],
    translate_e_p(Name, TName),
    translate_e_p(Parent, TParent),
    TNode = subset(BPOS,TName,TParent),
    extend_machine_acc(properties, MAcc1, TNode, NewMAcc).
% signature extends
define_sig_as_set_or_constant(MAcc, Name, Options, POS, NewMAcc) :-
    member(subsig(Parent), Options),
    !,
    (   member(one, Options) ->
        One = one
    ;   One = set
    ),
    assert_extending_signature(Name, Parent, One),
    translate_pos(POS, BPOS),
    define_sig_as_set_or_constant_aux(constants, MAcc, Name, Options, BPOS, MAcc1),
    translate_e_p(Name, TName),
    translate_e_p(Parent, TParent),
    TNode = subset(BPOS,TName,TParent),
    extend_machine_acc(properties, MAcc1, TNode, NewMAcc).
% default signature, but we introduced an artifical parent type, i.e., a deferred set in B
define_sig_as_set_or_constant(MAcc, Name, Options, POS, NewMAcc) :-
    translate_pos(POS, BPOS),
    artificial_parent_type(Name, ParentId),
    !,
    translate_e_p(Name, TName),
    TNode = subset(BPOS,TName,ParentId),
    extend_machine_acc(properties, MAcc, TNode, MAcc1),
    define_sig_as_set_or_constant_aux(constants, MAcc1, Name, Options, BPOS, NewMAcc).
% default signature
define_sig_as_set_or_constant(MAcc, Name, Options, POS, NewMAcc) :-
    translate_pos(POS, BPOS),
    define_sig_as_set_or_constant_aux(sets, MAcc, Name, Options, BPOS, NewMAcc).

define_sig_as_set_or_constant_aux(sets, MAcc, Name, Options, BPOS, NewMAcc) :-
    member(one, Options),
    !,
    atom_concat('_', Name, EnumElement),
    % TO DO: couldn't figure out untyped ast of enumerated set, define enumerated set S = {_S} instead of constant etc.
    extend_machine_acc(sets, MAcc, deferred_set(BPOS,Name), MAcc1),
    extend_machine_acc(constants, MAcc1, identifier(BPOS,EnumElement), MAcc2),
    extend_machine_acc(properties, MAcc2, equal(BPOS,identifier(BPOS,Name),set_extension(BPOS,[identifier(BPOS,EnumElement)])), NewMAcc).
define_sig_as_set_or_constant_aux(sets, MAcc, Name, Options, BPOS, NewMAcc) :-
    member(some, Options),
    !,
    extend_machine_acc(sets, MAcc, deferred_set(BPOS,Name), MAcc1),
    extend_machine_acc(properties, MAcc1, greater_equal(BPOS,card(BPOS,identifier(BPOS,Name)),integer(BPOS,1)), NewMAcc).
define_sig_as_set_or_constant_aux(sets, MAcc, Name, Options, BPOS, NewMAcc) :-
    member(lone, Options),
    !,
    extend_machine_acc(sets, MAcc, deferred_set(BPOS,Name), MAcc1),
    extend_machine_acc(properties, MAcc1, less_equal(BPOS,card(BPOS,identifier(BPOS,Name)),integer(BPOS,1)), NewMAcc).
define_sig_as_set_or_constant_aux(sets, MAcc, Name, _Options, BPOS, NewMAcc) :-
    extend_machine_acc(sets, MAcc, deferred_set(BPOS,Name), NewMAcc).
define_sig_as_set_or_constant_aux(constants, MAcc, Name, _Options, BPOS, NewMAcc) :-
    extend_machine_acc(constants, MAcc, identifier(BPOS,Name), NewMAcc).

% --------
% Semantic translation rules: translate expression or predicate
% --------
translate_e_p(A, TA) :-         %% functor(A,F,N),format('~n TRANSLATE: ~w/~w :  ~w ~n',[F,N,A]),
    translate_quantifier_e(A, TA),
    !.
translate_e_p(A, TA) :-
    translate_cst_e_p(A, TA),
    !.
translate_e_p(A, TA) :-
    translate_unary_e_p(A, TA),
    !.
translate_e_p(A, TA) :-
    translate_binary_e_p(A, TA),
    !.
translate_e_p(A, TA) :-
    translate_if_then_else(A, TA).
translate_e_p(A, _) :-
    add_error(load_alloy_model, 'Translation failed for:', A),
    fail.

% Translation in the context where we expect an integer
% This is the E_{int} translation function in the article
% The following rule is not necessary: untyped_b_ast_to_integer will peel off set extensions
%translate_e_p_integer(identifier(ID,_Type,POS), identifier(BPOS,ID)) :- %Type=integer, ?
%    is_singleton_set(ID),!,
%    translate_pos(POS, BPOS).
translate_e_p_integer(integer(A,POS), integer(BPOS,A)) :-
    !,
    translate_pos(POS, BPOS).
translate_e_p_integer(A, BAstInt) :-
    translate_e_p(A, BAst),
    untyped_b_ast_to_integer(BAst, BAstInt).

untyped_b_ast_to_integer(integer(BPOS,A), integer(BPOS,A)) :-
    !.
untyped_b_ast_to_integer(set_extension(_,[Single]),  Single) :-
    !.
% Alloy sums the elements of sets of integers so we can use SIGMA which also strips singleton sets like the MU operator.
% If preference alloy_strict_integers is set, we only accept singleton sets as integers and throw a WD error otherwise.
untyped_b_ast_to_integer(BAst, BAstInt) :-
    get_preference(alloy_strict_integers, Strict),
    Strict == false,
    !,
    get_b_position(BAst, BPOS),
    check_position(BPOS,untyped_b_ast_to_integer),
    BAstInt = general_sum(BPOS,[identifier(BPOS,z)],member(BPOS,identifier(BPOS,z),BAst),identifier(BPOS,z)).
untyped_b_ast_to_integer(BAst, BAstInt) :-
    get_b_position(BAst, BPOS),
    check_position(BPOS,untyped_b_ast_to_integer),
    BAstInt = mu(BPOS,BAst).



% Translation in the context where we expect a singleton set.
% Otherwise, a well-definedness error is thrown.
% This is the E_{one} translation function in the article.
% Note that the arguments are already translated (using \semE{.} in the article).
translate_e_p_one(set_extension(_,[One]), Res) :-
    !,
    Res = One.
translate_e_p_one(BAst, mu(none,BAst)).

% construct_choose(TSetA,external_function_call(none,'CHOOSE',[TSetA],string(none,'a member of X'),TypeParams,Decl)) :-
% % format of raw AST of CHOOSE copied from btype2
% TypeParams = rewrite_protected([identifier(none,T)]),
% Decl = rewrite_protected(total_function(none,pow_subset(none,identifier(none,T)),identifier(none,T))).
translate_if_then_else(if_then_else(ConditionPred,TruthExpr,FalsityExpr,_Type,POS), TIfThenElse) :-
    translate_e_p(ConditionPred, TConditionPred),
    translate_e_p(TruthExpr, TTruthExpr),
    translate_e_p(FalsityExpr, TFalsityExpr),
    translate_pos(POS, BPOS),
    get_type_and_arity_from_alloy_term(TruthExpr, Type, _),
    (   (   Type = [untyped]
        ;   is_primitive_type(Type, _)
        )
    ->
        TIfThenElse = conjunct(BPOS,
                               implication(BPOS,TConditionPred,TTruthExpr),
                               implication(BPOS,negation(BPOS,TConditionPred),TFalsityExpr))
    ;   TIfThenElse = if_then_else(BPOS,TConditionPred,TTruthExpr,TFalsityExpr)
    ).

translate_quantifier_e(Quantifier, TQuantifier) :-
    Quantifier =.. [Functor,Params,Fields,Body,_Type,POS],
    member(Functor, [all,no,some,one,lone,comprehension]),
    maplist(translate_e_p, Params, TParams),
    maplist(strip_singleton_set, TParams, TTParams), % strip singleton sets for parameter definition in the untyped B ast (only identifier/2 nodes allowed in there)
    translate_pos(POS, BPOS),
    maplist(translate_quantifier_field, Fields, TFieldsList, SingletonSetNames),
    (   is_disjoint_quantifier(Fields)
    ->
        maplist(translate_e_p, SingletonSetNames, TSets),
        AllDiff = equal(BPOS,card(BPOS,general_union(BPOS,set_extension(BPOS,TSets))),integer(BPOS,Len)),
        length(Fields, Len),
        join_untyped_ast_nodes(conjunct, [AllDiff|TFieldsList], TFields)
    ;   join_untyped_ast_nodes(conjunct, TFieldsList, TFields)
    ),
    translate_e_p(Body, TBody),
    translate_quantifier_e_aux(BPOS, Functor, TTParams, TFields, TBody, TQuantifier),
    % TO DO: retract singleton sets introduced in quantifiers on exit
    maplist(retract_singleton_set(BPOS), SingletonSetNames).

translate_quantifier_e_aux(Pos, all, TParams, TFields, TBody, forall(Pos,TParams,implication(Pos,TFields,TBody))).
translate_quantifier_e_aux(Pos, no, TParams, TFields, TBody, negation(Pos,exists(Pos,TParams,conjunct(Pos,TFields,TBody)))).
translate_quantifier_e_aux(Pos, some, TParams, TFields, TBody, exists(Pos,TParams,conjunct(Pos,TFields,TBody))).
translate_quantifier_e_aux(Pos, one, TParams, TFields, TBody, equal(Pos,card(Pos,comprehension_set(Pos,TParams,conjunct(Pos,TFields,TBody))),integer(Pos,1))).
translate_quantifier_e_aux(Pos, lone, TParams, TFields, TBody, less_equal(Pos,card(Pos,comprehension_set(Pos,TParams,conjunct(Pos,TFields,TBody))),integer(Pos,1))).
translate_quantifier_e_aux(Pos, comprehension, TParams, TFields, TBody, comprehension_set(Pos,TParams,conjunct(Pos,TFields,TBody))).

% Translate Constants
translate_cst_e_p(iden(POS), event_b_identity(BPOS)) :- % has no Span Info !
    translate_pos(POS, BPOS).
translate_cst_e_p(identifier(Cst,_,POS), BConstruct) :-
    alloy_constant_to_b(Cst, BPOS, BConstruct),
    !,
    translate_pos(POS, BPOS).
translate_cst_e_p(none(POS), TNone) :-
    !,
    translate_pos(POS, BPOS),
    TNone = empty_set(BPOS).
translate_cst_e_p(this, _) :-
    print('this not yet translated'),
    nl,
    fail.


alloy_constant_to_b(none, BPOS, empty_set(BPOS)).
alloy_constant_to_b('boolean''False', BPOS, boolean_false(BPOS)).
alloy_constant_to_b('boolean''True', BPOS, boolean_true(BPOS)).

translate_unary_e_p(cast2int(Expr,_Type,POS), Result) :-
    !,
    translate_pos(POS, BPOS),
    translate_e_p_integer(Expr, TExpr),
    Result = set_extension(BPOS, [TExpr]).
translate_unary_e_p(cast2sigint(Expr,_Type,_Pos), Result) :-
    % TO DO: do we need this? how is this written in Alloy?
    !,
    translate_e_p(Expr, TExpr),
    Result = TExpr.
translate_unary_e_p(string(StringCodes,POS), Result) :-
    atom_codes(String, StringCodes),
    !,
    translate_pos(POS, BPOS),
    Result = set_extension(BPOS,[string(BPOS,String)]).
translate_unary_e_p(seq(ID), Result) :-
    translate_unary_e_p(ID, TID),
    !,
    Result = sequence_extension(none,TID).
translate_unary_e_p(Int, Result) :-
    integer(Int),
    !,
    Result = integer(none,Int).
translate_unary_e_p(Term, Result) :-
    (   Term = identifier(ID,type(TypeList,_),POS),
        TypeList = [SigName|_]
    ;   Term = ID),
    atom(ID),
    sig_field_id(ID, SigName),
    !,
    atom_concat(ID, '_', NID),
    atom_concat(NID, SigName, NNID),
    translate_pos(POS, BPOS),
    (is_singleton_set_id(NNID)
    ->  Result = set_extension(BPOS,[identifier(BPOS,NNID)])
    ;   Result = identifier(BPOS,NNID)).
translate_unary_e_p(identifier(ID,Type,POS), Result) :-
    prepend_identifier_of_module_import(ID, Type, NID),
    % singleton enumerated set like S = {_S}
    enumerated_set(NID),
    is_singleton_set_id(NID),
    !,
    atom_concat('_', NID, NNID),
    translate_pos(POS, BPOS),
    Result = set_extension(BPOS,[identifier(BPOS,NNID)]).
translate_unary_e_p(identifier(ID,Type,POS), Result) :-
    prepend_identifier_of_module_import(ID, Type, NID),
    is_singleton_set_id(NID),
    !,
    translate_pos(POS, BPOS),
    Result = set_extension(BPOS,[identifier(BPOS,NID)]).
translate_unary_e_p(ID, Result) :-
    atom(ID),
    prepend_module_name(ID,NID),
    enumerated_set(NID),
    is_singleton_set_id(NID),
    !,
    atom_concat('_', NID, NNID),
    Result = set_extension(none,[identifier(none,NNID)]).
translate_unary_e_p(ID, Result) :-
    atom(ID),
    prepend_module_name(ID,NID),
    is_singleton_set_id(NID),
    !,                          % DOES THIS EVER TRIGGER ?
    Result = set_extension(none,[identifier(none,NID)]).
translate_unary_e_p(integer(A,POS), Result) :-
    !,
    translate_pos(POS, BPOS),
    Result = set_extension(BPOS,[integer(BPOS,A)]).
translate_unary_e_p(identifier('Int',_Type,POS), Result) :-
    !,
    translate_pos(POS, BPOS),
    integer_type(IntOrInteger),
    Result = integer_set(BPOS,IntOrInteger).
translate_unary_e_p(identifier('nat''Natural',_Type,POS), Result) :-
    !,
    translate_pos(POS, BPOS),
    natural_type(NatOrNatural),
    Result = integer_set(BPOS,NatOrNatural).
translate_unary_e_p(identifier('nat''One',_Type,POS), Result) :-
    !,
    translate_pos(POS, BPOS),
    Result = set_extension(BPOS,[integer(BPOS,1)]).
translate_unary_e_p(identifier('nat''Zero',_Type,POS), Result) :-
    !,
    translate_pos(POS, BPOS),
    Result = set_extension(BPOS,[integer(BPOS,0)]).
translate_unary_e_p(identifier('String',_,_), Result) :-
    !,
    Result = string_set(none).
translate_unary_e_p(identifier(ID,Type,POS), Result) :-
    !,
    prepend_identifier_of_module_import(ID, Type, NID),
    translate_pos(POS, BPOS),
    Result = identifier(BPOS,NID).
translate_unary_e_p('Int', Result) :-
    !,
    integer_type(IntOrInteger),
    Result = integer_set(none,IntOrInteger).
translate_unary_e_p('nat''Natural', Result) :-
    !,
    natural_type(NatOrNatural),
    Result = integer_set(none,NatOrNatural).
translate_unary_e_p('nat''One', Result) :-
    !,
    Result = set_extension(none,[integer(none,1)]).
translate_unary_e_p('nat''Zero', Result) :-
    !,
    Result = set_extension(none,[integer(none,0)]).
translate_unary_e_p(boolean(true,POS), Result) :-
    !,                          % was boolean_true, but Alloy has no booleans
    translate_pos(POS, BPOS),
    Result = truth(BPOS).
translate_unary_e_p(boolean(false,POS), Result) :-
    !,                          % was boolean_false
    translate_pos(POS, BPOS),
    Result = falsity(BPOS).
translate_unary_e_p(Type, Result) :-
    % lhs of sequence is integer, Alloy Types are like [seqInt,A] for set(couple(INTEGER,A))
    atom(Type),
    atom_concat(seq, _, Type),
    !,
    integer_type(IntOrInteger),
    Result = integer_set(none,IntOrInteger).
translate_unary_e_p(ID, Result) :-
    atom(ID),
    prepend_module_name(ID,NID),
    !,
    Result = identifier(none,NID).
translate_unary_e_p(UnaryP, TUnaryP) :-
    % and/or defines a list of ast nodes
    UnaryP =.. [Op,ArgList|_],
    memberchk(Op, [and,or]),    % translated to conjunct/disjunct
    is_list(ArgList),
    !,
    maplist(translate_e_p, ArgList, TArgList),
    join_predicates(Op, TArgList, TUnaryP).
translate_unary_e_p(UnaryP, TUnaryP) :-
    UnaryP =.. [Op,Arg,_Type,POS],
    member(Op, [no,one,some,lone,exactlyof,oneof,someof,loneof]),
    !,
    translate_e_p(Arg, TArg),
    translate_pos(POS, BPOS),
    translate_quantified_e(BPOS, Op, TArg, TUnaryP).
translate_unary_e_p(card(Arg,_Type,POS), set_extension(BPOS,[card(BPOS,TArg)])) :-
    !,
    translate_e_p(Arg, TArg),
    translate_pos(POS, BPOS).
translate_unary_e_p(UnaryP, TUnaryP) :-
    UnaryP =.. [Op,Arg,_Type,POS],
    translate_e_p(Arg, TArg),
    alloy_to_b_unary_operator(Op, BOp),
    translate_pos(POS, BPOS),
    TUnaryP =.. [BOp,BPOS,TArg].

% support for Alloy modules to maintain unique identifiers
prepend_module_name(ID,NID) :-
    atom(ID),
    \+ has_module_prefix(ID),
    current_module_in_scope(ModuleName),
    !,
    atom_concat(ModuleName, '''', ModuleNameP),
    atom_concat(ModuleNameP, ID, NID).
prepend_module_name(ID,ID).

has_module_prefix(ID) :-
    atom(ID),
    module_in_scope_but_root(ModuleName, Alias),
    (   atom_concat(ModuleName, _ , ID)
    ;   Alias \== 'alloy2b_none',
        atom_concat(Alias, _ , ID)
    ),
    !.

prepend_identifier_of_module_import(ID, Type, NID) :-
    % field names of signatures from imported modules are not unique,
    % their type (the signature name) is unique though
    atom(ID),
    Type = type([Inner|_],_),
    module_in_scope_but_root(ModuleName, Alias),
    (   Alias \== 'alloy2b_none'
    ->  Name = Alias
    ;   Name = ModuleName
    ),
    atom_concat(Name, _, Inner),
    atom_concat(Name, '''', NameP),
    \+atom_concat(NameP, _, ID),
    !,
    atom_concat(NameP, ID, NID).
prepend_identifier_of_module_import(ID, _, ID).

%% FUNCTIONS defined by func in util:
% alloy_to_b_natural_utility_function(Name, BPOS, Params, Result).
% Special cases with no direct counterpart in B.
alloy_to_b_natural_utility_function('''ord''first', BPOS, [], set_extension(BPOS,[integer(BPOS,0)])).
alloy_to_b_natural_utility_function('''ord''last', BPOS, [], set_extension(BPOS,[max_int(BPOS)])).
alloy_to_b_natural_utility_function('''ord''next', BPOS, [P1Int], set_extension(BPOS,[add(BPOS,P1Int,integer(BPOS,1))])).
alloy_to_b_natural_utility_function('''ord''prev', BPOS, [P1Int], set_extension(BPOS,[minus(BPOS,P1Int,integer(BPOS,1))])).
alloy_to_b_natural_utility_function('''One', BPOS, [], set_extension(BPOS,[integer(BPOS,1)])).
alloy_to_b_natural_utility_function('''Zero', BPOS, [], set_extension(BPOS,[integer(BPOS,0)])).
alloy_to_b_natural_utility_function('''min', BPOS, [], set_extension(BPOS,[max(BPOS,set_extension(BPOS,[integer(BPOS,0),min_int(BPOS)]))])).
alloy_to_b_natural_utility_function('''inc', BPOS, [P1Int], set_extension(BPOS,[add(BPOS,P1Int,integer(BPOS,1))])).
alloy_to_b_natural_utility_function('''dec', BPOS, [P1Int], set_extension(BPOS,[minus(BPOS,P1Int,integer(BPOS,1))])).

% alloy_to_b_integer_utility_function(Name, BPOS, Params, Result).
% Special cases with no direct counterpart in B.
alloy_to_b_integer_utility_function('''signum', BPOS, [P1Int], Result) :-
    !,
    % IF int(p) < 0 THEN -1 ELSE IF int(p) > 0 THEN 1 ELSE 0 END END
    Result = if_then_else(BPOS,less(BPOS,P1Int,integer(BPOS,0)),set_extension(BPOS,[integer(BPOS,-1)]),if_then_else(BPOS,greater(BPOS,P1Int,integer(BPOS,0)),set_extension(BPOS,[integer(BPOS,1)]),set_extension(BPOS,[integer(BPOS,0)]))).
alloy_to_b_integer_utility_function('''zero', BPOS, [P1Int], Result) :-
    !,
    % int(p) = 0
    Result = equal(BPOS,P1Int,integer(BPOS,0)).
alloy_to_b_integer_utility_function(Name, BPOS, [P1Int], Result) :-
    (   Name == '''pos', BOp = greater
    ;   Name == '''neg', BOp = less),
    !,
    translate_e_p_integer(P1Int, P1Int),
    % int(p) > 0 or int(p) < 0
    Result =.. [BOp,BPOS,P1Int,integer(BPOS,0)].
alloy_to_b_integer_utility_function(Name, BPOS, [P1Int], Result) :-
    (   Name == '''nonpos', BOp = less_equal
    ;   Name == '''nonneg', BOp = greater_equal),
    !,
    translate_e_p_integer(P1Int, P1Int),
    % int(p) <= 0 or int(p) >= 0
    Result =.. [BOp,BPOS,P1Int,integer(BPOS,0)].
alloy_to_b_integer_utility_function(Name, BPOS, [P1Int,P2Int], Result) :-
    (   Name == '''smaller', BOp = min
    ;   Name == '''larger', BOp = max),
    !,
    % {min({int(p1),int(p2)})} or {max({int(p1),int(p2)})}
    Inner =.. [BOp,BPOS,set_extension(BPOS,[P1Int,P2Int])],
    Result = set_extension(BPOS,[Inner]).
alloy_to_b_integer_utility_function(Name, BPOS, [P1Int], Result) :-
    (   Name == '''nexts', BOp = greater
    ;   Name == '''prevs', BOp = less),
    !,
    % {x | x : INTEGER & x > int(p)} or {x | x : INTEGER & x < int(p)}
    Inner =.. [BOp,BPOS,identifier(BPOS,x),P1Int],
    integer_type(IntOrInteger),
    check_position(BPOS,alloy_to_b_integer_utility_function(Name)),
    Result = comprehension_set(BPOS,[identifier(BPOS,x)],
                 conjunct(BPOS,member(BPOS,identifier(BPOS,x),integer_set(none,IntOrInteger)),Inner)).

%% alloy_to_b_integer_natural_utility_func(Name,Arity,BOperator,ArgType,ReturnType).
% Special cases with direct counterparts in B.
alloy_to_b_integer_natural_utility_func(FunName, Arity, BOp, ArgType, ReturnType) :-
    is_util_integer_natural_func(FunName, _, AlloyOp),
    (   alloy_to_b_integer_func(AlloyOp, Arity, BOp, ArgType, ReturnType)
    ;   alloy_to_b_integer_operator_func(AlloyOp, Arity, BOp, ArgType, ReturnType)
    ;   alloy_to_b_integer_to_integer_func(AlloyOp, Arity, BOp, ArgType, ReturnType)
    ;   alloy_to_b_operator_func(AlloyOp, Arity, BOp, ArgType, ReturnType)
    ),!.

is_util_integer_natural_func(FunName, UtilType, AlloyOp) :-
    atom(FunName),
    module_in_scope_but_root(ModuleName, Alias),
    atom(ModuleName),
    atom(Alias),
    (   atom_concat(ModuleName, AlloyOp, FunName)
    ;   atom_concat(Alias, AlloyOp, FunName)
    ),
    (   ModuleName == 'util''integer',
        UtilType = integer
    ;   ModuleName == 'util''natural',
        UtilType = natural
    ).

% functions with integer arguments and result
alloy_to_b_integer_to_integer_func('''minus', 2, minus, integer, integer).
alloy_to_b_integer_to_integer_func('''plus', 2, add, integer, integer).
alloy_to_b_integer_to_integer_func('''div', 2, div, integer, integer).
alloy_to_b_integer_to_integer_func('''mul', 2, multiplication, integer, integer).
alloy_to_b_integer_to_integer_func('''rem', 2, modulo, integer, integer).
% alloy_to_b_integer_to_integer_func('''sum',sigma, integer, integer). % TO DO: translate sum
alloy_to_b_integer_to_integer_func('''add', 2, add, integer, integer).
alloy_to_b_integer_to_integer_func('''sub', 2, minus, integer, integer).
% Unary:
alloy_to_b_integer_to_integer_func('''negate', 1, unary_minus, integer, integer).
alloy_to_b_integer_to_integer_func('''next', 1, successor, integer, integer). % not sure this is called; seems to call arity 0 and use join
alloy_to_b_integer_to_integer_func('''prev', 1, predecessor, integer, integer).

% functions with integer result but set arguments
alloy_to_b_integer_func('''min', 1, min, set, integer).
alloy_to_b_integer_func('''max', 1, max, set, integer).
alloy_to_b_integer_func('''min', 0, min_int, set, integer). % TO DO: investigate bit-width related translation
alloy_to_b_integer_func('''max', 0, max_int, set, integer).
% Note: for run Arithmetic for 6 Int the Alloy Evaluator produces {-32} for min and {+31} for max
% predicate functions with integer arguments
alloy_to_b_operator_func('''next', 0, successor, set, set).
alloy_to_b_operator_func('''prev', 0, predecessor, set, set).

% predicate functions with integer arguments
alloy_to_b_integer_operator_func('''gt', 2, greater, integer, set).
alloy_to_b_integer_operator_func('''lt', 2, less, integer, set).
alloy_to_b_integer_operator_func('''gte', 2, greater_equal, integer, set).
alloy_to_b_integer_operator_func('''lte', 2, less_equal, integer, set).
alloy_to_b_integer_operator_func('''eq', 2, equal, integer, set).

% Not[b] =      bool(b = FALSE)
alloy_to_b_boolean_utility_function('boolean''Not', [TP1], BPOS, Result) :-
    !,
    Result = convert_bool(BPOS,equal(BPOS,TP1,boolean_false(BPOS))).
alloy_to_b_boolean_utility_function('boolean''isTrue', [TP1], BPOS, Result) :-
    !,
    Result = equal(BPOS,TP1,boolean_true(BPOS)).
alloy_to_b_boolean_utility_function('boolean''isFalse', [TP1], BPOS, Result) :-
    !,
    Result = equal(BPOS,TP1,boolean_false(BPOS)).
% And[a,b] =    bool(a = TRUE & b = TRUE)
alloy_to_b_boolean_utility_function('boolean''And', [TP1,TP2], BPOS, Result) :-
    !,
    Result = convert_bool(BPOS,conjunct(BPOS,equal(BPOS,TP1,boolean_true(BPOS)),equal(BPOS,TP2,boolean_true(BPOS)))).
% Or[a,b] =     bool(a = TRUE or b = TRUE)
alloy_to_b_boolean_utility_function('boolean''Or', [TP1,TP2], BPOS, Result) :-
    !,
    Result = convert_bool(BPOS,disjunct(BPOS,equal(BPOS,TP1,boolean_true(BPOS)),equal(BPOS,TP2,boolean_true(BPOS)))).
% Nand[a,b] =   bool(a = FALSE or b = FALSE)
alloy_to_b_boolean_utility_function('boolean''Nand', [TP1,TP2], BPOS, Result) :-
    !,
    Result = convert_bool(BPOS,disjunct(BPOS,equal(BPOS,TP1,boolean_false(BPOS)),equal(BPOS,TP2,boolean_false(BPOS)))).
% Nor[a,b] =    bool(a = FALSE & b = FALSE)
alloy_to_b_boolean_utility_function('boolean''Nor', [TP1,TP2], BPOS, Result) :-
    !,
    Result = convert_bool(BPOS,conjunct(BPOS,equal(BPOS,TP1,boolean_false(BPOS)),equal(BPOS,TP2,boolean_false(BPOS)))).
% Xor[a,b] =    bool((a = FALSE & b = TRUE) or (a = TRUE & b = FALSE))
alloy_to_b_boolean_utility_function('boolean''Xor', [TP1,TP2], BPOS, Result) :-
    !,
    LHS = conjunct(BPOS,equal(BPOS,TP1,boolean_false(BPOS)),equal(BPOS,TP2,boolean_true(BPOS))),
    RHS = conjunct(BPOS,equal(BPOS,TP1,boolean_true(BPOS)),equal(BPOS,TP2,boolean_false(BPOS))),
    Result = convert_bool(BPOS,disjunct(BPOS,LHS,RHS)).
alloy_to_b_boolean_utility_function(FunName, _, BPOS, truth(BPOS)) :-
    atom_concat('boolean', _, FunName),
    !,
    add_error(alloy_to_b_boolean_utility_function, 'Function from util/boolean currently not supported', [], BPOS).

% dom[r] = dom(r)
alloy_to_b_relational_utility_function('relation''dom', [TRelation], BPOS, TCall) :-
    !,
    TCall = dom(BPOS,TRelation).
% ran[r] = ran(r)
alloy_to_b_relational_utility_function('relation''ran', [TRelation], BPOS, TCall) :-
    !,
    TCall = range(BPOS,TRelation).
% total[r,d] = !x.(x:d => r[{x}] /= {}})
alloy_to_b_relational_utility_function('relation''total', [TRelation,TDomain], BPOS, TCall) :-
    !,
    ID0 = identifier(BPOS,x),
    LHS = member(BPOS,ID0,TDomain),
    RHS = not_equal(BPOS,image(BPOS,TRelation,set_extension(BPOS,[ID0])),empty_set(BPOS)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
% functional[r,d] = !x.(x:d => card(r[{x}]) <= 1)
alloy_to_b_relational_utility_function('relation''functional', [TRelation,TDomain], BPOS, TCall) :-
    !,
    ID0 = identifier(BPOS,x),
    LHS = member(BPOS,ID0,TDomain),
    RHS = less_equal(BPOS,card(BPOS,image(BPOS,TRelation,set_extension(BPOS,[ID0]))),integer(BPOS,1)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
% function[r,d] = !x.(x:d => card(r[{x}]) = 1)
alloy_to_b_relational_utility_function('relation''function', [TRelation,TDomain], BPOS, TCall) :-
    !,
    ID0 = identifier(BPOS,x),
    LHS = member(BPOS,ID0,TDomain),
    RHS = equal(BPOS,card(BPOS,image(BPOS,TRelation,set_extension(BPOS,[ID0]))),integer(BPOS,1)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
% injective[r,c] = !(y).(y:c => card(r~[{y}]) <= 1)
alloy_to_b_relational_utility_function('relation''injective', [TRelation,TCodomain], BPOS, TCall) :-
    !,
    ID0 = identifier(BPOS,y),
    LHS = member(BPOS,ID0,TCodomain),
    RHS = less_equal(BPOS,card(BPOS,image(BPOS,reverse(BPOS,TRelation),set_extension(BPOS,[ID0]))),integer(BPOS,1)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
% surjective[r,c] = !(y).(y:c => r~[{y}] /= {})
alloy_to_b_relational_utility_function('relation''surjective', [TRelation,TCodomain], BPOS, TCall) :-
    !,
    ID0 = identifier(BPOS,y),
    LHS = member(BPOS,ID0,TCodomain),
    RHS = not_equal(BPOS,image(BPOS,reverse(BPOS,TRelation),set_extension(BPOS,[ID0])),empty_set(BPOS)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
% bijective[r,c] = !(y).(y:c => card(r~[{y}]) = 1)
alloy_to_b_relational_utility_function('relation''bijective', [TRelation,TCodomain], BPOS, TCall) :-
    !,
    ID0 = identifier(BPOS,y),
    LHS = member(BPOS,ID0,TCodomain),
    RHS = equal(BPOS,card(BPOS,image(BPOS,reverse(BPOS,TRelation),set_extension(BPOS,[ID0]))),integer(BPOS,1)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
% bijection[r,d,c] = function[r,d] & bijective[r,c]
alloy_to_b_relational_utility_function('relation''bijection', [Relation,Domain,Codomain], BPOS, TCall) :-
    !,
    alloy_to_b_relational_utility_function('relation''function', [Relation,Domain], BPOS, IsFunction),
    alloy_to_b_relational_utility_function('relation''bijective', [Relation,Codomain], BPOS, IsBijective),
    TCall = conjunct(BPOS,IsFunction,IsBijective).
% reflexive[r,s] = id(s) <: r
alloy_to_b_relational_utility_function('relation''reflexive', [TRelation1,TRelation2], BPOS, TCall) :-
    !,
    TCall = subset(BPOS,identity(TRelation2),TRelation1).
% irreflexive[r] = id(dom(r)) /\ r = {}
alloy_to_b_relational_utility_function('relation''irreflexive', [TRelation], BPOS, TCall) :-
    !,
    TCall = equal(BPOS,intersection(BPOS,identity(domain(BPOS,TRelation)),TRelation),empty_set(BPOS)).
% symmetric[r] = r~ <: r
alloy_to_b_relational_utility_function('relation''symmetric', [TRelation], BPOS, TCall) :-
    !,
    TCall = subset(BPOS,reverse(BPOS,TRelation),TRelation).
% antisymmetric[r] = (r~ /\ r) <: id(dom(r))
alloy_to_b_relational_utility_function('relation''antisymmetric', [TRelation], BPOS, TCall) :-
    !,
    Intersection = intersection(BPOS,reverse(BPOS,TRelation),TRelation),
    TCall = subset(BPOS,Intersection,identity(domain(BPOS,TRelation))).
% transitive[r] = (r ; r) <: r
alloy_to_b_relational_utility_function('relation''transitive', [TRelation], BPOS, TCall) :-
    !,
    TCall = subset(BPOS,composition(BPOS,TRelation,TRelation),TRelation).
% acyclic[r,s] = !x.(x:s => x|->x /: closure1(r))
alloy_to_b_relational_utility_function('relation''acyclic', [TRelation1,TRelation2], BPOS, TCall) :-
    !,
    ID0 = identifier(BPOS,x),
    TCall = forall(BPOS,[ID0],implication(BPOS,member(BPOS,ID0,TRelation2),
                                   not_member(BPOS,couple(BPOS,ID0,ID0),closure(BPOS,TRelation1)))).
% complete[r,s] = !(x,y).(x:s & y:s & x /= y => x|->y : (r \/ r~))
alloy_to_b_relational_utility_function('relation''complete', [TRelation1,TRelation2], BPOS, TCall) :-
    !,
    ID0 = identifier(BPOS,x),
    ID1 = identifier(BPOS,y),
    LHS = conjunct(BPOS,conjunct(BPOS,member(BPOS,ID0,TRelation2),member(BPOS,ID1,TRelation2)),not_equal(BPOS,ID0,ID1)),
    RHS = member(BPOS,couple(BPOS,ID0,ID1),union(BPOS,TRelation1,reverse(BPOS,TRelation1))),
    TCall = forall(BPOS,[ID0,ID1],implication(BPOS,LHS,RHS)).
% preorder[r,s] = reflexive[r,s] & transitive[r]
alloy_to_b_relational_utility_function('relation''preorder', [Relation1,Relation2], BPOS, TCall) :-
    !,
    alloy_to_b_relational_utility_function('relation''reflexive', [Relation1,Relation2], BPOS, IsReflexive),
    alloy_to_b_relational_utility_function('relation''transitive', [Relation1], BPOS, IsTransitive),
    TCall = conjunct(BPOS,IsReflexive,IsTransitive).
% equivalence[r,s] = preorder[r,s] & symmetric[r]
alloy_to_b_relational_utility_function('relation''equivalence', [Relation1,Relation2], BPOS, TCall) :-
    !,
    alloy_to_b_relational_utility_function('relation''preorder', [Relation1,Relation2], BPOS, IsReflexive),
    alloy_to_b_relational_utility_function('relation''symmetric', [Relation1], BPOS, IsTransitive),
    TCall = conjunct(BPOS,IsReflexive,IsTransitive).
% partialOrder[r,s] = preorder[r,s] & antisymmetric[r]
alloy_to_b_relational_utility_function('relation''partialOrder', [Relation1,Relation2], BPOS, TCall) :-
    !,
    alloy_to_b_relational_utility_function('relation''preorder', [Relation1,Relation2], BPOS, IsReflexive),
    alloy_to_b_relational_utility_function('relation''antisymmetric', [Relation1], BPOS, IsTransitive),
    TCall = conjunct(BPOS,IsReflexive,IsTransitive).
% totalOrder[r,s] = partialOrder[r,s] & complete[r,s]
alloy_to_b_relational_utility_function('relation''totalOrder', [Relation1,Relation2], BPOS, TCall) :-
    !,
    alloy_to_b_relational_utility_function('relation''partialOrder', [Relation1,Relation2], BPOS, IsReflexive),
    alloy_to_b_relational_utility_function('relation''complete', [Relation1,Relation2], BPOS, IsTransitive),
    TCall = conjunct(BPOS,IsReflexive,IsTransitive).
alloy_to_b_relational_utility_function(FunName, _, BPOS, truth(BPOS)) :-
    atom_concat('relation', _, FunName),
    !,
    add_error(alloy_to_b_relational_utility_function, 'Function from util/relation currently not supported', [], BPOS).

% remove_prime(Old,New) :- atom_codes(Old,OldC) , delete(OldC,39,NewC) , atom_codes(New,NewC).
get_clean_ordering_fun_name(OrderingFunctionName, CleanOrdFunName) :-
    member(CleanOrdFunName, [first,last,min,max,next,nexts,prev,prevs,larger,smaller,lt,lte,gt,gte]),
    atom_concat(_, CleanOrdFunName, OrderingFunctionName).

get_ordering_definition_name(TempFunctionName, SignatureName, FunctionName) :-
    member(TempFunctionName, [next,prev,nexts,prevs]),
    atom_concat(TempFunctionName, '_', TTempFunctionName),
    atom_concat(TTempFunctionName, SignatureName, FunctionName).

translate_binary_e_p(in(Arg1,Arg2,_Type,POS), Translation) :-
    !,
    translate_pos(POS, BPOS),
    translate_e_p(Arg1, TArg1),
    translate_binary_in(TArg1, Arg2, BPOS, Translation).
translate_binary_e_p(CartTerm, TCartesian) :-
    is_cartesian_product_term(CartTerm, Arg1, Arg2, POS),
    !,
    translate_pos(POS, BPOS),
    translate_cartesian(BPOS, Arg1, Arg2, TCartesian).
% S <: Rel --> {x1,..,xk| x1:S & (x1,...,xk):Rel}
translate_binary_e_p(dom_restr(Arg1,Arg2,_Type,POS), TDom) :-
    \+ is_binary_relation(Arg2), % for binary relation translation to B domain_restriction works
    !,
    get_type_and_arity_from_alloy_term(Arg2, _, Arity),
    gen_ids_and_couple(Arity, IDS, Couple),
    translate_pos(POS, BPOS),
    TDom = comprehension_set(BPOS,IDS,conjunct(BPOS,Mem1,Mem2)),
    IDS = [ID1|_],
    check_position(BPOS,translate_binary_e_p),
    Mem1 = member(BPOS,ID1,TArg1),
    Mem2 = member(BPOS,Couple,TArg2),
    translate_e_p(Arg1, TArg1),
    translate_e_p(Arg2, TArg2).
translate_binary_e_p(join(Arg1,Arg2,_Type,POS), TJoin) :-
    !,
    translate_pos(POS, BPOS),
    translate_join(BPOS, Arg1, Arg2, TJoin).
translate_binary_e_p(let(VarName,Expr,Sub,Type,POS), TLet) :-
    !,
    translate_e_p(VarName, TVarName),
    translate_e_p(Expr, TExpr),
    (   Type == type(['PrimitiveBoolean'],0) ->
        NewFunctor = let_predicate
    ;   NewFunctor = let_expression
    ),
    translate_e_p(Sub, TSub),
    translate_pos(POS, BPOS),
    TLet =.. [NewFunctor,BPOS,[TVarName],equal(BPOS,TVarName,TExpr),TSub].
translate_binary_e_p(Call, TCall) :-
    Call =.. [Functor,Name,Params,Type,POS],
    (   Functor = pred_call
    ;   Functor = fun_call
    ),
    !,
    translate_pos(POS, BPOS),
    translate_function_call(Name, Params, Type, BPOS, TCall).
translate_binary_e_p(Binary, TBinary) :-
    Binary =.. [Op,Arg1,Arg2,_Type,POS],
    alloy_to_b_integer_operator(Op, BOp),
    !,
    translate_e_p_integer(Arg1, TArg1),
    translate_e_p_integer(Arg2, TArg2),
    translate_pos(POS, BPOS),
    TBinary =.. [BOp,BPOS,TArg1,TArg2].
translate_binary_e_p(Binary, TBinary) :-
    Binary =.. [Op,Arg1,Arg2,_Type,POS],
    \+ add_error_for_univ_or_iden(translate_binary_e_p, Op, Arg1, Arg2),
    alloy_to_b_binary_operator(Op, BOp),
    translate_e_p(Arg1, TArg1),
    translate_e_p(Arg2, TArg2),
    translate_pos(POS, BPOS),
    TBinary =.. [BOp,BPOS,TArg1,TArg2].

% We do not allow keywords univ or iden for specific operators which we cannot translate to
% an equivalent B expression without introducing the universal type.
binary_op_not_allowing_iden_or_univ(equal).
binary_op_not_allowing_iden_or_univ(not_equal).
binary_op_not_allowing_iden_or_univ(in).
binary_op_not_allowing_iden_or_univ(not_in).

add_error_for_univ_or_iden(Src, Op, Arg1, Arg2) :-
    binary_op_not_allowing_iden_or_univ(Op),
    (   ground(Arg1),
        is_iden(Arg1)
    ;   ground(Arg2),
        is_iden(Arg2)
    ),
    atom_concat('iden keyword in Alloy operator ', Op, Msg),
    add_error_not_supported(Src, Msg).
add_error_for_univ_or_iden(Src, Op, Arg1, Arg2) :-
    binary_op_not_allowing_iden_or_univ(Op),
    (   ground(Arg1),
        is_univ(Arg1)
    ;   ground(Arg2),
        is_univ(Arg2)
    ),
    atom_concat('univ keyword in Alloy operator ', Op, Msg),
    add_error_not_supported(Src, Msg).

add_error_not_supported(Src, NotSupported) :-
    atom(NotSupported),
    add_error(Src, 'Expression not supported by our translation: ', [NotSupported]).

translate_binary_in(TArg1, Arg2, _, _) :-
    add_error_for_univ_or_iden(translate_binary_in, in, TArg1, Arg2),
    !.
% translate TArg1 in Arg2 (multplicities are not allowed for not in)
translate_binary_in(TArg1, Mult, BPOS, Translation) :-
    compound(Mult),
    functor(Mult, AlloyFunctor, 4),              % binary operator with type and pos info
    alloy_to_b_special_multiplicity_type(AlloyFunctor, _, _, _, _, _, _, _), % TO DO: clean up
    arg(1, Mult, MDom),
    arg(2, Mult, MRan),
    translate_e_p(MDom, TMDom),
    translate_e_p(MRan, TMRan),
    alloy_to_b_special_multiplicity_type(AlloyFunctor, BPOS, TArg1, TMDom, TMRan, TMult, AdditionalConstraint, UseInverse),
    !,
    (   UseInverse == yes
    ->  TTArg1 = reverse(BPOS,TArg1)
    ;   TTArg1 = TArg1
    ),
    (   AdditionalConstraint == none
    ->  Translation = member(BPOS,TTArg1,TMult) % use element of rather than subset for special type operators
    ;   Translation = conjunct(BPOS,member(BPOS,TTArg1,TMult),AdditionalConstraint)
    ).
% cartesian
translate_binary_in(TArg1, Arg2, BPOS, subset(BPOS,TArg1,TArg2)) :-
    translate_e_p(Arg2, TArg2).

% translate expressions of the form Arg1 MULT -> MULT Arg2  with MULT : {some,lone, one, ''}
% The result is to be used in member(Pos,LHS,TBinary) not for subset checks !
translate_special_field_multiplicity_binary_e_p(Binary, TBinary) :-
    Binary =.. [AlloyFunctor,Arg1,Arg2,_Type,POS],
    translate_pos(POS, BPOS),
    translate_e_p(Arg1, TArg1),
    translate_e_p(Arg2, TArg2),
    % some functions of alloy_to_b_special_multiplicity_type/8 do not apply to
    % field multiplicities as only one multiplicity is allowed there
    alloy_to_b_special_multiplicity_type(AlloyFunctor, BPOS, _, TArg1, TArg2, TBinary, _, no). % no inverse for field multiplicity

% Alloy: r in T -> some U
% B:     r : T <-> U & dom(r)=T
alloy_to_b_special_multiplicity_type(total_relation, BPOS, TRel, TMDom, TMRan, relations(BPOS,TMDom,TMRan), AdditionalConstraint, no) :-
    AdditionalConstraint = equal(BPOS,domain(BPOS,TRel),TMDom).
% Alloy: r in T -> one U
% B:     r : T --> U
alloy_to_b_special_multiplicity_type(total_function, BPOS, _, TMDom, TMRan, total_function(BPOS,TMDom,TMRan), none, no).
% Alloy: r in T -> lone U
% B:     r : T +-> U
alloy_to_b_special_multiplicity_type(partial_function, BPOS, _, TMDom, TMRan, partial_function(BPOS,TMDom,TMRan), none, no).
% Alloy: r in T some -> U
% B:     r : T <-> U & ran(r) = U
alloy_to_b_special_multiplicity_type(surjection_relation, BPOS, TRel, TMDom, TMRan, relations(BPOS,TMDom,TMRan), AdditionalConstraint, no) :-
    AdditionalConstraint = equal(BPOS,range(BPOS,TRel),TMRan).
% Alloy: r in T some -> some U
% B:     r : T <-> U & dom(r)=T & ran(r) = U
alloy_to_b_special_multiplicity_type(total_surjection_relation, BPOS, TRel, TMDom, TMRan, relations(BPOS,TMDom,TMRan), AdditionalConstraint, no) :-
    AdditionalConstraint = conjunct(BPOS,equal(BPOS,domain(BPOS,TRel),TMDom),equal(BPOS,range(BPOS,TRel),TMRan)).
% Alloy: r in T some -> lone U
% B:     r : T +->> U
alloy_to_b_special_multiplicity_type(partial_surjection, BPOS, _, TMDom, TMRan, partial_surjection(BPOS,TMDom,TMRan), none, no).
% Alloy: r in T some -> one U
% B:     r : T -->> U
alloy_to_b_special_multiplicity_type(total_surjection, BPOS, _, TMDom, TMRan, total_surjection(BPOS,TMDom,TMRan), none, no).
% Alloy: r in T lone -> U
% B:     r~ : U +-> T
alloy_to_b_special_multiplicity_type(injection_relation, BPOS, _, TMDom, TMRan, partial_function(BPOS,TMRan,TMDom), none, yes).
% Alloy: r in T lone -> some U
% B:     r~ : U +->> T
alloy_to_b_special_multiplicity_type(injection_surjection_relation, BPOS, _, TMDom, TMRan, partial_surjection(BPOS,TMRan,TMDom), none, yes).
% Alloy: r in T lone -> lone U
% B:     r : T >+> U
alloy_to_b_special_multiplicity_type(partial_injection, BPOS, _, TMDom, TMRan, partial_injection(BPOS,TMDom,TMRan), none, no).
% Alloy: r in T lone -> one U
% B:     r : T >-> U
alloy_to_b_special_multiplicity_type(total_injection, BPOS, _, TMDom, TMRan, total_injection(BPOS,TMDom,TMRan), none, no).
% Alloy: r in T one -> U
% B:     r~ : U --> T
alloy_to_b_special_multiplicity_type(bijection_relation, BPOS, _, TMDom, TMRan, total_function(BPOS,TMRan,TMDom), none, yes).
% Alloy: r in T one -> some U
% B:     r~ : U -->> T
alloy_to_b_special_multiplicity_type(total_bijection_relation, BPOS, _, TMDom, TMRan, total_surjection(BPOS,TMRan,TMDom), none, yes).
% Alloy: r in T one -> lone U
% B:     r : T >+>> U
alloy_to_b_special_multiplicity_type(partial_bijection, BPOS, _, TMDom, TMRan, partial_bijection(BPOS,TMDom,TMRan), none, no).
% Alloy: r in T one -> one U
% B:     r : T >->> U
alloy_to_b_special_multiplicity_type(total_bijection, BPOS, _, TMDom, TMRan, total_bijection(BPOS,TMDom,TMRan), none, no).

% Note: Sequences in B are indexed from 1 to n, in Alloy from 0 to n-1.
%       We do not use B sequences when translating from Alloy but define a partial function with domain 0..n-1
%       (we have to manually implement most functions anyway)
% Documentation available at http://alloytools.org/quickguide/seq.html
%% Sequence function calls without a parameter but the sequence itself.
% first: seq[{0}]
sequence_function_to_b('seq\'first', BPOS, [Seq], image(BPOS,Seq,set_extension(BPOS,[integer(BPOS,0)]))).
% last: seq[{card(seq)-1}]
sequence_function_to_b('seq\'last', BPOS, [Seq], image(BPOS,Seq,set_extension(BPOS,[minus(BPOS,card(BPOS,Seq),integer(BPOS,1))]))).
% rest: IF seq = <> THEN <> ELSE %x.(x:0..card(seq)-2 | seq(x+1) END
sequence_function_to_b('seq\'rest', BPOS, [Seq], if_then_else(BPOS,equal(BPOS,Seq,empty_sequence(BPOS)),empty_sequence(BPOS),lambda(BPOS,[identifier(BPOS,x)],member(BPOS,identifier(BPOS,x),interval(BPOS,integer(BPOS,0),minus(BPOS,card(BPOS,Seq),integer(BPOS,2)))),function(BPOS,Seq,add(BPOS,identifier(BPOS,x),integer(BPOS,1)))))).
% elems: ran(seq)
sequence_function_to_b('seq\'elems', BPOS, [Seq], range(BPOS,Seq)).
% butlast: seq[{card(seq)-2}]
sequence_function_to_b('seq\'butlast', BPOS, [Seq], image(BPOS,Seq,set_extension(BPOS,[minus(BPOS,card(BPOS,Seq),integer(BPOS,2))]))).
% isEmpty: seq = <>
sequence_function_to_b('seq\'isEmpty', BPOS, [Seq], equal(BPOS,Seq,empty_sequence(BPOS))).
% hasDups: card(seq) /= card(ran(seq))
sequence_function_to_b('seq\'hasDups', BPOS, [Seq], not_equal(BPOS,card(BPOS,Seq),card(BPOS,range(BPOS,Seq)))).
% inds: 0..card(seq)-1
sequence_function_to_b('seq\'inds', BPOS, [Seq], interval(BPOS,integer(BPOS,0),minus(BPOS,card(BPOS,Seq),integer(BPOS,1)))).
% lastIdx: {card(seq)-1} /\ 0..(card(seq)-1)
sequence_function_to_b('seq\'lastIdx', BPOS, [Seq], intersection(BPOS,set_extension(BPOS,[minus(BPOS,card(BPOS,Seq),integer(BPOS,1))]),interval(BPOS,integer(BPOS,0),minus(BPOS,card(BPOS,Seq),integer(BPOS,1))))).
% afterLastIdx: {card(seq)} /\ 0..(MAXSEQ-1)
sequence_function_to_b('seq\'afterLastIdx', BPOS, [Seq], intersection(BPOS,set_extension(BPOS,[card(BPOS,Seq)]),interval(BPOS,integer(BPOS,0),minus(BPOS,integer(BPOS,MaxSeq),integer(BPOS,1))))) :-
    maxseq(MaxSeq).
%% Sequence function calls with one parameter and the sequence itself.
% idxOf[X]: IF seq~[{E_{one}(X)}] /= {} THEN {min(seq~[{E_{one}(X)}])} ELSE {} END
sequence_function_to_b('seq\'idxOf', BPOS, [Seq,P1], if_then_else(BPOS,not_equal(BPOS,image(BPOS,reverse(BPOS,Seq),set_extension(BPOS,[P1One])),empty_set(BPOS)),set_extension(BPOS,[min(BPOS,image(BPOS,reverse(BPOS,Seq),set_extension(BPOS,[P1One])))]),empty_set(BPOS))) :-
    translate_e_p_one(P1, P1One).
% lastIdxOf[X]: IF seq~[{E_{one}(X)}] /= {} THEN {max(seq~[{E_{one}(X)}])} ELSE {} END
sequence_function_to_b('seq\'lastIdxOf', BPOS, [Seq,P1], if_then_else(BPOS,not_equal(BPOS,image(BPOS,reverse(BPOS,Seq),set_extension(BPOS,[P1One])),empty_set(BPOS)),set_extension(BPOS,[max(BPOS,image(BPOS,reverse(BPOS,Seq),set_extension(BPOS,[P1One])))]),empty_set(BPOS))) :-
    translate_e_p_one(P1, P1One).
% indsOf[X]: seq~[{E_{one}(X)}]
sequence_function_to_b('seq\'indsOf', BPOS, [Seq,P1], image(BPOS,reverse(BPOS,Seq),set_extension(BPOS,[P1One]))) :-
    translate_e_p_one(P1, P1One).
% append[seq2]: (0..(MAXSEQ-1)) <| (seq \/ %x.(x:card(seq)..(card(seq)+card(seq2)-1) | seq2(x-card(seq))))
sequence_function_to_b('seq\'append', BPOS, [Seq,P1], domain_restriction(BPOS,interval(BPOS,integer(BPOS,0),minus(BPOS,integer(BPOS,MaxSeq),integer(BPOS,1))),union(BPOS,Seq,lambda(BPOS,[identifier(BPOS,x)],member(BPOS,identifier(BPOS,x),interval(BPOS,card(BPOS,Seq),minus(BPOS,add(BPOS,card(BPOS,Seq),card(BPOS,P1)),integer(BPOS,1)))),function(BPOS,P1,minus(BPOS,identifier(BPOS,x),card(BPOS,Seq))))))) :-
    maxseq(MaxSeq).
% Throw WD error if X is not singleton.
% add[X]: IF card(seq) < MAXSEQ THEN seq \/ {card(seq) |-> E_{one}(X)} ELSE seq END
sequence_function_to_b('seq\'add', BPOS, [Seq,P1], if_then_else(BPOS,less(BPOS,card(BPOS,Seq),integer(BPOS,MaxSeq)),union(BPOS,Seq,set_extension(BPOS,[couple(BPOS,card(BPOS,Seq),P1One)])),Seq)) :-
    maxseq(MaxSeq),
    translate_e_p_one(P1, P1One).
% delete[i]: IF (i >= 0) THEN (0..(i-1) <| seq) \/ %x.(x:i..(card(seq)-2) & (x+1):dom(seq) | seq(x+1)) ELSE seq END
sequence_function_to_b('seq\'delete', BPOS, [Seq,P1], if_then_else(BPOS,greater_equal(BPOS,P1Int,integer(BPOS,0)),union(BPOS,domain_restriction(BPOS,interval(BPOS,integer(BPOS,0),minus(BPOS,P1Int,integer(BPOS,1))),Seq),lambda(BPOS,[identifier(BPOS,x)],conjunct(BPOS,member(BPOS,identifier(BPOS,x),interval(BPOS,P1Int,minus(BPOS,card(BPOS,Seq),integer(BPOS,2)))),member(BPOS,add(BPOS,identifier(BPOS,x),integer(BPOS,1)),domain(BPOS,Seq))),function(BPOS,Seq,add(BPOS,identifier(BPOS,x),integer(BPOS,1))))),Seq)) :-
    untyped_b_ast_to_integer(P1, P1Int).
% Throw WD error if X is not singleton.
% setAt[i,X]: IF (i >= 0 & i <= card(seq) ) THEN seq <+ {i |-> E_{one}(X)} ELSE seq END
sequence_function_to_b('seq\'setAt', BPOS, [Seq,P1,P2], if_then_else(BPOS,conjunct(BPOS,greater_equal(BPOS,P1Int,integer(BPOS,0)),less_equal(BPOS,P1Int,card(BPOS,Seq))),overwrite(Seq,set_extension(BPOS,[couple(BPOS,P1Int,P2One)])),Seq)) :-
    untyped_b_ast_to_integer(P1, P1Int),
    translate_e_p_one(P2, P2One).
% Throw WD error if X is not singleton.
% special case for i=0
% insert[0,X]: 0..(m-1) <| {0 |-> E_{one}(X)} \/ %z.(z:1..card(seq) ∧ (z-1):dom(seq) | seq(z-1)))
sequence_function_to_b('seq\'insert', BPOS, [Seq,P1,P2], domain_restriction(BPOS,interval(BPOS,integer(BPOS,0),minus(BPOS,integer(BPOS,MaxSeq),integer(BPOS,1))),union(BPOS,set_extension(BPOS,[couple(BPOS,integer(BPOS,0),P2One)]),lambda(BPOS,[identifier(BPOS,z)],member(BPOS,identifier(BPOS,z),interval(BPOS,integer(BPOS,1),card(BPOS,Seq))),function(BPOS,Seq,minus(BPOS,identifier(BPOS,z),integer(BPOS,1))))))) :-
    untyped_b_ast_to_integer(P1, P1Int),
    P1Int = integer(_,0),
    maxseq(MaxSeq),
    translate_e_p_one(P2, P2One).
% insert[i,X]: 0..(m-1) <| ((0..(i-1) <| seq) \/ {i |-> E_{one}(X)} \/ %z.(z:(i+1)..card(seq) ∧ (z-1):dom(seq) | seq(z-1)))
sequence_function_to_b('seq\'insert', BPOS, [Seq,P1,P2], domain_restriction(BPOS,interval(BPOS,integer(BPOS,0),minus(BPOS,integer(BPOS,MaxSeq),integer(BPOS,1))),union(BPOS,union(BPOS,domain_restriction(BPOS,interval(BPOS,integer(BPOS,0),minus(BPOS,P1Int,integer(BPOS,1))),Seq),set_extension(BPOS,[couple(BPOS,P1Int,P2One)])),lambda(BPOS,[identifier(BPOS,z)],member(BPOS,identifier(BPOS,z),interval(BPOS,add(BPOS,P1Int,integer(BPOS,1)),card(BPOS,Seq))),function(BPOS,Seq,minus(BPOS,identifier(BPOS,z),integer(BPOS,1))))))) :-
    P1 \= integer(_,0),
    maxseq(MaxSeq),
    untyped_b_ast_to_integer(P1, P1Int),
    translate_e_p_one(P2, P2One).
% subseq[from,to]: IF 0 <= from & from <= to & to < card(seq) THEN %z.(z:0..(to-from) & (z+from):dom(seq) | seq(z+from)) ELSE {} END
sequence_function_to_b('seq\'subseq', BPOS, [Seq,P1,P2], if_then_else(BPOS,conjunct(BPOS,conjunct(BPOS,less_equal(BPOS,integer(BPOS,0),P1Int),less_equal(BPOS,P1Int,P2Int)),less(BPOS,P2Int,card(BPOS,Seq))),lambda(BPOS,[identifier(BPOS,z)],member(BPOS,identifier(BPOS,z),interval(BPOS,integer(BPOS,0),minus(BPOS,P2Int,P1Int))),function(BPOS,Seq,add(BPOS,identifier(BPOS,z),P1Int))),empty_set(BPOS))) :-
    untyped_b_ast_to_integer(P1, P1Int),
    untyped_b_ast_to_integer(P2, P2Int).
sequence_function_to_b(FunName, BPOS, _, Res) :-
    atom_concat('seq', _, FunName),
    !,
    add_error(sequence_function_to_b, 'Function from util/sequences currently not supported', [], BPOS),
    Res = truth(BPOS).

translate_seq_function_param(join(identifier(this,type([Type],1),IPos),Arg2,JType,JPos), TParam) :-
    !,
    % remove join with this for signature fields
    translate_e_p(join(identifier(Type,type([Type],1),IPos),Arg2,JType,JPos), TTParam),
    TParam = TTParam.
translate_seq_function_param(Param, TParam) :-
    translate_e_p(Param, TParam).

% Translation of operations from utility libraries
% util/sequence
translate_function_call(Function, Params, _Type, BPOS, TCall) :-
    maplist(translate_seq_function_param, Params, TParams),
    sequence_function_to_b(Function, BPOS, TParams, TCall),
    !.
% util/boolean
translate_function_call(Name, Params, _Type, BPOS, Result) :-
    maplist(translate_e_p, Params, TParams),
    alloy_to_b_boolean_utility_function(Name, TParams, BPOS, Result).
% util/relation
translate_function_call(Name, Params, _Type, BPOS, Result) :-
    maplist(translate_e_p, Params, TParams),
    check_position(BPOS,alloy_to_b_relational_utility_function),
    alloy_to_b_relational_utility_function(Name, TParams, BPOS, Result).
% util/integer with no direct counterparts in B
translate_function_call(Name, Params, _Type, BPOS, Result) :-
    is_util_integer_natural_func(Name, UtilType, AlloyOp),
    UtilType == integer,
    maplist(translate_e_p_integer, Params, TParams),
    alloy_to_b_integer_utility_function(AlloyOp, BPOS, TParams, Result).
% util/natural with no direct counterparts in B
translate_function_call(Name, Params, _Type, BPOS, Result) :-
    is_util_integer_natural_func(Name, UtilType, AlloyOp),
    UtilType == natural,
    maplist(translate_e_p_integer, Params, TParams),
    alloy_to_b_natural_utility_function(AlloyOp, BPOS, TParams, Result),
    !.
% util/integer and util/natural with direct counterparts in B
translate_function_call(Name, Params, _Type, BPOS, Result) :-
    length(Params, Arity),
    alloy_to_b_integer_natural_utility_func(Name, Arity, BOp, ArgType, ReturnType),
    !,                                       % functions like plus, minus, mul, ... with integer arguments and result
    (   ArgType = integer ->
        maplist(translate_e_p_integer, Params, TParams)
    ;   maplist(translate_e_p, Params, TParams)
    ),
    (   ReturnType = integer ->
        Result = set_extension(BPOS,[TCall]) % we add set_extension wrapper
    ;   Result = TCall
    ),
    TCall =.. [BOp,BPOS|TParams].
% util/ordering
translate_function_call(FunName, Params, type([SignatureName|_],_), BPOS, TCall) :-
    translate_e_p(SignatureName, TSignatureName),
    get_clean_ordering_fun_name(FunName, CleanOrdFunName),
    translate_ordering_function(CleanOrdFunName, Params, SignatureName, TSignatureName, BPOS, TCall).
% definition call for calls to user defined Alloy predicates and functions
translate_function_call(Name, Params, _, BPOS, TCall) :-
    % predicate and function call
    maplist(translate_e_p, Params, TParams),
    maplist(strip_singleton_set_from_this, TParams, TTParams), % identifier 'this' is added for fields which is already wrapped in a set extension
    TCall = definition(BPOS,Name,TTParams).

% first: IF S_ord /= {} THEN {min(S_ord)} ELSE {}
translate_ordering_function(first, [], _, TSignatureName, BPOS, if_then_else(BPOS,not_equal(BPOS,TSignatureName,empty_set(BPOS)),set_extension(BPOS,[min(BPOS,TSignatureName)]),empty_set(BPOS))) :-
    !.
% last: IF S_ord /= {} THEN {max(S_ord)} ELSE {}
translate_ordering_function(last, [], _, TSignatureName, BPOS, if_then_else(BPOS,not_equal(BPOS,TSignatureName,empty_set(BPOS)),set_extension(BPOS,[max(BPOS,TSignatureName)]),empty_set(BPOS))) :-
    !.
% min[p]: IF p /= {} THEN {min(p)} ELSE {}
% max[p]: IF p /= {} THEN {min(p)} ELSE {}
translate_ordering_function(FunName, [Param], _, _, BPOS, TCall) :-
    (   FunName == min
    ;   FunName == max
    ),
    !,
    translate_e_p(Param, TParam),
    InnerCall =.. [FunName,BPOS,TParam],
    TCall = if_then_else(BPOS,not_equal(BPOS,TParam,empty_set(BPOS)),set_extension(BPOS,[InnerCall]),empty_set(BPOS)).
translate_ordering_function(FunName, [Param], SignatureName, _, BPOS, TCall) :-
    (   FunName == next
    ;   FunName == nexts
    ;   FunName == prev
    ;   FunName == prevs
    ),
    !,
    get_ordering_definition_name(FunName, SignatureName, DefName),
    translate_e_p(Param, TParam),
    TCall = definition(BPOS,DefName,[TParam]).
% larger[p1,p1]: IF p1 \/ p2 /= {} THEN {max(p1 \/ p2)} ELSE {} END
% smaller[p1,p2]: IF p1 \/ p2 /= {} THEN {min(p1 \/ p2)} ELSE {} END
translate_ordering_function(FunName, [P1,P2], _, _, BPOS, TCall) :-
    (   FunName == larger
    ;   FunName == smaller
    ),
    get_partial_b_functor_for_ordering(FunName, BFunctor),
    translate_e_p(P1, TP1),
    translate_e_p(P2, TP2),
    ParamUnion = union(BPOS,TP1,TP2),
    TInnerCall =.. [BFunctor,BPOS,ParamUnion],
    InnerCall = set_extension(BPOS,[TInnerCall]),
    TCall = if_then_else(BPOS,not_equal(BPOS,ParamUnion,empty_set(BPOS)),InnerCall,empty_set(BPOS)).
% lt[p1,p2]: (p1 = {}) or (p1 /= {} & p2 /= {} & p1 /= p2 & {min(p1 \/ p2)} = p1))
% gt[p1,p2]: (p1 = {}) or (p1 /= {} & p2 /= {} & p1 /= p2 & {max(p1 \/ p2)} = p1))
translate_ordering_function(FunName, [P1,P2], _, _, BPOS, TCall) :-
    (   FunName == lt
    ;   FunName == gt
    ),
    get_partial_b_functor_for_ordering(FunName, BFunctor),
    translate_e_p(P1, TP1),
    translate_e_p(P2, TP2),
    TInnerCall =.. [BFunctor,BPOS,union(BPOS,TP1,TP2)],
    InnerCall = set_extension(BPOS,[TInnerCall]),
    LHS = equal(BPOS,TP1,empty_set(BPOS)),
    RHS = conjunct(BPOS,conjunct(BPOS,conjunct(BPOS,not_equal(BPOS,TP1,empty_set(BPOS)),not_equal(BPOS,TP2,empty_set(BPOS))),not_equal(BPOS,TP1,TP2)),equal(BPOS,InnerCall,TP1)),
    TCall = disjunct(BPOS,LHS,RHS).
% lte[p1,p2]: (p1 = {}) or (p1 /= {} & p2 /= {} & {min(p1 \/ p2)} = p1)
% gte[p1,p2]: (p1 = {}) or (p1 /= {} & p2 /= {} & {max(p1 \/ p2)} = p1)
translate_ordering_function(FunName, [P1,P2], _, _, BPOS, TCall) :-
    (   FunName == lte
    ;   FunName == gte
    ),
    get_partial_b_functor_for_ordering(FunName, BFunctor),
    translate_e_p(P1, TP1),
    translate_e_p(P2, TP2),
    TInnerCall =.. [BFunctor,BPOS,union(BPOS,TP1,TP2)],
    InnerCall = set_extension(BPOS,[TInnerCall]),
    LHS = equal(BPOS,TP1,empty_set(BPOS)),
    RHS = conjunct(BPOS,conjunct(BPOS,not_equal(BPOS,TP1,empty_set(BPOS)),not_equal(BPOS,TP2,empty_set(BPOS))),equal(BPOS,InnerCall,TP1)),
    TCall = disjunct(BPOS,LHS,RHS).
translate_ordering_function(FunName, _, _, _, BPOS, empty_set(BPOS)) :-
    add_error(translate_ordering_function, 'Cannot translate ordering function', [FunName], BPOS).

get_partial_b_functor_for_ordering(larger, max).
get_partial_b_functor_for_ordering(smaller, min).
get_partial_b_functor_for_ordering(lt, min).
get_partial_b_functor_for_ordering(lte, min).
get_partial_b_functor_for_ordering(gt, max).
get_partial_b_functor_for_ordering(gte, max).

strip_singleton_set(set_extension(_,[identifier(Pos,Name)]), identifier(Pos,Name)) :-
    !.
strip_singleton_set(A, A).

strip_singleton_set_from_this(set_extension(_,[identifier(Pos,this)]), identifier(Pos,this)) :-
    !.
strip_singleton_set_from_this(A, A).

translate_cartesian(TPos, Arg1, Arg2, Res) :-
    translate_e_p(Arg1, TArg1),
    translate_cartesian2(TPos, TArg1, Arg2, Res).

translate_cartesian2(TPos, TArg1, Arg2, Res) :-
    translate_e_p(Arg2, TArg2),
    translate_cartesian3(TPos, TArg1, Arg2, TArg2, Res).
% P -> Q  --> P*Q
translate_cartesian3(TPos, TArg1, Arg2, TArg2, cartesian_product(TPos,TArg1,TArg2)) :-
    is_unary_relation(Arg2),
    !.
% P -> Q  --> {c0, c1,...,ck | c0:P & (c1,...,ck):Q}
translate_cartesian3(TPos, TArg1, Arg2, TArg2, comprehension_set(TPos,[ID0|IDs2],conjunct(TPos,Mem1,Mem2))) :-
    get_type_and_arity_from_alloy_term(Arg2, _Type, Arity),
    gen_ids_and_couple(Arity, IDs2, Couple2),
    gen_identifier(0, "_c_", ID0),
    check_position(TPos,translate_cartesian3),
    Mem1 = member(TPos,ID0,TArg1),
    Mem2 = member(TPos,Couple2,TArg2).

gen_ids_and_couple(Arity, [ID1|IDListRest], NestedCouple) :-
    Arity > 0,
    gen_identifier(1, "_c_", ID1),
    gen_ids_and_couple(2, Arity, IDListRest, ID1, NestedCouple).

gen_ids_and_couple(Nr, Arity, L, Acc, Res) :-
    Nr > Arity,
    !,
    L = [],
    Res = Acc.
gen_ids_and_couple(Nr, Arity, [IDN|T], Acc, Res) :-
    gen_identifier(Nr, "_c_", IDN),
    N1 is Nr+1,
    gen_ids_and_couple(N1, Arity, T, couple(none,Acc,IDN), Res).

gen_identifier(Nr, Prefix, identifier(none,ID)) :-
    number_codes(Nr, NrC),
    append(Prefix, NrC, AC),
    atom_codes(ID, AC).

% ordering function calls like s.next.next are converted to fun_call
translate_join(_, Arg1, Arg2, TBinaryJoin) :-
    Arg2 = fun_call(Name,[],Type,A2Pos),
    \+ atom_concat('integer', _, Name),
    translate_e_p(fun_call(Name,[Arg1],Type,A2Pos), TBinaryJoin).
% Translation of the dot join operator has several special cases depending on the arity of the arguments.
translate_join(TPos, Arg1, Arg2, TBinaryJoin) :-
    translate_e_p(Arg1, TArg1),
    translate_e_p(Arg2, TArg2),
    translate_join_aux(TPos, Arg1, Arg2, TArg1, TArg2, TBinaryJoin).

% Note: N-ary relation are encoded in B as follows:
% { (e1 |-> (e2 |-> (e3 |-> ...))) ,... }, for example:
% {(tracy|->(spanish|->french)),(carol|->(spanish|->french)), ...}
% univ.P  -->  ran(P)  % this does NOT work for ternary,... relations
translate_join_aux(TPos, Arg1, Arg2, _TArg1, TArg2, range(TPos,TArg2)) :-
    is_univ(Arg1),
    is_binary_relation(Arg2),
    !.
% P.univ  -->  dom(P)  % this works for ternary,... relations
translate_join_aux(TPos, _Arg1, Arg2, TArg1, _TArg2, domain(TPos,TArg1)) :-
    is_univ(Arg2),
    !.
% {el}.R --> {R(el)}
translate_join_aux(TPos, Arg1, Arg2, TArg1, TArg2, set_extension(TPos,[function(TPos,TArg2,[TTArg1])])) :-
    get_type_and_arity_from_alloy_term(Arg1, Type1, Arity),
    Arity = 1,                  % is_unary_relation(Arg1),
    is_binary_relation(Arg2),   % TO DO: Arg2 can probably be n-ary?
    is_total_function(TArg2, Type1),
    can_remove_singleton_set(TArg1, TTArg1),
    !.
% unary.R --> R[unary]
translate_join_aux(TPos, Arg1, Arg2, TArg1, TArg2, image(TPos,TArg2,TArg1)) :- % Unary.Arg2 --> Arg2[Unary]
    is_unary_relation(Arg1),
    is_binary_relation(Arg2),
    !.
% binary.unary
% Next RULE is broken, can lead to Type errors and test is wrong:
% translate_join_aux(Pos,Arg1,Arg2,TArg1,TArg2,function(Pos,reverse(Pos,TArg1),[TTArg2])) :-
% % function call if lhs is a total function and rhs a unary relation
% is_binary_relation(Arg1),
% is_unary_relation(Arg2),
% is_total_function(TArg1) , ! ,  % TEST SHOULD BE SURJECTIVE and INJECTIVE !!
% remove_singleton_set(TArg2,TTArg2).
% Arg1.Unary --> Arg1~[Unary]
translate_join_aux(TPos, _Arg1, Arg2, TArg1, TArg2, image(TPos,reverse(TPos,TArg1),TArg2)) :-
    is_unary_relation(Arg2),
    !.
% binary.binary:  Arg1.Arg2 --> (Arg1 ; Arg2)
translate_join_aux(TPos, Arg1, Arg2, TArg1, TArg2, composition(TPos,TArg1,TArg2)) :-
    is_binary_relation(Arg1),
    is_binary_relation(Arg2),
    !.
translate_join_aux(TPos, Arg1, Arg2, TArg1, TArg2, comprehension_set(TPos,IDS,Body)) :-
    Body = conjunct(TPos,TypingPred,exists(TPos,[JoinID],conjunct(TPos,JoinTypingPred,conjunct(TPos,Mem1,Mem2)))),
    get_type_and_arity_from_alloy_term(Arg2, [JoinIDType|RTypes], Arity), % first type refers to outter variable
    gen_ids_and_couple(Arity, IDs2, Couple2),
    IDs2 = [JoinID|Rest2],
    get_typing_pred_for_ids(TPos, RTypes, Rest2, TypingPred),
    get_typing_pred_for_ids(TPos, [JoinIDType], [JoinID], JoinTypingPred),
    (   is_univ(Arg1) ->
        Mem1 = truth(TPos),
        IDS = Rest2
    ;   is_unary_relation(Arg1) ->
        strip_singleton_set_from_this(TArg1, TTArg1),
        Mem1 = member(TPos,JoinID,TTArg1),
        IDS = Rest2
    ;   gen_identifier(0, "_c_", ID0),
        IDS = [ID0|Rest2],
        Mem1 = member(TPos,couple(TPos,ID0,JoinID),TArg1)
    ),
    check_position(TPos,translate_join_aux),
    Mem2 = member(TPos,Couple2,TArg2).
translate_join_aux(Pos, Arg1, Arg2, _TArg1, _TArg2, empty_set(Pos)) :-
    format('~nJoin not supported this way:~nLeft: ~w~nRight: ~w~n', [Arg1,Arg2]),
    add_error(load_alloy_model, 'Cannot translate this type of join yet', [Arg1,Arg2], Pos).

% This predicate relies on the order of elements in 2. and 3. argument for typing.
% TO DO: improve
get_typing_pred_for_ids(Pos, [], [], truth(Pos)).
get_typing_pred_for_ids(Pos, [Type|TT], [ID|IT], TypingPred) :-
    translate_e_p(Type, TType),
    get_typing_pred_for_ids_acc(Pos, TT, IT, member(Pos,ID,TType), TypingPred).

get_typing_pred_for_ids_acc(_, [], [], Acc, Acc).
get_typing_pred_for_ids_acc(Pos, [Type|TT], [ID|IT], Acc, TypingPred) :-
    translate_e_p(Type, TType),
    get_typing_pred_for_ids_acc(Pos, TT, IT, conjunct(Pos,Acc,member(Pos,ID,TType)), TypingPred).

% translate Alloy position to B position
translate_pos(pos(Col,Row), Res) :-
    !,
    Res = pos(0,1,Row,Col,Row,Col).
translate_pos(_, none).

%get_position(Term, BPos) :-
%    get_alloy_position(Term, APos),
%    !,
%    translate_pos(APos, BPos).
%get_position(_, none).

get_alloy_position(Term, APos) :-
    functor(Term, _, Arity),
    Arity > 0,
    arg(Arity, Term, APos).

get_b_position(UntypedBAst, BPOS) :-
    functor(UntypedBAst, _, Arity),
    Arity > 0,
    arg(1, UntypedBAst, BPOS).

check_position(Var,PP) :- var(Var),!,add_internal_error('Variable position at: ',PP),Var=none.
check_position(none,_) :- !.
check_position(pos(Id,F,SL,SC,EL,EC),_) :- number(Id),atomic(F),number(SL),number(SC),number(EL),number(EC),!.
check_position(T,PP) :-  add_warning(check_position,'Unknown position at: ',PP:T).

translate_quantified_e(Pos, no, TArg, equal(Pos,TArg,empty_set(none))).
translate_quantified_e(Pos, AlloyKw, TArg, equal(Pos,card(none,TArg),integer(none,1))) :-
    % naming possibly differs having the same semantics
    AlloyKw = one; AlloyKw = oneof.
translate_quantified_e(Pos, AlloyKw, TArg, greater(Pos,card(none,TArg),integer(none,0))) :-
    AlloyKw = some; AlloyKw = someof.
translate_quantified_e(Pos, AlloyKw, TArg, less_equal(Pos,card(none,TArg),integer(none,1))) :-
    AlloyKw = lone; AlloyKw = loneof.
translate_quantified_e(_, exactlyof, TArg, TArg).

% Join a list of types using cartesian_product like ((A*B)*C) for types [A,B,C].
create_cartesian_type([ID], TID) :-
    !,
    translate_unary_e_p(ID, TID).
create_cartesian_type([ID1,ID2], cartesian_product(none,TID1,TID2)) :-
    !,
    translate_unary_e_p(ID1, TID1),
    translate_unary_e_p(ID2, TID2).
create_cartesian_type([ID1,ID2|T], cartesian_product(none,cartesian_product(none,TID1,TID2),NT)) :-
    translate_unary_e_p(ID1, TID1),
    translate_unary_e_p(ID2, TID2),
    create_cartesian_type(T, NT).

% Field declarations have several special cases depending on the keywords set, one, some, lone or seq.
% We are inside Signature TSignatureName  sig NAME {  field : DeclTerm }
translate_field_decl(BPOS, SignatureName, TSignatureName, DeclTerm, FieldID, TFieldID, TFieldPred) :-
    % DeclTerm =.. [_,SetID|_],
    replace_this_in_signature_field(DeclTerm, NewDeclTerm),
    debug_format(19, 'Translating Field ~w inside signature ~w,~n   Term=~w,~n~n', [FieldID,SignatureName,NewDeclTerm]),
    (   translate_special_field_multiplicity_binary_e_p(NewDeclTerm, TFUNC)
    ->
        % we could do: translate_cartesian2(BPOS,TSignatureName,NewDeclTerm,Cart), TFieldPred = subset(BPOS,TFieldID,Cart), and add predicate
        % TO DO: better treatment, also detect nested special multiplicity annotations inside NewDeclTerm
        % generate forall x:SIG => x.field : TFUNC
        ID = identifier(BPOS,'_x_'),
        AID = identifier('_x_',type([SignatureName],1),none),
        (   singleton_set('_x_') ->
            true
        ;   assert_singleton_set('_x_')
        ),
        translate_e_p(AID, TAID), % will add set-extension wrapper
        LHS = member(BPOS,ID,TSignatureName),
        RHS = member(BPOS,TJoin,TFUNC),
        TFieldPred = conjunct(BPOS,FieldTyping,forall(BPOS,[ID],implication(none,LHS,RHS))),
        translate_join_aux(BPOS, AID, FieldID, TAID, TFieldID, TJoin),
        get_type_and_arity_from_alloy_term(FieldID, TypeList, _),
        create_cartesian_type(TypeList, CartesianType),
        FieldTyping = member(Pos,TFieldID,pow_subset(Pos,CartesianType))
    ;   translate_sig_field_rhs(FieldID, NewDeclTerm, BPOS, TSignatureName, TFieldID, FieldTypeExpr)
    ->
        add_constraint_for_someof_field(NewDeclTerm, member(BPOS,TFieldID,FieldTypeExpr), TSignatureName, TTFieldPred),
        add_constraint_for_seq_field(NewDeclTerm, TSignatureName, TTFieldPred, TFieldID, TFieldPred)
    ;   add_error(alloy2b, 'Field declaration not implemented:', NewDeclTerm, BPOS),
        TFieldPred = truth(none)
    ).

add_constraint_for_someof_field(someof(_,_,_), TempPred, TSignatureName, TPred) :-
    !,
    TempPred = member(BPOS,TFieldID,_),
    TPred = conjunct(BPOS,equal(BPOS,domain(BPOS,TFieldID),TSignatureName),TempPred).
add_constraint_for_someof_field(_, Pred, _, Pred).

add_constraint_for_seq_field('is_seq'(_,_,_), TSignatureName, TempPred, TFieldID, TPred) :-
    !,
    % !_s.(_s : TSignatureName => dom(TFieldID)[{_s}] = 0..card(dom(TFieldID)[{_s}])-1)
    TPred = conjunct(BPOS,TempPred,forall([identifier(BPOS,'_s')],implication(BPOS,member(identifier(BPOS,'_s'),TSignatureName),equal(BPOS,image(BPOS,domain(BPOS,TFieldID),set_extension(BPOS,[identifier(BPOS,'_s')])),interval(BPOS,integer(BPOS,0),minus(BPOS,card(BPOS,image(BPOS,domain(BPOS,TFieldID),set_extension(BPOS,[identifier(BPOS,'_s')]))),integer(BPOS,1))))))).
add_constraint_for_seq_field(_, _, Pred, _, Pred).

% sig SIG { field : set Set}  -->  field :  SIG <-> Set
translate_sig_field_rhs(_, setof(Set,_,_), _, SIG, _, relations(none,SIG,TSet)) :-
    translate_e_p(Set, TSet).
% sig SIG { field : one Set}  -->  field :  SIG --> Set
translate_sig_field_rhs(_, oneof(Set,_,_), _, SIG, TFieldID, total_function(none,SIG,TSet)) :-
    translate_e_p(Set, TSet),
    assert_total_function(TFieldID, SIG).
% sig SIG { field : lone Set}  -->  field :  SIG +-> Set
translate_sig_field_rhs(_, loneof(Set,_,_), _, SIG, _, partial_function(none,SIG,TSet)) :-
    translate_e_p(Set, TSet).
% sig SIG { field : some Set}  -->  field :  SIG +-> Set
translate_sig_field_rhs(_, someof(Set,_,_), _, SIG, _, relations(none,SIG,TSet)) :-
    translate_e_p(Set, TSet).
% sig SIG { field : seq Set}  -->  field :  (SIG * Integer) +-> Set
translate_sig_field_rhs(FieldID, 'is_seq'(_,Set,_,_), _, SIG, _, partial_function(none,cartesian_product(none,SIG,interval(none,integer(none,0),integer(none,MaxSeq1))),TSet)) :-
    maxseq(MaxSeq),
    MaxSeq1 is MaxSeq - 1,
    FieldID = identifier(Name,Type,_),
    prepend_identifier_of_module_import(Name, Type, NName),
    assert(is_seq(NName)),
    translate_e_p(Set, TSet).
% sig SIG { field : FieldTerm}  -->  field :  POW(SIG * FieldTerm)
translate_sig_field_rhs(_, FieldTerm, Pos, SIG, _, pow_subset(none,Cart)) :-
    % TO DO: check whether lone, one, ... can appear inside FieldTerm
    translate_cartesian2(Pos, SIG, FieldTerm, Cart).

replace_this_in_signature_field(identifier('this',type([SignatureName],1),Pos), Res) :-
    !,
    Res = identifier(SignatureName,type([SignatureName],1),Pos).
replace_this_in_signature_field(DeclTerm, NewDeclTerm) :-
    DeclTerm =.. [Functor,Arg1,Type,Pos],
    !,
    replace_this_in_signature_field(Arg1, NArg1),
    NewDeclTerm =.. [Functor,NArg1,Type,Pos].
replace_this_in_signature_field(DeclTerm, NewDeclTerm) :-
    DeclTerm =.. [Functor,Arg1,Arg2,Type,Pos],
    !,
    replace_this_in_signature_field(Arg1, NArg1),
    replace_this_in_signature_field(Arg2, NArg2),
    NewDeclTerm =.. [Functor,NArg1,NArg2,Type,Pos].
replace_this_in_signature_field(DeclTerm, DeclTerm).

% Translate TFieldID: Declaration
fun_or_pred_decl_special_cases(setof(Expr,_,_), TFieldID, Pos, subset(Pos,TFieldID,TExpr), set) :-
    translate_e_p(Expr, TExpr).
fun_or_pred_decl_special_cases(oneof(Expr,_,_), TFieldID, Pos, subset(Pos,TFieldID,TExpr), singleton) :- % Default
    translate_e_p(Expr, TExpr).
fun_or_pred_decl_special_cases(loneof(Expr,_,_), TFieldID, Pos, conjunct(Pos,subset(Pos,TFieldID,TExpr),ExtraPred), set) :-
    % x : lone S  ->  x <: S & card(x) <= 1
    ExtraPred = less_equal(Pos,card(Pos,TFieldID),integer(Pos,1)),
    translate_e_p(Expr, TExpr).
fun_or_pred_decl_special_cases(someof(Expr,_,_), TFieldID, Pos, conjunct(Pos,subset(Pos,TFieldID,TExpr),ExtraPred), set) :-
    % x : some S  ->  x <: S & x \= {}
    ExtraPred = not_equal(Pos,TFieldID,empty_set(Pos)),
    translate_e_p(Expr, TExpr).
fun_or_pred_decl_special_cases(Expr, TFieldID, Pos, Translation, IsSingleton) :-
    % treat identifier, cartesian, union, ...  use translate of TFieldID in Expr
    translate_binary_in(TFieldID, Expr, Pos, Translation),
    get_type_and_arity_from_alloy_term(Expr, _Type, Arity),
    !,
    (   Arity = 1 ->
        IsSingleton = singleton
    ;   IsSingleton = set
    ).
fun_or_pred_decl_special_cases(Term, _, Pos, falsity(Pos), set) :-
    add_error(alloy2b, 'Field declaration for function or predicate not implemented:', Term, Pos).

alloy_to_b_operator(Op, BOp) :-
    alloy_to_b_unary_operator(Op, BOp),
    !.
alloy_to_b_operator(Op, BOp) :-
    alloy_to_b_binary_operator(Op, BOp),
    !.
alloy_to_b_operator(Op, Op) :-
    format('~n~nTRANSLATING ~w to B without modification!~n', [Op]).

alloy_to_b_unary_operator(not, negation).
alloy_to_b_unary_operator(closure1, closure).
alloy_to_b_unary_operator(closure, reflexive_closure).
alloy_to_b_unary_operator(inverse, reverse). % works for binary relation only; indeed Alloy says: ~ can be used only with a binary relation.


alloy_to_b_binary_operator(in, subset).
alloy_to_b_binary_operator(not_in, not_subset).
alloy_to_b_binary_operator(plus, union).
alloy_to_b_binary_operator(or, disjunct). % actually treated as unary operator
alloy_to_b_binary_operator(and, conjunct). % actually treated as unary operator
alloy_to_b_binary_operator(minus, set_subtraction).
% alloy_to_b_binary_operator(cartesian,cartesian_product). % only works for  X->UNARY !
alloy_to_b_binary_operator(function, expression_definition).
alloy_to_b_binary_operator(predicate, predicate_definition).
alloy_to_b_binary_operator(not_equal, not_equal).
alloy_to_b_binary_operator(dom_restr, domain_restriction). % not ok for ternary, set must be unary
alloy_to_b_binary_operator(ran_restr, range_restriction). % also ok for ternary, ... relations
alloy_to_b_binary_operator(equal, equal).
alloy_to_b_binary_operator(conjunct, conjunct).
alloy_to_b_binary_operator(intersection, intersection).
alloy_to_b_binary_operator(implication, implication).
alloy_to_b_binary_operator(iff, equivalence).
alloy_to_b_binary_operator(override, overwrite). % ++

is_cartesian_product_term(Term, Arg1, Arg2, POS) :-
    functor(Term, Functor, 4),
    is_cartesian_product(Functor),
    arg(1, Term, Arg1),
    arg(2, Term, Arg2),
    arg(4, Term, POS).
is_cartesian_product(cartesian).


% predicates with integer arguments
alloy_to_b_integer_operator(less, less).
alloy_to_b_integer_operator(greater, greater).
alloy_to_b_integer_operator(less_equal, less_equal).
alloy_to_b_integer_operator(greater_equal, greater_equal).


%%%
% Accumulate the translated machine parts and signature names during the translation and build the machine AST afterwards.
% We may need the signature names later on if translating a global scope.
build_machine_ast(b_machine(ListOfMachineParts,_SignatureNames), machine(generated(none,AbstractMachine))) :-
    % filter empty machine parts
    findall(MachinePart, (   member(MachinePart, ListOfMachineParts),
                             MachinePart =.. [_,_,L],
                             L \== []
                         ), NonEmptyMachineSections),
    % properties need to be conjoined
    (   select(properties(none,L), NonEmptyMachineSections, RestListOfUsedMachineParts) ->
        true
    ;   L == [],
        RestListOfUsedMachineParts = NonEmptyMachineSections
    ),
    join_predicates(conjunct, L, FlatL),
    AbstractMachine = abstract_machine(none,machine(none),
                                       machine_header(none,alloytranslation,[]),
                                       [properties(none,FlatL)|RestListOfUsedMachineParts]),
    !.

empty_machine_acc(b_machine([sets(none,[]),constants(none,[]),
                             definitions(none,[]),properties(none,[]),
                             assertions(none,[]),operations(none,[])],[])).

% extend a machine section with a give list of conjuncts
extend_machine_with_conjuncts(_Section, [], Acc, Res) :-
    !,
    Res = Acc.
extend_machine_with_conjuncts(Section, Conjuncts, MAcc1, MAcc2) :-
    join_untyped_ast_nodes(conjunct, Conjuncts, TPred),
    extend_machine_acc(Section, MAcc1, TPred, MAcc2).

extend_machine_acc(signatures, b_machine(MachineParts,SignatureNames), New,
                   b_machine(MachineParts,NewSignatureNames)) :-
    !,
    append(SignatureNames, New, TSignatureNames),
    remove_dups(TSignatureNames, NewSignatureNames).
extend_machine_acc(Functor, b_machine(MachineParts,SignatureNames), New,
                   b_machine([NewMachinePart|RestMachineParts],SignatureNames)) :-
    MachinePart =.. [Functor,none,List],
    select(MachinePart, MachineParts, RestMachineParts),
    append(List, [New], NewList),                        % keeps order, but is inefficient!
    NewMachinePart =.. [Functor,none,NewList],
    !.
extend_machine_acc(Functor, _, _, _) :-
    add_internal_error('Failed: ', extend_machine_acc(Functor,_,_,_)),
    fail.

get_signature_names_from_machine_acc(b_machine(_MachineParts,SignatureNames), SignatureNames).

get_command_counter_atom(CommandCounterAtom) :-
    command_counter(CommandCounter),
    number_codes(CommandCounter, CommandCounterCodes),
    atom_codes(CommandCounterAtom, CommandCounterCodes),
    retractall(command_counter(_)),
    NewCommandCounter is CommandCounter+1,
    asserta(command_counter(NewCommandCounter)).

finalize_extending_signatures(MAcc, NewMAcc) :-
    extending_signatures(List),
    !,
    finalize_extending_signatures_aux(MAcc, List, NewMAcc).
finalize_extending_signatures(MAcc, MAcc).

finalize_extending_signatures_aux(MAcc, [], MAcc).
finalize_extending_signatures_aux(MAcc, [(Parent,SubSigs)|T], NewMAcc) :-
    length(SubSigs, AmountOfSubSigs),
    AmountOfSubSigs > 1,
    !,
    maplist(translate_sub_sig, SubSigs, TSubSigs),
    translate_e_p(Parent, TParent),
    extend_machine_acc(properties, MAcc, partition(none,TParent,TSubSigs), MAcc2),
    % equivalent to
    % parent = sub1 \/ sub2 \/ ... \/ subN   &  sub1 /\ sub2 = {} &  ...
    findall(OneSub, member(OneSub/one, SubSigs), OneSubSigs),
    length(OneSubSigs, AmountOneSubSigs),
    (   AmountOneSubSigs = AmountOfSubSigs % we basically have an enumerated set with only singleton elements
    ->
        extend_machine_acc(properties, MAcc2, equal(none,card(none,TParent),integer(none,AmountOneSubSigs)), MAcc3)
    ;   extend_machine_acc(properties, MAcc2, greater_equal(none,card(none,TParent),integer(none,AmountOneSubSigs)), MAcc3)
    ),
    finalize_extending_signatures_aux(MAcc3, T, NewMAcc).
finalize_extending_signatures_aux(MAcc, [_|T], NewMAcc) :-
    finalize_extending_signatures_aux(MAcc, T, NewMAcc).
%%%
translate_sub_sig(Name/_One, B) :-
    translate_e_p(Name, B).

assert_extending_signature(Name, Parent, One) :-
    (   is_sub_signature_of(Name, Parent, One) ->
        true
    ;   debug_format(19, '~w is sub signature of ~w (~w)~n', [Name,Parent,One]),
        assert(is_sub_signature_of(Name, Parent, One))
    ),
    extending_signatures(List),
    select((Parent,SubSigs), List, Rest),
    !,
    (   \+ memberchk(Name/One, SubSigs)
           ->
        retractall(extending_signatures(_)),
        asserta(extending_signatures([(Parent,[Name/One|SubSigs])|Rest]))
    ;   true
    ).
assert_extending_signature(Name, Parent, One) :-
    retract(extending_signatures(List)),
    !,
    asserta(extending_signatures([(Parent,[Name/One])|List])).
assert_extending_signature(Name, Parent, One) :-
    asserta(extending_signatures([(Parent,[Name/One])])).

assert_enumerated_set(Options, Name) :-
    % singleton top-level signatures are defined as an enumerated set like S = {_S}
    member(one, Options),
    % but extending singleton signatures are defined as constants and singleton subsets of their parent type
    \+ member(subset(_), Options),
    \+ member(subsig(_), Options),
    !,
    asserta(enumerated_set(Name)).
assert_enumerated_set(_, _).

assert_singleton_set(Options, Name) :-
    member(one, Options),
    !,
    assert_singleton_set(Name).
assert_singleton_set(_, _).

assert_singleton_set(Name) :-   % debug_format(19,'Singleton set : ~w~n',[Name]),nl,
    % \+ singleton_set(Name),
    asserta(singleton_set(Name)),
    !.
% assert_singleton_set(_).
retract_singleton_set(BPOS, Name) :-
    (   retract(singleton_set(Name)) ->
        true
    ;
        add_message(retract_singleton_set, 'WARNING: no singleton_set fact for: ', Name, BPOS) % can be ok for non-singleton ids
    ).
% retractall(singleton_set(Name)).
retract_state :-
    retractall(sig_field_id(_, _)),
    retractall(artificial_parent_type(_, _)),
    retractall(current_module_in_scope(_)),
    retractall(module_in_scope_but_root(_, _)),
    retractall(enumerated_set(_)),
    retractall(current_command(_)),
    retractall(command_counter(_)),
    asserta(command_counter(0)),
    retractall(singleton_set(_)),
    retractall(ordered_signature(_)),
    retractall(total_function(_, _)),
    retractall(extending_signatures(_)),
    retractall(is_seq(_)),
    retractall(maxseq(_)),
    retractall(is_sub_signature_of(_, _, _)),
    !.

% an identifier which in B refers not to a set but an element of a set, representing a singleton set in Alloy
is_singleton_set_id(this).      % this is always singleton (TO DO: check this ;-)
is_singleton_set_id(IDName) :-
    singleton_set(IDName).

assert_total_function(identifier(_,Name), SIG) :-
    !,
    assert_total_function(Name, SIG).
assert_total_function(Name, identifier(_,SIG)) :-
    !,
    assert_total_function(Name, SIG).
assert_total_function(Name, SIG) :-
    debug_format(19, 'Total function ~w over ~w~n', [Name,SIG]),
    asserta(total_function(Name, SIG)).

is_total_function(predecessor(_), _). % translation of prev
is_total_function(successor(_), _). % translation of next
is_total_function(function(_,ID,_), ArgType) :-
    !,
    is_total_function(ID, ArgType).
is_total_function(identifier(_,Name), ArgType) :-
    !,
    is_total_function(Name, ArgType).
is_total_function(Name, [ArgType]) :-
    total_function(Name, FunType), % TO DO: check if FunctionType is Superset of ArgType
    (   strip_singleton_set(FunType, SFunType),
        SFunType = identifier(_,ArgType)
     -> true
     ;  ArgType = FunType -> true
     ;  is_sub_signature_of(ArgType, FunType, _)
    ).

assert_ordered_sig_name(Name) :-
    asserta(ordered_signature(Name)).

is_ordered_signature(IDName) :-
    ordered_signature(IDName).
is_ordered_signature('ordering\'').

is_univ(identifier('univ',_,_)). % universe
is_univ(identifier(_,'univ')).

is_iden(iden(_)).
is_iden(event_b_identity(_)).

is_unary_relation(AlloyTerm) :-
    get_type_and_arity_from_alloy_term(AlloyTerm, _Type, Arity),
    Arity == 1.

is_binary_relation(AlloyTerm) :-
    get_type_and_arity_from_alloy_term(AlloyTerm, _Type, Arity),
    Arity == 2.

can_remove_singleton_set(set_extension(_,[Element]), Element).
% remove_singleton_set(AST,AST).
get_type_and_arity_from_alloy_term(integer(_,_), Type, Arity) :-
    !,
    Type = integer,
    Arity = 1.                  % will get translated to singleton set
get_type_and_arity_from_alloy_term(and(_,_), Type, Arity) :-
    !,
    Type = [untyped],
    Arity = 0.                  % predicate
get_type_and_arity_from_alloy_term(or(_,_), Type, Arity) :-
    !,
    Type = [untyped],
    Arity = 0.                  % predicate
get_type_and_arity_from_alloy_term(iden(_), Type, Arity) :-
    !,
    Type = univ,
    Arity = 2.
get_type_and_arity_from_alloy_term(AlloyTerm, Type, NArity) :-
    AlloyTerm =.. [_|Args],
    member(type(Type,Arity), Args),
    (   is_primitive_type(Type, NArity)
    ;
        (   length(Type, LType),
            LType == Arity,
            NArity = Arity
        )
    ),
    !.
get_type_and_arity_from_alloy_term(AlloyTerm, Type, Arity) :-
    add_warning(alloy2b, 'Could not extract type and arity from Alloy term:', AlloyTerm),
    Type = [untyped],
    Arity = 0.                  % like predicate

% TO DO: maybe there are more primitive types?
% primitive types have arity zero in Alloy for some reason
is_primitive_type(['PrimitiveBoolean'], 1).

% atom_or_identifier_term(ID,ID) :- atom(ID).
% atom_or_identifier_term(identifier(IDName,_,_),IDName).
join_predicates(Op, TArgList, TBinaryP) :-
    alloy_to_b_operator(Op, BOp),
    (   BOp = conjunct
    ;   BOp = disjunct
    ),
    join_untyped_ast_nodes(BOp, TArgList, TBinaryP).

join_untyped_ast_nodes(_, [], truth(none)).
join_untyped_ast_nodes(BOp, [H|T], Conjoined) :-
    join_untyped_ast_nodes(BOp, T, H, Conjoined).

join_untyped_ast_nodes(_, [], Acc, Acc).
join_untyped_ast_nodes(BOp, [H|T], Acc, Conjoined) :-
    NewAcc =.. [BOp,none,Acc,H],
    join_untyped_ast_nodes(BOp, T, NewAcc, Conjoined).


:- begin_tests(full_machine_translation).

test(cards, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/cards.pl').

test(filesystem, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/filesystem.pl').

test(filesystem2, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/filesystem2.pl').

test(self_grandpas, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/selfgrandpas.pl').

test(queens2, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/queens2.pl').

test(river_crossing, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/rivercrossingpuzzle.pl').

test(enum_test, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/enumtest.pl').

test(graphiso, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/graphiso.pl').

test(disj_field_test, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/disjfieldtest.pl').

test(crewAlloc, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/crewalloc.pl').

test(sudoku1, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/sudoku1.pl').

test(sequence_test, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/sequence_test.pl').

test(http, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/http.pl').

test(knights, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/knights_and_knaves.pl').

test(hanoi, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/hanoi.pl').

test(einstein, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/einstein.pl').

test(join_binary, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/joinbinary.pl').

test(slot_data, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/slot_data_v2.pl').

test(job_puzzle, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/job_puzzle.pl').

test(natural, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/natural.pl').

/*
   test(family_v1,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/family-v1.pl').

   test(hanoi,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/hanoi.pl').

   test(knights,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/knights_and_knaves.pl').

   test(directctrl,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/directctrl.pl').

   test(views,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/views.pl').

   test(mutex,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/mutex.pl').
 */
:- end_tests(full_machine_translation).
