% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(ast_inspector,[tcltk_get_node_info/6]).

:- use_module(library(codesio)).
:- use_module(library(lists)).
:- use_module(bmachine).
:- use_module(bsyntaxtree).

:- use_module(module_information,[module_info/2]).
:- module_info(group,visualization).
:- module_info(description,'This module provides the logic behind the AST inspector debugging view in Tcl').

% first argument is a path : top-level root is empty path [], afterwards we have numbers representing the number of the child chosen
tcltk_get_node_info([],'','','','',NSubs) :-
    !,top_level_astnodes(Nodes),length(Nodes,NSubs).
tcltk_get_node_info([TopIndex|RIndex],write(Text),write(Type),write(Infos),write(Quant),NSubs) :-
    top_level_astnodes(Tops),
    nth0(TopIndex,Tops,Top),
    lookup_astnode(Top,RIndex,info(Text,Type,Infos,Quant),NSubs),!.
tcltk_get_node_info(_Index,'Failure when trying to get node info','','','',0).

top_level_astnodes(Tops) :- findall(T,top_level(T),Tops).

top_level(properties) :- b_get_properties_from_machine(P),\+ is_truth(P).
top_level(static_assertions) :- b_get_static_assertions_from_machine([_|_]).
top_level(dynamic_assertions) :- b_get_dynamic_assertions_from_machine([_|_]).
top_level(invariant).
top_level(operations).

astnode_info(properties,Info,[P]) :-
    !,text_info('Properties',Info),b_get_properties_from_machine(P).
astnode_info(invariant,Info,[I]) :-
    !,text_info('Invariant',Info),b_get_invariant_from_machine(I).
astnode_info(operations,Info,Ops) :-
    !,text_info('Operations',Info),
    findall(op(Name),b_is_operation_name(Name),Ops).
astnode_info(op(Name),Info,Result) :-
    !,b_get_machine_operation(Name,Returns,Parameters,_),
    ( Returns   =[] -> Result=PRes; Result=[returns(Name)   |PRes]),
    ( Parameters=[] -> PRes  =Subs; PRes  =[parameters(Name)|Subs]),
    Subs = [subst(Name)],text_info(Name,Info).
astnode_info(returns(Name),info('Returns','','',Ids),TIds) :-
    !,b_get_machine_operation(Name,TIds,_,_),get_texpr_ids(TIds,Ids).
astnode_info(parameters(Name),info('Parameters','','',Ids),TIds) :-
    !,b_get_machine_operation(Name,_,TIds,_),get_texpr_ids(TIds,Ids).
astnode_info(subst(Name),Info,[Subst]) :-
    !,text_info('Substitution',Info), b_get_machine_operation(Name,_,_,Subst).
astnode_info(static_assertions,Info,Ass) :-
    !,text_info('Static Assertions',Info),b_get_static_assertions_from_machine(Ass).
astnode_info(dynamic_assertions,Info,Ass) :-
    !,text_info('Dynamic Assertions',Info),b_get_dynamic_assertions_from_machine(Ass).
astnode_info(TExpr,info(Node,Type,Infos,Ids),Subs) :-
    create_texpr(Expr,Type,Infos,TExpr),!,
    safe_syntaxelement(Expr,Subs,Names,_,Constant),
    ( Names = [] -> Ids=''
    ; otherwise  -> get_texpr_ids(Names,Ids)),
    functor(Expr,Functor,_),
    ( Constant=[] ->
        Node=Functor
    ; otherwise ->
        functor(Node,Functor,1),
        arg(1,Node,Constant)
    ).
astnode_info(TExpr,info('INVALID NODE','',TExpr,''),[]).

text_info(Text,info(Text,'','','')).

lookup_astnode(Top,Index,Info,N) :-
    astnode_info(Top,TopInfo,Subs),
    lookup_subnode(Index,Subs,TopInfo,Info,N).

lookup_subnode([],Subs,Info,Info,N) :-
    length(Subs,N).
lookup_subnode([H|T],Subs,_TopInfo,Info,N) :-
    nth0(H,Subs,Sub),
    lookup_astnode(Sub,T,Info,N).
