% (c) 20019-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(b_enumeration_order_analysis, [find_do_not_enumerate_variables/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module computes DO_NOT_ENUMERATE variables.').

:- use_module(error_manager).
:- use_module(debug).
:- use_module(self_check).
:- use_module(bsyntaxtree).

:- use_module(library(lists)).
:- use_module(library(ordsets)).

:- use_module(library(ugraphs),[vertices_edges_to_ugraph/3 %, top_sort/2
]).

% try and detect variables which do not have to be enumerated:
% Example: {x,y,z|z:1..10 & y=z*z & #v.(v:{0,1} & x=y+v)}
% Here we want to annotate x and y as DO_NOT_ENUMERATE: y can be computed from z, x can be computed from y and v
find_do_not_enumerate_variables(TIds,Body,VariablesToAnnotate) :-
     get_texpr_ids(TIds,Ids),sort(Ids,SIds),
     % format('Detecting lambda_result / DO_NOT_ENUMERATE ~w : ',[SIds]),translate:print_bexpr(b(comprehension_set(TIds,Body),any,[])),nl,
     recursive_conjunction_to_list(Body,SIds,LBody),
     %translate:l_print_bexpr_or_subst(LBody),nl,
     % first find equalities of the form ID = Expr(UsedIds) with ID not in UsedIds
     find_lambda_equalities(LBody,SIds,Equalities,Rest),
     Equalities \= [],
     Equalities \= [equality('_lambda_result_',_,_)],
     % find the identifiers used in the the rest predicate
     % any variable used in the rest cannot be marked as DO_NOT_ENUMERATE
     find_identifier_uses_l(Rest,[],SortedRestIds),
     %format('Detecting lambda result over ~w with rest ids = ~w~n',[SIds,SortedRestIds]),
     remove_illegal_equalities(Equalities,SortedRestIds,RemainingEqualities,ResultUsedIds),
     RemainingEqualities \= [],
     RemainingEqualities \= [equality('_lambda_result_',_,_)],
     findall(Vertice,member(equality(Vertice,_,_),RemainingEqualities),Vs),
     sort(Vs,SortedVs),
     debug_format(19,'Topological sorting of Vertices = ~w, rest ids=~w~n',[SortedVs,ResultUsedIds]),
     findall(V-W,(member(V,SortedVs),
                  member(equality(W,UsedIds,_),RemainingEqualities),
                  member(V,UsedIds)), Edges),
     %format('Edges = ~w~n',[Edges]),
     vertices_edges_to_ugraph(SortedVs,Edges,Graph),
     topological_sorting(Graph,OrderedVertices),
     OrderedVertices \= ['_lambda_result_'],
     !,
     sort(OrderedVertices,SOV), % vertices were sorted according to enumeration order
     ord_intersection(SOV,SIds,VariablesToAnnotate). % retain only those ids which are freshly quantified for annotation
find_do_not_enumerate_variables(_,_,[]). % return empty list of variables to annotate

:- use_module(tools_printing,[format_with_colour/4]).
topological_sorting(Graph,OrderedVertices) :-
     top_sort(Graph,OrderedVertices,UnsortedNr),
     % allow cycles but do only return the vertices that can be topologically sorted
     % {x,y,z|z:1..10 & x=z+1 & y=x+1 & x=y-1} -> cycle over x and y
     % {x,y,r,z|z:1..10 & x=z+1 & y=x+1 & x=y-1 & r=z*z}  -> cycle over x and y, but r can be annotated
     % see also test 1824
     !,
     debug_format(19,'Topological Sorting successful = ~w (unsorted due to cycles: ~w)~n',[OrderedVertices,UnsortedNr]).
topological_sorting(Graph,_) :-
     format_with_colour(user_output,[red,bold],'Topological sorting failed for ~w~n~n',[Graph]),
     fail.

% detect assignments of the Form ID = Expr, where ID does not occur in Expr and occurs in SIds
is_lambda_equality(_SIds,Equality,ID,UsedIDs) :-
    b_ast_cleanup:identifier_equality(Equality,ID,_TID,TExpr), % TO DO: move to separate module
    %ID \= '_lambda_result_', % include as equality, IDs in TExpr should not count
    %ord_member(ID,SIds), % if we find lambda equalities for other identifiers not in SIds it is useful to consider them too
    % Example: {x,y,z|z:1..10 & #(w,v).(x=z+z &w:1..9 & v=card({x,x+x,w}) & y=v+v )}
    % When analyzing #(w,v) it is also useful to consider the equalities over x and y in the process; otherwise we are not able to detect that v should not be enumerated
    % example is also identifier NV in operation caval__rule__1__compute in private_examples/ClearSy/2019_Nov/rule_perf4
    find_identifier_uses(TExpr,[],UsedIDs),
    \+ ord_member(ID,UsedIDs).


% separate predicates into lambda equality candidates ID = Expr and other predicates
% find_lambda_equalities(ListOfPreds,QuantifiedIDs,LambdaEqualities,ListOfRest)
find_lambda_equalities([],_SIds,[],[]).
find_lambda_equalities([TPred|Rest],SIds,A,B) :-
    (is_lambda_equality(SIds,TPred,ID,UsedIDs)
      -> A=[equality(ID,UsedIDs,TPred)|AR], B=BR
         %, format('Equality ~w  = Expr(~w)~n',[ID,UsedIDs])
      ; pred_can_be_ignored(TPred)
       -> A=AR, B=BR
        ; A=AR, B=[TPred|BR]),
    find_lambda_equalities(Rest,SIds,AR,BR).

% predicate can be safely ignored for enumeration order analysis:
pred_can_be_ignored(b(E,_,_)) :- pred_can_be_ignored_aux(E).
pred_can_be_ignored_aux(external_pred_call(EP,_)) :- ext_pred_can_be_ignored(EP).
ext_pred_can_be_ignored('DO_NOT_ENUMERATE').
ext_pred_can_be_ignored('ENUMERATE').
ext_pred_can_be_ignored('FORCE').
ext_pred_can_be_ignored('printf'). % TO DO: debugging_only fact in external_functions
ext_pred_can_be_ignored('fprintf').
ext_pred_can_be_ignored('printf1').
ext_pred_can_be_ignored('printf_opt_trace').
ext_pred_can_be_ignored('iprintf').
ext_pred_can_be_ignored('printf_two').
ext_pred_can_be_ignored('tprintf').
ext_pred_can_be_ignored('printf_nonvar').
ext_pred_can_be_ignored('observe').
ext_pred_can_be_ignored('tobserve').
ext_pred_can_be_ignored('observe_fun').


remove_illegal_equalities(Equalities,RestUsedIds,RemainingEqualities,ResultUsedIds) :-
     select(equality(ID,UsedIDs,_),Equalities,RestEqualities),
     ord_member(ID,RestUsedIds),
     %format('Removing equality over ~w, id used in rest ~w~n',[ID,RestUsedIds]),
     !,
     ord_add_element(UsedIDs,ID,UsedIDs2),
     ord_union(RestUsedIds,UsedIDs2,NewRestUsedIds),
     remove_illegal_equalities(RestEqualities,NewRestUsedIds,RemainingEqualities,ResultUsedIds).
remove_illegal_equalities(Eq,Used,Eq,Used).


% a version of conjunction to list which lifts existential quantifier bodies to the top-level
recursive_conjunction_to_list(C,BoundIds,L) :- var(C),!,
   add_error_fail(conjunction_to_list,'Internal error: var :',conjunction_to_list(C,BoundIds,L)).
recursive_conjunction_to_list(C,BoundIds,List) :-
    rec_conjunction_to_list2(C,BoundIds,List,[]).
rec_conjunction_to_list2(X,BoundIds,I,O) :- X=b(E,_,_),
   (E=conjunct(LHS,RHS) -> rec_conjunction_to_list2(LHS,BoundIds,I,Inter),
                           rec_conjunction_to_list2(RHS,BoundIds,Inter,O)
    ; E = truth -> I=O
    ; E = exists(Ids,Body),
      check_clash(Ids,BoundIds,NewBoundIds)
       -> rec_conjunction_to_list2(Body,NewBoundIds,I,O)
    ; E = let_predicate(Ids,Exps,Body),
      check_clash(Ids,BoundIds,NewBoundIds) -> % ditto
                I = [b(let_predicate(Ids,Exps,b(truth,pred,[])),pred,[]) | Inter],
                rec_conjunction_to_list2(Body,NewBoundIds,Inter,O)
    ; I = [X|O]).

% check whether the new identifiers clash
% Without this check we can get errors: See test 983, 1380 prop_COMPUTE_5, esatrack  (detected by check_typed_ids in bsyntaxtree.pl)
% TO DO: we could rename the identifiers for the analysis; but complex and is it worth it?
check_clash(TIds,BoundIds,NewBoundIds) :- def_get_texpr_ids(TIds,Ids),
    sort(Ids,SIds),
    (ord_intersection(SIds,BoundIds,Clash), Clash \= []
      -> debug_format(19,'Not expanding quantifier for enumeration analysis due to clash with ~w~n',[Clash]),
         fail
      ;  ord_union(SIds,BoundIds,NewBoundIds)). % we could use ord_disjoint_union

% tests were detection works: 790, 799, 982, 1302, 1303, 1380, 1394, 1492, 1751
% 1751 slightly slower
% 1307 639
% 1935, 1936, 1946, 1950, 1981, 1983

% TO DO:
%  can the analysis be counter-productive??
%   what about  {x | #v.(v:1..1000000 &  x = bool(v=0))} in CLPFD FALSE mode
%      here {x | (x=TRUE or x=FALSE) & #v.(v:1..1000000 &  x = bool(v=0))} is faster 2 vs 3.5 seconds
%
%  more refined analysis, to detect issue in perf_3264/rule_186.mch
% block:dom(ic___OPS_SDS_3264_OTHER_DP_IN_BLOCK |> {FALSE}) is enumerated before
%  signal :  ran(SORT(dom(ic___OPS_SDS_3264_BBS))/|\186)
%  bad because from signal we can derive block, from the predicate
%      block : ic___OPS_SDS_3264_BBS(signal)




:- use_module(library(avl)).

% a slight variation of top_sort from library(ugraphs) which allows the sorting to be not completed
% top_sort(G,S,0) corresponds to top_sort in ugraphs
top_sort(Graph, Sorted,UnsortedNr) :-
    ugraphs:fanin_counts(Graph, Counts),
    ugraphs:get_top_elements(Counts, Top, 0, I),
    ord_list_to_avl(Counts, Map),
    top_sort_aux(Top, I, Map, Sorted,UnsortedNr).

top_sort_aux([], I, _, [], UnsortedNr) :- UnsortedNr = I.
top_sort_aux([V-VN|Top0], I, Map0, [V|Sorted],UnsortedNr) :-
    ugraphs:dec_counts(VN, I, J, Map0, Map, Top0, Top),
    top_sort_aux(Top, J, Map, Sorted,UnsortedNr).

