% (c) 2013-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(disprover, [disprove/5, disprove_with_opts/6]).

:- use_module(module_information).
:- module_info(group,cbc).
:- module_info(description,'Rodin Prover / Disprover').

% provide access to the Disprover algorithm with all hypotheses, selected hypotheses and goal

:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2]).

:- use_module(solver_interface, [solve_predicate/5,type_check_in_machine_context/2]).
:- use_module(preferences, [temporary_set_preference/3, reset_temporary_preference/2,
                            get_preference/2]).
:- use_module(bsyntaxtree, [conjunct_predicates/2, create_negation/2,
                            find_typed_identifier_uses/3,
                            find_typed_identifier_uses_l/3,
                            get_texpr_types/2]).
:- use_module(error_manager, [reset_errors/0,add_warning/3, add_internal_error/2]).
:- use_module(library(lists)).
:- use_module(tools,[start_ms_timer/1, stop_ms_timer/1,stop_ms_timer_with_msg/2]).
:- use_module(debug, [debug_println/2,debug_mode/1, debug_format/3]).
:- use_module(b_global_sets,[contains_unfixed_deferred_set/1]).
:- use_module(external_functions,[observe_state/1]).
:- use_module(unsat_cores,[unsat_core/2]).

disprove(Goal,AllHypotheses,SelectedHypotheses,Timeout,OutResult) :-
    % default options:
    Opts = [use_smt_mode/true,use_clpfd_solver/true,use_chr_solver/true],
    % Note: disprover_mode is always set to true
    disprove_with_opts(Goal,AllHypotheses,SelectedHypotheses,Timeout,Opts,OutResult).

% ----------

disprove_with_opts(Goal,AllHypotheses,SelectedHypotheses,Timeout,Opts,OutResult) :-
    check_options(Opts),
    temporary_set_preference(disprover_mode,true,DC),
    temporary_set_preference(time_out,Timeout,TC),
    start_ms_timer(Timer),
    user_interruptable_call_det(disprove1(Goal,AllHypotheses,SelectedHypotheses,Opts,Result),InterruptResult),
    %write(ir(InterruptResult)), nl,
    (InterruptResult = interrupted -> OutResult = interrupted ; OutResult = Result),
    %catch_internal_timeout(InterruptResult,Result,OutResult),
    (debug_mode(on) -> write(disprove_result(OutResult)),nl,stop_ms_timer(Timer),nl ; true),
    reset_temporary_preference(disprover_mode,DC),
    reset_temporary_preference(time_out,TC).

check_options([]).
check_options([O|T]) :- !, (valid_option(O) -> true ; add_warning(disprover,'Illegal disprover option: ',O)),
                         check_options(T).
check_options(O) :- add_internal_error('Illegal disprover options: ',check_options(O)).

:- use_module(solver_interface, [valid_solver_option/1]).
% Valid options
valid_option(disprover_option(X)) :- !,valid_disprover_option(X).
valid_option(X) :- valid_solver_option(X).

valid_disprover_option(observe_state).
valid_disprover_option(unsat_core).
valid_disprover_option(unsat_core_as_machine).
valid_disprover_option(export_po_as_machine(_)).
valid_disprover_option(wd_prover_timeout(_)).

% ----------

:- use_module(error_manager,[add_all_perrors/3]).
disprove1(Goal,AllHypotheses,SelectedHypotheses,Opts,Result) :-
    on_exception(type_errors(E),
          disprove2(Goal,AllHypotheses,SelectedHypotheses,Opts,Result),
          (add_all_perrors(E,[],disprover), Result=type_error)).

%catch_internal_timeout(ok,contradiction_found,Out) :-
%    error_manager:check_warning_occured(warning(_X),Y),
%    is_time_out_message(Y),  % Sebastian: why do we catch this ?? This is not necessarily a problem ??
%    !,                       % Might be a problem for proofs, invalidate only if Result = contradiction_found
%    Out = time_out.          % However, this might be old code that is superseded by the enum warnings
%catch_internal_timeout(ok,Result,Result).

% is_time_out_message(Y) :- atom(Y), atom_codes(Y,Codes),
%    append("TIME-OUT occurred while ProB was expanding", _, Codes).

solve_disprover_predicate(Predicate,Opts,CBCResult) :-
   TimeoutFactor = 1,
   exclude(disprover_option,Opts,SOpts),
   (member(disprover_option(observe_state),Opts) -> observe_state(State) ; true),
   solve_predicate(Predicate,State,TimeoutFactor,SOpts,CBCResult).
disprover_option(disprover_option(_)).


:- use_module(wdsrc(well_def_analyser),[prove_sequent/3]).
:- use_module(tools_meta,[safe_time_out/3]).
prove_using_wd_prover(Goal,Hypotheses,TimeOutMs,contradiction_found) :-
   % quite often 0 or 1 ms are enough for TimeOutMs
   debug_format(19,'Starting WD Prover (timeout ~w ms)~n',[TimeOutMs]),
   start_ms_timer(Timer),
   safe_time_out(prove_sequent(proving,Goal,Hypotheses),TimeOutMs,TimeOutRes),
   (TimeOutRes=time_out ->
      format('Timeout in WD Prover (~w ms)~n',[TimeOutMs]),fail ; true),
   (debug_mode(off) -> true ; stop_ms_timer_with_msg(Timer,'WD Prover successful')).


:- public disprove2/5. % it is called by user_interruptable_call_det above; we should also add a meta_predicate annotation
:- use_module(bsyntaxtree, [predicate_identifiers/2]).
disprove2(Goal,AllHypotheses,SelectedHypotheses,Opts,ResultOut) :-
    type_check_in_machine_context([Goal|AllHypotheses],[TGoal|TAllHypotheses]),
    type_check_in_machine_context(SelectedHypotheses,TSelectedHypotheses),
    disprove3(TGoal,TAllHypotheses,TSelectedHypotheses,Opts,ResultOut).

disprove3(TGoal,TAllHypotheses,TSelectedHypotheses,Opts,ResultOut) :-
    % build predicate H1 & H2 .... & not Goal
    create_negation(TGoal,NegatedGoal),
    %(debug_mode(on) -> write(NegatedGoal), nl ; true),
    (debug_mode(on) -> nl,write('Negated Goal:'), nl,
                             translate:nested_print_bexpr_as_classicalb(NegatedGoal),
                             write_hypothesis(TAllHypotheses)
                          ;  true),
    conjunct_predicates([NegatedGoal|TAllHypotheses],Predicate),
    (member(disprover_option(export_po_as_machine(File)),Opts)
      ->  export_po_as_machine(File,TAllHypotheses,TSelectedHypotheses,TGoal)
      ; true),
    %tools:print_bt_message(solving),
    ( get_wd_prover_timeout(Opts,Timeout),
      Timeout>0,
      prove_using_wd_prover(TGoal,TAllHypotheses,Timeout,Result)
      % adds 5 extra PO proven in test 1800, and 4 more in test 1447
      -> true % DISCHARGED by WD Prover
       ; solve_disprover_predicate(Predicate,Opts,Result)
    ),
    %tools:print_bt_message(solver_result(CBCResult)),
    debug_println(10,disprover_result_with_all_hypotheses(Result)),
    disprove_double_check(Result,NegatedGoal,TAllHypotheses,TSelectedHypotheses,Predicate,Opts,ResultOut).

get_wd_prover_timeout(Opts,Timeout) :- member(disprover_option(wd_prover_timeout(Timeout)),Opts),!.
get_wd_prover_timeout(_,WDTimeout) :- get_preference(time_out,GlobalTO),
     WDTimeout is 20 + GlobalTO // 100.

% res <— solve(  All Hyp & not G )
% if res = solution then
%    return CounterExample Found,  Goal definitely not provable
% else if unfixed deferred set used in selected Hypotheses or Goal then
%    return Unknown
% else if res = fail (without time-out & without enumeration warning & no unfixed deferred set in All Hypotheses) then
%    return Proof Found, Goal proven
% else
%    res <— solve (Selected Hyp & not G)    /* maybe use smaller time-out here ?? */
%    if res = fail (without time-out & without enumeration warning) then
%      return Proof Found, Goal proven
%    else if res = solution then
%      return CounterExample for selected Hypotheses found, Goal not provable from selected Hypotheses (but may be provable with all Hyp)
%    else
%      return Unknown
%    fi
%  fi

% we found a solution / counter example. no need to double check here.
% just clear up the resulting state, keeping only variables occuring in the goal.
% although: sometimes it is useful to have other IDs as well, at least in the same component as the Goal !
disprove_double_check(solution(X),NegatedGoal,_Hypotheses,_SelectedHypotheses,_,_,ResultOut) :-
    predicate_identifiers(NegatedGoal,GoalIdentifiers),
    findall(binding(Id,Value,PP), get_binding_from_solution(X,GoalIdentifiers,0,Id,Value,PP), FilteredState),
    ResultOut = solution(FilteredState).
% check if the unfixed deferred sets occur in selected hyps or goal
disprove_double_check(no_solution_found(unfixed_deferred_sets),NegatedGoal,_Hypotheses,SelectedHypotheses,_,_,no_solution_found(unfixed_deferred_sets)) :-
    find_typed_identifier_uses(NegatedGoal,[],GoalIdentifiers),
    get_texpr_types(GoalIdentifiers,GoalTypes),
    find_typed_identifier_uses_l(SelectedHypotheses,[],HypIdentifiers),
    get_texpr_types(HypIdentifiers,HypTypes),
    append(GoalTypes,HypTypes,Types),
    % exclude all types that are not unfixed
    include(contains_unfixed_deferred_set,Types,UnfixedTypes),
    % see if an unfixed type is left
    UnfixedTypes = [_|_],
    debug_println(4,goal_or_sel_hyp_contains_unfixed_deferred_sets(UnfixedTypes)),
    !.
% contradiction was found, i.e. there was neither a timeout nor an enumeration warning
disprove_double_check(contradiction_found,NegatedGoal,Hypotheses,SelectedHypotheses,FullPredicate,Opts,FinalResult) :-
    get_preference(double_evaluation_when_analysing,true)
    -> check_negated(NegatedGoal,Hypotheses,SelectedHypotheses,FullPredicate,Opts,FinalResult)
    ;  FinalResult = contradiction_found.
disprove_double_check(Result,_NegatedGoal,Hypotheses,SelectedHypotheses,_,_Opts,ResultOut) :-
    same_conjunction_list(Hypotheses,SelectedHypotheses),!,
    (debug_mode(on) -> nl,write('Selected hypotheses identical to all hypotheses.'),nl ; true),
    Result=ResultOut.
% see what solving only selected hypotheses & not G gives us
disprove_double_check(_OldResult,NegatedGoal,Hypotheses,SelectedHypotheses,_,Opts,ResultOut) :-
    % No proof found with all hypotheses, now try selected ones only
    % reset existing enumeration warnings,
    % that might have been generated in a previous run
    reset_errors,
    (debug_mode(on)
      -> nl,write('Trying to use only selected hypotheses.\nNegated Goal:'), nl,
         translate:nested_print_bexpr_as_classicalb(NegatedGoal),
         write_hypothesis(SelectedHypotheses)
       ; true),
    conjunct_predicates([NegatedGoal|SelectedHypotheses],Predicate),
    %external_functions:observe_state(State),
    solve_disprover_predicate(Predicate,Opts,Result),
    !,
    disprove_triple_check(Result,NegatedGoal,Hypotheses,SelectedHypotheses,Predicate,Opts,ResultOut).


%  we still have a contradiction -> double check it if preference set
disprove_triple_check(contradiction_found,NegatedGoal,Hypotheses,SelectedHypotheses,FullPredicate,Opts,FinalResult) :-
    (get_preference(double_evaluation_when_analysing,true) % preference set by Rodin Disprover
    -> check_negated(NegatedGoal,Hypotheses,SelectedHypotheses,FullPredicate,Opts,FinalResult)
    ;  FinalResult = contradiction_found,
       compute_unsat_core_if_requested(FullPredicate,Opts)
    ).
disprove_triple_check(solution(X),NegatedGoal,_,_,_FullPredicate,_,ResultOut) :-
    % Counter-Example to Goal for selected Hypotheses found
    predicate_identifiers(NegatedGoal,GoalIdentifiers),
    findall(binding(Id,Value,PP), get_binding_from_solution(X,GoalIdentifiers,10,Id,Value,PP), FilteredState),
    ResultOut = solution_on_selected_hypotheses(FilteredState).
disprove_triple_check(time_out,_,_,_,_,_,time_out). % TODO: We could try Kodkod here!
disprove_triple_check(no_solution_found(X),_,_,_,_,_,no_solution_found(X)). % Or here?

:- use_module(library(ordsets),[ord_member/2]).
% get bindings from solution; showing all identifiers from goal and a few others as indicated by MaxIndexNonGoal
get_binding_from_solution(Solution,GoalIdentifiers,MaxIndexNonGoal,Id,Value,PP) :- MaxIndexNonGoal>0,
    sort(GoalIdentifiers,SG),
    nth1(Nr,Solution,binding(Id,Value,PP)),
    (Nr =< MaxIndexNonGoal -> true ; ord_member(Id,SG)).
    % TO DO: rather than limiting this way, we could analyse the variables which are in the same component
    % either by calling predicate_components or by storing component info in

check_negated(NegatedGoal,Hypotheses,_,FullOrigPredicate,Opts,ResultOut) :-
    create_negation(NegatedGoal,Goal),
    conjunct_predicates([Goal|Hypotheses],Predicate),
    %tools:print_bt_message(check_negated),
    debug_println(20,'Check whether we can also prove negation of goal'),
    solve_disprover_predicate(Predicate,Opts,Result),
    %% tools:print_bt_message(solver_result(CBCResult)),
    !,
    (Result = contradiction_found % proven both (H & not G) and (H & G)
     -> ResultOut = contradiction_in_hypotheses,
        translate:translate_bexpression_with_limit(NegatedGoal,GS),
        add_warning(disprover_inconsistent_hypotheses,
                   'Disprover double check result: Contradiction in hypotheses (or Bug) detected for goal: ',
                    GS), %% check_negated(NegatedGoal,Hypotheses))
        % translate:nested_print_bexpr_as_classicalb(Hypotheses),nl,nl,
        compute_unsat_core_if_requested(Predicate,Opts)
     ;  ResultOut = contradiction_found,
        compute_unsat_core_if_requested(FullOrigPredicate,Opts)
     ).

compute_unsat_core_if_requested(Predicate,Opts) :-
    (member(disprover_option(unsat_core),Opts) -> compute_unsat_core(Opts,Predicate)  ; true).
:- use_module(preferences,[call_with_preference/3]).
:- use_module(translate,[translate_predicate_into_machine/3,nested_print_bexpr_as_classicalb/1]).
compute_unsat_core(Opts,Predicate) :-
    (member(unsat_core_algorithm/P,Opts)
      -> call_with_preference(compute_unsat_core_aux(Predicate,CorePredicate),unsat_core_algorithm,P)
      ; compute_unsat_core(Predicate,CorePredicate)),
    !,
    print('UNSAT CORE: '),nl,print('--------'),nl,
    (member(disprover_option(unsat_core_as_machine),Opts)
      -> translate_predicate_into_machine(CorePredicate,'UnsatCore',ResultAtom),
         write(ResultAtom),nl
       ; nested_print_bexpr_as_classicalb(CorePredicate)),
    nl, print('--------'),nl.

compute_unsat_core_aux(Predicate,CorePredicate) :-
    print('Computing UNSAT CORE'),nl,
    unsat_core(Predicate,CorePredicate). % from unsat_cores

:- use_module(probsrc(bsyntaxtree),[same_texpr/2]).
% check if two lists of conjuncts represent the same conjunction (modulo info fields)
same_conjunction_list(List1,List2) :-
    same_length(List1,List2,_), % avoid possibly expensive comparisons if lists clearly different
    % we could strip info field and the sort; but here we just want to catch the obvious case where the disprover is called with SelectedHyp = AllHyp
    maplist(same_texpr,List1,List2).

:- use_module(translate,[nested_print_sequent_as_classicalb/5,nested_print_bexpr_as_classicalb/1]).
write_hypothesis([]).
write_hypothesis([Hyp|Hyps]) :-
    write(' & /* Hypothesis: */ '), nl,
    %(debug_mode(on) -> write(Hyp), nl ; true),
    nested_print_bexpr_as_classicalb(Hyp),
    write_hypothesis(Hyps).

:- use_module(tools_io,[safe_intelligent_open_file/3]).
export_po_as_machine(Filename,THypothesesList,TSelectedHypothesesList,TGoal) :-
   %nested_print_sequent_as_classicalb(Stream,THypothesesList,TGoal,'ProofObligation_AllHyps'),
   format(user_output,'Exporting selected hypotheses to B machine: ~w~n',[Filename]),
   safe_intelligent_open_file(Filename,write,Stream),
   nested_print_sequent_as_classicalb(Stream,TSelectedHypothesesList,TGoal,
                                             THypothesesList,'ProofObligation_SelectedHyps'),
   close(Stream).

