:- module(dpllt_pred_to_sat, [predicate_to_sat/8,
                              predicate_to_sat/6,
                              sat_to_predicate_from_solution/3,
                              get_amount_of_sat_variables/1,
                              next_sat_var_name/1,
                              reset_sat_var_id/0]).

:- use_module(library(lists)).
:- use_module(probsrc(module_information), [module_info/2]).
:- use_module(probsrc(bsyntaxtree), [remove_all_infos/2,
                                     add_texpr_infos/3,
                                     conjunct_predicates/2,
                                     conjunction_to_list/2,
                                     disjunct_predicates/2,
                                     disjunction_to_list/2]).

:- module_info(group, dpllt).
:- module_info(description,'This module provides the conversion from B predicates to SAT formulae and vice-versa to be used for DPLL(T).').

:- dynamic sat_var_id/1.
:- volatile sat_var_id/1.

sat_var_id(0).

%% get_amount_of_sat_variables(-AmountOfVars).
get_amount_of_sat_variables(AmountOfVars) :-
    sat_var_id(AmountOfVars).

reset_sat_var_id :-
    retractall(sat_var_id(_)),
    asserta(sat_var_id(0)).

%% predicate_to_sat(+Type, +Pred, -NewEnv, -NewPred, -BoolFormula, -NVarAcc).
% Same as predicate_to_sat/8 but initializing an empty environment and accumulator for variables.
predicate_to_sat(Type, Pred, NewEnv, AnnotatedPred, BoolFormula, VarAcc) :-
    predicate_to_sat(Type, [], [], Pred, NewEnv, AnnotatedPred, BoolFormula, VarAcc).

% TO DO: remove AnnotatedPred argument from predicate_to_sat

%% predicate_to_sat(+Type, +VarAcc, +Env, +Pred, -NewEnv, -AnnotatedPred, -BoolFormula, -NVarAcc).
% Create a boolean formula from a B predicate, e.g., return 'A=TRUE & B=TRUE' for 'x:INT & x>1'.
% Introduces fresh unique identifiers if Type is not equal to 'only_reuse'.
% If Type is 'only_reuse': do not introduce new SAT variables but skip parts that need a new SAT variable
% (some unsat core computations might deduce new constraints and we do not want to introduce new SAT variables during CDCL).
% Env stores pairs of B ASTs (without info field) and its assigned SAT variables.
% Add corresponding SAT variable names to AST's info fields (returned in NewPred).
% Each SAT variable contains a term smt_formula/1 in its info field providing the corresponding SMT formula.
% Note: SMT formula is split on the level of conjunction, disjunction, implication and equivalence.
predicate_to_sat(_, Acc, Env, b(truth,pred,Info), EnvOut, Truth, Truth, VarAccOut) :- 
    !,
    EnvOut = Env,
    VarAccOut = Acc,
    Truth = b(truth,pred,Info).
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(conjunct(_,_),_,_),
    conjunction_to_list(Pred, PList),
    map_predicate_to_sat(Type, VarAcc, Env, PList, NewEnv, NPList, SatList, NVarAcc),
    !,
    conjunct_predicates(SatList, BoolFormula),
    conjunct_predicates(NPList, NewPred).
predicate_to_sat(_, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(disjunct(_,_),_,_),
    remove_all_infos(Pred, CPred),
    member((CPred,SatVar), Env),
    % equality from symmetry breaking is a disjunction
    !,
    NewEnv = Env,
    NewPred = CPred,
    BoolFormula = b(equal(SatVar,b(boolean_true,boolean,[])),pred,[]),
    NVarAcc = VarAcc.
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(disjunct(_,_),_,_),
    disjunction_to_list(Pred, PList),
    map_predicate_to_sat(Type, VarAcc, Env, PList, NewEnv, NPList, SatList, NVarAcc),
    !,
    disjunct_predicates(SatList, BoolFormula),
    disjunct_predicates(NPList, NewPred).
predicate_to_sat(Type, VarAcc, Env, b(negation(Pred),pred,_), NewEnv, NewPred, BoolFormula, NVarAcc) :-
    reuse_or_introduce_bool_var(Type, VarAcc, Env, Pred, boolean_false, NewEnv, NewPred, BoolFormula, NVarAcc),
    NewPred \== b(truth,pred,_),
    !.
predicate_to_sat(Type, VarAcc, Env, b(negation(Pred),pred,_), NewEnv, NewPred, BoolFormula, NVarAcc) :-
    predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, TNewPred, BoolFormula, NVarAcc),
    (   TNewPred \== b(truth,pred,_)
    ->  NewPred = b(negation(TNewPred),pred,[])
    ;   NewPred = b(truth,pred,[])
    ).
predicate_to_sat(_, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, VarAccOut) :-
    Pred = b(Functor,pred,Info),
    (   Functor == value(pred_true)
    ;   Functor == value(pred_false)
    ),
    !,
    VarAccOut = VarAcc,
    BoolFormula = b(Functor,pred,Info),
    NewEnv = Env,
    NewPred = Pred.
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(implication(Lhs,Rhs),_,Info),
    map_predicate_to_sat(Type, VarAcc, Env, [Lhs,Rhs], NewEnv, NSmtArgs, NSatArgs, NVarAcc),
    !,
    NSmtArgs = [NSmtLhs,NSmtRhs],
    NewPred = b(implication(NSmtLhs,NSmtRhs),pred,Info),
    NSatArgs = [NSatLhs,NSatRhs],
    BoolFormula = b(implication(NSatLhs,NSatRhs),pred,[smt_formula(Pred)|Info]).
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(equivalence(Lhs,Rhs),_,Info),
    map_predicate_to_sat(Type, VarAcc, Env, [Lhs,Rhs], NewEnv, NSmtArgs, NSatArgs, NVarAcc),
    !,
    NSmtArgs = [NSmtLhs,NSmtRhs],
    NewPred = b(equivalence(NSmtLhs,NSmtRhs),pred,Info),
    NSatArgs = [NSatLhs,NSatRhs],
    BoolFormula = b(equivalence(NSatLhs,NSatRhs),pred,[smt_formula(Pred)|Info]).
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(Node,_,_),
    functor(Node, _, Arity),
    (   Arity == 2
    ;   Arity == 3
    ),
    !,
    reuse_or_introduce_bool_var(Type, VarAcc, Env, Pred, boolean_true, NewEnv, NewPred, BoolFormula, NVarAcc).
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    % formula is a singleton pred which is not splitted such as a subset
    reuse_or_introduce_bool_var(Type, VarAcc, Env, Pred, boolean_true, NewEnv, NewPred, BoolFormula, NVarAcc).
% TO DO: if going into quantifications, we have to watch out for local identifier scopes
%       (i.e., x<7 might not be the same x<7 within a quantification if x is a local id)
% Note that we currently do not go into quantifications so this is not an issue.

map_predicate_to_sat(_, VarAcc, Env, [], Env, [], [], VarAcc).
map_predicate_to_sat(Type, VarAcc, Env, [Pred|T], NewEnv, [NPred|PNT], [BoolFormula|NT], NVarAcc) :-
    predicate_to_sat(Type, VarAcc, Env, Pred, TempEnv, NPred, BoolFormula, TVarAcc),
    map_predicate_to_sat(Type, TVarAcc, TempEnv, T, NewEnv, PNT, NT, NVarAcc).

% Note: BoolFormula is 'a = TRUE' or 'a = FALSE' for (negated) literal a.
reuse_or_introduce_bool_var(_, VarAcc, Env, Pred, BoolAtom, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    remove_all_infos(Pred, CPred),
    member((CPred,SatVar), Env),
    !,
    SatVar = b(identifier(SatVarName),_,_),
    add_texpr_infos(Pred, [sat_var(SatVarName)], NewPred),
    BoolFormula = b(equal(SatVar,b(BoolAtom,boolean,[])),pred,[]),
    NewEnv = Env,
    NVarAcc = VarAcc.
reuse_or_introduce_bool_var(Type, VarAcc, Env, Pred, BoolAtom, [(CleanPred,UniqueId)|Env], NewPred, BoolFormula, NVarAcc) :-
    Type \== only_reuse,
    !,
    next_sat_var_name(SatVarName),
    remove_all_infos(Pred, CleanPred),
    TUniqueId = b(identifier(SatVarName),boolean,[]),
    %find_identifier_uses(Pred, [], UsedIds),
    add_texpr_infos(Pred, [sat_var(SatVarName)], NewPred),
    add_texpr_infos(TUniqueId, [smt_formula(Pred)], UniqueId),
    BoolFormula = b(equal(UniqueId,b(BoolAtom,boolean,[])),pred,[]),
    NVarAcc = [UniqueId|VarAcc].
reuse_or_introduce_bool_var(only_reuse, VarAcc, Env, Pred, _, Env, Pred, b(truth,pred,[]), VarAcc).

next_sat_var_name(SatVarName) :-
    retract(sat_var_id(Old)),
    New is Old + 1,
    asserta(sat_var_id(New)),
    number_codes(New, NC),
    atom_codes(Index, NC),
    atom_concat('sat', Index, SatVarName).

%% sat_to_predicate_from_solution(+SatBindings, +SatVars, -Conjunction).
% Create conjunction of SAT bindings using their corresponding SMT formulae
% which are stored in the bool variables' information fields.
sat_to_predicate_from_solution([], _, b(truth,pred,[])).
sat_to_predicate_from_solution([Binding|T], SatVars, Conjunction) :-
    get_smt_formula_from_binding(Binding, SatVars, SMT, RestSatVars),
    !,
    sat_to_predicate_from_solution(T, SMT, RestSatVars, Conjunction).
sat_to_predicate_from_solution([_|T], SatVars, Conjunction) :-
    % skip: binding might be a variable if instantiation is not required
    sat_to_predicate_from_solution(T, SatVars, Conjunction).

sat_to_predicate_from_solution([], SmtAcc, _, SmtAcc).
sat_to_predicate_from_solution([Binding|T], SmtAcc, SatVars, Conjunction) :-
    get_smt_formula_from_binding(Binding, SatVars, SMT, RestSatVars),
    !,
    NewSmtAcc = b(conjunct(SMT,SmtAcc),pred,[]),
    sat_to_predicate_from_solution(T, NewSmtAcc, RestSatVars, Conjunction).
sat_to_predicate_from_solution([_|T], SmtAcc, SatVars, Conjunction) :-
    sat_to_predicate_from_solution(T, SmtAcc, SatVars, Conjunction).

get_var_and_value_from_binding(bind(SatVarName,Value), SatVarName, Value).
get_var_and_value_from_binding(binding(SatVarName,Value,_), SatVarName, Value).

get_smt_formula_from_binding(Binding, SatVars, SMT, RestSatVars) :-
    ground(Binding),
    get_var_and_value_from_binding(Binding, SatVarName, Value),
    !,
    select(b(identifier(SatVarName),_,Info), SatVars, RestSatVars),
    member(smt_formula(TSMT), Info),!,
    (   Value == pred_true
    ->  SMT = TSMT
    ;   Value == pred_false,
        SMT = b(negation(TSMT),pred,[])
    ).
/*get_smt_formula_from_binding(Binding, _, _, _) :-
    add_error(get_smt_formula_from_binding, 'SAT binding not ground:', [Binding]),
    fail.*/
