:- module(dpllt_solver, [dpllt_solve_predicate/2]).

:- use_module(library(lists)).
:- use_module(library(timeout)).
:- use_module(library(sets), [subtract/3]).
:- use_module(library(clpfd), [fd_var/1, fd_min/2, fd_max/2]).

:- use_module(probsrc('dpllt_solver/symmetry_check/smt_symmetry_breaking')).
:- use_module(probsrc('dpllt_solver/dpllt_preprocessing')).
:- use_module(probsrc('dpllt_solver/dpllt_pred_to_sat')).
:- use_module(probsrc('dpllt_solver/dpllt_sat_solver')).
:- use_module(probsrc('dpllt_solver/difference_logic/difference_logic_solver')).
:- use_module(probsrc('dpllt_solver/difference_logic/ast_to_difference_logic')).
:- use_module(probsrc(b_ast_cleanup), [clean_up_pred/3]).
:- use_module(probsrc(preferences), [get_preference/2,
                                     temporary_set_preference/2,
                                     reset_temporary_preference/1]).
:- use_module(probsrc(b_enumerate), [b_tighter_enumerate_values/2]).
:- use_module(probsrc(kernel_waitflags), [init_wait_flags/2,ground_wait_flags/1]).
:- use_module(probsrc(kernel_objects), [infer_value_type/2]).
:- use_module(probsrc(b_interpreter), [set_up_typed_localstate/6,
                                       b_test_boolean_expression/4,
                                       b_convert_bool/5]).
:- use_module(probsrc(error_manager), [add_error/3,
                                       check_error_occured/2,
                                       enter_new_error_scope/2,
                                       exit_error_scope/3,
                                       clear_enumeration_warnings/0,
                                       critical_enumeration_warning_occured_in_error_scope/4]).
:- use_module(probsrc(bsyntaxtree), [find_typed_identifier_uses/2,
                                     get_texpr_info/2,
                                     conjunct_predicates/2]).
:- use_module(probsrc(module_information), [module_info/2]).

:- module_info(group, dpllt).
:- module_info(description,'This module provides a DPLL(T) based solver for B.').

:- dynamic initial_solution/0, use_idl_theory_solver/0.
:- volatile initial_solution/0, use_idl_theory_solver/0.

%use_idl_theory_solver.

init :-
    (   dpllt_settings:static_smt_symmetry_breaking(true)
    ->  % enumeration has to be linear for symmetry breaking
        temporary_set_preference(randomise_enumeration_order, false)
    ;   true
    ),
    error_manager:reset_errors,
    error_manager:clear_enumeration_warnings,
    dpllt_pred_to_sat:reset_sat_var_id,
    retractall(initial_solution),
    retractall(idl_candidate_constants(_,_)),
    %temporary_set_preference(use_chr_solver, true),
    temporary_set_preference(unsat_core_algorithm, divide_and_conquer),
    temporary_set_preference(use_smt_mode, false),
    temporary_set_preference(use_clpfd_solver, true),
    temporary_set_preference(optimize_ast, true),
    temporary_set_preference(use_common_subexpression_elimination, false),
    temporary_set_preference(normalize_ast, true). % important

reset_preferences :-
    reset_temporary_preference(randomise_enumeration_order),
    reset_temporary_preference(unsat_core_algorithm),
    reset_temporary_preference(use_smt_mode),
    reset_temporary_preference(use_clpfd_solver),
    reset_temporary_preference(optimize_ast),
    reset_temporary_preference(use_common_subexpression_elimination),
    reset_temporary_preference(normalize_ast).

%% dpllt_solve_predicate(+Pred, -Result).
% Main interface predicate for DPLL(T) based solver which times out
% with respect to ProB's time_out preference.
dpllt_solve_predicate(Pred, Result) :-
    init,
    get_preference(time_out, Timeout),
    time_out(dpllt_solve_predicate_initialized(Pred, SolverResult),
        Timeout,
        TimeoutResult),
    (   TimeoutResult == time_out
    ->  Result = time_out
    ;   ground(SolverResult),
        Result = SolverResult
    ).

dpllt_solve_predicate_initialized(Pred, Result) :-
    on_exception(Exception,
        dpllt_solve_predicate_no_timeout(Pred, Result),
        handle_clpfd_overflow(Exception, Pred, Result)),
    reset_preferences.

%% handle_clpfd_overflow(+Exception, +Pred, -Result).
% Disable CLP(FD) interface and restart DPLL(T) if an overflow occured.
handle_clpfd_overflow(Exception, Pred, Result) :-
    Exception = error(representation_error(Err),_),
    memberchk(Err, ['CLPFD integer overflow','max_clpfd_integer','min_clpfd_integer']),
    !,
    temporary_set_preference(use_clpfd_solver, false),
    dpllt_solve_predicate_initialized(Pred, Result).
handle_clpfd_overflow(_, _, unknown).

%% dpllt_solve_predicate(+Pred, -SolverResult).
% Interface predicate for DPLL(T) based solver without a timeout.
dpllt_solve_predicate_no_timeout(Pred, SolverResult) :-
    %nl, write('Initial pred: '), nl, translate:print_bexpr(Pred), nl,
    %nl, write('Initial pred: '),nl,writeq(Pred),nl,
    %unfold_let(Pred, Unfolded),
    %nl, write('Unfolded pred: '), nl, translate:print_bexpr(Unfolded), nl,
    %(   preferences:get_preference(smtlib2b_cleanup_predicate, true)
    %->
    b_ast_cleanup:clean_up_pred(Pred, _, CleanPred),
    nl, write('Clean pred: '), nl, translate:print_bexpr(CleanPred), nl,
    (   dpllt_settings:static_smt_symmetry_breaking(true)
    ->  add_symmetry_breaking_predicates(CleanPred, SymPred)
    ;   SymPred = CleanPred
    ),
    %;   CleanPred = Pred
    %),
    %nl, write('Clean pred: '),nl,writeq(CleanPred),nl,
    %nl, write('Sym pred: '),nl,translate:print_bexpr(SymPred),nl,
    (   CleanPred = b(truth,pred,_)
    ->  SolverResult = solution([])
    ;   CleanPred = b(falsity,pred,_)
    ->  SolverResult = contradiction_found
    ;   is_rewrite_to_idl(RewriteToIdl),
        preprocess_predicate(RewriteToIdl, SymPred, LiftedPred, FilteredCandidateImplsConj),
        predicate_to_sat(normal, [], [], LiftedPred, Env1, AnnotatedPred, SmtBoolFormula, SatVars1),
        predicate_to_sat(normal, SatVars1, Env1, FilteredCandidateImplsConj, NewEnv, _, ImplBoolFormula, SatVars),
        %nl, write('Impl Formula: '), nl, translate:print_bexpr(FilteredCandidateImplsConj), nl,
        %nl, write('ImplBoolFormula: '), nl, translate:print_bexpr(ImplBoolFormula), nl,
        !,
        %nl, write('Lifted: '), translate:print_bexpr(LiftedPred), nl,
        conjunct_predicates([ImplBoolFormula,SmtBoolFormula], BoolFormula),
        find_typed_identifier_uses(AnnotatedPred, SmtVars), % TO DO: improve
        %nl, write('Bool formula: '), nl, translate:print_bexpr(BoolFormula), nl,
        enter_new_error_scope(ScopeID, dpllt_solve_predicate),
        call_cleanup(conflict_driven_clause_learning_online(NewEnv, SmtVars, SatVars, BoolFormula, SolverResult),
                     exit_error_scope(ScopeID, _, dpllt_solve_predicate))
    ).
dpllt_solve_predicate_no_timeout(Pred, _) :-
    add_error(dpllt_solve_predicate, 'Cannot create SAT formula from B predicate:', [Pred]).

is_rewrite_to_idl(Res) :-
    use_idl_theory_solver,
    !,
    Res = true.
is_rewrite_to_idl(false).

% Given a set of boolean B identifiers. Create a list of bindings containing terms bind/2
% and a list of tuples of B identifier and corresponding Prolog variable used for sat solving.
get_bindings_from_ids(SatVars, SatBindings, IdPrologVarTuples, PrologSatVarTriple) :-
    get_bindings_from_ids_acc(SatVars, [], [], [], SatBindings, IdPrologVarTuples, PrologSatVarTriple), !.

get_bindings_from_ids_acc([], AccBindings, AccTuples, AccPlVars, AccBindings, AccTuples, AccPlVars).
get_bindings_from_ids_acc([b(identifier(Name),IdType,IdInfo)|T], AccBindings, AccTuples, AccPlVars, SatBindings, IdPrologVarTuples, PrologSatVarTriple) :-
    !,
    NewAccBindings = [bind(Name,PrologVar)|AccBindings],
    NewAccTuples = [(b(identifier(Name),IdType,IdInfo),PrologVar)|AccTuples],
    member(smt_formula(SmtFormula), IdInfo),
    NewAccPlVars = [PrologVar-Name-SmtFormula|AccPlVars],
    get_bindings_from_ids_acc(T, NewAccBindings, NewAccTuples, NewAccPlVars, SatBindings, IdPrologVarTuples, PrologSatVarTriple).
get_bindings_from_ids_acc([_|T], AccBindings, AccTuples, AccPlVars, SatBindings, IdPrologVarTuples, PrologSatVarTriple) :-
    get_bindings_from_ids_acc(T, AccBindings, AccTuples, AccPlVars, SatBindings, IdPrologVarTuples, PrologSatVarTriple).

%% conflict_driven_clause_learning_online(+Env, +SmtVars, +SatVars, +BoolFormula, -SolverResult).
% Conflict-driven clause learning from incomplete assignments of the boolean formula by setting up reification constraints
% connecting the SAT and theory solvers.
conflict_driven_clause_learning_online(Env, SmtVars, SatVars, BoolFormula, SolverResult) :-
    get_bindings_from_ids(SatVars, SatBindings, IdPrologVarTuples, PrologSatVarTriple),
    b_to_cnf(BoolFormula, SatBindings, CnfBoolFormula),
    setup_theory_wf_store(SmtVars, SmtBindings, WfStoreSmt),
    (   use_idl_theory_solver
    ->   difference_logic_solver:init_idl_solver(IdlGraphMut)
    ;   true
    ),
    dpllt_sat_solver:create_solver_state(SatStateMutable),
    setup_reification(SatStateMutable, IdlGraphMut, IdPrologVarTuples, SmtBindings, WfStoreSmt),
    findall(SatVarName, member(bind(SatVarName,_), SatBindings), SatVarNames),
    SolverEnv = [sat_vars:SatVars,sat_varnames:SatVarNames,sat_bindings:SatBindings,t2b_env:Env,pl_vars:PrologSatVarTriple],
    dpllt_sat_solver:sat(SolverEnv, SatStateMutable, CnfBoolFormula),
    (   use_idl_theory_solver
    ->  propagate_fd_bounds_to_idl_solver(IdlGraphMut, SmtBindings),% TO DO: propagate from ProB to IDL solver earlier?
        get_idl_solution_bindings(IdlGraphMut, IdlBindings),
        (   % try one idl solution first since ProB could timeout, but do not enumerate idl solutions here
            set_bindings(IdlBindings, SmtBindings),
            ground_wait_flags(WfStoreSmt)
        ;   exclude_idl_solution(IdlBindings, SmtBindings, WfStoreSmt),
            (   ground_wait_flags(WfStoreSmt)
            ;   % fallback to idl solver and possibly enumerate all solutions
                error_manager:critical_enumeration_warning_occured_in_error_scope(_, _, _, _),
                error_manager:reset_errors,
                propagate_idl_solution_to_bindings_bt(IdlGraphMut, SmtBindings),
                ground_wait_flags(WfStoreSmt)
            )
        )
    ;   ground_wait_flags(WfStoreSmt)
    ),
    (   \+ error_manager:check_error_occured(_, _)
    ;   error_manager:reset_errors,
        fail
    ),
    SolverResult = solution(SmtBindings),
    log_solution,
    announce_bt_from_smt_solution.
conflict_driven_clause_learning_online(_, _, _, _, Res) :-
    \+ error_manager:check_error_occured(well_definedness_error, _),
    error_manager:critical_enumeration_warning_occured_in_error_scope(A, B, C, D),
    !,
    clear_enumeration_warnings,
    Res = no_solution_found(enumeration_warning(A,B,C,D,critical)).
conflict_driven_clause_learning_online(_, _, _, _, Res) :-
    error_manager:check_error_occured(_, _),
    \+ error_manager:check_error_occured(b_to_cnf, _),
    !,
    Res = error.
conflict_driven_clause_learning_online(_, _, _, _, contradiction_found) :-
    \+ error_manager:check_error_occured(b_to_cnf, _),
    \+ initial_solution.

%% setup_reification(+SatStateMutable, +IdlGraphMut, +IdPrologVarTuples, +SmtBindings, +WfStoreSmt).
% Set up reification constraints to connect SAT and SMT solver.
% Propagates in both directions (b_convert_bool/5 implements theory propagation).
setup_reification(_, _, [], _, _).
setup_reification(SatStateMutable, IdlGraphMut, [(SatId,SatPrologVar)|T], SmtBindings, WfStoreSmt) :-
    get_texpr_info(SatId, Info),
    memberchk(smt_formula(SmtFormula), Info),
    SatId = b(identifier(SatVarName),_,_),
    setup_reification_for_solver(SatStateMutable, IdlGraphMut, SmtFormula, SmtBindings, WfStoreSmt, SatVarName, SatPrologVar),
    setup_reification(SatStateMutable, IdlGraphMut, T, SmtBindings, WfStoreSmt).

setup_reification_for_solver(SatStateMutable, IdlGraphMut, SmtFormula, SmtBindings, WfStoreSmt, SatVarName, SatPrologVar) :-
    log_theory_propagation(SatStateMutable, SatVarName, SatPrologVar),
    (   use_idl_theory_solver
        % TO DO: Problem beide Solver aufzusetzen: was ist wenn idl solver lösung findet aber prob timeout?
        ->  (   ast_to_difference_logic:rewrite_to_idl(SmtFormula, [DLConstraint])
            ->  nl,write('Constraint for IDL: '), nl, translate:print_bexpr(SmtFormula), nl,
                idl_solver_interface(IdlGraphMut, SatPrologVar, [DLConstraint]),
                b_convert_bool(SmtFormula, SmtBindings, [], WfStoreSmt, SatPrologVar)
            ;   log_idl_candidates_from_constraint(SmtFormula),
                nl,write('Constraint for ProB: '), nl, translate:print_bexpr(SmtFormula), nl,
                propagate_non_idl_to_idl(IdlGraphMut, SatPrologVar, SmtFormula),
                b_convert_bool(SmtFormula, SmtBindings, [], WfStoreSmt, SatPrologVar)
            )
    ;   nl,write('Constraint for ProB: '), nl, translate:print_bexpr(SmtFormula), nl,
        b_convert_bool(SmtFormula, SmtBindings, [], WfStoreSmt, SatPrologVar)
    ).

:- block log_theory_propagation(?, ?, -).
log_theory_propagation(SatStateMutable, SatVarName, SatPrologVar) :-
    ground(SatPrologVar),
    (   is_last_propagated_sat_var(SatStateMutable, SatVarName)
    ->  true
    ;   dpllt_sat_solver:log_theory_propagation_sat_stack(SatStateMutable, SatVarName, SatPrologVar)
    ).

:- dynamic idl_candidate_constants/2.
:- volatile idl_candidate_constants/2.

log_idl_candidates_from_constraint(Constraint) :-
    get_ids_and_int_constants(Constraint, Ids, IntConstants),
    Ids \== [],
    IntConstants \== [],
    !,
    assert_idl_candidates(Ids, IntConstants).
log_idl_candidates_from_constraint(_).

assert_idl_candidates([], _).
assert_idl_candidates([Id|T], IntConstants) :-
    retract(idl_candidate_constants(Id, Candidates)),
    !,
    subtract(IntConstants, Candidates, New),
    append(Candidates, New, NewCandidates),
    asserta(idl_candidate_constants(Id, NewCandidates)),
    assert_idl_candidates(T, IntConstants).
assert_idl_candidates([Id|T], IntConstants) :-
    asserta(idl_candidate_constants(Id, IntConstants)),
    assert_idl_candidates(T, IntConstants).

get_ids_and_int_constants(b(Node,pred,_), Ids, IntConstants) :-
    comparison_op(Node, Lhs, Rhs),
    get_ids_and_int_constants_expr(Lhs, [], [], IdsAcc, IntConstantsAcc),
    get_ids_and_int_constants_expr(Rhs, IdsAcc, IntConstantsAcc, Ids, IntConstants).

get_ids_and_int_constants_expr(Id, IdsAcc, IntConstantsAcc, [Name|IdsAcc], IntConstantsAcc) :-
    Id = b(identifier(Name),integer,_).
get_ids_and_int_constants_expr(Int, IdsAcc, IntConstantsAcc, IdsAcc, [Value,Value1|IntConstantsAcc]) :-
    (   Int = b(integer(Value),integer,_)
    ;   Int = b(unary_minus(b(integer(Value),integer,_)),integer,_)
    ),
    % select positive and negative value
    Value1 is Value * -1.
get_ids_and_int_constants_expr(b(Node,integer,_), IdsAcc, IntConstantsAcc, Ids, IntConstants) :-
    arithmetic_expr(Node, Lhs, Rhs),
    get_ids_and_int_constants_expr(Lhs, IdsAcc, IntConstantsAcc, NIdsAcc, NIntConstantsAcc),
    get_ids_and_int_constants_expr(Rhs, NIdsAcc, NIntConstantsAcc, Ids, IntConstants).

comparison_op(less(Lhs,Rhs), Lhs, Rhs).
comparison_op(less_equal(Lhs,Rhs), Lhs, Rhs).
comparison_op(equal(Lhs,Rhs), Lhs, Rhs).
comparison_op(not_equal(Lhs,Rhs), Lhs, Rhs).

arithmetic_expr(minus(Lhs,Rhs), Lhs, Rhs).
arithmetic_expr(add(Lhs,Rhs), Lhs, Rhs).
arithmetic_expr(div(Lhs,Rhs), Lhs, Rhs).
arithmetic_expr(floored_div(Lhs,Rhs), Lhs, Rhs).
arithmetic_expr(power_of(Lhs,Rhs), Lhs, Rhs).
arithmetic_expr(multiplication(Lhs,Rhs), Lhs, Rhs).
arithmetic_expr(modulo(Lhs,Rhs), Lhs, Rhs).

:- block propagate_non_idl_to_idl(?, -, ?).
propagate_non_idl_to_idl(IdlGraphMut, SatPrologVar, SmtFormula) :-
    (   SatPrologVar == pred_true % TO DO: pred_false?
    % e.g., x:NAT -> x>=0
    ->  (   infer_constraints_for_idl_solver(SmtFormula, ConjList)
        ->  register_constraints(IdlGraphMut, ConjList)
        ;   true
        )
    ;   true
    ).

%% idl_solver_interface(+IdlGraphMut, +SatPrologVar, +Constraint).
:- block idl_solver_interface(?, -, ?).
idl_solver_interface(IdlGraphMut, SatPrologVar, [DLConstraint]) :-
    SatPrologVar == pred_true,
    difference_logic_solver:remove_unsat_core,
    dpllt_sat_solver:remove_idl_unsat_core,
    (   difference_logic_solver:register_constraint(IdlGraphMut, DLConstraint)
    ;   \+ dpllt_sat_solver:is_backjumping,
        propagate_idl_unsat_core_to_sat_solver(IdlGraphMut),
        fail
    ).
idl_solver_interface(IdlGraphMut, SatPrologVar, [DLConstraint]) :-
    SatPrologVar == pred_false,
    difference_logic_solver:remove_unsat_core,
    dpllt_sat_solver:remove_idl_unsat_core,
    (   difference_logic_solver:register_constraint(IdlGraphMut, b(negation(DLConstraint),pred,[]))
    ;   \+ dpllt_sat_solver:is_backjumping,
        propagate_idl_unsat_core_to_sat_solver(IdlGraphMut),
        fail
    ).

propagate_idl_unsat_core_to_sat_solver(IdlGraphMut) :-
    difference_logic_solver:get_solver_result(IdlGraphMut, Res),
    Res == contradiction_found,
    difference_logic_solver:get_unsat_core(IdlCore),
    difference_logic_solver:remove_unsat_core,
    dpllt_sat_solver:store_idl_unsat_core(IdlCore).

%% get_idl_solution_bindings(+IdlGraphMut, -Bindings).
get_idl_solution_bindings(IdlGraphMut, Bindings) :-
    get_solver_result(IdlGraphMut, SolverResult),
    SolverResult = solution(Bindings).

%% propagate_idl_solution_to_bindings_bt(+IdlGraphMut, +SmtBindings).
propagate_idl_solution_to_bindings_bt(IdlGraphMut, SmtBindings) :-
    use_idl_theory_solver,
    !,
    % TO DO: try the current solution first
    (   get_candidate_bounds_from_non_idl_constraints(CandidateTuples),
        difference_logic_solver:try_candidate_bounds(IdlGraphMut, CandidateTuples, Result)
    ;   difference_logic_solver:get_all_solutions_on_bt(IdlGraphMut, Result)
    ),
    Result = solution(Bindings),
    set_bindings(Bindings, SmtBindings).
propagate_idl_solution_to_bindings_bt(_, _).

%% propagate_fd_bounds_to_idl_solver(+IdlGraphMut, +SmtBindings).
propagate_fd_bounds_to_idl_solver(IdlGraphMut, SmtBindings) :-
    use_idl_theory_solver,
    !,
    difference_logic_solver:get_registered_vars(IdlGraphMut, Vars),
    get_constraints_from_fd_bounds(SmtBindings, Vars, [], ConjList),
    difference_logic_solver:register_constraints(IdlGraphMut, ConjList).
propagate_fd_bounds_to_idl_solver(_, _).

get_constraints_from_fd_bounds(_, [], Acc, Acc).
get_constraints_from_fd_bounds(SmtBindings, [Var|T], Acc, ConjList) :-
    memberchk(bind(Var, int(Int)), SmtBindings),
    integer(Int),
    !,
    Equality = b(equal(b(identifier(Var),integer,[]),b(integer(Int),integer,[])),pred,[]),
    get_constraints_from_fd_bounds(SmtBindings, T, [Equality|Acc], ConjList).
get_constraints_from_fd_bounds(SmtBindings, [Var|T], Acc, ConjList) :-
    memberchk(bind(Var, int(FDVar)), SmtBindings),
    fd_var(FDVar),
    fd_min(FDVar, Min),
    fd_max(FDVar, Max),
    !,
    (   var_geq_min(Var, Min, MinBound)
    ->  (   var_leq_max(Var, Max, MaxBound)
        ->  NAcc = [b(conjunct(MinBound,MaxBound),pred,[])|Acc]
        ;   NAcc = [MinBound|Acc]
        )
    ;   (   var_leq_max(Var, Max, MaxBound)
        ->  NAcc = [MaxBound|Acc]
        ;   NAcc = Acc
        )
    ),
    get_constraints_from_fd_bounds(SmtBindings, T, NAcc, ConjList).
get_constraints_from_fd_bounds(SmtBindings, [_|T], Acc, ConjList) :-
    get_constraints_from_fd_bounds(SmtBindings, T, Acc, ConjList).

var_geq_min(Var, Min, MinBound) :-
    % -v <= -min
    integer(Min),
    Min1 is Min * -1,
    MinBound = b(less_equal(b(minus(b(identifier('_zero'),integer,[]),b(identifier(Var),integer,[])),integer,[]),b(integer(Min1),integer,[])),pred,[]).

var_leq_max(Var, Max, MaxBound) :-
    % v <= Max
    integer(Max),
    MaxBound = b(less_equal(b(identifier(Var),integer,[]),b(integer(Max),integer,[])),pred,[]).

get_candidate_bounds_from_non_idl_constraints(CandidateTuples) :-
    findall((VarName,Candidates), idl_candidate_constants(VarName, Candidates), CandidateTuples).

set_bindings([], _).
set_bindings([binding(VarName,Val,_)|T], SmtBindings) :-
    member(bind(VarName,Val), SmtBindings),
    set_bindings(T, SmtBindings).

%% exclude_idl_solution(+IdlBindings, +SmtBindings, +WfStoreSmt).
exclude_idl_solution(IdlBindings, SmtBindings, WfStoreSmt) :-
    initial_solution,
    exclude_solution(IdlBindings, Exclusion),
    b_test_boolean_expression(Exclusion, SmtBindings, [], WfStoreSmt).
exclude_idl_solution(_, _, _) :-
    \+ initial_solution.

%% exclude_solution(+Bindings, -Exclusion).
exclude_solution([], b(truth,pred,[])).
exclude_solution([Binding|T], Exclusion) :-
    get_equality_from_binding(Binding, EQ),
    exclude_solution(T, EQ, Exclusion).

exclude_solution([], Acc, Acc).
exclude_solution([Binding|T], Acc, Exclusion) :-
    get_equality_from_binding(Binding, EQ),
    NAcc = b(disjunct(b(negation(EQ),pred,[]),Acc),pred,[]),
    exclude_solution(T, NAcc, Exclusion).

%% get_equality_from_binding(+Binding, -EQ).
get_equality_from_binding(Binding, EQ) :-
    (   Binding = bind(VarName,Val)
    ;   Binding = binding(VarName,Val,_)
    ),
    kernel_objects:infer_value_type(Val, Type),
    EQ = b(equal(b(identifier(VarName),Type,[]),b(value(Val),Type,[])),pred,[]).

%% setup_theory_wf_store(+SmtVars, -SmtBindings, -WFStoreSMT).
setup_theory_wf_store(SmtVars, SmtBindings, WFStoreSMT) :-
    set_up_typed_localstate(SmtVars, _, SmtTypedVals, [], SmtBindings, positive),
    init_wait_flags(WFStoreSMT, [wf_smt_cdcl]),
    b_tighter_enumerate_values(SmtTypedVals, WFStoreSMT),!.

log_solution :-
    \+ initial_solution,
    !,
    asserta(initial_solution).
log_solution.

unfold_let(Pred, Unfolded) :-
   unfold_let([], Pred, Unfolded).

unfold_let(IdTuples, b(identifier(Name),_,_), Replacement) :-
    member((b(identifier(Name),_,_),Ast), IdTuples),
    !,
    Replacement = Ast.
unfold_let(IdTuples, b(Node,_,_), Replacement) :-
    Node =.. [Functor,Ids,Asts,Body],
    (   Functor == let_predicate
    ;   Functor == let_expression
    ),
    !,
    maplist(unfold_let(IdTuples), Asts, NAsts),
    zip_acc(Ids, NAsts, IdTuples, NewIdTuples),
    unfold_let(NewIdTuples, Body, Replacement).
unfold_let(IdTuples, b(Node,Type,Info), Replacement) :-
    Node =.. [Functor,Arg1,Arg2],
    !,
    unfold_let(IdTuples, Arg1, NArg1),
    unfold_let(IdTuples, Arg2, NArg2),
    NNode =.. [Functor,NArg1,NArg2],
    Replacement = b(NNode,Type,Info).
unfold_let(IdTuples, b(Node,Type,Info), Replacement) :-
    Node =.. [Functor,Arg1],
    !,
    unfold_let(IdTuples, Arg1, NArg1),
    NNode =.. [Functor,NArg1],
    Replacement = b(NNode,Type,Info).
unfold_let(IdTuples, b(forall(Ids,Lhs,Rhs),Type,Info), Replacement) :-
    !,
    unfold_let(IdTuples, Lhs, NLhs),
    unfold_let(IdTuples, Rhs, NRhs),
    Replacement = b(forall(Ids,NLhs,NRhs),Type,Info).
unfold_let(IdTuples, b(exists(Ids,Body),Type,Info), Replacement) :-
    !,
    unfold_let(IdTuples, Body, NBody),
    Replacement = b(exists(Ids,NBody),Type,Info).
unfold_let(_, Ast, Ast).

zip_acc([], [], Acc, Acc).
zip_acc([A|T1], [B|T2], Acc, Zipped) :-
    zip_acc(T1, T2, [(A,B)|Acc], Zipped).

%%%%%%%%%%%%%%%%%%%%% Fuzzing and Unit Tests %%%%%%%%%%%%%%%%%%%%%

:- use_module(extension('prolog_fuzzer/fuzzing')).
:- use_module(probsrc(solver_interface), [solve_predicate/3]).

verify_dpllt_solutions_with_prob(Pred) :-
    dpllt_solve_predicate(Pred, Solution),
    !,
    prob_accepts_result(Pred, Solution).

solve_clean_predicate(Pred, Bindings, Res) :-
    clean_up_pred(Pred, _, CPred),
    solve_predicate(CPred, Bindings, Res).

prob_accepts_result(_, Res) :-
    % skip erroneous constraints or timeout
    (   Res == error
    ;   Res == time_out
    ),
    !.
prob_accepts_result(Pred, solution([])) :-
    solve_clean_predicate(Pred, _, Res),
    (   Res = solution([])
    ;   Res = no_solution_found(enumeration_warning(_,_,_,_,_))
    ;   Res = time_out
    ).
prob_accepts_result(Pred, solution(Bindings)) :-
    ground(Bindings),
    solve_clean_predicate(Pred, Bindings, Res),
    (   Res = solution(_)
    ;   Res = no_solution_found(enumeration_warning(_,_,_,_,_))
    ;   Res = time_out
    ).
prob_accepts_result(Pred, SolverSolution) :-
    SolverSolution == contradiction_found,
    solve_clean_predicate(Pred, _, Res),
    (   Res == contradiction_found
    ;   Res = no_solution_found(enumeration_warning(_,_,_,_,_))
    ).
prob_accepts_result(Pred, SolverSolution) :-
    SolverSolution = no_solution_found(enumeration_warning(_,_,_,_,_)),
    solve_clean_predicate(Pred, _, Res),
    (   Res = no_solution_found(enumeration_warning(_,_,_,_,_))
    ;   Res = solution(_)
    ;   Res = contradiction_found
    ).
/*prob_accepts_result(Pred, SolverSolution) :-
    SolverSolution == error,
    solve_clean_predicate(Pred, _, Res),
    Res == error.
*/

fuzzy_verify_solver :-
    fuzz(verify_dpllt_solutions_with_prob, 1, 1000, prob_ast_pred([])).
