% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(eventhandling, [register_event_listener/3,announce_event/1, store_virtual_event/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'The eventhandling module is used to register predicates that are called when certain events (like loading a specification) occurr.').

:- use_module(library(lists)).

:- use_module(error_manager,[add_internal_error/2]).
:- use_module(tools_strings,[ajoin/2]).

:- dynamic registered_listener/3.

:- meta_predicate register_event_listener(+,0,+).
% This predicate is called when starting up; the other modules may not be initialised
% Do not use other modules yet
register_event_listener(Event,Module:Pred,Description) :- %print(register(Event,Module)),nl,
    ( known_events(Event,_) ->
        register_event_listener2(Event,Module:Pred,Description)
    ; otherwise -> print(unknown(Event)),nl,
        format('~n!!! Unknown event ~w (from ~w:~w) : ~w~n~n',[Event,Module,Pred,Description])
        %add_internal_error('Unknown event',register_event_listener(Event,Module:Pred,Description))
    ).
register_event_listener2(Event,Module:Pred,Description) :-
    (registered_listener(Event,Module:Pred,Description) -> true
      ; assert(registered_listener(Event,Module:Pred,Description))).

known_events(compile_prob,'ProB is being compiled').
known_events(startup_prob,'ProB is starting up').
known_events(clear_specification,'After we have finished with a specification').
known_events(reset_specification,'Resetting animator, without changing specification').
known_events(start_initialising_specification,'We start loading a (new) specification').
known_events(specification_initialised,'Directly after the specificiation has been loaded').
known_events(change_of_animation_mode,'Changing Animation Mode (e.g., adding CSP guide)').
known_events(start_solving,'Begin solving a new predicate').
known_events(end_solving,'End of solving a predicate').
known_events(play_counterexample,'Getting counterexample from cache').
known_events(start_unit_tests,'Starting ProB Unit Tests').
known_events(stop_unit_tests,'Stop ProB Unit Tests').

% not announced, just stored for lifecycle; TO DO: try and get rid of this (used for self-check)
store_virtual_event(Event) :-
    ( known_events(Event,_) ->
        check_lifecycle(Event)
    ; otherwise ->
        add_internal_error('Unknown event',announce_event(Event))).

announce_event(Event) :- %print(announce(Event)),nl,
    ( known_events(Event,_) ->
        announce_event2(Event)
    ; otherwise ->
        add_internal_error('Unknown event',announce_event(Event))).
:- meta_predicate call_for_event(-,0).
announce_event2(Event) :-
    check_lifecycle(Event),
    findall(Module:Predicate, registered_listener(Event,Module:Predicate,_Desc), Calls),
    maplist(call_for_event(Event), Calls)
    . %, format('Finished processing ~w~n',[Event]).
call_for_event(Event,Module:Predicate) :-
    % format('Event ~w --triggered--> ~w~n',[Event,Predicate]),
    ( call(Module:Predicate) -> true
    ; otherwise ->
        ajoin(['Call for event ',Event,' failed.'],Msg),
        add_internal_error(Msg,Module:Predicate)).


% life-cycle management & monitoring
% at the moment very simple: as all events are specification related

:- dynamic last_event/1.
last_event(none).

check_lifecycle(Event) :-  last_event(X), %nl,print(event(Event,last(X))),nl,
  !,
  (missing_transition(X,Event) -> true ; true),
  retractall(last_event(_)),
  assert(last_event(Event)).
check_lifecycle(Event) :- add_internal_error('No last event for ',check_lifecycle(Event)),
  assert(last_event(Event)).

:- use_module(error_manager,[add_warning/3, add_warning/2]).
missing_transition(specification_initialised,start_initialising_specification) :-
    add_warning(event_handling,'Missing event: ',clear_specification).
    %announce_event(clear_specification). % here we can remedy this
missing_transition(clear_specification,specification_initialised) :-
    add_warning(event_handling,'Missing event: ',start_initialising_specification).
missing_transition(specification_initialised,specification_initialised) :-
     add_warning(event_handling,'Previous specification not cleared').
missing_transition(start_initialising_specification,start_initialising_specification) :-
     add_warning(event_handling,'Double Start of Specification Initialisation').
%missing_transition(start_initialising_specification,clear_specification). % we do not need to announce spec initialised
