% (c) 2020-2021 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

%:- module(external_functions_svg,['svg_points'/4, svg_axis/7, svg_set_polygon/7]).

% external functions which aid in producing attributes for SVG/HTML VisB objects

% generate a SVG "points" attribute string (e.g. for polygons or polylines) from a sequence of pairs of reals:
% svg_points2str(points) == conc(%i.(i:dom(points)|
%                                FORMAT_TO_STRING("~w,~w ",[prj1(REAL,REAL)(points(i)),
%                                                           prj2(REAL,REAL)(points(i))])));
 % example: svg_points([(1.0,2.0),(2.0,2.0)]) = "1.0,2.0 2.0,2.0 "

expects_waitflag('svg_points').
expects_spaninfo('svg_points').
external_fun_type('svg_points',[T1,T2],[seq(couple(T1,T2)),string]). % T1, T2 should be numbers
:- block 'svg_points'(-,?,?,?).
svg_points(Seq,StrResult,Span,WF) :- !,
   custom_explicit_sets:expand_custom_set_to_list_wf(Seq,ESeq,Done,'svg_points',WF),
   svg_points2str_aux(Done,ESeq,StrResult,Span,WF).

:- block svg_points2str_aux(-,?,?,?,?).
svg_points2str_aux(_Done,ESeq,StrResult,Span,_WF) :-
    sort(ESeq,Sorted),
    translate_pairs(Sorted,Span,List,[]),
    ajoin(List,ResStr),
    StrResult = string(ResStr).

% translate pairs of numbers to SVG format "x1,y1 x2,y2 ..."
translate_pairs([],_) --> [].
translate_pairs([(_,(V1,V2))|T],Span) -->
    {get_number(V1,V1A,Span)},
    [V1A],
    [','],
    {get_number(V2,V2A,Span)},
    [V2A],
    [' '],
    translate_pairs(T,Span).

    % to do: use pp_value(V,LimitReached,Codes,[])
    % to do: check that we have numbers
% external_functions:svg_points([(int(1),(int(1),int(3)))],Res,unknown,WF), Res == string('1,3 ')
% external_functions:svg_points([(int(1),(int(1),int(3))), (int(2),(int(2),int(4)))],Res,unknown,WF), Res == Res = string('1,3 2,4 ')

% -----------

% render a simple train as a polygon
% XScale is applied to start and length
% B Example: svg_train(10,5,1.0,2.0,3.0) = "10.0,3.0 15.0,3.0 13.0,0 10.0,0 10.0,3.0 "
expects_waitflag('svg_train').
expects_spaninfo('svg_train').
external_fun_type('svg_train',[T1,T2],[T1,T1,T2,T2,T2,string]).  % T1, T2 should be numbers
:- block 'svg_train'(-,?,?,?,?, ?,?,?),
         'svg_train'(?,-,?,?,?, ?,?,?),
         'svg_train'(?,?,-,?,?, ?,?,?),
         'svg_train'(?,?,?,-,?, ?,?,?),
         'svg_train'(?,?,?,?,-, ?,?,?).
svg_train(Start,Length,XScale,Slant,Height,StrResult,Span,_WF) :-
    get_number(Start,St,Span), get_number(Length,Len,Span),
    get_number(XScale,XMultiplier,Span),
    get_number(Slant,S,Span), get_number(Height,H,Span),
    svg_train_aux(St,Len,XMultiplier,S,H,StrResult,Span).

:- block 'svg_train_aux'(-,?,?,?,?, ?,?),
         'svg_train_aux'(?,-,?,?,?, ?,?),
         'svg_train_aux'(?,?,-,?,?, ?,?),
         'svg_train_aux'(?,?,?,-,?, ?,?),
         'svg_train_aux'(?,?,?,?,-, ?,?).
svg_train_aux(St,Len,XMultiplier,Slant,Height,StrResult,Span) :-
    X1 is St*XMultiplier,
    X2 is (St+Len)*XMultiplier, X2S is X2-Slant,
    translate_pairs([(1,(X1,Height)), (2,(X2,Height)), (3,(X2S,0)), (4,(X1,0)), (5,(X1,Height))],Span,List,[]),
    ajoin(List,ResStr),
    StrResult = string(ResStr).

% --------------------------

% draw an axis from 0 to MaxX with TickMarks of height Height at the TickMarks positions
% the TickMarks set values are multiplied by XScale (for convenience, e.g., for B models without reals)
% StrResult is a string to be used for a SVG polyline or polygon points attribute
% Note: polylines start at 0,0
% B example: svg_axis({50},1,0,100,1) = "0,0 50,0 50,0.5 50,-0.5 50,0 100,0 "
%            svg_axis({50},1.5,0.0,100.0,2.0) = "0.0,0 75.0,0 75.0,1.0 75.0,-1.0 75.0,0 100.0,0 "
expects_waitflag('svg_axis').
expects_spaninfo('svg_axis').
external_fun_type('svg_axis',[T1,T2],[set(T1),T2,T2,T2,string]).  % T1, T2 should be numbers
:- block 'svg_axis'(-,?,?,?,?,?,?),
         'svg_axis'(?,-,?,?,?,?,?),
         'svg_axis'(?,?,-,?,?,?,?),
         'svg_axis'(?,?,?,-,?,?,?).
svg_axis(TickMarks,XScale,MaxX,Height,StrResult,Span,WF) :-
    get_number(MaxX,Max,Span),
    get_number(XScale,XMultiplier,Span),
    get_number(Height,H,Span),
    custom_explicit_sets:expand_custom_set_to_list_wf(TickMarks,ESet,Done,'svg_axis',WF),
    svg_axis_aux(Done,ESet,XMultiplier,Max,H,StrResult,Span).

:- block svg_axis_aux(-,?,?,?, ?,?,?),
         svg_axis_aux(?,-,?,?, ?,?,?),
         svg_axis_aux(?,?,-,?, ?,?,?),
         svg_axis_aux(?,?,?,-, ?,?,?).
svg_axis_aux(_Done,ESet,XMultiplier,Max,H,StrResult,Span) :-
    sort(ESet,Sorted),
    H2 is H / 2,
    Tail =  [Max, ',0 '],
    gen_ticks(Sorted,XMultiplier,H2,Span,List,Tail),
    ajoin(List,ResStr),
    StrResult = string(ResStr).


gen_ticks([],_,_,_) --> [].
gen_ticks([TickMark|T],XMultiplier,H2,Span) -->
     {get_number(TickMark,TM,Span), XTick is TM*XMultiplier},
     [XTick], [',0 '],
     [XTick], [','], [H2], [' '],
     [XTick], [',-'], [H2], [' '],
     [XTick], [',0 '],
     gen_ticks(T,XMultiplier,H2,Span).


% --------------------------

% shows an integer set as a step-function/polygon of given height
% StrResult is a string to be used for a SVG polyline or polygon points attribute
% B example:   svg_set_polygon({50,51,52,55},1.0,100.0,2.0) =
%              "50.0,0 50.0,2.0 53.0,2.0 53.0,0 55.0,0 55.0,2.0 56.0,2.0 56.0,0 100.0,0 "
expects_waitflag('svg_set_polygon').
expects_spaninfo('svg_set_polygon').
external_fun_type('svg_set_polygon',[T2],[set(integer),T2,T2,T2,string]).  % T2 should be number type
:- block 'svg_set_polygon'(-,?,?,?,?,?,?),
         'svg_set_polygon'(?,-,?,?,?,?,?),
         'svg_set_polygon'(?,?,-,?,?,?,?),
         'svg_set_polygon'(?,?,?,-,?,?,?).
svg_set_polygon(TickMarks,XScale,MaxX,Height,StrResult,Span,WF) :-
    get_number(MaxX,Max,Span),
    get_number(XScale,XMultiplier,Span),
    get_number(Height,H,Span),
    custom_explicit_sets:expand_custom_set_to_list_wf(TickMarks,ESet,Done,'svg_set_polygon',WF),
    svg_set_polygon_aux(Done,ESet,XMultiplier,Max,H,StrResult,Span).

:- block svg_set_polygon_aux(-,?,?,?, ?,?,?),
         svg_set_polygon_aux(?,-,?,?, ?,?,?),
         svg_set_polygon_aux(?,?,-,?, ?,?,?),
         svg_set_polygon_aux(?,?,?,-, ?,?,?).
svg_set_polygon_aux(_Done,ESet,XMultiplier,Max,H,StrResult,Span) :-
    sort(ESet,Sorted),
    Tail =  [Max, ',0 '],
    gen_polygon(Sorted,XMultiplier,H,Span,List,Tail),
    ajoin(List,ResStr),
    StrResult = string(ResStr).


gen_polygon([],_,_,_) --> [].
gen_polygon([TickMark|T],XMultiplier,H,Span) -->
     {get_number(TickMark,TM,Span), XTick is TM*XMultiplier},
     [XTick], [',0 '],
     [XTick], [','], [H], [' '],
     {scan_for_next(T,TM,Span,TR,TM1), XTickNxt is TM1*XMultiplier},
     [XTickNxt], [','], [H], [' '],
     [XTickNxt], [',0 '],
     gen_polygon(TR,XMultiplier,H,Span).

% skip as long as X+1 is also in the set
scan_for_next([Nxt|T],XTick,Span,Res,XTickNxt) :- get_number(Nxt,XTick1,Span), XTick1 is XTick+1,!,
    scan_for_next(T,XTick1,Span,Res,XTickNxt).
scan_for_next(List,XTick,_,List,XTickNxt) :- XTickNxt is XTick+1.

% -------------------------
% Utilities


:- use_module(kernel_reals,[is_real/2]).
:- block get_number(-,?,?).
get_number(int(I),Res,_) :- !, Res=I.
get_number(Term,Res,_) :- is_real(Term,Real),!,Res=Real.
get_number(Nr,Res,_) :- number(Nr),!, Res=Nr. % convenience function for internal use in this module
get_number(Term,_,Span) :-
   add_error(get_number,'Illegal value for external function, is not a number:',Term,Span),fail.


