% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_cardinality_attr,[finite_cardinality_as_int_wf/3,
        finite_cardinality_as_int_with_type_wf/5,
        clpfd_card_domain_for_var/3
        ]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides operations for cardinality of sets, storing them as attributes.').


:- use_module(custom_explicit_sets,[explicit_set_cardinality_wf/3]).
:- use_module(kernel_waitflags,[add_wd_error/3]).
:- use_module(kernel_objects,[empty_set_wf/2, max_cardinality/2,
                              disjoint_sets/3, add_new_element_wf/4, not_element_of_wf/3,
                              exhaustive_kernel_check_wf/2]).
:- use_module(library(clpfd)).
:- use_module(clpfd_interface,[clpfd_max_bounded/1]).
:- use_module(self_check).


:- use_module(library(atts)).

/*
Examples:
x<:1..200000 & card(x) = 2*card(y) & card(y):3..44 & x/\y = {} & card(x) = 130-card(y)
  3.8 seconds before, now 1 ms to detect failure
card(x) : 10..200 & card(y) : card(x)..300 & card(y)+card(x) = 500 & x : POW(1..10000) & y: POW(1..10000)
  now solved much faster 1 second vs ???

STATUS:
 TO DO: replace cardinality_as_int in kernel_objects completely and add version here that returns inf/...
 TO DO: call finite_cardinality_as_int_with_type_wf if possible
 TO DO: Test 1631 NBishops has become slower (943d37585be8eeac46c93705d51c2392fde0fda3)

failed tests:
** Failed tests:
33 without SMT mode
 x:POW(7..8) & card(x)=1 & x /= {7}  --> no longer deterministically sets x={8} with SMT=FALSE
 calls:  kernel_objects:not_equal_explicit_set/2: AVL{int(7)}: VARIABLE: _17573

now fixed:
1627
1828 Check card operator does not generate solutions with repeated entries
     union(s) = {} & x<:INTEGER & x:s & card(s)=2  succeeds
union(s) = {} & x<:INTEGER & x:s & card(s)=2 succeeds
38 -> probably we need to set up list skel before infinite enumeration
223 -> overflow
496 Many explicit computations which check that all operators work as expected on various concrete data.
1187 Ensure proper enumeration of sequences (CLPFD=FALSE).
1418 Test this is solved quickly using new ordered_value check for card.
     card(x)=10 & !y.(y:x & y<50 => y+1:x) & x<:1..50
1419 Test this is solved quickly using new ordered_value check for card & avoid bug in expand all quantifeir.
1611 Ensure proper enumeration of sequences (CLPFD=TRUE).
1625 Check card of interval propagates
568 ProZ test
1540 Test Integer Z laws and that the division semantics of Z is preserved.
1850 Test loading .tex file works
1776 Check compilation of while with op call and inner while works
*/


:- attribute kernel_card_attr/2.


finite_cardinality_as_int_with_type_wf(Set,ObjectType,Info,int(Card),WF) :-
   max_cardinality(ObjectType,MaxCard),
    %format('card with type=~w, maxcard=~w, Info=~w~n',[ObjectType,MaxCard,Info]),
    %tools_printing:print_term_summary(Set),
   finite_card_wf(Set,0,MaxCard,Info,[],Card,WF).
   % print('card = '),tools_printing:print_integer(Card),nl,print(set(Set)),nl,nl.


:- assert_must_succeed((finite_cardinality_as_int_wf(Y,int(X),_WF), Y = [fd(1,'Name'),fd(2,'Name')],X==2)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_cardinality_attr:finite_cardinality_as_int_wf([int(2),int(4),int(1)],int(3),WF),WF)).
:- assert_must_succeed((finite_cardinality_as_int_wf(Y,int(2),no_wf_available), Y = [fd(1,'Name'),fd(2,'Name')])).
:- assert_must_fail((finite_cardinality_as_int_wf(Y,int(3),no_wf_available), Y = [fd(1,'Name'),fd(2,'Name')])).
:- assert_must_succeed((finite_cardinality_as_int_wf(Y,int(X),no_wf_available),
  Y = [H1|YY], YY=[H2], X==2, H1=int(0), H2=int(3) )).
:- assert_must_succeed((finite_cardinality_as_int_wf([A|Y],int(3),no_wf_available),
  Y = [B|YY], YY=[C], kernel_objects:equal_object([A,B,C], [int(1),int(3),int(2)])
   %,A=int(1),B=int(3),C=int(2) not possible because of ordered list skeleton setup below
   )).
:- assert_must_succeed((finite_cardinality_as_int_wf(Y,int(1),no_wf_available), Y = [fd(1,'Name')])).
:- assert_must_succeed((finite_cardinality_as_int_wf(Y,int(0),no_wf_available), Y = [])).
:- assert_must_succeed((finite_cardinality_as_int_wf(X,int(3),no_wf_available), kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_fail((finite_cardinality_as_int_wf(Y,int(X),no_wf_available), Y = [fd(1,'Name'),fd(2,'Name')],dif(X,2))).
:- assert_must_succeed_any((preferences:preference(use_clpfd_solver,false) ;
    finite_cardinality_as_int_wf(S,int(C),no_wf_available), clpfd_interface:force_post_constraint('#='(C,2)), nonvar(S),S=[A|T],nonvar(T),A=int(1),T=[int(2)])).
% does not work yet:
%:- assert_must_succeed_any((preferences:preference(use_clpfd_solver,false) ;
%    finite_cardinality_as_int_wf(S,int(C),no_wf_available), clpfd_interface:force_post_constraint('#>='(C,2)), nonvar(S),S=[_|T],nonvar(T))).
:- assert_must_succeed((finite_cardinality_as_int_wf([int(1)|avl_set(node(int(3),true,0,empty,empty))],int(2),no_wf_available))).
:- assert_must_succeed((finite_cardinality_as_int_wf([int(1)|avl_set(node(int(3),true,0,empty,empty))],X,no_wf_available),X==int(2))).
% check that we deal with repeated elements, in case no other predicate sets up a list !
:- assert_must_fail((finite_cardinality_as_int_wf([int(1),int(1)],int(2),no_wf_available))).
:- assert_must_fail((finite_cardinality_as_int_wf([int(1),int(1)],_,no_wf_available))).
:- assert_must_fail((finite_cardinality_as_int_wf(X,int(2),no_wf_available),X=[int(1),int(1)])).
:- assert_must_fail((finite_cardinality_as_int_wf([int(3)|avl_set(node(int(3),true,0,empty,empty))],_,no_wf_available))).
:- assert_must_fail((finite_cardinality_as_int_wf([X|avl_set(node(int(3),true,0,empty,empty))],int(2),no_wf_available),X=int(3))).

:- assert_must_succeed(( finite_cardinality_as_int_wf(V,int(X),no_wf_available), X in 1..5, finite_cardinality_as_int_wf(V,int(X2),no_wf_available), X2 in 5..10, X2==5 , V=[int(1),int(2),int(3),int(4),int(5)])).
:- assert_must_succeed(( finite_cardinality_as_int_wf(V,int(X),no_wf_available), X in 1..5, finite_cardinality_as_int_wf(V,int(X2),no_wf_available), X2 in 5..10, V=[_A,_B,_C,_D,_E|T], T==[],
   V=[int(1),int(2),int(3),int(4),int(5)])).


finite_cardinality_as_int_wf(Set,int(Card),WF) :-
   finite_card_wf(Set,0,inf,[contains_wd_condition],[],Card,WF).

finite_card_wf(Var,Acc,MaxCard,_Info,ElementsSoFar,Card,WF) :- var(Var),
    !,
    Card #= VCard+Acc,
    card_for_var(Var,VCard,ElementsSoFar,WF),
    check_no_duplicates(Var,ElementsSoFar,WF),
    restrict_card(Card,MaxCard).
finite_card_wf([],Acc,_,_Info,_,Card,_WF) :- !, Card=Acc.
finite_card_wf([H|T],Acc,MaxCard,Info,ElementsSoFar,Card,WF) :- !, A1 is Acc+1,
    check_and_add_new_element(H,ElementsSoFar,ElementsSoFar2,WF),
    finite_card_wf(T,A1,MaxCard,Info,ElementsSoFar2,Card,WF).
finite_card_wf(CS,Acc,MaxCard,Info,ElementsSoFar,Card,WF) :-
    explicit_set_cardinality_wf(CS,CSCard,WF), %print(cs_card(Acc,CSCard)),nl,
    add_cs_card(CS,CSCard,Acc,MaxCard,Info,Card,WF),
    disjoint_sets(ElementsSoFar,CS,WF).

:- block check_no_duplicates(-,?,?).
check_no_duplicates([],_,_) :- !.
check_no_duplicates([H|T],ElementsSoFar,WF) :- !,
    check_and_add_new_element(H,ElementsSoFar,ElementsSoFar2,WF),
    check_no_duplicates(T,ElementsSoFar2,WF).
check_no_duplicates(CS,ElementsSoFar,WF) :-
    disjoint_sets(CS,ElementsSoFar,WF).
% we have to check here that the list is really a set if we return cardinality
% sometimes the outer predicates do not ensure this
% e.g. for test 1828 we have that union(s) = {} & x<:INTEGER & x:s & card(s)=2 may otherwise succeed
% and generate a list with two identical elements: [ [],  [] ]
% ideally the rest of the kernel should ensure that such lists are not generated; we may generate some redundant work here
check_and_add_new_element(H,ElementSoFar,SoFar2,WF) :-
    not_element_of_wf(H,ElementSoFar,WF),
    add_new_element_wf(H,ElementSoFar,SoFar2,WF).

card_for_var(Var,Card,_ElementsSoFar,_WF) :- get_atts(Var,+kernel_card_attr(StoredCard,_)), % could be other WF !
   !,
   Card=StoredCard.
card_for_var(Var,Card,ElementsSoFar,WF) :- Card #>= 0, put_atts(Var,+kernel_card_attr(Card,WF)),
   force_empty_and_set_skel(Card,Var,ElementsSoFar,WF).

% Attribute unification:
% verify_attributes is called whenever a variable Var that might have attributes is about to be bound to Value (it might have none).
% verify(VariableBeforeUnification, ValueWithWhichVariableUnifies,  GoalWhichSICStusShouldCallAfterUnif)
verify_attributes(ThisVar,Value,Goal) :-
   get_atts(ThisVar,+kernel_card_attr(Card,WF)),!,
   update_card(Value,Card,Goal,WF).
verify_attributes(_, _, [] ).


update_card(Value,Card,Goal,WF) :- var(Value),!,
   (get_atts(Value,+kernel_card_attr(Card2,_)) % could be other WF
     -> Goal = [Card=Card2]
     ;  put_atts(Value,+kernel_card_attr(Card,WF)), Goal=[]
    ).
update_card([],Card,Goal,_WF) :- !, Goal=[Card=0].
update_card([_|T],Card,Goal,WF) :- !, get_list_tail(T,1,Tail,Delta),
   %print(update(T,Tail,Delta)),nl,
   Card #= C1+Delta, C1 #>= 0, force_empty(C1,Tail,WF),
   update_card(Tail,C1,Goal,WF).
update_card(CS,Card,Goal,WF) :-
   Goal = [safe_explicit_set_cardinality_wf(CS,Card,WF)].


get_list_tail(Tail,Acc,Tail,Res) :- var(Tail),!,Res=Acc.
get_list_tail([_|T],Acc,Tail,Res) :- !, A1 is Acc+1, get_list_tail(T,A1,Tail,Res).
get_list_tail(Tail,Acc,Tail,Res) :- !,Res=Acc.

:- block force_empty(-,-,?).
force_empty(Card,Set,WF) :- %print(emp(Card,Set)),nl,
    Card==0,
    !,
    empty_set_wf(Set,WF).
force_empty(_,_,_).


:- use_module(kernel_waitflags,[get_wait_flag/4]).
% wake up if Card is set and then force empty set or setup list skeleton
:- block force_empty_and_set_skel(-,-,?,?).
force_empty_and_set_skel(Card,Set,_ElementsSoFar,WF) :- % print(force_empty_and_set_skel(Card,Set)),nl,
    Card==0,
    !,
    empty_set_wf(Set,WF).
force_empty_and_set_skel(Card,Set,ElementsSoFar,WF) :- var(Set), number(Card),
    !,
    (preferences:preference(use_smt_mode,true)
      -> Prio is Card*2 % instantiate skeleton earlier, relevant for propagation in test 33
       ; Prio is Card*100),
    get_wait_flag(Prio,force_empty_and_set_skel(Set,Card),WF,LWF2),
    instantiate_set(Set,Card,ElementsSoFar,LWF2,WF).
force_empty_and_set_skel(_,_,_,_). % note hat verify_attributes will call update_card and check cardinality!


% instantiate a variable
:- block instantiate_set(-,?,?,-,?).
instantiate_set(Var,Card,ElementsSoFar,LWF2,WF) :- %print(instantiate_set(Var,Card,ElementsSoFar,LWF2)),nl,
   var(Var),!,
   (Card>1
     -> instantiate_var_set(Var,Card,ElementsSoFar,LWF2,WF)
     ;  Var = [H],
        not_element_of_wf(H,ElementsSoFar,WF)). %, print(instantiated_set(Var,Card,LWF2)),nl.
instantiate_set([H|T],Card,ElementsSoFar,LWF2,WF) :- !,
   (Card>1
    ->  check_and_add_new_element(H,ElementsSoFar,ElementsSoFar2,WF),
        C1 is Card-1,
        instantiate_set(T,C1,ElementsSoFar2,LWF2,WF)
     ; empty_set_wf(T,WF)).
instantiate_set(_,_,_,_,_WF). % TO DO: what if we have an open list skeleton? should we instantiate end?
  % note hat verify_attributes will call update_card and check cardinality!

instantiate_var_set(VarSet,Card,ElementsSoFar,_LWF2,WF) :-
   kernel_objects:unbound_variable_for_card(VarSet),
   !,
   kernel_objects:setup_ordered_list_skeleton(Card,Skel,closed,WF),
   disjoint_sets(ElementsSoFar,Skel,WF),
   %print(skel(Card,Skel)),nl,
   VarSet = Skel. % bypass equal_object: assign variable in one-go
instantiate_var_set(VarSet,Card,ElementsSoFar,LWF2,WF) :-
   C1 is Card-1, VarSet = [H|TVar],
   check_and_add_new_element(H,ElementsSoFar,ElementsSoFar2,WF),
   instantiate_set(TVar,C1,ElementsSoFar2,LWF2,WF).

:- public safe_explicit_set_cardinality_wf/3.
safe_explicit_set_cardinality_wf(CS,Card,WF) :-
   explicit_set_cardinality_wf(CS,CSCard,WF),  % TO DO: use finite version which itself raises WD error when inf and use clpfd addition
    add_cs_card(CS,CSCard,0,inf,[contains],Card,WF).


add_cs_card(_CS,CSCard,Acc,_MaxCard,_Info,Res,_WF) :- number(CSCard),!,
   Res is Acc+CSCard.
add_cs_card(_CS,CSCard,Acc,MaxCard,Info,Res,_WF) :-
   (clpfd_max_bounded(CSCard)
    ; var(CSCard),nonmember(contains_wd_condition,Info)),
   !,
   % no risk of CLPFD error with inf -> we can unify straightaway; better for propagation
   % We could also look at CS itself: e.g., interval closure: no risk of infinite card
   Res #= Acc+CSCard,
   restrict_card(Res,MaxCard).
add_cs_card(CS,CSCard,_Acc,_MaxCard,_Info,_Res,WF) :- CSCard == inf,!,
   add_wd_error('card applied to very large set, cardinality not representable in ProB: ',CS,WF).
add_cs_card(CS,CSCard,Acc,MaxCard,_Info,Res,WF) :-
   Res #>= Acc,
   restrict_card(Res,MaxCard),
   block_add_cs_card(CS,CSCard,Acc,Res,WF).

:- block block_add_cs_card(?,-,?,?,?).
block_add_cs_card(_CS,CSCard,Acc,Res,_WF) :- number(CSCard),!,Res #= Acc+CSCard.
block_add_cs_card(CS,_,_Acc,_Res,WF) :- add_wd_error('card applied to infinite (or very large) set: ',CS,WF).

:- use_module(tools,[max_tagged_integer/1]).
% restrict Card to be smaller or equal to MaxCard
restrict_card(Card,MaxCard) :- var(Card),
  number(MaxCard),
  max_tagged_integer(Max), MaxCard =< Max, % check if CLP(FD) overflow would occur when posting constraint
  !,
  Card #=< MaxCard.
restrict_card(_,_).



/*
| ?- finite_cardinality_as_int_wf(Var,C,WF), Var=[a,b,c|T], finite_cardinality_as_int_wf(T,CT,WF), C=int(3).
update([b,c|_1241],_1241,3)
emp([a,b,c|_3493],_2855)
emp(0,_3493)
Var = [a,b,c],
C = 3,
T = [],
CT = 0 ?
yes


*/


:- use_module(clpfd_interface,[clpfd_domain/3]).

% obtain the CLP(FD) variable bounds for a cardinality variable
clpfd_card_domain_for_var(Var,MinCard,MaxCard) :-
   var(Var),
   (get_atts(Var,+kernel_card_attr(StoredCard,_))
     -> clpfd_domain(StoredCard,MinCard,MaxCard) %, print(clpfd_domain(Var,MinCard,MaxCard)),nl
     ;  (MinCard,MaxCard) = (0,sup)
   ).

% -----------------

% to do: implement and ensure that it also works for infinite sets
%not_empty_set_wf(X,WF) :- var(X), !,
%   card_not_empty_set(X,WF).
%not_empty_set_wf([_|_]).

:- use_module(library(clpfd),[fd_min/2]).
% check if a variable is annotated with cardinality which implies that the set is not empty
clpfd_card_is_gt0_for_var(Var) :- var(Var),
   get_atts(Var,+kernel_card_attr(StoredCard,_)),
   fd_min(StoredCard,Low),
   number(Low), Low>0.

card_not_empty_set(Set,WF) :- var(Set),
   card_for_var(Set,Card,[],WF),
   Card #>0.

reify_empty_set_test(Set,PredRes,WF) :- var(Set),
   card_for_var(Set,Card,[],WF),
   (Card #> 0) #<=> R01,
   b_interpreter_check:prop_pred_01(PredRes,R01).

% ----------------------

