% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_freetypes,
          [registered_freetype/2
          ,reset_freetypes/0
          ,add_freetype/2
          ,enumerate_freetype/3
          ,max_cardinality_freetype/2
          ,set_freetype_depth/1
          ,is_a_freetype_wf/3
          ,freetype_cardinality/2
          ,is_infinite_freetype/1
          ,is_maximal_freetype/1
          ,is_empty_freetype/1
          ,is_non_empty_freetype/1
          ,test_empty_freetype/2
          ,get_freeval_type/3]).

/* This module contains predicates for the support for free types as they are used in Z specifications */

:- use_module(module_information,[module_info/2]).
:- module_info(group,proz).
:- module_info(description,'This module contains predicates that cope with the internal freetype. This is currently only used by the Z extension.').

:- use_module(library(lists)).

:- use_module(error_manager,[add_internal_error/2]).
:- use_module(debug,[debug_format/3]).
:- use_module(self_check).
:- use_module(preferences,[get_preference/2]).
:- use_module(kernel_objects,[basic_type/2,
                              enumerate_type/3,
                              max_cardinality/2,safe_add/3]).
:- use_module(typing_tools,[is_infinite_type/1]).


:- dynamic freetype_register_db/2, max_iteration_depth/1, marked_cases/2, max_cardinality_db/3.

/* freetype_register queries the database of registered free types for the list of cases */
registered_freetype(Id, Cases) :- freetype_register_db(Id,Cases).

/* the extended register contains for each case a mark if it is recursive or not,
   for internal use only (regarding the module) */
extended_register(Id, ECases) :- marked_cases(Id,Cs),!,ECases = Cs.
extended_register(Id, ECases) :-
	registered_freetype(Id,Cases),mark_recursive_cases(Cases,Id,ECases),
	assert(marked_cases(Id,ECases)).

/* reset_freetypes clears the database of registered free types */
reset_freetypes :-
	retractall(freetype_register_db(_,_)),
	retractall(marked_cases(_,_)),
	retractall(max_iteration_depth(_)),
	retractall(max_cardinality_db(_,_,_)).

/* with set_freetype_depth it is possible to set the maximum recursion depth of
   free types when enumerating them. If the depth is not set manually the
   depth is taken from the preference globalsets_fdrange. Mainly for testing */
set_freetype_depth(Depth) :-
	retractall(max_iteration_depth(_)),assert(max_iteration_depth(Depth)).

/* add_freetypes appends a new free type to the database of free types */
% example (test 1304):
% add_freetype('List'(T),[case(nil,constant([nil])),case(cons,couple(T,freetype(List(T))))]).
add_freetype(Id, Cases) :-
    debug_format(19,'Add freetype ~w with cases ~w~n',[Id,Cases]),
    ( is_caselist(Cases,Id) ->
        ( registered_freetype(Id,_) ->
            add_internal_error('freetype already registered: ', add_freetype(Id, Cases)),fail
        ; registered_freetype(OId,Cs),member(case(C,_),Cs),member(case(C,_),Cases) ->
            add_internal_error('freetype case registered twice: ', (C,OId,Id)),fail
        ; otherwise ->
            assert(freetype_register_db(Id,Cases))
        )
    ; otherwise ->
        add_internal_error('list of cases invalid', add_freetype(Id, Cases)),fail).

/* enumerates a free type with regard to the maximum recursive iteration of free types.
   The first argument is "basic" or "tight", depending on wich enumeration method
   should be used for the value of the included types. */
enumerate_freetype(EType,freeval(Id,C,Value),freetype(Id)) :-
	max_freetype_depth(MaxDepth),
	enumerate_freetype(EType,freeval(Id,C,Value),freetype(Id,MaxDepth)).
enumerate_freetype(EType,freeval(Id,C,Value),freetype(Id,Depth)) :-
	Depth > 0,
	extended_register(Id,Cases),
	member(case(C,CT,Mark),Cases),
	/* If the case is recursive we have to append information about the
	   maximum recursion depth to limit the size of the data structure.
           In the non-recursive case we leave the type untouched */
	(Mark = recursive ->
	    (NewDepth is Depth - 1, insert_recursion_depth(CT,NewDepth,NewCT));
	    NewCT = CT),
	enumerate_type(Value, NewCT, EType).


/* gives the maximumm cardinality of the free type with regard to the maximum
   recursive iteration of free types */
max_cardinality_freetype(freetype(Id),Card) :-
	max_freetype_depth(MaxDepth),
	max_cardinality_freetype(freetype(Id,MaxDepth),Card).
max_cardinality_freetype(freetype(Id,Depth),Card) :-
	(Depth > 0 ->
	    /* we save the computed cardinality for a certain maximum depth to
	       prevent exponential computation steps */
	    (max_cardinality_db(Id,Depth,Card) -> true;
		(NewDepth is Depth - 1,
		    extended_register(Id,Cases),
		    sum_cases_cardinalities(Cases,NewDepth,Card),
		    assert(max_cardinality_db(Id,Depth,Card))));
	    Card = 0).

sum_cases_cardinalities([],_,0).
sum_cases_cardinalities([case(_,Type,Mark)|Rest],Depth,Card) :-
	sum_cases_cardinalities(Rest,Depth,CRest),
	/* like in enumerate_freetype we distinguish between recursive and
	   non-recursive cases and add information about the maximum recursion
	   depth in the former case. */
	(Mark = recursive ->
	    insert_recursion_depth(Type,Depth,NewType);
	    NewType = Type),
	max_cardinality(NewType,C),
	add_with_inf(CRest,C,Card).

add_with_inf(inf,_,R) :- !,R=inf.
add_with_inf(_,inf,R) :- !,R=inf.
add_with_inf(A,B,R) :- R is A+B.

max_freetype_depth(Depth) :-
	max_iteration_depth(Depth) -> true; get_preference(globalsets_fdrange,Depth).

:- use_module(typing_tools,[valid_ground_type/1]).
/* Is the list a valid list of cases? */
is_caselist([],_).
is_caselist([case(Case,CaseType)|Rest],Id) :-
   atom(Case),
   (ground(Id) -> valid_ground_type(CaseType)
    ; true % TODO: we could check if variables are subset of variables of Id
   ),
   is_caselist(Rest,Id).


/* Appends to all freetype references a maximum depth */
insert_recursion_depth(freetype(Id),Depth,freetype(Id,Depth)) :- !.
insert_recursion_depth(set(A),Depth,set(B)) :- !,insert_recursion_depth(A,Depth,B).
insert_recursion_depth(seq(A),Depth,seq(B)) :- !,insert_recursion_depth(A,Depth,B).
insert_recursion_depth(couple(A,B),Depth,couple(C,D)) :- !,
	insert_recursion_depth(A,Depth,C),insert_recursion_depth(B,Depth,D).
insert_recursion_depth(record(Fields),Depth,record(RFields)) :- !,insert_recursion_depth_fields(Fields,Depth,RFields).
insert_recursion_depth(X,_,X).

insert_recursion_depth_fields([],_,[]).
insert_recursion_depth_fields([field(Name,Type)|Rest],Depth,[field(Name,RType)|RRest]) :-
	insert_recursion_depth(Type,Depth,RType),
	insert_recursion_depth_fields(Rest,Depth,RRest).


:- assert_must_succeed((kernel_freetypes:find_references([case(a,integer),
					 case(b,set(freetype(aa))),
					 case(c,couple(boolean,seq(freetype(bb)))),
					 case(d,record([field(f1,constant),
							field(f2,freetype(cc)),
							field(f3,freetype(aa)),
							field(f4,integer)]))],[cc,bb,aa]))).
/* finds all references to other free types */
find_references(Cases,Ids) :- find_references(Cases,[],Ids).
find_references([],Refs,Refs).
find_references([case(_,Type)|Rest],Acc,Refs) :- find_references_type(Type,Acc,TRefs),find_references(Rest,TRefs,Refs).
find_references_type(Type,Refs) :- find_references_type(Type,[],Refs).
find_references_type(freetype(Id),Acc,Res) :- !,(member(Id,Acc) -> Res=Acc ; Res=[Id|Acc]).
find_references_type(set(X),Acc,Refs) :- !, find_references_type(X,Acc,Refs).
find_references_type(seq(X),Acc,Refs) :- !, find_references_type(X,Acc,Refs).
find_references_type(couple(X,Y),Acc,Refs) :- !, find_references_type(X,Acc,XRefs),find_references_type(Y,XRefs,Refs).
find_references_type(record(Fields),Acc,Refs) :- !, find_references_fields(Fields,Acc,Refs).
find_references_type(_,Refs,Refs).
find_references_fields([],Refs,Refs).
find_references_fields([field(_,Type)|Rest],Acc,Refs) :-
	find_references_type(Type,Acc,TRefs),
	find_references_fields(Rest,TRefs,Refs).

/* find all references of a type (also indirect references) by breadth first search */
find_transitive_references(Type,Refs) :-
	find_references_type(Type,Refs1),
	find_transitive_references(Refs1,[],Refs).
find_transitive_references([],Refs,Refs).
find_transitive_references([Id|Rest],VisitedRefs,AllRefs) :-
	registered_freetype(Id,Cases),
	find_references(Cases,Refs),
	append([Id|Rest],VisitedRefs,FoundRefs),
	find_new_ids(Refs,FoundRefs,NewRefs),
	append(Rest,NewRefs,UnvisitedRefs),
	find_transitive_references(UnvisitedRefs,[Id|VisitedRefs],AllRefs).

find_new_ids([],_,[]).
find_new_ids([Id|Rest],Old,New) :-
	member(Id,Old),!,
	find_new_ids(Rest,Old,New).
find_new_ids([Id|Rest],Old,[Id|New]) :-
	find_new_ids(Rest,Old,New).

/* translate each case/2 into a case/3 where the third argument is "recursive" or "plain" */
mark_recursive_cases([],_,[]).
mark_recursive_cases([case(C,Type)|Rest],Id,[case(C,Type,Mark)|MRest]) :-
	(recursive_case(Id,Type) ->
	    Mark = recursive;
	    Mark = plain),
	mark_recursive_cases(Rest,Id,MRest).

recursive_case(Id,Type) :-
	find_transitive_references(Type,Refs),
	member(Id,Refs).

/* used by kernel_objects:check_element_of */
is_a_freetype_wf(freeval(Id,Case,Value),Id,_WF) :-
	registered_freetype(Id,Cases),
	is_a_freetype_wf2(Case,Cases,Value).
:- block is_a_freetype_wf2(-,?,?).
is_a_freetype_wf2(Case,Cases,Value) :-
	member(case(Case,CT),Cases),!,
	basic_type(Value,CT).

freetype_cardinality(Id,Card) :-
    extended_register(Id,Cases),
    % if there is a recursive case, the cardinality is infinite
    ( memberchk( case(_,_,recursive), Cases ) ->
        Card = inf
    ; otherwise ->
        sum_case_cardinality(Cases,0,Card)).
sum_case_cardinality([],Card,Card).
sum_case_cardinality([case(_Name,Type,_)|Crest],In,Out) :-
    max_cardinality(Type,TypeCard),
    safe_add(In,TypeCard,NewCard),
    sum_case_cardinality(Crest,NewCard,Out).

is_infinite_freetype(Id) :-
    extended_register(Id,Cases),
    has_infinite_case(Cases).
has_infinite_case([case(_Name,Type,Mark)|_]) :-
    is_infinite_case(Type,Mark),!.
has_infinite_case([_|Crest]) :-
    has_infinite_case(Crest).
is_infinite_case(_Type,recursive) :- !. % recursive types are always infinite
is_infinite_case(Type,_) :-
    is_infinite_type(Type).

is_maximal_freetype(_).
% we assume a freetype is always maximal (i.e., covers all possible values with its basic type)
% TO DO: Daniel check this !

is_empty_freetype(_) :- fail.
is_non_empty_freetype(_).
test_empty_freetype(_,pred_false).
% we assume all freetypes are not empty; TO DO: DANIEL please check

% get the basic type of a freeval(Id,Case,Val) for Val
get_freeval_type(Id,Case,CaseType):-
   registered_freetype(Id,Cases),
   member(case(Case,CaseType),Cases).
