% (c) 2012-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_strings, [empty_string/1,
                           string_length/2,
                           string_to_int_wf/4,
                           int_to_string/2,
                           int_to_dec_string/3,
                           string_is_int/2,
                           string_is_number/2,
                           string_is_decimal/2,
                           string_is_alphanumerical/2,
                           to_string/2,
                           string_append_wf/4,
                           string_reverse_wf/3,
                           string_split_wf/4,
                           string_join_wf/5,
                           string_chars/2,
                           string_codes/2,
                           string_to_uppercase/2,
                           string_to_lowercase/2,
                           string_equal_case_insensitive/3,
                           substring_wf/6,
                           format_to_string/3, convert_b_sequence_to_list_of_atoms/3,

                           % utilities:
                           split_atom_string/3,
                           generate_code_sequence/3
  ]).

% Strings in ProB are represented by terms of the form string(PrologAtom)

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(library(lists)).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides (external) functions to manipulate B strings.').

empty_string(string('')).

:- use_module(kernel_objects,[exhaustive_kernel_succeed_check/1,exhaustive_kernel_fail_check/1,
   exhaustive_kernel_check_wf/2,exhaustive_kernel_fail_check_wf/2]).

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_length(string(''),int(0)))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_length(string('a'),int(1)))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_length(string('aa'),int(2)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:string_length(string('a'),int(0)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:string_length(string('a'),int(2)))).

:- use_module(kernel_objects,[greater_than_equal/2]).

:- block string_length(-,-).
string_length(SA,int(L)) :-
   greater_than_equal(int(L),int(0)),
   string_len2(SA,L).
:- block string_len2(-,-).
string_len2(SA,L)  :-
   L==0,!,
   empty_string(SA).
string_len2(string(A),L) :-
   string_len3(A,L).
:- block string_len3(-,-).
string_len3(A,L)  :-
   L==0,
   !,
   empty_string_atom(A).
% in case A is not known and L=1 we could enumerate chars ??
string_len3(A,L) :-
   string_len4(A,L).
:- block string_len4(-,?).
string_len4(A,L) :-
   atom_length(A,LL), LL=L. % delay unification due to bug in SICStus atom_length
% bug in SICStus: dif(X,1), atom_length(a,X) succeeds in 4.2.0 and 4.2.1


% ----------------------------


:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_to_int_wf(string('11'),int(11),unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(kernel_strings:string_to_int_wf(string('11'),int(1),unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(kernel_strings:string_to_int_wf(string('1'),int(11),unknown,WF),WF)).
:- assert_must_fail((kernel_strings:string_to_int_wf(S,int(11),u,WF),kernel_strings:string_to_int_wf(S,int(12),u,WF))).

:- block string_to_int_wf(-,-,?,?).
string_to_int_wf(string(S),int(I),Span,WF) :-
   string_to_int2(S,I,Span,WF).

:- block string_to_int2(-,-,?,?).
string_to_int2(S,Res,Span,WF) :- var(S), % we know the number; we cannot construct the string as leading spaces/0s are ok
   % with an injective string_to_int conversion we could invert the function
   !,
   frozen(S,Goal),
   (incompatible_goal(Goal,S,Res) ->
     fail ; true),
   strint_to_int3(S,Res,Span,WF).
string_to_int2(S,Res,Span,WF) :- strint_to_int3(S,Res,Span,WF).

:- use_module(kernel_waitflags,[add_wd_error_set_result/6]).
:- block strint_to_int3(-,?,?,?).
strint_to_int3(S,Res,Span,WF) :-
    atom_codes(S,C),
    on_exception(error(syntax_error(_),_),
         integer_number_codes(C,S,Res,Span,WF),
         add_wd_error_set_result('### Could not convert string to integer: ',S,Res,0,Span,WF)).
         %add_error_and_fail(external_functions,'### Could not convert string to integer: ',S)),

integer_number_codes(C,S,Res,Span,WF) :-
  number_codes(Num,C),
  (integer(Num) -> Res=Num
   ; %add_error_and_fail(external_functions,'### String represents a floating point number (expected integer): ',S)).
     add_wd_error_set_result('### String represents a floating point number (expected integer): ',S,Res,0,Span,WF)).

% check if another pending co-routine transforms the same string into another number
incompatible_goal((A,B),S,Res) :-
   (incompatible_goal(A,S,Res) -> true ; incompatible_goal(B,S,Res)).
incompatible_goal(kernel_strings:strint_to_int3(S2,Res2,_,_),S,Res) :-
   number(Res2),
    S==S2, Res2 \= Res.

% ----------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_string(int(0),string('0')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_string(int(10),string('10')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_string(int(-10),string('-10')))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:int_to_string(int(0),string('1')))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:int_to_string(int(1),string('01')))).

% difference to string_to_int: do not throw error when string cannot be converted to integer

:- block int_to_string(-,?).
int_to_string(int(I),S) :- int_to_string2(I,S).

:- block int_to_string2(-,?).
int_to_string2(Num,Res) :-
   number_codes(Num,C),
   atom_codes(S,C), Res=string(S).


:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('0'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('-10'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('1267650600228229401496703205376'),pred_true))). %// 2^100
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('1.0'),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string(''),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('-'),pred_false))).

:- block string_is_int(-,?).
string_is_int(string(S),Res) :-
    string_is_int2(S,Res).

:- block string_is_int2(-,?).
string_is_int2(S,Res) :-
    atom_codes(S,C),
    on_exception(error(syntax_error(_),_),
           (number_codes(Num,C),
            (integer(Num)->Res=pred_true;Res=pred_false)),
         Res=pred_false).

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_number(string('0'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_number(string('-10'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_number(string('1267650600228229401496703205376'),pred_true))). %// 2^100
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_number(string('1.0'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_number(string(''),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_number(string('-'),pred_false))).

:- block string_is_number(-,?).
string_is_number(string(S),Res) :-
    string_is_number2(S,Res).

:- block string_is_number2(-,?).
string_is_number2(S,Res) :-
    atom_codes(S,C),
    on_exception(error(syntax_error(_),_),
           (number_codes(_Num,C),
             Res=pred_true),
         Res=pred_false).


:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string('0.0'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string('-10.2'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string('1267650600228229401496703205376.000'),pred_true))). %// 2^100
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string('01.10'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string('1.99'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string('1'),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string('.99'),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string('.'),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string(''),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_decimal(string('-'),pred_false))).

% test if we have a pure decimal number, with leading and trailing digits around the dot
:- block string_is_decimal(-,?).
string_is_decimal(string(S),Res) :-
    string_is_decimal2(S,Res).

:- block string_is_decimal2(-,?).
string_is_decimal2(S,Res) :-
    atom_codes(S,C),
    (is_dec_nr(C) -> Res=pred_true ; Res=pred_false).

is_dec_nr([H|T]) :- is_digit(H),!,is_dec_nr2(T).
is_dec_nr([45,H|T]) :-  % 45 = minus
    is_digit(H),is_dec_nr2(T).

is_dec_nr2([H|T]) :- is_digit(H),!,is_dec_nr2(T).
is_dec_nr2([46,D|T]) :- % 46 = dot
    is_digit(D),
    is_dec_nr3(T).

is_dec_nr3([]).
is_dec_nr3([H|T]) :- is_digit(H),is_dec_nr3(T).


is_digit(X) :- X>=48, X=<57.

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(0),int(1),string('0.0')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(21),int(1),string('2.1')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(121),int(2),string('1.21')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(101),int(2),string('1.01')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(-101),int(2),string('-1.01')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(-101),int(3),string('-0.101')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(121),int(0),string('121')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(121),int(-2),string('12100')))).

:- block int_to_dec_string(-,?,?), int_to_dec_string(?,-,?).
int_to_dec_string(int(I),int(Prec),S) :- int_to_dec_string2(I,Prec,S).

:- block int_to_dec_string2(-,?,?), int_to_dec_string2(?,-,?).
int_to_dec_string2(I,Prec,String) :-
    Prec=<0,
    !,
    IP is I * (10^abs(Prec)),
    int_to_string2(IP,String).
int_to_dec_string2(I,Prec,String) :- %Prec>0,
    PowTen is 10^Prec,
    IntVal is I // PowTen,
    number_codes(IntVal,IVC),
    ((IntVal=0, I<0) -> Prefix = [45|IVC] % need to add leading -
     ; Prefix = IVC),
    DecVal is abs(I) mod PowTen,
    number_codes(DecVal,DVC),
    length(DVC,Digits),
    NrZeros is Prec-Digits,
    length(Zeros,NrZeros),
    maplist(is_zero,Zeros),
    append(Zeros,DVC,Suffix),
    append(Prefix,[46|Suffix],Codes), % 46 is the dot .
    atom_codes(Atom,Codes),
    String = string(Atom).

is_zero(48). % ascii code of zero 0
% -------------------


:- use_module(kernel_tools,[ground_value_check/2]).

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:to_string(int(10),string('10')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:to_string(pred_true,string('TRUE')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:to_string([],string('{}')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:to_string(string('01'),string('01')))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:to_string(int(1),string('01')))).

:- block to_string(-,?).
to_string(int(I),S) :- !,
  int_to_string2(I,S).
to_string(string(S),Res) :- !,
  Res=string(S). % we already have a string; nothing needs to be done
to_string(Value,S) :- ground_value_check(Value,GrValue), to_string_aux(GrValue,Value,S).

:- block to_string_aux(-,?,?).
to_string_aux(_,Value,Str) :- to_string_aux(Value,Str).

:- use_module(preferences,[temporary_set_preference/3,reset_temporary_preference/2]).
:- use_module(translate,[translate_bvalue/2]).
% convert_to_avl
to_string_aux(Value,Str) :-
   normalise_value_for_to_string(Value,NValue),
   temporary_set_preference(expand_avl_upto,100000,CHNG),
   translate_bvalue(NValue,Atom),
   reset_temporary_preference(expand_avl_upto,CHNG),!,
   Str=string(Atom).
to_string_aux(Value,Str) :-
   add_internal_error('Translation to string failed: ',Value),
   trace, to_string_aux(Value,Str),
   Str=string('???').


:- use_module(store,[normalise_value_for_var/3]).
% normalise_value_for_var normalises values for storing; for printing we need to do less work
% e.g., we do not need to normalise AVL values; we could add further cases for records ...
normalise_value_for_to_string(avl_set(A),R) :- !, R=avl_set(A).
normalise_value_for_to_string((A,B),R) :- !, R=(NA,NB),
    normalise_value_for_to_string(A,NA),
    normalise_value_for_to_string(B,NB).
normalise_value_for_to_string(A,R) :- normalise_value_for_var(to_string,A,R).



% -------------------

:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_append_wf(string('a'),string('b'),string('ab'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_append_wf(string('0'),string('1'),string('01'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_append_wf(string('aa'),string(''),string('aa'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_append_wf(string(''),string('aa'),string('aa'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(kernel_strings:string_append_wf(string('0'),string('1'),string('1'),WF),WF)).

:- block string_append_wf(-,?,-,?), string_append_wf(?,-,-,?).
string_append_wf(string(A),string(B),string(C),WF) :-
  app2(A,B,C,WF).
:- block app2(-,?,-,?), app2(?,-,-,?).
app2(A,B,C,_WF) :-
   nonvar(A),nonvar(B),!,atom_concat(A,B,CC), /* overcome bug in SICStus; delay unification */
   CC=C.
app2(A,B,C,WF) :- atom_codes(C,CC), app3(A,B,CC,WF).

:- use_module(kernel_waitflags,[get_wait_flag/4]).
app3(A,B,CC,WF) :-
   (   nonvar(B) -> atom_codes(B,BB), append(AA,BB,CC), atom_codes(A,AA)
     ; nonvar(A) -> atom_codes(A,AA), append(AA,BB,CC), atom_codes(B,BB)
     ; length(CC,CLen),
       get_wait_flag(CLen,'STRING_APPEND',WF,LWF),
       app4(A,B,CC,WF,LWF)
    ).
:- block app4(-,?,-,?,-), app4(?,-,-,?,-).
app4(A,B,CC,WF,_) :- (nonvar(A) ; nonvar(B)), !, % no need to enumerate
   app3(A,B,CC,WF).
app4(A,B,CC,_WF,_) :- %print(enumerating(CC)),nl,
   append(AA,BB,CC), % will be non-deterministic
   atom_codes(A,AA), atom_codes(B,BB).

% -------------------------------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_reverse_wf(string('01001'),string('10010'),no_wf_available))).

:- block string_reverse_wf(-,-,?).
string_reverse_wf(string(A),string(B),_) :-
   string_reverse2(A,B).

:- block string_reverse2(-,-).
string_reverse2(A,B) :- nonvar(A),!, atom_codes(A,AA), reverse(AA,RA), atom_codes(B,RA).
string_reverse2(B,A) :- atom_codes(A,AA), reverse(AA,RA), atom_codes(B,RA).

% -------------------------------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_split_wf(string('01001'),string('1'),[(int(1),string('0')),(int(2),string('00')),(int(3),string(''))],no_wf_available))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_split_wf(string('789'),string('1'),[(int(1),string('789'))],no_wf_available))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:string_split_wf(string('aaa'),string('a'),[(int(1),string('a')),(int(2),string('a'))],no_wf_available))).

% function to split a string into a list of strings which were delimited by a separator
% WARNING: if the seperator is of length more than one, then first match-strategy will be used
:- block string_split_wf(?,-,?,?), string_split_wf(-,?,-,?).
string_split_wf(string(A),string(B),R,WF) :-
   string_split2(A,B,R,WF).
:- block string_split2(?,-,?,?),string_split2(-,?,-,?).
string_split2(Atom,Seperator,SplitAtomList,WF) :-
  (var(Atom) ; var(Seperator)),
  !, % currently : separator always known
  expand_custom_set_to_list_wf(SplitAtomList,ExpandedSplitAtomList,Done,string_split2,WF),
  string_split3(Atom,Seperator,SplitAtomList,Done,ExpandedSplitAtomList,WF).
string_split2(Atom,Separator,SplitAtomList,WF) :- % normal forward mode: atom and separator known:
  string_split_forward(Atom,Separator,SplitAtomList,WF).

string_split_forward(Atom,Separator,SplitAtomList,WF) :-
  split_atom_string(Atom,Separator,List), % safe_call ?
  convert_prolog_to_b_list(List,SplitAtomList,WF).

:- block string_split3(?,-,?,?,?,?),string_split3(-,?,?,-,?,?). % we need to know the seperator: TO DO : improve this
string_split3(Atom,Seperator,SplitAtomList,Done,_ExpandedSplitAtomList,WF) :-
   var(Done),
   !,
   string_split_forward(Atom,Seperator,SplitAtomList,WF).
string_split3(Atom,Seperator,SplitAtomList,_Done,ExpandedSplitAtomList,WF) :-
   ExpandedSplitAtomList \= [],  % split("",sep) --> [""] not the empty list; note: this is not a WD error, the constraint STRING_SPLIT(a,b) = [] is simply unsatisfiable
   !,
   sort(ExpandedSplitAtomList,SL),
   maplist(drop_index,SL,IL), % also: no WD error needs to be raised if this is not a sequence
   convert_b_to_prolog_atoms(IL,PL,Done),
   atom_codes(Seperator,SepCodes),
   append(SepCodes,_,SepCodes2),
   split4(Done,SepCodes2,PL,Seperator,Atom,SplitAtomList,WF).


:- use_module(probsrc(tools_strings),[ajoin/2]).

:- block split4(-,?,?,?,-,?,?), split4(-,?,?,-,?,?,?).
% unblock either when Done or when both Atom and Sperator known
split4(Done,_SepCodes2,_PL,Seperator,Atom,SplitAtomList,WF) :-
   var(Done),
   !,
   % we can now compute forwards anyhow; ignore backwards direction
   string_split_forward(Atom,Seperator,SplitAtomList,WF).
split4(_,SepCodes2,PL,Seperator,Atom,_,_) :-
      maplist(not_suffix_atom(SepCodes2),PL), % check that seperator occurs in no split atom: e.g. STRING_SPLIT(r,"_") = ["a","_","c"] should fail
       insert_sep(PL,Seperator,PL2),
       ajoin(PL2,Atom).

insert_sep([],_,[]).
insert_sep([H],_,R) :- !, R=[H].
insert_sep([H|T],Sep,[H,Sep|IT]) :- insert_sep(T,Sep,IT).

not_suffix_atom(SepCodes,Atom) :- \+ suffix_atom(SepCodes,Atom).
suffix_atom(SepCodes,Atom) :-
   atom_codes(Atom,AL),
   append(_,SepCodes,AL).

% ------------------------

:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_join_wf([(int(1),string('0')),(int(2),string('00')),(int(3),string(''))],string('1'),string('01001'),unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_join_wf([(int(1),string('0')),(int(2),string('00'))],string('1'),string('0100'),unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_join_wf([(int(1),string('0'))],string('1'),string('0'),unknown,WF),WF)).

:- block string_join_wf(?,-,?,?,?), string_join_wf(-,?,?,?,?).
string_join_wf(SplitAtoms,string(Sep),Res,Span,WF) :- string_join2(SplitAtoms,Sep,Res,Span,WF).

% this is not reversible  ["a","b"],"_"  ->  "a_b"  but ["a_b"],"_" -> "a_b" has same result
:- block string_join2(?,-,?,?,?),string_join2(-,?,?,?,?).
string_join2(SplitAtomList,Seperator,Result,Span,WF) :-
  expand_custom_set_to_list_wf(SplitAtomList,ExpandedSplitAtomList,Done,string_join2,WF),
  string_join3(Result,Seperator,SplitAtomList,Done,ExpandedSplitAtomList,Span,WF).

:- use_module(kernel_objects,[equal_object/3, equal_object_optimized_wf/4, equal_object_wf/4]).
:- block string_join3(?,-,?,?,?,?,?),string_join3(?,?,?,-,?,?,?). % we need to know the seperator: TO DO : improve this
string_join3(Result,Seperator,_SplitAtomList,_Done,ExpandedSplitAtomList,Span,WF) :-
   %ExpandedSplitAtomList \= [],   % commented out this means that STRING_JOIN([],sep) = ""
   % result of split("",sep) --> [""] : this is not the empty list; i.e., STRING_JOIN is then no longer injective
   !,
   sort(ExpandedSplitAtomList,SL),
   drop_index_with_seq_check(SL,1,IL,Span,WF),
   convert_b_to_prolog_atoms(IL,PL,Done2),
   when(nonvar(Done2),
     (insert_sep(PL,Seperator,PL2),
      ajoin(PL2,Atom),
      equal_object_optimized_wf(Result,string(Atom),string_join,WF))).
%string_join3(Result,_Seperator,SplitAtomList,_Done,[],Span,WF) :-
%   add_wd_error_set_result('### STRING_JOIN not defined for empty sequence: ',SplitAtomList,Result,string(''),Span,WF).

% -----------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_chars(string(''),[]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_chars(string('010'),[(int(1),string('0')),(int(2),string('1')),(int(3),string('0'))]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_chars(string('a'),[(int(1),string('a'))]))).

:- block string_chars(-,-).
string_chars(Str,SeqRes) :- nonvar(Str), Str=string(A), ground(A),!,
   string_chars2(A,SeqRes).
string_chars(Str,SeqRes) :- when((ground(Str);ground(SeqRes)),string_chars1(Str,SeqRes)).

string_chars1(Str,SeqRes) :- nonvar(Str), Str=string(A), ground(A),!,
   % construct sequence from string:
   string_chars2(A,SeqRes).
string_chars1(Str,SeqRes) :-
   expand_custom_set_to_list(SeqRes,ExpandedAtomList,Done,string_chars1),
   string_chars3(Str,SeqRes,ExpandedAtomList,Done).

:- use_module(kernel_objects,[equal_object_optimized/3]).
string_chars2(A,SeqRes) :- atom_codes(A,AA), generate_char_seq(AA,1,CharSeq),
   equal_object_optimized(CharSeq,SeqRes,string_chars2).

:- use_module(probsrc(tools_strings),[ajoin/2]).

:- block string_chars3(-,?,?,-).
string_chars3(Str,SeqRes,_ExpandedAtomList,_Done) :-
   % construct sequence from string:
   nonvar(Str), Str=string(A), ground(A),
   !,
   string_chars2(A,SeqRes).
string_chars3(Str,_SeqRes,ExpandedAtomList,Done) :-
   nonvar(Done),
   % construct string from sequence:
   !,
   sort(ExpandedAtomList,SL),
   maplist(drop_index,SL,IL),
   kernel_strings:convert_b_to_prolog_atoms(IL,PL,Done2),
   when(nonvar(Done2),
      (ajoin(PL,Atom),
       equal_object(Str,string(Atom),string_chars))).
string_chars3(Str,SeqRes,ExpandedAtomList,Done) :- % Str is only partially instantiated
   when((ground(Str);nonvar(Done)),string_chars3(Str,SeqRes,ExpandedAtomList,Done)).
generate_char_seq([],_,[]).
generate_char_seq([Code|T],Nr,[(int(Nr),string(CS))|TSeq]) :-
   atom_codes(CS,[Code]),
   N1 is Nr+1, generate_char_seq(T,N1,TSeq).

drop_index((int(_),R),R).

:- block drop_index_with_seq_check(-,?,?,?,?).
drop_index_with_seq_check([],_,[],_,_).
drop_index_with_seq_check([(int(Nr),R)|T],Expected,[R|TR],Span,WF) :-
   (Nr=Expected -> E1 is Expected+1, drop_index_with_seq_check(T,E1,TR,Span,WF)
     ; ajoin(['### Unexpected index: ',Nr,'! Argument for STRING_JOIN is not a sequence! Expected next index to be: '],Msg),
       add_wd_error_set_result(Msg,Expected,TR,[],Span,WF)
    ).

% ------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_codes(string(''),[]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_codes(string('010'),[(int(1),int(48)),(int(2),int(49)),(int(3),int(48))]))).

:- block string_codes(-,-).
string_codes(string(A),SeqRes) :- string_codes2(A,SeqRes).

:- use_module(custom_explicit_sets,[expand_custom_set_to_list/4, expand_custom_set_to_list_wf/5]).

:- block string_codes2(-,-).
string_codes2(A,SeqRes) :-
   nonvar(A),
   !,
   string_codes4(A,SeqRes).
string_codes2(A,SeqRes) :-
   SeqRes==[],
   !,
   empty_string_atom(A).
string_codes2(A,SeqRes) :- expand_custom_set_to_list(SeqRes,SeqList,_,string_codes2),
   when((nonvar(A);ground(SeqList)), string_codes3(A,SeqList)).

string_codes3(A,SeqRes) :-
  nonvar(A),
  !,
  string_codes4(A,SeqRes).
string_codes3(A,SeqRes) :-
   sort(SeqRes,SSeqRes),
   extract_codes(SSeqRes,1,Codes),
   atom_codes(A,Codes).
string_codes4(A,SeqRes) :-
   atom_codes(A,AA), generate_code_sequence(AA,1,CodeSeq),
   equal_object_optimized(CodeSeq,SeqRes,string_codes4).

generate_code_sequence([],_,[]).
generate_code_sequence([Code|T],Nr,[(int(Nr),int(Code))|TSeq]) :-
   N1 is Nr+1, generate_code_sequence(T,N1,TSeq).

extract_codes([],_,[]).
extract_codes([(int(Nr),int(Code))|T],N,[Code|CT]) :-
  (Nr==N -> true ; add_error(extract_codes,'Unexpected index: ',(Nr,N))),
  N1 is N+1, extract_codes(T,N1,CT).


empty_string_atom('').

% ------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_equal_case_insensitive(string(abCdAZ),string('ABcDAZ'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_equal_case_insensitive(string(a),string(aa),pred_false))).

% does not seem to be really faster than doing twice string_to_uppercase and comparing the result
% in principle we avoid building up two atoms for the upper case string
:- block string_equal_case_insensitive(-,?,?), string_equal_case_insensitive(?,-,?).
string_equal_case_insensitive(string(A),string(B),Res) :-
   str_eq_nocase(A,B,Res).

:- block str_eq_nocase(-,?,?), str_eq_nocase(?,-,?).
str_eq_nocase(A,A,Res) :- !, Res=pred_true.
%str_eq_nocase(A,B,Res) :- % performance not improved by this rule
% atom_length(A,L1), atom_length(B,L2), L1\=L2,
%   !, % in case upcase replaces one char by two we need to adapt this rule;
%      % check that atom_length deals with unicode chars correctly
%   Res=pred_false.
str_eq_nocase(A,B,Res) :-
   atom_codes(A,CA),
   atom_codes(B,CB),
   (l_eq_upcase(CA,CB) -> Res=pred_true ; Res=pred_false).

l_eq_upcase([],[]).
l_eq_upcase([H1|T1],[H2|T2]) :-
   (H1=H2 -> true ; upcase(H1,HU), upcase(H2,HU)),
   l_eq_upcase(T1,T2).


:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_to_uppercase(string(abcdAZ),string('ABCDAZ')))).

:- block string_to_uppercase(-,?).
string_to_uppercase(string(A),Res) :- string_to_uppercase2(A,Res).
% TO DO: add flag to make unicode/umlaut conversions optional

:- block string_to_uppercase2(-,?).
string_to_uppercase2(A,Res) :- atom_codes(A,AA),
  l_upcase(AA,UpCodes),
  atom_codes(AU,UpCodes),
  Res = string(AU).

l_upcase([],[]).
l_upcase([H|T],[HU|TU]) :- upcase(H,HU), l_upcase(T,TU).

upcase(223,R) :- !, R is "S". % ß
upcase(199,R) :- !, R is "C". % upper case ç
upcase(231,R) :- !, R is "C". % ç
upcase(208,R) :- !, R is "D". % special D
upcase(209,R) :- !, R is "N". % Ñ
upcase(241,R) :- !, R is "N". % ñ
upcase(221,R) :- !, R is "Y". % upper case ý
upcase(253,R) :- !, R is "Y". % ý
upcase(255,R) :- !, R is "Y". % ÿ

upcase(H,R) :- H<"a", !,R=H. % Z code is 90, a code is 97
upcase(H,R) :- H >="a", H=<"z", !, R is H+"A"-"a".
upcase(H,R) :- H >=192, H=<197, !, R is "A".  % upper-case A
upcase(H,R) :- H >=224, H=<229, !, R is "A".  % H >="à", H=<"å"
upcase(H,R) :- H >=200, H=<203, !, R is "E".  % upper-case E
upcase(H,R) :- H >=232, H=<235, !, R is "E".  % H >="è", H=<"ë"
upcase(H,R) :- H >=204, H=<207, !, R is "I".  % upper-case I
upcase(H,R) :- H >=236, H=<239, !, R is "I".  % H >="ì", H=<"ï"
upcase(H,R) :- H >=210, H=<214, !, R is "O".  % upper-case O
upcase(H,R) :- H >=242, H=<246, !, R is "O".  % H >="ò", H=<"ö"
upcase(H,R) :- H >=217, H=<220, !, R is "U".  % upper-case U
upcase(H,R) :- H >=249, H=<252, !, R is "U".  % H >="ù", H=<"ü"
upcase(H,R) :- H =< 255,!, R=H.
% some special variations of characters; there are a few chars in between which represent multiple chars ae,...
upcase(H,R) :- H >=256, H=<261, !, R is "A".
upcase(H,R) :- H >=262, H=<269, !, R is "C".
upcase(H,R) :- H >=270, H=<273, !, R is "D".
upcase(H,R) :- H >=274, H=<283, !, R is "E".
upcase(H,R) :- H >=284, H=<291, !, R is "G".
upcase(H,R) :- H >=292, H=<295, !, R is "H".
upcase(H,R) :- H >=296, H=<305, !, R is "I".
upcase(H,R) :- H >=308, H=<309, !, R is "J".
upcase(H,R) :- H >=310, H=<312, !, R is "K".
upcase(H,R) :- H >=313, H=<322, !, R is "L".
upcase(H,R) :- H >=323, H=<331, !, R is "N".
upcase(H,R) :- H >=332, H=<337, !, R is "O".
upcase(H,R) :- H >=340, H=<345, !, R is "R".
upcase(H,R) :- H >=346, H=<353, !, R is "S".
upcase(H,R) :- H >=354, H=<359, !, R is "T".
upcase(H,R) :- H >=360, H=<371, !, R is "U".
upcase(H,R) :- H >=372, H=<373, !, R is "W".
upcase(H,R) :- H >=374, H=<376, !, R is "Y".
upcase(H,R) :- H >=377, H=<382, !, R is "Z".
upcase(H,R) :- H >=384, H=<389, !, R is "B".
upcase(Code,Code).

% between:between(190,300,Char), format(' ~w = ~s~n',[Char,[Char]]),fail.
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_to_lowercase(string('ABCD-az'),string('abcd-az')))).

:- block string_to_lowercase(-,?).
string_to_lowercase(string(A),Res) :- string_to_lowercase2(A,Res).
% TO DO: add flag to make unicode/umlaut conversions optional

:- block string_to_lowercase2(-,?).
string_to_lowercase2(A,Res) :- atom_codes(A,AA),
  l_upcase(AA,UpCodes),
  l_lowcase(UpCodes,LowCodes),
  atom_codes(AU,LowCodes),
  Res = string(AU).

l_lowcase([],[]).
l_lowcase([H|T],[HU|TU]) :- simple_lowcase(H,HU), l_lowcase(T,TU).

simple_lowcase(H,R) :- H >="A", H=<"Z", !, R is H+"a"-"A".
simple_lowcase(Code,Code).



% ------------------------

:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:substring_wf(string(abcd),int(1),int(2),string(ab),unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:substring_wf(string(abcd),int(1),int(6),string(abcd),unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:substring_wf(string(abcd),int(4),int(6),string(d),unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:substring_wf(string(abcd),int(4),int(0),string(''),unknown,WF),WF)).


:- block substring_wf(-,?,?,?,?,?),substring_wf(?,-,?,?,?,?),substring_wf(?,?,-,?,?,?).
substring_wf(string(S),int(From),int(Len),Res,Span,WF) :-
    substring(S,From,Len,Res,Span,WF).

:- block substring(-,?,?,?,?,?),substring(?,-,?,?,?,?),substring(?,?,-,?,?,?).
substring(_,From,_Len,Res,Span,WF) :- From<1,!,
  add_wd_error_set_result('### From index for SUB_STRING must be positive: ',From,Res,string(''),Span,WF).
substring(S,From,Len,Res,_Span,_WF) :-
   PrefixLen is From-1, Length=Len,
   (Length < 1 -> empty_string_atom(ResAtom)
     ; atom_codes(S,Codes),
       (sublist(Codes, SelectedCodes, PrefixLen , Length, _)
         -> true
         ; sublist(Codes, SelectedCodes, PrefixLen , RealLength, 0),
           RealLength < Length
           -> true
         ; empty_string_atom(ResAtom) %  Deal with case that PrefixLen beyond length of string
       ),
       atom_codes(ResAtom,SelectedCodes)
   ),
   Res = string(ResAtom).




% ------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:format_to_string(string('abc'),[],string('abc')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:format_to_string(string('abc~wfg'),[(int(1),string('de'))],string('abcdefg')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:format_to_string(string('abc~wfg~w'),[(int(1),string('de')),(int(2),string('h'))],string('abcdefgh')))).

:- block format_to_string(-,?,?).
format_to_string(string(FormatString),BSeqOfValues,Res) :-
    convert_b_sequence_to_list_of_atoms(BSeqOfValues,ListOfAtoms,Done),
    format_to_string_aux(Done,FormatString,ListOfAtoms,Res).

:- use_module(library(codesio),[format_to_codes/3]).
:- block format_to_string_aux(-,?,?,?), format_to_string_aux(?,-,?,?).
format_to_string_aux(_,FormatString,ListOfAtoms,Res) :-
   format_to_codes(FormatString,ListOfAtoms,Codes),
   atom_codes(Atom,Codes),
   Res = string(Atom).

:- use_module(custom_explicit_sets,[is_set_value/2, expand_custom_set_to_list_gg/4]).

% convert a B sequence into a list of atoms; pretty printing if necessary
:- block convert_b_sequence_to_list_of_atoms(-,?,?).
convert_b_sequence_to_list_of_atoms(BSeqOfValues,Res,Done) :-
    is_set_value(BSeqOfValues,convert_b_sequence_to_list_of_atoms),
    !,
    expand_custom_set_to_list_gg(BSeqOfValues,ESet,GG,kernel_strings), % could use _gg version
    (GG=guaranteed_ground -> GrESet=true ; ground_value_check(ESet,GrESet)),
    convert_aux(GrESet,ESet,Res,Done).
convert_b_sequence_to_list_of_atoms(SingleValue,[S],Done) :-
    translate_bvalue(SingleValue,XS),
    add_warning(kernel_strings,'B sequence expected, obtained single value: ',XS),
    ground_value_check(SingleValue,GrValue), to_string_aux(GrValue,SingleValue,string(S)), Done=GrValue.

:- block convert_aux(-,?,?,?), convert_aux(?,-,?,?).
convert_aux(_,ESet,ListOfAtoms,Done) :-
   sort(ESet,SortedESet),
   maplist(get_string,SortedESet,ListOfAtoms),
   Done=true.

get_string((_,string(S)),R) :- !,R=S.
get_string((_,X),R) :- !,to_string_aux(X,string(R)).
get_string(X,R) :-
   translate_bvalue(X,XS),
   add_warning(kernel_strings,'B sequence expected, obtained set containing: ',XS),
   to_string_aux(X,string(R)).


:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_alphanumerical(string('a10'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_alphanumerical(string(''),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_alphanumerical(string('1.0'),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_alphanumerical(string('a_z'),pred_false))).

:- block string_is_alphanumerical(-,?).
string_is_alphanumerical(string(S),Res) :-
    string_is_alpha_aux(S,Res).
:- block string_is_alpha_aux(-,?).
string_is_alpha_aux(S,Res) :-
    atom_codes(S,Codes),
    (Codes \= [],
     is_alphnum_aux(Codes)
     -> Res=pred_true
      ; Res=pred_false).

is_alphnum_aux([]).
is_alphnum_aux([H|T]) :- is_alpha_numerical(H),is_alphnum_aux(T).

is_alpha_numerical(Code) :- Code >= 48, Code =< 57.  % 0-9
is_alpha_numerical(Code) :- Code >= 65, Code =< 90.  % A-Z
is_alpha_numerical(Code) :- Code >= 97, Code =< 122. % a-z

% ------------------------
% UTILITIES
% ------------------------


:- assert_must_succeed((kernel_strings:split_atom_string('ef,g',',',R), R==[ef,g])).
:- assert_must_succeed((kernel_strings:split_atom_string('ab,cd,ef,g',',',R), R==['ab','cd',ef,g])).
:- assert_must_succeed((kernel_strings:split_atom_string('ab','a',R), R==['','b'])).
:- assert_must_succeed((kernel_strings:split_atom_string('','a',R), R==[''])).
:- assert_must_succeed((kernel_strings:split_atom_string('STRING1','',R), R==['STRING1'])).
:- assert_must_succeed((kernel_strings:split_atom_string('mod274,mod276,mod277,mod282,mod283,mod284,mod285,mod286',',',R), R==[mod274,mod276,mod277,mod282,mod283,mod284,mod285,mod286])).

split_atom_string(Atom,Sep,SplitList) :-
  atom_chars(Sep,SepAscii),
  (SepAscii=[] -> SplitList = [Atom]
   ; SepAscii = [H|T], atom_chars(Atom,ListAscii),
     split3(ListAscii,H,T,Match,Match,SplitList)).

% MatchSoFar is passed in two variables: one to instantiate and one with the Result of the match
% this avoids calling reverse
split3([],_,_,MatchSoFarIn,MatchSoFarRes,R) :- !,
   MatchSoFarIn=[],  % match complete, ground tail of match
   atom_chars(Atom,MatchSoFarRes),R=[Atom].
split3([H|List],H,Sep,MatchSoFarIn,MatchSoFarRes,Res) :-
   append(Sep,Tail,List),
   !, % we have a match with a separator
   MatchSoFarIn=[], % match complete
   atom_chars(Atom,MatchSoFarRes),
   Res=[Atom|R2], split3(Tail,H,Sep,NewMatch,NewMatch,R2).
split3([H|T],HS,Sep,[H|MatchSoFarIn],MatchSoFarRes,Res) :- % no match
   split3(T,HS,Sep,MatchSoFarIn,MatchSoFarRes,Res).



% -----------------------
:- use_module(custom_explicit_sets,[try_expand_and_convert_to_avl/2]).

convert_prolog_to_b_list(PL,BL,WF) :-
  convert_prolog_to_b_list_aux(PL,1,CPL),
  try_expand_and_convert_to_avl(CPL,CPL2),
  equal_object_wf(CPL2,BL,convert_prolog_to_b_list,WF).


convert_prolog_to_b_list_aux([],_,[]).
convert_prolog_to_b_list_aux([H|T],Index,[(int(Index),CH)|CT]) :-
  convert_prolog_to_b_term(H,CH),
  I1 is Index+1, convert_prolog_to_b_list_aux(T,I1,CT).

convert_prolog_to_b_term(N,R) :-
   number(N),!,
   R=int(N).
convert_prolog_to_b_term(A,R) :-
   atomic(A),!,
   R=string(A).
convert_prolog_to_b_term(A,R) :-
   add_internal_error('Illegal Prolog term: ',convert_prolog_to_b_term(A,R)), R=A.


% a version that delays converting and sets Done to done when all B Atoms have been grounded
:- block convert_b_to_prolog_atoms(-,?,?).
convert_b_to_prolog_atoms([],[],done).
convert_b_to_prolog_atoms([BAtom|T],[PrologAtom|PT],Done) :-
   convert_b_to_prolog_atoms_aux(BAtom,PrologAtom,DoneAtom),
   convert_b_to_prolog_atoms(T,PT,DoneT),
   both_done(DoneAtom,DoneT,Done).

:- block both_done(-,?,?), both_done(?,-,?).
both_done(_,_,done).

:- block convert_b_to_prolog_atoms_aux(-,?,?).
convert_b_to_prolog_atoms_aux(pred_true,'TRUE',done).
convert_b_to_prolog_atoms_aux(pred_false,'FALSE',done).
convert_b_to_prolog_atoms_aux(string(S),PrologAtom,Done) :-
   convert_b_to_prolog_atoms_aux2(S,PrologAtom,Done).
convert_b_to_prolog_atoms_aux(int(S),PrologAtom,Done) :-
   convert_b_to_prolog_atoms_aux2(S,PrologAtom,Done).

:- block convert_b_to_prolog_atoms_aux2(-,?,?).
convert_b_to_prolog_atoms_aux2(Atom,Atom,done).
