% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(ltl_tools,[temporal_parser/3,temporal_parser_l/3,
                     parse_ltlfile/2,check_csp_event/4,
                     parse_and_pp_ltl_formula/2,
                     typecheck_temporal_formula/3,extract_formulas/4,
                     preprocess_formula/2,
                     preprocess_tp/2]).

/* SICSTUS Libraries */
:- use_module(library(lists)).

/* ProB Libraries */
% B
:- use_module(probsrc(bsyntaxtree), [predicate_identifiers/2, get_texpr_type/2]).
:- use_module(probsrc(bmachine),[b_type_expression/5, b_get_machine_operation/4]).
% CSP
:- use_module(probcspsrc(haskell_csp),[channel/2, channel_type_list/2,parse_single_csp_expression/3]).
% utility modules
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(debug)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(tools),[split_atom/3, string_concatenate/3]).
% state space modules
:- use_module(probsrc(state_space),[current_state_id/1]).
:- use_module(probsrc(specfile),[animation_mode/1,csp_mode/0, csp_with_bz_mode/0]).
% other
:- use_module(probsrc(parsercall),[call_ltl_parser/3, parse_expression/2,parse_predicate/2]).
/* Promela Tools */
%:- use_module(probpromela(promela_tools),[parse_promela_strings/4]).
/* ProB LTL modules */
:- use_module(probltlsrc(ltl_translate)).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides utility predicates for the LTL and CTL model checkers.').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parse an LTL or CTL formula
% Mode: ltl or ctl
temporal_parser(Formula,Mode,Result) :-
    temporal_parser_l([Formula],Mode,[Result]).

temporal_parser_l(Formulas,Mode,Result) :-
    % call the Java parser for LTL
    call_ltl_parser(Formulas,Mode,Res1),
    % check if everything was fine
    check_syntax_result_l(Res1,Parsed,ParseFlag,Mode), ParseFlag=ok,
    % typecheck atomic propositions and such things
    typecheck_temporal_formula_l(Parsed,Result,TypeFlag), TypeFlag=ok.

check_syntax_result_l([],[],_,_).
check_syntax_result_l([R|Rrest],[F|Frest],Res,Mode) :-
    check_syntax_result(R,F,Res,Mode),
    check_syntax_result_l(Rrest,Frest,Res,Mode).
check_syntax_result(ltl(F),F,_,_Mode).
check_syntax_result(syntax_error(E),_,error,Mode) :-
    add_error(Mode,'Syntax error: ',E).

typecheck_temporal_formula_l([],[],_).
typecheck_temporal_formula_l([F|Frest],[R|Rrest],Res) :-
    typecheck_temporal_formula(F,R,Res),
    typecheck_temporal_formula_l(Frest,Rrest,Res).
typecheck_temporal_formula(ap(AP),ap(Typed),Res) :-
    !,typecheck_ap(AP,Typed,Res).
typecheck_temporal_formula(action(TP),action(Typed),Res) :-
    !,typecheck_tp(TP,Typed,Res).
typecheck_temporal_formula(ena(TP,F),ena(Typed,R),Res) :-  % used in CTL only
    !,typecheck_tp(TP,Typed,Res),
    typecheck_temporal_formula(F,R,Res).
typecheck_temporal_formula(F,R,Res) :-
    F =.. [Functor|InArgs],
    same_length(InArgs,OutArgs),
    R =.. [Functor|OutArgs],
    typecheck_temporal_formula_l(InArgs,OutArgs,Res).

typecheck_ap(BasicAP,AP,_Res) :- is_basic_ap(BasicAP),!,AP=BasicAP.
typecheck_ap(enabled(TP),enabled(R),Res) :-
    !,typecheck_tp(TP,R,Res).
typecheck_ap(available(TP),available(R),Res) :-
    !,typecheck_tp(TP,R,Res).
typecheck_ap(weak_fair(TP),weak_fair(R),Res) :-
    !,typecheck_tp(TP,R,Res).
typecheck_ap(strong_fair(TP),strong_fair(R),Res) :-
    !,typecheck_tp(TP,R,Res).
typecheck_ap(dlk(TPList),dlk(R),Res) :-
    !,typecheck_tp_l(TPList,R,Res).
typecheck_ap(ctrl(TPList),ctrl(R),Res) :-
    !,typecheck_tp_l(TPList,R,Res).
typecheck_ap(det(TPList),det(R),Res) :-
    !,typecheck_tp_l(TPList,R,Res).
typecheck_ap(bpred(Raw),bpred(Typed),Res) :-
    b_type_expression(Raw,[variables],pred,Typed,Errors),
    add_all_perrors(Errors,[],typecheck_atomic_property),
    ( no_real_perror_occurred(Errors) -> true
    ; otherwise    -> Res=error).
typecheck_ap(unparsed_pred(Atom),xtl_predicate_check(Atom),ResultFlag) :-
    ( animation_mode(xtl)  -> true
    ; otherwise -> ResultFlag=error).

is_basic_ap(deadlock).
is_basic_ap(sink).
is_basic_ap(current).
is_basic_ap(stateid(_)).

typecheck_tp(unparsed_trans(Op),xtl(Opname,Arity,Args),ResultFlag) :-
    animation_mode(xtl), !,
    ( compound(Op) -> functor(Op,Opname,Arity), Op =.. [_|Args]
    ; split_atom(Op,['.'],[Opname|Args]) ->
        length(Args,Arity)
    ; otherwise ->
        add_error(ltl,'Failed to parse operation specification'),
        ResultFlag = error).
typecheck_tp(unparsed_trans(Opname),csp(Channel,Arity,ResArgs),ResultFlag) :-
    csp_mode,!,
    ( Opname = tick
        -> Channel=Opname, Arity=0, ResArgs=[] % should we set Arity ??
    ; Opname = tau
        -> Channel=Opname, Arity=0, ResArgs=[] % should we set Arity ??
    ; (split_atom(Opname,['.','!'],[Channel|Args]), channel(Channel,_))
        ->
         (channel_type_list(Channel, TypeList),length(TypeList,Arity),length(Args,Arity)
             -> (Arity = 0
                 -> ResArgs = []
                 ; (\+memberchk('_',Args) % we want to call the CSP-M parser only once when there is no '_' symbol in Args
                     -> mergeArgsToDotTuple([Channel|Args],DotTupleEvent),
                        parse_single_csp_expression(ltl,DotTupleEvent,ParsedEvent),
                        splitToSingleArgs(ParsedEvent,[Channel|ResArgs],Arity)
                     ;  l_parse_single_csp_expression_undescore(ltl,Args,ResArgs)
                     )
                )
             ;  add_error(ltl,'Number of arguments not correct for: ', Opname),
                ResultFlag=error, ResArgs=[], Arity = 0
         )
    ; otherwise
        -> add_error(ltl,'Channel does not exist: ',Opname),
           ResultFlag = error, ResArgs=[], Arity = 0
    ).
typecheck_tp(bop(oppattern(_Pos,Name,Args)),Result,ResultFlag) :-
    ( b_get_machine_operation(Name,Results,Parameters,_) ->
        Result = bop(Name,NumberArgs,NumberResults,ArgPatterns,ResPatterns),
        length(Parameters,NumberArgs), length(Results,NumberResults),
        % Pattern matching for results is not yet supported
        ResPatterns = [],
        ( Args=[] ->
            ArgPatterns = [],
            ResPatterns = []
        ; length(Args,NumberArgs) ->
            create_b_op_patterns(Args,Parameters,ArgPatterns,ResultFlag)
        ; otherwise ->
            add_error(ltl,'Number of arguments not correct for: ', Name),ResultFlag=error)
    ; csp_mode, Args=[] ->
        debug_println(9,unparsed_trans(Name)),
        typecheck_tp(unparsed_trans(Name),Result,ResultFlag)
    ; otherwise ->
        add_error(ltl,'Could not find operation: ', Name),ResultFlag=error).
typecheck_tp(btrans(event(Name)),btrans(event(Name)),_ResultFlag) :- !.
typecheck_tp(btrans(event(Name,RawPredicate)),btrans(event(Name,TPredicate,Poststate)),ResultFlag) :-
    ( b_get_machine_operation(Name,_,_Parameters,_) ->
        ( bmachine:type_with_errors(RawPredicate,[prepost,operation(Name)],pred,TPredicate) ->
            find_poststate_access(TPredicate,Poststate)
        ; otherwise -> ResultFlag=error)
    ; otherwise ->
        add_error(ltl,'Could not find operation: ', Name),ResultFlag=error).

typecheck_tp_l([],[],_Res).
typecheck_tp_l([TP|TPs],[TTP|TTPs],Res) :-
    typecheck_tp(TP,TTP,ResultFlag),
    (ResultFlag == error -> Res = error
    ; otherwise -> Res=Res1),
    typecheck_tp_l(TPs,TTPs,Res1).

:- use_module(probsrc(bsyntaxtree),[syntaxtraversion/6]).
find_poststate_access(TExpr,Poststate) :-
    find_poststate_access2(TExpr,P,[]),
    sort(P,Poststate).
find_poststate_access2(TExpr) -->
    {syntaxtraversion(TExpr,Expr,_,Infos,Subs,_)},
    find_poststate_access3(Expr,Infos),
    find_poststate_access_l(Subs).
find_poststate_access3(identifier(Id),Infos) -->
    {memberchk(poststate,Infos),!,atom_concat(Orig,'\'',Id)},
    [poststate(Orig,Id)].
find_poststate_access3(_,_) --> [].
find_poststate_access_l([]) --> !.
find_poststate_access_l([TExpr|Rest]) -->
    find_poststate_access2(TExpr),
    find_poststate_access_l(Rest).

/* Parsing CSP expressions where '_' will be not parsed */
parse_single_csp_expression_undescore(Context,Expr,Res) :-
    (Expr=='_' -> Res = Expr; parse_single_csp_expression(Context,Expr,Res)).

l_parse_single_csp_expression_undescore(Context,LExpr,LRes) :-
    maplist(parse_single_csp_expression_undescore(Context),LExpr,LRes).
/*******************************************************/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% do some preprocessing on a formula:
%  - insert current state id
%  - rewrite CSP declarations in some cases

preprocess_formula(In,Out) :-
    ( preprocess_formula2(In,Out) -> true
    ; otherwise ->
        add_error(ltl,'preprocessing formula failed: ',In),fail).

:- assert_must_succeed(preprocess_formula2(globally(next(false)),false)).
:- assert_must_succeed(preprocess_formula2(globally(next(next(true))),globally(true))).
:- assert_must_succeed(preprocess_formula2(finally(next(true)),true)).
:- assert_must_succeed(preprocess_formula2(finally(next(next(false))),finally(false))).

% base cases
preprocess_formula2(ap(A),ap(B)) :- !,
    (A=bpred(Pred) ->
        predicate_identifiers(Pred,Ids),
        debug_println(predicate_identifiers(A,Ids))
    ; true),
    preprocess_ap(A,B).
preprocess_formula2(action(A),action(B)) :- !,preprocess_tp(A,B).
preprocess_formula2(ena(SpecA,FA),action(SpecB,FB)) :-
    !,preprocess_tp(SpecA,SpecB),preprocess_formula2(FA,FB).
% (X true = true) and (X false = false)
preprocess_formula2(next(F),PNextF) :-
    !,preprocess_formula2(F,PF),
    (PF = true  -> PNextF = true
    ;PF = false -> PNextF = false
    ; otherwise -> PNextF = next(PF)).
% idempotency laws
preprocess_formula2(globally(globally(F)),globally(PF)) :-
    !,preprocess_formula2(F,PF).
preprocess_formula2(finally(finally(F)),finally(PF)) :-
    !,preprocess_formula2(F,PF).
preprocess_formula2(until(F,until(F,G)),until(PF,PG)) :-
    !,preprocess_formula2(F,PF),
      preprocess_formula2(G,PG).
preprocess_formula2(until(until(F,G),G),until(PF,PG)) :-
    !,preprocess_formula2(F,PF),
      preprocess_formula2(G,PG).
% absorption laws
preprocess_formula2(finally(globally(finally(F))),globally(finally(PF))) :-
    !,preprocess_formula2(F,PF).
preprocess_formula2(globally(finally(globally(F))),finally(globally(PF))) :-
    !,preprocess_formula2(F,PF).
% F (f U g) == F g
preprocess_formula2(finally(until(_F,G)),finally(PG)) :-
    !,preprocess_formula2(G,PG).
% G (f R g) == G g
preprocess_formula2(globally(release(_F,G)),globally(PG)) :-
    !,preprocess_formula2(G,PG).
% F true = true
preprocess_formula2(finally(F),PFinallyF) :-
    !,preprocess_formula2(F,PF),
    (PF = true  -> PFinallyF = true
    ; otherwise -> PFinallyF = finally(PF)).
% G false = false
preprocess_formula2(globally(F),PGloballyF) :-
    !,preprocess_formula2(F,PF),
    (PF = false  -> PGloballyF = false
    ;  otherwise -> PGloballyF = globally(PF)).
% otherwise
preprocess_formula2(Term,NewTerm) :-
    functor(Term,Functor,Arity),
    functor(NewTerm,Functor,Arity),
    preprocess_formula_args(Arity,Term,NewTerm).

preprocess_formula_args(0,_,_) :- !.
preprocess_formula_args(N,Term,NewTerm) :-
    arg(N,Term,Arg),
    arg(N,NewTerm,NewArg),
    preprocess_formula2(Arg,NewArg),
    N2 is N-1,
    preprocess_formula_args(N2,Term,NewTerm).

preprocess_ap(current,stateid(StateId)) :- !,
    current_state_id(StateId).
preprocess_ap(enabled(A),enabled(B)) :- !,
    preprocess_tp(A,B).
preprocess_ap(AP,AP).

preprocess_tp(csp(Channel,Arity,CspArgs),
              bop(Channel,NumberArgs,NumberResults,ArgPatterns,ResPatterns)) :-
    csp_with_bz_mode,
    b_get_machine_operation(Channel,Results,Parameters,_Body),
    !,
    % check if more arguments are specified that allowed
    length(Parameters,NumberArgs),
    length(Results,NumberResults),
    MaxNumberAllowed is NumberArgs + NumberResults,
    ( Arity > MaxNumberAllowed ->
        add_error(ltl,'Too many arguments for LTL/CTL channel: ', Channel),fail
    ; Arity =< NumberResults ->
        ParamArgs = CspArgs,ResArgs = []
    ; otherwise ->
        MinArgs is min(NumberArgs,Arity),
        append_length(ParamArgs,ResArgs,CspArgs,MinArgs)),
    csp_arguments(ParamArgs,1,Parameters,ArgPatterns,Exprs1),
    csp_arguments(ResArgs,1,Results,ResPatterns,Exprs2),
    append(Exprs1,Exprs2,Exprs),
    parse_preds_exprs(Exprs).
preprocess_tp(TP,TP).

csp_arguments([],_Index,_Parameters,[],[]).
csp_arguments([P|ParamArgs],Index,[Param|ParamRest],ArgPatterns,Exprs) :-
    ( is_csp_placeholder(P) ->
        ArgPatterns = ArgPatternsRest,
        Exprs = ExprsRest
    ; otherwise ->
        ArgPatterns = [Index/ArgPattern|ArgPatternsRest],
        Exprs = [Expr|ExprsRest],
        csp_argument(P,Param,ArgPattern,Expr)),
    I2 is Index+1,
    csp_arguments(ParamArgs,I2,ParamRest,ArgPatternsRest,ExprsRest).
csp_argument(Argument,Parameter,ParsedExpr,expr(ToParse,Type,ParsedExpr)) :-
    get_texpr_type(Parameter,Type),
    atom_codes(Argument,ToParse).

is_csp_placeholder('_').
is_csp_placeholder('?').

% currently not called: nor is unparsed_ap called
%translate_preds(In,Out) :-
%    unparsed_ap_l(In,Out,Translation,[]),
%    parse_preds_exprs(Translation).

parse_preds_exprs([]) :- !. % nothing to parse
parse_preds_exprs(Translation) :-
    split_preds(Translation,PredStrings,Preds,ExprStrings,Exprs),
    animation_mode(AnimMode),
    parse_preds_exprs2(AnimMode,PredStrings,Preds,ExprStrings,Exprs).

% default: parse the predicates and expressions as B
parse_preds_exprs2(_,PredStrings,Preds,ExprStringsWithTypes,Exprs) :-
    annotate_ap_type(Preds,bpred,BPreds),
    split_exprs_types(ExprStringsWithTypes,ExprStrings,Types),
    maplist(parse_expression,ExprStrings,UExprs),
    maplist(parse_predicate,PredStrings,UPreds),
    type_exprs(UExprs,Types,Exprs,EErrors),
    type_preds(UPreds,BPreds,PErrors),
    append(EErrors,PErrors,Errors),
    add_all_perrors(Errors,[],type_error).

split_exprs_types([],[],[]).
split_exprs_types([Expr/Type|ETRest],[Expr|ERest],[Type|TRest]) :-
    split_exprs_types(ETRest,ERest,TRest).

type_exprs([],[],[],[]).
type_exprs([Expr|Erest],[Type|TypeRest],[Typed|Trest],Errors) :-
    b_type_expression(Expr,[variables],Type,Typed,LErrors),
    append(LErrors,ErrorRest,Errors),
    type_exprs(Erest,TypeRest,Trest,ErrorRest).
type_preds([],[],[]).
type_preds([Pred|Prest],[Typed|Trest],Errors) :-
    b_type_expression(Pred,[variables],pred,Typed,LErrors),
    append(LErrors,ErrorRest,Errors),
    type_preds(Prest,Trest,ErrorRest).

annotate_ap_type([],_,[]).
annotate_ap_type([AP|APrest],Type,[R|Rrest]) :-
    functor(AP,Type,1),arg(1,AP,R),
    annotate_ap_type(APrest,Type,Rrest).



split_preds([],[],[],[],[]).
split_preds([pred(A,B)|Rest],[A|ARest],[B|BRest],X,Y) :- !,
   split_preds(Rest,ARest,BRest,X,Y).
split_preds([expr(A,T,B)|Rest],X,Y,[A/T|ARest],[B|BRest]) :-
   split_preds(Rest,X,Y,ARest,BRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parse a file with several LTL formulas:

parse_ltlfile(Filename, Formulas) :-
    ( parse_ltlfile2(Filename,Formulas) -> true
    ; otherwise ->
        add_error(ltl, 'Error while parsing the LTL formula file')).
parse_ltlfile2(Filename, Formulas) :-
    read_file(Filename,Codes),
    parse_lines(Lines,Codes,[]),
    parse_sections(Raws,Lines,[]),
    parse_file_formulas(Raws,Formulas).
parse_file_formulas(In,Out) :-
    extract_formulas(In,Strings,Parsed,Out),
    temporal_parser_l(Strings,ltl,Parsed).

extract_formulas([],[],[],[]).
extract_formulas([formula(Name,Raw)|Rin],[Raw|Rraw],
                 [Parsed|Rparsed],[formula(Name,Parsed)|Rout]) :-
    extract_formulas(Rin,Rraw,Rparsed,Rout).

/*
parse_file2([]) --> [].
parse_file2([F|Rest]) -->
    parse_section(F),
    parse_file2(Rest). */

parse_sections([formula(Name,F)|Rest]) -->
    munch_comments,
    [header(Name)],
    parse_section_formulas("",FCodes),!,
    { atom_codes(F,FCodes) },
    parse_sections(Rest).
parse_sections([]) --> [].

munch_comments --> [comment(_)],!,munch_comments.
munch_comments --> [].

parse_section_formulas(Prev,F) -->
    munch_comments,
    [formula(Snippet)],!,
    {(Prev==[] -> S2=Snippet; append("\n",Snippet,S2) ),append(Prev,S2,F2)},
    parse_section_formulas(F2,F).
parse_section_formulas(F,F) --> munch_comments,[].

parse_lines([]) --> [].
parse_lines([L|Rest]) -->
    parse_line(L),
    parse_lines(Rest).

parse_line(comment(C)) -->
    pl_spaces,
    "#",!,
    pl_until_newline(C).
parse_line(comment("")) -->
    pl_spaces,
    pl_newline,!.
parse_line(header(Title)) -->
    pl_title(Title),
    pl_spaces, pl_newline, !.
parse_line(formula([C|F])) -->
    % no empty formulas
    [C],
    pl_until_newline(F).

pl_title(Title) -->
    "[", pl_spaces, pl_title2(T), {atom_codes(Title,T)}.
pl_title2([]) --> pl_spaces,"]",!.
pl_title2([]) --> pl_newline,!,{fail}. % maybe an exception?
pl_title2([C|Rest]) --> [C],pl_title2(Rest).

pl_spaces --> " ",!,pl_spaces.
pl_spaces --> "\t",!,pl_spaces.
pl_spaces --> [].

pl_until_newline([]) --> pl_newline,!.
pl_until_newline([C|Rest]) --> [C],!,pl_until_newline(Rest).
pl_until_newline([]) --> [].

pl_newline --> "\n\r",!.
pl_newline --> "\n",!.
pl_newline --> "\r".

read_file(Filename,Codes) :-
    open(Filename,read,S,[encoding('UTF-8')]),
    read_codes(S,Codes),
    close(S).
read_codes(S,Codes) :-
    get_code(S,Code),
    ( Code < 0 ->
        Codes = []
    ; otherwise ->
        Codes = [Code|Rest],
        read_codes(S,Rest)).

create_b_op_patterns(Args,Parameters,ArgPatterns,Res) :-
    create_b_op_patterns2(Args,Parameters,ArgPatterns,1,Res).
create_b_op_patterns2([],[],[],_,_).
create_b_op_patterns2([Arg|ARest],[Param|ParamRest],Patterns,Index,Res) :-
    create_b_op_pattern(Arg,Param,Pattern,Res),
    ( var(Pattern) -> Patterns = PRest
    ; otherwise -> Patterns = [Index/Pattern|PRest]),
    Index2 is Index+1,
    create_b_op_patterns2(ARest,ParamRest,PRest,Index2,Res).
create_b_op_pattern(undef(_Pos),_,_,_).
create_b_op_pattern(def(_Pos,Raw),TParam,Typed,Res) :-
    get_texpr_type(TParam,Type),
    b_type_expression(Raw,[variables],Type,Typed,Errors),
    add_all_perrors(Errors,[],typecheck_operation_argument),
    ( no_real_perror_occurred(Errors) -> true
    ; otherwise    -> Res=error).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


parse_and_pp_ltl_formula(FormulaAsAtom,Text) :-
   temporal_parser(FormulaAsAtom,ltl,Ltl),
   ( pp_ltl_formula_classicb(Ltl,Text)
      -> true
   ; add_error_fail(ltl, 'Failed to pretty print LTL formula: ', Ltl)
   ).

:- use_module(probsrc(specfile),[unset_animation_minor_modes/1,reset_animation_minor_modes/1]).
:- use_module(probsrc(translate),[set_unicode_mode/0, unset_unicode_mode/0]).
pp_ltl_formula_classicb(Ltl,Text) :-
   unset_animation_minor_modes(L), % unset eventb mode, as we will later parse the LTL formula in classic mode!
   % TO DO: improve and ensure things like finite(.) from Event-B work in atomic propositions
   set_unicode_mode,
   call_cleanup(pp_ltl_formula(Ltl,Text),
                  (reset_animation_minor_modes(L),unset_unicode_mode)).

%%%%%%%% Other predicates %%%%%%%%%%

:- assert_must_succeed((ltl_tools: mergeArgsToDotTuple([1,2,'<bob,alice>','{bob}'],Res), Res = '1.2.<bob,alice>.{bob}')).
:- assert_must_succeed((ltl_tools: mergeArgsToDotTuple(['(bob,alice)'],Res), Res = '(bob,alice)')).
:- assert_must_succeed(ltl_tools: mergeArgsToDotTuple([],'')).

mergeArgsToDotTuple(List,Res) :- mergeArgsToDotTuple(List,'',Res).

mergeArgsToDotTuple([],Str,Str).
mergeArgsToDotTuple([H|T],Str,Res) :-
        (T\=[] -> string_concatenate(H,'.',HRes) ; HRes=H),
         string_concatenate(Str,HRes,StrRes),mergeArgsToDotTuple(T,StrRes,Res).

splitToSingleArgs(dotTuple(List),List,Arity) :-
    N is Arity +1,
    length(List,N).
splitToSingleArgs(X,[X],0).
splitToSingleArgs(X,_L,N) :-
    add_internal_error('Internal Error: Arity does not match for channel: ',X/N).

:- assert_must_succeed((check_csp_event(tick,0,[],R), R=tick(_))).
:- assert_must_succeed((check_csp_event(tau,0,[],R), R=tau(_))).
:- assert_must_fail(check_csp_arguments([setValue([alice,bob]),tupleExp([s,t,u])], [setValue([alice,bob]),in(na_tuple([s,t]))],some_event)).
:- assert_must_succeed(check_csp_arguments([setValue([alice,bob]),tupleExp([s,t,u])], [setValue([alice,bob]),in(na_tuple([s,t,u]))],some_event)).
:- assert_must_succeed(check_csp_arguments([setValue([bob,alice]),tupleExp([s,t,u])], [setValue([alice,bob]),in(na_tuple([s,t,u]))],some_event)).
:- assert_must_succeed(check_csp_arguments([a,'_','_'],[in(a),in(int(10)),in(int(20))],true)).

%% Predicate check csp event
check_csp_event(tick,0,[],R) :- !, R=tick(_).
check_csp_event(tau,0,[],R) :- !, R=tau(_).
check_csp_event(Event,Arity,PatArgs,io(Vals,Event,_SrcSpan)) :- !,
    (channel(Event,_) ->
        ((channel_type_list(Event,List),length(List,Arity)) ->
            check_csp_arguments(PatArgs,Vals,Event)
        ; otherwise -> add_error(ltl,'Internal Error: Missing/extra arguments in CSP LTL/CTL pattern: ',Event:PatArgs),fail
        )
    ; otherwise -> add_internal_error('Internal Error: Channel not defined: ',Event/Arity)).

:- block check_csp_arguments(?,-,?).
check_csp_arguments([],[],_).
check_csp_arguments([],[H|T],Event) :-
  print('* Missing arguments in CSP LTL/CTL pattern (use .?): '),print(Event),
  print(' : '),print([H|T]),nl.
check_csp_arguments([H|T],[],Event) :-
  add_error(ltl,'Extra arguments in CSP LTL/CTL pattern: ',Event:[H|T]),fail.
check_csp_arguments([H|T],[PatH|PatT],Event) :-
  check_csp_arg(H,PatH),!,
  check_csp_arguments(T,PatT,Event).

check_csp_arg(X,Y) :-
    var(Y),!,
    add_internal_error('Internal Error: Event argument was a variable, expected to be: ',X).
check_csp_arg('_',_Y). % underscore should match any argument
check_csp_arg(X,in(Y)) :-
    check_csp_arg(X,Y).
check_csp_arg(X,out(Y)) :-
    check_csp_arg(X,Y).

check_csp_arg(list(L),list(L1)) :- !,
  l_check_csp_arg(L,L1,list).
check_csp_arg(setValue(L),setValue(L1)) :- !,
  sort(L,R),
  sort(L1,R1),
  l_check_csp_arg(R,R1,setValue). % elements order is not important within set expressions
check_csp_arg(tupleExp(L),na_tuple(L1)) :- !,
  l_check_csp_arg(L,L1,na_tuple).
check_csp_arg(X,Y) :- eq_csp_val(Y,X).

eq_csp_val(int(X),int(Y)) :- !, Y=X.
eq_csp_val(X,X).

l_check_csp_arg([],[],_):- !.
l_check_csp_arg([],[H|T],Functor) :-
    add_error_fail(ltl,'Missing elements in CSP LTL pattern: ',Functor:[H|T]).
l_check_csp_arg([H|T],[H1|T1],Functor) :-
    check_csp_arg(H,H1),!,
    l_check_csp_arg(T,T1,Functor).
