% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(memoization,[store_memo_expansion/3,
                       get_stored_memo_expansion/3,
                       compute_memo_hash/2,
                       store_memo_computation_time/2, % for profiling

                       register_memoization_function/6,
                       get_registered_function/2,
                       is_memoization_closure/4,
                       get_memoization_closure_value/4,
                       %expand_memoize_stored_function_reference/2,
                       store_memo_result/6,
                       get_stored_memo_result/4,
                       apply_to_memoize/8,
                       get_complete_memoization_expansion/6,
                       element_of_memoization_closure/6,
                       not_element_of_memoization_closure/6,

                       print_memo_profile/0,
                       reset_memo_table/0,
                       reset_memo_with_statistics/0]).

:- use_module(error_manager).
:- use_module(debug).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides memoization features (for closure expansion, closure1,...).').
% used when use_closure_expansion_memoization is true



:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(startup_prob,init_memo_table,
                           'Initialise Preferences').
:- register_event_listener(clear_specification,reset_memo_table,
                           'Reset Temporary Preferences.').

:- use_module(library(terms),[term_hash/3]).



% ---------------- Memoization for set comprehension / lambda / closure1 expansion:

get_stored_memo_expansion(Hash,Term,Result) :-
   stored_memo_expansion(Hash,Term,Result),
   inc_counter(memoization_reuse_counter,_).

:- dynamic stored_memo_expansion/3.

store_memo_expansion(Hash,LookupTerm,StoredResult) :-
   assert(stored_memo_expansion(Hash,LookupTerm,StoredResult)).

compute_memo_hash(LookupTerm,Hash) :-
    inc_counter(memoization_hash_counter,_),
    term_hash(LookupTerm,
            [range(smallint),algorithm(sdbm), depth(infinite),if_var(ignore)],Hash).
    %hashing:my_term_hash(closure(Parameters,ParameterTypes,ClosureBody),Hash),
    %inc(Hash), %

% ---------------- Memoization for function application

:- dynamic registered_value/3, memoization_expansion_completed/1.
% we register the functions to be memoized
get_registered_function(Id,Val) :- registered_value(Id,_,Val).

:- use_module(tools,[remove_variables/3]).
:- use_module(kernel_tools,[ground_value_check/2, value_variables/2]).
% register a (usually symbolic closure) value as a function
% ElType is the type of the elements of the function and MemoResultClosure is a closure that
% can and should be used instead of BValue to ensure memoization works
register_memoization_function(Name,BValue,ElType,RecursionInfo,MemoID,MemoResultClosure) :-
    value_variables(BValue,Vars),
    %print(vars(Name,Vars,LHSValue)),nl, print(bvalue(BValue)),nl,
    (RecursionInfo='$recursion_value'(RecValue)
     -> remove_variables(Vars,[RecValue],WVars), % do not wait on RecValue; it will be set by registering MEMOID
        RecValue = MemoResultClosure,
        InnerMemoID=MemoID % already instantiate MemoID before asserting closure; cyclic use
      ; WVars=Vars
    ),
    ground_value_check(WVars,Ground),
    %print(register(Name,MemoID,Ground,BValue)),nl,nl,
    construct_memoization_closure(MemoID,BValue,ElType,MemoResultClosure),
    register_memoization_function_aux(Ground,Name,BValue,InnerMemoID,MemoID).

:- block register_memoization_function_aux(-,?,?,?,?).
register_memoization_function_aux(_,Name,BValue,InnerMemoID,MemoID) :-
    inc_counter(memoization_function_counter,InnerMemoID),
    format('Registering function ~w for memoization: ~w~n',[InnerMemoID,Name]),
    (ground(BValue) -> true ; add_internal_error('Non ground memo closure: ',InnerMemoID:BValue)),
    compile_value(BValue,InnerMemoID),
    assert(registered_value(InnerMemoID,Name,BValue)),
    %if(MemoID=InnerMemoID, format('Finish ~w~n',[MemoID]),format('FAILURE ~w~n',[InnerMemoID])).
    MemoID=InnerMemoID. % only unify after assert; in case co-routines are triggered which call get_registered_function

 :- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(cogen,true)).
compile_value(closure(Paras,Ty,Body),InnerMemoID) :-
     compile_closure(Paras,Ty,Body,InnerMemoID),!.
compile_value(_,_).
:- use_module(b_cogen).
:- else.
compile_value(_,_).
reset_cogen.
:- endif.

construct_memoization_closure(MemoID,BValue,ElType,Result) :-
    get_infos(BValue,BInfos),
    XFC = b(external_function_call('MEMOIZE_STORED_FUNCTION',[b(value(int(MemoID)),integer,[])]),set(ElType),[]),
    TID = b(identifier('_zzzz_unary'),ElType,[]),
    Result = closure(['_zzzz_unary'],[ElType],
                          b(member(TID,XFC),pred,
                            [prob_annotation('MEMOIZE'(MemoID))|BInfos])).

:- use_module(bsyntaxtree,[get_texpr_info/2, extract_info/2, extract_pos_infos/2]).
% extract important infos,
get_infos(Var,Res) :- var(Var),!,Res=[].
get_infos(closure(_,_,B),Res) :- extract_info(B,Infos),!,
   get_texpr_info(B,AllInfos),extract_pos_infos(AllInfos,Pos),
   append(Pos,Infos,Res).
get_infos(_,[]). % :- print(no_infos(A)),nl.

:- use_module(bsyntaxtree, [get_texpr_id/2]).
is_memoization_closure([ID],_T,b(Body,_,INFO),MemoID) :-
   member(prob_annotation('MEMOIZE'(MemoID)),INFO),
   % check Body has not been rewritten:
   Body = member(TID,b(RHS,_,_)),
   get_texpr_id(TID,ID),
   is_memoize_stored_function_reference(RHS,MemoID).

is_memoize_stored_function_reference(external_function_call('MEMOIZE_STORED_FUNCTION',[Arg]),MemoID) :-
   Arg=b(value(int(MemoID)),_,_).

% check if we have a memoization closure and get its stored value
get_memoization_closure_value(P,T,B,Value) :-
    is_memoization_closure(P,T,B,MemoID),
    number(MemoID),
    registered_value(MemoID,_,Value).

% expand an (untyped) expression term referring to a stored function into full value:
%expand_memoize_stored_function_reference(Expr,Value) :-
%   is_memoize_stored_function_reference(Expr,MemoID),
%   number(MemoID), % in case this is called before function registered
%   registered_value(MemoID,_,Value).


% ----------------------------------------------
% providing stored_memo_result_id/4 fact interface
% some MemoID are stored in separate table
:- dynamic stored_memo_result1/3, stored_memo_result2/3,
           stored_memo_result3/3, stored_memo_result4/3,
           stored_memo_result5/3, stored_memo_result6/3,
           stored_memo_result7/3, stored_memo_result8/3,
           stored_memo_result/4.
stored_memo_result_id(1,ArgHash,Argument,StoredResult) :- !,
   stored_memo_result1(ArgHash,Argument,StoredResult).
stored_memo_result_id(2,ArgHash,Argument,StoredResult) :- !,
   stored_memo_result2(ArgHash,Argument,StoredResult).
stored_memo_result_id(3,ArgHash,Argument,StoredResult) :- !,
   stored_memo_result3(ArgHash,Argument,StoredResult).
stored_memo_result_id(4,ArgHash,Argument,StoredResult) :- !,
   stored_memo_result4(ArgHash,Argument,StoredResult).
stored_memo_result_id(5,ArgHash,Argument,StoredResult) :- !,
   stored_memo_result5(ArgHash,Argument,StoredResult).
stored_memo_result_id(6,ArgHash,Argument,StoredResult) :- !,
   stored_memo_result6(ArgHash,Argument,StoredResult).
stored_memo_result_id(7,ArgHash,Argument,StoredResult) :- !,
   stored_memo_result7(ArgHash,Argument,StoredResult).
stored_memo_result_id(8,ArgHash,Argument,StoredResult) :- !,
   stored_memo_result8(ArgHash,Argument,StoredResult).
stored_memo_result_id(MemoID,ArgHash,Argument,StoredResult) :-
   stored_memo_result(ArgHash,MemoID,Argument,StoredResult).

assert_stored_memo_result_id(1,ArgHash,Argument,StoredResult) :- !,
   assert(stored_memo_result1(ArgHash,Argument,StoredResult)).
assert_stored_memo_result_id(2,ArgHash,Argument,StoredResult) :- !,
   assert(stored_memo_result2(ArgHash,Argument,StoredResult)).
assert_stored_memo_result_id(3,ArgHash,Argument,StoredResult) :- !,
   assert(stored_memo_result3(ArgHash,Argument,StoredResult)).
assert_stored_memo_result_id(4,ArgHash,Argument,StoredResult) :- !,
   assert(stored_memo_result4(ArgHash,Argument,StoredResult)).
assert_stored_memo_result_id(5,ArgHash,Argument,StoredResult) :- !,
   assert(stored_memo_result5(ArgHash,Argument,StoredResult)).
assert_stored_memo_result_id(6,ArgHash,Argument,StoredResult) :- !,
   assert(stored_memo_result6(ArgHash,Argument,StoredResult)).
assert_stored_memo_result_id(7,ArgHash,Argument,StoredResult) :- !,
   assert(stored_memo_result7(ArgHash,Argument,StoredResult)).
assert_stored_memo_result_id(8,ArgHash,Argument,StoredResult) :- !,
   assert(stored_memo_result8(ArgHash,Argument,StoredResult)).
assert_stored_memo_result_id(MemoID,ArgHash,Argument,StoredResult) :-
   assert(stored_memo_result(ArgHash,MemoID,Argument,StoredResult)).

retractall_stored_memo_result :-
    retractall(stored_memo_result1(_,_,_)),
    retractall(stored_memo_result2(_,_,_)),
    retractall(stored_memo_result3(_,_,_)),
    retractall(stored_memo_result4(_,_,_)),
    retractall(stored_memo_result5(_,_,_)),
    retractall(stored_memo_result6(_,_,_)),
    retractall(stored_memo_result7(_,_,_)),
    retractall(stored_memo_result8(_,_,_)),
    retractall(stored_memo_result(_,_,_,_)).

% this clause is only ok if we have just a single function; otherwise we get collisions between different functions
compute_funcall_memo_hash(int(X),MemoID,R) :- MemoID <9,  !,
   R=X. % avoid computing hash for simple integer values; only drawback: indexing not good for too large integers
% TO DO: we could register with each registered_value fact also a dynamic table_memo fact
% then we can avoid clashes between different memoized functions
compute_funcall_memo_hash(X,MemoID,ArgHash) :-
   compute_memo_hash((MemoID,X),ArgHash).


% --------------------------------------------

% last argument should be left free when calling
get_stored_memo_result(MemoID,Argument,ArgHash,StoredResult) :-
   stored_memo_result_id(MemoID,ArgHash,Argument,StoredResult),
   %print(get_stored(MemoID,ArgHash,StoredResult)),nl,
   count_reuse(MemoID).


:- use_module(kernel_waitflags,[add_wd_error_span/4]).
:- use_module(tools_strings,[ajoin/2]).
:- use_module(store,[normalise_value_for_var/3]).
store_memo_result(MemoID,Argument,ArgHash,Result,Span,WF) :-
   normalise_value_for_var(MemoID,Result,NormResult),
   store_norm_memo_result(MemoID,Argument,ArgHash,NormResult,Span,WF).

:- use_module(kernel_objects,[equal_object/2]).
store_norm_memo_result(MemoID,Argument,ArgHash,Result,Span,WF) :-
   stored_memo_result_id(MemoID,ArgHash,Argument,StoredResult),
   !,
   (Result=StoredResult -> true
    ; equal_object(Result,StoredResult) -> format('Two representations for memoized function (~w) result.~n',[MemoID])
     ;  ajoin(['Memoized function ',MemoID,' has multiple results for: '],Msg),
        Relation = [(Argument,StoredResult),(Argument,Result)],
        add_wd_error_span(Msg,'@fun'(Argument,Relation),Span,WF)
        %,print('stored: '),translate:print_bvalue(StoredResult),nl,print('new: '),translate:print_bvalue(Result),nl
   ).
store_norm_memo_result(MemoID,Argument,ArgHash,Result,_,_) :-
   %format('Storing memo result for ~w, argument =',[MemoID]),
   %translate:print_bvalue(Argument),nl, translate:print_bvalue(Result),nl,
   assert_stored_memo_result_id(MemoID,ArgHash,Argument,Result).
   %assert(stored_memo_result(ArgHash,MemoID,Argument,Result)).
   %print(stored_memo_result(ArgHash,MemoID,Argument,Result)),nl.


% function application for memoized function:
apply_to_memoize(MemoID,P,T,B,X,Y,Span,WF) :-
        ground_value_check(X,XV),
        block_apply_closure_memo_groundx(XV,MemoID,X,Y,P,T,B,raise_wd,Span,WF).

:- use_module(kernel_tools,[ground_value/1]).
:- use_module(kernel_objects,[equal_object_wf/3]).

:- block block_apply_closure_memo_groundx(-,?, ?,?, ?,?,?, ?,?,?),
         block_apply_closure_memo_groundx(?,-, ?,?, ?,?,?, ?,?,?).


block_apply_closure_memo_groundx(_,MemoID, X,Y, P,T,B, RaiseWD, Span,WF) :-
    % MemoID is ground, meaning body B is ground as well
    % TODO: obtain stored value directly
    compute_funcall_memo_hash(X,MemoID,ArgHash),
    if(get_stored_memo_result(MemoID,X,ArgHash,StoredResult),
        equal_object_wf(Y,StoredResult,WF),
        (start_profile_function_application(MemoID,Timer),
         if(check_element_of_function_closure_nowf(MemoID,X,ComputedResult,P,T,B),
           % ground WF to compute ground value for ComputedResult
             (stop_profile_function_application(MemoID,Timer),
              (ground_value(ComputedResult)
               -> store_memo_result(MemoID,X,ArgHash,ComputedResult,Span,WF)
               ;  add_error(memoization,'Could not compute ground value for function application',MemoID,Span),
                  print('Argument: '),translate:print_bvalue(X),nl,
                  print('Result: '), print(ComputedResult),nl,
                  print('Function: '),registered_value(MemoID,_Name,FunValue), translate:print_bvalue(FunValue),nl,nl
               ),
               % To do: check if we compute two or more solutions for f(X) and add WD error
               equal_object_wf(Y,ComputedResult,WF)
             ),
             (RaiseWD=raise_wd,
              add_wd_error_span('function applied outside of domain (#8-Memo): ', '@fun'(X,closure(P,T,B)),Span,WF)
             )
           )
        )
      ).

:- use_module(kernel_waitflags,    [init_wait_flags/2,ground_wait_flags/1]).
:- use_module(custom_explicit_sets,[check_element_of_function_closure/6]).
check_element_of_function_closure_nowf(MemoID,X,ComputedResult,_,_,_) :-
   registered_value(MemoID,_,closure(P,T,B)),!,
   init_wait_flags(WF,[check_element_of_function_closure_nowf(MemoID)]),
   check_element_of_function_closure(X,ComputedResult,P,T,B,WF),
   ground_wait_flags(WF).
check_element_of_function_closure_nowf(_,X,ComputedResult,P,T,B) :-
   init_wait_flags(WF,[check_element_of_function_closure_nowf]),
   check_element_of_function_closure(X,ComputedResult,P,T,B,WF),
   ground_wait_flags(WF).

:- use_module(library(avl),[avl_domain/2]).
:- use_module(library(lists),[maplist/2]).

% store the complete expansion of a memoized function for later re-use
% and also passes result and done flag to an outer call
:- block store_memoization_closure_expansion(?,-,?,?,?,?,?).
store_memoization_closure_expansion(MemoID,_,Res,OuterDone,OuterRes,Span,WF) :-
   my_avl_domain(Res,List),
   !,
   debug_println(19,storing_expansion_of_memoized_function(MemoID)),
   (maplist(store_function_result(MemoID,Span,WF),List)
    -> assert(memoization_expansion_completed(MemoID))
     ; add_internal_error(memoization,'Error storing expansion for memoized function:',MemoID,Span)),
   equal_object_wf(Res,OuterRes,WF), OuterDone=true.
store_memoization_closure_expansion(MemoID,_,R,OuterDone,OuterRes,Span,WF) :-
   add_warning(memoization,'Cannot store expansion for memoized function:',MemoID:R,Span),
   equal_object_wf(R,OuterRes,WF), OuterDone=true.

my_avl_domain(avl_set(A),List) :- avl_domain(A,List).
my_avl_domain([],[]).

store_function_result(MemoID,Span,WF,(X,ComputedResult)) :-
    compute_funcall_memo_hash(X,MemoID,ArgHash),
    store_memo_result(MemoID,X,ArgHash,ComputedResult,Span,WF). % TO DO: we could use special version that also allows relations

:- use_module(kernel_objects,[equal_object_optimized_wf/4]).
% succeeds if function has been completely memoized and returns expansion
try_get_complete_memoization_expansion(MemoID,Result,WF) :-
    memoization_expansion_completed(MemoID),
    debug_println(19,recovering_full_value_for_memoized_function(MemoID)),
    findall((X,Y),stored_memo_result_id(MemoID,_,X,Y),List), % probably not sorted !
    equal_object_optimized_wf(List,Result,expand_normal_closure_memo,WF).

% always succeeds to obtain the full expansion; also stores it for reuse
:- block get_complete_memoization_expansion(-,?,?,?,?,?).
get_complete_memoization_expansion(MemoID,Result,Done,_,_PP,WF) :-
    try_get_complete_memoization_expansion(MemoID,Cached,WF),
    !,
    Result=Cached, Done=true.
get_complete_memoization_expansion(MemoID,Result,Done,Span,PP,WF) :-
    registered_value(MemoID,Name,Val), nonvar(Val),
    !,
    get_complete_memoization_expansion_aux(Val,MemoID,Name,Result,Done,Span,PP,WF).
get_complete_memoization_expansion(MemoID,_,_,Span,PP,_) :-
    add_error(get_complete_memoization_expansion,'Memoized function ID not registered:',MemoID:PP,Span),
    fail.


get_complete_memoization_expansion_aux(Val,_MemoID,_,Result,Done,_Span,_PP,_WF) :- expanded_value(Val), !,
   Result=Val,Done=true.
get_complete_memoization_expansion_aux(closure(P,T,B),MemoID,Name,Result,Done,Span,PP,WF) :- !,
    debug_println(19,expanding_memoized_function(MemoID,PP)),
    (custom_explicit_sets:is_infinite_closure(P,T,B)
      -> add_warning(memoization(PP),'Expanding infinite memoized function: ',Name,B)
      ; custom_explicit_sets:is_symbolic_closure(P,T,B)
      -> add_message(memoization(PP),'Fully expanding symbolic memoized function: ',Name,B)
      ; true),
    custom_explicit_sets:expand_normal_closure_direct(P,T,B,Expansion,DoneExp,WF),
    store_memoization_closure_expansion(MemoID,DoneExp,Expansion,Done,Result,Span,WF).
get_complete_memoization_expansion_aux(Val,MemoID,Name,_,_,Span,PP,_) :-
    add_error(memoization(PP),'Illegal value stored for memoized function:',MemoID:Name,Span),
    print(Val),nl,
    fail.

expanded_value(avl_set(_)).
expanded_value(global_set(_)).
expanded_value([]).
expanded_value([_|_]).

:- use_module(kernel_tools,[ground_value/1]).
:- use_module(kernel_objects,[check_element_of_wf/3]).
% special treatment for element of memoization closure; possibly not a function call
element_of_memoization_closure(MemoID,Element,WF,P,T,B) :-
    nonvar(Element), Element = (X,Y),
    ground_value(X),
    % we can do a function application:
    !,  %print(fun_apply_reuse(MemoID,Element)),nl,
    Span=B,
    block_apply_closure_memo_groundx(ground,MemoID, X,Y, P,T,B, do_not_raise_wd, Span,WF).
element_of_memoization_closure(MemoID,Element,WF,_,_,B) :-
% TO DO: delay before computing complete memoization
    get_complete_memoization_expansion(MemoID,Set,_Done,B,element_of_memoization_closure,WF),
    !,  %print(el_of(MemoID,Element)),nl,
    check_element_of_wf(Element,Set,WF).

:- use_module(kernel_objects,[not_equal_object_wf/3]).
:- use_module(kernel_objects,[not_element_of_wf/3]).

not_element_of_memoization_closure(MemoID,Element,WF,P,T,B) :-
    nonvar(Element), Element = (X,Y),
    ground_value(X),
    !,
    if(block_apply_closure_memo_groundx(ground,MemoID, X,YRes, P,T,B, do_not_raise_wd, B,WF),
       not_equal_object_wf(Y,YRes,wf),
       true).
not_element_of_memoization_closure(MemoID,Element,WF,_,_,B) :-
    % TO DO: delay before computing complete memoization
    get_complete_memoization_expansion(MemoID,Set,_Done,B,not_element_of_memoization_closure,WF),
    !,
    not_element_of_wf(Element,Set,WF).
% -------------------------------

:- dynamic function_application_time/2.
:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_profile,true)).
%:- print('Profiling Memoizations'),nl.
count_reuse(MemoID) :-
   my_bb_get_reuse_counter(MemoID,Cnt), C1 is Cnt+1, bb_put(MemoID,C1),
   inc_counter(memoization_function_reuse_counter,_).
my_bb_get_reuse_counter(Key,Val) :- (bb_get(Key,V) -> V=Val ; Val=0).
bb_reset_reuse_counter(Key) :- bb_put(Key,0).
start_profile_function_application(_,Walltime) :- statistics(walltime,[Walltime,_]).
stop_profile_function_application(MemoID,Walltime1) :- statistics(walltime,[Walltime2,_]),
  Delta is Walltime2-Walltime1,
  (Delta>0 ->
     (retract(function_application_time(MemoID,Old))
       -> New is Old+Delta ; New = Delta),
     assert(function_application_time(MemoID,New))
   ; true
  ).
function_application_time_available :- true.
:- else.
count_reuse(_MemoID) :-
   inc_counter(memoization_function_reuse_counter,_).
my_bb_get_reuse_counter(_,'?').
bb_reset_reuse_counter(_).
start_profile_function_application(_,0).
stop_profile_function_application(_,_).
function_application_time_available :- fail.
:- endif.

% -------------------------------
% utilities to profile expansions
:- public inc/1.
:- dynamic hash_count/2.
inc(Hash) :- (retract(hash_count(Hash,Nr)) -> N1 is Nr+1 ; N1 = 1), assert(hash_count(Hash,N1)).
:- dynamic hash_time/2.
store_memo_computation_time(Hash,Time) :- % for closure expansions, not for funct. applications
   retractall(hash_time(Hash,_)), assert(hash_time(Hash,Time)).

print_memo_profile :-
   ((stored_memo_expansion(_,_,_) ; registered_value(_,_,_))
     -> print('MEMO Table:'),nl,
        print_memo_table
     ; true).

% memoization:print_memo.
print_memo_table :-
  stored_memo_expansion(Hash,LookupTerm,PackedValue),
  (hash_count(Hash,Count) ->  print(Count), print(' : ') ; true),
  (hash_time(Hash,Time) -> print(Time),print(' --> ') ; true),
  print_memo_entry(LookupTerm),
  nl, print('     :  '),translate:print_bvalue(PackedValue),nl,fail.
print_memo_table :-
  debug_mode(on),
  registered_value(MemoID,Name,FunValue),
  (memoization_expansion_completed(MemoID)
   -> Compl=' (completely evaluated)' ; Compl=''),
  format('MemoID ~w stored FUNCTION ~w~w ',[MemoID,Name,Compl]),
  nl, print('     :  '),translate:print_bvalue(FunValue),nl,fail.
print_memo_table :-
  debug_mode(on),
  registered_value(MemoID,Name,_),
  stored_memo_result_id(MemoID,_Hash,LookupTerm,PackedValue),
  format('MemoID ~w (~w) result for argument ',[MemoID,Name]),
  print_memo_entry(LookupTerm),
  nl, print('     :  '),translate:print_bvalue(PackedValue),nl,fail.
print_memo_table :-
  stored_memo_result(Hash,MemoID1,T1,_),
  stored_memo_result(Hash,MemoID2,T2,_),
  dif((MemoID1,T1),(MemoID2,T2)),
  format('Hash COLLISION ~w, ~w:~w vs ~w:~w~n',[Hash,MemoID1,T1,MemoID2,T2]),
  fail.
print_memo_table :- format('Summary of reuse per memoization ID:~n',[]),
  registered_value(MemoID,Name,_),
  my_bb_get_reuse_counter(MemoID,Cnt),
  findall(1,stored_memo_result_id(MemoID,_Hash,_,_),Stored), length(Stored,NrStored),
  (\+ function_application_time_available -> WTime='?'
   ; function_application_time(MemoID,WTime) -> true ; WTime=0),
  format('MemoID ~w (~w) : Values Stored: ~w (~w ms to compute), Reused ~w~n',[MemoID,Name,NrStored,WTime,Cnt]),
  fail.
print_memo_table :- get_counter(memoization_hash_counter,Hash),
  get_counter(memoization_reuse_counter,RHit),
  format('Hashes computed: ~w, expansions reused: ~w~n',[Hash,RHit]),
  get_counter(memoization_function_counter,F),
  get_counter(memoization_function_reuse_counter,FR),
  format('Memoization functions registered: ~w, results reused: ~w~n',[F,FR]).

print_memo_entry(closure(Parameters,ParameterTypes,ClosureBody)) :- !,
  translate:print_bvalue(closure(Parameters,ParameterTypes,ClosureBody)).
print_memo_entry(closure1_for_explicit_set(A)) :- !,
  print('closure1('),translate:print_bvalue(avl_set(A)), print(')').
print_memo_entry(A) :- translate:print_bvalue(A).

:- use_module(extension('counter/counter'),[counter_init/0, reset_counter/1,
                                            new_counter/1, inc_counter/2,
                                            get_counter/2]).
init_memo_table :-
   counter_init,
   new_counter(memoization_reuse_counter), new_counter(memoization_hash_counter),
   new_counter(memoization_function_counter), new_counter(memoization_function_reuse_counter),
   reset_memo_table.

reset_memo_table :-
  retract(registered_value(MemoID,_,_)),
  bb_reset_reuse_counter(MemoID),
  fail.
reset_memo_table :-
   reset_counter(memoization_reuse_counter),
   reset_counter(memoization_hash_counter),
   reset_counter(memoization_function_counter),
   reset_counter(memoization_function_reuse_counter),
   (stored_memo_expansion(_,_,_) -> debug_println(19,'Resetting MEMO table') ; true),
   retractall(hash_count(_,_)),
   retractall(hash_time(_,_)),
   retractall(stored_memo_expansion(_,_,_)),
   retractall(registered_value(_,_,_)),
   retractall(function_application_time(_,_)),
   retractall(memoization_expansion_completed(_)),
   retractall_stored_memo_result,
   reset_cogen.

:- use_module(probsrc(tools),[retract_with_statistics/2]).
reset_memo_with_statistics :-
   retract_with_statistics(memoization,
      [stored_memo_expansion(_,_,_),registered_value(_,_,_),memoization_expansion_completed(_),
        stored_memo_result1(_,_,_), stored_memo_result2(_,_,_),
        stored_memo_result3(_,_,_), stored_memo_result4(_,_,_),
        stored_memo_result5(_,_,_), stored_memo_result6(_,_,_),
        stored_memo_result7(_,_,_), stored_memo_result8(_,_,_),
       stored_memo_result(_,_,_,_)]),
   reset_memo_table.

