% (c) 2016-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(meta_interface,[command_description/3, command_call/4,
                          command_unavailable/2,
                          command_arguments/2, command_optional_arguments/4,
                          command_category/2,
                          command_preferences_category/2,
                          command_additional_info/2,
                          get_command_preferences/2,
                          call_command/5,
                          is_dot_command/1, is_dot_command/6,
                          call_dot_command/2, call_dot_command/3,
                          is_dot_command_for_expr/1, call_dot_command_for_expr/3, call_dot_command_for_expr/4,
                          is_table_command/6]).

:- use_module(module_information).
:- module_info(group,visualization).
:- module_info(description,'This module provides access and description to a variety of Prolog commands.').

:- use_module(error_manager).
:- use_module(debug).
:- use_module(tools,[ajoin/2]).
:- use_module(specfile,[
                    b_or_z_mode/0,
                    eventb_mode/0,
                    csp_mode/0,
                    %csp_with_bz_mode/0,
                    spec_file_has_been_successfully_loaded/0]).
:- use_module(state_space, [current_state_corresponds_to_fully_setup_b_machine/0, current_state_corresponds_to_setup_constants_b_machine/0]).

:- use_module(bmachine,[b_machine_has_constants_or_properties/0, b_machine_has_assertions/0,
                        b_machine_has_constants/0, b_machine_has_variables/0,
                        b_machine_has_operations/0]).

% dot command with no arguments
is_dot_command(Command) :- command_category(Command,dot), command_arguments(Command,[dotfile]).
call_dot_command(Command,File) :- call_dot_command(Command,File,[]).
call_dot_command(Command,File,OptionalArgs) :- %print(cmd(Command,File,OptionalArgs)),nl,
   call_command(dot,Command,[dotfile],[File],OptionalArgs).

is_dot_command_for_expr(Command) :-
   command_category(Command,dot), command_arguments(Command,[expression_atom,dotfile]).
call_dot_command_for_expr(Command,Expr,File) :-
   call_dot_command_for_expr(Command,Expr,File,[]).
call_dot_command_for_expr(Command,Expr,File,OptionalArgs) :-  % print(cmd(Command,Expr,File,OptionalArgs)),nl,
   call_command(dot,Command,[expression_atom,dotfile],[Expr,File],OptionalArgs).

% dot commands with no or with extra arguments:
% find out which dot commands are available and what arguments and preferences they take:
is_dot_command(Command,Name,Description,NumberOfFormulaArgs,Preferences,AdditionalInfo) :-
   command_description(Command,Name,Description), % move forward to ensure ordering of found commands is same as in file
   (is_dot_command(Command),NumberOfFormulaArgs=0 ;
    is_dot_command_for_expr(Command),NumberOfFormulaArgs=1),
   get_command_preferences(Command, Preferences),
   findall(Info, command_additional_info(Command, Info), AdditionalInfo).

:- use_module(library(lists)).
% table commands (return as last argument list of lists)
is_table_command(Command,Name,Description,NumberOfFormulaArgs,Preferences,AdditionalInfo) :-
   command_category(Command,table),
   command_arguments(Command,Args), last(Args,table_result),
   length(Args,Len), NumberOfFormulaArgs is Len - 1,
   command_description(Command,Name,Description),
   get_command_preferences(Command, Preferences),
   findall(Info, command_additional_info(Command, Info), AdditionalInfo).





% call a command described in this module
call_command(Category,Command,FormalArgs,ActualArgs,OptionalArgs) :-
   command_category(Command,Category),
   command_arguments(Command,FormalArgs), % to do: more flexible matching of args ?
   !,
   (command_unavailable(Command,Msg)
    -> ajoin(['Command ', Command, ' not available: '],Msg1),
       add_error(meta_interface,Msg1,Msg),
       fail
    ; setup_optional_args(Command,OptionalArgs,OptArgValueList),
      (command_call(Command,ActualArgs,OptArgValueList,Call)
         -> debug_println(4,command_call(Call)),
            call(Call)
        ;  command_call(Command,ArgsExpected,_,_),length(ArgsExpected,Exp),length(ActualArgs,Form) %, Exp \= Form
           ->
           ajoin(['Command ', Command, ' expects ', Exp, ' arguments, but obtained ',Form,'!'],Msg1),
           add_internal_error(Msg1,command_call(Command,ActualArgs,OptArgValueList,_)),
           fail
        ;  add_internal_error('Cannot construct command call: ',command_call(Command,ActualArgs,OptArgValueList,_)),
           fail
      )
    ).
call_command(Category,Command,_FormalArgs,_ActualArgs,_OptArgs) :- command_category(Command,Category),
    !,
    add_error(meta_interface,'Argument mis-match when calling command: ',Command),fail.
call_command(Category,Command,_,_,_) :- command_category(Command,_OtherCategory),!,
    add_error(meta_interface,'Command belongs to other category: ',Category:Command),fail.
call_command(Category,Command,_,_,_) :-
    add_error(meta_interface,'Unknown command: ',Category:Command),fail.


setup_optional_args(Command,OptionalArgs,OptArgValueList) :-
     command_optional_arguments(Command,Names,_Types,DefaultValues),!,
     setup_opt_args_aux(Names,DefaultValues,OptionalArgs,Command,OptArgValueList).
setup_optional_args(_,[],Res) :- !, Res=[].
setup_optional_args(Cmd,Opt,Res) :- add_internal_error('Illegal optional arguments: ',setup_optional_args(Cmd,Opt,Res)),
   Res=[].

:- use_module(library(lists),[select/3]).

setup_opt_args_aux([],[],OptArgs,Command,Res) :-
    (OptArgs = [] -> true ; add_internal_error('Too many optional arguments: ',Command:OptArgs)),
    Res = [].
setup_opt_args_aux([Name|NT],[Default|DT],OptArgs,Command,Res) :-
    (select('='(Name,Value),OptArgs,OA2)
      -> Res = [Value|RT], setup_opt_args_aux(NT,DT,OA2,Command,RT)
      ;  Res = [Default|RT], setup_opt_args_aux(NT,DT,OptArgs,Command,RT)
    ).


% TO DO: provide automatic conversion ?  expression_atom -> expression_codes -> typed_expression ?

:- use_module(preferences,[virtual_preference_category/2]).
% get all preferences for a command
get_command_preferences(Command,Preferences) :-
   findall(Pref, (command_preferences_category(Command,PrefCat),
                  virtual_preference_category(Pref,PrefCat)), Preferences).

% --------------------
% COMMAND DESCRIPTIONS
% --------------------

:- discontiguous command_description/3, command_call/4,
                 command_unavailable/2, command_arguments/2,
                 command_optional_arguments/4,
                 command_category/2, command_preferences_category/2,
                 command_additional_info/2.

:- use_module(b_machine_hierarchy,[write_dot_event_hierarchy_to_file/1,write_dot_hierarchy_to_file/1]).

command_description(machine_hierarchy,'Machine Hierarchy','Shows the machine hierarchy of a classical B model').
command_call(machine_hierarchy,[File],[],b_machine_hierarchy:write_dot_hierarchy_to_file(File)).
command_unavailable(machine_hierarchy,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(machine_hierarchy,[dotfile]).
command_category(machine_hierarchy,dot).
command_preferences_category(machine_hierarchy,dot_machine_hierarchy).

command_description(operations,'Operation Call Graph','Shows the call graph of a classical B model').
command_call(operations,[File],[],b_machine_hierarchy:write_dot_op_hierarchy_to_file(File)).
command_unavailable(operations,'only available for B or Z models with OPERATIONS') :-
     (\+ (b_or_z_mode,b_machine_has_operations) ; eventb_mode).
command_arguments(operations,[dotfile]).
command_category(operations,dot).
command_preferences_category(operations,dot_graph_generator).

command_description(event_hierarchy,'Event Hierarchy','Shows the event hierarchy of an Event-B model').
command_call(event_hierarchy,[File],[],b_machine_hierarchy:write_dot_event_hierarchy_to_file(File)).
command_unavailable(event_hierarchy,'only available for Event-B models with events') :-
     \+ (eventb_mode, b_machine_has_operations).
command_arguments(event_hierarchy,[dotfile]).
command_category(event_hierarchy,dot).
command_preferences_category(event_hierarchy,dot_event_hierarchy).


:- use_module(visualize_graph,[tcltk_print_states_for_dot/2]).
command_description(state_space,'State Space','Show state space').
command_call(state_space,[File],[ColourTransitions],visualize_graph:tcltk_print_states_for_dot(File,ColourTransitions)).
command_unavailable(state_space,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(state_space,[dotfile]).
command_optional_arguments(state_space,[colouring],[atom],[no_colour_for_transitions]).
command_category(state_space,dot).
command_preferences_category(state_space,dot_state_space).

command_description(state_space_sfdp,'State Space (Fast)','Show state space (fast)').
command_call(state_space_sfdp,[File],[ColourTransitions],visualize_graph:tcltk_print_states_for_dot_sfdp(File,ColourTransitions)).
command_unavailable(state_space_sfdp,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(state_space_sfdp,[dotfile]).
command_optional_arguments(state_space_sfdp,[colouring],[atom],[no_colour_for_transitions]).
command_category(state_space_sfdp,dot).
command_preferences_category(state_space_sfdp,dot_state_space). % TODO: maybe remove those that are fixed and have no effect
command_additional_info(state_space_sfdp,preferred_dot_type(sfdp)). % sfdp preferred

:- use_module(visualize_graph,[tcltk_print_current_state_for_dot/1]).
command_description(current_state,'Current State in State Space','Show current state and successors in state space').
command_call(current_state,[File],[],visualize_graph:tcltk_print_current_state_for_dot(File)).
command_unavailable(current_state,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(current_state,[dotfile]).
command_category(current_state,dot).
command_preferences_category(current_state,dot_state_space).

:- use_module(visualize_graph,[tcltk_print_history_to_current_state_for_dot/1]).
command_description(history,'Path to Current State','Show a path leading to current state').
% in ProB2-UI the Tcl/Tk history is not the ProB2 history !!!
command_call(history,[File],[],visualize_graph:tcltk_print_history_to_current_state_for_dot(File)).
command_unavailable(history,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(history,[dotfile]).
command_category(history,dot).
command_preferences_category(history,dot_state_space).

%:- use_module(visualize_graph,[tcltk_print_history_to_current_state_with_neighbors_for_dot/1]).
%command_description(history_with_neighbours,'History of States (with Neighbours for last State)','Show history with neighbours leading to current state').
%command_call(history_with_neighbours,[File],[],visualize_graph:tcltk_print_history_to_current_state_with_neighbors_for_dot(File)).
%command_unavailable(history_with_neighbours,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
%command_arguments(history_with_neighbours,[dotfile]).
%command_category(history_with_neighbours,dot).
%command_preferences_category(history_with_neighbours,dot_state_space).

:- use_module(state_space_reduction,[write_transition_diagram_for_expr_to_dotfile/2,
                                     write_signature_merge_to_dotfile/2]).
command_description(signature_merge,'Signature Merge','Show signature-merged reduced state space').
command_call(signature_merge,[File],[IgnoredEvents],state_space_reduction:write_signature_merge_to_dotfile(IgnoredEvents,File)).
command_unavailable(signature_merge,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(signature_merge,[dotfile]).
command_optional_arguments(signature_merge,[ignored_events],[list(operations)],[[]]).
command_category(signature_merge,dot).
command_preferences_category(signature_merge,dot_graph_generator).


:- use_module(reduce_graph_state_space,[print_dot_for_dfa_from_nfa/1]).
command_description(dfa_merge,'DFA Merge','Show state space as deterministic automaton (DFA)').
command_call(dfa_merge,[File],[],reduce_graph_state_space:print_dot_for_dfa_from_nfa(File)).
command_unavailable(dfa_merge,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(dfa_merge,[dotfile]).
command_category(dfa_merge,dot).

:- use_module(state_space_reduction,[write_transition_diagram_for_expr_to_dotfile/2]).
command_description(transition_diagram,'State Space Expression Projection...','Project state space onto expression values and show transition diagram').
command_call(transition_diagram,[Expression,File],[],
             state_space_reduction:write_transition_diagram_for_expr_to_dotfile(Expression,File)).
command_unavailable(transition_diagram,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(transition_diagram,[expression_atom,dotfile]). % also accepts expression_raw_ast
command_category(transition_diagram,dot).
command_preferences_category(transition_diagram,dot_projection).

:- use_module(probporsrc(dot_graphs_static_analysis), [tcltk_create_dependence_graph/1, tcltk_create_enable_graph/1]).
command_description(enable_graph,'Enable Graph','Show enabling graph of events').
command_call(enable_graph,[File],[],dot_graphs_static_analysis:tcltk_create_enable_graph(File)).
command_unavailable(enable_graph,Txt) :- command_unavailable(operations,Txt).
command_arguments(enable_graph,[dotfile]).
command_category(enable_graph,dot).
command_preferences_category(enable_graph,dot_graph_generator).

:- use_module(graph_canon,[print_cstate_graph/1]).
command_description(state_as_graph,'Current State as Graph','Show values in current state as a graph').
command_call(state_as_graph,[File],[],graph_canon:print_cstate_graph(File)).
command_unavailable(state_as_graph,'only available for B,Z or Event-B models after constant setup or initialisation') :- \+ current_state_corresponds_to_setup_constants_b_machine.
command_arguments(state_as_graph,[dotfile]).
command_category(state_as_graph,dot).
command_preferences_category(state_as_graph,state_as_graph).
% we could use tcltk_print_current_state_as_graph_for_dot

:- use_module(state_custom_dot_graph,[tcltk_generate_state_custom_dot_graph/1,state_custom_dot_graph_available/0]).

command_description(custom_graph,'Customized Current State as Graph','Show values in current state as a graph using CUSTOM_GRAPH_EDGES').
command_call(custom_graph,[File],[],state_custom_dot_graph:tcltk_generate_state_custom_dot_graph(File)).
command_unavailable(custom_graph,'only available when CUSTOM_GRAPH_NODES and CUSTOM_GRAPH_EDGES are defined in the DEFINITIONS of a B machine') :-
   \+ state_custom_dot_graph_available.
command_unavailable(custom_graph,'only available for initialised B,Z or Event-B models') :-
   state_custom_dot_graph_available,
   \+ current_state_corresponds_to_fully_setup_b_machine.
command_arguments(custom_graph,[dotfile]). % also accepts expression_raw_ast
command_category(custom_graph,dot).
command_preferences_category(custom_graph,dot_graph_generator).

% from tcltk_interface.pl:
command_description(expr_as_graph,'(Relational) Expression as Graph...','Show (relational) expression value as a graph').
command_call(expr_as_graph,[Expression,File],[],tcltk_interface:tcltk_show_expression_as_dot(Expression,File)).
command_unavailable(expr_as_graph,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(expr_as_graph,[expression_atom,dotfile]). % also accepts expression_raw_ast
command_category(expr_as_graph,dot).
command_preferences_category(expr_as_graph,state_as_graph).


command_description(formula_tree,'Custom Predicate/Expression Formula Tree...','Show predicate/expressions and sub-formulas as a tree').
command_call(formula_tree,[Expression,File],[],tcltk_interface:generate_dot_from_formula(Expression,File)).
command_unavailable(formula_tree,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(formula_tree,[expression_atom,dotfile]). % is actually predicate_atom, also accepts expression_raw_ast
command_category(formula_tree,dot).
command_preferences_category(formula_tree,dot_formula_tree).

command_description(invariant,'Invariant Formula Tree','Show invariant as a formula tree').
command_call(invariant,[File],[],tcltk_interface:generate_dot_from_invariant(File)).
command_unavailable(invariant,'only available for initialised B,Z or Event-B models') :- \+ current_state_corresponds_to_fully_setup_b_machine.
command_arguments(invariant,[dotfile]).
command_category(invariant,dot).
command_preferences_category(invariant,dot_formula_tree).

command_description(properties,'Properties Formula Tree','Show properties as a formula tree').
command_call(properties,[File],[],tcltk_interface:generate_dot_from_properties(File)).
%command_unavailable(properties,'only available for initialised B,Z or Event-B models') :- \+ b_or_z_mode. % for non-initialised machines an existential quantifier is added
command_unavailable(properties,'only available for B,Z or Event-B models with PROPERTIES after constant setup or initialisation') :-
    \+ (current_state_corresponds_to_setup_constants_b_machine,b_machine_has_constants_or_properties).
command_arguments(properties,[dotfile]).
command_category(properties,dot).
command_preferences_category(properties,dot_formula_tree).

command_description(assertions,'Assertions Formula Tree','Show assertions as a formula tree').
command_call(assertions,[File],[],tcltk_interface:generate_dot_from_assertions(File)).
%command_unavailable(assertions,'only available for initialised B,Z or Event-B models') :- \+ b_or_z_mode. % for non-initialised machines an existential quantifier is added
command_unavailable(assertions,'only available for B,Z or Event-B models with ASSERTIONS after constant setup or initialisation') :-
   \+ (current_state_corresponds_to_setup_constants_b_machine, b_machine_has_assertions).
command_arguments(assertions,[dotfile]).
command_category(assertions,dot).
command_preferences_category(assertions,dot_formula_tree).

command_description(deadlock,'Deadlock Formula Tree','Show deadlocking status as a formula tree').
command_call(deadlock,[File],[],tcltk_interface:generate_dot_from_deadlock_po(File)).
command_unavailable(deadlock,'only available for initialised B,Z or Event-B models with OPERATIONS') :-
 \+ (current_state_corresponds_to_fully_setup_b_machine, b_machine_has_operations).
command_arguments(deadlock,[dotfile]).
command_category(deadlock,dot).
command_preferences_category(deadlock,dot_formula_tree).

:- use_module(bmachine,[b_get_machine_goal/1]).
command_description(goal,'Goal Formula Tree','Show GOAL as a formula tree').
command_call(goal,[File],[],tcltk_interface:generate_dot_from_goal(File)).
command_unavailable(goal,'only available for initialised B,Z or Event-B models with a GOAL DEFINITION') :-
  (\+ b_or_z_mode ;  \+ b_get_machine_goal(_)).
command_arguments(goal,[dotfile]).
command_category(goal,dot).
command_preferences_category(goal,dot_formula_tree).

command_description(dependence_graph,'Dependence Graph','Show dependence graph of events').
command_call(dependence_graph,[File],[],dot_graphs_static_analysis:tcltk_create_dependence_graph(File)).
command_unavailable(dependence_graph,Txt) :- command_unavailable(operations,Txt).
command_arguments(dependence_graph,[dotfile]).
command_category(dependence_graph,dot).
command_preferences_category(dependence_graph,dot_graph_generator).

command_description(variable_modification_graph,'Variable Read/Write Graph','Show variable modification by operations and reading in guards').
command_call(variable_modification_graph,[File],[],b_read_write_info:tcltk_dot_variable_modification_analysis(File)).
command_unavailable(variable_modification_graph,Txt) :- command_unavailable(variables,Txt).
command_arguments(variable_modification_graph,[dotfile]).
command_category(variable_modification_graph,dot).
command_preferences_category(variable_modification_graph,dot_variable_modification).

:- use_module(visualize_graph,[tcltk_print_definitions_as_graph_for_dot/1]).
command_description(definitions,'Definitions Graph','Show dependence graph of DEFINITIONS').
command_call(definitions,[File],[],visualize_graph:tcltk_print_definitions_as_graph_for_dot(File)).
command_unavailable(definitions,'only available for B or Z models') :- (b_or_z_mode -> eventb_mode ; true).
command_arguments(definitions,[dotfile]).
command_category(definitions,dot).
command_preferences_category(definitions,dot_definitions).

:- use_module(visualize_graph,[tcltk_print_predicate_dependency_as_graph_for_dot/2]).
command_description(predicate_dependency,'Predicate Dependency Graph...','Show dependence graph of conjuncts of a predicate').
command_call(predicate_dependency,[Pred,File],[],visualize_graph:tcltk_print_predicate_dependency_as_graph_for_dot(Pred,File)).
command_unavailable(predicate_dependency,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(predicate_dependency,[expression_atom,dotfile]). % also accepts expression_raw_ast
command_category(predicate_dependency,dot).

command_description(last_error,'Last Error Formula Tree','Show last error source as a formula tree').
command_call(last_error,[File],[],tcltk_interface:generate_dot_from_last_span_predicate(File)).
command_unavailable(last_error,'only available when error occured') :- \+ tcltk_interface:can_generate_dot_from_last_state_error.
command_arguments(last_error,[dotfile]).
command_category(last_error,dot).


% TABLE COMMANDS:

% from tcltk_interface.pl:
command_description(expr_as_table,'Expression Table...','Show expression value as a table').
command_call(expr_as_table,[Expression,Result],[],tcltk_interface:tcltk_eval_as_table(Expression,Result)).
command_unavailable(expr_as_table,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(expr_as_table,[expression_atom,table_result]). % also accepts expression_raw_ast
command_category(expr_as_table,table).

:- use_module(tcltk_interface,[tcltk_operations_covered_info/2]).

command_description(quick_operation_coverage,'Operation Coverage','Show which operations are covered').
command_call(quick_operation_coverage,[Result],[],tcltk_interface:tcltk_operations_covered_info(Result,quick)).
command_unavailable(quick_operation_coverage,'only available for B,Z, Event-B with OPERATIONS, or CSP models') :-
  \+ (b_or_z_mode, b_machine_has_operations), \+ csp_mode.
command_arguments(quick_operation_coverage,[table_result]). % also accepts expression_raw_ast
command_category(quick_operation_coverage,table).

command_description(precise_operation_coverage,'Operation Coverage and Feasibility','Show which feasible operations are covered').
command_call(precise_operation_coverage,[Result],[],tcltk_interface:tcltk_operations_covered_info(Result,precise)).
command_unavailable(precise_operation_coverage,Txt) :- command_unavailable(operations,Txt).
command_arguments(precise_operation_coverage,[table_result]). % also accepts expression_raw_ast
command_category(precise_operation_coverage,table).

% not really useful for ProB2: already in Statistics View when more details shown:
%command_description(operation_coverage_stats,'Operation Coverage Statistics','Show operation coverage statistics').
%command_call(operation_coverage_stats,[Result],[],meta_interface:get_operation_coverage_stats(Result)).
%command_arguments(operation_coverage_stats,[table_result]).
%command_category(operation_coverage_stats,table).
%:- use_module(coverage_statistics,[tcltk_compute_coverage/1]).
%get_operation_coverage_stats(list([list(['Info'])|LRes])) :-
%    tcltk_compute_coverage(list(Res)),
%    maplist(wrap_into_list,Res,LRes).


command_description(show_typing,'Show Typing','Show Types for Variables and Constants').
command_call(show_typing,[Result],[],meta_interface:show_typing_table(Result)).
command_unavailable(show_typing,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(show_typing,[table_result]). % also accepts expression_raw_ast
command_category(show_typing,table).
:- public show_typing_table/1.
show_typing_table(list([list(['Typing'])|LRes])) :- tcltk_interface:tcltk_show_typing_of_variables_and_constants(list(Res)),
   maplist(wrap_into_list,Res,LRes).

command_description(variable_coverage,'Variable Coverage','Show Number of Distinct Covered Values for Variables').
command_call(variable_coverage,[Result],[],state_space_reduction:tcltk_compute_nr_covered_values_for_all_variables(Result)).
command_unavailable(variable_coverage,Txt) :- command_unavailable(variables,Txt).
command_arguments(variable_coverage,[table_result]). % also accepts expression_raw_ast
command_category(variable_coverage,table).

command_description(expression_coverage,'Expression Coverage...','Evaluate Expression over State Space').
command_call(expression_coverage,[Expr,Result],[],meta_interface:expression_coverage(Expr,Result)).
command_unavailable(expression_coverage,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(expression_coverage,[expression_atom,table_result]).
command_category(expression_coverage,table).

:- public expression_coverage/2.
:- use_module(state_space_reduction,[compute_covered_values_for_expression/4]).
expression_coverage(ExprToEvaluate,list(LRes)) :-
   compute_covered_values_for_expression(ExprToEvaluate,_MaxTypeCard,_TotalNrOfValuesFound,list(Res)),
   maplist(wrap_into_list,Res,LRes).

command_description(minmax_table,'Min/Max Values','Show Min/Max Values for Constants and Variables').
command_call(minmax_table,[Result],[],meta_interface:minmax_table(Result)).
command_unavailable(minmax_table,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(minmax_table,[table_result]).
command_category(minmax_table,table).
:- public minmax_table/1.
minmax_table(list([list(['Min/Max'])|LRes])) :- coverage_statistics:tcltk_compute_min_max(list(Res)),
   maplist(wrap_into_list,Res,LRes).

wrap_into_list(list(L),R) :- !, R=list(L).
wrap_into_list([],R) :- !, R=list([]).
wrap_into_list([H|T],R) :- !, R=list([H|T]).
wrap_into_list(X,list([X])).

command_description(vacuous_invariants,'Vacuous Invariants','Show List of Vacuous Invariants (given current state space)').
command_call(vacuous_invariants,[Result],[],meta_interface:vacuous_invariants(Result)).
command_unavailable(vacuous_invariants,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(vacuous_invariants,[table_result]).
command_category(vacuous_invariants,table).
:- public vacuous_invariants/1.
vacuous_invariants(list([list(['Vacuous Invariant'])|LRes])) :-
   tcltk_interface:tcltk_get_vacuous_invariants(list(Res)),
   maplist(wrap_into_list,Res,LRes).

command_description(specialized_invariants,'Specialized Invariants','Show List of Invariants (specialized using proof information for each operation)').
command_call(specialized_invariants,[Result],[],meta_interface:specialized_invariants(Result)).
command_unavailable(specialized_invariants,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(specialized_invariants,[table_result]).
command_category(specialized_invariants,table).
:- public specialized_invariants/1.
specialized_invariants(list([list(['Operation','Unproven','Predicate'])|LRes])) :-
   bmachine:tcltk_get_specialized_invariants_for_ops(list(Res)),
   maplist(wrap_into_list,Res,LRes).


command_description(vacuous_guards,'Vacuous Guards','Show List of Vacuous Guards (given current state space)').
command_call(vacuous_guards,[Result],[],meta_interface:vacuous_guards(Result)).
command_unavailable(vacuous_guards,Txt) :- command_unavailable(operations,Txt).
command_arguments(vacuous_guards,[table_result]).
command_category(vacuous_guards,table).
:- public vacuous_guards/1.
vacuous_guards(list([list(['Vacuous Guards'])|LRes])) :-
   tcltk_interface:tcltk_get_vacuous_guards(list(Res)),
   maplist(wrap_into_list,Res,LRes).

command_description(inv_coverage,'Invariant Analysis','Analyse individual invariants truth value in state space').
command_call(inv_coverage,[Result],[],mcdc_coverage:tcltk_get_invariant_coverage(Result)).
command_unavailable(inv_coverage,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(inv_coverage,[table_result]).
command_category(inv_coverage,table).

command_description(mcdc_coverage,'MC/DC Operation Coverage','Show MC/DC Operation Coverage').
command_call(mcdc_coverage,[Result],[],mcdc_coverage:tcltk_compute_mcdc_operation_coverage(Result)).
command_unavailable(mcdc_coverage,Txt) :- command_unavailable(operations,Txt).
command_arguments(mcdc_coverage,[table_result]).
command_category(mcdc_coverage,table).
command_preferences_category(mcdc_coverage,mc_dc_commands).

command_description(mcdc_inv_coverage,'MC/DC Invariant Coverage','Show MC/DC Invariant Coverage').
command_call(mcdc_inv_coverage,[Result],[],mcdc_coverage:tcltk_compute_mcdc_invariant_coverage(Result)).
command_unavailable(mcdc_inv_coverage,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(mcdc_inv_coverage,[table_result]).
command_category(mcdc_inv_coverage,table).
command_preferences_category(mcdc_inv_coverage,mc_dc_commands).


command_description(read_write_matrix,'Operation Read/Write Matrix','Show identifiers read and written by operations').
command_call(read_write_matrix,[Result],[],b_read_write_info:tcltk_read_write_matrix(Result)).
command_unavailable(read_write_matrix,Txt) :- command_unavailable(operations,Txt).
command_arguments(read_write_matrix,[table_result]).
command_category(read_write_matrix,table).

command_description(variable_read_write_matrix,'Variable Read/Write Matrix','Show for variables which operations read and write them').
command_call(variable_read_write_matrix,[Result],[],b_read_write_info:tcltk_variable_read_write_matrix(no_check,Result)).
command_unavailable(variable_read_write_matrix,Txt) :- command_unavailable(variables,Txt).
command_arguments(variable_read_write_matrix,[table_result]).
command_category(variable_read_write_matrix,table).

command_description(machine_statistics,'Machine Statistics','Show statistics for main machine/specification').
command_call(machine_statistics,[Result],[],meta_interface:meta_get_machine_stats(Result)).
command_unavailable(machine_statistics,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(machine_statistics,[table_result]).
command_category(machine_statistics,table).

:- use_module(bmachine,[b_machine_statistics/2]).
:- public meta_get_machine_stats/1.
meta_get_machine_stats(list([list(['Key','Nr'])|L])) :- findall(list([Key,Nr]),b_machine_statistics(Key,Nr),L).


command_description(wd_pos,'WD POs','Show Well-Definedness Proof Obligations').
command_call(wd_pos,[Result],[],well_def_analyser:tcltk_get_machine_wd_pos(only_goal,Result,_,_)).
command_unavailable(wd_pos,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(wd_pos,[table_result]).
command_category(wd_pos,table).
command_preferences_category(wd_pos,wd_commands).

command_description(wd_pos_and_hyps,'WD POs and Hyps','Show Well-Definedness Proof Obligations and Hypotheses').
command_call(wd_pos_and_hyps,[Result],[],well_def_analyser:tcltk_get_machine_wd_pos(goal_and_hyps,Result,_,_)).
command_unavailable(wd_pos_and_hyps,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(wd_pos_and_hyps,[table_result]).
command_category(wd_pos_and_hyps,table).
command_preferences_category(wd_pos_and_hyps,wd_commands).

command_description(det_check_constants,'Compute Forced Constants','Analyse which values of CONSTANTS are forced and optionally compute an explanation').
command_call(det_check_constants,[Result],[],b_state_model_check:tcltk_cbc_constants_det_check(Result)).
command_unavailable(det_check_constants,Txt) :- command_unavailable(constants,Txt).
command_arguments(det_check_constants,[table_result]).
command_category(det_check_constants,table).
command_preferences_category(det_check_constants,cbc_commands).

command_description(unsat_core_properties,'Unsat Core of Properties/Axioms','Analyse a minimal unsatisfiable core of the PROPERTIES/Axioms').
command_call(unsat_core_properties,[Result],[],unsat_cores:unsat_core_properties_table(Result)).
command_unavailable(unsat_core_properties,'only available for B,Z or Event-B models with CONSTANTS or PROPERTIES') :-
       \+ (b_or_z_mode, b_machine_has_constants_or_properties).
command_arguments(unsat_core_properties,[table_result]).
command_category(unsat_core_properties,table).


:- use_module(runtime_profiler,[runtime_profile_available/0, tcltk_get_profile_info/1]).
command_description(prob_profile_info,'ProB Profile Info','Show runtime information for operations').
command_call(prob_profile_info,[Result],[],runtime_profiler:tcltk_get_profile_info(Result)).
command_unavailable(prob_profile_info,'no profiling information available, recompile probcli with -Dprob_profile=true') :- \+ runtime_profile_available,!.
command_unavailable(prob_profile_info,Txt) :- command_unavailable(b_mode,Txt).
command_arguments(prob_profile_info,[table_result]). % also accepts expression_raw_ast
command_category(prob_profile_info,table).

% a few typical reasons why commands are not available:
command_unavailable(constants,'only available for B,Z or Event-B models with CONSTANTS') :-
    \+ (b_or_z_mode, b_machine_has_constants).
command_unavailable(variables,'only available for B,Z or Event-B models with VARIABLES') :-
    \+ (b_or_z_mode, b_machine_has_variables).
command_unavailable(operations,'only available for B,Z or Event-B models with OPERATIONS/events') :-
   \+ (b_or_z_mode, b_machine_has_operations).
command_unavailable(b_mode,'only available for B,Z or Event-B models') :- \+ b_or_z_mode.
