% (c) 2020-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(optimizing_solver,[optimizing_solve_predicate/5,
             optimizing_solve_predicate/7, % with extra optimizing expression
             maxsolve_predicate/4,
             minsolve_predicate/4
             ]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module can be used to find optimal (maximal/minimal) solutions to a predicate.').


:- use_module(b_interpreter,[b_test_boolean_expression/6, set_up_typed_localstate/6]).
:- use_module(bsyntaxtree, [get_texpr_type/2]).
:- use_module(kernel_waitflags).
:- use_module(error_manager).
:- use_module(b_enumerate, [b_tighter_enumerate_values/2]).

% find a maximal model for TypedPred with free variables Parameters in given expanded state EState
% return solution found for Parameters in LocalState
maxsolve_predicate(EState, Parameters,TypedPred,LocalState) :-
    optimizing_solve_predicate(maximizing,EState, Parameters,TypedPred,LocalState).
% find a minimal model
minsolve_predicate(EState, Parameters,TypedPred,LocalState) :-
    optimizing_solve_predicate(minimizing,EState, Parameters,TypedPred,LocalState).

:- use_module(bsyntaxtree, [create_couple/2]).
% similar to test_bool_exists in eval_strings, but finds optimal solution:
optimizing_solve_predicate(OptMode,EState, Parameters,TypedPred,LocalState) :-
        create_couple(Parameters,OptimizingExpr),
        optimizing_solve_predicate(OptMode,EState, Parameters,TypedPred,OptimizingExpr,_,LocalState).

optimizing_solve_predicate(OptMode,EState, Parameters,TypedPred,OptimizingExpr,OptValue,LocalState) :-
        init_wait_flags(WF,[expansion_context(optimize_pred,Parameters)]),
        set_up_typed_localstate(Parameters,_SolTuple,TypedVals,[],LocalState,positive),
        b_tighter_enumerate_values(TypedVals,WF),
        b_interpreter:b_test_boolean_expression(TypedPred,LocalState,EState,WF), % test predicate to satisfy
        b_interpreter:b_compute_expression(OptimizingExpr,LocalState,EState,OptValue,WF), % compute value to optimise
        get_texpr_type(OptimizingExpr,OptType), % type of value to optimise
        format('Starting ~w solver (~w)~n',[OptMode,LocalState]),
        format('Optimizing: ',[]),translate:print_bexpr(OptimizingExpr),nl,
        safe_ground_wait_flag0(WF), % avoid doing deterministic computations every time in ground_for_optimal_sol

        % TO DO: we could select a subset of Parameters and SolTuple to optimize

        on_exception(user_interrupt_signal,
            ground_for_optimal_sol(1,OptMode,OptType,OptValue, WF,LocalState),
            (add_warning(optimizing_solver,'Search for optimal solution interrupted by user (CTRL-C): ',OptMode,Parameters),
             fail)
            ).
optimizing_solve_predicate(_, _,_,_,_,OptValue,LocalState) :-
        get_solution(OptValue,LocalState).



% TO DO: check for enumeration warnings, see e.g. gol_v2
% TO DO: add version with time-out and maximal number of improvements?
% TO DO: allow to pass initial solution (which could be solved using partitioning into components)
% TO DO: try to improve by component ? or we simply assume partitioning is done beforehand.

:- dynamic non_det_flag/0.
safe_ground_wait_flag0(WF) :- retractall(non_det_flag),
   ground_wait_flag0(WF),
   (non_det_flag -> add_warning(optimizing_solver,'Non-determinsm during WF0 grounding: ',WF)
    ; assert(non_det_flag)).

:- dynamic best_sol_found/2.
% we store the best solution found
% we avoid setting up the constraints completely fresh, we just backtrack to before
% grounding waitflags and add the constraint that we should improve
ground_for_optimal_sol(Count,OptMode,OptType,OptValue,WF,LocalState) :-
    ((ground_wait_flags(WF)
      -> format('Found solution at step ~w: ~w~n',[Count,OptValue]),
         retractall(best_sol_found(_,_)),
         asserta(best_sol_found(OptValue,LocalState)),
         fail % backtrack and undo ground_wait_flags above
              % TO DO: we could experiment with also backtracking beyond last best_sol_found predicate
      ;  true % no more optimization possible
     )
      -> fail
      ; C1 is Count+1,
        %print(improve_upon(BestSolTuple)),nl,
        ground_for_optimal_sol_with_reset(C1,OptMode,OptType,OptValue,WF,LocalState)
    ).

:-if(1=2).
% a version which accumulates the improve constraints and does not remove them
ground_for_optimal_sol_with_reset(C1,OptMode,OptType,OptValue,WF,LocalState) :-
    add_improve_constraint(OptMode,OptType,OptValue,WF),
    ground_for_optimal_sol(C1,OptMode,OptType,OptValue,WF,LocalState).
:- else.
% a version which removes the previous improvement constraint and adds a new one
% requires transitivity of the improve constraints
ground_for_optimal_sol_with_reset(Count,OptMode,OptType,OptValue,WF,LocalState) :-
    (( add_improve_constraint(OptMode,OptType,OptValue,WF),
       ground_wait_flags(WF)
      -> format('Found improved solution at step ~w: ~w~n',[Count,OptValue]),
         retractall(best_sol_found(_,_)),
         asserta(best_sol_found(OptValue,LocalState)),
         fail % backtrack and undo ground_wait_flags and add_improve_constraint above
      ;  true % no more optimization possible
     )
      -> fail
      ; C1 is Count+1,
        ground_for_optimal_sol_with_reset(C1,OptMode,OptType,OptValue,WF,LocalState)
    ).
:-endif.

:- use_module(kernel_objects,[not_equal_object_wf/3]).
add_improve_constraint(OptMode,OptType,OptValue,WF) :-
    best_sol_found(BestSolSoFar,_),!,
    copy_wf_start(WF,improve_upon,CWF), % unset WF0 in CWF in case it is set in WF
    not_equal_object_wf(OptValue,BestSolSoFar,CWF), % avoid finding same solution again
    (OptMode = maximizing
     -> better_or_equal(OptType,OptValue,BestSolSoFar,CWF) % find at least as good a solution
      ; better_or_equal(OptType,BestSolSoFar,OptValue,CWF)
    ),
    copy_wf_finish(WF,CWF).

get_solution(OptVal,LocalState) :-
         format('No more optimizing possible~n',[]),
         retract(best_sol_found(OptVal,LocalState)).
         %format('Solution=~w~n',[LocalState]).





% TO DO: provide possibility to provide custom optimization function, treat total functions, ... differently
% code related to LEQ_SYM_BREAK

:- use_module(kernel_objects,[greater_than_equal/2, less_than_equal_direct/2, check_subset_of_wf/3]).
better_or_equal(integer,NewVal,BestVal,_WF) :- greater_than_equal(NewVal,BestVal). % TODO: use absolute value?
better_or_equal(boolean,B1,B2,_WF) :- !, bool_geq(B1,B2).
better_or_equal(boolean,_,pred_false,_WF).
better_or_equal(couple(T1,T2),(N1,N2),(B1,B2),WF) :- better_or_equal(T1,N1,B1,WF), better_or_equal(T2,N2,B2,WF).
better_or_equal(set(_),NewVal,BestVal,WF) :- check_subset_of_wf(BestVal,NewVal,WF).
better_or_equal(seq(_),NewVal,BestVal,WF) :- check_subset_of_wf(BestVal,NewVal,WF).
better_or_equal(string,NewVal,BestVal,_WF) :- NewVal=BestVal. % possibly use length, prefix or @< order ?
better_or_equal(global(G),fd(V1,G),fd(V2,G),_WF) :- less_than_equal_direct(V2,V1). % use fd(_,_) numbering
better_or_equal(record(Fields),rec(F1),rec(F2),WF) :- better_or_equal_fields(Fields,F1,F2,WF).
better_or_equal(freetype(_),V,V,_). % TO DO: add support for freetypes
better_or_equal(freetype(_,_),V,V,_).


% one of the arguments is instantiated at least; otherwise we would need block declaration
bool_geq(B1,B2) :- var(B1),!, (B2=pred_true -> B1=pred_true ; true).
bool_geq(pred_true,_).
bool_geq(pred_false,pred_false).

better_or_equal_fields([],[],[],_).
better_or_equal_fields([field(Name,Type)|TT],[field(Name,V1)|T1],[field(Name,V2)|T2],WF) :-
   better_or_equal(Type,V1,V2,WF),
   better_or_equal_fields(TT,T1,T2,WF).

% examples for REPL:
% :max x:BOOL & y:BOOL & z:BOOL & (x/=y or y/=z)
% :max x <: 1..10 & y <: 1..10 & x /\ y = {} & 5 : x \/ y & card(x)=card(y)
% :max x:BOOL & y:BOOL & z:BOOL & (x=TRUE => y=TRUE) & (y=TRUE => z=TRUE)
% :max n=8 & n1=n-1 & q<:(1..n1)*(1..n1) & !(x,y).(x|->y:q => x/=y)

