% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(pref_definitions,[b_get_preferences_from_machine/0, b_get_preferences_from_machine/1,
         b_get_important_preferences_from_raw_machine/2,
         b_get_set_pref_definition/3,
         b_get_definition_string_list_from_spec/2, b_get_definition_string_from_spec/2]).

:- use_module(module_information).
:- module_info(group,tools).
:- module_info(description,'Tools for extracting information from DEFINITIONS.').

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(tools,[ajoin/2, ajoin_with_sep/3]).
:- use_module(bmachine,[b_definition_prefixed/5,b_get_typed_definition/3]).
:- use_module(preferences,[temporary_set_preference/2,eclipse_preference/2,preference_val_type/2]).
:- use_module(btypechecker,[unify_types_werrors/5]).
:- use_module(bsyntaxtree).
:- use_module(debug,[debug_println/2]).
% get and set important preferences from a raw, untyped machine (as they
% may influence type checking,...);
% example: IGNORE_PRJ_TYPES, WARN_IF_DEFINITION_HIDES_VARIABLE
b_get_important_preferences_from_raw_machine(RawMachine,RawModelType) :-
    (b_get_important_pref_type(RawModelType,Pref,PVal) ;
     b_get_important_pref_mach(RawMachine,Pref,PVal)),
    %print(pref(Pref,PType,PVal)),nl,
    (temporary_set_preference(Pref,PVal)
         -> debug_println(9,set_important_pref(Pref,PVal))
         ;  add_error(pref_definitions,'Could not set DEFINITION preference to specified value: ',PVal)
    ),
    fail.
b_get_important_preferences_from_raw_machine(_,_).

b_get_important_pref_mach(RawMachine,Pref,PVal) :-
    member(definitions(_Pos,Defs),RawMachine),
    member(expression_definition(_DPOS,DefName,[],RawVALUE),Defs),
    atom_concat('SET_PREF_',PREFString,DefName),
    eclipse_preference(PREFString,Pref),
    preference_val_type(Pref,PType),
    (PType=bool ; PType=xbool), % at the moment only boolean preferences supported
    convert_raw_pref_val(PType,RawVALUE,PVal).

b_get_important_pref_type(system(Pos),Pref,PVal) :-
   % SET_PREF_ALLOW_NEW_OPERATIONS_IN_REFINEMENT == TRUE;  for Event-B Atelier-B style; alternatively we could set animation_minor_mode ?
   add_message(pref_definitions,'Detected Event-B model; setting ALLOW_NEW_OPERATIONS_IN_REFINEMENT to ','TRUE',Pos),
   Pref=allow_new_ops_in_refinement, PVal=true.


convert_raw_pref_val(bool,boolean_true(_),true).
convert_raw_pref_val(bool,boolean_false(_),false).
%convert_raw_pref_val(int,integer(_,V),V).
%convert_raw_pref_val(nat,integer(_,V),V). % range checking done in temporary_set_preference
%convert_raw_pref_val(nat1,integer(_,V),V).
%convert_raw_pref_val(neg,integer(_,V),V).
%convert_raw_pref_val(range(_,_),integer(_,V),V).
%convert_raw_pref_val(string(_),string(_,S),S).

:- use_module(translate,[translate_bexpression/2]).
b_get_preferences_from_machine(List) :-
   findall(set_pref(String,ValS),(get_def_aux(String,TExpr),translate_bexpression(TExpr,ValS)),List).
% now a version that does not need to obtain a list with all preferences set:
b_get_preferences_from_machine :- get_def_aux(_,_),fail.
b_get_preferences_from_machine.

get_def_aux(String,TExpr) :-
   b_definition_prefixed(_,'SET_PREF_',String,DefName,DefPos),
   b_get_set_pref_definition(DefName,String,TExpr),
   set_pref_from_machine(String,DefPos,TExpr).

b_get_set_pref_definition(DefName,String,TExpr) :-
   b_definition_prefixed(_,'SET_PREF_',String,DefName,_DefPos),
   b_get_typed_definition(DefName,[],TExpr).

:- use_module(tools_matching,[get_possible_preferences_matches_msg/2]).
set_pref_from_machine(String,DefPos,TExpr) :-
   (eclipse_preference(String,Pref) -> true
    ; preference_val_type(String,_), Pref=String % user uses directly internal Prolog Name
    ),
   !,
   preference_val_type(Pref,PType),
   convert_pref_val(PType,TExpr,PVal),
   (temporary_set_preference(Pref,PVal)
     -> true
     ;  add_error(pref_definitions,'Could not set DEFINITION preference: ',(String:PVal),DefPos)
   ).
set_pref_from_machine(String,DefPos,_TExpr) :-
   get_possible_preferences_matches_msg(String,FuzzyMatches),
   !,
   ajoin(['Error in SET_PREF DEFINITION. Preference ',String,' does not exist! Did you mean: '],Msg),
   add_warning(pref_definitions,Msg,FuzzyMatches,DefPos),fail.
set_pref_from_machine(String,DefPos,_TExpr) :-
   add_warning(pref_definitions,'Error in SET_PREF DEFINITION. Preference does not exist: ',String,DefPos),fail.

convert_pref_val(int,TExpr,Val) :- !,
	check_type(integer,TExpr),
	compute_expr(TExpr,int(Val)).
convert_pref_val(nat,TExpr,Val) :- !,
	convert_pref_val(int,TExpr,Val),
	( (ground(Val),Val >= 0) -> true
	; otherwise -> error(TExpr,'Value must be a natural (>= 0)')).
convert_pref_val(neg,TExpr,Val) :- !,
	convert_pref_val(int,TExpr,Val),
	( (ground(Val),Val =< 0) -> true
	; otherwise -> error(TExpr,'Value must be non-positive (<= 0)')).
convert_pref_val(nat1,TExpr,Val) :- !,
	convert_pref_val(int,TExpr,Val),
	( (ground(Val),Val >= 1) -> true
	; otherwise -> error(TExpr,'Value must be a natural >= 1')).
convert_pref_val(range(Low,Up),TExpr,Val) :-!,
	convert_pref_val(int,TExpr,Val),
	( ground(Val),Val >= Low, Val =< Up -> true
	; otherwise -> ajoin(['Value must be in the interval [',Low,',',Up,']'],Msg),
	    error(TExpr,Msg)).
convert_pref_val(bool,TExpr,Val) :- !,
	check_type(boolean, TExpr),
	compute_expr(TExpr,BVal),
	( BVal = pred_true /* bool_true */ -> Val = true
	; BVal = pred_false /* bool_false */ -> Val = false
	; otherwise -> add_error(pref_definitions,'Unexpected result (neither true nor false)',BVal)).
convert_pref_val(xbool,TExpr,Val) :- !, % extended boolean
	(get_texpr_type(TExpr,boolean) -> true ; check_type(string,TExpr)),
	compute_expr(TExpr,BVal),
	( BVal = pred_true /* bool_true */ -> Val = true
	; BVal = pred_false /* bool_false */ -> Val = false
	; BVal=string(Val),member(Val,[false,true,full]) -> true
	; otherwise -> add_error(pref_definitions,'Unexpected result (neither true, false nor full)',BVal)).
convert_pref_val(string,TExpr,Val) :- !,
	check_type(string, TExpr),
	get_texpr_expr(TExpr,string(Val)).
convert_pref_val(path,TExpr,Val) :- !,
	check_type(string, TExpr),
	get_texpr_expr(TExpr,string(Val)).
convert_pref_val([H|T],TExpr,Val) :- !,
	check_type(string, TExpr),
	get_texpr_expr(TExpr,string(Val)),
	(member(Val,[H|T]) -> true
	 ; ajoin_with_sep(['Value must be one of:',H|T],' ',Msg),
	   error(TExpr,Msg)).
convert_pref_val(PrefType,_,_) :-
	add_error(pref_definitions,'Unsupported preference type in definitions:',PrefType),fail.

:- use_module(b_interpreter,[b_compute_expression_nowf/4]).
compute_expr(TExpr,Res) :-
             if( b_compute_expression_nowf(TExpr,[],[],R) , Res=R ,
             add_error_fail(pref_definitions,'Computing expression failed: ',TExpr)).
error(TExpr,Msg) :-
	get_texpr_pos(TExpr,Pos),
	add_all_perrors([error(Msg,Pos)]),
	fail.

check_type(Type,TExpr) :-
	get_texpr_type(TExpr,EType),
	get_texpr_pos(TExpr,Pos),
	unify_types_werrors(Type,EType,Pos,Errors,[]),
	( Errors = [] -> true
	; otherwise   -> add_all_perrors(Errors),fail).

% -------------------


:- use_module(tools,[safe_atom_codes/2, string_concatenate/3]).
:- use_module(bmachine,[b_definition_prefixed/5,b_get_definition_string_from_machine/2]).
:- use_module(specfile,[animation_mode/1]).
:- use_module(xtl_interface,[xtl_get_definition_string/2]).

   /* for tcltk: */
b_get_definition_string_list_from_spec(Def_Name_Prefix,Result) :-
    atom_codes(Def_Name_Prefix,Codes), append(Codes,_,CodesPrefix),
    findall(S, (b_get_definition_string_from_spec(Def_Name,DS),
                safe_atom_codes(Def_Name,CodesPrefix), /* check that name begins with the same characters as Def_Name_Prefix */
                string_concatenate(DS,';',S)), Result).

b_get_definition_string_from_spec(Def_Name,DS) :- animation_mode(xtl),!,
     xtl_get_definition_string(Def_Name,DS).
b_get_definition_string_from_spec(Def_Name,DS) :-
   b_get_definition_string_from_machine(Def_Name,DS).

