% (c) 2015-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(prob_state_predicates, [create_state_pred/4, create_state_pred/5]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smt_solvers).
:- module_info(description,'This module translates ProB\'s (local) states to predicates understandable for CVC4/Z3').

:- use_module(library(lists), [maplist/2]).

:- use_module(probsrc(bsyntaxtree), [get_texpr_id/2,
                                     safe_create_texpr/3,
                                     conjunct_predicates/2]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(debug), [debug_format/3]).
:- use_module(probsrc(b_global_sets), [is_b_global_constant/3]).
:- use_module(probsrc(custom_explicit_sets), [expand_custom_set_to_list_now/2]).
:- use_module(probsrc(clpfd_interface), [clpfd_domain/3]).
:- use_module(probsrc(error_manager), [add_error/4]).

create_state_pred(IDs,LS,S,P) :-
   create_state_pred(IDs,LS,S,P,[allow_to_skip_vars(_)]).

create_state_pred(IDs,LS,S,P,Opts) :-
    create_state_preds(IDs,LS,S,Preds,Opts),
    conjunct_predicates(Preds,P),
    translate_bexpression(P,PP),
    debug_format(9,'smt_solvers_interface state_pred: ~w~n', [PP]).
create_state_preds([],_,_,[b(truth,pred,[])],_).
create_state_preds([b(identifier(Name),Type,Infos)|IDs],LS,S,[SubPred|Preds],Opts) :-
    (  member(bind(Name,Value),LS)
     ; member(bind(Name,Value),S)
    ),!,
    create_state_preds_aux(Name,Type,Infos,Value,SubPred,Opts),
    create_state_preds(IDs,LS,S,Preds,Opts).
create_state_preds([_|IDs],LS,S,Preds,Opts) :-
    create_state_preds(IDs,LS,S,Preds,Opts).

:- use_module(probsrc(kernel_tools),[ground_value/1]).

% for integers: create fd interval
create_state_preds_aux(Name,integer,Infos,int(Value),Pred,_) :- !,
    (number(Value) ->
       safe_create_texpr(integer(Value),integer,TVal),
       safe_create_texpr(greater_equal(b(identifier(Name),integer,Infos),TVal),pred,Pred)
     ;
       clpfd_domain(Value,Min,Max),
       (number(Min)
       -> safe_create_texpr(integer(Min),integer,MinValue),
          safe_create_texpr(greater_equal(b(identifier(Name),integer,Infos),MinValue),pred,Greater)
       ;  Greater = b(truth,pred,[])),
      (number(Max)
       -> safe_create_texpr(integer(Max),integer,MaxValue),
          safe_create_texpr(less_equal(b(identifier(Name),integer,Infos),MaxValue),pred,Less)
       ;  Less = b(truth,pred,[])),
       conjunct_predicates([Less,Greater],Pred)
     ).
% non-integers: create equality if ground and simple enough
create_state_preds_aux(Name,Type,Infos,Value,Pred,_) :-
    ground_value(Value),
    useful_value_for_state_pred(Value), !,
    Pred = b(equal(b(identifier(Name),Type,Infos),b(value(Value),Type,[])),pred,[]).
% else: create truth (i.e. no futher state information)
create_state_preds_aux(ID,_,_,Val,b(truth,pred,[]),Opts) :-
    (var(Val) % for :z3 these are the quantified variables themselves; in general we should ensure there are no constraints, co-routines pending on Val
     ; member(allow_to_skip_vars(true),Opts)
       -> debug_format(19,'Not creating state predicate for ~w (value too complex)~n',[ID])
     ; member(skipped_variables(L),Opts), member(ID,L)
       -> debug_format(19,'Not creating state predicate for ~w (value too complex)~n',[ID])
    ),
    !.
create_state_preds_aux(ID,_,Infos,Val,b(truth,pred,[]),_) :-
    translate:print_bvalue(Val),nl,
    add_error(create_state_pred,'Ignoring value (too complex) for identifier',ID,Infos). % TO DO: enum warning ; if no solution found ok


useful_value_for_state_pred([]). % empty sets
useful_value_for_state_pred([H|T]) :- % sets as lists
    maplist(useful_value_for_state_pred,[H|T]).
useful_value_for_state_pred(avl_set(Val)) :-
    expand_custom_set_to_list_now(avl_set(Val),Values),
    maplist(useful_value_for_state_pred,Values).
useful_value_for_state_pred(int(Nr)) :- number(Nr).
useful_value_for_state_pred(term(floating(Nr))) :- number(Nr).
useful_value_for_state_pred(pred_true).
useful_value_for_state_pred(pred_false).
useful_value_for_state_pred(global_set(_)).
useful_value_for_state_pred(fd(Num,Name)) :-
    is_b_global_constant(Name,Num,_).
%useful_value_for_state_pred(rec(Fields)) :- useful_fields(Fields). % mk_value failed for test 1719 if we comment this in
useful_value_for_state_pred((A,B)) :-
    useful_value_for_state_pred(A),
    useful_value_for_state_pred(B).

%useful_fields([]).
%useful_fields([field(_,Val)|T]) :- useful_value_for_state_pred(Val),useful_fields(T).
