% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(state_custom_dot_graph,[tcltk_generate_state_custom_dot_graph/1, state_custom_dot_graph_available/0]).

:- use_module(module_information).
:- module_info(group,dot).
:- module_info(description,'This module provides a way to generate custom state graphs (using info from DEFINITIONS).').

:- use_module(dot_graph_generator).

%:- use_module(self_check).
:- use_module(error_manager).

:- use_module(state_space,[current_expression/2]).
:- use_module(bsyntaxtree, [get_texpr_id/2]).
:- use_module(specfile, [state_corresponds_to_fully_setup_b_machine/2]).
:- use_module(debug, [debug_println/2]).
:- use_module(bmachine, [b_get_machine_custom_nodes_function/2,b_get_machine_custom_edges_function/2]).
:- use_module(preferences,[valid_rgb_color/1, valid_dot_shape/1,
                           valid_dot_line_style/1, valid_dot_node_style/1]).


state_custom_dot_graph_available :-
   b_get_machine_custom_nodes_function(_,_),
   b_get_machine_custom_edges_function(_,_).

tcltk_generate_state_custom_dot_graph(File) :-
       current_expression(_CurID,State),
       state_corresponds_to_fully_setup_b_machine(State,BState),!,
       (b_get_machine_custom_nodes_function(_,_)
        -> true
        ; add_error_and_fail(tcltk_generate_state_custom_dot_graph,'No CUSTOM_GRAPH_NODES Function in DEFINITIONS')),
       (b_get_machine_custom_edges_function(_,_)
        -> true
        ; add_error_and_fail(tcltk_generate_state_custom_dot_graph,'No CUSTOM_GRAPH_EDGES Function in DEFINITIONS')),
       retractall(custom_nodes(_,_,_,_)),
       retractall(custom_edges(_,_,_,_)),
       retractall(custom_node_id(_,_)), retractall(nodectr(_)),
       retractall(custom_trans_label(_,_)), retractall(trans_ctr(_)), assert(trans_ctr(1)),
       eval_defs(BState),
       gen_dot_graph(File,state_custom_dot_graph,node_predicate,trans_predicate,none,none),
       print(finished),nl.
tcltk_generate_state_custom_dot_graph(_) :-
       add_error_and_fail(tcltk_generate_state_custom_dot_graph,'Only possible for initialised B machine.').

:- use_module(library(lists),[select/3]).
eval_defs(BState) :-
       b_get_machine_custom_nodes_function(NodesFunction,Nr),
       eval_definition_fun(NodesFunction, BState, Infos, NodesRes),
       debug_println(19,custom_nodes(Nr,NodesRes)),
       (select(label/EFID,Infos,RestInfos)
         -> assert(custom_nodes(EFID,Nr,RestInfos,NodesRes))
        ; get_texpr_id(NodesFunction,EFID)
         -> assert(custom_nodes(EFID,Nr,Infos,NodesRes))
          ; % we cannot derive name, then generate one
            number_codes(Nr,NC), append("custom",NC,CC), atom_codes(Custom,CC),
            assert(custom_nodes(Custom,Nr,Infos,NodesRes))),
       fail.
eval_defs(BState) :-
       b_get_machine_custom_edges_function(EdgesFunction,Nr),
       eval_definition_fun(EdgesFunction, BState, Infos, EdgesRes),
       debug_println(19,custom_edges(Nr,EdgesRes)),
       (select(label/Label,Infos,RestInfos) -> true
        ; get_texpr_id(EdgesFunction,Label) -> RestInfos=Infos
        ; b_get_machine_custom_edges_function(_,Nr2), Nr2 \= Nr % other relation exists; makes sense to use label
         -> number_codes(Nr,NC), append("custom",NC,CC),
            atom_codes(Label,CC),RestInfos=Infos
        ; otherwise -> Label='',RestInfos=Infos
       ),
       assert(custom_edges(Label,Nr,RestInfos,EdgesRes)),
       fail.
eval_defs(_).

:- dynamic custom_nodes/4, custom_edges/4.


:- use_module(tools,[string_escape/2]).
:- public node_predicate/6.
node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,Color) :- SubGraph=none,
     custom_nodes(_ID,_Nr,Infos,Nodes),
     (var(Nodes) -> add_error(node_predicate,'Variable custom_nodes: ',Nodes),fail ; true),
     get_node_and_colour(Nodes,Infos,NodeVal,Color,Shape,Style),
     (translate_value_and_escape(NodeVal,NodeDesc) -> true ; NodeDesc = '???'),
     %tools:print_message(node(NodeVal,ColVal,Color,NodeID)),
     gen_id(NodeVal,NodeID). %tools:print_message(nodeid(NodeID)).

get_node_and_colour(CustomNodes,Infos,NodeVal,ColVal,Shape,Style) :-
   (member(style/Style,Infos) -> true ; Style=filled), % to do: extract multiple styles
   (member(shape/DefShape,Infos) -> true ; DefShape=box),
   (member(color/DefaultCol,Infos) -> true ; DefaultCol=none),
   member(Node,CustomNodes),
   deconstruct_node(Node,DefaultCol,NodeVal,ColVal,DefShape,Shape).

deconstruct_node(((NodeVal,string(Shape)),ColVal),_,ResNodeVal,ResColVal,_,Shape) :-
   valid_dot_shape(Shape),
   !, % we have a triple (Node,"box","colour") = ((Node,"box"),"colour")
   ResNodeVal=NodeVal,
   translate_bvalue_to_colour(ColVal,ResColVal).
deconstruct_node((NodeVal,string(ColVal)),_,ResNodeVal,ResColVal,Shape,Shape) :-
   valid_rgb_color(ColVal),
   !, % we have a pair (Node,"colour")
   ResNodeVal=NodeVal, ResColVal=ColVal.
deconstruct_node((NodeVal,string(Shape)),DefaultCol,ResNodeVal,ResColVal,_,Shape) :-
   valid_dot_shape(Shape),
   !, % we have a pair (Node,"shape")
   ResNodeVal=NodeVal, ResColVal=DefaultCol.
deconstruct_node(NodeVal,DefaultCol,NodeVal,DefaultCol,Shape,Shape).
   % shapes: triangle,ellipse,box,diamond,hexagon,octagon,house,invtriangle,invhouse,invtrapez,doubleoctagon,egg,parallelogram,pentagon,trapezium...

:- dynamic custom_node_id/2, nodectr/1.
nodectr(0).

get_ctr(Res) :- retract(nodectr(C)),!, C1 is C+1, assert(nodectr(C1)), Res=C.
get_ctr(0) :- assert(nodectr(1)).

% generate or lookup ID for node
gen_id(NodeVal,ID) :-
 (custom_node_id(NodeVal,ID) -> true
   ; get_ctr(C),
     assert(custom_node_id(NodeVal,C)),ID=C).

% lookup ID and generate message if it does not exist
lookup_id(NodeVal,ID) :-
 (custom_node_id(NodeVal,ID) -> true
   ; format(user_error,'Node not specified in CUSTOM_GRAPH_NODES: ~w~n',[NodeVal]),
     gen_id(NodeVal,ID)
 ).

:- dynamic custom_trans_label/2, trans_ctr/1.
trans_ctr(1).
gen_trans_color(Label,Col,_,IsColor) :- custom_trans_label(Label,C),!,Col=C,IsColor=false.
gen_trans_color(_Label,Col,Infos,IsColor) :- member(color/C,Infos),!,Col=C,IsColor=false.
gen_trans_color(Label,Col,_,IsColor) :- try_translate_bvalue_to_colour(Label,C),!,Col=C,IsColor=true.
gen_trans_color(Label,Col,_,IsColor) :-  retract(trans_ctr(Ctr)), C1 is Ctr+1, assert(trans_ctr(C1)),
   translate_bvalue_to_colour(int(Ctr),C),
   assert(custom_trans_label(Label,C)), Col=C,IsColor=false.


:- public trans_predicate/5.
trans_predicate(NodeID,Label,SuccID,Color,Style) :-
   custom_edges(DefaultLabel,_,Infos,Edges),
   (member(style/Style,Infos) -> true ; Style=solid),
   (member(color/DefaultCol,Infos) -> true ; DefaultCol=blue),
   (var(Edges) -> add_error(trans_predicate,'Variable custom_edges: ',Edges),fail ; true),
   %format(user_output,'Edges (~w,~w):~w~n',[DefaultLabel,Style,Edges]),
   member((From,To),Edges),
   trans_pred_aux(From,(DefaultLabel,DefaultCol),Infos,To,NodeID,SuccID,Label,Color).

% we have a transition where the colour is specified (e.g., CUSTOM_GRAPH_EDGES == {n1,col,n2 | ... }
% we assume trans_predicate has been called first
trans_pred_aux(FromValue,_Defaults,Infos,To,NodeID,SuccID,ELabel,Color) :-
   % ((From,LabelCol) |-> To) case
   get_trans_label_and_color(FromValue,Infos,From,Color,ELabel), % the From Value contains color and label
   !,
   lookup_id(From,NodeID),
   lookup_id(To,SuccID).
trans_pred_aux(From,_Defaults,Infos,(LabelVal,ToVal),NodeID,SuccID,ELabel,Color) :-
   % (From |-> (LabelCol,To)) case
   get_trans_label_and_color((ToVal,LabelVal),Infos,To,Color,ELabel),  % the To Value contains color and label
   !,
   lookup_id(From,NodeID),
   lookup_id(To,SuccID).
trans_pred_aux(From,(Label,Col),_Infos,To,NodeID,SuccID,Label,Col) :-
   % no label or color specified in transition pair
   lookup_id(From,NodeID),
   lookup_id(To,SuccID).

% try and decompose a from value, detecting color and label string  ((From,"label"),"color")
get_trans_label_and_color((FromVal,LabelVal),Infos,From,Color,ELabel) :-
    gen_trans_color(LabelVal,Infos,Color,IsColor),
    (IsColor
     -> get_trans_label(FromVal,From,ELabel)
     ; translate_value_and_escape(LabelVal,ELabel),
       From=FromVal
    ).

get_trans_label((From,string(LabelVal)),From,ELabel) :- string_escape(LabelVal,ELabel).
get_trans_label(From,From,'').  % or should we use DefaultLabel

translate_value_and_escape(string(S),ELabel) :- !, string_escape(S,ELabel).
translate_value_and_escape(LabelVal,ELabel) :-
    translate:translate_bvalue(LabelVal,Label),string_escape(Label,ELabel).

:- use_module(b_interpreter,[b_compute_expression_nowf/4]).
:- use_module(custom_explicit_sets,[try_expand_custom_set_with_catch/3]).

eval_definition_fun(AnimFunction,BState, SInfos, FunctionRes) :- %trace,
       b_compute_expression_nowf(AnimFunction,[],BState,FunctionResCl),
       %nl, print('FunctionResult'(FunctionResCl)),nl,
       extract_dot_function(FunctionResCl, Infos, FunctionRes),
       sort(Infos,SInfos).

extract_dot_function((Label,Value),Info,FunctionRes) :-  % treat e.g. ("red","F",F)
       extract_info(Label,Info,Info2),!,
       extract_dot_function(Value,Info2,FunctionRes).
extract_dot_function((Value,Label),Info,FunctionRes) :-
       extract_info(Label,Info,Info2),!,
       extract_dot_function(Value,Info2,FunctionRes).
extract_dot_function(rec(Fields),Info,FunctionRes) :- !,
       extract_info_from_fields(Fields,Info,Vals),
       (Vals = [field(_,Value)] % just a single record field remaining
       % TO DO: check if field name is ok: value, edges, nodes, ...?
        -> try_expand_custom_set_with_catch(Value, FunctionRes,extract_dot_function)
        ;  add_message(state_custom_graph,'CUSTOM_GRAPH record not ok (use fields color,shape,style,label,description and either edges or nodes): ',Vals),
           fail
       ).
extract_dot_function(Value,[],FunctionRes) :- try_expand_custom_set_with_catch(Value,FunctionRes,extract_dot_function).

:- use_module(translate, [translate_bvalue/2]).
extract_info((A,B),I0,I2) :- !, extract_info(A,I0,I1), extract_info(B,I1,I2).
extract_info(string(Str),[Type/Str|T],T) :-
    (infer_string_type(Str,Type) -> true ; Type=label).
extract_info(Value,[label/Str|T],T) :- simple_label_value(Value),
    translate_bvalue(Value,Str).
simple_label_value(fd(_,_)).

infer_string_type(Str,color) :- valid_rgb_color(Str).
infer_string_type(Str,shape) :- valid_dot_shape(Str).
infer_string_type(Str,style) :- valid_dot_line_style(Str).
infer_string_type(Str,style) :- valid_dot_node_style(Str).

extract_info_from_fields([],[],[]).
extract_info_from_fields([field(FName,string(Str))|TF],[Type/Str|TI],Val) :-
    dot_attribute_field(FName,Type),!,
    (  Type=label -> true
     ; Type = description -> true
     ; infer_string_type(Str,Type) -> true
     ; add_message(state_custom_dot_graph,'Unexpected value for dot attribute: ',Type/Str)
    ),
    extract_info_from_fields(TF,TI,Val).
extract_info_from_fields([F|TF],Info,[F|TVal]) :- extract_info_from_fields(TF,Info,TVal).

dot_attribute_field(color,color).
dot_attribute_field(colour,color).
dot_attribute_field(shape,shape).
dot_attribute_field(style,style).
dot_attribute_field(description,description).
dot_attribute_field(label,label).
