% (c) 2018-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(tools_matching,
                 [fuzzy_match_codes_lower_case/2,
                  fuzzy_match_codes/2,
                  codes_to_lower_case/2, % to lower case, also performs Unicode simplifications
                  get_current_keywords/1,
                  get_possible_preferences/1, get_possible_preferences_matches_msg/2,
                  get_possible_fuzzy_matches_msg/3,
                  get_possible_completions_msg/3
                  ]).

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(library(lists)).

:- use_module(module_information).

:- module_info(group,infrastructure).
:- module_info(description,'A few utilities for fuzzy matching and completion.').



:- assert_must_succeed(tools_matching:fuzzy_match_codes_lower_case("a","A")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes_lower_case("aBcD","ABCd")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes_lower_case("aBcD","ABxCd")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes_lower_case("aBxcD","ABCd")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes_lower_case("aBcD","ABCdx")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes_lower_case("aBcDx","ABCd")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes_lower_case("a_Bc_D","AB__Cd")).
%:- assert_must_succeed(tools_matching:fuzzy_match_codes_lower_case("äÄ","aA")).
:- assert_must_fail(tools_matching:fuzzy_match_codes_lower_case("abc","cba")).


fuzzy_match_codes_lower_case(Codes1,Codes2) :-
   codes_to_lower_case(Codes1,LCodes1),
   codes_to_lower_case(Codes2,LCodes2),
   fuzzy_match_codes(LCodes1,LCodes2).

:- assert_must_succeed(tools_matching:fuzzy_match_codes("aBcD","aBcD")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes("aBxcD","aBcD")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes("aBcD","aBcxD")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes("aBcD","aBcDx")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes("xaBcD","aBcD")).
:- assert_must_succeed(tools_matching:fuzzy_match_codes("aBcD","xaBcD")).
:- assert_must_fail(tools_matching:fuzzy_match_codes("abc","ABC")).

fuzzy_match_codes([],[]).
fuzzy_match_codes([H|T1],[H|T2]) :- !,fuzzy_match_codes(T1,T2).
fuzzy_match_codes([_|T],[_|T]) :- !. % one character rewritten
fuzzy_match_codes([H1|T1],L2) :- possible_skip_char(H1),!,fuzzy_match_codes(T1,L2).
fuzzy_match_codes(L1,[H2|T2]) :- possible_skip_char(H2),!,fuzzy_match_codes(L1,T2).
fuzzy_match_codes([_|T],T) :- !. % one character too much
fuzzy_match_codes(T,[_|T]) :- !. % one character too few


%:- assert_must_succeed(tools_matching:codes_to_lower_case("äÄöAa","aaoaa")).

codes_to_lower_case(Codes,LC) :- maplist(code_to_lower_case,Codes,LC).
% TO DO: normalise more UNICODE symbols, ...

code_to_lower_case(Char,LC_Char) :- Char >= 65, Char =< 90,!, LC_Char is Char+32.
code_to_lower_case(Char,LC_Char) :- Char >= 8320, Char =< 8329,!, LC_Char is Char-8272. % Unicode Subscripts
code_to_lower_case(8242,R) :- !, R=8242. % Unicode Prime
code_to_lower_case(8216,R) :- !, R=8242.
code_to_lower_case(8217,R) :- !, R=8242.
code_to_lower_case(Char,R) :- Char >= 192, Char =< 197,!, R=97. % upper-case a
code_to_lower_case(Char,R) :- Char >= 224, Char =< 229,!, R=97. % lower-case a
code_to_lower_case(Char,R) :- Char >= 200, Char =< 203,!, R=101. % upper-case e
code_to_lower_case(Char,R) :- Char >= 232, Char =< 235,!, R=101. % lower-case e
code_to_lower_case(Char,R) :- Char >= 204, Char =< 207,!, R=105. % upper-case i
code_to_lower_case(Char,R) :- Char >= 236, Char =< 239,!, R=105. % lower-case i
code_to_lower_case(Char,R) :- Char >= 210, Char =< 214,!, R=111. % upper-case o
code_to_lower_case(Char,R) :- Char >= 242, Char =< 246,!, R=111. % lower-case o
code_to_lower_case(Char,R) :- Char >= 217, Char =< 220,!, R=117. % upper-case u
code_to_lower_case(Char,R) :- Char >= 249, Char =< 252,!, R=117. % lower-case u
code_to_lower_case(253,R) :- !, R=121. % ý -> y
code_to_lower_case(209,R) :- !, R=110. % Ñ -> n
code_to_lower_case(241,R) :- !, R=110. % ñ -> n
code_to_lower_case(231,R) :- !, R=99. % ç -> c
code_to_lower_case(223,R) :- !, R=115. % ß -> s
code_to_lower_case(C,C).

% use_module(library(between)), between(150,255,R), atom_codes(A,[R]), format("~w : ~w~n",[R,A]),fail.

possible_skip_char(95). % _

:- use_module(specfile,[b_or_z_mode/0, animation_minor_mode/1]).

get_current_keywords(List) :- b_or_z_mode,!,
   (animation_minor_mode(Minor) -> Mode=Minor ; Mode=b),
   get_keywords(Mode,[expr,predicate],List).
get_current_keywords([]). % TO DO: support csp mode

get_keywords(Mode,Types,List) :-
   findall(ID,(keyword(ID,Type,Modes), member(Mode,Modes), member(Type,Types)),Ids),
   sort(Ids,List).

% list of language specific and context specific keywords
keyword(not,predicate,[b,eventb]).
keyword(or,predicate,[b,eventb]).
keyword('TRUE',expr,[b,eventb]).
keyword('FALSE',expr,[b,eventb]).
keyword('BOOL',expr,[b,eventb]).
keyword('bool',expr,[b,eventb]).
keyword('POW',expr,[b,eventb]).
keyword('POW1',expr,[b,eventb]).
keyword('FIN',expr,[b,eventb]).
keyword('FIN1',expr,[b,eventb]).
keyword('union',expr,[b,eventb]).
keyword('inter',expr,[b,eventb]).
keyword('UNION',expr,[b,eventb]).
keyword('INTER',expr,[b,eventb]).
keyword('INTEGER',expr,[b]).
keyword('NATURAL',expr,[b]).
keyword('NATURAL1',expr,[b]).
keyword('INT',expr,[b,eventb]).
keyword('NAT',expr,[b,eventb]).
keyword('NAT1',expr,[b,eventb]).
keyword('MININT',expr,[b]).
keyword('MAXINT',expr,[b]).
keyword('min',expr,[b,eventb]).
keyword('max',expr,[b,eventb]).
keyword('SIGMA',expr,[b]).
keyword('PI',expr,[b]).
keyword('dom',expr,[b,eventb]).
keyword('ran',expr,[b,eventb]).
keyword('id',expr,[b,eventb]).
keyword('prj1',expr,[b,eventb]).
keyword('prj2',expr,[b,eventb]).
keyword('pred',expr,[b,eventb]).
keyword('succ',expr,[b,eventb]).
keyword('closure',expr,[b]).
keyword('closure1',expr,[b]).
keyword('iterate',expr,[b]).
keyword('fnc',expr,[b]). % also Event-B ?
keyword('rel',expr,[b]).

keyword('seq',expr,[b]).
keyword('seq1',expr,[b]).
keyword('iseq',expr,[b]).
keyword('iseq1',expr,[b]).
keyword('perm',expr,[b]).
keyword('size',expr,[b]).
keyword('rev',expr,[b]).
keyword('first',expr,[b]).
keyword('last',expr,[b]).
keyword('front',expr,[b]).
keyword('tail',expr,[b]).
keyword('conc',expr,[b]).
keyword('struct',expr,[b]).
keyword('rec',expr,[b]).
keyword('STRING',expr,[b]).

keyword('skip',subst,[b]).
keyword('CASE',subst,[b]).
keyword('SELECT',subst,[b]).
keyword('PRE',subst,[b]).

keyword('IF',_,[b]).
keyword('THEN',_,[b]).
keyword('ELSE',_,[b]).
keyword('LET',_,[b]).
keyword('BE',_,[b]).
keyword('END',_,[b]).

% TO DO: add keywords for TLA, Z minor modes

:- use_module(preferences,[eclipse_preference/2]).
get_possible_preferences(SPrefs) :-
   findall(Pref,eclipse_preference(Pref,_),P),
   sort(P,SPrefs).

get_possible_preferences_matches_msg(String,FuzzyMatchMsg) :-
    get_possible_preferences(Prefs),
    (get_possible_fuzzy_matches_msg(String,Prefs,FuzzyMatchMsg) ;
     get_possible_completions_msg(String,Prefs,FuzzyMatchMsg)).

get_possible_fuzzy_matches(ID,AllIDs,FuzzyMatches) :-
    atom_codes(ID,IDCodes),
    findall(Target,(member(Target,AllIDs),atom_codes(Target,TargetCodes),
                   fuzzy_match_codes_lower_case(IDCodes,TargetCodes)),FuzzyMatches).

% get possible matches as atom which can be used after phrase: Did you mean:
get_possible_fuzzy_matches_msg(ID,AllIDs,Msg) :-
   get_possible_fuzzy_matches(ID,AllIDs,FuzzyMatches),
   get_match_msg(FuzzyMatches,Msg).

get_match_msg(FuzzyMatches,Msg) :-
   length(FuzzyMatches,Nr), Nr>0,
   get_msg(FuzzyMatches,Nr,Msg).

:- use_module(tools_strings,[ajoin/2,ajoin_with_sep/3]).
get_msg([Match],1,Res) :- !, Res=Match.
get_msg(List,Nr,Msg) :- Nr < 6, !,
   ajoin_with_sep(List,',',Msg).
get_msg([First|_],Nr,Msg) :- N1 is Nr-1,
   ajoin([First,' (',N1,' more matches)'],Msg).



% get possible completions as atom which can be used after phrase: Did you mean:
get_possible_completions_msg(ID,SortedAllIDs,Msg) :-
    atom_codes(ID,IDCodes0),
    codes_to_lower_case(IDCodes0,IDCodes),
    findall(Target,(member(Target,SortedAllIDs),atom_codes(Target,TargetCodes),
                    codes_to_lower_case(TargetCodes,TC2), prefix(TC2,IDCodes)
                   ),Completions),
   get_match_msg(Completions,Msg).

