% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(typing_tools,[
     is_infinite_type/1, % check if a type is infinite for animation purposes
     is_provably_finite_type/1, % check if a type is provably finite (not just for animation)
     is_finite_type_in_context/2, % check if a type is finite, either in animation or proving context
     valid_ground_type/1, % check if a type is a valid ground type
     contains_infinite_type/1,
     any_value_for_type/2,
     normalize_type/2,
     create_maximal_type_set/2, % converts sequences to cartesian product
     create_type_set/2, % keeps sequence types
     create_type_set/3 % with flag for keeping sequences types or not
]).
:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module provides utilities for B types.').

%:- use_module(bsyntaxtree, [get_texpr_types/2, syntaxtraversion/6,  find_typed_identifier_uses/3]).
:- use_module(library(lists), [maplist/2, maplist/3]).

% check if a type is infinite for ProB animation; deferred sets are not counted as infinite here!
is_infinite_type(integer).
is_infinite_type(string).
is_infinite_type(freetype(_)).
is_infinite_type(set(X)) :- is_infinite_type(X).
is_infinite_type(seq(_)).
is_infinite_type(couple(X,Y)) :- (is_infinite_type(X) -> true ; is_infinite_type(Y)).
is_infinite_type(record(Fields)) :-
   ((member(field(_,Type),Fields), is_infinite_type(Type)) -> true).

contains_infinite_type([H|T]) :-
  (is_infinite_type(H) -> true ; contains_infinite_type(T)).

:- use_module(b_global_sets,[unfixed_deferred_set/1]).

% this a finite type for the prover, not just for animation
% e.g., deferred sets in Event-B are not finite; in classical B they are in principle finite
is_provably_finite_type(boolean).
is_provably_finite_type(set(X)) :- is_provably_finite_type(X).
is_provably_finite_type(couple(X,Y)) :- is_provably_finite_type(X),is_provably_finite_type(Y).
is_provably_finite_type(global(G)) :-
  \+ unfixed_deferred_set(G). % Note: in Classical B in principle all SETS are finite !
  % note that scope_ DEFINITIONS make the set G finite here: TODO : investigate what we want/need
is_provably_finite_type(record(Fields)) :- maplist(finite_field,Fields).

finite_field(field(_,Type)) :- is_provably_finite_type(Type).



is_finite_type_in_context(animation,Type) :-
    !,\+ is_infinite_type(Type). % all deferred sets counted as finite here; assumes record_construction already ran
is_finite_type_in_context(proving,Type) :-
    is_provably_finite_type(Type). % here deferred sets can be in principle infinite; see Classical B comment above


:- use_module(error_manager).

valid_ground_type(X) :- var(X),!, add_error(valid_ground_type,'Variable as type: ',X),fail.
valid_ground_type(X) :- valid_ground_type_aux(X),!.
valid_ground_type(X) :- add_error(valid_ground_type,'Illegal type term: ',X),fail.

%:- use_module(b_global_sets,[b_global_set/1]).

valid_ground_type_aux(any) :- format(user_error,'! Type contains *any*~n',[]). % can happen, e.g., for test 949 due to assertion prj1({{}},BOOL)({}|->TRUE) = {}
valid_ground_type_aux(integer).
valid_ground_type_aux(real).
valid_ground_type_aux(string).
valid_ground_type_aux(boolean).
valid_ground_type_aux(global(G)) :- atom(G).
 %bmachine:b_get_machine_set(G).
 %b_global_sets:b_global_set(G).
valid_ground_type_aux(freetype(FT)) :- ground(FT).
valid_ground_type_aux(set(X)) :- valid_ground_type(X).
valid_ground_type_aux(seq(X)) :- valid_ground_type(X).
valid_ground_type_aux(couple(X,Y)) :- valid_ground_type(X), valid_ground_type(Y).
valid_ground_type_aux(record(Fields)) :- valid_fields(Fields,'$prev_field_name').
valid_ground_type_aux(constant([X])) :- ground(X). % appears in Z freetype cases, test 738

valid_fields([],_).
valid_fields([field(Name,_)|_],_) :- \+ atom(Name),!,
   add_error(valid_ground_type,'Illegal record field name: ',Name),fail.
valid_fields([field(Name,_)|_],PrevName) :- Name @=< PrevName,!,
   add_error(valid_ground_type,'Record field names not sorted: ',PrevName:Name),fail.
valid_fields([field(Name,Type)|T],_) :-
   valid_ground_type(Type),
   valid_fields(T,Name).


% --------------

:- use_module(kernel_freetypes,[get_freeval_type/3]).
:- use_module(b_global_sets,[b_get_fd_type_bounds/3,get_global_type_value/3]).
any_value_for_type(T,V) :- if(inst2(T,V),true, (nl,print('*** instantiate_to_any_value failed: '), print(T:V),nl,nl %,trace,inst2(T,V)
                              )).
%inst2(_,V) :- ground(V),!.
%inst2(Type,V) :- print(inst2(Type,V)),nl,trace,fail.
inst2(boolean,V) :- !, inst3(V,pred_true).
inst2(pred,V) :- !, inst3(V,pred_true).
inst2(integer,V) :- !, V=int(X),inst_fd(X,0).
inst2(set(_),V) :- !,  inst3(V,[]). % TO DO: what if V already partially instantiated
inst2(seq(_),V) :- !,  inst3(V,[]). % TO DO: what if V already partially instantiated
inst2(couple(TA,TB),V) :- !, V=(A,B), inst2(TA,A), inst2(TB,B).
%inst2(real,V) :- !, V=... % TO DO
inst2(string,V) :- !, V=string(X), inst3(X,'').
inst2(global(T),V) :- !, get_global_type_value(V,T,X),
  b_get_fd_type_bounds(T,LowBnd,_UpBnd),
  inst_fd(X,LowBnd).
inst2(record(Fields),rec(Vals)) :- !, instantiate_fields(Fields,Vals).
inst2(constant([A]),V) :- !, V=term(A).
inst2(freetype(FT),V) :- !, V=freeval(FT,CASE,Val),
  (nonvar(CASE) -> (get_freeval_type(FT,CASE,Type) -> any_value_for_type(Type,Val)) % ensure no backtracking
    ; nonvar(Val) -> print((nonvar_value(V))),nl, get_freeval_type(FT,CASE,Type), any_value_for_type(Type,Val)
    ; (get_freeval_type(FT,C,Type) -> CASE=C, any_value_for_type(Type,Val))
  ).
inst2(X,V) :- print(cannot_inst2(X,V)),nl.

inst3(X,V) :- (var(X) -> X=V ; true).

:- use_module(probsrc(clpfd_interface), [clpfd_some_element_in_domain/2]).
inst_fd(X,_) :- nonvar(X),!.
inst_fd(X,_Default) :- clpfd_some_element_in_domain(X,El),!, X=El.
inst_fd(X,X).

instantiate_fields([],[]).
instantiate_fields([field(Name,Type)|T],[field(Name,Val)|TV]) :-
   any_value_for_type(Type,Val),
   instantiate_fields(T,TV).

% ----------------------

% replace seq(X) by set(couple(integer,X))

normalize_type(X,Y) :-
  (normalize_type_aux(X,Y) -> true ; add_internal_error('Illegal type: ',normalize_type(X,Y)),Y=X).

normalize_type_aux(V,R) :- var(V),!,R=V.
normalize_type_aux(any,any).
normalize_type_aux(pred,pred).
normalize_type_aux(subst,subst).
normalize_type_aux(boolean,boolean).
normalize_type_aux(integer,integer).
normalize_type_aux(string,string).
normalize_type_aux(real,real).
normalize_type_aux(constant(C),constant(C)).
normalize_type_aux(global(F),global(F)).
normalize_type_aux(freetype(F),freetype(F)).
normalize_type_aux(set(X),set(N)) :- normalize_type_aux(X,N).
normalize_type_aux(seq(X),set(couple(integer,N))) :- normalize_type_aux(X,N).
normalize_type_aux(couple(X,Y),couple(NX,NY)) :- normalize_type_aux(X,NX),normalize_type_aux(Y,NY).
normalize_type_aux(record(F),record(NF)) :- normalize_type_fields(F,NF).
normalize_type_aux(op(A,B),op(NA,NB)) :- maplist(normalize_type_aux,A,NA), maplist(normalize_type_aux,B,NB).

normalize_type_fields(V,R) :- var(V),!,R=V.
normalize_type_fields([],[]).
normalize_type_fields([field(N,T)|R],[field(N,NT)|RN]) :- normalize_type_aux(T,NT),normalize_type_fields(R,RN).

% -------------------------

:- use_module(bsyntaxtree, [create_texpr/4]).

% translate a ProB type into a B expression so that we can write x:TSet
% translate_type:
create_type_set(Type,TSet) :- create_type_set(Type,true,TSet).
create_type_set(Type,KeepSeq,TSet) :-
    ( var(Type) -> add_error_and_fail(translate,'type_set Type is variable: ',Type)
    ; otherwise ->
        type_set2(Type,KeepSeq,Set,Infos),
        create_texpr(Set,set(Type),Infos,TSet)).
type_set2(set(T),KeepSeq,pow_subset(Set),[]) :-
    create_type_set(T,KeepSeq,Set).
type_set2(seq(T),KeepSeq,Res,Info) :-
  (KeepSeq = true
   -> Res=seq(Set), Info=[], create_type_set(T,KeepSeq,Set) % keep sequence types as is
    ; type_set2(set(couple(integer,T)),KeepSeq,Res,Info)). % write as POW(INTEGER*T)
type_set2(couple(TA,TB),KeepSeq,cartesian_product(SetA,SetB),[]) :-
    create_type_set(TA,KeepSeq,SetA), create_type_set(TB,KeepSeq,SetB).
%type_set2(global(G),_,identifier(G),[given_set]).
type_set2(global(G),_,value(global_set(G)),[given_set]). % avoids variable capture/clash
type_set2(integer,_,integer_set('INTEGER'),[]).
type_set2(string,_,string_set,[]).
type_set2(real,_,real_set,[]).
type_set2(boolean,_,bool_set,[]).
type_set2(record(TFields), KeepSeq, struct(Rec),[]) :-
    create_texpr(rec(SFields),record(TFields),[],Rec),
    type_set_fields(TFields,KeepSeq,SFields).
%type_set2(freetype(Id),identifier(Id),[given_set]).
% Missing: no case for constant([C]) term
type_set2(freetype(Id),_,freetype_set(Id),[]).
type_set2(constant(L),_,value(Terms),[]) :-  % these types occur in Event-B for Theory plugin inductive datatypes
    maplist(create_term,L,Terms).
create_term(T,term(T)).
type_set_fields([],_,[]).
type_set_fields([field(Id,Type)|TFrest],KeepSeq,[field(Id,TSet)|EFrest]) :-
    create_type_set(Type,KeepSeq,TSet),
    type_set_fields(TFrest,KeepSeq,EFrest).

% translate a ProB type to a typed expression representing the maximal type
% set_type_to_maximal_texpr
create_maximal_type_set(ProBType,TExpr) :- create_type_set(ProBType,false,TExpr).
  %replace_seq(ProBType,MType),create_type_set(MType,TExpr).




