% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(unsat_cores,[unsat_core/2,
                       unsat_core_list_with_time_limit/5, unsat_chr_core_list_with_auto_time_limit/5,
                       unsat_core_with_fixed_conjuncts/3, unsat_chr_core_with_fixed_conjuncts_auto_time_limit/4,
                       minimum_unsat_core/2,
                       minimum_unsat_core_with_fixed_conjuncts/3,
                       tcltk_unsat_core_properties/2, unsat_core_properties_table/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,misc).
:- module_info(description,'Computation of unsatisfiable cores of predicates.').

:- use_module(library(lists),[min_member/3, reverse/2, select/3]).

:- use_module(bsyntaxtree,[conjunction_to_list/2,
                           conjunct_predicates/2,
                           is_truth/1,
                           create_negation/2,
                           create_implication/3,
                           add_texpr_infos/3,
                           add_labels_to_texpr/3]).
:- use_module(solver_interface, [solve_predicate/4]).
:- use_module(preferences, [get_preference/2]).
:- use_module(debug, [debug_println/2, silent_mode/1, set_silent_mode/1]).
:- use_module(self_check, [assert_must_succeed/1,assert_must_fail/1]).

:- assert_must_succeed((unsat_core(b(falsity,pred,[]),Core), (ground(Core), Core = b(falsity,pred,_)))).
:- assert_must_succeed((unsat_core(b(conjunct(b(falsity,pred,[]),b(truth,pred,[])),pred,[]),Core), (ground(Core), Core = b(falsity,pred,_)))).

:- use_module(translate,[translate_bexpression_with_limit/3, set_unicode_mode/0, unset_unicode_mode/0]).
:- use_module(specfile,[get_specification_description/2]).
:- use_module(error_manager,[get_tk_table_position_info/2]).
:- use_module(tools_strings,[ajoin/2]).
% for meta_interface and ProB2:
unsat_core_properties_table(list([list([PropsHeader,'Source'])|ResList])) :-
   get_specification_description(properties,PROPS),
   unsat_core_of_properties(CorePredicates,[no_warning_if_sol_found],StrRes),
   ajoin([PROPS,' : ',StrRes],PropsHeader),
   set_unicode_mode,
   findall(list([TS,PosInfo]),
         (member(CP,CorePredicates), translate_bexpression_with_limit(CP,200,TS),
          get_tk_table_position_info(CP,PosInfo)),ResList),
    unset_unicode_mode.

:- use_module(library(lists), [maplist/2,maplist/3]).
tcltk_unsat_core_properties(list(CorePredStrings),StrRes) :-
   % TO DO: transmit runtime from Runtime to FINALISE SETUP_CONSTANTS to avoid re-checking Properties
   unsat_core_of_properties(CorePredicates,[],StrRes),
   set_unicode_mode,
   maplist(translate:translate_bexpression,CorePredicates,CorePredStrings),
   unset_unicode_mode,
   maplist(add_warn,CorePredicates). % so that we see the spans of the unsat core in the edit window

unsat_core_of_properties(CorePredicates,Options,StrRes) :-
   bmachine:b_get_properties_from_machine(Properties), % TO DO: also provide for unsat component
   write('Computing (efficient) unsat core for PROPERTIES'),nl,
   unsat_chr_core_list_with_auto_time_limit(Properties,5000,Options,GlobalResult,CorePredicates),
   (explain_result(GlobalResult,Explanation) -> StrRes=Explanation ; StrRes='**UNKNOWN**').

add_warn(Conjunct) :-
   add_warning(unsat_cores,'Part of Unsat core of PROPERTIES: ',Conjunct,Conjunct).

:- use_module(preferences, [temporary_set_preference/3, reset_temporary_preference/2]).

% calls unsat_core but with CHR and other preferences set
%unsat_core_chr(Pred,GlobalResult,Core) :-
%    temporary_set_preference(use_chr_solver,true,CHNG1),
%    call_cleanup(unsat_core(Pred,GlobalResult,Core),
%                 reset_temporary_preference(use_chr_solver,CHNG1)).

% TODO: cleanup / rewrite predicate in order to minimize cores
% i.e. implications could be rewritten into conjuncts

:- assert_must_fail((minimum_unsat_core(b(conjunct(b(falsity,pred,[]),
  b(conjunct(b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,[]),
  b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,[])),pred,[])),pred,[]),Core),
   (ground(Core), Core = b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,_)))).
:- assert_must_fail((minimum_unsat_core(b(conjunct(b(falsity,pred,[]),
                                           b(conjunct(b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,[]),b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,[])),pred,[])),pred,[]),Core),
                                           (ground(Core), Core = b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,_)))).
% there should be two cores
:- assert_must_succeed((findall(Core, unsat_core(b(conjunct(b(falsity,pred,[]),
                                      b(conjunct(b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,[]),b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,[])),pred,[])),pred,[]),Core),
                                      ListOfCores), length(ListOfCores,2))). % was 5

% call if you know how much time (in ms) it took to find inconsistency in Predicate
unsat_core_list_with_time_limit(Predicate,TimeToFindContradiction,Options,GlobalRes, CorePredicates) :-
    temporary_set_timeout(TimeToFindContradiction,CHNG1),
    % this time will be multiplied by 2 below
    call_cleanup(unsat_core_list(Predicate,Options,GlobalRes,CorePredicates),
                 reset_temporary_preference(time_out,CHNG1)).

:- use_module(error_manager, [add_warning/4, add_message/4]).
unsat_chr_core_list_with_auto_time_limit(Predicate,MaxTime,Options,GlobalRes,CorePredicates) :-
    temporary_set_preference(use_chr_solver,true,CHNG1),
    call_cleanup(unsat_core_wth_auto_time_limit(Predicate,MaxTime,Options,GlobalRes,CorePredicates),
                 reset_temporary_preference(use_chr_solver,CHNG1)).

% automatically determine time limit by checking that full predicate is false and registering time
unsat_core_wth_auto_time_limit(Predicate,MaxTime,Options,GlobalRes,CorePredicates) :-
    get_time_out_for_comp_pred(Predicate,_Nr,CompPred,TimeToFindContradiction,GlobalRes),
    progress_info(GlobalRes),
    (solution_found(GlobalRes)
      -> (member(no_warning_if_sol_found,Options)
          -> add_message(unsat_cores,'Predicate has a solution, unsat core cannot be computed: ',Predicate,Predicate)
           ; add_warning(unsat_cores,'Predicate has a solution, unsat core cannot be computed: ',Predicate,Predicate)
          ),
          CorePredicates=[Predicate]
      ;  (member(auto_time_out_factor(Factor),Options) -> true ; Factor=120),
         Limit is min(((TimeToFindContradiction*Factor)/100),MaxTime), %print(time_limit(Limit)),nl,
         unsat_core_list_with_time_limit(CompPred,Limit,Options,GlobalRes,CorePredicates)
      ).

:- use_module(library(lists),[nth1/3]).
:- use_module(bsyntaxtree,[predicate_components/2]).
%get_time_out_for_comp_pred(SubPred,1,SubPred,TimeToFindContradiction,Res) :- !,get_time_out_for_pred(SubPred,TimeToFindContradiction,Res).
get_time_out_for_comp_pred(Predicate,Nr,SubPred,TimeToFindContradiction,Res) :- % partition
    predicate_components(Predicate,Components),
    if( (nth1(Nr,Components,component(SubPred,_Ids)), % select component via backtracking
         get_time_out_for_pred(SubPred,TimeToFindContradiction,Res),
         progress_info(Res),
         %print(sol_comp(Nr,Res)),nl,
         \+ solution_found(Res)),
         true,
         (SubPred=Predicate,
          get_time_out_for_pred(SubPred,TimeToFindContradiction,Res)
         )
      ).

get_time_out_for_pred(Predicate,TimeToFindContradiction,Res) :-
    silent_mode(CurMode), set_silent_mode(on),
    statistics(walltime,[T1,_]),
    solve_predicate(Predicate, _State, 2, Res),
    statistics(walltime,[T2,_]), TimeToFindContradiction is T2-T1,
    set_silent_mode(CurMode).

temporary_set_timeout(TimeToFindContradiction,CHNG1) :-
    number(TimeToFindContradiction), TimeToFindContradiction>=0,
    NewTimeOut is TimeToFindContradiction+10, % in case we have 0; just add a few ms
    temporary_set_preference(time_out,NewTimeOut,CHNG1).

unsat_core(Predicate,CorePredicate) :-
    unsat_core(Predicate,no_solution_found,CorePredicate).
unsat_core(Predicate,GlobalResult,CorePredicate) :-
     unsat_core_list(Predicate,[],GlobalResult,CorePreds),
     conjunct_predicates(CorePreds,CorePredicate).

unsat_core_list(Predicate,Options,GlobalResult,CorePreds) :-
    uc_conjunction_to_list(Predicate,PredicateList),
    length(PredicateList,OrigLen),
    silent_mode(CurMode), set_silent_mode(on),
    call_cleanup(unsat_core_with_options_lists(PredicateList,Options,GlobalResult,CorePreds),
                 set_silent_mode(CurMode)),
    (CurMode=on -> true; length(CorePreds,NewLen), format('~nFound core with ~w/~w conjuncts.~n',[NewLen,OrigLen])).

% strip outermost existential quantifiers so that we can examine the conjunction inside
uc_conjunction_to_list(b(exists(_V,Pred),pred,_),List) :- !,
   uc_conjunction_to_list(Pred,List).
% TO DO: add let_predicate (decomposed further below) ? forall ?
uc_conjunction_to_list(Pred,List) :- conjunction_to_list(Pred,List).

% called by ProB2
unsat_chr_core_with_fixed_conjuncts_auto_time_limit(Predicate,FixedConjuncts,MaxTimeLimit,CorePredicate) :-
    temporary_set_preference(use_chr_solver,true,CHNG1),
    call_cleanup(unsat_core_with_fixed_conjuncts_auto_time_limit(Predicate,FixedConjuncts,MaxTimeLimit,CorePredicate),
                 reset_temporary_preference(use_chr_solver,CHNG1)).
unsat_core_with_fixed_conjuncts_auto_time_limit(Predicate,FixedConjuncts,MaxTimeLimit,CorePredicate) :-
    is_truth(FixedConjuncts),!,
    Options=[],
    unsat_core_wth_auto_time_limit(Predicate,MaxTimeLimit,Options,_,CorePreds),
    conjunct_predicates(CorePreds,CorePredicate).
unsat_core_with_fixed_conjuncts_auto_time_limit(Predicate,FixedConjuncts,MaxTimeLimit,CorePredicate) :-
    conjunct_predicates([Predicate,FixedConjuncts],FullPred),
    get_time_out_for_pred(FullPred,TimeToFindContradiction,GlobalRes), % we cannot use component version due to FixedConjuncts
    progress_info(GlobalRes),
    (solution_found(GlobalRes)
      -> add_warning(unsat_cores,'Predicate has a solution, unsat core cannot be computed: ',Predicate,Predicate),
          CorePredicate=Predicate
      ;  Limit is min((TimeToFindContradiction*110/100),MaxTimeLimit),
         temporary_set_timeout(Limit,CHNG1),
         call_cleanup(unsat_core_with_fixed_conjuncts(Predicate,FixedConjuncts,GlobalRes,CorePredicate),
                      reset_temporary_preference(time_out,CHNG1))
      ).

unsat_core_with_fixed_conjuncts(P,FC,Core) :- unsat_core_with_fixed_conjuncts(P,FC,no_solution_found,Core).
unsat_core_with_fixed_conjuncts(Predicate,FixedPreds,GlobalRes,CorePredicate) :-
    conjunction_to_list(FixedPreds,FixedList),
    Options = [fixed_conjuncts(FixedList)],
    unsat_core_list(Predicate,Options,GlobalRes,CorePreds),
    conjunct_predicates(CorePreds,CorePredicate).

unsat_core_with_options_lists(PredicateList,Options,GlobalResult,CorePreds) :-
    get_preference(unsat_core_algorithm,divide_and_conquer), !, % TO DO: pass GlobalResult
    temporary_set_preference(allow_enumeration_of_infinite_types,false,OldValueOfPref),
    (member(fixed_conjuncts(FixedList),Options) -> true ; FixedList=[]),
    call_cleanup(unsat_core_fixed_conjuncts_lists_divide_and_conquer(PredicateList,FixedList,GlobalResult, CorePreds),
                 reset_temporary_preference(allow_enumeration_of_infinite_types,OldValueOfPref)).
unsat_core_with_options_lists(PredicateList,Options,GlobalResult,CorePreds) :-
    reset_max_core_size,
    temporary_set_preference(allow_enumeration_of_infinite_types,true, OldValueOfPref),
    %print('START: '),nl,translate:nested_print_bexpr(RPredicateList),nl,
    call_cleanup(unsat_core_with_options_linear(PredicateList,Options, GlobalResult, CorePreds),
                 reset_temporary_preference(allow_enumeration_of_infinite_types, OldValueOfPref)),
    (member(branch_and_bound,Options) -> length(CorePreds,Len), set_max_core_size(Len) ; true).


unsat_core_with_options_linear(PredicateList,Options, GlobalRes, CorePreds) :-
    reverse(PredicateList, RPredicateList), % so that we properly treat WD we start removing from the back
    compute_cores_linear(RPredicateList, Options, GlobalRes, CorePreds).

compute_cores_linear(RPredicateList, Options, GlobalRes, CorePreds) :-
    compute_core_linear_loop(RPredicateList, 0, [], Options, GlobalRes, InCore,NotInCore,AllPreds),!,
    (  CorePreds = InCore
     ; % find more cores:
       write('R'),
       restart_core_linear(InCore,NotInCore,AllPreds,Options,GlobalRes,CorePreds)
     ).

% restart the unsat core algorithm with one conjunct from core removed from the original predicate list
restart_core_linear([C|Cs],Keep,AllPreds,Options,GlobalRes,CorePreds) :-
    append(Keep,Cs,LL0), add_fixed_conj(Options,LL0,LL),
    conjunct_predicates(LL, ConjunctionToCheck),
    %  print('TRY TO REMOVE CORE CONJUNCT FOR RESTART: '), translate:print_bexpr(C),nl,
    solve_predicate(ConjunctionToCheck, _State, 1, Res),
    progress_info(Res),
    must_keep_conjunct(Res,GlobalRes),!,
    % we need to keep C in PredicateList for restart; otherwise we change global result
    restart_core_linear(Cs,[C|Keep],AllPreds,Options,GlobalRes,CorePreds).
restart_core_linear([C|_],_,AllPreds,Options,GlobalRes,CorePreds) :-
    % we can remove C from the global list
    (get_preference(unsat_core_algorithm,linear_greedy) -> ! ; true),
    remove_conjunct(C,AllPreds,NewAllPreds), % remove C, keeping original order
    write('C'),
    compute_cores_linear(NewAllPreds, Options, GlobalRes, CorePreds).
restart_core_linear([C|Cs],Keep,AllPreds,Options,GlobalRes,CorePreds) :-
    % try to remove another conjunct; we may find the same core multiple times
    restart_core_linear(Cs,[C|Keep],AllPreds,Options,GlobalRes,CorePreds).

add_fixed_conj(Options,P,Res) :- member(fixed_conjuncts(FixedList),Options),!,
    (append(FixedList,P,FP) -> Res=FP
      ; add_internal_error('Illegal fixed list: ',FixedList), Res=P).
add_fixed_conj(_,P,P).

:- use_module(probsrc(bsyntaxtree),[same_texpr/2]).
:- use_module(probsrc(error_manager),[add_internal_error/2]).
remove_conjunct(C,RPredicateList,NewRPredicateList) :-
    select(CC,RPredicateList,NewRPredicateList), % TO DO: remember position of C and use nth1
    same_texpr(C,CC),!.
remove_conjunct(C,RPredicateList,New) :-
    add_internal_error('Could not find conjunct from unsat core: ',remove_conjunct(C,RPredicateList,New)),fail.


% implies_optional_hyp(C,DefinitelyInCore,_Hyp)
% implied_by_core(C,DefinitelyInCore)
% implies_optional_hyp(C,DefinitelyInCore,_Hyp)

% simple unsat core computation: remove conjuncts till satisfiability is reached
compute_core_linear_loop([], _,DefinitelyInCore, _Options, _, DefinitelyInCore,[],[]).
compute_core_linear_loop([C|Cs], Counter,DefinitelyInCore, Options, GlobalRes, CorePreds,NotInCore,AllPreds) :-
    % remove C and check for satisfiability
    append(DefinitelyInCore,Cs,LL0), add_fixed_conj(Options,LL0,LL),
    conjunct_predicates(LL, ConjunctionToCheck),
    % print('TRY TO REMOVE: '), translate:print_bexpr(C),nl,
    solve_predicate(ConjunctionToCheck, _State, 1, Res),
    debug_println(4, solve_predicate_result(Res)),
    progress_info(Res),
    must_keep_conjunct(Res,GlobalRes), % a solution has been found or a time-out occured - C can not be removed;
      % normally corresponds to solve_result: contradiction_found or no_solution_found(unfixed_deferred_sets)
    !,
     % print('KEEPING: '), translate:print_bexpr(C),nl,
     % TO DO: when looking for minimal unsat core we could provide an upper-bound for the length
     increase_core_counter(Counter,Options,IncCtr),
     (can_be_further_decomposed(C, CA, CB)
      -> % print('% Split: '), translate:print_bexpr(CA),nl, print('%        '),translate:print_bexpr(CB),nl,
         compute_core_linear_loop([CA,CB|Cs], Counter,DefinitelyInCore, Options, GlobalRes, CorePreds,NotInCore,AllPreds)
      ;  add_uc_info(Res, C, IC), AllPreds = [C|TAP],
         compute_core_linear_loop(Cs, IncCtr,[IC|DefinitelyInCore], Options, GlobalRes, CorePreds,NotInCore,TAP)
     ).
compute_core_linear_loop([C|Cs], Ctr, DefinitelyInCore, Options, GlobalRes, CorePreds,NotInCore,[C|AllPreds]) :-
    % format('CAN BE REMOVED (~w): ',[GlobalRes]), translate:print_bexpr(C),nl,
    % C can be removed - we no longer introduce a choice point to not remove C; further cores are found by restarting
     NotInCore = [C|RestNotInCore],
     compute_core_linear_loop(Cs, Ctr, DefinitelyInCore, Options, GlobalRes, CorePreds,RestNotInCore,AllPreds).

:- dynamic max_core_size/1.
increase_core_counter(Cur,Options,_) :- max_core_size(Max), Cur >= Max,
       member(branch_and_bound,Options),!,format('B~w',[Max]),fail.
increase_core_counter(Cur,_,C1) :- C1 is Cur+1.

reset_max_core_size :- retractall(max_core_size(_)).
set_max_core_size(Max) :- retractall(max_core_size(_)), assert(max_core_size(Max)).

% check if an optional hypothesis in the current core is implied by the rest and the new hyp C
%implies_optional_hyp(C,CurrentCore,Hyp) :-
%     \+ is_truth(CurrentCore),
%      select(Hyp,CurrentCore,Rest), % DO THIS TEST also for C which must be kept
%      bsyntaxtree:get_texpr_info(Hyp,IL), member(optionally_in_core,IL),
%      % print('    --> '),translate:print_bexpr(Hyp), print(' <== '),translate:print_bexpr(Rest),nl,
%      prove_sequent(animation,Hyp,[C|Rest]).
%
%:- use_module(wdsrc(well_def_analyser),[prove_sequent/3]).
%implied_by_core(C,CurrentCore) :- prove_sequent(animation,C,CurrentCore).


add_uc_info(SolverRes, C, IC2) :-
  (  ( ground(SolverRes),
       SolverRes = solution(_), Labels = [],
       UCInfo = [description('definitely in unsat core (solution)')])
  ;  ( SolverRes == error, Labels = ['WD'],
       UCInfo = [description('definitely in unsat core (WD error)')])
  ;  ( SolverRes == time_out, Labels = ['TO'],
       UCInfo = [description('possibly in unsat core (timeout)')])
  ;  ( ground(SolverRes), Labels = ['VTO'],
       SolverRes = no_solution_found(enumeration_warning(_,_,_,_,_)),
       UCInfo = [description('possibly in unsat core (virtual timeout)')])
  ;  ( SolverRes == no_solution_found(unfixed_deferred_sets), Labels = ['DEFSETS'],
       UCInfo = [description('definitely in unsat core (unfixed deferred sets)')])
  ), !,
  add_texpr_infos(C, UCInfo, IC1),
  add_labels_to_texpr(IC1,Labels,IC2).

solution_found(Res) :-
    Res \= contradiction_found,
    Res \= no_solution_found(_),
    Res \= time_out,
    Res \= error,
    Res \= clpfd_overflow.

% check if a removed conjunct has to be kept in unsat core given solver_result (w/o conjunct) and global result
must_keep_conjunct(solution(_),_) :- !.
must_keep_conjunct(contradiction_found,_) :- !,fail. %
must_keep_conjunct(no_solution_found(_),_) :- !,fail. % X \= enumeration_warning(_,_,_,_,critical) ?
must_keep_conjunct(X,X) :- !,fail. % same result as global result, e.g., WD error or time_out
must_keep_conjunct(_,_).

can_be_further_decomposed(TE,A,B) :- can_be_further_decomposed(TE,positive,A,B).
can_be_further_decomposed(b(E,pred,Info),NegPos,A,B) :- can_be_further_decomposed_aux(E,Info,NegPos,A,B).
can_be_further_decomposed_aux(conjunct(A,B),_,positive,A,B).
can_be_further_decomposed_aux(disjunct(A,B),_,negative,A,B).
% TO DO: :core ((x=2 & y=3) or (x=3 & y=4)) & x>3   --> being able to remove y=3 and y=4 from unsat core
can_be_further_decomposed_aux(implication(A,B),_,positive,I1,I2) :- % A=>B1&B2  <-->  A=>B1 & A=>B2
    can_be_further_decomposed(B,positive,B1,B2),
    create_implication(A,B1,I1),
    create_implication(A,B2,I2).
can_be_further_decomposed_aux(implication(A,B),_,negative,NA,B) :- % A=>B  <-->  not(A) or B
    create_negation(A,NA).
can_be_further_decomposed_aux(negation(D),_,NegPos,NA,NB) :- neg_context(NegPos,PosNeg),
    can_be_further_decomposed(D,PosNeg,A,B),
    create_negation(A,NA),
    create_negation(B,NB).
can_be_further_decomposed_aux(let_predicate(V,E,P),Info,NegPos,LEA,LEB) :-
   can_be_further_decomposed(P,NegPos,A,B),
   LEA = b(let_predicate(V,E,A),pred,Info), % used_ids could be wrong; TO DO: think about exists
   LEB = b(let_predicate(V,E,B),pred,Info).
can_be_further_decomposed_aux(lazy_let_pred(V,E,P),Info,NegPos,LEA,LEB) :-
   can_be_further_decomposed(P,NegPos,A,B),
   LEA = b(lazy_let_pred(V,E,A),pred,Info),
   LEB = b(lazy_let_pred(V,E,B),pred,Info).

neg_context(positive,negative). neg_context(negative,positive).

% a binary search algorithm in order to only need log(n) instead of n solver calls
% Idea: split list of conjuncts in half, C = [A,B]
unsat_core_fixed_conjuncts_lists_divide_and_conquer([],_,_,_) :- !, fail. % we split too far!
unsat_core_fixed_conjuncts_lists_divide_and_conquer([C],_FixedConjuncts,_GlobalResult,[C]) :-
    !. % if there is only one conjunct left, it has to belong to the core
unsat_core_fixed_conjuncts_lists_divide_and_conquer(Clauses,FixedConjuncts,GlobalResult,Core) :-
    split_in_half(Clauses,A,B),
    % print('SPLIT: '), print(A),nl,print(B),nl,
    unsat_core_fixed_conjuncts_lists_divide_and_conquer_split(A,B,FixedConjuncts,GlobalResult,Core).

unsat_core_fixed_conjuncts_lists_divide_and_conquer_split(A,B,FixedConjuncts,GlobalResult,Core) :-
    append(A,FixedConjuncts,AllConjuncts),
    conjunct_predicates(AllConjuncts,APred),
    % print('SOLVING LHS: '), translate:print_bexpr(APred),nl,
    solve_predicate(APred,_AState,1,ARealRes),
    progress_info(ARealRes),
    % print(result(ARealRes)),nl,
    !, % do not backtrack into solving or multiple results can be found leading to duplicated cores
    unsat_core_fixed_conjuncts_lists_divide_and_conquer_recur1(ARealRes,A,B,FixedConjuncts,GlobalResult,Core).

progress_info(ARes) :- progress_info_aux(ARes),!,flush_output(user_output).
progress_info_aux(time_out) :- write('T').
progress_info_aux(error) :- write('WD').
progress_info_aux(no_solution_found(unfixed_deferred_sets)) :- write('-').
progress_info_aux(contradiction_found) :- write('-').
progress_info_aux(solution(_)) :- write('+').
progress_info_aux(no_solution_found(_)) :- write('V').
progress_info_aux(clpfd_overflow) :- write('O').
progress_info_aux(UNKNOWN) :- write('*'),write(UNKNOWN),write('*').

explain_result(time_out,'UNKNOWN (TIME_OUT)').
explain_result(error,'UNKNOWN (ERROR)').
explain_result(clpfd_overflow,'UNKNOWN (CLPFD OVERFLOW)').
explain_result(no_solution_found(_),'FALSE').
explain_result(contradiction_found,'FALSE').
explain_result(solution(_),'TRUE').
explain_result(_,'UNKNOWN').

unsat_core_fixed_conjuncts_lists_divide_and_conquer_recur1(ARealRes,A,_B,FixedConjuncts,GlobalResult,Core) :-
        \+ must_keep_conjunct(ARealRes,GlobalResult), % A is (virtually / effectively) unsat. Recur on A.
        unsat_core_fixed_conjuncts_lists_divide_and_conquer(A,FixedConjuncts,GlobalResult,Core).
unsat_core_fixed_conjuncts_lists_divide_and_conquer_recur1(ARealRes,A,B,FixedConjuncts,GlobalResult,Core) :-
        append(B,FixedConjuncts,AllConjuncts),
        conjunct_predicates(AllConjuncts,BPred),
        % print('SOLVING RHS: '), translate:print_bexpr(BPred),nl,
        solve_predicate(BPred,_BState,1,BRealRes),
        progress_info(BRealRes),
        % print(result(BRealRes)),nl,
        !,
        unsat_core_fixed_conjuncts_lists_divide_and_conquer_recur2(ARealRes,BRealRes,A,B,FixedConjuncts,GlobalResult,Core).

unsat_core_fixed_conjuncts_lists_divide_and_conquer_recur2(_ARealRes,BRealRes,_A,B,FixedConjuncts,GlobalResult,Core) :-
    \+ must_keep_conjunct(BRealRes,GlobalResult), % B is (virtually / effectively) unsat. Recur on B.
    !,
    unsat_core_fixed_conjuncts_lists_divide_and_conquer(B,FixedConjuncts,GlobalResult,Core).
unsat_core_fixed_conjuncts_lists_divide_and_conquer_recur2(ARealRes,BRealRes,A,B,FixedConjuncts,GlobalResult,Core) :-
    % the interesting case: if both are sat
    must_keep_conjunct(ARealRes,GlobalResult),
    must_keep_conjunct(BRealRes,GlobalResult),
    % minimize A assuming usat core includes B
    append(B,FixedConjuncts,FixedAndB),
    unsat_core_fixed_conjuncts_lists_divide_and_conquer(A,FixedAndB,GlobalResult,ACore),
    % now ACore & B is unsat and minimal but not minimum core
    % minimize B assuming ACore
    append(ACore,FixedConjuncts,FixedAndA),
    unsat_core_fixed_conjuncts_lists_divide_and_conquer(B,FixedAndA,GlobalResult,BCore),
    append(ACore,BCore,Core).

%divide_and_conquer_recur_condition(contradiction_found).
%divide_and_conquer_recur_condition(no_solution_found(X)) :-
%  % to do: if the outermost conjunct already generates a time_out or critical enumeration warning then this should be enabled as a recur condition
%    X \= enumeration_warning(_,_,_,_,critical).
%%divide_and_conquer_recur_condition(time_out).

split_in_half(C, A, B) :-
    length(C,L),
    Half is L // 2,
    length(A, Half),
    append(A, B, C).

% there is currently no known good algorithm to compute the minimum unsat core
% we can just compute every core and take the minimum one out of the list
:- assert_must_succeed((minimum_unsat_core(b(conjunct(b(falsity,pred,[]),
                                           b(conjunct(b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,[]),b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,[])),pred,[])),pred,[]),Core),
                                           (ground(Core), Core = b(falsity,pred,_)))).
minimum_unsat_core(Predicate,CorePredicate) :-
    minimum_unsat_core_with_fixed_conjuncts(Predicate,[],CorePredicate).

minimum_unsat_core_with_fixed_conjuncts(Predicate,FixedConjuncts,CorePredicate) :-
    Options = [branch_and_bound,fixed_conjuncts(FixedConjuncts)],
    findall(CoreList,
            unsat_core_list(Predicate,Options,no_solution_found,CoreList),
            Cores),
    reset_max_core_size,
    min_member(compare_length,MinCs,Cores),
    conjunct_predicates(MinCs,CorePredicate).

:- use_module(library(terms),[term_size/2]).
compare_length(P1,P2) :- length(P1,L1), length(P2,L2),
    (L1<L2 -> true
     ; L1=L2 ->
         term_size(P1, T1),
         term_size(P2, T2),
         T1 < T2
    ).
