% (c) 2020-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(well_def_analyser, [analyse_wd_for_machine/3, analyse_wd_for_machine/4,
  compute_wd/4, compute_all_wd_pos/4,
  analyse_wd_for_expr/3,
  annotate_wd/2,
  tcltk_get_machine_wd_pos/4,
  prove_sequent/1,
  prove_sequent/3
  ]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module computes WD conditions in AST.').

:- use_module(probsrc(error_manager)).
:- use_module(probsrc(debug)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(external_functions),[external_fun_has_wd_condition/1]).
:- use_module(probsrc(specfile),[z_or_tla_minor_mode/0, eventb_mode/0]).

:- use_module(library(ordsets)).
:- use_module(library(avl)).
:- use_module(well_def_hyps).


prove_sequent(Goal) :-
   prove_sequent(proving,Goal,[]).


prove_sequent(Context,Goal,Hyps) :-
   (Context=animation -> Val=true
    ; Context=proving -> Val=false
    ; add_internal_error('Unknown context:',Context),
      Val=false),
   temporary_set_preference(wd_analysis_for_animation,Val,Chng),
   call_cleanup(prove_sequent_cur_context(Goal,Hyps),
                reset_temporary_preference(wd_analysis_for_animation,Chng)).


% utility for disprover
prove_sequent_cur_context(_,Hyps) :-
   member(Falsity,Hyps), is_falsity(Falsity),!. % TO DO: upon pushing hyps
prove_sequent_cur_context(Goal,Hypotheses) :-
   empty_hyps(E),
   list_to_ord_set([discharge_po],Options),
   push_hyps(E,Hypotheses,Options,Hyps),
   prove_sequent_goal(Goal,Hyps,Options).

% process complex sequent goals which arise in Rodin; not generated by our POG
prove_sequent_goal(b(Formula,pred,I),Hyps,Options) :-
   process_sequent_aux(Formula,I,Hyps,Options).

process_sequent_aux(implication(A,B),_,Hyps,Options) :- !,
   (debug_mode(off) -> true ; print('Peeling implication in sequent goal: '), translate:print_bexpr(A),nl),
   push_hyp(Hyps,A,Options,Hyps2),
   prove_sequent_goal(B,Hyps2,Options).
process_sequent_aux(conjunct(A,B),_I,Hyps,Options) :- !,
   prove_sequent_goal(A,Hyps,Options),
   prove_sequent_goal(B,Hyps,Options).
process_sequent_aux(disjunct(A,B),_I,Hyps,Options) :- !,
   (negate_hyp(B,NegB),
    push_hyp(Hyps,NegB,Options,Hyps2), % not(B) => A we have A or B ;
                                       % is a combination of OR_R proof rule and proof by case B, not(B)
    prove_sequent_goal(A,Hyps2,Options)
    -> true
    ; negate_hyp(A,NegA),
      push_hyp(Hyps,NegA,Options,Hyps2), % not(A) => B we have A or B
      prove_sequent_goal(B,Hyps2,Options)).
process_sequent_aux(forall(Ids,A,B),_,Hyps,Options) :- !,
   (debug_mode(off) -> true ; print('Peeling forall in sequent goal: '), translate:print_bexpr(A),nl),
   add_new_hyp_variables(Hyps,Ids,Hyps1),
   push_hyp(Hyps1,A,Options,Hyps2),
   prove_sequent_goal(B,Hyps2,Options).
process_sequent_aux(Goal,I,Hyps,Options) :-
   get_normalized_and_renamed_predicate(b(Goal,pred,I),Hyps,_RenTarget,NT),
   % print('PROVING: '),translate:print_bexpr(_RenTarget),nl,
   (get_current_po_label(POLabel,Options) -> true ; POLabel = ''),
   prove_proof_obligation(POLabel,Goal,NT,Hyps,ProofTree),
   debug_format(19,'WD PROVER SUCCESSFUL: ~w~n',[ProofTree]).



compute_all_wd_pos(TExpr,OriginalHyps,Options,POs) :-
  findall(PO, compute_wd(TExpr,OriginalHyps,Options,PO), POs).


% analyze and print WD POs for a given typed expression or predicate or subst:
analyse_wd_for_expr(TExpr,_,_) :-
   reset_stats,
   formatsilent('WD Proof Obligations:~n',[]),
   empty_hyps(OriginalHyps),
   list_to_ord_set([create_full_po,create_not_discharged_msg(message),discharge_po,ignore_wd_infos
                   %,minimize_full_po
                   ],Options),
   compute_wd(TExpr,OriginalHyps,Options,PO),
   (silent_mode(on) -> true ; translate:nested_print_bexpr(PO),nl),
   fail.
analyse_wd_for_expr(TExpr,ResStr,AllDischarged) :- formatsilent('----~n',[]),
   print_stats,
   get_discharged_result(ResStr,_,_,AllDischarged),
   empty_hyps(OriginalHyps),
   transform_wd(TExpr,OriginalHyps,[discharge_po],NewTExpr),
   (silent_mode(on) -> true
     ; format('Annotated formula:~n',[]), temporary_set_preference(pp_wd_infos,true,Chng),
       translate:nested_print_bexpr(NewTExpr),nl,
       reset_temporary_preference(pp_wd_infos,Chng)
    ).

tcltk_get_machine_wd_pos(only_goal,list([Header|POs]),ResStr,AllDischarged) :- !,
    reset_stats,
    Header = list(['PO Label','Discharged','Proof Obligation Goal', 'Source']),
    list_to_ord_set([discharge_po,ignore_wd_infos,reorder_conjuncts],Options),
    findall(list([POLabel,Discharged,POStr,Source]),
             (get_machine_wd_po(Options,PO,PO_Name),
              get_texpr_info(PO,Infos),
              (member(discharged(D,POLabel),Infos)
               -> (D=true -> Discharged='yes' ; Discharged='no')
                ; Discharged='UNKNOWN',POLabel=PO_Name),
              translate:translate_bexpression_with_limit(PO,100,POStr),
              get_tk_table_position_info(Infos,Source)
              ), POs),
    get_discharged_result(ResStr,_,_,AllDischarged).
tcltk_get_machine_wd_pos(goal_and_hyps,list([Header|POs]),ResStr,AllDischarged) :-
    reset_stats,
    Header = list(['PO Label','Discharged','PO Goal', '(Necessary) Hypotheses', 'Source']),
    % in case a PO is proven we only show the hyps necessary for the proof
    list_to_ord_set([discharge_po,ignore_wd_infos,reorder_conjuncts,  create_full_po,minimize_full_po],Options),
    findall(list([POLabel,Discharged,GoalStr,HypsStr,Source]),
             (get_machine_wd_po(Options,PO,PO_Name),
              get_texpr_info(PO,Infos),
              (member(discharged(D,POLabel),Infos)
               -> (D=true -> Discharged='yes' ; Discharged='no')
                ; Discharged='UNKNOWN',POLabel=PO_Name),
              decompose_po(PO,Hyps,Goal),
              translate:translate_bexpression_with_limit(Hyps,100,HypsStr),
              translate:translate_bexpression_with_limit(Goal,100,GoalStr),
              get_tk_table_position_info(Infos,Source)
              ), POs),
    get_discharged_result(ResStr,_,_,AllDischarged).

decompose_po(b(implication(Hyps,Goal),pred,_),Hyps,Goal) :- !.
decompose_po(Goal,b(truth,pred,[]),Goal).

% analyse all WD POs for the loaded B machine
analyse_wd_for_machine(NrDischarged,NrTot,Discharged) :-
    Options = [create_not_discharged_msg(warning),discharge_po,ignore_wd_infos,reorder_conjuncts],
    analyse_wd_for_machine(NrDischarged,NrTot,Discharged,Options).

analyse_wd_for_machine(NrDischarged,NrTot,Discharged,UnsortedOptions) :- reset_stats,
    list_to_ord_set(UnsortedOptions,Options),
    %list_to_ord_set([discharge_po,reorder_conjuncts],Options),
    statistics(walltime,[Start,_]),
    ( get_machine_wd_po(Options,_PO,_),
      fail
      ;
        statistics(walltime,[Stop,_]), Delta is Stop-Start,
        format('Analysis walltime: ~w ms~n',[Delta]),
        print_stats,
        get_discharged_result(_ResStr,NrDischarged,NrTot,Discharged)
    ).

:- use_module(probsrc(bmachine), [b_get_properties_from_machine/1, b_get_machine_all_constants/1,
                         b_get_machine_variables/1, b_get_invariant_from_machine/1,
                         b_get_static_assertions_from_machine/1, b_get_dynamic_assertions_from_machine/1,
                         b_get_machine_operation/6, b_get_initialisation_from_machine/2]).

% get POs for loaded B machine
get_machine_wd_po(LOpt,PO,PO_Name) :-
    reset_stats,
    list_to_ord_set(LOpt,Options),
    b_get_properties_from_machine(Properties),
    % TO DO: move Conjuncts without WD condition to front in first pass
    empty_hyps(E),
    b_get_machine_all_constants(Csts),
    add_new_hyp_variables(E,Csts,OriginalHyps),
    % TO DO: move Conjuncts without WD condition to front in first pass
    ( PO_Name = 'AXM',
      debug_format(19,'Checking WD of PROPERTIES~n',[]),
      ord_add_element(Options,po_label(PO_Name),Opt2),
      compute_wd_with_reordering(Properties,OriginalHyps,Opt2,PO)

       %debug:time(well_def_analyser:transform_wd(Properties,OriginalHyps,Opt2,NewProperties)),nl,fail)

     ; % TO DO: for transform: only use assertions if assertion checking is on!?

       push_hyp(OriginalHyps,Properties,Options,Hyps1),
       b_get_static_assertions_from_machine(StaticAssertions),

       ( PO_Name = 'THM-AXM',
         ord_add_element(Options,po_label(PO_Name),Opt2),
         conjunct_predicates(StaticAssertions,SAssPred),
         compute_wd(SAssPred,Hyps1,Opt2,PO)
         ;
         push_hyps(Hyps1,StaticAssertions,Options,Hyps2),
         %portray_hyps(Hyps2),nl,
         b_get_machine_variables(Vars),
         add_new_hyp_variables(Hyps2,Vars,Hyps3),
         b_get_invariant_from_machine(Invariant),
         (  PO_Name = 'INV',
            debug_format(19,'Checking WD of INVARIANT~n',[]),
            ord_add_element(Options,po_label(PO_Name),Opt2),
            compute_wd_with_reordering(Invariant,Hyps3,Opt2,PO)
          ;
            push_hyp(Hyps3,Invariant,Options,Hyps4),
            b_get_dynamic_assertions_from_machine(DynAssertions),
            ( PO_Name = 'THM-INV',
              debug_format(19,'Checking WD of ASSERTIONS~n',[]),
              ord_add_element(Options,po_label(PO_Name),Opt2),
              conjunct_predicates(DynAssertions,AssPred),
              compute_wd(AssPred,Hyps4,Opt2,PO)
              % TO DO: use assertions for operations
            ; push_hyps(Hyps4,DynAssertions,Options,Hyps5),
              (
                b_get_initialisation_from_machine(Body,_OType),
                PO_Name = 'INIT',
                debug_format(19,'Checking WD of INITIALISATION~n',[]),
                ord_add_element(Options,po_label(PO_Name),Opt2),
                compute_wd(Body,Hyps5,Opt2,PO)
              ; b_get_machine_operation(OpName,OpResults,OpFormalParameters,Body,_OType,_TopLevel),
                PO_Name = OpName,
                debug_format(19,'Checking WD of OPERATION ~w~n',[OpName]),
                ord_add_element(Options,po_label(PO_Name),Opt2),
                add_new_hyp_variables(Hyps5,OpFormalParameters,Hyps6),
                add_new_hyp_variables(Hyps6,OpResults,Hyps7),
                compute_wd(Body,Hyps7,Opt2,PO)
              )
           )
         )
       )
     ),
     inc_counter(PO_Name).


% --------------------

:- use_module(probsrc(tools_strings),[ajoin/2,ajoin_with_sep/3]).
% get label of current PO from Options
get_current_po_label(POLabel,Options) :-
   findall(Label,get_label(Label,Options),Labels),
   Labels \= [],
   ajoin_with_sep(Labels,'/',POLabel).
get_label(Label,Options) :- member(po_label(L),Options),
   (L = [_|_] -> member(Label,L) ; Label=L).

% --------------------

:- use_module(well_def_prover,[prove_po/3]).
:- use_module(probsrc(tools_lists),[ord_member_nonvar_chk/2]).

create_po(H,Target,Infos,Options,PO) :- create_po(H,Target,Infos,Options,PO,_).

create_po(H,Target,Infos,Options,PO,D) :- (var(H) ; H \= hyp_rec(_,_) ; var(Options)),!,
    add_internal_error('Illegal hypotheses or options: ', create_po(H,Target,Infos,Options,PO,D)),fail.
create_po(Hyps,Target,Infos,Options,PO,Discharged) :-
    get_normalized_and_renamed_predicate(Target,Hyps,RenTarget,NT),
    (get_current_po_label(POLabel,Options) -> true ; POLabel = ''),
    (ord_member(print_goal,Options) -> translate:print_bexpr(RenTarget),nl ; true),
    (ord_member(discharge_po,Options)
     -> (%nl,add_message(well_def_analyser,'Proving ',Target,Infos),
         prove_proof_obligation(POLabel,Target,NT,Hyps,ProofTree)
          -> debug_format(19,'PO ~w discharged: ~w~n',[POLabel,ProofTree]), %debug:nl_time,
             inc_counter(discharged_po),
             Discharged=true
          ;  inc_counter(not_discharged_po),
             Discharged=false,
             (ord_member_nonvar_chk(create_not_discharged_msg(MsgType),Options)
               ->  ajoin(['WD-PO [',POLabel,'] not discharged: '],Msg),
                   (MsgType=message -> add_message(well_def_analyser,Msg,Target,Infos)
                    ; MsgType=warning -> add_warning(well_def_analyser,Msg,Target,Infos)
                    ; add_error(well_def_analyser,Msg,Target,Infos)
                   ),
                   (debug_mode(off) -> true ; print(NT),nl)
               ; true
             )
          )
      ; Discharged=false % no discharging of POs
    ),
    extract_pos_infos(Infos,PosInfos),
    (ord_member(create_full_po,Options)
      -> Hyps = hyp_rec(AvlNH,_),
         avl_range(AvlNH,Hyp),
         (Discharged=true, ord_member(minimize_full_po,Options)
           -> minimize_hyps(Hyp,RenTarget,NecHyps)
            ; NecHyps=Hyp
          ),
         conjunct_predicates(NecHyps,HypC),
             % optionally minimize hyps
         create_texpr(implication(HypC,RenTarget),pred,[discharged(Discharged,POLabel)|PosInfos],PO),
          %nl,translate:print_bexpr(PO),nl, prove_po_with_atelierb(POLabel,[ml],PO,_),
          % terms:term_hash(PO,Hash), print(Hash),nl, prove_po_with_smt(POLabel,z3,PO,_),
          (ord_member(prove_with_ml,Options) -> prove_po_with_atelierb(POLabel,[ml],PO,_) ; true),
          (ord_member(prove_with_pp,Options) -> prove_po_with_atelierb(POLabel,[pp],PO,_) ; true)
      ;  RenTarget = b(POE,POT,POI),
         append([discharged(Discharged,POLabel)|PosInfos],POI,POInfos),
         PO = b(POE,POT,POInfos)
     ).

% compute unsat core of hypotheses
minimize_hyps(Hyps,Goal,NecHyps) :- minimize_hyps_aux(Hyps,Goal,NecHyps,NecHyps),nl.

% TO DO: maybe do binary search / divide and conquer
% we could also examine which hypotheses were retrieved during the proof
%    (but not easy; what if access inside a Prolog negation)
minimize_hyps_aux([],_,_,[]).
minimize_hyps_aux([H|T],Goal,NecessaryHyps,Tail) :-
     \+ (Tail=T,prove_sequent_cur_context(Goal,NecessaryHyps)), % we need H to prove Goal
     !,
     write('+'),
     Tail = [H|TT],
     minimize_hyps_aux(T,Goal,NecessaryHyps,TT).
minimize_hyps_aux([_|T],Goal,NecessaryHyps,Tail) :-
     write('-'),
     minimize_hyps_aux(T,Goal,NecessaryHyps,Tail).

:- use_module(probsrc(runtime_profiler),[register_profiler_runtime/4]).
:- use_module(probsrc(atelierb_provers_interface),[prove_sequent_with_provers/3]).
prove_po_with_atelierb(POLabel,Provers,PO,ProofRes) :- runtime_profiler:print_runtime_profile,
      format('Proving PO ~w with ~w~n',[POLabel,Provers]),
      statistics(walltime,[T1,_]),
      on_exception(user_interrupt_signal,
              prove_sequent_with_provers(Provers,PO,ProofRes), % we cannot use time_out as PP/ML/KRT are a separate process
              ProofRes = user_interrupt_signal),
      statistics(walltime,[T2,_]), WT is T2-T1,
      !,
      register_profiler_runtime(atelierb(ProofRes),POLabel,WT,WT),
      format('Result of ~w on PO = ~w (walltime ~w ms)~n',[Provers,ProofRes,WT]),
      (ProofRes=user_interrupt_signal
       -> write(' Continue? [y/n] => '), flush_output, read(Cont),
          (Cont='y' -> true ; Cont='yes' -> true ; throw(user_interrupt_signal))
        ; true).

:- use_module(probsrc(bsyntaxtree),[create_negation/2]).
:- use_module(smt_solvers_interface(smt_solvers_interface),[smt_solve_predicate/4]).
prove_po_with_smt(POLabel,Solver,PO,ProofRes) :- runtime_profiler:print_runtime_profile,
      format('Proving PO ~w with ~w~n',[POLabel,Solver]),
      statistics(walltime,[T1,_]),
      create_negation(PO,NegPO),
      smt_solve_predicate(Solver,NegPO,_State,ProofRes),!,
      statistics(walltime,[T2,_]), WT is T2-T1,
      register_profiler_runtime(smt(Solver,ProofRes),POLabel,WT,WT),
      format('Result of ~w on PO = ~w (walltime ~w ms)~n',[Solver,ProofRes,WT]).

% a wrapper to prove_po which also checks for obvious POs using type info:
prove_proof_obligation(POLabel,Target,_NT,Hyps,ProofTree) :-
     obvious_po(Target,Hyps),!,
     debug_format(19,'Obvious PO (due to typing info): ~w~n',[POLabel]),
     register_profiler_runtime(wd_prob_prover(obvious_po),POLabel,0,0),
     ProofTree = obvious_po.
prove_proof_obligation(POLabel,_Target,NT,Hyps,ProofTree) :-
    statistics(runtime,[RT1,_]),
    statistics(walltime,[T1,_]),
    (prove_po(NT,Hyps,ProofTree) -> Res=discharged
    ; Res=unknown,ProofTree=''  % ,format('~n *********** WD PO UNKNOWN ~w~n~n',[POLabel])
    ),
    statistics(runtime,[RT2,_]), RT is RT2-RT1,
    statistics(walltime,[T2,_]), WT is T2-T1,
    register_profiler_runtime(wd_prob_prover(Res),POLabel,RT,WT),
    (WT>10 -> format('Proof of ~w took ~w ms, result=~w (~w)~n',[POLabel,WT,Res,ProofTree])
      %, translate:print_bexpr(_Target),nl, trace, assert(well_def_prover:dt),(prove_po(NT,Hyps,_) -> true ; true), trace
     ; true),
    Res=discharged.

%:- use_module(well_def_hyps,[is_finite_type_for_wd/2]).
obvious_po(b(finite(Set),pred,_),Hyps) :-
    get_texpr_type(Set,Type), % check this proof rule here, as prove_po has no longer types in the AST
    is_finite_type_for_wd(Type,Hyps).

% ------------------------------

:- dynamic counter/2.
reset_stats :- retractall(counter(_,_)),
   assert(counter(discharged_po,0)),
   assert(counter(not_discharged_po,0)).

inc_counter(C) :- (retract(counter(C,Nr)) -> N1 is Nr+1 ; N1 = 1),
  assert(counter(C,N1)).

print_stats :- findall(C/N,counter(C,N),L), format('WD Stats: ~w~n',[L]).

%tcltk_get_stats(list([list(['Category','Nr'])|Stats])) :-
%   findall(list([C,N]),counter(C,N),Stats).

get_discharged_result(ResStr,D,Tot,AllDischarged) :- counter(not_discharged_po,ND),
   counter(discharged_po,D), Tot is ND+D,
   ajoin([D,'/',Tot,' DISCHARGED'],ResStr),
   (ND=0 -> AllDischarged='TRUE' ; AllDischarged='FALSE').

% ------------------------------

annotate_wd(TExpr,AnnotatedTExpr) :-
   reset_stats,
   empty_hyps(EH),
   Options = [discharge_po],
   transform_wd(TExpr,EH,Options,AnnotatedTExpr),
   print_bexpr_with_wd_info(AnnotatedTExpr).


:- use_module(probsrc(preferences),[temporary_set_preference/3, reset_temporary_preference/2, get_preference/2]).
print_bexpr_with_wd_info(TExpr) :-
   temporary_set_preference(pp_wd_infos,true,Chng),
   call_cleanup(translate:nested_print_bexpr(TExpr),reset_temporary_preference(pp_wd_infos,Chng)),
   nl.

% transform an AST by adding annotations about discharged POs
transform_wd(b(Expr,Type,Info),Hypotheses,Options,b(NewExpr,Type,NewInfo)) :-
    member(contains_wd_condition,Info),
    !,
    syntaxtransformation(Expr,Subs,TNames,NSubs,NewExpr),
    add_new_hyp_variables(Hypotheses,TNames,Hyps2),
    (wd_transparent(Expr)
      -> l_transform_wd(Subs,Hyps2,Options,NSubs), adapt_info(Info,NSubs,NewInfo)
     ; wd_like_conjunction(Expr,Ids,A,B)
      -> wd_like_conjunction(NewExpr,Ids,TA,TB),
         l_transform_conj_wd([A,B],Hyps2,Options,[TA,TB]),
         check_top_level_wd(conj,Expr,Type,Info,Hyps2,Options, [TA,TB], NewInfo)
     ; transform_special_expr(Expr,Type,Info,Hyps2,Options,NewExpr,NewInfo) -> true
     ; l_transform_wd(Subs,Hyps2,Options,NSubs)
       ->
       check_top_level_wd(other,Expr,Type,Info,Hyps2,Options, NSubs, NewInfo)
    ).
transform_wd(TExpr,_,_,TExpr). % copy entire expression if no contains_wd_condition info

check_top_level_wd(conj,Expr,Type,Info,Hyps,Options, _NSubs, NewInfo) :-
   create_top_level_wd_po_conj_like(Expr,Type,Info,Hyps,Options,_PO,false),!, % PO not proven; keep Info as is
   %write(-), flush_output, %print(not_discharged(PO)),nl,
   NewInfo=Info.
check_top_level_wd(other,Expr,Type,Info,Hyps,Options, _NSubs, NewInfo) :-
   create_top_level_wd_po(Expr,Type,Info,Hyps,Options,_PO,false),!, % PO not proven; keep Info as is
   %write(-), flush_output, %print(not_discharged(PO)),nl,
   NewInfo=Info.
check_top_level_wd(_,_Expr,_Type,Info,_Hyps,_Options, NSubs, NewInfo) :- % no top-level PO or PO proven
   adapt_info(Info,NSubs,NewInfo).

:- use_module(library(lists),[delete/3]).
% adapt WD Info depending on result of analysis
adapt_info(Info,Subs,NewInfo) :- subs_contain_wd_condition(Subs),!,NewInfo=Info.
adapt_info(Info,_,[discharged_wd_po|NewInfo]) :- delete(Info,contains_wd_condition,NewInfo).

subs_contain_wd_condition(Subs) :-
   member(b(_,_,Infos),Subs),
   memberchk(contains_wd_condition,Infos).

% transform independent sub-arguments
l_transform_wd([],_Hyps,_Options,[]).
l_transform_wd([Sub|T],Hyps,Options,[NSub|NT]) :-
   transform_wd(Sub,Hyps,Options,NSub),
   l_transform_wd(T,Hyps,Options,NT).

% transform conjunction like sub-arguments, with left-to-right adding of hypotheses
l_transform_conj_wd([],_Hyps,_Options,[]).
l_transform_conj_wd([Sub|T],Hyps,Options,[NSub|NT]) :-
   transform_wd(Sub,Hyps,Options,NSub),
   push_hyp(Hyps,Sub,Options,Hyps2),
   l_transform_conj_wd(T,Hyps2,Options,NT).

% transform special cases where sub-arguments are not independent nor conjunction-like
% TO DO: assign
transform_special_expr(disjunct(A,B),_,Info,Hyps,Options,disjunct(TA,TB),NewInfo) :- NewInfo=Info,
   transform_wd(A,Hyps,Options,TA),
   negate_hyp(A,NA),
   push_hyp(Hyps,NA,Options,Hyps2),
   transform_wd(B,Hyps2,Options,TB).
transform_special_expr(if_then_else(A,B,C),_,Info,Hyps,Options,if_then_else(TA,TB,TC),NewInfo) :- NewInfo=Info,
   transform_wd(A,Hyps,Options,TA),
   push_hyp(Hyps,A,Options,Hyps2),
   transform_wd(B,Hyps2,Options,TB),
   negate_hyp(A,NA),
   push_hyp(Hyps,NA,Options,Hyps3),
   transform_wd(C,Hyps3,Options,TC).
transform_special_expr(recursive_let(ID,B),_,Info,Hyps,Options,recursive_let(ID,TB),NewInfo) :- NewInfo=Info,
   transform_wd(B,Hyps,Options,TB).
transform_special_expr(general_sum(Ids,P,E),integer,Info,Hyps,Options,general_sum(Ids,TP,TE),NewInfo) :-
   NewInfo=Info, % Note: we do not try to discharge finite PO
   transform_wd(P,Hyps,Options,TP),
   push_hyp(Hyps,P,Options,Hyps2),
   transform_wd(E,Hyps2,Options,TE).
transform_special_expr(general_product(Ids,P,E),integer,Info,Hyps,Options,general_product(Ids,TP,TE),NewInfo) :-
   NewInfo=Info, % Note: we do not try to discharge finite PO
   transform_wd(P,Hyps,Options,TP),
   push_hyp(Hyps,P,Options,Hyps2),
   transform_wd(E,Hyps2,Options,TE).
transform_special_expr(if(List),_,Info,Hyps,Options,if(TList),NewInfo) :- NewInfo=Info,
     list_transform_if_elsif_wd(List,Hyps,Options,TList).

list_transform_if_elsif_wd([],_,_,[]).
list_transform_if_elsif_wd([b(if_elsif(A,Body),Type,I) | T], Hyps,Options,Res) :-
     is_truth(A),!,
     Res=[b(if_elsif(A,TBody),Type,I) | T],
     transform_wd(Body,Hyps,Options,TBody).
list_transform_if_elsif_wd([b(if_elsif(A,Body),Type,I) | T], Hyps,Options,
                           [b(if_elsif(TA,TBody),Type,I) | TT]) :-
     transform_wd(A,Hyps,Options,TA),
     push_hyp(Hyps,A,Options,Hyps2),
     transform_wd(Body,Hyps2,Options,TBody),
     negate_hyp(A,NA),
     push_hyp(Hyps,NA,Options,Hyps3),
     list_transform_if_elsif_wd(T,Hyps3,Options,TT).

% ------------------------------

:- use_module(probsrc(bsyntaxtree),[conjunction_to_list/2]).
:- use_module(probsrc(tools),[split_list/4]).

% process those conjuncts with no WD condition first; this means we have more hypotheses for proving and is safe for ProB
compute_wd_with_reordering(TExpr,Hypotheses,Options,PO) :- \+ ord_member(reorder_conjuncts,Options),!,
   compute_wd(TExpr,Hypotheses,Options,PO).
compute_wd_with_reordering(TExpr,Hypotheses,Options,PO) :-
   conjunction_to_list(TExpr,Conjuncts),
   split_list(is_well_defined_conjunct,Conjuncts,Conj1,Conj2),
   Conj2 \= [], % otherwise no WD condition anyway
   (ord_member(ignore_wd_infos,Options)
    ->  append(Conj1,Conj2,Conj12),
        list_compute_wd(Conj12,Options,Hypotheses,PO)
     ;  push_hyps(Hypotheses,Conj1,Options,Hyps2), % push the Hypotheses without WD checking
        list_compute_wd(Conj2,Options,Hyps2,PO)
    ).

%is_well_defined_conjunct(TE) :- print('Checking conjunct: '), translate:print_bexpr(TE),nl,fail.
is_well_defined_conjunct(b(_,_,Infos)) :- nonmember(contains_wd_condition,Infos),!.
%is_well_defined_conjunct(TE) :- print('Not well defined: '), translate:print_bexpr(TE),nl,fail.

compute_wd(TExpr,Hypotheses,Options,PO) :-
   syntaxtraversion(TExpr,Expr,Type,Infos,Subs,TNames),
   (ord_member(ignore_wd_infos,Options) -> true ; member(contains_wd_condition,Infos)),
   %(print(Expr),nl ; print(backtrack(Expr)),nl,fail),
   (get_info_labels(Infos,Label) -> ord_add_element(Options,po_label(Label),Opt2) ; Opt2=Options),
   compute_wd_aux(Expr,Type,Infos,Subs,TNames,Hypotheses,Opt2,PO),
   true. %check_if_po_expected(PO,TExpr,Infos).

check_if_po_expected(PO,TExpr,Infos) :-
   ( member(contains_wd_condition,Infos) -> true
     ; get_texpr_info(PO,POInfo), POInfo=[discharged(true,_)|_] -> true
     % Note: Sigma currently gets no contains_wd_condition, but has finiteness WD constraint
     ; add_error(well_def_analyser,'Unexpected WD-PO for expression marked to have no WD condition:',PO,TExpr)
   ).

:- use_module(library(lists),[maplist/3]).
% treat assignment substitution
% see b_assign_values_or_functions in b_interpreter
compute_wd_assign(Hyps,Options,PO,LHS,RHS) :- LHS=b(function(F,X),_,_),
     !,
     % special case for f(X) := RHS
    (  compute_wd(F,Hyps,Options,PO) % ?
     ; compute_wd(X,Hyps,Options,PO)
     ; compute_wd(RHS,Hyps,Options,PO) ).
compute_wd_assign(Hyps,Options,PO,_,RHS) :- % LHS must be an identifier
   compute_wd(RHS,Hyps,Options,PO).

compute_wd_aux(assign(LHSList,RHSList),_,_,_,[],Hypotheses,Options,PO) :-
    !,
    maplist(compute_wd_assign(Hypotheses,Options,PO),LHSList,RHSList).
compute_wd_aux(Expr,_,_,Subs,TNames,Hypotheses,Options,PO) :-
    wd_transparent(Expr),!,
    % cut: only look at WD condition of all sub-arguments
    add_new_hyp_variables(Hypotheses,TNames,Hyps2),
    member(Sub,Subs),
    compute_wd(Sub,Hyps2,Options,PO).
% Now the rules which do not look at all sub-arguments with original Hypotheses:
compute_wd_aux(conjunct(A,B),_,_,_,_,Hypotheses,Options,PO) :- !,
    flatten_conjunct(A,B,List), % avoid re-pushing hyps if A is nested conjunct
    compute_wd_conjuncts(List,Hypotheses,Options,PO).
compute_wd_aux(sequence(List),subst,_,_,_,Hypotheses,Options,PO) :- % sequential composition
    !,
    sequence_compute_wd(List,Options,Hypotheses,PO).
compute_wd_aux(while(COND,STMT,INV,VARIANT),_,Infos,_,_,Hypotheses,Options,PO) :- !,
    (member(modifies(VarsModified),Infos) -> debug_println(9,computed_wd_while(VarsModified))
      ; add_error(well_def_analyser,'While does not contain modifies info:',Infos,Infos),
        VarsModified=[]
    ),
    copy_hyp_variables(Hypotheses,VarsModified,Hyps2), % we can no longer make any assumption about the modified vars
    % this corresponds to a universal quantification
    ( compute_wd(INV,Hyps2,Options,PO) % Invariant must hold before and after WHILE loop
      ;
      push_hyp(Hyps2,INV,Options,Hyps3), % add INV to the hypotheses now for the other parts of the WHILE
      (
        compute_wd(COND,Hyps3,Options,PO)
        ;
        compute_wd(VARIANT,Hyps3,Options,PO) % NAT proof rule for While must also hold if P is false according to Atelier-B handbook (otherwise we could have pushed COND onto the stack)
         ;
          push_hyp(Hyps3,COND,Options,Hyps4),
          compute_wd(STMT,Hyps4,Options,PO) % Statement is only executed if COND is true; INV is checked
       )
      ).
compute_wd_aux(Expr,Type,Infos,Subs,TNames,Hypotheses,Options,PO) :-
    wd_like_conjunction(Expr,_,A,B),!,
    add_new_hyp_variables(Hypotheses,TNames,Hyps2),
    ( create_top_level_wd_po_conj_like(Expr,Type,Infos,Hyps2,Options,PO,_Discharged) % e.g., for assertion_expression
     ;
     compute_wd_aux(conjunct(A,B),Type,Infos,Subs,TNames,Hyps2,Options,PO)
    ).
compute_wd_aux(if_then_else(A,B,C),_,_,_,_,Hypotheses,Options,PO) :- !,
    (compute_wd(A,Hypotheses,Options,PO)
     ;
     push_hyp(Hypotheses,A,Options,NewHyp),
     compute_wd(B,NewHyp,Options,PO)
     ;
     negate_hyp(A,NA),
     push_hyp(Hypotheses,NA,Options,NewHyp),
     compute_wd(C,NewHyp,Options,PO)).
compute_wd_aux(if(List),_,_,_,_,Hypotheses,Options,PO) :- !,
     list_compute_if_elsif_wd(List,Hypotheses,Options,PO).
compute_wd_aux(disjunct(A,B),_,_,_,_,Hypotheses,Options,PO) :- !,
    (compute_wd(A,Hypotheses,Options,PO)
     ;
     negate_hyp(A,NA),
     push_hyp(Hypotheses,NA,Options,NewHyp),
     compute_wd(B,NewHyp,Options,PO)).
compute_wd_aux(general_sum(Ids,P,E),integer,Infos,_,_,Hypotheses,Options,PO) :- !, % SIGMA
    ( % first the sub arguments
      add_new_hyp_variables(Hypotheses,Ids,Hyps2),
      ( compute_wd(P,Hyps2,Options,PO)
       ;
        push_hyp(Hyps2,P,Options,NewHyp),
        compute_wd(E,NewHyp,Options,PO)
      )
    ; % now the PO of general_sum itself
      create_texpr(finite(TS),pred,[],Target),
      create_comprehension_set(Ids,P,[],TS),
      create_po(Hypotheses,Target,Infos,Options,PO)
    ).
compute_wd_aux(general_product(Ids,P,E),integer,Infos,Subs,TNames,Hypotheses,Options,PO) :- !, % PI
    compute_wd_aux(general_sum(Ids,P,E),integer,Infos,Subs,TNames,Hypotheses,Options,PO).
compute_wd_aux(rlevent(_Name,_Sect,_Status,_Params,Guard,Theorems,Actions,VWit,PWit,_Unmod,_AbstractEvents),
                _,_Infos,_Subs,TNames,Hypotheses,Options,PO) :- !,
      add_new_hyp_variables(Hypotheses,TNames,Hyps2),
      % TO DO: treat nested lists VWit,...
      Act = b(parallel(Actions),subst,[]),
      append([[Guard],Theorems,VWit,PWit,[Act]],List),
      list_compute_wd(List,Options,Hyps2,PO).
compute_wd_aux(recursive_let(_ID,Body),_,_,_,_,Hypotheses,Options,PO) :- !,
      % the ID does not have to be added to hypothesis stack !
      compute_wd(Body,Hypotheses,Options,PO).
% ----------------
% Now the rules where we also look at the sub-arguments with original Hypotheses:
compute_wd_aux(_,_,_,Subs,TNames,Hypotheses,Options,PO) :-
    add_new_hyp_variables(Hypotheses,TNames,Hyp2),
    % look at WD condition of all sub-arguments independently
    member(Sub,Subs),
    compute_wd(Sub,Hyp2,Options,PO).
% Now the specific rules:
compute_wd_aux(Expr,Type,Infos,_,_,Hypotheses,Options,PO) :-
    create_top_level_wd_po(Expr,Type,Infos,Hypotheses,Options,PO,_Discharged).

% treat a list of conjuncts (after flattening)
%compute_wd_conjuncts([],_,_,_) :- fail.
compute_wd_conjuncts([A|T],Hypotheses,Options,PO) :-
    (compute_wd(A,Hypotheses,Options,PO)
     ;
     T \= [],
     push_hyp(Hypotheses,A,Options,NewHyp),
     compute_wd_conjuncts(T,NewHyp,Options,PO)).

% flatten conjuncts to avoid re-pushing the same hypotheses
flatten_conjunct(ConjA,ConjB,List) :-
    flatten_conjunct_dl(ConjA,List,[ConjB]). % Tail not flattened; not necessary for performance
flatten_conjunct_dl(b(conjunct(A,B),_,_)) --> !, flatten_conjunct_dl(A), flatten_conjunct_dl(B).
flatten_conjunct_dl(X) --> [X].


% create top-level WD conditions for conjunction-like operators; not for sub-arguments
create_top_level_wd_po_conj_like(assertion_expression(A,_Msg,_),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
    Target = A,
    ord_add_element(Options,po_label('ASSERT'),Options2),
    create_po(Hypotheses,Target,Infos,Options2,PO,Discharged).

% create top-level WD conditions; not for sub-arguments
create_top_level_wd_po(function(A,B),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
    (create_texpr(member(B,DomA),pred,[],Target),
     get_texpr_type(B,TB),
     create_texpr(domain(A),set(TB),[],DomA),
     create_po(Hypotheses,Target,Infos,Options,PO,Discharged)
     %, tools_printing:print_term_summary(created_function_po(Target,PO))
    ;
     create_partial_fun_target(A,Target),
     create_po(Hypotheses,Target,Infos,Options,PO,Discharged)
    ).
create_top_level_wd_po(div(_,B),integer,Infos,Hypotheses,Options,PO,Discharged) :- !,
    create_texpr(not_equal(B,Zero),pred,[],Target),
    create_texpr(integer(0),integer,[],Zero),
    create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(div_real(_,B),integer,Infos,Hypotheses,Options,PO,Discharged) :- !,
    create_texpr(not_equal(B,Zero),pred,[],Target),
    create_texpr(real('0.0'),real,[],Zero),
    create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(floored_div(_,B),integer,Infos,Hypotheses,Options,PO,Discharged) :- !,
    create_texpr(not_equal(B,Zero),pred,[],Target),
    create_texpr(integer(0),integer,[],Zero),
    create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(modulo(A,B),integer,Infos,Hypotheses,Options,PO,Discharged) :- !,
    create_texpr(integer(0),integer,[],Zero),
    (  create_texpr(greater(B,Zero),pred,[],Target),  % does TLA require this to be positive??
       create_po(Hypotheses,Target,Infos,Options,PO)
     ; \+ z_or_tla_minor_mode,
       create_texpr(greater_equal(A,Zero),pred,[],Target),
       create_po(Hypotheses,Target,Infos,Options,PO,Discharged)
    ).
create_top_level_wd_po(power_of(A,B),integer,Infos,Hypotheses,Options,PO,Discharged) :- !,
    create_texpr(integer(0),integer,[],Zero),
    create_texpr(greater_equal(AB,Zero),pred,[],Target),
    (  AB=B, % B >= 0 must hold in B and Event-B
       create_po(Hypotheses,Target,Infos,Options,PO,Discharged)
     ; eventb_mode, % TO DO: what happens in Z and TLA+?
       AB=A, % A >= 0 must hold in Event-B
       create_po(Hypotheses,Target,Infos,Options,PO,Discharged)
    ).
create_top_level_wd_po(card(S),integer,Infos,Hypotheses,Options,PO,Discharged) :- !,
    create_texpr(finite(S),pred,[],Target),
    create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(max(S),integer,Infos,Hypotheses,Options,PO,Discharged) :- !,
    create_texpr(finite(S),pred,[],Target1),
    get_texpr_type(S,TS), create_texpr(not_equal(S,b(empty_set,TS,[])),pred,[],Target2),
    Target = b(conjunct(Target1,Target2),pred,[]), % this is a sufficient condition, not a necessary one
    % create a single PO, because of use in check_top_level_wd
    create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(min(S),integer,Infos,Hypotheses,Options,PO,Discharged) :- !,
    create_texpr(finite(S),pred,[],Target1),
    get_texpr_type(S,TS), create_texpr(not_equal(S,b(empty_set,TS,[])),pred,[],Target2),
    Target = b(conjunct(Target1,Target2),pred,[]), % this is a sufficient condition, not a necessary one
    create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(general_intersection(A),TA,Infos,Hypotheses,Options,PO,Discharged) :- !, % inter
    create_texpr(not_equal(A,b(empty_set,set(TA),[])),pred,[],Target),
    create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
% Sequence operators:
create_top_level_wd_po(size(S),integer,Infos,Hypotheses,Options,PO,Discharged) :- !,
    % we do not have to create a finite PO: sequences are by construction finite
     create_sequence_target(S,Target),
     create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(insert_tail(S,_),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     create_sequence_target(S,Target), create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(insert_front(_,S),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     % page 174 of Abrial's book just requires that S is a partial function NATURAL1 +-> TYPE
     create_sequence_target(S,Target), create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(rev(S),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     create_sequence_target(S,Target), create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(concat(S1,S2),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     (create_sequence_target(S1,Target) ; create_sequence_target(S2,Target)),
     create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(general_concat(S),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     create_seq_seq_target(S,Target), create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(first(S),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     create_sequence1_target(S,Target), create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(last(S),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     create_sequence1_target(S,Target), create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(front(S),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     create_sequence1_target(S,Target), create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(tail(S),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     create_sequence1_target(S,Target), create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(restrict_front(S,N),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     (  create_sequence_target(S,Target) ; create_seq_interval_target(S,N,Target) ),
     create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(restrict_tail(S,N),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     (  create_sequence_target(S,Target) ; create_seq_interval_target(S,N,Target) ),
     create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(mu(S),_,Infos,Hypotheses,Options,PO,Discharged) :- !, % Z MU Operator
    create_texpr(finite(S),pred,[],Target1),
    get_texpr_type(S,TS), create_texpr(not_equal(S,b(empty_set,TS,[])),pred,[],Target2),
    create_texpr(integer(1),integer,[],One),
    create_texpr(card(S),set(TS),[],CardS),
    create_texpr(less_equal(CardS,One),pred,[],Target3),
    conjunct_predicates([Target1,Target2,Target3],Target),
    create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
% External Functions:
create_top_level_wd_po(external_function_call('MU',[S]),T,Infos,Hypotheses,Options,PO,Discharged) :- !,
    create_top_level_wd_po(mu(S),T,Infos,Hypotheses,Options,PO,Discharged).
create_top_level_wd_po(external_function_call('CHOOSE',[A]),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
    get_texpr_type(A,TA),
    create_texpr(not_equal(A,b(empty_set,set(TA),[])),pred,[],Target),
    create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(external_function_call('SQUASH',[A]),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
     create_partial_fun_target(A,Target),
     create_po(Hypotheses,Target,Infos,Options,PO,Discharged).
create_top_level_wd_po(external_function_call(C,_),_,Infos,Hypotheses,Options,PO,Discharged) :- !,
    % external_fun_has_wd_condition is true
    get_preference(wd_ignore_external_funs,false),
    create_po(Hypotheses,b(unknown_truth_value(C),pred,[]),Infos,Options,PO,Discharged).
create_top_level_wd_po(external_subst_call(C,_),_,Infos,Hypotheses,Options,PO,Discharged) :- !, % ditto
    get_preference(wd_ignore_external_funs,false),
    create_po(Hypotheses,b(unknown_truth_value(C),pred,[]),Infos,Options,PO,Discharged).
create_top_level_wd_po(external_pred_call(C,_),_,Infos,Hypotheses,Options,PO,Discharged) :- !, % ditto
    get_preference(wd_ignore_external_funs,false),
    create_po(Hypotheses,b(unknown_truth_value(C),pred,[]),Infos,Options,PO,Discharged).
% Unsupported:
create_top_level_wd_po(Expr,T,I,Hypotheses,Options,PO,Discharged) :-
    functor(Expr,F,N),
    ajoin(['Unsupported expression ',F,'/',N,': '],Msg),
    (ord_member_nonvar_chk(create_not_discharged_msg(_),Options)
     -> add_message(well_def_analyser,Msg,Expr,I) % we create warning below
     ; known_to_be_unsupported(F,N)
      -> add_warning(well_def_analyser,Msg,Expr,I)
      ; add_internal_error(Msg,create_top_level_wd_po(Expr,T,I,_,Options,PO,Discharged))
    ),
    inc_counter(unsupported),
    create_po(Hypotheses,b(unknown_truth_value(F),pred,[wd(Expr)]),I,Options,PO,Discharged).

known_to_be_unsupported(freetype_destructor,3). % has WD condition, see check_freetype_case in b_interpreter
known_to_be_unsupported(lazy_let_expr,3).
known_to_be_unsupported(lazy_let_pred,3).
known_to_be_unsupported(lazy_let_subst,3).

create_partial_fun_target(A,Target) :-
     % see create_function_check in bvisual2
     get_texpr_type(A,TA), is_set_type(TA,couple(TDom,TRan)),
     create_texpr(typeset,TDom,[],TDT), % we could also use create_maximal_type_set/2
     create_texpr(typeset,TRan,[],TRT),
     create_texpr(partial_function(TDT,TRT),set(TA),[],PF),
     create_texpr(member(A,PF),pred,[],Target).

:- use_module(probsrc(bsyntaxtree),[is_set_type/2]).
% get texpr for set of all sequences:
create_seq_type_expr(Seq,Sequences) :-
     get_texpr_type(Seq,ST),
     is_set_type(ST,couple(_,TRan)),
     create_texpr(typeset,set(TRan),[],TRT),
     create_texpr(seq(TRT),ST,[],Sequences).
create_seq1_type_expr(Seq,Sequences) :-
     get_texpr_type(Seq,ST),
     is_set_type(ST,couple(_,TRan)),
     create_texpr(typeset,set(TRan),[],TRT),
     create_texpr(seq1(TRT),ST,[],Sequences).
create_seq_seq_type_expr(SeqSeq,SequencesOfSequences) :-
     get_texpr_type(SeqSeq,SST),
     is_set_type(SST,couple(_,SeqT)),  % SeqT = type of the inner sequences
     is_set_type(SeqT,couple(_,TRan)), % TRan is type elements in the inner sequence
     create_texpr(typeset,set(TRan),[],TRT1),
     create_texpr(seq(TRT1),SeqT,[],Sequences),
     create_texpr(seq(Sequences),SST,[],SequencesOfSequences).

create_sequence_target(S,Target) :-
     create_seq_type_expr(S,Sequences),
     extract_pos_infos(S,PosInfos),
     create_texpr(member(S,Sequences),pred,PosInfos,Target).
create_sequence1_target(S,Target) :-
     create_seq1_type_expr(S,Sequences),
     extract_pos_infos(S,PosInfos),
     create_texpr(member(S,Sequences),pred,PosInfos,Target).
create_seq_seq_target(S,Target) :-
     create_seq_seq_type_expr(S,SeqSeq),
     extract_pos_infos(S,PosInfos),
     create_texpr(member(S,SeqSeq),pred,PosInfos,Target).

create_seq_interval_target(S,N,Target) :-
      create_texpr(integer(0),integer,[],Zero),
      create_texpr(size(S),integer,[],MaxIdx),
      create_texpr(interval(Zero,MaxIdx),set(integer),[],ValidIdxs),
      create_texpr(member(N,ValidIdxs),pred,[],Target).

% compute WD of a conjunction/sequence of formulas
list_compute_wd([A|_],Options,Hyps,PO) :-
    compute_wd(A,Hyps,Options,PO).
list_compute_wd([A|T],Options,Hyps,PO) :- T \= [],
    push_hyp(Hyps,A,Options,Hyps2),
    list_compute_wd(T,Options,Hyps2,PO).

% compute WD of a list of independent formulas
%list_indep_compute_wd([A|T],Options,Hyps,PO) :-
%    (compute_wd(A,Hyps,Options,PO)
%     ;
%     list_indep_compute_wd(T,Options,Hyps,PO)).

% compute WD of a list of IF-THEN-ELSE cases:
list_compute_if_elsif_wd([b(if_elsif(A,Body),_,_) | T], Hyps,Options,PO) :-
    ( is_truth(A) -> compute_wd(Body,Hyps,Options,PO)
     ; ( compute_wd(A,Hyps,Options,PO)
        ;
        push_hyp(Hyps,A,Options,NewHyp),
        compute_wd(Body,NewHyp,Options,PO)
        ;
        T \= [],
        negate_hyp(A,NA),
        push_hyp(Hyps,NA,Options,NewHyp),
        list_compute_if_elsif_wd(T, NewHyp,Options,PO)
      )
    ).

% compute WD of a conjunction/sequence of formulas
sequence_compute_wd([A|_],Options,Hyps,PO) :-
    compute_wd(A,Hyps,Options,PO).
sequence_compute_wd([A|T],Options,Hyps,PO) :- T \= [],
    update_hyp_for_sequence_lhs(Hyps,A,Options,Hyps2),
    sequence_compute_wd(T,Options,Hyps2,PO).

:- use_module(probsrc(b_read_write_info), [get_accessed_vars/4]).
% update the hypotheses depending on effect of LHS of a sequential composition:
update_hyp_for_sequence_lhs(Hyps,b(LHS,subst,_),Options,Hyps2) :-
   process_seq_lhs_aux(LHS,Hyps,Options,Hyps2),
   !.
update_hyp_for_sequence_lhs(Hyps,TStmt,_Options,Hyps2) :-
   add_modified_vars_to_hyp(Hyps,TStmt,Hyps2).
   %add_message(well_def_analyser,'Uncovered substitution in sequential composition: ',TStmt,TStmt).

% add all modified vars as new hyp vars: new hypotheses and goals will thus refer to new version (like priming variables)
add_modified_vars_to_hyp(Hyps,TLHS,Hyps2) :-
   (get_accessed_vars(TLHS,[],ModifiedVars,_Read) -> true
      ; add_error(well_def_analyser,'Could not obtain modified vars:',TLHS,TLHS), ModifiedVars=[]
   ),
   maplist(construct_typed_var(Hyps),ModifiedVars,TModVars),
   add_new_hyp_variables(Hyps,TModVars,Hyps2).

construct_typed_var(Hyps,ID,b(identifier(ID),Type,[])) :-
   (get_hyp_var_type(ID,Hyps,Type) -> true
    ; add_internal_error('Unknown identifier in hyps:',ID), Type=any).

rename_expr(Hyps,Expr,RenExpr) :- get_renamed_expression(Expr,Hyps,RenExpr).
:- use_module(probsrc(bsyntaxtree),[get_texpr_id/2]).
construct_equal(TID,RenExpr, b(equal(TID,RenExpr),pred,[])) :- get_texpr_id(TID,_),!.
construct_equal(_,_,b(truth,pred,[])). % TO DO: deal with more complex assignments f(i) := E  -> f' = f <+ {i|->E}

process_seq_lhs_aux(assign_single_id(TID,Expr),Hyps,Options,Hyps2) :- !,
    get_renamed_expression(Expr,Hyps,RenExpr), % rename in old context; e.g., if we have x:=x+1
    add_new_hyp_variables(Hyps,[TID],Hyps1),
    get_renamed_expression(TID,Hyps1,RenTID),
    Eq = b(equal(RenTID,RenExpr),pred,[]),  % print('Add: '), translate:print_bexpr(Eq),nl,
    push_hyps_wo_renaming(Hyps1,[Eq],Options,Hyps2). %, portray_hyps(Hyps2).
process_seq_lhs_aux(assign(FTIDs,Exprs),Hyps,Options,Hyps2) :- !,
    maplist(rename_expr(Hyps),Exprs,RenExprs),% rename in old context; e.g., if we have x:=x+1
    maplist(get_assigned_ids,FTIDs,TIDs),
    add_new_hyp_variables(Hyps,TIDs,Hyps1),
    maplist(rename_expr(Hyps1),TIDs,RenTIDs),
    maplist(well_def_analyser:construct_equal,RenTIDs,RenExprs,Equalities),
    push_hyps_wo_renaming(Hyps1,Equalities,Options,Hyps2).

% get assigned ids for things like f(x) := E  or r'f := E
get_assigned_ids(b(Expr,_,_),TID) :- get_assigned_id(Expr,F), !,get_assigned_ids(F,TID).
get_assigned_ids(TID,TID).

get_assigned_id(function(F,_),F).
get_assigned_id(record_field(R,_),R).

% wd_like_conjunction(Construct,NewTypedVars,Conjunct1,Conjunct2)
wd_like_conjunction(conjunct(A,B),[],A,B).
wd_like_conjunction(implication(A,B),[],A,B).
wd_like_conjunction(precondition(A,B),[],A,B).
wd_like_conjunction(assertion(A,B),[],A,B).
wd_like_conjunction(select_when(A,B),[],A,B).
wd_like_conjunction(assertion_expression(A,_Msg,B),[],A,B). % Note: this has associated WD condition
wd_like_conjunction(any(Ids,A,B),Ids,A,B).
wd_like_conjunction(forall(Ids,A,B),Ids,A,B).
wd_like_conjunction(quantified_union(Ids,A,B),Ids,A,B). % rewritten
wd_like_conjunction(let_predicate(Ids,E,P),Ids,A,P) :- construct_let_eq_pred(Ids,E,A).
wd_like_conjunction(let_expression(Ids,E,P),Ids,A,P) :- construct_let_eq_pred(Ids,E,A).
wd_like_conjunction(let(Ids,E,P),Ids,A,P) :- A=E. % this is already a predicate


:- use_module(library(lists),[maplist/4,append/2]).
construct_let_eq_pred(Ids,E,Pred) :-
   maplist(construct_eq,Ids,E,Ps),
   conjunct_predicates(Ps,Pred),!.
construct_let_eq_pred(Ids,E,Pred) :- add_internal_error('Call failed: ',construct_let_eq_pred(Ids,E,Pred)),fail.

construct_eq(TID,E,EqPred) :- safe_create_texpr(equal(TID,E),pred,EqPred).

% The constructs which have no WD condition of their own; they just require checking the arguments:
wd_transparent(truth).
wd_transparent(falsity).
wd_transparent(negation(_)).
wd_transparent(equivalence(_,_)).
wd_transparent(equal(_,_)).
wd_transparent(not_equal(_,_)).
wd_transparent(member(_,_)).
wd_transparent(not_member(_,_)).
wd_transparent(subset(_,_)).
wd_transparent(subset_strict(_,_)).
wd_transparent(not_subset(_,_)).
wd_transparent(not_subset_strict(_,_)).
wd_transparent(less_equal(_,_)).
wd_transparent(less(_,_)).
wd_transparent(less_equal_real(_,_)).
wd_transparent(less_real(_,_)).
wd_transparent(greater_equal(_,_)).
wd_transparent(greater(_,_)).
wd_transparent(finite(_)).
wd_transparent(partition(_,_)).
wd_transparent(value(_)).
wd_transparent(boolean_true).
wd_transparent(boolean_false).
wd_transparent(max_int).
wd_transparent(min_int).
wd_transparent(empty_set).
wd_transparent(bool_set).
wd_transparent(float_set).
wd_transparent(real_set).
wd_transparent(string_set).
wd_transparent(convert_bool(_)).
wd_transparent(add(_,_)).
wd_transparent(add_real(_,_)).
wd_transparent(minus(_,_)).
wd_transparent(minus_real(_,_)).
wd_transparent(minus_or_set_subtract(_,_)).
wd_transparent(unary_minus(_)).
wd_transparent(unary_minus_real(_)).
wd_transparent(multiplication(_,_)).
wd_transparent(multiplication_real(_,_)).
wd_transparent(mult_or_cart(_,_)).
wd_transparent(cartesian_product(_,_)).
wd_transparent(successor).
wd_transparent(predecessor).
wd_transparent(couple(_,_)).
wd_transparent(pow_subset(_)).
wd_transparent(pow1_subset(_)).
wd_transparent(fin_subset(_)).
wd_transparent(fin1_subset(_)).
wd_transparent(general_union(_)).
wd_transparent(interval(_,_)).
wd_transparent(union(_,_)).
wd_transparent(intersection(_,_)).
wd_transparent(set_subtraction(_,_)).
wd_transparent(relations(_,_)).
wd_transparent(identity(_)).
wd_transparent(event_b_identity).
wd_transparent(reverse(_)).
wd_transparent(first_projection(_,_)).
wd_transparent(first_of_pair(_)).
wd_transparent(event_b_first_projection(_)).
wd_transparent(event_b_first_projection_v2).
wd_transparent(second_projection(_,_)).
wd_transparent(event_b_second_projection(_)).
wd_transparent(event_b_second_projection_v2).
wd_transparent(second_of_pair(_)).
wd_transparent(composition(_,_)).
wd_transparent(ring(_,_)).
wd_transparent(direct_product(_,_)).
wd_transparent(parallel_product(_,_)).
wd_transparent(trans_function(_)).
wd_transparent(trans_relation(_)).
wd_transparent(iteration(_,_)).
wd_transparent(reflexive_closure(_)).
wd_transparent(closure(_)).
wd_transparent(domain(_)).
wd_transparent(range(_)).
wd_transparent(image(_,_)).
wd_transparent(domain_restriction(_,_)).
wd_transparent(domain_subtraction(_,_)).
wd_transparent(range_restriction(_,_)).
wd_transparent(range_subtraction(_,_)).
wd_transparent(overwrite(_,_)).
wd_transparent(partial_function(_,_)).
wd_transparent(total_function(_,_)).
wd_transparent(partial_injection(_,_)).
wd_transparent(total_injection(_,_)).
wd_transparent(partial_surjection(_,_)).
wd_transparent(total_surjection(_,_)).
wd_transparent(total_bijection(_,_)).
wd_transparent(partial_bijection(_,_)).
wd_transparent(total_relation(_,_)).
wd_transparent(surjection_relation(_,_)).
wd_transparent(total_surjection_relation(_,_)).
wd_transparent(seq(_)).
wd_transparent(seq1(_)).
wd_transparent(iseq(_)).
wd_transparent(iseq1(_)).
wd_transparent(perm(_)).
wd_transparent(empty_sequence).
wd_transparent(identifier(_)).
wd_transparent(lazy_lookup_expr(_)).
wd_transparent(lazy_lookup_pred(_)).
wd_transparent(integer(_)).
wd_transparent(integer_set(_)).
wd_transparent(string(_)).
wd_transparent(set_extension(_)).
wd_transparent(sequence_extension(_)).
wd_transparent(struct(_)).
wd_transparent(rec(_)).
wd_transparent(record_field(_,_)).
wd_transparent(typeset).

wd_transparent(power_of_real(_,_)). % TODO: check if we need WD POs

wd_transparent(external_function_call(F,_)) :- \+ external_fun_has_wd_condition(F).
wd_transparent(external_pred_call(F,_)) :- \+ external_fun_has_wd_condition(F).
wd_transparent(external_subst_call(F,_)) :- \+ external_fun_has_wd_condition(F).

wd_transparent(operation_call_in_expr(_,_)). % the operations will be checked for WD independently
% PRE conditions are not part of WD checking

wd_transparent(freetype_set(_)).
wd_transparent(freetype_constructor(_,_,_)).
wd_transparent(freetype_case(_,_,_)). % predicate to check if we match a given case

% substitutions:
wd_transparent(skip).
wd_transparent(block(_)).
wd_transparent(parallel(_)).
wd_transparent(choice(_)).
wd_transparent(assign(_,_)). % TO DO: f(X) := E ==>  f := f <+ {X|->E}, etc,...
wd_transparent(assign_single_id(_,_)).
wd_transparent(select(_)).
wd_transparent(select(_,_)). % TOD: Else only reached if all cases do not apply
wd_transparent(becomes_element_of(_,_)). % in classical B we do not have to check set not empty?
wd_transparent(becomes_such(_,_)). % in classical B we do not have to check satisfiability
wd_transparent(operation_call(_,_,_)). % I.e., we do not see checking PRE condition checking as part of WD checking
wd_transparent(var(_,_)).

wd_transparent(witness(_,_)).

% transparent with new quantified variables
wd_transparent(comprehension_set(_,_)).
wd_transparent(exists(_,_)). % WD must be proven universally for all subexpression


/*
Not yet covered:

syntaxelement(kodkod(PId,Ids),       Ids,[],[Ids],PId, pred).

syntaxelement(operation_call_in_expr(Id,As), [Id|As], [], [As], [], expr).

% rewritten:
syntaxelement(event_b_comprehension_set(Ids,E,P),[E,P|Ids],  Ids,[Ids], [], expr/only_typecheck).
syntaxelement(lambda(Ids,P,E),         [P,E|Ids],Ids,[Ids], [], expr).

syntaxelement(quantified_intersection(Ids,P,E), [P,E|Ids],Ids,[Ids], [], expr).


TREE operations

syntaxelement(let_expression_global(Ids,As,Expr), Exprs, Ids, [Ids,As], [], expr) :-  % version used by b_compiler
   append([Ids,As,[Expr]],Exprs).
syntaxelement(lazy_let_expr(TID,A,Expr), Exprs, [TID], [[TID],[A]], [], expr) :-
   append([[TID],[A],[Expr]],Exprs).
syntaxelement(lazy_let_pred(TID,A,Expr), Exprs, [TID], [[TID],[A]], [], pred) :-
   append([[TID],[A],[Expr]],Exprs).
syntaxelement(lazy_let_subst(TID,A,Expr), Exprs, [TID], [[TID],[A]], [], subst) :-
   append([[TID],[A],[Expr]],Exprs).

syntaxelement(compaction(A), [A],  [], [], [], expr).
syntaxelement(mu(A),         [A],  [], [], [], expr).
syntaxelement(bag_items(A),  [A],  [], [], [], expr).

syntaxelement(freetype_set(Id), [], [], [], Id, expr).
syntaxelement(freetype_case(Type,Case,Expr), [Expr], [], [], [Type,Case], pred).
syntaxelement(freetype_constructor(Type,Case,Expr), [Expr], [], [], [Type,Case], expr).
syntaxelement(freetype_destructor(Type,Case,Expr), [Expr], [], [], [Type,Case], expr).

rlevent(Name,Section,_Status,_Params,Guard,_Theorems,_Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents),



seval /Users/leuschel/git_root/private_examples/ClearSy/2020/03_Mar/wd_timeout/invalidRef_welldef.mch
:wd
Analysis walltime: 183 ms

Stats: [AXM/1084,THM/104,unsupported/7,not_discharged_po/545,discharged_po/704,caval__rule__1__compute/61]
Stats: [unsupported/2,AXM/1085,THM/104,not_discharged_po/542,discharged_po/708,caval__rule__1__compute/61]

Analysis walltime: 31 ms
Stats: [AXM/1085,THM/104,not_discharged_po/542,discharged_po/708,caval__rule__1__compute/61]
when reordering and not ignoring infos:
Analysis walltime: 17 ms
Stats: [AXM/1085,THM/104,not_discharged_po/540,discharged_po/710,caval__rule__1__compute/61]

Tests:

 :wd a>0 & x>0 & x:1..10 & f:1..10-->INTEGER & f(x)=2 & x>1
 :wd %x.(x:2..4 | %x.(x:4..8|10/x)(2*x)) = x

 :wd x:NATURAL1 & y>=x & x mod y = x/y
 -> same PO generated twice

:wd f6 = %x.(x>10 & x<20|100) & res=f6(11)
:wd i5 = %x.(x>10|100+x) & i5(100) = 200
 seval /Users/leuschel/git_root/prob_examples/examples/B/ASTD/wetransfer-545a33/Case_Study_Handmade/TRAIN_CONTROL_M6.mch

:wd n:NATURAL & a:1..n --> 1..10 & a /= {} & a(1)=1

-> support for $0 missing
-> support for ; and while (adding condition and invariant to loop body hyps) for transform
-> support for operation all in expression
-> Parallel in classical B: we could propagate guards from first to second subst; CarlaTravelAgency; but Atelier-B considers this to be not well-defined; each parallel construct should be WD on its own; we should make no assumption about order
 -> CASE statement: treat cases specially

 % Processing file 2/3: /Users/leuschel/git_root/private_examples/ClearSy/2020/01_Jan/logxml_system_error/rule.mch
Proof of AXM took 169 ms, result=discharged ...
*/
