% (c) 2020-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(well_def_prover, [prove_po/3]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module proves WD POs.').

:- use_module(library(avl)).
:- use_module(library(lists)).

:- use_module(well_def_hyps,[get_hyp_var_type/3, portray_hyps/1, is_hyp_var/2, is_finite_type_for_wd/2]).

:- use_module(well_def_tools, [rename_norm_term/3, member_in_norm_conjunction/2, not_occurs/2, occurs/2]).
:- use_module(probsrc(debug)).
:- use_module(probsrc(error_manager),[add_error/3, add_internal_error/2]).

:- use_module(probsrc(custom_explicit_sets),[domain_of_explicit_set_wf/3, equal_avl_tree/2,
                range_of_explicit_set_wf/3,
                invert_explicit_set/2,check_interval_in_custom_set/4, is_interval_closure/5,
                check_avl_in_interval/3, check_avl_subset/2, is_one_element_avl/2,
                expand_custom_set_to_list/2, quick_definitely_maximal_set_avl/1,
                expand_and_convert_to_avl_set/4, is_avl_sequence/1, is_avl_partial_function/1]).

% PROVING:
% --------

% some more rules are covered in process_sequent_aux of prove_sequent/3
prove_po(truth,_,truth).
prove_po(NormTarget,hyp_rec(AVL,_),hyp) :- avl_fetch(NormTarget,AVL).
prove_po(member(X,Y),Hyps,mem(PT)) :- % Y is usually domain(Func)
    simplify_expr(Y,Hyps,SY),
    simplify_expr(X,Hyps,SX),
    check_member_of_set(SY,SX,Hyps,PT).
prove_po(not_member(X,Y),Hyps,mem(PT)) :-
    simplify_expr(Y,Hyps,SY),
    simplify_expr(X,Hyps,SX),
    check_not_member_of_set(SY,SX,Hyps,PT).
prove_po(finite(Set),Hyp,finite_set(PT)) :- check_finite(Set,Hyp,PT).
prove_po(not_equal(A,B),Hyp,not_equal) :-  %print(check_not_equal(A,B)),nl,
    simplify_expr(A,Hyp,SA),
    simplify_expr(B,Hyp,SB),
    check_not_equal(SA,SB,Hyp).
prove_po(equal(A,B),Hyp,equal) :- % not generated by our POG
    simplify_expr(A,Hyp,SA),
    simplify_expr(B,Hyp,SB),
    check_equal(SA,SB,Hyp,_).
prove_po(greater(A,B),Hyp,PT) :- prove_po(less(B,A),Hyp,PT).
prove_po(greater_equal(A,B),Hyp,greater_equal) :- % print(check_leq(B,A)),nl,
    check_leq(B,A,Hyp).
prove_po(less_equal(A,B),Hyp,less_equal) :-
    check_leq(A,B,Hyp).
prove_po(less(A,B),Hyp,less) :-
    check_leq(A,B,Hyp),!,
    check_not_equal(A,B,Hyp).
prove_po(subset(A,B),Hyp,PT) :-
    check_is_subset(A,B,Hyp,PT).
prove_po(subset_strict(A,B),Hyp,subset_strict(PT)) :-
    check_is_subset(A,B,Hyp,PT),!,
    check_not_equal(A,B,Hyp).
prove_po(conjunct(A,B),Hyp,conj(T1,T2)) :- % generated by Rodin
   prove_po(A,Hyp,T1),
   !,
   prove_po(B,Hyp,T2).
prove_po(disjunct(A,B),Hyp,conj(T1,T2)) :- % could be generated by Rodin
   (prove_po(A,Hyp,T1) -> true
    ; prove_po(B,Hyp,T2)). % we could add not(A) as hypothesis, this is done in prove_sequent_goal
%prove_po(implication(A,B),Hyp,imply(T2)) :- % generated by Rodin; now treated in prove_sequent_goal
%   push_normalized_hyp(A,Hyp,Hyp2),
%   prove_po(B,Hyp2,T2).
prove_po(exists([$(B0)],forall([$(X)],member($(X),Set),less_equal($(X),$(B0)))),Hyp,finite_max(PT)) :- X \= B0,
    % generated by Rodin for max(Set)
    debug_println(19,checking_finite_for_max(Set)),
    check_finite(Set,Hyp,PT).
prove_po(exists([$(B0)],forall([$(X)],member($(X),Set),less_equal($(B0),$(X)))),Hyp,finite_min(PT)) :- X \= B0,
    % generated by Rodin for min(Set)
    debug_println(19,checking_finite_for_min(Set)),
    check_finite(Set,Hyp,PT).
%prove_po(NT,_,_) :- format('Unproven by WD Prover: ~w~n~n',[NT]),fail.

simple_value(Nr) :- number(Nr).
simple_value('$'(_)).
simple_value(boolean_true).
simple_value(boolean_false).
simple_value(record_field('$'(_),_)).
simple_value(value(_)).
simple_value(string(_)).
simple_value(function(F,_)) :- simple_value(F).
simple_value(couple(A,B)) :- simple_value(A), simple_value(B).

%get_set_of_possible_values(X,Hyps,XSet) :-
%   if(try_get_set_of_possible_values(X,Hyps,R), XSet=R,
%      XSet=set_extension([X])). % was typeset

try_get_set_of_possible_values(Nr,Hyps,R,Hyps) :- number(Nr),!,R=interval(Nr,Nr).
try_get_set_of_possible_values(X,Hyps,R,Hyps1) :-
   avl_fetch_binop_from_hyps(X,equal,Hyps,Nr,Hyps1), number(Nr),!, % TODO: also treat is_explicit_value
   R=interval(Nr,Nr).
try_get_set_of_possible_values(X,Hyps,XSet,Hyps2) :-
   avl_fetch_binop_from_hyps(X,member,Hyps,XSet1,Hyps1),
   (improve_integer_set_precision(X,XSet1,Hyps1,XSet,Hyps2)
     -> true ; Hyps2=Hyps1, XSet=XSet1).
try_get_set_of_possible_values(X,Hyps,Res,Hyps3) :-
   avl_fetch_equal_from_hyps(X,Hyps,X2,Hyps1),
   quick_not_occurs_check(X,X2),
   rewrite_local_loop_check(X,try_get_set_of_possible_values,X2,Hyps1,Hyps2),
   (X2='$'(_) -> X = '$'(_) ; true), % avoid rewriting x -> card(...) -> x; TO DO: better cyclic equality prevention
   try_get_set_of_possible_values(X2,Hyps2,Res,Hyps3), !.
try_get_set_of_possible_values(function(Func,_),Hyps,RangeSet,Hyps2) :- !,
   get_range_or_superset(Func,Hyps,RangeSet,Hyps2).
try_get_set_of_possible_values(first(Seq),Hyps,RangeSet,Hyps2) :- !, % first(S) === S(1)
   get_range_or_superset(Seq,Hyps,RangeSet,Hyps2).
try_get_set_of_possible_values(last(Seq),Hyps,RangeSet,Hyps2) :- !, % last(S) === S(size(S))
   get_range_or_superset(Seq,Hyps,RangeSet,Hyps2).
% TO DO other sequence operations
try_get_set_of_possible_values(couple(A,B),Hyps,cartesian_product(SA,SB),Hyps2) :- !,
  try_get_set_of_possible_values(A,Hyps,SA,Hyps1),
  try_get_set_of_possible_values(B,Hyps1,SB,Hyps2).
try_get_set_of_possible_values(max(Set),H,Set,H). % we could do special rules for interval
try_get_set_of_possible_values(min(Set),H,Set,H).
try_get_set_of_possible_values(mu(Set),H,Set,H).
try_get_set_of_possible_values(max_int,H,'NATURAL1',H).
% TO DO: min_int
try_get_set_of_possible_values(value(avl_set(A1)),Hyps,value(avl_set(A2)),Hyps) :- !,
   expand_and_convert_to_avl_set([avl_set(A1)],A2,get_set_of_possible_values,'WD(?)').
try_get_set_of_possible_values(add(X,Y),Hyps,Set,Hyps2) :-
   add_with_number(add(X,Y),A,Nr),
   try_get_set_of_possible_values(A,Hyps,ValA,Hyps2),
   add_to_value_set(ValA,Nr,Set),!.
try_get_set_of_possible_values(minus(A,Nr),Hyps,Set,Hyps2) :-
   try_get_set_of_possible_values(A,Hyps,ValA,Hyps2), number(Nr), Nr1 is -Nr,
   add_to_value_set(ValA,Nr1,Set),!.
   % TO DO : add unary_minus, multiplication with nr
try_get_set_of_possible_values(Seq,Hyps,Set,Hyps) :-
   infer_sequence_type_of_expr(Seq,Hyps,SeqType),!,
   (SeqType=seq1 -> Set=seq1(typeset) ; Set=seq(typeset)). % TO DO: examine type of list elements
try_get_set_of_possible_values(X,Hyps,XSet,Hyps2) :-
   avl_fetch_binop_from_hyps(X,greater_equal,Hyps,Low,Hyps1), !,
   (avl_fetch_binop_from_hyps(X,less_equal,Hyps1,Up,Hyps2) -> XSet = interval(Low,Up)
    ; Low=0 -> XSet = 'NATURAL', Hyps2=Hyps1
    ; number(Low),Low>0 -> XSet= 'NATURAL1', Hyps2=Hyps1). % TO DO: improve precision
try_get_set_of_possible_values(X,Hyps,XSet,Hyps2) :-
   rewrite_integer(X,Hyps,X2,Hyps1),!,
   try_get_set_of_possible_values(X2,Hyps1,XSet,Hyps2).
try_get_set_of_possible_values(size(A),Hyps,XSet,Hyps) :-
   (check_not_empty_set(A,Hyps) -> XSet = 'NATURAL1' ; XSet='NATURAL').
try_get_set_of_possible_values(card(A),Hyps,XSet,Hyps) :-
   (check_not_empty_set(A,Hyps) -> XSet = 'NATURAL1' ; XSet='NATURAL').

is_integer_set(interval(_,_)).
is_integer_set('NATURAL').
is_integer_set('NATURAL1').
%  TO DO: add avl_set

% detect sets which can profit from narrowing down:
is_infinite__or_large_integer_set('NATURAL',0,inf).
is_infinite__or_large_integer_set('NATURAL1',1,inf).
is_infinite__or_large_integer_set(interval(Low,max_int),Low,max_int). % one cannot prove a lot with max_int anyway!?

% try and improve precision of integer set
% limitation: only looks for one other hypotheses; maybe we should do this in well_defs_hyps.pl
improve_integer_set_precision(X,XSet1,Hyps1,NewSet,Hyps3) :-
   is_integer_set(XSet1),
   avl_fetch_binop_from_hyps(X,member,Hyps1,XSet2,Hyps2),
   XSet2 \= XSet1,
   !,
   intersect_sets(XSet1,XSet2,XSet12),
   try_improve_interval(XSet12,X,Hyps2,NewSet,Hyps3). % TO DO: we could try and find another member
   % TO DO: also look at less_equal, greater_equal constraints
improve_integer_set_precision(X,XSet1,Hyps1,NewSet,Hyps3) :-
   is_infinite__or_large_integer_set(XSet1,Low,LargeUp),
   avl_fetch_binop_from_hyps(X,less_equal,Hyps1,Up,Hyps2),
   Up \= LargeUp, % we really improve upon earlier value
   !,
   try_improve_interval(interval(Low,Up),X,Hyps2,NewSet,Hyps3).
improve_integer_set_precision(X,XSet1,Hyps1,NewSet,Hyps2) :-
   try_improve_interval(XSet1,X,Hyps1,NewSet,Hyps2).

try_improve_interval(interval(OldLow,OldUp),X,Hyps1,interval(NewLow,NewUp),Hyps2) :- !,
   improve_interval(X,OldLow,OldUp,Hyps1,NewLow,NewUp,Hyps2).
try_improve_interval(Set,_,Hyps,Set,Hyps).

% phase 1: try improve upper bound
improve_interval(X,OldLow,OldUp,Hyps1,NewLow,NewUp,Hyps3) :-
   improve_upper_bound(X,OldUp,Hyps1,NewUp1,Hyps2),!,
   improve_interval(X,OldLow,NewUp1,Hyps2,NewLow,NewUp,Hyps3).
improve_interval(X,OldLow,Up,Hyps1,NewLow,Up,Hyps2) :-
   improve_interval2(X,OldLow,Hyps1,NewLow,Hyps2).
%
improve_upper_bound(X,OldUp,Hyps1,NewUp,Hyps2) :-
   avl_fetch_binop_from_hyps(X,less_equal,Hyps1,Up,Hyps2),
   order_values(OldUp,Up,NewUp,OldUp),
   NewUp \= OldUp.

% now try and improve lower bound:
improve_interval2(X,OldLow,Hyps1,NewLow,Hyps3) :-
   improve_lower_bound(X,OldLow,Hyps1,Low1,Hyps2),!,
   improve_interval2(X,Low1,Hyps2,NewLow,Hyps3).
improve_interval2(_,Low,Hyps,Low,Hyps).
%
improve_lower_bound(X,OldLow,Hyps1,NewLow,Hyps2) :-
   avl_fetch_binop_from_hyps(X,greater_equal,Hyps1,Low,Hyps2),
   order_values(Low,OldLow,OldLow,NewLow),
   NewLow \= OldLow.


% try and intersect two sets:
intersect_sets(interval(L1,U1),B,Res) :- !, intersect_interval(B,L1,U1,Res).
intersect_sets(B,interval(L1,U1),Res) :- intersect_interval(B,L1,U1,Res).
intersect_sets('NATURAL1','NATURAL','NATURAL1').
intersect_sets('NATURAL','NATURAL1','NATURAL1').

intersect_interval(interval(L2,U2),L1,U1,interval(L3,U3)) :-
  order_values(L1,L2,_,L3), % choose larger value as lower bound
  order_values(U1,U2,U3,_). % choose smaller value as upper bound
intersect_interval('NATURAL1',L1,U1,interval(L3,U1)) :- order_values(L1,1,L3,_).
intersect_interval('NATURAL',L1,U1,interval(L3,U1)) :- order_values(L1,0,L3,_).

% order values for interval intersection
order_values(N1,N2,R1,R2) :- number(N1),!, order_aux_nr(N2,N1,R1,R2).
order_values(N1,N2,R1,R2) :- number(N2),!, order_aux_nr(N1,N2,R1,R2).
order_values(min_int,N2,R1,R2) :- !, R1=min_int,R2=N2.
order_values(max_int,N2,R1,R2) :- !, R1=N2,R2=max_int.
order_values(N1,N2,N1,N2). % just choose N1

order_aux_nr(N2,N1,R1,R2) :- number(N2),!,
    (N2>N1 -> R1=N1,R2=N2 ; R1=N2,R2=N1).
order_aux_nr(max_int,N1,R1,R2) :- N1 < 1, !, R1=N1, R2=max_int.
order_aux_nr(_N2,N1,N1,N1). % choose the number as the bound


%try_get_set_of_possible_values(X,_,_,_) :- print(try_get_set_of_possible_values_failed(X)),nl,fail.
% TO DO: more precise representation of open-ended intervals interval(Low,'$infinity'))
% TO DO: intersect multiple solutions; e.g., intervals and >=, <= constraints

add_to_value_set(interval(L,U),Nr,interval(L2,U2)) :- add_nr(L,Nr,L2), add_nr(U,Nr,U2).
add_to_value_set('NATURAL',1,'NATURAL1'). % TO DO: extend
%add_to_value_set(value(avl_set(A1),Nr,value(avl_set(A2)) :- TO DO: add Nr to all values in A2

% adding a known number to an arithmetic expression; could be extended
% this is mainly for dealing with index arithmetic for arrays
add_nr(Nr1,ToAdd,Nr2) :- number(Nr1),!, Nr2 is Nr1+ToAdd.
add_nr(minus(Expr,Nr1),ToAdd,Res) :- number(Nr1),!, Nr2 is Nr1-ToAdd,
   (Nr2=0 -> Res=Expr ; Res= minus(Expr,Nr2)).
add_nr(add(Expr,Nr1),ToAdd,Res) :- number(Nr1),!, Nr2 is Nr1+ToAdd,
   (Nr2=0 -> Res=Expr ; Res= add(Expr,Nr2)).
add_nr(add(Nr1,Expr),ToAdd,Res) :- number(Nr1),!, Nr2 is Nr1+ToAdd,
   (Nr2=0 -> Res=Expr ; Res= add(Expr,Nr2)).


% check if an expression is a sequence
check_is_sequence(El,Hyps) :- check_is_sequence(El,Hyps,_).
check_is_non_empty_sequence(El,Hyps) :- check_is_sequence(El,Hyps,seq1).

check_is_sequence(empty_set,_,seq) :- !.
check_is_sequence(empty_sequence,_,seq) :- !.
check_is_sequence(El,Hyps,RequiredType) :-
  infer_sequence_type_of_expr(El,Hyps,Type),
  (Type=seq1 -> true ; RequiredType=seq -> true
     ; check_not_equal(El,empty_set,Hyps)).
check_is_sequence(A,Hyps,RequiredType) :-
   avl_fetch_worthwhile_equal_from_hyps(A,Hyps,Value,Hyps2),
   check_is_sequence(Value,Hyps2,RequiredType).
check_is_sequence(domain_restriction(Dom,S),Hyps,Res) :- !,
    is_interval(Dom,Hyps,1,N),
    check_is_sequence(S,Hyps,SeqType),
    (SeqType=seq1, check_leq(1,N,Hyps) -> Res=seq1 ; Res=seq).
check_is_sequence(El,Hyps,RequiredType) :-
   avl_fetch_mem_or_struct(El,Hyps,Set,Hyps2),
   is_set_of_sequences_type(Set,Hyps2,Type), % should we move this to subset? dealt with in subset_transitivity_rule
   % required for :wd Right:seq(BOOL) & (Right/=[] => tail(Right)=res) in test 2018
   (Type=seq1 -> true ; RequiredType=seq -> true
     ; check_not_equal(El,empty_set,Hyps2)),
    !.
check_is_sequence(X,Hyps,RequiredType) :- try_get_set_of_possible_values(X,Hyps,XSet,Hyps2),
   (RequiredType==seq1 -> check_is_subset(XSet,seq1(typeset),Hyps2,_PT)
    ; check_is_subset(XSet,seq(typeset),Hyps2,_PT)).
% check if something is an interval
is_interval(Expr,Hyps,A,B) :- simplify_expr(Expr,Hyps,SE), is_interval_aux(SE,A,B).
is_interval_aux(interval(A,B),A,B).
is_interval_aux(value(avl_set(AVL)),A,B) :- avl_is_interval(AVL,A,B).

is_set_of_sequences_type(seq1(_),_,seq1) :- !.
is_set_of_sequences_type(seq(_),_,seq) :- !.
is_set_of_sequences_type(iseq(_),_,seq) :- !.
is_set_of_sequences_type(iseq1(_),_,seq) :- !.
is_set_of_sequences_type(perm(A),Hyps,Type) :- !,
   (check_not_equal(A,empty_set,Hyps) -> Type=seq1 ; Type=seq).
is_set_of_sequences_type(Func,Hyps,Type) :- % a total function 1..Up --> Range is a sequence
   get_exact_domain_of_func_or_rel_type(Func,Hyps,Dom,Hyps1),
   check_equal_pattern(Dom,interval(1,Up),Hyps1,Hyps2),
   % we could call check_equal for Low; relevant for :wd BV=16 & II=1 & BIdx = II..BV & s:BIdx --> BOOL & res=size(s)
   is_partial_function_type(Func,Hyps2,_),
   (number(Up), Up>0  % we could call check_leq
    -> Type=seq1 ; Type=seq).

% a simple equality check
check_equal_pattern(A,Target,Hyps,Hyps1) :-
   check_equal_h(A,Target,not_ground,[],Hyps,Hyps1).
check_equal(A,Target,Hyps,Hyps1) :-
   check_equal_h(A,Target,ground,[],Hyps,Hyps1).

% TargetGround=ground means Target is a ground, fully known expressions and not a pattern with variables
check_equal_h(A,Target,_,_,Hyps,Hyps1) :- A=Target,!, Hyps1=Hyps.
check_equal_h(intersection(A,B),EmptySet,ground,_,Hyps,Hyps) :- is_empty_set(EmptySet),!,
   check_disjoint(A,B,Hyps).
check_equal_h(couple(A1,A2),couple(B1,B2),TGr,History,Hyps,Hyps2) :- % TO DO: records, ...
   check_equal_h(A1,B1,TGr,History,Hyps,Hyps1),!,
   check_equal_h(A2,B2,TGr,History,Hyps1,Hyps2).
check_equal_h(A,Target,TGr,History,Hyps,Hyps2) :-
   avl_fetch_equal_from_hyps(A,Hyps,A2,Hyps1), nonmember(A2,History),
   check_equal_h(A2,Target,TGr,[A|History],Hyps1,Hyps2).
check_equal_h(A,Target,ground,_,Hyps,Hyps) :-
   avl_fetch_from_hyps(subset(A,Target),Hyps),
   avl_fetch_from_hyps(subset(Target,A),Hyps).

infer_sequence_type_of_expr(sequence_extension([_|_]),_,seq1).
infer_sequence_type_of_expr(sorted_set_extension(SList),_,seq1) :-
   sorted_set_extension_is_seq(SList,1).
infer_sequence_type_of_expr(set_extension(List),_,seq1) :- sort(List,SList),
   sorted_set_extension_is_seq(SList,1).
infer_sequence_type_of_expr(insert_tail(_,_),_,seq1).
   % we do not need to check Seq is a sequence; this will be checked in its own PO, ditto for operators below
infer_sequence_type_of_expr(insert_front(_,_),_,seq1).
infer_sequence_type_of_expr(concat(A,B),Hyps,R) :-
   (  infer_sequence_type_of_expr(A,Hyps,seq1) -> R=seq1
    ; infer_sequence_type_of_expr(B,Hyps,seq1) -> R=seq1
    ;  R=seq).
infer_sequence_type_of_expr(restrict_front(_,_),_,seq).
infer_sequence_type_of_expr(restrict_tail(_,_),_,seq).
infer_sequence_type_of_expr(rev(A),Hyps,R) :-
   (infer_sequence_type_of_expr(A,Hyps,seq1) -> R=seq1 ; R=seq).
infer_sequence_type_of_expr(front(_),_,seq). % we could call check_not_empty_set(front(A),Hyps)
infer_sequence_type_of_expr(tail(_),_,seq).  % ditto
infer_sequence_type_of_expr(general_concat(_),_,seq).
infer_sequence_type_of_expr(value(avl_set(SeqAVL)),_,seq1) :- !, \+ SeqAVL=empty,
   is_avl_sequence(SeqAVL).
infer_sequence_type_of_expr(if_then_else(_,A,B),Hyps,Type) :- !,
   (infer_sequence_type_of_expr(A,Hyps,S1)
    -> infer_sequence_type_of_expr(B,Hyps,S2), lub_seq(S1,S2,Type)).
infer_sequence_type_of_expr(Expr,Hyps,R) :-
   is_lambda_function_with_domain(Expr,Domain),
   Domain = interval(1,N),
   (check_leq(1,N,Hyps) -> R = seq1 ; R=seq).
% TO DO: rule for composition

lub_seq(seq1,seq1,seq1).
lub_seq(seq1,seq,seq).
lub_seq(seq,seq1,seq).
lub_seq(seq,seq,seq).

% check if a sorted set extension represent a proper sequence
sorted_set_extension_is_seq([],_).
sorted_set_extension_is_seq([couple(Nr,_)|T],Nr) :- N1 is Nr+1, sorted_set_extension_is_seq(T,N1).

% --------
% DOMAIN

% compute exact domain
% currently there can be multiple solutions for $(_) case below; first one is usually more precise
%compute_exact_domain(Value,Hyps2,Res) :- debug:print_quoted_with_max_depth(compute_exact_domain(Value,Hyps2,Res),4),nl,fail.


compute_exact_domain(reverse(Func),Hyps,Res,Hyps2) :- !,
   compute_exact_range(Func,Hyps,Res,Hyps2).
compute_exact_domain(rev(Func),Hyps,Res,Hyps2) :- !, % reverse of a sequence; domain identical
   compute_exact_domain(Func,Hyps,Res,Hyps2).
compute_exact_domain(identity(Domain),Hyps,Res,Hyps2) :- !, Hyps2=Hyps, Res=Domain.
compute_exact_domain(restrict_front(_Seq,N),Hyps,Res,Hyps2) :- !,
   % WD Condition requires N : 0..size(Seq)
   Hyps2=Hyps, Res = interval(1,N). % TODO: similar rule for restrict_tail
compute_exact_domain(Func,Hyps,Res,Hyps2) :-
   compute_exact_domain_direct(Func,Hyps,Res,Hyps2),!. % No recursive application of equal or hyp
compute_exact_domain(Func,Hyps,Domain,Hyps2) :-
   avl_fetch_worthwhile_equal_from_hyps(Func,Hyps,Func2,Hyps1),
   compute_exact_domain(Func2,Hyps1,Domain,Hyps2).
%compute_exact_domain(Expr,_,Domain,_) :- print(compute_exact_domain_failed(Expr,_,Domain)),nl,fail.


compute_exact_domain_direct(Func,Hyps,Res,Hyps2) :-
   avl_fetch_mem_or_struct(Func,Hyps,Function,Hyps1), % look for  Func : Res --> Range ; e.g. Func:perm(1..10) -> DomSet=1..10
   % f : _ +-> ( Dom --> _ ) & x:Dom   ==>  x:dom(f(_))
   % f : _ +-> ( Dom --> _ )  =>  dom(f(_)) = Dom
   get_exact_domain_of_func_or_rel_type(Function,Hyps1,Res,Hyps2),!. % is thus also minimal domain
compute_exact_domain_direct(Func,Hyps,Res,Hyps3) :- Func = '$'(_), % Look for Func = Value definition
   avl_fetch_equal_from_hyps(Func,Hyps,Value,Hyps2),
   compute_exact_domain(Value,Hyps2,Res,Hyps3).
compute_exact_domain_direct(value(CS),Hyps,value(Res),Hyps) :- !, nonvar(CS),
   domain_of_explicit_set_wf(CS,Res,no_wf_available).
compute_exact_domain_direct(overwrite(F1,F2),Hyps,D12,Hyps2) :- !,  % dom(F1 <+ F2) = dom(F1) \/ dom(F2)
   compute_exact_domain(F1,Hyps,D1,Hyps1), compute_exact_domain(F2,Hyps1,D2,Hyps2),
   construct_union(D1,D2,Hyps2,D12).
compute_exact_domain_direct(domain_restriction(S,F),Hyps,intersection(S,D),Hyps2) :- !,  % dom(S <| F) = S /\ dom(F)
   compute_exact_domain(F,Hyps,D,Hyps2).
compute_exact_domain_direct(domain_subtraction(S,F),Hyps,set_subtraction(D,S),Hyps2) :- !,  % dom(S <<| F) = dom(F) - S
   compute_exact_domain(F,Hyps,D,Hyps2).
compute_exact_domain_direct(composition(F1,F2),Hyps,Domain,Hyps4) :- !,  % dom((F1;F2)) = dom(F1) if ran(F1) <: dom(F2)
   compute_exact_domain(F1,Hyps,Domain,Hyps2),
   compute_exact_domain(F2,Hyps2,D2,Hyps3), % or_subset would also be ok
   (maximal_set(D2,Hyps3) -> Hyps4=Hyps3
     ; get_range_or_superset(F1,Hyps3,R1,Hyps4),
       check_is_subset(R1,D2,Hyps4,_PT)
   ).
compute_exact_domain_direct(union(F,G),Hyps,UnionDFDG,Hyps2) :- !,  % dom(F \/ G) = dom(F) \/ dom(G)
   compute_exact_domain(F,Hyps,DF,Hyps1),
   compute_exact_domain(G,Hyps1,DG,Hyps2),
   construct_union(DF,DG,Hyps2,UnionDFDG).
compute_exact_domain_direct(sorted_set_extension(List),Hyps,Res,Hyps2) :- !,
   compute_exact_domain_direct(set_extension(List),Hyps,Res,Hyps2).
compute_exact_domain_direct(set_extension(List),Hyps,Res,Hyps) :- !, maplist(get_dom_el,List,Domain),
   construct_set_extension(Domain,Hyps,Res).
compute_exact_domain_direct(Expr,Hyps,Domain,Hyps) :- is_lambda_function_with_domain(Expr,Domain),!.
compute_exact_domain_direct(Func,Hyps,Domain,Hyps2) :-
   avl_fetch_equal_from_hyps(domain(Func),Hyps,Domain,Hyps2).

% get domain element of a couple
get_dom_el(couple(A,_),A).
% get range element of a couple
get_ran_el(couple(_,B),B).

% construct union term with a few optimisations
construct_union(empty_set,B,_Hyps,Res) :- !,Res=B.
construct_union(set_extension(A),set_extension(B),Hyps,Res) :- !,
   append(A,B,AB),
   construct_set_extension(AB,Hyps,Res).
construct_union(A,empty_set,_,Res) :- !,Res=A.
construct_union(A,B,_,union(A,B)).

% get maximal domain of a function (i.e., domain or superset thereof)
get_domain_or_superset(reverse(Func),Hyps,Res,Hyps2) :- !,
   get_range_or_superset(Func,Hyps,Res,Hyps2).
get_domain_or_superset(Func,Hyps,Res,Hyps3) :-
   compute_exact_domain_direct(Func,Hyps,Res,Hyps2),
   rewrite_local_loop_check(Func,get_domain_or_superset,Res,Hyps2,Hyps3),
   !.
get_domain_or_superset(domain_restriction(A,_),Hyps,Res,Hyps) :- Res=A. % in case compute_exact_domain_direct fails
get_domain_or_superset(Func,Hyps,DomSet,Hyps2) :- simple_value(Func),
   avl_fetch_mem_or_struct(Func,Hyps,FunctionType,Hyps1),
   get_domain_or_superset_of_func_or_rel_type(FunctionType,Hyps1,DomSet,Hyps2).
   %get_possible_domain_of_func_or_rel_type(FunctionType,Hyps,DomSet,_).
get_domain_or_superset(Func,Hyps,Domain,Hyps2) :-
   avl_fetch_worthwhile_equal_from_hyps(Func,Hyps,Func2,Hyps1),
   get_domain_or_superset(Func2,Hyps1,Domain,Hyps2).
get_domain_or_superset(Func,Hyps,DomSuperSet,Hyps2) :-
   avl_fetch_binop_from_hyps(domain(Func),subset,Hyps,DomSuperSet,Hyps2).

% get exact (thus also minimal) domain of a function type
get_exact_domain_of_func_or_rel_type(FunType,Hyps,A,Hyps) :-
   get_possible_domain_of_func_or_rel_type(FunType,Hyps,A,exact),!.
get_exact_domain_of_func_or_rel_type(FunType,Hyps,Domain,Hyps2) :-
   avl_fetch_worthwhile_equal_from_hyps(FunType,Hyps,FunType2,Hyps1), % in case we have a definition like X = 1..n --> R
   get_exact_domain_of_func_or_rel_type(FunType2,Hyps1,Domain,Hyps2).
get_exact_domain_of_func_or_rel_type(sorted_set_extension(F),Hyps,Domain,Hyps2) :- !,
   get_exact_domain_of_func_or_rel_type(set_extension(F),Hyps,Domain,Hyps2).
get_exact_domain_of_func_or_rel_type(set_extension([Func|TF]),Hyps,Domain,Hyps2) :-
   compute_exact_domain(Func,Hyps,Domain,Hyps2), % now check that all other functions have the same domain
   (member(Func2,TF), \+ compute_exact_domain(Func2,Hyps2,Domain,_) -> fail
    ; true).

get_possible_domain_of_func_or_rel_type(iseq(_),_,'NATURAL1',subset).
get_possible_domain_of_func_or_rel_type(iseq1(_),_,'NATURAL1',subset).
get_possible_domain_of_func_or_rel_type(partial_bijection(A,_),_,A,subset).
get_possible_domain_of_func_or_rel_type(partial_function(A,_),_,A,subset).
get_possible_domain_of_func_or_rel_type(partial_injection(A,_),_,A,subset).
get_possible_domain_of_func_or_rel_type(partial_surjection(A,_),_,A,subset).
get_possible_domain_of_func_or_rel_type(perm(A),Hyps,Domain,Type) :-
     (compute_card_of_set(A,Hyps,CardA,_) % we could do check_finite and use card(A) instead of CardA
       -> Domain = interval(1,CardA), Type=exact
        ; check_finite(A,Hyps,_) -> Domain = interval(1,card(A)), Type=exact
        ; print(could_not_compute_card_for_perm(A)),nl,
          Domain = 'NATURAL1', Type=subset
      ).
get_possible_domain_of_func_or_rel_type(relations(A,_),_,A,subset).
get_possible_domain_of_func_or_rel_type(seq(_),_,'NATURAL1',subset).
get_possible_domain_of_func_or_rel_type(seq1(_),_,'NATURAL1',subset).
get_possible_domain_of_func_or_rel_type(surjection_relation(A,_),_,A,subset).
get_possible_domain_of_func_or_rel_type(total_bijection(A,_),_,A,exact).
get_possible_domain_of_func_or_rel_type(total_function(A,_),_,A,exact).
get_possible_domain_of_func_or_rel_type(total_injection(A,_),_,A,exact).
get_possible_domain_of_func_or_rel_type(total_relation(A,_),_,A,exact).
get_possible_domain_of_func_or_rel_type(total_surjection_relation(A,_),_,A,exact).
get_possible_domain_of_func_or_rel_type(total_surjection(A,_),_,A,exact).


% variation of get_possible_domain_of_func_or_rel_type, which uses Hyps and can deal with set_extensions
get_domain_or_superset_of_func_or_rel_type(sorted_set_extension(List),Hyps,Dom,Hyps2) :- !,
   get_domain_or_superset_of_func_or_rel_type(set_extension(List),Hyps,Dom,Hyps2).
get_domain_or_superset_of_func_or_rel_type(set_extension(List),Hyps,Dom,Hyps2) :- !,
   % if we have f: {f1,f2,...}  =>  dom(f) <: dom(f1) \/ dom(f2) \/ ...
   merge_possible_domains_of_list(List,Hyps,empty_set,Dom,Hyps2).
get_domain_or_superset_of_func_or_rel_type(Func,Hyps,Res,Hyps) :-
   get_possible_domain_of_func_or_rel_type(Func,Hyps,D,_),!,Res=D.

% merge domains of a list of possible functions
merge_possible_domains_of_list([],Hyps,Acc,Acc,Hyps).
merge_possible_domains_of_list([H|T],Hyps,Acc,Res,Hyps2) :-
   get_domain_or_superset(H,Hyps,Domain,Hyps1),!,
   construct_union(Acc,Domain,Hyps1,Acc1),
   merge_possible_domains_of_list(T,Hyps1,Acc1,Res,Hyps2).

% RANGE
% -----

% compute range or subset thereof

compute_exact_range(reverse(Func),Hyps,Res,Hyps2) :-
   compute_exact_domain(Func,Hyps,Res,Hyps2).
compute_exact_range(rev(Func),Hyps,Res,Hyps2) :- % reverse of a sequence: same range
   compute_exact_range(Func,Hyps,Res,Hyps2).
compute_exact_range(identity(Domain),Hyps,Res,Hyps2) :- !, Hyps2=Hyps, Res=Domain.
compute_exact_range(Func,Hyps,Res,Hyps2) :-
   compute_exact_range_direct(Func,Hyps,Res,Hyps2),!. % No recursive application of equal or hyp
compute_exact_range(Func,Hyps,Res,Hyps2) :- Func = '$'(_),
   avl_fetch_mem_or_struct(Func,Hyps,FunctionType,Hyps2), % Func : _ --> Res
   get_exact_range_of_func_type_direct(FunctionType,Res).
compute_exact_range(Func,Hyps,Range,Hyps2) :-
   avl_fetch_worthwhile_equal_from_hyps(Func,Hyps,Func2,Hyps1),
   compute_exact_range(Func2,Hyps1,Range,Hyps2).
%compute_exact_range(Expr,H,Domain) :- nl,portray_hyps(H),nl,print(compute_range_failed(Expr,_,Domain)),nl,fail.
% TO DO: rule for composition (exact case)


compute_exact_range_direct(empty_set,Hyps,empty_set,Hyps) :- !.
compute_exact_range_direct(empty_sequence,Hyps,empty_set,Hyps) :- !.
compute_exact_range_direct(function(Func2,_),Hyps,Res,Hyps2) :-
   % f : _ +-> ( _ --> Ran ) & x:Ran   ==>  x:ran(f(_))
   % f : _ +-> ( _ -->> Ran )  =>  ran(f(_)) = Ran
   get_range_or_superset(Func2,Hyps,Range,Hyps2),
   get_exact_range_of_func_type_direct(Range,Res). % is thus also minimal domain
compute_exact_range_direct(value(CS),Hyps,value(Res),Hyps) :- !, nonvar(CS), % TO DO: maybe only if small enough
   range_of_explicit_set_wf(CS,Res,no_wf_available).
compute_exact_range_direct(sequence_extension(L),Hyps,Res,Hyps) :- !,
   construct_set_extension(L,Hyps,Res).
compute_exact_range_direct(union(F,G),Hyps,UnionRFRG,Hyps2) :- !,  % ran(F \/ G) = ran(F) \/ ran(G)
   compute_exact_range(F,Hyps,RF,Hyps1),
   compute_exact_range(G,Hyps1,RG,Hyps2),
   construct_union(RF,RG,Hyps2,UnionRFRG).
compute_exact_range_direct(sorted_set_extension(List),Hyps,Res,Hyps2) :- !,
   compute_exact_range_direct(set_extension(List),Hyps,Res,Hyps2).
compute_exact_range_direct(set_extension(List),Hyps,Res,Hyps) :- !, maplist(get_ran_el,List,Domain),
   construct_set_extension(Domain,Hyps,Res).
compute_exact_range_direct(Func,Hyps,Range,Hyps2) :-
   avl_fetch_equal_from_hyps(range(Func),Hyps,Range,Hyps2).

% get maximal range of a function (i.e., range or superset thereof)
get_range_or_superset(reverse(Func),Hyps,Res,Hyps2) :- !,
   get_domain_or_superset(Func,Hyps,Res,Hyps2).
get_range_or_superset(Func,Hyps,Res,Hyps3) :-
   compute_exact_range_direct(Func,Hyps,Res,Hyps2),
   rewrite_local_loop_check(Func,get_range_or_superset,Res,Hyps2,Hyps3),
   !. % can be a loop dom(f) = ran(g)
get_range_or_superset(function(Func2,_),Hyps,Res,Hyps2) :-
   % f2 : _ +-> ( _ --> Res ) ==>  ran(f2(.)) <: Res
   get_range_or_superset(Func2,Hyps,Range,Hyps1),
   get_possible_range_of_func_or_rel_type(Range,Hyps1,Res,_,Hyps2).
get_range_or_superset(range_restriction(_,B),Hyps,Res,Hyps) :- Res=B. % in case compute_exact_range_direct fails
get_range_or_superset(Func,Hyps,RangeSet,Hyps2) :- simple_value(Func),
  avl_fetch_mem_or_struct(Func,Hyps,FunctionType,Hyps1),
  get_possible_range_of_func_or_rel_type(FunctionType,Hyps1,RangeSet,_,Hyps2).
get_range_or_superset(tail(Seq),Hyps,Res,Hyps2) :- !,  % ran(tail(S)) <: ran(S)
  get_range_or_superset(Seq,Hyps,Res,Hyps2).
get_range_or_superset(front(Seq),Hyps,Res,Hyps2) :- !, % ran(front(S)) <: ran(S)
  get_range_or_superset(Seq,Hyps,Res,Hyps2).
get_range_or_superset(restrict_front(Seq,_),Hyps,Res,Hyps2) :- !,
  get_range_or_superset(Seq,Hyps,Res,Hyps2).
get_range_or_superset(restrict_tail(Seq,_),Hyps,Res,Hyps2) :- !,
  get_range_or_superset(Seq,Hyps,Res,Hyps2).
get_range_or_superset(concat(Seq1,Seq2),Hyps,Res12,Hyps2) :- !, % ran(S1^S2) = ran(S1) \/ ran(S2)
  get_range_or_superset(Seq1,Hyps,Res1,Hyps2),
  get_range_or_superset(Seq2,Hyps,Res2,Hyps2),
  construct_union(Res1,Res2,Hyps2,Res12).
get_range_or_superset(composition(_,Func2),Hyps,Res2,Hyps2) :- !, % ran((F1;F2)) <: ran(F2)
  get_range_or_superset(Func2,Hyps,Res2,Hyps2).
get_range_or_superset(Func,Hyps,Range,Hyps2) :-
   avl_fetch_worthwhile_equal_from_hyps(Func,Hyps,Func2,Hyps1),
   get_range_or_superset(Func2,Hyps1,Range,Hyps2).
get_range_or_superset(Func,Hyps,RangeSuperSet,Hyps2) :-
   avl_fetch_binop_from_hyps(range(Func),subset,Hyps,RangeSuperSet,Hyps2).
% to do: more sequence operations: insert_front, insert_tail

% get exact range without equality rewrites
get_exact_range_of_func_type_direct(Func,R) :-
   get_possible_range_of_func_or_rel_type_direct(Func,R,exact).
% TO DO: maybe do same treatment for set_extension as in get_exact_domain_of_func_or_rel_type

% get possible range with equality rewrites
get_possible_range_of_func_or_rel_type(Func,Hyps,Range,ResType,Hyps2) :-
    get_possible_range_of_func_or_rel_type_direct(Func,Range,Type),!, ResType=Type,Hyps2=Hyps.
get_possible_range_of_func_or_rel_type(Func,Hyps,Range,ResType,Hyps2) :-
   avl_fetch_worthwhile_equal_from_hyps(Func,Hyps,Func2,Hyps1),
   get_possible_range_of_func_or_rel_type(Func2,Hyps1,Range,ResType,Hyps2).

% get possible range without equality rewrites
get_possible_range_of_func_or_rel_type_direct(total_function(_,B),B,subset).
get_possible_range_of_func_or_rel_type_direct(total_injection(_,B),B,subset).
get_possible_range_of_func_or_rel_type_direct(total_surjection(_,B),B,exact).
get_possible_range_of_func_or_rel_type_direct(total_bijection(_,B),B,exact).
get_possible_range_of_func_or_rel_type_direct(total_relation(_,B),B,subset).
get_possible_range_of_func_or_rel_type_direct(total_surjection_relation(_,B),B,exact).
get_possible_range_of_func_or_rel_type_direct(partial_function(_,B),B,subset).
get_possible_range_of_func_or_rel_type_direct(partial_injection(_,B),B,subset).
get_possible_range_of_func_or_rel_type_direct(partial_surjection(_,B),B,exact).
get_possible_range_of_func_or_rel_type_direct(partial_bijection(_,B),B,exact).
get_possible_range_of_func_or_rel_type_direct(perm(B),B,exact).
get_possible_range_of_func_or_rel_type_direct(iseq(B),B,subset).
get_possible_range_of_func_or_rel_type_direct(iseq1(B),B,subset).
get_possible_range_of_func_or_rel_type_direct(seq(B),B,subset).
get_possible_range_of_func_or_rel_type_direct(seq1(B),B,subset).
get_possible_range_of_func_or_rel_type_direct(relations(_,B),B,subset).
get_possible_range_of_func_or_rel_type_direct(surjection_relation(_,B),B,exact).


% EXACT REWRITING/SIMPLIFICATION RULES

% simplifier, useful rules independent of context
simplify_expr(A,Hyps,Res) :-
    rewrite_set_expression_exact(A,Hyps,A2,_),!, Res=A2.
simplify_expr(A,Hyps,Res) :-
    rewrite_integer(A,Hyps,A2,_),!, Res=A2.
simplify_expr(record_field(rec(Fields),Field),Hyps,SExpr) :-
   member(field(Field,Expr),Fields),!,
   simplify_expr(Expr,Hyps,SExpr).
simplify_expr(domain(reverse(A)),_,range(A)) :- !.
simplify_expr(range(reverse(A)),_,domain(A)) :- !.
simplify_expr(E,_,E).


get_lambda_args_and_body(IDS,Body,Args,RestBodyList) :-
   append(Args,['$'(Lambda)],IDS), % TO DO: pass lambda info from typed unnormalized expression!
   conj_to_list(Body,BodyList,[]),
   select(equal(A,B),BodyList,RestBodyList),
   (  A='$'(Lambda), not_occurs(B,Lambda)
    ; B='$'(Lambda), not_occurs(A,Lambda)
   ).

% just check if something is a lambda function or similar, without computing exact domain
is_lambda_function(comprehension_set(IDS,Body)) :- !,
   get_lambda_args_and_body(IDS,Body,_,_).
is_lambda_function(Expr) :- is_lambda_function_with_domain(Expr,_).

% determine is something is a lambda function and determine exact domain:
is_lambda_function_with_domain(comprehension_set(IDS,Body),Set) :-
   get_lambda_args_and_body(IDS,Body,Args,RestBodyList),
   get_argument_types(Args,Args,RestBodyList,ArgTypes),
   create_cartesian_product(ArgTypes,Set).
is_lambda_function_with_domain(cartesian_product(Domain,Set),Domain) :-
   singleton_set(Set,_).
is_lambda_function_with_domain(set_extension([couple(El,_)]),set_extension([El])).
is_lambda_function_with_domain(Set,singleton_set([El])) :- singleton_set(Set,couple(El,_)). % TO DO: longer lists and check no multiple domain elements
is_lambda_function_with_domain(sequence_extension(List),interval(1,Len)) :- length(List,Len).
% we could treat domain_restriction, domain_subtraction here

singleton_set(set_extension([El]),El).
singleton_set(sorted_set_extension([El]),El).

conj_to_list(conjunct(A,B)) --> !, conj_to_list(A),conj_to_list(B).
conj_to_list(X) --> [X].

:- use_module(probsrc(tools),[map_split_list/4]).
% we support Arg:Set and we support an argument not appearing at all (equivalent to  Arg:typeset)
get_argument_types([],_,[],[]). % no other conjuncts remain in body
get_argument_types(['$'(ID1)|T],AllArgs,BodyList,[Set1|TS]) :-
    map_split_list(typing_predicate_for(ID1,AllArgs),BodyList,TypingSetList,RestBody),
    create_intersection(TypingSetList,Set1),
    get_argument_types(T,AllArgs,RestBody,TS).

% check if we have a typing predicate for a given identifier
typing_predicate_for(ID1,AllArgs,member('$'(ID1),Set1),Set1) :- l_not_occurs(Set1,AllArgs).
typing_predicate_for(ID1,AllArgs,subset('$'(ID1),SSet1),pow_subset(SSet1)) :- l_not_occurs(SSet1,AllArgs).

% check if any argument appears in expression; if so we have a link between arguments and no proper type
l_not_occurs(Expr,AllArgs) :- member('$'(ID),AllArgs), occurs(Expr,ID),!,fail.
l_not_occurs(_,_).

create_intersection([],typeset). % no constraints on identifier: use typeset
create_intersection([A],Res) :- !, Res=A.
create_intersection([A|T],intersection(A,Rest)) :- create_intersection(T,Rest).

create_cartesian_product([Type],Res) :- !, Res=Type.
create_cartesian_product([Type|T],Res) :- create_cartesian_product3(T,Type,Res).

create_cartesian_product3([],Res,Res).
create_cartesian_product3([Type|T],Acc,Res) :-
    create_cartesian_product3(T,cartesian_product(Acc,Type),Res).
% Note: dom(%(x,y,z).(x:BOOL & y:1..2 & z:BOOL|1)) = (BOOL*(1..2))*BOOL


% Partial Function Check:

% check if we can deduce from the Hypotheses that something is a partial function
check_is_partial_function(Func,Hyps) :-
  avl_fetch_mem_or_struct(Func,Hyps,Function,Hyps1),
   % also deals with function(_)  f : _ +-> ( _ +-> _ )  =>  f(_) : _ +-> _
  is_partial_function_type(Function,Hyps1,_),!.
check_is_partial_function(reverse(Func),Hyps) :-
  check_is_injective(Func,Hyps),!.
check_is_partial_function(value(avl_set(AVL)),_) :- !,
  nonvar(AVL),
  is_avl_partial_function(AVL).
check_is_partial_function(composition(F1,F2),Hyp) :- !,
   % composition of two partial functions is a partial function
   (check_is_partial_function(F1,Hyp)
    -> check_is_partial_function(F2,Hyp)
    ).
check_is_partial_function(overwrite(F1,F2),Hyp) :- !,
   % overwrite of two partial functions is a partial function
   (check_is_partial_function(F1,Hyp)
    -> check_is_partial_function(F2,Hyp)
    ).
check_is_partial_function(identity(_),_Hyp) :- !.
check_is_partial_function(domain_subtraction(_,F),Hyp) :- !, check_is_partial_function(F,Hyp).
check_is_partial_function(domain_restriction(_,F),Hyp) :- !, check_is_partial_function(F,Hyp).
check_is_partial_function(range_subtraction(F,_),Hyp) :- !, check_is_partial_function(F,Hyp).
check_is_partial_function(range_restriction(F,_),Hyp) :- !, check_is_partial_function(F,Hyp).
check_is_partial_function(set_subtraction(F,_),Hyp) :- !, check_is_partial_function(F,Hyp).
check_is_partial_function(intersection(F1,F2),Hyp) :- !,
   (check_is_partial_function(F1,Hyp) -> true ; check_is_partial_function(F2,Hyp)).
check_is_partial_function(sorted_set_extension(List),Hyp) :- !,
   check_set_extension_is_partial_function(List,Hyp).
check_is_partial_function(set_extension(List),Hyp) :- !,
   check_set_extension_is_partial_function(List,Hyp).
check_is_partial_function(Expr,_) :-
   is_lambda_function(Expr),!. % also treats cartesian_product and sequence_extension
check_is_partial_function(Expr,Hyps) :-
   infer_sequence_type_of_expr(Expr,Hyps,_),!. % any sequence expression is a partial function; e.g. a <- b, front(.)
check_is_partial_function(Func,Hyps) :- rewrite_set_expression_exact(Func,Hyps,NewFunc,Hyps2),!,
   check_is_partial_function(NewFunc,Hyps2).
check_is_partial_function(union(F1,F2),Hyp) :- !,
   check_disjoint(F1,F2,Hyp),
   check_is_partial_function(F1,Hyp),
   check_is_partial_function(F2,Hyp).
check_is_partial_function(Func,Hyps) :-
   avl_fetch_equal_from_hyps(Func,Hyps,Value,Hyps2),
   quick_not_occurs_check(Func,Value),
   check_is_partial_function(Value,Hyps2).
%check_is_partial_function(Func,_) :- print(check_is_partial_function_failed(Func)),nl,fail.

% check if this is a partial function type or something defined to be equal to a function type
is_partial_function_type(PF,Hyps,Hyps1) :- is_partial_function(PF,_,_),!,Hyps1=Hyps.
is_partial_function_type(range(Func),Hyps,Hyps2) :-
   get_range_or_superset(Func,Hyps,RanFunc,Hyps1),!,
   is_partial_function_type(RanFunc,Hyps1,Hyps2).
is_partial_function_type(domain(Func),Hyps,Hyps2) :-
   get_domain_or_superset(Func,Hyps,DomFunc,Hyps1),!,
   is_partial_function_type(DomFunc,Hyps1,Hyps2).
is_partial_function_type(sorted_set_extension(Funcs),Hyps,Hyps2) :- !,
   is_partial_function_type(set_extension(Funcs),Hyps,Hyps2).
is_partial_function_type(set_extension(Funcs),Hyps,Hyps2) :- !,
   (member(F,Funcs), \+ check_is_partial_function(F,Hyps) -> fail
    ; Hyps2=Hyps). % all elements of Funcs are partial functions
is_partial_function_type(Func,Hyps,Hyps2) :-
   get_superset(Func,Hyps,SuperSet,Hyps1),!,
   is_partial_function_type(SuperSet,Hyps1,Hyps2).
is_partial_function_type(PF,Hyps,Hyps2) :-
   avl_fetch_worthwhile_equal_from_hyps(PF,Hyps,PF2,Hyps1), % in case we have a definition like X = 1..n --> R
   is_partial_function_type(PF2,Hyps1,Hyps2).

% get worthwhile superset
get_superset(comprehension_set([ID],Body),Hyps,Set,Hyps) :-
    get_parameter_superset_in_body(ID,[ID],Body,Set).
get_superset(set_subtraction(A,_),Hyps,A,Hyps).
get_superset(intersection(A,B),Hyps,R,Hyps) :- (R=A ; R=B).

is_partial_function(total_function(A,B),A,B).
is_partial_function(partial_function(A,B),A,B).
is_partial_function(total_injection(A,B),A,B).
is_partial_function(partial_injection(A,B),A,B).
is_partial_function(total_surjection(A,B),A,B).
is_partial_function(partial_surjection(A,B),A,B).
is_partial_function(total_bijection(A,B),A,B).
is_partial_function(partial_bijection(A,B),A,B).
is_partial_function(perm(A),'NATURAL1',A).
is_partial_function(seq(B),'NATURAL1',B).
is_partial_function(seq1(B),'NATURAL1',B).
is_partial_function(iseq(B),'NATURAL1',B).
is_partial_function(iseq1(B),'NATURAL1',B).

% if First = f(1,GS) -> we can check if function is total; we could store summary of set_extension in hyps
check_set_extension_is_partial_function([_],_) :- !. % one element set extension is a function
check_set_extension_is_partial_function(List,Hyps) :-
   maplist(get_explicit_dom_value(Hyps),List,VList),!,
   sort(VList,SList),
   SList = [couple(First,_)|TS],
   check_set_ext_pf(TS,First,Hyps).
check_set_extension_is_partial_function([couple(A,_),couple(B,_)],Hyps) :-
   check_not_equal(A,B,Hyps). % TO DO: all_different for longer lists

check_set_ext_pf([],_,_).
check_set_ext_pf([couple(Next,_)|TS],Last,Hyp) :-
   Next \= Last,
   check_set_ext_pf(TS,Next,Hyp).

get_explicit_dom_value(Hyps,couple(Val,RanVal),couple(Val2,RanVal)) :- get_explicit_value(Val,Hyps,Val2).

get_explicit_value(couple(A,B),Hyps,couple(A2,B2)) :- !,
    get_explicit_value(A,Hyps,A2), get_explicit_value(B,Hyps,B2).
get_explicit_value(rec(Fields),Hyps,rec(SFields2)) :- !,
     maplist(get_field_value(Hyps),Fields,Fields2),
     sort(Fields2,SFields2).
get_explicit_value(Val,Hyps,R) :- is_explicit_value(Val,Hyps),!,R=Val.
get_explicit_value('$'(ID),Hyps,Val2) :-
    avl_fetch_equal_from_hyps('$'(ID),Hyps,Val2,Hyps2),
    is_explicit_value(Val2,Hyps2). % should we allow recursion through multiple equations?
get_explicit_value(value(R),_,Nr) :- nonvar(R),R=int(Nr), number(Nr). % TODO: more values

% is value which can be compared using Prolog equality
is_explicit_value(boolean_true,_).
is_explicit_value(boolean_false,_).
is_explicit_value(Nr,_) :- number(Nr).
is_explicit_value(string(_),_).
is_explicit_value(couple(A,B),Hyp) :- is_explicit_value(A,Hyp), is_explicit_value(B,Hyp).
is_explicit_value('$'(ID),Hyp) :- is_global_constant_id(ID,Hyp).

get_field_value(Hyps,field(Name,Val),field(Name,Val2)) :- get_explicit_value(Val,Hyps,Val2).

:- use_module(probsrc(b_global_sets),   [lookup_global_constant/2]).
% enumerated set element name
is_global_constant_id(ID,Hyp) :-
    lookup_global_constant(ID,_),
    \+ is_hyp_var(ID,Hyp). % global enumerated set constant visible

is_enumerated_set(ID,Hyp) :-
    enumerated_set(ID),
    \+ is_hyp_var(ID,Hyp). % global enumerated set constant visible

% Disjoint check:
check_disjoint(A,B,Hyps) :-
  (check_disjoint_aux(A,B,Hyps) -> true ; check_disjoint_aux(B,A,Hyps)).
check_disjoint_aux(empty_set,_,_).
check_disjoint_aux(domain_subtraction(A,_),B,Hyps) :- % A <<| f /\ B = {}  if dom(B) <: A
   get_domain_or_superset(B,Hyps,DomB,Hyps2),
   check_is_subset(DomB,A,Hyps2,_).
check_disjoint_aux(set_subtraction(_,A),B,Hyps) :-  % x \ A /\ B = {}  if  B <: A
   check_is_subset(B,A,Hyps,_).
% TO DO: move union of set_extension here?

% Injective check:

check_is_injective(Func,Hyps) :-
  get_type_from_hyps(Func,Hyps,Function,Hyps1),
  %print(check_rev_fun(Func,Function)),nl,
  is_injective_function_type(Function,Hyps1,_).
check_is_injective(value(avl_set(AVL)),_) :- !,
   nonvar(AVL), invert_explicit_set(avl_set(AVL),Inv),
   Inv=avl_set(AVL2), is_avl_partial_function(AVL2).
check_is_injective(identity(_),_).
check_is_injective(Set,_) :- singleton_set(Set,_). % TO DO: extend to more than singleton set_extension
check_is_injective(sequence_extension([_]),_). % TO DO: check all elements are different
check_is_injective(Func, Hyps) :-
  avl_fetch_equal_from_hyps(Func,Hyps,Value,Hyps2),
  %print(check_inj_value(Func,Value)),nl,
  check_is_injective(Value,Hyps2).

% check if this is a partial function type or something defined to be equal to a function type
is_injective_function_type(PF,Hyps,Hyps1) :- is_injective(PF),!,Hyps1=Hyps.
is_injective_function_type(PF,Hyps,Hyps2) :-
   avl_fetch_worthwhile_equal_from_hyps(PF,Hyps,PF2,Hyps1), % in case we have a definition like X = 1..n --> R
   is_injective_function_type(PF2,Hyps1,Hyps2).

is_injective(total_injection(_,_)).
is_injective(partial_injection(_,_)).
is_injective(total_bijection(_,_)).
is_injective(partial_bijection(_,_)).
is_injective(iseq(_)).
is_injective(iseq1(_)).


% check if something is a subset of something else
check_is_subset(H,H,_,equal).
check_is_subset(_,MAX,Hyps,maximal_set) :- maximal_set(MAX,Hyps),!.
check_is_subset(empty_set,_,_,empty_set) :- !.
check_is_subset(empty_sequence,_,_,empty_sequence) :- !.
check_is_subset(cartesian_product(A,B),cartesian_product(A2,B2),Hyps,cart(PTA,PTB)) :- !,
    % A <: A2 & B <: B2  =>  (A*B) <: (A2*B2)
    (check_is_subset(A,A2,Hyps,PTA)
     -> check_is_subset(B,B2,Hyps,PTB)).
check_is_subset('NATURAL1','NATURAL',_,nat1_nat) :- !.
check_is_subset(interval(L,U),B,Hyps,interval(PT)) :- !, check_subset_interval(B,L,U,Hyps,PT).
check_is_subset(intersection(A,B),Super,Hyps,intersection(PT)) :- !,
  ( check_is_subset(A,Super,Hyps,PT) -> true ; check_is_subset(B,Super,Hyps,PT)).
check_is_subset(union(A,B),Super,Hyps,union(PTA,PTB)) :- !,
  ( check_is_subset(A,Super,Hyps,PTA) -> check_is_subset(B,Super,Hyps,PTB)).
check_is_subset(domain_subtraction(_,B),Super,Hyps,dom_sub(PT)) :- !,check_is_subset(B,Super,Hyps,PT).
check_is_subset(domain_restriction(_,B),Super,Hyps,dom_res(PT)) :- !,check_is_subset(B,Super,Hyps,PT).
check_is_subset(range_subtraction(A,_),Super,Hyps,ran_sub(PT)) :- !,check_is_subset(A,Super,Hyps,PT).
check_is_subset(range_restriction(A,_),Super,Hyps,ran_res(PT)) :- !,check_is_subset(A,Super,Hyps,PT).
check_is_subset(set_subtraction(A,_),Super,Hyps,set_sub(PT)) :- !,check_is_subset(A,Super,Hyps,PT).
check_is_subset(value(avl_set(AVL)),B,Hyps,avl) :- !,check_subset_avl(B,AVL,Hyps).
check_is_subset('$'(ID),B,Hyps,eq(ID,PT)) :-
   (Operator = equal ; Operator = subset), % for subset_strict we also have subset in Hyps
   % TO DO: similar rule for B
   avl_fetch_binop_from_hyps('$'(ID),Operator,Hyps,S2,Hyps2),
   rewrite_local_loop_check(ID,check_is_subset,S2,Hyps2,Hyps3),
   check_is_subset(S2,B,Hyps3,PT),!.
check_is_subset('$'(ID),B,Hyps,eq(ID,PT)) :-
   get_type_from_hyps('$'(ID),Hyps,Set,Hyps2),
   extract_element_super_set_type(Set,Hyps2,S2),
   rewrite_local_loop_check(ID,check_is_subset,S2,Hyps2,Hyps3),
   check_is_subset(S2,B,Hyps3,PT),!.
check_is_subset(domain(Func),B,Hyps,domain(PT)) :-
   get_domain_or_superset(Func,Hyps,DomFunc,Hyps2),
   %rewrite_local_loop_check(domain(Func),check_is_subset,DomFunc,Hyps2,Hyps3),
   check_is_subset(DomFunc,B,Hyps2,PT),!.
check_is_subset(range(Func),B,Hyps,range(PT)) :-
   get_range_or_superset(Func,Hyps,RanFunc,Hyps2),
   %rewrite_local_loop_check(range(Func),check_is_subset,RanFunc,Hyps2,Hyps3),
   check_is_subset(RanFunc,B,Hyps2,PT),!.
check_is_subset(function(Func,_),B,Hyps,function_range(PT)) :- !,
   get_range_or_superset(Func,Hyps,RanFunc,Hyps2),  % f : _ +-> POW(Ran)  & Ran <: B =>  f(.) <: B
   subset_transitivity_rule(RanFunc,pow_subset(B),A2,B2), % extract pow_subset from Range
   check_is_subset(A2,B2,Hyps2,PT).
check_is_subset(image(Func,_),B,Hyps,image(PT)) :- % or B=range(Range)
   (B = range(FuncB),check_equal(Func,FuncB,Hyps,_) -> !, PT=range_of_same_func  % f[.] <: ran(f)
    ; get_range_or_superset(Func,Hyps,Range,Hyps2) -> !, check_is_subset(Range,B,Hyps2,PT)).
check_is_subset(A,B,Hyps,transitivity(PT)) :- subset_transitivity_rule(A,B,A2,B2),
   !, % unary subset rules like POW(A2) <: POW(B2)  if A2 <: B2
   check_is_subset(A2,B2,Hyps,PT).
check_is_subset(A,B,Hyps,transitivity(PT1,PT2)) :- subset_bin_transitivity_rule(A,B,A1,A2,B1,B2),
   !, % binary subset rules like A1+->B1 <: A2+->B2  if A1 <:B1 & A2 <: B2
   (check_is_subset(A1,B1,Hyps,PT1) -> check_is_subset(A2,B2,Hyps,PT2)).
check_is_subset(sorted_set_extension(List),B,Hyps,PT) :- !,
   check_is_subset(set_extension(List),B,Hyps,PT).
check_is_subset(set_extension(List),B,Hyps,set_extension) :-
   simplify_expr(B,Hyps,BS), % simplify expression once
   %portray_hyps(Hyps),nl,
   l_check_is_member(List,BS,Hyps).
check_is_subset(Sub,union(A,B),Hyps,sub_union(PT)) :- !,
  ( check_is_subset(Sub,A,Hyps,PT) -> true ; check_is_subset(Sub,B,Hyps,PT)).
% get_set_of_possible_values;  treat   sequence_extension
%check_is_subset(A,B,_,_) :- print(check_is_subset_failed(A,B)),nl,nl,fail.

l_check_is_member([],_,_).
l_check_is_member([El|T],B,Hyps) :- check_member_of_set(B,El,Hyps,_ProofTree), l_check_is_member(T,B,Hyps).

% extract set type of the elements of a set: x: POW(A)  ==> x<:A
extract_element_super_set_type(FuncType,Hyps,cartesian_product(A,B)) :-
   get_possible_domain_of_func_or_rel_type(FuncType,Hyps,A,_),!,
   get_possible_range_of_func_or_rel_type_direct(FuncType,B,_).
extract_element_super_set_type(fin_subset(A),_,A).
extract_element_super_set_type(fin1_subset(A),_,A).
extract_element_super_set_type(pow_subset(A),_,A).
extract_element_super_set_type(pow1_subset(A),_,A).

% simple not member of set check
check_not_member_of_set(Set,_,Hyps,empty_set) :- check_equal(Set,empty_set,Hyps,_),!.
check_not_member_of_set(Set,El,Hyps,hyp) :-
   avl_fetch_from_hyps(not_member(El,Set),Hyps),!. % hyp
check_not_member_of_set(if_then_else(_,A,B),El,Hyps,if_then_else(PTA,PTB)) :-
   (check_not_member_of_set(A,El,Hyps,PTA) -> check_not_member_of_set(B,El,Hyps,PTB)),!.
check_not_member_of_set(intersection(A,B),El,Hyps,inter(PT)) :-
   (check_not_member_of_set(A,El,Hyps,PT) -> true ; check_not_member_of_set(B,El,Hyps,PT)),!.
check_not_member_of_set(set_subtraction(A,B),El,Hyps,inter(PT)) :-
   (check_not_member_of_set(A,El,Hyps,PT) -> true ; check_member_of_set(B,El,Hyps,PT)),!.
check_not_member_of_set(union(A,B),El,Hyps,inter(PTA,PTB)) :-
   (check_not_member_of_set(A,El,Hyps,PTA) -> check_not_member_of_set(B,El,Hyps,PTB)),!.
check_not_member_of_set(overwrite(A,B),El,Hyps,overwrite(PTA,PTB)) :-
   (check_not_member_of_set(A,El,Hyps,PTA) -> check_not_member_of_set(B,El,Hyps,PTB)),!.
check_not_member_of_set('NATURAL1',El,Hyps,nat1) :-
   check_leq(El,0,Hyps).
check_not_member_of_set('NATURAL',El,Hyps,nat1) :-
   check_leq(El,-1,Hyps).
check_not_member_of_set(interval(From,To),El,Hyps,interval) :-
   (check_leq(El,minus(From,1),Hyps) -> true
     ; check_leq(add(To,1),El,Hyps) -> true). % TODO: or interval empty
check_not_member_of_set(A,El,Hyps,eq(ProofTree)) :-
   avl_fetch_worthwhile_equal_from_hyps(A,Hyps,Value,Hyps2),
   rewrite_local_loop_check(A,check_member_of_set,Value,Hyps2,Hyps3),
   check_not_member_of_set(Value,El,Hyps3,ProofTree).
% TO DO: process equalities, set_extension?, value(avl_set(AVL)), ...

% check_member_of_set(Set,Element,Hyps,ProofTree)
check_member_of_set(Set,_,Hyps,maximal_set) :- maximal_set(Set,Hyps), !.
check_member_of_set(Set,if_then_else(_,A,B),Hyps,if(P1,P2)) :- !, % if-then-else exprssion
   (check_member_of_set(Set,A,Hyps,P1) -> check_member_of_set(Set,B,Hyps,P2)).
check_member_of_set(Set,El,Hyps,hyp) :-
   % we could do avl_fetch_binop_from_hyps(El,member,Hyps,Set2,Hyps2), and check_subset(Set2,Set)
   avl_fetch_from_hyps(member(El,Set),Hyps),!. % hyp
   % TO DO: sometimes value(El) stored !
check_member_of_set(sorted_set_extension(List),El,Hyps,PT) :- !, % ordsets:ord_member(El,List),!.
   check_member_of_set(set_extension(List),El,Hyps,PT).
check_member_of_set(set_extension(List),El,Hyps,set_extension) :- member(El2,List),
    check_equal(El,El2,Hyps,_),!. % TO DO: avoid multiple equality rewriting of El for long lists ?
check_member_of_set(partial_function(T1,T2),El,Hyps,partial_function) :-
   (is_empty_set(El) -> true
    ; maximal_set(T1,Hyps), maximal_set(T2,Hyps),
      !,
      check_is_partial_function(El,Hyps) % Use special code for partial function check, typical WD PO
   ).
check_member_of_set(range(Func),El,Hyps,mem_range(PT)) :-
   check_member_of_range(El,Func,Hyps,PT),!. % check before function application below, can do symbolic range check
check_member_of_set(image(Func,set_extension([S1|_])),El,Hyps,mem_range_for_image(PT0,PT1)) :-
   % El:ran(F) & S /\ dom(F) \= {} => El:F[S]
   check_member_of_set(range(Func),El,Hyps,PT0),!,
   check_member_of_set(domain(Func),S1,Hyps,PT1). % TO DO: more general check S /\ dom(F) \= {}
check_member_of_set(A,ElFunc,Hyps,typing_membership(PT)) :-
   get_type_from_hyps(ElFunc,Hyps,Range,Hyps2), % !, % moving cut later proves on additional PO for test 2039
   % e.g. f(.) : A  if  ran(f) <: Range & Range <: A
   %rewrite_local_loop_check(A,check_member_of_set,Range,Hyps2,Hyps3),
   check_is_subset(Range,A,Hyps2,PT),!.
check_member_of_set(A,El,Hyps,eq(ProofTree)) :-
   avl_fetch_worthwhile_equal_from_hyps(A,Hyps,Value,Hyps2),
   rewrite_local_loop_check(A,check_member_of_set,Value,Hyps2,Hyps3),
   check_member_of_set(Value,El,Hyps3,ProofTree).
check_member_of_set(interval(L,U),El,Hyps,interval(PT)) :- !, check_in_interval(El,L,U,Hyps,PT).
check_member_of_set('NATURAL1',El,Hyps,nat1(PT)) :- !, check_subset_interval('NATURAL1',El,El,Hyps,PT).
check_member_of_set('NATURAL',El,Hyps,nat(PT)) :- !, check_subset_interval('NATURAL',El,El,Hyps,PT).
check_member_of_set(union(A,B),El,Hyps,union(PTA,PTB)) :- !,
   (check_member_of_set(A,El,Hyps,PTA) -> true ; check_member_of_set(B,El,Hyps,PTB)).
check_member_of_set(intersection(A,B),El,Hyps,intersection(PTA,PTB)) :- !,
   (check_member_of_set(A,El,Hyps,PTA) -> check_member_of_set(B,El,Hyps,PTB)).
check_member_of_set(set_subtraction(A,B),El,Hyps,set_subtraction(PTA,PTB)) :- !,
   (check_member_of_set(A,El,Hyps,PTA) -> check_not_member_of_set(B,El,Hyps,PTB)).
check_member_of_set(pow_subset(T1),El,Hyps,pow(PT)) :- !,
   check_is_subset(El,T1,Hyps,PT).
check_member_of_set(fin_subset(T1),El,Hyps,fin(PT1,PT2)) :- !,
   check_is_subset(El,T1,Hyps,PT1),!,
   check_finite(El,Hyps,PT2).
check_member_of_set(pow1_subset(T1),El,Hyps,pow1(PT)) :- !,
   check_not_empty_set(El,Hyps),!,
   check_is_subset(El,T1,Hyps,PT).
check_member_of_set(fin1_subset(T1),El,Hyps,fin1(PT1,PT2)) :- !,
   check_not_empty_set(El,Hyps),!,
   check_is_subset(El,T1,Hyps,PT1),!,
   check_finite(El,Hyps,PT2).
check_member_of_set(seq(T1),El,Hyps,seq(PT)) :- !,
   check_is_sequence(El,Hyps),
   check_is_subset(range(El),T1,Hyps,PT).
check_member_of_set(seq1(T1),El,Hyps,seq1(PT)) :- !,
   check_is_non_empty_sequence(El,Hyps),
   check_is_subset(range(El),T1,Hyps,PT).
check_member_of_set(cartesian_product(T1,T2),couple(El1,El2),Hyps,cart(PT1,PT2)) :- !,
   check_member_of_set(T1,El1,Hyps,PT1),
   check_member_of_set(T2,El2,Hyps,PT2).
check_member_of_set(value(avl_set(AVL)),El,Hyps,PT) :-
    (avl_can_fetch(El,BVal) -> !,PT=avl_fetch(El),avl_fetch(BVal,AVL)
      ; avl_is_interval(AVL,Min,Max) ->  !, PT=avl_interval(PT2),
            % useful is El is not a number, but e.g. an arithmetic expression
            % print(avl_interval(Min,Max,El)),nl,
            check_integer(El,check_member_of_set_avl_interval),
            check_in_interval(El,Min,Max,Hyps,PT2)
     ).
check_member_of_set(A,El,Hyps,rewrite(PT)) :-  rewrite_set_expression_exact(A,Hyps,A2,Hyps2),!,
   check_member_of_set(A2,El,Hyps2,PT).
check_member_of_set(domain(Func),Index,Hyps,mem_domain(PT)) :-
   check_member_of_domain(Index,Func,Hyps,PT),!.
check_member_of_set(Set,X,Hyps,value_set(PT)) :- try_get_set_of_possible_values(X,Hyps,XSet,Hyps2),
   check_is_subset(XSet,Set,Hyps2,PT),!.
%check_member_of_set(Set,X,Hyps,eq(PT)) :- Set = '$'(_),
%   avl_fetch_equal_from_hyps(Set,Hyps,Set2,Hyps2),  % maybe perform direct rewrite ancestor cycle check here
%   check_member_of_set(Set2,X,Hyps2,PT),!.
check_member_of_set(Set,X,Hyps,trans(PT)) :-
   avl_fetch_binop_from_hyps(Set,superset,Hyps,SubSet,Hyps2), % X:B & B <: A =>  X:A
   check_member_of_set(SubSet,X,Hyps2,PT),!.
check_member_of_set(Set2,max(Set1),Hyps,mem_max(PT)) :- !, check_is_subset(Set1,Set2,Hyps,PT).
check_member_of_set(Set2,min(Set1),Hyps,mem_min(PT)) :- !, check_is_subset(Set1,Set2,Hyps,PT).
%check_member_of_set(A,B,_H,_ProofTree) :- print(check_member_of_set_failed(A,B)),nl,fail.


:- use_module(probsrc(kernel_reals),[construct_real/2]).
% check if we can fetch an expression as a B value (second arg) in an AVL set
avl_can_fetch(El,Res) :- number(El),!, Res=int(El).
avl_can_fetch(boolean_true,pred_true).
avl_can_fetch(boolean_false,pred_false).
avl_can_fetch(real(Atom),R) :- construct_real(Atom,R).
avl_can_fetch(string(S),string(S)) :- ground(S).
avl_can_fetch(couple(A,B),(VA,VB)) :- avl_can_fetch(A,VA), avl_can_fetch(B,VB).

check_member_of_domain(El,reverse(Func2),Hyps,reverse(PT)) :- !,check_member_of_range(El,Func2,Hyps,PT).
check_member_of_domain(Index,Func,Hyps,size_in_dom_seq) :-  % x:seq1(T) => size(x) : dom(x)
   index_in_non_empty_sequence(Index,Func,Hyps),
   check_is_non_empty_sequence(Func,Hyps),!.
% TO DO: f~(x) : dom(f)  ??
check_member_of_domain(El,union(A,B),Hyps,dom_of_union(PT)) :-
    check_member_of_union(domain(A),domain(B),El,Hyps,PT).
check_member_of_domain(El,overwrite(A,B),Hyps,dom_of_overwrite(PT)) :-
    check_member_of_union(domain(A),domain(B),El,Hyps,PT).
check_member_of_domain(El,Func,Hyps,dom_of_subset(PT)) :- % Func2 <: Func & El:dom(Func2) => El:dom(Func)
   % counter part of rule with superset for check_member_of_set
   (Op = equal ; Op = superset),
   avl_fetch_binop_from_hyps(Func,Op,Hyps,Func2,Hyps1),
   rewrite_local_loop_check(Func,check_member_of_domain,Func2,Hyps1,Hyps2),
   check_member_of_set(domain(Func2),El,Hyps2,PT).
check_member_of_domain(El,comprehension_set(IDS,Body),Hyps,dom_of_lambda(PTs)) :-
   get_lambda_args_and_body(IDS,Body,Args,RestBodyList),
   %nl,print(lambda(Args,El,RestBodyList)),nl,
   generate_funapp_binding(Args,El,Subst),
   l_rename_and_prove_goals(RestBodyList,Subst,Hyps,PTs).

% we could do intersection, subtraction

% check if an element is an element of a union of two sets
check_member_of_union(Set1,_,El,Hyps,PT) :- check_member_of_set(Set1,El,Hyps,PT),!.
check_member_of_union(_,Set2,El,Hyps,PT) :- check_member_of_set(Set2,El,Hyps,PT),!.
check_member_of_union(Set1,Set2,El,Hyps,union(PT1,PT2)) :-
   % x : A \/ B   &  A <: S1 & B <: S2   =>  x : S1 \/ S2
   avl_fetch_mem_from_hyps(El,Hyps,union(A,B),Hyps2), % TO DO: other conditions ?
   (check_is_subset(A,Set1,Hyps2,PT1) -> check_is_subset(B,Set2,Hyps2,PT2)
    ; check_is_subset(A,Set2,Hyps2,PT1)  -> check_is_subset(B,Set1,Hyps2,PT2)).

generate_funapp_binding(['$'(X)],El,[X/El]).
generate_funapp_binding(['$'(X),'$'(Y)],couple(El1,El2),[X/El1,Y/El2]).
generate_funapp_binding(['$'(X),'$'(Y),'$'(Z)],couple(couple(El1,El2),El3),[X/El1,Y/El2,Z/El3]).
% TO DO: create substitution for more arguments and other parameters


check_member_of_range(El,reverse(Func2),Hyps,reverse(PT)) :- !,check_member_of_domain(El,Func2,Hyps,PT).
check_member_of_range('$'(ID),Func2,Hyps,PT) :-
   avl_fetch_worthwhile_equal_from_hyps('$'(ID),Hyps,Value,Hyps2),
   check_member_of_range(Value,Func2,Hyps2,PT).
check_member_of_range(function(Func1,_),Func2,Hyps,func_app_in_range) :- % f(.) : ran(f)
   check_equal(Func1,Func2,Hyps,_).

avl_is_interval(AVL,Min,Max) :-
    avl_min(AVL,int(Min)),avl_max(AVL,int(Max)),
    avl_size(AVL,Size), Size is 1+Max-Min.

% subset transitivity for unary operators:
subset_transitivity_rule(pow_subset(A),pow_subset(B),A,B).

subset_transitivity_rule(pow1_subset(A),pow1_subset(B),A,B).
subset_transitivity_rule(pow1_subset(A),pow_subset(B),A,B).

subset_transitivity_rule(fin_subset(A),fin_subset(B),A,B).
subset_transitivity_rule(fin_subset(A),pow_subset(B),A,B).

subset_transitivity_rule(fin1_subset(A),fin1_subset(B),A,B).
subset_transitivity_rule(fin1_subset(A),fin_subset(B),A,B).
subset_transitivity_rule(fin1_subset(A),pow1_subset(B),A,B).
subset_transitivity_rule(fin1_subset(A),pow_subset(B),A,B).

subset_transitivity_rule(seq(A),seq(B),A,B).
subset_transitivity_rule(seq(A),partial_function(typeset,B),A,B).

subset_transitivity_rule(seq1(A),seq1(B),A,B).
subset_transitivity_rule(seq1(A),seq(B),A,B).
subset_transitivity_rule(seq1(A),partial_function(typeset,B),A,B).

subset_transitivity_rule(iseq(A),iseq(B),A,B).
subset_transitivity_rule(iseq(A),seq(B),A,B).
subset_transitivity_rule(iseq(A),partial_function(typeset,B),A,B).

subset_transitivity_rule(iseq1(A),iseq1(B),A,B).
subset_transitivity_rule(iseq1(A),iseq(B),A,B).
subset_transitivity_rule(iseq1(A),seq1(B),A,B).
subset_transitivity_rule(iseq1(A),seq(B),A,B).
subset_transitivity_rule(iseq1(A),partial_function(typeset,B),A,B).

subset_transitivity_rule(perm(A),perm(B),A,B).
subset_transitivity_rule(perm(A),iseq(B),A,B).
subset_transitivity_rule(perm(A),seq(B),A,B).
subset_transitivity_rule(perm(A),partial_function(typeset,B),A,B).


subset_transitivity_rule(range(A),domain(reverse(B)),A,B).
subset_transitivity_rule(range(A),range(B),A,B).
subset_transitivity_rule(range(reverse(A)),domain(B),A,B).
subset_transitivity_rule(domain(reverse(A)),range(B),A,B).
subset_transitivity_rule(domain(A),domain(B),A,B).  % dom(A) <: dom(B)  if A <:B
subset_transitivity_rule(domain(A),range(reverse(B)),A,B).
subset_transitivity_rule(reverse(A),reverse(B),A,B).
subset_transitivity_rule(rev(A),rev(B),A,B).
subset_transitivity_rule(identity(A),identity(B),A,B).

% TO DO: add rules for binary operators
subset_bin_transitivity_rule(partial_function(A1,A2),partial_function(B1,B2),A1,A2,B1,B2).
subset_bin_transitivity_rule(partial_injection(A1,A2),partial_function(B1,B2),A1,A2,B1,B2).
subset_bin_transitivity_rule(partial_surjection(A1,A2),partial_function(B1,B2),A1,A2,B1,B2).
subset_bin_transitivity_rule(partial_bijection(A1,A2),partial_function(B1,B2),A1,A2,B1,B2).
subset_bin_transitivity_rule(total_function(A1,A2),partial_function(B1,B2),A1,A2,B1,B2).
subset_bin_transitivity_rule(total_injection(A1,A2),partial_function(B1,B2),A1,A2,B1,B2).
subset_bin_transitivity_rule(total_surjection(A1,A2),partial_function(B1,B2),A1,A2,B1,B2).
subset_bin_transitivity_rule(total_bijection(A1,A2),partial_function(B1,B2),A1,A2,B1,B2).
subset_bin_transitivity_rule(image(A1,A2),image(B1,B2),A1,A2,B1,B2). % A1[A2] <: B1[B2] if A1 <: B1 & A2 <: B2
subset_bin_transitivity_rule(domain_restriction(A1,A2),domain_restriction(B1,B2),A1,A2,B1,B2). % A1 <| A2 <: B1 <| B2 if A1 <: B1 & A2 <: B2
subset_bin_transitivity_rule(range_restriction(A1,A2),range_restriction(B1,B2),A1,A2,B1,B2).
subset_bin_transitivity_rule(domain_subtraction(A1,A2),domain_subtraction(B1,B2),B1,A2,A1,B2). % A1 <<| A2 <: B1 <<| B2 if B1 <: A1 & A2 <: B2
subset_bin_transitivity_rule(range_subtraction(A1,A2),range_subtraction(B1,B2),A1,B2,A2,B1). % A1 |>> A2 <: B1|>> B2 if A1 <: B1 & B2 <: A2
% TO DO: add more

% TO DO: instead of is_set_of_sequences_type
%subset_mixed_transitivity_rule(total_function(A1,A2),seq(B2),A2,B2) :- is_interval(A1).


check_in_interval(El,Min,Max,Hyps,PT) :- check_subset_interval(interval(Min,Max),El,El,Hyps,PT).

% check if an interval is a subset of the first argument
check_subset_interval(union(A,B),L1,U1,Hyps,union(PT)) :-  !,
   % TO DO: try and merge A,B : union(interval(1,10),set_extension([11]))
   (check_subset_interval(A,L1,U1,Hyps,PT) -> true ; check_subset_interval(B,L1,U1,Hyps,PT)).
check_subset_interval(sorted_set_extension(L),L1,U1,Hyps,PT) :-  !,
    check_subset_interval(set_extension(L),L1,U1,Hyps,PT).
check_subset_interval(set_extension(L),L1,U1,Hyps,set_extension(Nr)) :-  !,
   % TO DO: maybe merge L into an interval
   nth1(Nr,L,El), check_sub_intervals(L1,U1,El,El,Hyps),!.
check_subset_interval(intersection(A,B),L1,U1,Hyps,inter(PTA,PTB)) :-  !,
   % L1..U1 <: A /\ B  if L1..U1 <: A  & L1..U1 <: B
   (check_subset_interval(A,L1,U1,Hyps,PTA) -> check_subset_interval(B,L1,U1,Hyps,PTB)).
check_subset_interval(interval(L2,U2),L1,U1,Hyps,interval) :-
    !,check_sub_intervals(L1,U1,L2,U2,Hyps).
check_subset_interval('NATURAL',L1,_,Hyps,nat) :- !, check_leq(0,L1,Hyps).
check_subset_interval('NATURAL1',L1,_,Hyps,nat1) :- !, check_leq(1,L1,Hyps).
check_subset_interval(value(avl_set(A)),L1,U1,Hyps,avl(PT)) :- !,
   (number(L1), number(U1)
    -> PT=in(L1,U1),
       check_interval_in_custom_set(L1,U1,avl_set(A),no_wf_available)
     ; avl_min(A,int(L2)), avl_max(A,int(U2)), PT=min_max(L2,U2,PT2),
       check_subset_interval(interval(L2,U2),L1,U1,Hyps,PT2)
   ).
check_subset_interval(A,L1,U1,Hyps,rewrite(PT)) :-  rewrite_set_expression_exact(A,Hyps,A2,Hyps2),!,
   check_subset_interval(A2,L1,U1,Hyps2,PT).
check_subset_interval(domain(Expr),Low,Up,Hyps,dom_seq1) :- % a special rule when using SEQ(1) rather than first(SEQ)
   (check_leq(1,Low,Hyps), check_leq(Up,size(Expr),Hyps)  % 1..size(s) <: dom(s)
      -> check_is_sequence(Expr,Hyps)
    ; index_in_non_empty_sequence(Low,Expr,Hyps),
      index_in_non_empty_sequence(Up,Expr,Hyps)  % 1..1 or size(s)..size(s)  <: dom(s)  if s:seq1(.)
      -> check_is_non_empty_sequence(Expr,Hyps)
    ).
check_subset_interval(range(reverse(Expr)),Low,Up,Hyps,PT) :- !,
   check_subset_interval(domain(Expr),Low,Up,Hyps,PT).
check_subset_interval(A,Low,Up,Hyps,eq(PT)) :-
   avl_fetch_worthwhile_equal_from_hyps(A,Hyps,A2,Hyps2),
   rewrite_local_loop_check(A,check_subset_interval,A2,Hyps2,Hyps3),
   check_subset_interval(A2,Low,Up,Hyps3,PT).
%check_subset_interval(A,L1,U1,_,_) :- print(check_subset_interval_failed(A,L1,U1)),nl,fail.

% s:seq1(.) => 1:dom(s) & size(s):dom(s)
index_in_non_empty_sequence(1,_,_).
index_in_non_empty_sequence(card(E),E,_).
index_in_non_empty_sequence(size(E),E,_).
index_in_non_empty_sequence('$'(X),E,Hyps) :-
   avl_fetch_equal_from_hyps('$'(X),Hyps,Y,Hyps2),
   rewrite_local_loop_check(X,index_in_non_empty_sequence,Y,Hyps2,Hyps3),
   index_in_non_empty_sequence(Y,E,Hyps3).

% check if L1..U1 <: L2..U2
check_sub_intervals(L1,L1,L2,U2,Hyps) :- (L1=L2 ; L1=U2),!,
    check_not_empty_set(interval(L2,U2),Hyps).
check_sub_intervals(L1,U1,L2,U2,Hyps) :- % L1..U1 <: L2..U2 if L2 <= L1 & U1 <= U2
    check_leq(L2,L1,Hyps),!,
    check_leq(U1,U2,Hyps).



% some exact rewrite steps
rewrite_set_expression_exact(domain(A),Hyps,Res,Hyps2) :-  compute_exact_domain(A,Hyps,Dom,Hyps2),!,
   %print(rewrote(domain(A))),nl, print(Dom),nl,
   (A='$'(ID) -> not_occurs(Dom,ID) ; true), % prevent silly rewrites
   Res=Dom.
rewrite_set_expression_exact(range(A),Hyps,Res,Hyps2) :- compute_exact_range(A,Hyps,Ran,Hyps2),!,
   %print(rewrote(range(A))),nl, print(Ran),nl,
   (A='$'(ID) -> not_occurs(Ran,ID) ; true), % prevent silly rewrites
   Res=Ran.
rewrite_set_expression_exact(intersection(A,B),Hyps,Res,Hyps) :-
   (is_empty_set(A) -> Res=empty_set ; is_empty_set(B) -> Res=empty_set).
rewrite_set_expression_exact(set_subtraction(A,B),Hyps,Res,Hyps) :-
   (is_empty_set(A) -> Res=empty_set ; is_empty_set(B) -> Res=A).
rewrite_set_expression_exact(union(A,B),Hyps,Res,Hyps) :-
    merge_set_extensions(union(A,B),List,[]), construct_set_extension(List,Hyps,Res).
rewrite_set_expression_exact(value(closure(P,T,B)),Hyps,Res,Hyps) :- nonvar(P),
   is_interval_closure(P,T,B,LOW,UP), number(LOW),number(UP),!,
   Res = interval(LOW,UP).
rewrite_set_expression_exact(assertion_expression(_,_,Expr),Hyps,Expr,Hyps). % TO DO: add Predicate to Hyps ?
% Note one can have equalities like f = f~~  (in FunLawsWithLambda.mch); hence important to pass Hyps for cycle detection

is_empty_set(empty_set).
is_empty_set(empty_sequence).
is_empty_set(value(X)) :- X==[].

merge_set_extensions(empty_set) --> [].
merge_set_extensions(empty_sequence) --> [].
merge_set_extensions(set_extension(L)) --> L.
merge_set_extensions(sorted_set_extension(L)) --> L.
merge_set_extensions(union(A,B)) --> merge_set_extensions(A), merge_set_extensions(B).

% check if AVL is a subset of the first argument
check_subset_avl(union(A,B),AVL1,Hyps) :- !,  % TO DO: try and merge A,B
   (check_subset_avl(A,AVL1,Hyps) -> true ; check_subset_avl(B,AVL1,Hyps)).
check_subset_avl(intersection(A,B),AVL1,Hyps) :- !, % AVL <: A /\ B  if AVL <: A  & AVL <: B
   (check_subset_avl(A,AVL1,Hyps) -> check_subset_avl(B,AVL1,Hyps)).
check_subset_avl(interval(L2,U2),AVL,_) :- number(L2),number(U2),!,
    check_avl_in_interval(AVL,L2,U2).
check_subset_avl(value(avl_set(AVL2)),AVL1,_) :-  !, check_avl_subset(AVL1,AVL2).
check_subset_avl(seq(MAX),AVL,Hyps) :- maximal_set(MAX,Hyps), !, is_avl_set_of_sequences(AVL,seq).
check_subset_avl(seq1(MAX),AVL,Hyps) :- maximal_set(MAX,Hyps), !, is_avl_set_of_sequences(AVL,seq1).
check_subset_avl(seq(seq(MAX)),AVL,Hyps) :-  maximal_set(MAX,Hyps),
    % comes from general concat
    custom_explicit_sets:is_one_element_avl(AVL,Element), % usually one value from try_get_set_of_possible_values
    is_sequence(Element,seq),
    expand_custom_set_to_list(Element,ListOfSeqs),
    maplist(is_subsequence,ListOfSeqs).
check_subset_avl(A,AVL,Hyps) :-  rewrite_set_expression_exact(A,Hyps,A2,Hyps2),!,
   check_subset_avl(A2,AVL,Hyps2).
check_subset_avl(A,AVL,Hyps) :-
   avl_fetch_worthwhile_equal_from_hyps(A,Hyps,A2,Hyps2),
   rewrite_local_loop_check(A,check_subset_avl,A2,Hyps2,Hyps3),
   check_subset_avl(A2,AVL,Hyps3).
%check_subset_avl(A,AVL,_) :- print(check_subset_avl_failed(A,AVL)),nl,fail.

is_subsequence((int(_Index),Sequence)) :- is_sequence(Sequence,seq).

is_sequence(avl_set(SeqAVL),_) :- is_avl_sequence(SeqAVL).
is_sequence([],seq).

% check if all elements of the AVL are sequences
is_avl_set_of_sequences(AVL,SeqType) :- avl_height(AVL,Height), Height<7,
   expand_custom_set_to_list(avl_set(AVL),ListOfSeqs),
   l_is_sequence(ListOfSeqs,SeqType).
l_is_sequence([],_).
l_is_sequence([S1|T],SeqType) :- is_sequence(S1,SeqType), l_is_sequence(T,SeqType).


:- use_module(probsrc(b_global_sets),[b_global_set/1]).
maximal_set('INTEGER',_). % integer_set('INTEGER') ?
maximal_set(real_set,_).
maximal_set(string_set,_).
maximal_set(bool_set,_).
maximal_set('typeset',_).
maximal_set(cartesian_product(A,B),Hyps) :- maximal_set(A,Hyps), maximal_set(B,Hyps).
maximal_set(relations(A,B),Hyps) :- maximal_set(A,Hyps), maximal_set(B,Hyps).
maximal_set(pow_subset(A),Hyps) :- maximal_set(A,Hyps).
maximal_set('$'(ID),Hyps) :- is_global_set_id(ID,Hyps).
maximal_set(value(avl_set(AVL)),_) :-
    quick_definitely_maximal_set_avl(AVL).
maximal_set(set_extension(A),Hyps) :- maximal_set_extension(A,Hyps).
% sorted_set_extension is never maximal
maximal_set(comprehension_set(_,truth),_).
%maximal_set(X,_) :- print(max_fail(X)),nl,fail.

construct_set_extension([],_,Res) :- !, Res=empty_set.
construct_set_extension(L,Hyps,Res) :- maximal_set_extension(L,Hyps),!, Res='typeset'.
construct_set_extension(L,_,sorted_set_extension(SL)) :-
    %length(L,Len), format('Construct set_extension ~w~n',[Len]),
    sort(L,SL).

maximal_set_extension([boolean_true|T],_) :- !, member(boolean_false,T).
maximal_set_extension([boolean_false|T],_) :- !, member(boolean_true,T).
maximal_set_extension(['$'(ID)|T],Hyps) :-
   is_global_constant_id(ID,Hyps),
   sort(['$'(ID)|T],Sorted),
   maplist(is_glob_const_id(Hyps),Sorted), % all elements are global constants
   lookup_global_constant(ID,fd(_,GlobalSet)),
   enumerated_set(GlobalSet),b_global_set_cardinality(GlobalSet,Size),
   length(Sorted,Size).
%maximal_set_extension(X,_) :- print(maximal_failed(X)),nl,fail.

is_glob_const_id(Hyps,'$'(ID)) :- is_global_constant_id(ID,Hyps).


is_global_set_id(ID,Hyps) :-
    b_global_set(ID),
    \+ is_hyp_var(ID,Hyps). % global enumerated set visible

% often called with 0 or 1 in first position
check_leq(I,I,_) :- !.
check_leq(if_then_else(_,A1,A2),B,Hyp) :- !, % TODO?: treat if-then-else for B
   (check_leq(A1,B,Hyp) -> check_leq(A2,B,Hyp)).
check_leq(N1,N2,_) :- number(N1), number(N2), !, N1 =< N2.
check_leq(N1,N2,hyp_rec(AVL,_)) :-
   (avl_fetch(less_equal(N1,N2),AVL)
    -> true
    ; avl_fetch(equal(N1,N2),AVL)),!.
check_leq(min(List),N2,Hyps) :- !,
   member(N1,List), check_leq(N1,N2,Hyps),!.
check_leq(N1,N2,Hyps) :-
   rewrite_integer(N2,Hyps,RN2,Hyps2),!,
   check_leq(N1,RN2,Hyps2).
check_leq(add(N1,1),N2,Hyps) :-
    check_not_equal(N1,N2,Hyps),
    !, % N1+1 <= N2  if N1 <= N2 & N1 \= N2  ; happens quite often in array traversals
    check_leq(N1,N2,Hyps).
check_leq(N1,minus(N2,1),Hyps) :- % variation of rule above
    check_not_equal(N1,N2,Hyps),
    !, % N1 <= N2-1  if N1 <= N2 & N1 \= N2  ; happens in array traversals
    check_leq(N1,N2,Hyps).
check_leq(Nr,X,Hyps) :-
    \+ number(X),
    try_get_set_of_possible_values(X,Hyps,SetX,Hyps2),
    check_all_values_geq_val(SetX,Nr,Hyps2),!.
check_leq(Nr,X,Hyps) :- number(Nr), !,
   check_leq_nr(Nr,X,Hyps).
check_leq(N1,N2,Hyps) :- rewrite_integer(N1,Hyps,RN1,Hyps2),!,
   check_leq(RN1,N2,Hyps2).
check_leq(Add,N2,Hyps) :-  % A+N1 <= N2   <=>  A <= N2-N1
   number(N2),
   add_with_number(Add,A,N1),!,
   N21 is N2-N1,
   check_leq(A,N21,Hyps).
check_leq(Mul,N2,Hyps) :-  % A*N1 <= N2   if  A <= N2/N1 if N1>0 and N2 mod N1=0
   number(N2),
   mul_with_number(Mul,A,N1),
   % symmetrical case to check_leq_nr(N1,Mul,Hyps), with N1=-N2
   !,
   (  N1=0 -> check_leq(0,N2,Hyps)
    ; N1>0 -> N21 is N2 div N1, % A <= 1.5  means we have to have A <= 1;
               % A <= -1.5  means we have to have A <= -2  -3 div 2 =:= -2
              check_leq(A,N21,Hyps)
    ;         cdiv(N2,N1,N21),  % A >= 1.5  means we have to have A >= 2 ; cdiv
              check_leq(N21,A,Hyps)
   ).
check_leq(div(A,N1),N2,Hyps) :-  % A/N1 <= N2   <=>  A <= (N1+1)*N2-1
   number(N1),number(N2), N1>0,
   !,
   N12 is (N2+1)*N1-1,
   check_leq(A,N12,Hyps).
check_leq(div(A1,N1),A2,Hyps) :- number(N1), N1>0,  % A/N1 <= A  if  N1>0 & A>=0
   check_equal(A1,A2,Hyps,Hyps1),!,
   check_leq(0,A1,Hyps1).
check_leq(modulo(A1,A2),B,Hyps) :-
    \+ z_or_tla_minor_mode, % args to mod must be non-negative, modulo is between 0..A2-1
   ((number(A2),A21 is A2-1 -> check_leq(A21,B,Hyps)
     ; B=minus(B1,1) -> check_leq(A2,B1,Hyps)
     ; check_leq(A2,B,Hyps)
     ) -> true
     ; check_leq(A1,B,Hyps)).
% TO DO: modulo as RHS
check_leq(unary_minus(A),unary_minus(B),Hyps) :- !, % -A <= -B   --->  A >= B
    check_leq(B,A,Hyps).
check_leq(X,Nr,Hyps) :- \+ number(X),
   try_get_set_of_possible_values(X,Hyps,SetX,Hyps2),
   check_all_values_leq_val(SetX,Nr,Hyps2),!. % cut here; get set of possible values can give multiple solutions
check_leq(Minus,N2,Hyps) :- minus_with_number(Minus,N1,Nr),
    Nr >= 0,!, % N1-Nr <= N2  if N1 <= N2
    % Both N1 and N2 are usually not numbers here
    check_leq(N1,N2,Hyps).
check_leq(N1,Add,Hyps) :-
    add_with_number(Add,N2,Nr),Nr >= 0,!, % N1 <= N2+Nr  if N1 <= N2
    % Both N1 and N2 are usually not numbers here
    check_leq(N1,N2,Hyps).
check_leq(add(A,B),E,Hyps) :- decompose_floor(E,X,Y), % e.g. divide a number E by 2
    check_leq(A,X,Hyps), % TO DO: other combinations like A <= 0, B <= Nr; or we could try_get_set_of_possible_values
    check_leq(B,Y,Hyps).
check_leq('$'(X),N2,Hyps) :-
   avl_fetch_binop_from_hyps('$'(X),less_equal,Hyps,Y,Hyps2),
   (number(N2),avl_fetch_not_equal('$'(X),Y,Hyps) % as we know X and Y we can use regular avl_fetch
    -> N21 is N2+1 % we have X<Y in the Hypotheses, we just require that Y <= N2+1
     ; N21=N2),
   check_leq(Y,N21,Hyps2).
check_leq(Nr,'$'(X),Hyps) :-
   (  avl_fetch_equal_from_hyps('$'(X),Hyps,Y,Hyps2), check_leq(Nr,Y,Hyps2) -> true
   % ; avl_fetch_binop_from_hyps('$'(X),greater,Hyps,Y,Hyps2), N1 is Nr-1, check_leq(N1,Y,Hyps2) -> true
    ; avl_fetch_binop_from_hyps('$'(X),greater_equal,Hyps,Y,Hyps2),
      % note: Nr is not a number, hence probably not useful to check not_equal in Hyps, as we cannot compute Nr-1
      check_leq(Nr,Y,Hyps2)
     -> true
   ),
   !.
%check_leq(A,B,_H) :- print(check_leq_failed(A,B)),nl, portray_hyps(_H),nl,fail.

% decompose an expression E into A and B so that A+B <= E
decompose_floor(Nr,A,B) :- number(Nr),!,
    A is Nr div 2, B=A. % -11 div 2 -> -6, -1 div 2 = -1, 11 div 2 = 5
decompose_floor(add(A,B),A,B). % TO DO: we could try other order
decompose_floor(Mul,A,A) :- mul_with_number(Mul,A,Nr), Nr>=2. % we could divide Nr by 2

% ceiling division utility
cdiv(N1,N2,Res) :-
   (N1 mod N2 =:= 0 -> Res is N1//N2
    ; Res is (N1 div N2)+1).

% Number <= Expression
check_leq_nr(N1,Add,Hyps) :-  % N1 <= A+N2   <=>  N1-N2 <= A
   add_with_number(Add,A,N2), !,
   N12 is N1-N2,
   check_leq(N12,A,Hyps).
check_leq_nr(Nr,add(N1,N2),Hyps) :- !, % 0 <= A+B  if 0 <= A & 0 <= B
    % Both N1 and N2 are usually not numbers here
    cdiv(Nr,2,Nr2), % Note: cdiv(-3,2) = 1, cdiv(3,2)=2
    check_leq(Nr2,N1,Hyps),
    check_leq(Nr2,N2,Hyps).
check_leq_nr(N1,minus(N2,B),Hyps) :-  % N1 <= N2-B   <=>  B <= N2-N1
   number(N2), !,
   N21 is N2-N1,
   check_leq(B,N21,Hyps).
check_leq_nr(N1,Mul,Hyps) :-  % N1 <= A*N2   if  N1/N2 <= A  and N2>0
   mul_with_number(Mul,A,N2),
   !,
   (  N2=0 -> check_leq(N1,0,Hyps)
    ; N2>0 -> cdiv(N1,N2,N12), % cdiv
               % if 1.5 <= A --> 2 <= A ; if -1.5 <= A  --> -1 <= A
              check_leq(N12,A,Hyps)
    ;          N12 is N1 div N2,
               % if A <= 1.5 --> A <= 1 ; if -1.5 <= A  --> -1 <= A
               % A <= -1.5  means we have to have A <= -2  -3 div 2 =:= -2
              check_leq(A,N12,Hyps)
    ).
check_leq_nr(0,multiplication(A,B),Hyps) :- !, % 0 <= A*B  if A and B have same parity
   (check_leq(0,A,Hyps) -> check_leq(0,B,Hyps)
    ; check_leq(A,0,Hyps) -> check_leq(B,0,Hyps)).
check_leq_nr(N1,div(A,N2),Hyps) :-  % N1 <= A/N2   <=>  N1*N2 <= A
   number(N2), N2>0,
   !,
   N12 is N1*N2,
   check_leq(N12,A,Hyps).
check_leq_nr(0,div(A,B),Hyps) :- !, % 0 <= A/B  if A and B have same parity
   (check_leq(0,A,Hyps) -> check_leq(0,B,Hyps)  % B \= 0 checked by other WD condition
    ; check_leq(A,0,Hyps) -> check_leq(B,0,Hyps)). % ditto
check_leq_nr(Nr,'$'(X),Hyps) :-
   (  avl_fetch_equal_from_hyps('$'(X),Hyps,Y,Hyps2), check_leq(Nr,Y,Hyps2) -> true
   % ; avl_fetch_binop_from_hyps('$'(X),greater,Hyps,Y,Hyps2), N1 is Nr-1, check_leq(N1,Y,Hyps2) -> true
    ; avl_fetch_binop_from_hyps('$'(X),greater_equal,Hyps,Y,Hyps2),
      (avl_fetch_not_equal('$'(X),Y,Hyps2) % we have X < Y => sufficient to prove N-1 <= Y
        -> N1 is Nr-1, check_leq(N1,Y,Hyps2)
         ; check_leq(Nr,Y,Hyps2)
      )
   ),
   !.
check_leq_nr(Nr,modulo(A,B),Hyps) :- \+ z_or_tla_minor_mode, % A and B must be non-negative, modulo is between 0..B-1
   (Nr =< 0 -> true % modulo always positive or 0
    ; check_leq_nr(Nr,A,Hyps),  % Nr <= A
      check_leq(B,A,Hyps)). % and B <= A so that modulo does not take effect
check_leq_nr(Nr,size(Seq),Hyps) :- check_leq_nr_size(Nr,Seq,Hyps).
check_leq_nr(1,power_of(A,_),Hyps) :- check_leq(1,A,Hyps). % Nr <= 1 <= x**y  if  x >= 1
check_leq_nr(Nr,power_of(A,_),Hyps) :- number(Nr), Nr =< 0,
    check_leq(0,A,Hyps). % 0 <= x**y  if  x >= 0
%check_leq_nr(A,B,_H) :- print(check_leq_nr_failed(A,B)),nl,fail.

:- use_module(probsrc(specfile),[z_or_tla_minor_mode/0]).


check_leq_nr_size(Nr,restrict_front(_,RestrN),Hyps) :- !, % X <= size( Seq /|\ N)  if X <= N  as WD condition implies N : 0..size(Seq)
     check_leq_nr(Nr,RestrN,Hyps).
check_leq_nr_size(1,Seq,Hyps) :- check_is_non_empty_sequence(Seq,Hyps).

add_with_number(add(A,B),X,Nr) :- (number(A) -> Nr=A, X=B ; number(B) -> Nr=B, X=A).
add_with_number(minus(A,B),A,Nr) :- number(B), Nr is -B.
mul_with_number(multiplication(A,B),X,Nr) :- (number(A) -> Nr=A, X=B ; number(B) -> Nr=B, X=A).
mul_with_number(unary_minus(A),A,Nr) :- Nr is -1.
minus_with_number(add(A,B),A,Nr) :- number(B), Nr is -B.
minus_with_number(minus(A,Nr),A,Nr) :- number(Nr).

%get_possible_values('$'(X),Hyps,SetX,Hyps2) :-
%    avl_fetch_binop_from_hyps('$'(X),member,Hyps,SetX,Hyps2).

% a few rewrite rules for integer expressions
% addition/multiplication is dealt with in other places (and is usually done symbolically)
rewrite_integer(size(Seq),Hyps,Size,Hyps2) :- % can happen for sequence POs, like restrict_front,tail
   compute_card_of_set(Seq,Hyps,Size,Hyps2),!.
rewrite_integer(card(Seq),Hyps,Size,Hyps2) :- !, rewrite_card_of_set(Seq,Hyps,Size,Hyps2).
rewrite_integer(assertion_expression(_,_,Expr),Hyps,Expr,Hyps). % TO DO: add Predicate to Hyps?
% the following may be done by ast_cleanup, but e.g., when applying functions no cleanup is run in l_rename_and_prove_goals
rewrite_integer(add(A,B),Hyps,Res,Hyps2) :- compute_integer(A,Hyps,A1,Hyps1), number(A1),
   compute_integer(B,Hyps1,B1,Hyps2), number(B1),
   Res is A1+B1.
rewrite_integer(unary_minus(A),Hyps,Res,Hyps2) :- compute_integer(A,Hyps,A1,Hyps2), number(A1),
   Res is -A1.
rewrite_integer(minus(A,B),Hyps,Res,Hyps2) :- compute_integer(A,Hyps,A1,Hyps1), number(A1),
   compute_integer(B,Hyps1,B1,Hyps2), number(B1),
   Res is A1-B1.
rewrite_integer(power_of(A,B),Hyps,Res,Hyps2) :- compute_integer(A,Hyps,A1,Hyps1), number(A1),
   compute_integer(B,Hyps1,B1,Hyps2), number(B1), B1 >=0,
    % check if not too large:
   (abs(A) < 2 -> true
    ; A=2 -> B1 =< 64
    ; A < 4294967296 -> B1 =< 2
    ; B1 =< 0
   ),
   Res is A1 ^ B1.
rewrite_integer(modulo(A,B),Hyps,Res,Hyps2) :- compute_integer(A,Hyps,A1,Hyps1),number(A1),
   A1 >= 0,
   compute_integer(B,Hyps1,B1,Hyps2), number(B1), B1 >0,
   Res is A mod B.
rewrite_integer(division(A,B),Hyps,Res,Hyps2) :- compute_integer(A,Hyps,A1,Hyps1), number(A1),
   compute_integer(B,Hyps1,B1,Hyps2), number(B1), B1 \= 0,
   Res is A // B. % Prolog division corresponds to B division


rewrite_card_of_set(Set,Hyps,Size,Hyps2) :-
   compute_card_of_set(Set,Hyps,Size,Hyps2),!.
rewrite_card_of_set(interval(1,Up),Hyps,Size,Hyps) :- !, % useful if Up is a symbolic expression
   Size=Up.
rewrite_card_of_set(Set,Hyps,Size,Hyps2) :- rewrite_set_expression_exact(Set,Hyps,S2,Hyps1),
   rewrite_card_of_set(S2,Hyps1,Size,Hyps2).

compute_integer(A,H,A,H) :- number(A),!.
compute_integer(A,H,Res,H2) :- rewrite_integer(A,H,Res,H2).

:- use_module(probsrc(b_global_sets),   [enumerated_set/1, b_global_set_cardinality/2]).
compute_card_of_set(empty_set,Hyps,0,Hyps).
compute_card_of_set(empty_sequence,Hyps,0,Hyps).
compute_card_of_set(bool_set,Hyps,2,Hyps).
compute_card_of_set(interval(L,U),Hyps,Size,Hyps) :- number(L), number(U), Size is U+1-L.
compute_card_of_set(value(Val),Hyps,Size,Hyps) :- get_set_val_size(Val,Size).
compute_card_of_set(sequence_extension(List),Hyps,Size,Hyps) :- length(List,Size).
compute_card_of_set(set_extension([_]),Hyps,Size,Hyps) :- Size=1. % to do check if all elements definitely different
compute_card_of_set(sorted_set_extension([_]),Hyps,Size,Hyps) :- Size=1. % ditto
compute_card_of_set(rev(A),Hyps,Size,Hyps2) :- !, compute_card_of_set(A,Hyps,Size,Hyps2).
compute_card_of_set(front(A),Hyps,Size,Hyps2) :- !, compute_card_of_set(tail(A),Hyps,Size,Hyps2).
compute_card_of_set(tail(A),Hyps,Size,Hyps2) :- !,
    compute_card_of_set(A,Hyps,Size1,Hyps2), number(Size1), Size1>0,
    Size is Size1-1.
compute_card_of_set(concat(A,B),Hyps,Size,Hyps2) :-
    compute_card_of_set(A,Hyps,SA,Hyps1),!,
    compute_card_of_set(B,Hyps1,SB,Hyps2),
    Size is SA+SB.
compute_card_of_set('$'(ID),Hyps,Size,Hyps) :- is_enumerated_set(ID,Hyps),
    !,
    b_global_set_cardinality(ID,Size).
compute_card_of_set('$'(ID),Hyps,Size,Hyps2) :-
   avl_fetch_equal_from_hyps('$'(ID),Hyps,X2,Hyps1),
   compute_card_of_set(X2,Hyps1,Size,Hyps2),!.
compute_card_of_set('$'(ID),Hyps,Size,Hyps4) :-  % e.g., f:1..10 --> BOOL   --> card(f) = 10
   avl_fetch_binop_from_hyps('$'(ID),member,Hyps,FunctionType,Hyps1),
   is_partial_function_type(FunctionType,Hyps1,Hyps2),
   get_exact_domain_of_func_or_rel_type(FunctionType,Hyps2,Dom,Hyps3),
   compute_card_of_set(Dom,Hyps3,Size,Hyps4).
%compute_card_of_set(S,_,_,_) :- print(card_fail(S)),nl,fail.

get_set_val_size([],0).
get_set_val_size(avl_set(AVL),Size) :- avl_size(AVL,Size).


check_all_values_geq_val(intersection(A,B),Nr,Hyps) :-
    (check_all_values_geq_val(A,Nr,Hyps) -> true ; check_all_values_geq_val(B,Nr,Hyps)).
check_all_values_geq_val(union(A,B),Nr,Hyps) :-
    (check_all_values_geq_val(A,Nr,Hyps) -> check_all_values_geq_val(B,Nr,Hyps)).
check_all_values_geq_val(set_subtraction(A,_),Nr,Hyps) :-
    check_all_values_geq_val(A,Nr,Hyps).
check_all_values_geq_val(interval(From,_),Nr,Hyps) :- check_leq(Nr,From,Hyps).
check_all_values_geq_val(value(avl_set(AVL)),Nr,Hyps) :- avl_min(AVL,int(Min)), check_leq(Nr,Min,Hyps).
check_all_values_geq_val('NATURAL',Nr,Hyps) :- check_leq(Nr,0,Hyps).
check_all_values_geq_val('NATURAL1',Nr,Hyps) :- check_leq(Nr,1,Hyps).
check_all_values_geq_val(domain(Func),Nr,Hyps) :-
   get_domain_or_superset(Func,Hyps,DomFunc,Hyps2),
   check_all_values_geq_val(DomFunc,Nr,Hyps2).
check_all_values_geq_val(range(Func),Nr,Hyps) :-
   get_range_or_superset(Func,Hyps,RanFunc,Hyps2),
   check_all_values_geq_val(RanFunc,Nr,Hyps2).
check_all_values_geq_val(sorted_set_extension(L),Nr,Hyps) :- !, check_all_values_geq_val(set_extension(L),Nr,Hyps).
check_all_values_geq_val(set_extension(L),Nr,Hyps) :-
   (member(Val,L), \+ check_leq(Nr,Val,Hyps) -> fail ; true).
check_all_values_geq_val('$'(X),Nr,Hyps) :-
   avl_fetch_equal_from_hyps('$'(X),Hyps,Y,Hyps2),
   rewrite_local_loop_check(X,check_all_values_geq_val,Y,Hyps2,Hyps3),
   check_all_values_geq_val(Y,Nr,Hyps3).
%check_all_values_geq_val(A,B,_) :- print(check_all_values_geq_val_failed(A,B)),nl,fail.

check_all_values_neq_nr(intersection(A,B),Nr,Hyps) :-
    (check_all_values_neq_nr(A,Nr,Hyps) -> true ; check_all_values_neq_nr(B,Nr,Hyps)).
check_all_values_neq_nr(union(A,B),Nr,Hyps) :-
    (check_all_values_neq_nr(A,Nr,Hyps) -> check_all_values_neq_nr(B,Nr,Hyps)).
check_all_values_neq_nr(set_subtraction(A,_),Nr,Hyps) :-
    check_all_values_neq_nr(A,Nr,Hyps).
check_all_values_neq_nr(interval(From,_),Nr,Hyps) :- number(From),F1 is From-1, check_leq(Nr,F1,Hyps).
check_all_values_neq_nr(interval(_,To),Nr,Hyps) :- number(To),T1 is To+1, check_leq(T1,Nr,Hyps).
check_all_values_neq_nr('NATURAL',Nr,Hyps) :- check_leq(Nr,-1,Hyps).
check_all_values_neq_nr('NATURAL1',Nr,Hyps) :- check_leq(Nr,0,Hyps).
check_all_values_neq_nr(sorted_set_extension(L),Nr,Hyps) :- !, check_all_values_neq_nr(set_extension(L),Nr,Hyps).
check_all_values_neq_nr(set_extension(L),Nr,Hyps) :-
   (member(Val,L), \+ check_not_equal(Val,Nr,Hyps) -> fail ; true).
check_all_values_neq_nr('$'(X),Nr,Hyps) :-
   avl_fetch_equal_from_hyps('$'(X),Hyps,Y,Hyps2),
   rewrite_local_loop_check(X,check_all_values_neq_nr,Y,Hyps2,Hyps3),
   check_all_values_neq_nr(Y,Nr,Hyps3).
%check_all_values_neq_nr(A,B,_) :- print(check_all_values_neq_nr_failed(A,B)),nl,fail.


check_all_values_leq_val(intersection(A,B),Nr,Hyps) :-
    (check_all_values_leq_val(A,Nr,Hyps) -> true ; check_all_values_leq_val(B,Nr,Hyps)).
check_all_values_leq_val(union(A,B),Nr,Hyps) :-
    (check_all_values_leq_val(A,Nr,Hyps) -> check_all_values_leq_val(B,Nr,Hyps)).
check_all_values_leq_val(set_subtraction(A,_),Nr,Hyps) :-
    check_all_values_leq_val(A,Nr,Hyps).
check_all_values_leq_val(interval(_,To),Nr,Hyps) :- check_leq(To,Nr,Hyps).
check_all_values_leq_val(value(avl_set(AVL)),Nr,Hyps) :- avl_max(AVL,int(Max)), check_leq(Max,Nr,Hyps).
check_all_values_leq_val(domain(Func),Nr,Hyps) :-
   get_domain_or_superset(Func,Hyps,DomFunc,Hyps2),
   check_all_values_leq_val(DomFunc,Nr,Hyps2).
check_all_values_leq_val(range(Func),Nr,Hyps) :-
   get_range_or_superset(Func,Hyps,RanFunc,Hyps2),
   check_all_values_leq_val(RanFunc,Nr,Hyps2).
check_all_values_leq_val(sorted_set_extension(L),Nr,Hyps) :- !, check_all_values_leq_val(set_extension(L),Nr,Hyps).
check_all_values_leq_val(set_extension(L),Nr,Hyps) :-
   (member(Val,L), \+ check_leq(Val,Nr,Hyps) -> fail ; true).
check_all_values_leq_val('$'(X),Nr,Hyps) :-
   avl_fetch_equal_from_hyps('$'(X),Hyps,Y,Hyps2),
   rewrite_local_loop_check(X,check_all_values_leq_val,Y,Hyps2,Hyps3),
   check_all_values_leq_val(Y,Nr,Hyps3).
%check_all_values_leq_val(A,B,_) :- print(check_all_values_leq_val(A,B)),nl,fail.

% check if two expressions are definitely different
% usually called for check_not_equal 0 or empty_set
check_not_equal(A,B,_Hyp) :-
   ground_value(A), ground_value(B), !, A \= B.
check_not_equal(X,Y,Hyp) :- sym_unify(X,Y,if_then_else(_,A1,A2),B),!,
   (check_not_equal(A1,B,Hyp) -> check_not_equal(A2,B,Hyp)).
check_not_equal(N1,N2,Hyps) :-
   avl_fetch_not_equal(N1,N2,Hyps),!.
check_not_equal(couple(A1,A2),couple(B1,B2),Hyps) :- !,
   (check_not_equal(A1,B1,Hyps) -> true ; check_not_equal(A2,B2,Hyps)).
check_not_equal(X,B,Hyps) :- number(B),
   try_get_set_of_possible_values(X,Hyps,SetX,Hyps2),
   check_all_values_neq_nr(SetX,B,Hyps2),!.
      % TO DO: compute also things like domain(...) for :wd s:perm(1..10) & x:dom(s) & res = 10/x
check_not_equal(X,Y,Hyps) :- sym_unify(X,Y,unary_minus(A),B),number(B),!, BM is -B,
   check_not_equal(A,BM,Hyps).
check_not_equal(X,Y,Hyps) :- sym_unify(X,Y,multiplication(A,B),0),!,  % A*B /= 0  if A/=0 & B/=0
   check_not_equal(A,0,Hyps),check_not_equal(B,0,Hyps).
check_not_equal(X,Y,Hyps) :- sym_unify(X,Y,power_of(A,_),0),!,  % A**B /= 0  if A/=0
   check_not_equal(A,0,Hyps).
check_not_equal(X,Y,Hyps) :- sym_unify(X,Y,Add,B),
   add_with_number(Add,A,Nr),!,
   (Nr>0 -> check_leq(B,A,Hyps)  % A >= B =>  A+Nr > B  => A+Nr /= B
    ; Nr=0 -> check_not_equal(A,B,Hyps)
    ; check_leq(A,B,Hyps)
   ).
check_not_equal(X,Y,Hyps) :- sym_unify(X,Y,A,B),number(B),!,
   B1 is B+1,
   (check_leq(B1,A,Hyps) -> true  % B < A
    ; B2 is B-1,
      check_leq(A,B2,Hyps)). % A < B
check_not_equal(XX,YY,Hyps) :- sym_unify(XX,YY,'$'(A),B),
   avl_fetch_binop_from_hyps('$'(A),less_equal,Hyps,Y,Hyps2),
   (number(B) -> (B1 is B-1, check_leq(Y,B1,Hyps2) -> true)
    ; avl_fetch_not_equal('$'(A),Y,Hyps) % we have $(A) < Y  => prove Y <= B
       ->  check_leq(Y,B,Hyps2) % we can prove x<y & y<=z => x<z  but we cannot yet prove x<=y & y<z => x<z
   ).
check_not_equal(XX,YY,Hyps) :- sym_unify(XX,YY,'$'(A),B),
   avl_fetch_binop_from_hyps('$'(A),greater_equal,Hyps,Y,Hyps2),
   (number(B) -> (B1 is B+1, check_leq(B1,Y,Hyps2) -> true)
    ; avl_fetch_not_equal('$'(A),Y,Hyps) % we have $(A) < Y  => prove Y <= B
       ->  check_leq(B,Y,Hyps2) % see comments above
   ).
check_not_equal(A,empty_set,Hyp) :- !, check_not_empty_set(A,Hyp).
check_not_equal(empty_set,A,Hyp) :- !, check_not_empty_set(A,Hyp).
check_not_equal(value(avl_set(A)),value(avl_set(B)),_) :- nonvar(A), nonvar(B),!, % nonvar should always be true
    \+ equal_avl_tree(A,B).
check_not_equal(A,B,Hyps) :-
    (A=set_extension(LA) -> check_not_equal_set_extension(B,LA,Hyps)
     ; B=set_extension(LB) -> check_not_equal_set_extension(A,LB,Hyps)).
check_not_equal(A,B,Hyps) :-
    avl_fetch_worthwhile_equal_from_hyps(A,Hyps,Value,Hyps2),!,
    check_not_equal(Value,B,Hyps2).
check_not_equal(A,B,Hyps) :-
    avl_fetch_worthwhile_equal_from_hyps(B,Hyps,Value,Hyps2),!,
    check_not_equal(A,Value,Hyps2).
%check_not_equal(A,B,Hyps) :- print(check_not_equal_failed(A,B)),nl,fail.

check_not_equal_set_extension(set_extension([B|TB]),[A|TA],Hyps) :- (TA=[];TB=[]),!,
    check_not_equal(A,B,Hyps). % TO DO: we can generalize this treatment to find one element in one set not in the other
check_not_equal_set_extension(value(avl_set(AVL)),LA,Hyps) :- length(LA,MaxSizeA),
    (avl_size(AVL,Sze),Sze>MaxSizeA -> true % AVL has at least one element more
     ; is_one_element_avl(AVL,B), LA=[A|_], check_not_equal(A,B,Hyps)).

avl_fetch_not_equal(N1,N2,hyp_rec(AVL,_)) :-
   (avl_fetch(not_equal(N1,N2),AVL) -> true
    ; avl_fetch(not_equal(N2,N1),AVL)). % we do not store both directions for not_equal

% unify two variables with other two variables; useful for symmetric rules
sym_unify(A,B,A,B).
sym_unify(A,B,B,A).

% TO DO: get equalities; maybe we should harmonise this for all rules
% we could add rules about min_int, max_int

% ground_value which can be compared using Prolog equality:
ground_value(Nr) :- number(Nr),!.
ground_value(string(S)) :- atom(S).
ground_value(boolean_true).
ground_value(boolean_false).
ground_value(couple(A,B)):- ground_value(A),ground_value(B).
%ground_value(empty_set). there are other representations of the empty_set: empty_sequence,...
% TO DO: value,...

check_not_empty_set(A,Hyps) :- avl_fetch_from_hyps(not_equal(A,empty_set),Hyps),!.
check_not_empty_set(A,Hyps) :- %Note: size(A) should be changed to card(A) in normalization
   (CardA = card(A) ; CardA = size(A)),
   avl_fetch_binop_from_hyps(CardA,greater_equal,Hyps,Nr,Hyps2), %Nr \= 0,
   check_leq(1,Nr,Hyps2),!. % cut here, relevant for test 2043
check_not_empty_set(set_extension([_|_]),_Hyps).
check_not_empty_set(sorted_set_extension([_|_]),_Hyps).
check_not_empty_set(sequence_extension([_|_]),_Hyps).
check_not_empty_set(cartesian_product(A,B),Hyps) :- (check_not_empty_set(A,Hyps) -> check_not_empty_set(B,Hyps)).
check_not_empty_set(interval(A,B),Hyps) :- check_leq(A,B,Hyps).
check_not_empty_set(value(avl_set(AVL)),_) :- AVL \= empty.
check_not_empty_set(union(A,B),Hyp) :- !, (check_not_empty_set(A,Hyp) -> true ; check_not_empty_set(B,Hyp)).
check_not_empty_set(overwrite(A,B),Hyp) :- !, (check_not_empty_set(A,Hyp) -> true ; check_not_empty_set(B,Hyp)).
check_not_empty_set(domain(A),Hyp) :- !, check_not_empty_set(A,Hyp).
check_not_empty_set(range(A),Hyp) :- !, check_not_empty_set(A,Hyp).
check_not_empty_set(identity(A),Hyp) :- !, check_not_empty_set(A,Hyp).
check_not_empty_set(image(R,interval(L,U)),Hyp) :- !,
   check_not_empty_set(interval(L,U),Hyp),
   (check_member_of_set(domain(R),L,Hyp,_) -> true
    ; check_member_of_set(domain(R),U,Hyp,_)
   ).
check_not_empty_set(reverse(A),Hyp) :- !, check_not_empty_set(A,Hyp).
check_not_empty_set(rev(A),Hyp) :- !, check_not_empty_set(A,Hyp).
check_not_empty_set(concat(A,B),Hyp) :- !, (check_not_empty_set(A,Hyp) -> true ; check_not_empty_set(B,Hyp)).
check_not_empty_set(bool_set,_Hyp) :- !.
check_not_empty_set(float_set,_Hyp) :- !.
check_not_empty_set(real_set,_Hyp) :- !.
check_not_empty_set(string_set,_Hyp) :- !.
check_not_empty_set('NATURAL1',_Hyp) :- !.
check_not_empty_set('NATURAL',_Hyp) :- !.
check_not_empty_set(typeset,_Hyp) :- !. % all basic sets are non empty in B and Event-B
check_not_empty_set(Expr,Hyps) :-
   is_lambda_function_with_domain(Expr,Domain),!,
   check_not_empty_set(Domain,Hyps).
check_not_empty_set('$'(ID),Hyps) :-
    enumerated_set(ID),
    \+ is_hyp_var(ID,Hyps),!. % global enumerated set visible
check_not_empty_set(Eq,Hyps) :-
   (Eq='$'(_) ; Eq=interval(_,_)),
   avl_fetch_equal_from_hyps(Eq,Hyps,Value,Hyps2),
   rewrite_local_loop_check(Eq,check_not_empty_set,Value,Hyps2,Hyps3),
   check_not_empty_set(Value,Hyps3),!.
check_not_empty_set(Seq,Hyp) :- infer_sequence_type_of_expr(Seq,Hyp,seq1),!.
check_not_empty_set(Func,Hyps) :- Func = '$'(_),
   avl_fetch_binop_from_hyps(Func,member,Hyps,FunctionType,Hyps2), % Func : . --> .
   check_not_empty_elements(FunctionType,Hyps2),!.
check_not_empty_set(function(Func2,_),Hyps) :-
   get_range_or_superset(Func2,Hyps,Range,Hyps2),
   check_not_empty_elements(Range,Hyps2),!.
check_not_empty_set(tail(A),Hyps) :- rewrite_card_of_set(A,Hyps,CardA,Hyps1),!,
   check_leq(2,CardA,Hyps1).
check_not_empty_set(front(A),Hyps) :- rewrite_card_of_set(A,Hyps,CardA,Hyps1),!,
   check_leq(2,CardA,Hyps1).
check_not_empty_set(A,Hyps) :-
   ( Lookup = A, Operator = superset
     ;
     (Lookup=domain(A) ; Lookup=range(A)),
     (Operator = superset ; Operator = equal)
   ),
   avl_fetch_binop_from_hyps(Lookup,Operator,Hyps,B,Hyps2), % B /= {} & B <: A =>  A /= {}
   rewrite_local_loop_check(A,check_not_empty_set,B,Hyps2,Hyps3),
   check_not_empty_set(B,Hyps3),!.
% TO DO: rule for dom(r)<:A  and r not empty implies A not empty; problem: we need lookup for A=dom(r), or dom(r)<:A, could be of form: r:A+->B
% check_not_empty_set(A,_) :- print(check_not_empty_set_failed(A)),nl,fail.
% TO DO: more rules for sequence operators; infer_sequence_type_of_expr does not look at values of ids


% check if elements of a function type or set are guaranteed to be not empty

check_not_empty_elements(fin1_subset(_),_).
check_not_empty_elements(pow1_subset(_),_).
check_not_empty_elements(seq1(_),_).
check_not_empty_elements(iseq1(_),_).
check_not_empty_elements(perm(A),Hyps) :- check_not_empty_set(A,Hyps).
check_not_empty_elements(total_function(A,_),Hyps) :- check_not_empty_set(A,Hyps).
check_not_empty_elements(total_injection(A,_),Hyps) :- check_not_empty_set(A,Hyps).
check_not_empty_elements(total_surjection(A,B),Hyps) :-
  (check_not_empty_set(A,Hyps) -> true ; check_not_empty_set(B,Hyps)).
check_not_empty_elements(total_bijection(A,B),Hyps) :-
  (check_not_empty_set(A,Hyps) -> true ; check_not_empty_set(B,Hyps)).
check_not_empty_elements(total_relation(A,_),Hyps) :- check_not_empty_set(A,Hyps).
check_not_empty_elements(total_surjection_relation(A,_),Hyps) :- check_not_empty_set(A,Hyps).
check_not_empty_elements(partial_surjection(_,B),Hyps) :- check_not_empty_set(B,Hyps).
check_not_empty_elements(partial_bijection(_,B),Hyps) :- check_not_empty_set(B,Hyps).
check_not_empty_elements(surjection_relation(_,B),Hyps) :- check_not_empty_set(B,Hyps).
% more cases, set_extension,...

:- use_module(probsrc(b_global_sets),[enumerated_set/1]).
% check if an expression is definitely finite
check_finite(bool_set,_,bool_set) :- !.
check_finite(empty_set,_,empty_set) :- !.
check_finite(empty_sequence,_,empty_sequence) :- !.
check_finite(float_set,_,bool_set) :- !.
% check_finite(integer_set(X),_,bool_set) :- !.  INT, NAT, NAT1 are translated to intervals
check_finite(set_extension(_),_,set_extension) :- !.
check_finite(sorted_set_extension(_),_,set_extension) :- !.
check_finite(sequence_extension(_),_,seq_extension) :- !.
check_finite(fin_subset(X),Hyps,fin(PT)) :- !, check_finite(X,Hyps,PT).
check_finite(fin1_subset(X),Hyps,fin1(PT)) :- !, check_finite(X,Hyps,PT).
check_finite(pow_subset(X),Hyps,pow(PT)) :- !, check_finite(X,Hyps,PT).
check_finite(pow1_subset(X),Hyps,pow1(PT)) :- !, check_finite(X,Hyps,PT).
check_finite(iseq(X),Hyps,iseq(PT)) :- !, check_finite(X,Hyps,PT).
check_finite(iseq1(X),Hyps,iseq1(PT)) :- !, check_finite(X,Hyps,PT).
check_finite(mu(Set),Hyps,mu) :- !, has_finite_elements(Set,Hyps).
check_finite(perm(X),Hyps,perm(PT)) :- !, check_finite(X,Hyps,PT).
check_finite(Set,Hyps,hyp) :-
    avl_fetch_from_hyps(finite(Set),Hyps),!.
check_finite(domain(A),Hyp,dom(PT)) :- !,
  (check_finite(A,Hyp,PT) -> true ; finite_domain(A,Hyp,PT)).
check_finite(range(A),Hyp,ran(PT)) :- !,
  (check_finite(A,Hyp,PT) -> true ; finite_range(A,Hyp,PT)).
check_finite(reverse(A),Hyp,rev(PT)) :- !, check_finite(A,Hyp,PT).
check_finite(identity(A),Hyp,id(PT)) :- !,check_finite(A,Hyp,PT). % finite(id(A)) if finite(A)
check_finite(function(Func,_),Hyps,function_finite_elements) :- !,
   get_range_or_superset(Func,Hyps,RanFunc,Hyps2),
   has_finite_elements(RanFunc,Hyps2).
check_finite(image(Func,B),Hyp,image(PT)) :- !,
    (check_finite(Func,Hyp,PT) -> true  % finite(Func[.]) <= finite(Func)
     ; check_finite(B,Hyp,PTB)
        -> PT = pfun(PTB), check_is_partial_function(Func,Hyp) % finite(Func[B]) <= finite(B) & Func : TD +-> TR
    ).
check_finite(union(A,B),Hyp,union(PTA,PTB)) :- !, (check_finite(A,Hyp,PTA) -> check_finite(B,Hyp,PTB)).
check_finite(if_then_else(_,A,B),Hyps,if(PTA,PTB)) :- !, (check_finite(A,Hyps,PTA) -> check_finite(B,Hyps,PTB)).
check_finite(intersection(A,B),Hyps,intersection(D,PT)) :- !,
    (D=left,check_finite(A,Hyps,PT) -> true ; D=right,check_finite(B,Hyps,PT)).
check_finite(cartesian_product(A,B),Hyp,PT) :-
    (check_finite(A,Hyp,PT) -> (A=empty_set -> true ; check_finite(B,Hyp,_)) ; B=empty_set,PT=cart_empty_set).
check_finite(Rel,Hyp,rel(PTA,PTB)) :- is_relations_type(Rel,A,B),!,
    (check_finite(A,Hyp,PTA) -> check_finite(B,Hyp,PTB)). % add other relations
check_finite(direct_product(A,B),Hyp,direct_product(PTA,PTB)) :- !,
    (check_finite(A,Hyp,PTA) -> check_finite(B,Hyp,PTB)).
check_finite(parallel_product(A,B),Hyp,parallel_product(PTA,PTB)) :- !,
    (check_finite(A,Hyp,PTA) -> check_finite(B,Hyp,PTB)).
check_finite(overwrite(A,B),Hyp,overwrite(PTA,PTB)) :- !, (check_finite(A,Hyp,PTA) -> check_finite(B,Hyp,PTB)).
check_finite(set_subtraction(A,_),Hyps,set_subtraction(PT)) :- !, check_finite(A,Hyps,PT).
check_finite(domain_restriction(A,BRel),Hyp,domain_restriction(PT)) :- !,
   (check_finite(BRel,Hyp,PT) ->  true
    ; check_is_partial_function(BRel,Hyp), check_finite(A,Hyp,PT)
    ; finite_range(BRel,Hyp,_) -> check_finite(A,Hyp,PT)  % finite(a <| brel) if finite(a) & finite(ran(brel))
    ).
check_finite(domain_subtraction(_,B),Hyp,domain_subtraction(PT)) :- !, check_finite(B,Hyp,PT).
check_finite(range_restriction(ARel,B),Hyp,range_restriction(PT)) :- !,
   (check_finite(ARel,Hyp,PT) ->  true
    ; check_is_injective(ARel,Hyp) -> check_finite(B,Hyp,PT)
    ; finite_domain(ARel,Hyp,_) -> check_finite(B,Hyp,PT)  % finite(arel |> b) if finite(b) & finite(dom(arel))
    ).
check_finite(image(A,B),Hyp,image(PT)) :- % A[B] is finite if A is finite or if B is finite and A a function
   (check_finite(A,Hyp,PT) ->  true ; check_is_partial_function(A,Hyp), check_finite(B,Hyp,PT)).
check_finite(range_subtraction(A,_),Hyp,range_subtraction(PT)) :-  check_finite(A,Hyp,PT).
check_finite(interval(_,_),_,interval) :- !.
check_finite(value(X),_,avl_set) :- nonvar(X),X=avl_set(_),!.
check_finite('$'(ID),Hyps,finite_type) :-
    get_hyp_var_type(ID,Hyps,Type),
    is_finite_type_for_wd(Type,Hyps),!.
check_finite('$'(ID),Hyps,enumerated_set) :-
    enumerated_set(ID),
    \+ is_hyp_var(ID,Hyps),!. % global enumerated set visible
%check_finite('$'(ID),Hyp,partition) :-
%    avl_fetch_binop_from_hyps('$'(ID),partition,Hyp,Values,Hyp2), % is now normalized
%    not_occurs(Values,ID),
%    l_check_finite(Values,Hyp2),!.
check_finite('$'(ID),Hyps,rewrite(Operator,PT)) :-
    (Operator = equal ; Operator = subset), % for subset_strict we also have subset in Hyp
    avl_fetch_binop_from_hyps('$'(ID),Operator,Hyps,Set2,Hyps2),
    not_occurs(Set2,ID), % avoid silly, cyclic rewrites $x -> reverse(reverse($x)) (FunLawsStrings.mch)
     % however, in SetLawsNatural this prevents proving 2 POs due to  SS <: min(SS)..max(SS)
    rewrite_local_loop_check(ID,check_finite,Set2,Hyps2,Hyps3),
    check_finite(Set2,Hyps3,PT),!.
check_finite(Set,Hyp,finite_elements) :- id_or_record_field(Set),
    avl_fetch_mem_or_struct(Set,Hyp,Set2,Hyp2),
    Set2 \= Set,
    has_finite_elements(Set2,Hyp2).
check_finite(Func,Hyp,pfun(PTA,PTB)) :- is_partial_function(Func,A,B),!,
   % a set of partial functions from A to B is finite if both A and B are finite
   (check_finite(A,Hyp,PTA) -> check_finite(B,Hyp,PTB)),!.
check_finite(Seq,Hyp,seq_type) :- infer_sequence_type_of_expr(Seq,Hyp,_),!. % a sequence is always finite
check_finite(comprehension_set(Paras,Body),Hyp,comprehension_set) :-
    finite_comprehension_set(Paras,Body,Hyp),!.
check_finite(struct(rec(Fields)),Hyp,struct) :- maplist(check_finite_field(Hyp),Fields).
% TODO: is_lambda_function_with_domain; improve some proof trees above in style of intersection
%check_finite(F,_Hs,_) :- print(check_finite_failed(F)),nl,fail.

check_finite_field(Hyp,field(_,Set)) :- check_finite(Set,Hyp,_PT).

finite_comprehension_set(Paras,Body,Hyp) :-
    % first exclude the parameters which can be proven finite on their own
    % now deal with the rest; we do not pass AllParas as second arg; as all references to excluded IDs is ok (only finitely many values possible)
    finite_comprehension_set_rest(Paras,Body,Hyp,[],Rest),
    Rest \= Paras,
    finite_comprehension_set_rest(Rest,Body,Hyp,[],[]). % do a second pass, e.g., for {x,y|x:{y,y+1} & y:1..2 & x:INTEGER}


finite_comprehension_set_rest([],_,_,_,[]).
% finite_comprehension_set(['$'(ID)|TID],Body,Hyp) :-  finite_type !
finite_comprehension_set_rest([ParaID1|TID],Body,Hyp,UnProven,Rest) :-
    get_parameter_superset_in_body(ParaExpr,[ParaID1|TID],Body,Values),
    l_not_occurs(Values,UnProven), % do not rely on not yet finitely proven paras; e.g. for {x,y|x:INTEGER & y=x}
    match_parameter(ParaExpr,ParaID1),
    check_finite(Values,Hyp,_PT),!,
    finite_comprehension_set_rest(TID,Body,Hyp,UnProven,Rest).
finite_comprehension_set_rest([ParaID1|TID],Body,Hyp,UnProven,[ParaID1|Rest]) :-
    finite_comprehension_set_rest(TID,Body,Hyp,[ParaID1|UnProven],Rest).

% match_parameter(Expr,ID) -> ID occurs in Expr and finite number of values for Expr implies finite values for ID
match_parameter(ParaID,ParaID).
match_parameter(couple(ParaID,_),ParaID).  % x|->y : Values  finite implies finitely many values for x
match_parameter(couple(_,ParaID),ParaID).
match_parameter(set_extension(Ext),ParaID) :-
     member(El,Ext), match_parameter(El,ParaID). % {x,..} : Values finite implies finitely many values for x
match_parameter(sorted_set_extension(Ext),ParaID) :- match_parameter(set_extension(Ext),ParaID).
match_parameter(sequence_extension(Ext),ParaID) :- member(ParaID,Ext).
match_parameter(rev(RF),ParaID) :- match_parameter(RF,ParaID).
match_parameter(reverse(RF),ParaID) :- match_parameter(RF,ParaID).
match_parameter(unary_minus(RF),ParaID) :- match_parameter(RF,ParaID).
match_parameter(concat(RF1,RF2),ParaID) :- (match_parameter(RF1,ParaID) -> true ; match_parameter(RF2,ParaID)).
match_parameter(union(RF1,RF2),ParaID) :- (match_parameter(RF1,ParaID) -> true ; match_parameter(RF2,ParaID)).
match_parameter(overwrite(_,RF),ParaID) :- match_parameter(RF,ParaID). % f <+ RF = (... <<| f ) \/ RF
match_parameter(Add,ParaID) :- add_with_number(Add,A,_Number), match_parameter(A,ParaID).
match_parameter(Mul,ParaID) :- mul_with_number(Mul,A,Number), Number \= 0, match_parameter(A,ParaID).
% TO DO: more injective functions where a solution determines the ParaID, identity? direct_product, ...
% cartesian_product : only if other set not empty

get_parameter_superset_in_body(ParaID,AllParas,Body,Values) :-
    (  member_in_norm_conjunction(Body,member(ParaID,Values))
     ; member_in_norm_conjunction(Body,subset(ParaID,Values)) % there are only finitely many subsets of a finite set
     ; member_in_norm_conjunction(Body,subset_strict(ParaID,Values))
     ; member_in_norm_conjunction(Body,equal(X,Y)),
       sym_unify(X,Y,ParaID,Value), Values=set_extension([Value]) ),
    l_not_occurs(Values,AllParas). % as an alternative: check for finite_type of set elements; e.g., detect ID=bool(...)


:- use_module(probsrc(bsyntaxtree),[is_set_type/2]).
%we suppose this has already failed: finite_domain(A,Hyp) :- check_finite(A,Hyp,PT).
finite_domain('$'(ID),Hyps,finite_type) :-
    get_hyp_var_type(ID,Hyps,Type),
    is_set_type(Type,couple(DomType,_)),
    is_finite_type_for_wd(DomType,Hyps),!.
finite_domain(domain_restriction(A,Rel),Hyps,domain_restriction(PT)) :- !,
    (check_finite(A,Hyps,PT) -> true ; finite_domain(Rel,Hyps,PT)).
finite_domain(A,Hyp,PT) :- get_domain_or_superset(A,Hyp,DA,Hyp2),!, check_finite(DA,Hyp2,PT).

finite_range('$'(ID),Hyps,finite_type) :-
    get_hyp_var_type(ID,Hyps,Type),
    is_set_type(Type,couple(_,RanType)),
    is_finite_type_for_wd(RanType,Hyps),!.
finite_range(A,Hyp,PT) :- get_range_or_superset(A,Hyp,RA,Hyp2),!, check_finite(RA,Hyp2,PT).

l_check_finite([],_).
l_check_finite([H|T],Hyp) :- (check_finite(H,Hyp,_) -> l_check_finite(T,Hyp)).

% is a set containing only finite sets
has_finite_elements(empty_set,_) :- !. % has no elements
has_finite_elements(empty_sequence,_) :- !. % has no elements
has_finite_elements(fin_subset(_),_) :- !.
has_finite_elements(fin1_subset(_),_) :- !.
has_finite_elements(pow_subset(X),Hyps) :- !, check_finite(X,Hyps,_).
has_finite_elements(pow1_subset(X),Hyps) :- !, check_finite(X,Hyps,_).
has_finite_elements(seq(_),_) :- !. % every sequence is always finite (of finite length)
has_finite_elements(seq1(_),_) :- !.
has_finite_elements(iseq(_),_) :- !.
has_finite_elements(iseq1(_),_) :- !.
has_finite_elements(perm(_),_) :- !.
has_finite_elements(union(A,B),Hyps) :- !, has_finite_elements(A,Hyps), has_finite_elements(B,Hyps).
has_finite_elements(intersection(A,B),Hyps) :- !, (has_finite_elements(A,Hyps) -> true ; has_finite_elements(B,Hyps)).
has_finite_elements(set_subtraction(A,_),Hyps) :- !, has_finite_elements(A,Hyps).
has_finite_elements(sorted_set_extension(L),Hyps) :- !, l_check_finite(L,Hyps).
has_finite_elements(set_extension(L),Hyps) :- !, l_check_finite(L,Hyps).
has_finite_elements(Func,Hyps) :- is_partial_function(Func,A,B),!,
  (check_finite(A,Hyps,_) -> true ; is_injective(Func), check_finite(B,Hyps,_)).
has_finite_elements(Rel,Hyps) :- is_relations_type(Rel,A,B),!,check_finite(A,Hyps,_),check_finite(B,Hyps,_).
%has_finite_elements(F,Hs) :- print(has_finite_elements_failed(F)),nl, portray_hyps(Hs),fail.


is_relations_type(relations(A,B),A,B).
is_relations_type(surjection_relation(A,B),A,B).
is_relations_type(total_relation(A,B),A,B).
is_relations_type(total_surjection_relation(A,B),A,B).

% TO DO: more rules for functions
% ------------------------------

:- use_module(probsrc(avl_tools),[avl_fetch_bin/4]).

% fetch member(Ground,Free) construct
%avl_fetch_mem(Key, AVL ,Res) :- avl_fetch_bin(Key, member, AVL ,Res).
%avl_fetch_equal(Key, AVL ,Res) :- avl_fetch_bin(Key, equal, AVL ,Res).


avl_fetch_mem_from_hyps(ID,Hyps,Value,Hyps2) :-
   avl_fetch_binop_from_hyps(ID,member,Hyps,Value,Hyps2).

avl_fetch_equal_from_hyps(ID,Hyps,Value,Hyps2) :-
   avl_fetch_binop_from_hyps(ID,equal,Hyps,Value,Hyps2).

avl_fetch_worthwhile_equal_from_hyps(ID,Hyps,Value,Hyps2) :-
   worth_rewriting_with_equal(ID),
   avl_fetch_binop_from_hyps(ID,equal,Hyps,Value,Hyps2),
   quick_not_occurs_check(ID,Value).

% fetch member predicate or indirect member via record fields
avl_fetch_mem_or_struct(Func,Hyps,Function,Hyps1) :-
  get_type_from_hyps(Func,Hyps,Function,Hyps1).
avl_fetch_mem_or_struct(record_field(Rec,Field),Hyps,FieldType,Hyps2) :-
  get_record_type_fields(Rec,Fields,Hyps,Hyps2),
  (member(field(Field,FieldType),Fields) -> true).

% find record type and extract fields for a given expression
get_record_type_fields(function(Func,_),Fields,Hyps,Hyps2) :-
   get_range_or_superset(Func,Hyps,Range,Hyps1),
   check_equal_pattern(Range,struct(rec(Fields)),Hyps1,Hyps2).
get_record_type_fields(Rec,Fields,Hyps,Hyps2) :-
   get_type_from_hyps(Rec,Hyps,RecType,Hyps1),
   check_equal_pattern(RecType,struct(rec(Fields)),Hyps1,Hyps2).

% get type from hyps x:XType  or  x:ran(F) with F : A->B
get_type_from_hyps(X,Hyps,XType,Hyps2) :-
   avl_fetch_mem_from_hyps(X,Hyps,XSet,Hyps1),
   get_type2(XSet,Hyps1,XType,Hyps2).
get_type_from_hyps(function(Func2,_),Hyps,Range,Hyps2) :-
  % f : _ +-> ( _ >-> _ )  =>  f(_) : _ >-> _
  get_range_or_superset(Func2,Hyps,Range,Hyps2).
get_type_from_hyps(second_of_pair(X),Hyps,Type,Hyps2) :- %prj2
   get_type_from_hyps(X,Hyps,XType,Hyps1),
   check_equal_pattern(XType,cartesian_product(_,Type),Hyps1,Hyps2).
get_type_from_hyps(first_of_pair(X),Hyps,Type,Hyps2) :- %prj1
   get_type_from_hyps(X,Hyps,XType,Hyps1),
   check_equal_pattern(XType,cartesian_product(Type,_),Hyps1,Hyps2).
get_type_from_hyps(assertion_expression(_,_,X),Hyps,XType,Hyps1) :-
   get_type_from_hyps(X,Hyps,XType,Hyps1).

get_type2(domain(Func),Hyps,XType,Hyps2) :- !,
   get_domain_or_superset(Func,Hyps,XType,Hyps2).
get_type2(range(Func),Hyps,XType,Hyps2) :- !,
   get_range_or_superset(Func,Hyps,XType,Hyps2).
get_type2(Type,Hyps,Type,Hyps).

id_or_record_field('$'(_)).
id_or_record_field(record_field(_,_)).

% perform occurs check if first arg is an identifier:
quick_not_occurs_check('$'(ID),Value) :- !, not_occurs(Value,ID).
quick_not_occurs_check(_,_).

% worth rewriting with equality hyps
worth_rewriting_with_equal('$'(_)).
worth_rewriting_with_equal(record_field('$'(_),_)).
worth_rewriting_with_equal(couple(_,_)).
worth_rewriting_with_equal(size(_)).
worth_rewriting_with_equal(card(_)).
worth_rewriting_with_equal(function(_,_)).

% utility to fetch fully ground term from hyp avl
avl_fetch_from_hyps(Term,hyp_rec(AVL,_)) :- avl_fetch(Term,AVL).

avl_fetch_binop_from_hyps(ID,BinOp,hyp_rec(AVL,HInfos),Value,hyp_rec(AVL,HInfos2)) :-
   avl_fetch_bin(ID,BinOp,AVL,Value),
   (avl_fetch(prevent_cycle_count,HInfos,CycleCount) % avoid cycles x=y, y=x
     -> (CycleCount < 5 -> true ;  % print(prevented_cycle(ID,CycleCount)),nl,   trace,
      % in test 2018: :wd target = [2,1,1,2,1] & n=size(target) & i:1..n & target(i)=res requires cycle count < 5
         !, fail),
        C1 is CycleCount+1
      ; C1 is 1
    ),
    avl_store(prevent_cycle_count,HInfos,C1,HInfos2).
% detect local loops; should not be used where Hyps are passed to unrelated goals or one has to reset_local_loop_check
rewrite_local_loop_check(_,_,Value,HI,HI1) :- useful_value(Value),!,HI1=HI.
rewrite_local_loop_check(Term,ProverPredicate,_,Hyps,_) :- var(Hyps),!,
   add_internal_error('Var hyps: ',rewrite_local_loop_check(Term,ProverPredicate,_,Hyps,_)),trace,fail.
rewrite_local_loop_check(Term,ProverPredicate,_,hyp_rec(AVL,HInfos),hyp_rec(AVL,HInfos1)) :- !,
    %(avl_fetch(rewritten(Term,ProverPredicate),HInfos) -> print(loop(Term,ProverPredicate)),nl,fail ; true),
    \+ avl_fetch(rewritten(Term,ProverPredicate),HInfos),
    avl_store(rewritten(Term,ProverPredicate),HInfos,true,HInfos1).
    % :wd a : 1 .. sz --> INTEGER & sz=5 & p : perm(dom(a)) & i : 1 .. sz - 1 & res= p(i)  % sz rewritten multiple times
    % :wd f: BOOL --> 1..10 & g : 0..20 --> BOOL & bb:BOOL & (f;g)(bb)=res

%reset_local_loop_check()

% values where there is no risk of looping when rewriting to:
useful_value(Value) :- number(Value).
useful_value(interval(A,B)) :- number(A), number(B).
useful_value(value(_)).


% rename and prove a list of goals
l_rename_and_prove_goals([],_,_,[]).
l_rename_and_prove_goals([H|T],Subst,Hyps,[PTH|PTT]) :-
   rename_norm_term(H,Subst,RH),!,
   prove_po(RH,Hyps,PTH),!, % TO DO: use version of prove_po that does not print info
   l_rename_and_prove_goals(T,Subst,Hyps,PTT).


% small utility for sanity checking
check_integer(A,PP) :- not_integer(A),!, add_error(PP,'Not an integer: ',A),fail.
check_integer(_,_).
not_integer(empty_set).
not_integer(interval(_,_)).
not_integer(couple(_,_)).
not_integer(union(_,_)).
not_integer(intersection(_,_)).
not_integer(domain(_)).
not_integer(range(_)).
% TO DO: extend
