% (c) 2020-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(well_def_tools, [rename_norm_term/3, member_in_norm_conjunction/2, not_occurs/2, occurs/2]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module provides utilities for the WD prover.').


:- use_module(probsrc(debug)).
% --------------------
% tools for normalized expressions:

% check if a known/bound ID occurs in a normalized expression
not_occurs(Expr,ID) :- (occurs(Expr,ID) -> debug_println(19,occurs_in(ID,Expr)),fail ; true).
occurs('$'(ID1),ID2) :- !, ID1 = ID2.
occurs(conjunct(A,B),ID) :- !, (occurs(A,ID) ; occurs(B,ID)).
occurs(couple(A,B),ID) :- !, (occurs(A,ID) ; occurs(B,ID)).
occurs(equal(A,B),ID) :- !, (occurs(A,ID) ; occurs(B,ID)).
occurs(field(_,A),ID) :- !, occurs(A,ID).
occurs(function(A,B),ID) :- !, (occurs(A,ID) ; occurs(B,ID)).
occurs(interval(A,B),ID) :- !, (occurs(A,ID) ; occurs(B,ID)).
occurs(member(A,B),ID) :- !, (occurs(A,ID) ; occurs(B,ID)).
occurs(record_field(A,B),ID) :- !, (occurs(A,ID) ; occurs(B,ID)).
occurs(value(_),_ID) :- !, fail.
occurs(string(_),_ID) :- !, fail.
occurs(reverse(A),ID) :- !, occurs(A,ID).
occurs(domain(A),ID) :- !, occurs(A,ID).
occurs(range(A),ID) :- !, occurs(A,ID).
occurs(comprehension_set(Paras,Body),ID) :- !,
    (member('$'(ID),Paras) -> fail % ID becomes invisible
      ; occurs(Body,ID)).
% TO DO: treat other bound IDs exists,... [but not crucial, as we err on the safe side here]
occurs([H|T],ID) :- !, (occurs(H,ID) -> true ; occurs(T,ID)).
occurs(boolean_true,_) :- !,fail.
occurs(boolean_false,_) :- !,fail.
occurs(empty_set,_) :- !,fail.
occurs(Expr,ID) :- Expr =.. [_|Args], member(A,Args), occurs(A,ID).

member_in_norm_conjunction(Pred,Pred).
member_in_norm_conjunction(conjunct(A,B),Pred) :-
    member_in_norm_conjunction(A,Pred) ; member_in_norm_conjunction(B,Pred).



% apply a substitution to a normalised term:
rename_norm_term('$'(X),Subst,Res) :-!,
   (member(X/NewTerm,Subst) -> Res=NewTerm ; Res='$'(X)).
rename_norm_term(conjunct(A,B),Subst,conjunct(SA,SB)) :-
   rename_norm_term(A,Subst,SA), rename_norm_term(B,Subst,SB).
rename_norm_term(if_then_else(A,B,C),Subst,if_then_else(SA,SB,SC)) :-
   rename_norm_term(A,Subst,SA), rename_norm_term(B,Subst,SB), rename_norm_term(C,Subst,SC).
rename_norm_term(equal(A,B),Subst,equal(SA,SB)) :-
   rename_norm_term(A,Subst,SA), rename_norm_term(B,Subst,SB).
rename_norm_term(function(A,B),Subst,function(SA,SB)) :-
   rename_norm_term(A,Subst,SA), rename_norm_term(B,Subst,SB).
rename_norm_term(member(A,B),Subst,member(SA,SB)) :-
   rename_norm_term(A,Subst,SA), rename_norm_term(B,Subst,SB).
rename_norm_term(Nr,_,Res) :- simple_norm_term(Nr),!,Res=Nr.
rename_norm_term(Term,Subst,RenTerm) :- functor(Term,BOP,2), bin_op(BOP),!,
   functor(RenTerm,BOP,2),
   arg(1,Term,A), arg(1,RenTerm,RA), rename_norm_term(A,Subst,RA),
   arg(2,Term,B), arg(2,RenTerm,RB), rename_norm_term(B,Subst,RB).
rename_norm_term(Term,Subst,RenTerm) :- functor(Term,UOP,1), un_op(UOP),!,
   functor(RenTerm,UOP,1),
   arg(1,Term,A), arg(1,RenTerm,RA), rename_norm_term(A,Subst,RA).
rename_norm_term(Term,Subst,RenTerm) :- list_op(Term,Kind,List),!,
   list_op(RenTerm,Kind,RList),
   l_rename_norm_terms(List,Subst,RList).
rename_norm_term(A,_,_) :- print(uncovered_renaming(A)),nl,fail.


l_rename_norm_terms([],_,[]).
l_rename_norm_terms([H|T],Subst,[RH|RT]) :-
   rename_norm_term(H,Subst,RH),!,
   l_rename_norm_terms(T,Subst,RT).


list_op(set_extension(L),set_extension,L).
list_op(sequence_extension(L),sequence_extension,L).

% TO DO: add more bin and unary ops; records, quantified operators
bin_op(add).
bin_op(add_real).
bin_op(card).
bin_op(cartesian_product).
bin_op(composition).
bin_op(concat).
bin_op(couple).
bin_op(div).
bin_op(div_real).
bin_op(disjunct).
bin_op(domain_restriction).
bin_op(domain_subtraction).
bin_op(equivalence).
bin_op(greater).
bin_op(greater_equal).
bin_op(floored_div).
bin_op(image).
bin_op(implication).
bin_op(insert_front).
bin_op(insert_tail).
bin_op(intersection).
bin_op(interval).
bin_op(iteration).
bin_op(less).
bin_op(less_equal).
bin_op(less_real).
bin_op(less_equal_real).
bin_op(minus).
bin_op(minus_real).
bin_op(modulo).
bin_op(multiplication).
bin_op(multiplication_real).
bin_op(overwrite).
bin_op(not_equal).
bin_op(not_member).
bin_op(partial_bijection).
bin_op(partial_function).
bin_op(partial_injection).
bin_op(partial_surjection).
bin_op(power_of).
bin_op(power_of_real).
bin_op(range_restriction).
bin_op(range_subtraction).
bin_op(relations).
bin_op(restrict_front).
bin_op(restrict_tail).
bin_op(set_subtraction).
bin_op(size).
bin_op(subset).
bin_op(subset_strict).
bin_op(surjection_relation).
bin_op(total_function).
bin_op(total_injection).
bin_op(total_surjection).
bin_op(total_relation).
bin_op(total_surjection_relation).
bin_op(union).

un_op(bag_items).
un_op(card).
un_op(closure). % closure1
un_op(compaction).
un_op(convert_bool).
un_op(convert_real).
un_op(convert_int_floor).
un_op(convert_int_ceiling).
un_op(domain).
un_op(finite).
un_op(fin_subset).
un_op(fin1_subset).
un_op(first_of_pair).
un_op(front).
un_op(identity).
un_op(iseq).
un_op(iseq1).
un_op(negation).
un_op(max).
un_op(min).
un_op(mu).
un_op(perm).
un_op(pow_subset).
un_op(pow1_subset).
un_op(range).
un_op(rev).
un_op(reverse).
un_op(reflexive_closure).
un_op(second_of_pair).
un_op(seq).
un_op(seq1).
un_op(size).
un_op(tail).
un_op(unary_minus).
un_op(unary_minus_real).

% do not require renaming:
simple_norm_term(truth).
simple_norm_term(boolean_true).
simple_norm_term(boolean_false).
simple_norm_term(bool_set).
simple_norm_term(float_set).
simple_norm_term(real_set).
simple_norm_term(string_set).
simple_norm_term(typeset).
simple_norm_term(value(_)).
simple_norm_term(string(_)).
simple_norm_term(empty_set).
simple_norm_term(empty_sequence).
simple_norm_term(max_int).
simple_norm_term(min_int).
simple_norm_term('NATURAL').
simple_norm_term('NATURAL1').
simple_norm_term(lazy_lookup_pred(_)).
simple_norm_term(lazy_lookup_expr(_)).
simple_norm_term(Nr) :- number(Nr).
