% (c) 2016-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(xml2b,[convert_xml_to_b/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,external_functions).
:- module_info(description,'This module transforms XML into a B representation.').

%:- use_module(library(xml)).
:- use_module(xml_prob).
:- use_module(library(lists)).
:- use_module(custom_explicit_sets,[convert_to_avl/2]).
:- use_module(error_manager).
:- use_module(debug).

% type of generated XML elements:
/*      struct(
        recId: NATURAL1,
        pId:NATURAL,
        element:STRING,
        attributes: STRING +-> STRING,
        meta: STRING +-> STRING
        )
*/

% convert a code list to a B sequence/set of records
convert_xml_to_b(Codes,BContent) :-
     reset_counter(xml2b_current_line),
     set_counter(xml2b_current_line,1),
     peel_and_count_newlines(Codes,PeeledCodes),
     (debug_mode(off) -> true
       ; get_counter(xml2b_current_line,InitialLineNr), format('XML starting at line: ~w~n',[InitialLineNr])),
     %statistics(walltime,[W1,_]),
     xml_parse(PeeledCodes,xml(_Atts,Content),[format(false)]), % format(false) means we also see all newlines
     %statistics(walltime,[W2,_]), W is W2-W1, formatsilent('% Walltime ~w ms to parse XML using Prolog library~n',[W]),
     % xml_pp(xml(_Atts,Content)),nl,
     % xml_parse(Codes2,xml(_Atts,Content),[]), format('XML:~n~s~n',[Codes2]),
     %print(Content),nl,nl,
     !,
     convert_xml_content_to_b(Content,0,1,_,BContent,[]).
convert_xml_to_b(_,_) :- add_error(convert_xml_to_b,'XML Parsing failed'),fail.


convert_xml_content_to_b([],_ParentId,NextId,OutId) --> {OutId=NextId}.
convert_xml_content_to_b([H|T],ParentId,NextId,OutId) -->
   convert_xml_element(H,ParentId,NextId,Id2),
   convert_xml_content_to_b(T,ParentId,Id2,OutId).

% TO DO: do we need to parse namespaces...
% TO DO: store meta-information
% TO DO: adapt line numbers when newlines inside initial tag: <?xml version="1.0" encoding="ASCII"?>
% TO DO: deal with new lines inside tags <Data version= "03.04"> : Prolog library hides those newlines !

convert_xml_element(Element, ParentId,NextId,OutId) -->
    % {get_counter(xml2b_current_line,CurNr),nl,print(el(CurNr,Element)),nl},
    {xml_element(Element,Tag,Attributes,Content,LineNr)},
    !,
    [(int(NextId),
       rec([ % note: fields have to be in order
          field(attributes,BAttributes), % partial function of Attributes
          field(element,string(Tag)),
          field(meta,MetaAttributes),  % such as xmlLineNumber
          field(pId,int(ParentId)), % id of the parent XML record
          field(recId,int(NextId)) % id of the XML element/record
           ])
     )],
    {Id2 is NextId+1,
     number_codes(LineNr,LCC), atom_codes(LineNrAtom,LCC),
     convert_to_avl([(string(xmlLineNumber),string(LineNrAtom))],MetaAttributes),
     maplist(convert_xml_attributes,Attributes,BAttributesList),
     convert_to_avl(BAttributesList,BAttributes)
    },
     convert_xml_content_to_b(Content,NextId,Id2,OutId).
convert_xml_element(namespace( URL, _, Element), ParentId,NextId,OutId) --> % what should we do with the name space ?
    {format('*** Ignoring XML namespace annotation: ~w~n',[URL])},
    convert_xml_element(Element, ParentId,NextId,OutId).
convert_xml_element(pcdata(_Codes),_,Id,Id) --> !. % should only happen when Codes consists solely of newlines
    % no need to count_newlines; already done below in xml_element before C1 \= [].
convert_xml_element(comment(Codes),_,Id,Id) --> !,{count_newlines(Codes)}.
convert_xml_element(instructions(Name, Chars),_,Id,Id) --> !,
    {format('Ignoring XML instructions annotation: ~w : ~w~n',[Name,Chars])}.
convert_xml_element(doctype(Tag, DoctypeId),_,Id,Id) --> !,
    {format('Ignoring XML doctype annotation: ~w : ~w~n',[Tag,DoctypeId])}.
convert_xml_element(cdata(Codes),ParentId,Id,OutId) --> !,
    % TO DO: tab expansion? see http://binding-time.co.uk/wiki/index.php/Parsing_XML_with_Prolog
    convert_xml_element(pcdata(Codes),ParentId,Id,OutId).
convert_xml_element(El,_,Id,Id) -->
    {format('*** Unknown XML element: ~w~n',[El])}.


xml_element(element(Tag,XAttributes,Content),Tag,Attributes,Content,LineNr) :-
   get_counter(xml2b_current_line,CurLineNr),
   (select('='('$attribute_linefeeds',LF),XAttributes,Rest)
     -> LineNr is CurLineNr+LF, Attributes = Rest, %print(inc_line_nr(CurLineNr,LF)),nl,
        set_counter(xml2b_current_line,LineNr)
     ;  LineNr = CurLineNr, Attributes=XAttributes).
xml_element(pcdata(Codes),'CText',['='(text,NewCodes)],[],LineNr) :-
    peel_and_count_newlines(Codes,C1),
    C1 \= [],
    get_counter(xml2b_current_line,LineNr),
    reverse(C1,RC),
    peel_and_count_newlines(RC,RC2), % remove newlines at end; try avoid B strings with newlines in them
    reverse(RC2,NewCodes).

convert_xml_attributes('='(Attr,Codes),(string(Attr),string(Atom))) :- !,atom_codes(Atom,Codes),
   count_newlines(Codes).
convert_xml_attributes(A,(string(error),string(A))) :- format('**** UNKNOWN ATTRIBUTE: ~w~n',[A]).


:- use_module(extension('counter/counter')).

count_newlines([]).
count_newlines(HT) :- newline(HT,T),!,  inc_counter(xml2b_current_line), count_newlines(T).
count_newlines([_|T]) :- count_newlines(T).

% peel leading newlines
peel_and_count_newlines([],[]).
peel_and_count_newlines(HT,Res) :- newline(HT,T),!, inc_counter(xml2b_current_line), peel_and_count_newlines(T,Res).
peel_and_count_newlines([H|T],Res) :- whitespace(H),!, peel_and_count_newlines(T,Res).
peel_and_count_newlines(R,R).


newline([10|T],T).
newline([13|X],T) :- (X=[10|TX] -> T=TX ; T=X). % TO DO: should we check if we are on Windows ?

whitespace(9). % tab
whitespace(32).

xml2b_startup :- % call once at startup to ensure all counters exist
      counter_init,
      new_counter(xml2b_current_line).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(startup_prob,xml2b_startup,
                           'Initialise xml2b Counters.').
