% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(memoization,[store_memo_expansion/3,
                       get_stored_memo_expansion/3,
                       compute_memo_hash/2,
                       store_memo_computation_time/2, % for profiling

                       texpr_is_marked_for_memoization/1,
                       register_memoization_function/6,
                       get_registered_function/2,
                       is_memoization_closure/4,
                       get_memoization_closure_value/4,
                       %expand_memoize_stored_function_reference/2,
                       store_memo_result/6,
                       get_stored_memo_result/5,
                       apply_to_memoize/8,

                       print_memo_profile/0,
                       reset_memo_table/0,
                       reset_memo_with_statistics/0]).

:- use_module(error_manager).
:- use_module(debug).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides memoization features (for closure expansion, closure1,...).').
% used when use_closure_expansion_memoization is true



:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(startup_prob,init_memo_table,
                           'Initialise Preferences').
:- register_event_listener(clear_specification,reset_memo_table,
                           'Reset Temporary Preferences.').

:- use_module(library(terms),[term_hash/3]).



% ---------------- Memoization for set comprehension / lambda / closure1 expansion:

get_stored_memo_expansion(Hash,Term,Result) :-
   stored_memo_expansion(Hash,Term,Result),
   inc_counter(memoization_reuse_counter,_).

:- dynamic stored_memo_expansion/3.

store_memo_expansion(Hash,LookupTerm,StoredResult) :-
   assert(stored_memo_expansion(Hash,LookupTerm,StoredResult)).

compute_memo_hash(LookupTerm,Hash) :-
    inc_counter(memoization_hash_counter,_),
    term_hash(LookupTerm,
            [range(smallint),algorithm(sdbm), depth(infinite),if_var(ignore)],Hash).
    %hashing:my_term_hash(closure(Parameters,ParameterTypes,ClosureBody),Hash),
    %inc(Hash), %

% ---------------- Memoization for function application

:- use_module(probsrc(bsyntaxtree),[get_texpr_description/2]).
% this is currently applied to identifiers in the ABSTRACT_CONSTANTS section
% marking is done using a pragma /*@desc memo */
texpr_is_marked_for_memoization(TExpr) :-
   get_texpr_description(TExpr,memo).

:- dynamic registered_value/3, registered_memo_pred/2.
% we register the functions to be memoized
get_registered_function(Id,Val) :- registered_value(Id,_,Val).

:- use_module(tools,[remove_variables/3]).
:- use_module(kernel_tools,[ground_value_check/2, value_variables/2]).
% register a (usually symbolic closure) value as a function
% ElType is the type of the elements of the function and MemoResultClosure is a closure that
% can and should be used instead of BValue to ensure memoization works
register_memoization_function(Name,BValue,ElType,RecursionInfo,MemoID,MemoResultClosure) :-
    value_variables(BValue,Vars),
    %print(vars(Name,Vars,LHSValue)),nl, print(bvalue(BValue)),nl,
    (RecursionInfo='$recursion_value'(RecValue)
     -> remove_variables(Vars,[RecValue],WVars), % do not wait on RecValue; it will be set by registering MEMOID
        RecValue = MemoResultClosure
      ; WVars=Vars
    ),
    ground_value_check(WVars,Ground),
    %print(register(Name,MemoID,Ground,BValue)),nl,nl,
    construct_memoization_closure(MemoID,BValue,ElType,MemoResultClosure),
    register_memoization_function_aux(Ground,Name,BValue,MemoID).

:- block register_memoization_function_aux(-,?,?,?).
register_memoization_function_aux(_,Name,BValue,MemoID) :-
    inc_counter(memoization_function_counter,MemoID),
    (ground(BValue) -> true ; add_internal_error('Non ground memo closure: ',MemoID:BValue)),
    assert(registered_value(MemoID,Name,BValue)),
    gen_memo_pred(MemoID,MemoPred),
    assert(registered_memo_pred(MemoID,MemoPred)),
    print(registered_memo_pred(Name,MemoID,MemoPred)),nl,
    Fact =.. [MemoPred,'$void','$void','$void'],
    assert((Fact :- fail)),
    retract((Fact :- fail)).

% generate a predicate for asserting the table/memoizations for the function
gen_memo_pred(MemoID,MemoPred) :-
    number_codes(MemoID,NC), append("m",NC,MC), atom_codes(MemoPred,MC).

construct_memoization_closure(MemoID,BValue,ElType,Result) :-
    get_infos(BValue,BInfos),
    XFC = b(external_function_call('MEMOIZE_STORED_FUNCTION',[b(value(int(MemoID)),integer,[])]),set(ElType),[]),
    TID = b(identifier('_zzzz_unary'),ElType,[]),
    Result = closure(['_zzzz_unary'],[ElType],
                          b(member(TID,XFC),pred,
                            [prob_annotation('MEMOIZE'(MemoID))|BInfos])).

:- use_module(bsyntaxtree,[get_texpr_info/2, extract_info/2, extract_pos_infos/2]).
% extract important infos,
get_infos(closure(_,_,B),Res) :- extract_info(B,Infos),!,
   get_texpr_info(B,AllInfos),extract_pos_infos(AllInfos,Pos),
   append(Pos,Infos,Res).
get_infos(_,[]). % :- print(no_infos(A)),nl.

:- use_module(bsyntaxtree, [get_texpr_id/2]).
is_memoization_closure([ID],_T,b(Body,_,INFO),MemoID) :-
   member(prob_annotation('MEMOIZE'(MemoID)),INFO),
   % check Body has not been rewritten:
   Body = member(TID,b(RHS,_,_)),
   get_texpr_id(TID,ID),
   is_memoize_stored_function_reference(RHS,MemoID).

is_memoize_stored_function_reference(external_function_call('MEMOIZE_STORED_FUNCTION',[Arg]),MemoID) :-
   Arg=b(value(int(MemoID)),_,_).

% check if we have a memoization closure and get its stored value
get_memoization_closure_value(P,T,B,Value) :-
    is_memoization_closure(P,T,B,MemoID),
    number(MemoID),
    registered_value(MemoID,_,Value).

% expand an (untyped) expression term referring to a stored function into full value:
expand_memoize_stored_function_reference(Expr,Value) :-
   is_memoize_stored_function_reference(Expr,MemoID),
   number(MemoID), % in case this is called before function registered
   registered_value(MemoID,_,Value).

%:- dynamic stored_memo_result/4.
% we now store for each function ID the result in mID/3.

% last argument should be left free when calling
get_stored_memo_result(MemoID,MemoPred,Argument,ArgHash,StoredResult) :-
   %stored_memo_result(ArgHash,MemoID,Argument,StoredResult),
   call(MemoPred,ArgHash,Argument,StoredResult),
   %print(stored(MemoID,ArgHash,StoredResult)),nl,
   inc_counter(memoization_function_reuse_counter,_).

stored_memo_result(ArgHash,MemoID,Argument,StoredResult) :-
   registered_memo_pred(MemoID,MemoPred),
   call(MemoPred,ArgHash,Argument,StoredResult).

:- use_module(kernel_waitflags,[add_wd_error_span/4]).
store_memo_result(MemoID,MemoPred,Argument,ArgHash,Result,Span,WF) :-
   %stored_memo_result(ArgHash,MemoID,Argument,StoredResult),
   call(MemoPred,ArgHash,Argument,StoredResult),
   !,
   (Result=StoredResult -> true
     ; add_wd_error_span('memoised function has multiple results for: ','@fun'(Argument,MemoID),Span,WF)
   ).
store_memo_result(MemoID,MemoPred,Argument,ArgHash,Result,_,_) :-
   %registered_memo_pred(MemoID,MemoPred),
   Fact =.. [MemoPred,ArgHash,Argument,Result],
   assert(Fact).
   %assert(stored_memo_result(ArgHash,MemoID,Argument,Result)).
   %print(stored_memo_result(ArgHash,MemoID,Argument,Result)),nl.

%callm(1,_,ArgHash,Argument,StoredResult) :- !, m1(ArgHash,Argument,StoredResult).
%callm(2,_,ArgHash,Argument,StoredResult) :- !, m2(ArgHash,Argument,StoredResult).
%callm(3,_,ArgHash,Argument,StoredResult) :- !, m3(ArgHash,Argument,StoredResult).
%callm(_,MemoPred,ArgHash,Argument,StoredResult) :- call(MemoPred,ArgHash,Argument,StoredResult).
%
%
%assert_table(1,_,ArgHash,Argument,Result) :- !,assert(m1(ArgHash,Argument,Result)).
%assert_table(2,_,ArgHash,Argument,Result) :- !,assert(m2(ArgHash,Argument,Result)).
%assert_table(3,_,ArgHash,Argument,Result) :- !,assert(m3(ArgHash,Argument,Result)).
%assert_table(4,_,ArgHash,Argument,Result) :- !,assert(m4(ArgHash,Argument,Result)).
%assert_table(5,_,ArgHash,Argument,Result) :- !,assert(m5(ArgHash,Argument,Result)).
%assert_table(_,MemoPred,ArgHash,Argument,Result)  :-
%   Fact =.. [MemoPred,ArgHash,Argument,Result],
%   assert(Fact).

% function application for memoized function:
apply_to_memoize(MemoID,P,T,B,X,Y,Span,WF) :-
        ground_value_check(X,XV),
        block_apply_closure_memo_groundx(XV,MemoID,X,Y,P,T,B,Span,WF).

:- use_module(kernel_tools,[ground_value/1]).
:- use_module(kernel_objects,[equal_object_wf/3]).

:- block block_apply_closure_memo_groundx(-,?, ?,?, ?,?,?, ?,?),
         block_apply_closure_memo_groundx(?,-,?,?,?,?,?,?,?).

% this clause is only ok if we have just a single function; otherwise we get collisions between different functions
compute_funcall_memo_hash(int(X),_,R) :- !,
   R=X. % avoid computing hash for simple integer values; only drawback: indexing not good for too large integers
% TO DO: we could register with each registered_value fact also a dynamic table_memo fact
% then we can avoid clashes between different memoized functions
compute_funcall_memo_hash(X,MemoID,ArgHash) :-
   %compute_memo_hash((MemoID,X),ArgHash).
   compute_memo_hash(X,ArgHash).

block_apply_closure_memo_groundx(_,MemoID, X,Y, P,T,B, Span,WF) :-
    % MemoID is ground, meaning body B is ground as well
    % TODO: obtain stored value directly
    compute_funcall_memo_hash(X,MemoID,ArgHash),
    registered_memo_pred(MemoID,MemoPred),
    if(get_stored_memo_result(MemoID,MemoPred,X,ArgHash,StoredResult),
        equal_object_wf(Y,StoredResult,WF),
        if(check_element_of_function_closure_nowf(MemoID,X,ComputedResult,P,T,B),
           % ground WF to compute ground value for ComputedResult
             ((ground_value(ComputedResult)
               -> store_memo_result(MemoID,MemoPred,X,ArgHash,ComputedResult,Span,WF)
               ;  add_error(memoization,'Could not compute ground value for function application',MemoID,Span),
                  print('Argument: '),translate:print_bvalue(X),nl,
                  print('Result: '), print(ComputedResult),nl,
                  print('Function: '),registered_value(MemoID,_Name,FunValue), translate:print_bvalue(FunValue),nl,nl
               ),
               % To do: check if we compute two or more solutions for f(X) and add WD error
               equal_object_wf(Y,ComputedResult,WF)
             ),
             add_wd_error_span('function applied outside of domain (#8-Memo): ', '@fun'(X,closure(P,T,B)),Span,WF)
           )
      ).

:- use_module(kernel_waitflags,    [init_wait_flags/1,ground_wait_flags/1]).
:- use_module(custom_explicit_sets,[check_element_of_function_closure/6]).
check_element_of_function_closure_nowf(MemoID,X,ComputedResult,_,_,_) :-
   registered_value(MemoID,_,closure(P,T,B)),!,
   init_wait_flags(WF),
   check_element_of_function_closure(X,ComputedResult,P,T,B,WF),
   ground_wait_flags(WF).
check_element_of_function_closure_nowf(_,X,ComputedResult,P,T,B) :-
   init_wait_flags(WF),
   check_element_of_function_closure(X,ComputedResult,P,T,B,WF),
   ground_wait_flags(WF).


% -------------------------------
% utilities to profile expansions
:- public inc/1, store_memo_computation_time/2.
:- dynamic hash_count/2.
inc(Hash) :- (retract(hash_count(Hash,Nr)) -> N1 is Nr+1 ; N1 = 1), assert(hash_count(Hash,N1)).
:- dynamic hash_time/2.
store_memo_computation_time(Hash,Time) :-
   retractall(hash_time(Hash,_)), assert(hash_time(Hash,Time)).

print_memo_profile :-
   ((stored_memo_expansion(_,_,_) ; registered_value(_,_,_))
     -> print('MEMO Table:'),nl,
        print_memo_table
     ; true).

% memoization:print_memo.
print_memo_table :-
  stored_memo_expansion(Hash,LookupTerm,PackedValue),
  (hash_count(Hash,Count) ->  print(Count), print(' : ') ; true),
  (hash_time(Hash,Time) -> print(Time),print(' --> ') ; true),
  print_memo_entry(LookupTerm),
  nl, print('     :  '),translate:print_bvalue(PackedValue),nl,fail.
print_memo_table :-
  registered_value(MemoID,Name,FunValue),
  format('MemoID ~w stored FUNCTION ~w ',[MemoID,Name]),
  nl, print('     :  '),translate:print_bvalue(FunValue),nl,fail.
print_memo_table :-
  stored_memo_result(_Hash,MemoID,LookupTerm,PackedValue),
  format('MemoID ~w result for argument ',[MemoID]),
  print_memo_entry(LookupTerm),
  nl, print('     :  '),translate:print_bvalue(PackedValue),nl,fail.
print_memo_table :-
  stored_memo_result(Hash,MemoID1,T1,_),
  stored_memo_result(Hash,MemoID2,T2,_),
  dif((MemoID1,T1),(MemoID2,T2)),
  format('Hash COLLISION ~w, ~w:~w vs ~w:~w~n',[Hash,MemoID1,T1,MemoID2,T2]),
  fail.
print_memo_table :- get_counter(memoization_hash_counter,Hash),
  get_counter(memoization_reuse_counter,RHit),
  format('Hashes computed: ~w, expansions reused: ~w~n',[Hash,RHit]),
  get_counter(memoization_function_counter,F),
  get_counter(memoization_function_reuse_counter,FR),
  format('Memoization functions registered: ~w, results reused: ~w~n',[F,FR]).

print_memo_entry(closure(Parameters,ParameterTypes,ClosureBody)) :- !,
  translate:print_bvalue(closure(Parameters,ParameterTypes,ClosureBody)).
print_memo_entry(closure1_for_explicit_set(A)) :- !,
  print('closure1('),translate:print_bvalue(avl_set(A)), print(')').
print_memo_entry(A) :- translate:print_bvalue(A).

:- use_module(extension('counter/counter'),[counter_init/0, reset_counter/1,
                                            new_counter/1, inc_counter/2,
                                            get_counter/2]).
init_memo_table :-
   counter_init,
   new_counter(memoization_reuse_counter), new_counter(memoization_hash_counter),
   new_counter(memoization_function_counter), new_counter(memoization_function_reuse_counter),
   reset_memo_table.
reset_memo_table :-
   reset_counter(memoization_reuse_counter),
   reset_counter(memoization_hash_counter),
   reset_counter(memoization_function_counter),
   reset_counter(memoization_function_reuse_counter),
   (stored_memo_expansion(_,_,_) -> debug_println(19,'Resetting MEMO table') ; true),
   retractall(hash_count(_,_)),
   retractall(hash_time(_,_)),
   retractall(stored_memo_expansion(_,_,_)),
   retractall(registered_value(_,_,_)),
   ( retract(registered_memo_pred(_,MemoPred)),
     Fact =.. [MemoPred,_,_,_],
     retractall(Fact),fail
     ; true).

:- use_module(probsrc(tools),[retract_with_statistics/2]).
reset_memo_with_statistics :-
   retract_with_statistics(memoization,
      [stored_memo_expansion(_,_,_),registered_value(_,_,_),stored_memo_result(_,_,_,_)]),
   reset_memo_table.

