% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% This module provides a plugin mechanism for ProB
% Each plugin has an unique ID. It should be registered in the module
% plugin_registry

:- module(plugins, [initialise_plugins/0
                   ,is_registered_plugin/1,is_registered_plugin/2
                   ,active_plugin/1,tcl_plugin_syntax_colouring/2
                   ,plugin_file_extensions/1
                   ,available_plugins_for_mode/1,plugin_is_available_for_mode/0
                   ,plugin_has_preferences/0,plugins_with_preferences/1
                   ,plugin_has_output/1
                   ,plugin_get_output/2
                   ,is_plugin_file_extension/2
                   ,load_plugin_file/2
                   ,setup_transition_skeleton/2
                   ,plugin_transition/5
                   ,plugin_state_property/3
                   ,plugin_parse/7
                   ,plugin_compute_expression/4
                   ,plugin_evaluate_predicate/3
                   ,plugin_evaluate_transition_predicate/3
                   ,notify_new_transition_to_plugin/4 ,notify_new_transition_to_plugin/5
                   ,plugin_is_initialised_state/2
                   ,start_plugin_for_current_mode/1,stop_plugin/0
                   ,plugin_prettyprint_transition/3
                   ,plugin_prettyprint_property/3
                   ,plugin_prettyprint_value/3
                   % ,plugin_advanced_eclipse_preference/2 % TO DO: not used nor defined at the moment
                   ,plugin_preference_default_value/2
                   ,plugin_preference_description/2
                   ,plugin_preference_val_type/2
                   ,plugin_preference_category/2
                   ,plugin_check_invariantKO/3
                   ,plugin_get_internal/2
                   ,compiletime_init_plugins/0
                   ]).

:- use_module(library(lists)).
:- use_module(library(codesio)).
:- use_module(library(file_systems)).

:- use_module(tools_strings,[ajoin/2]).
:- use_module(module_information,[module_info/2]).

:- use_module(specfile,[animation_mode/1,animation_minor_mode/1,set_animation_mode/1,set_animation_minor_mode/1]).

:- use_module(error_manager,[add_error/2]).
:- use_module(preferences,[get_preference/2]).

:- module_info(group,infrastructure).
:- module_info(description,'A plugin mechanism for ProB.').

:- public author/2. % for Spider
:- dynamic
    plugin/2,
    pname/2,
    status/2,
    author/2,
    file_extensions/2,
    file_loader/2,
    transition_skeleton/2,
    transition_pred/2,
    initialisation_pred/2,
    state_property_pred/2,
    parser_pred/2,
    compute_expression_pred/2,
    evaluate_predicate_pred/2,
    evaluate_transition_pred/2,
    is_initialised_state_pred/2,
    prettyprint_transition_pred/2,
    prettyprint_property_pred/2,
    prettyprint_value_pred/2,
    syntax_colouring/2,
    plugin_for_modes/2,
    plugin_init_pred/2,
    plugin_new_transition_pred/2,
    plugin_preferences_pred/2,
    plugin_check_invariant_pred/2,
    plugin_output_pred/2,
    plugin_internal_representation_pred/2.

% TODO: deactivated plugin mechanism for next release -- keeping the number of
% modules smaller until the report is better structured
% load_compiletime_plugins :- !.
load_compiletime_plugins :- print(load_compiletime_plugins),nl,
    findall(Module,
            ( directory_member_of_directory(plugins('.'),Dir,Fullname),
                atom_concat(Dir,'.pl',PrologFilename),
                file_member_of_directory(Fullname,PrologFilename,_),
                ajoin([Dir,'/',PrologFilename], Modulename),
                Module = plugins(Modulename) ),
            Plugins),
    %print(loading_plugins(Plugins)),nl,
    load_plugins2(Plugins).

% these plug-ins are not stable enough to load for the normal user:
do_not_load_plugin(plugins('probvm/probvm.pl')).

load_plugins2([]).
load_plugins2([Module|Prest]) :- do_not_load_plugin(Module),!, load_plugins2(Prest).
load_plugins2([Module|Prest]) :-
    load_plugin(Module,Id),
    (nonvar(Id) -> assert(plugin(Id,Module)); true),
    load_plugins2(Prest).

load_plugin(Module,Id) :-
    catch( ( load_plugin2(Module,Id) ->
               ajoin(['Loaded plugin ', Id], Msg),
               print_message(informational, Msg)
           ; otherwise ->
               ajoin(['Loading of module ', Module, ' failed: '], Msg),
               print_message(error, Msg)),
           E,
           ( ajoin(['Loading of module ', Module, ' raised an exception: ', E], Msg),
               print_message(error, Msg))).

load_plugin2(Module,Id) :-
    use_module(IdVar,Module,[]), % SICStus 4.3.3 now keeps IdVar a variable !
    (var(IdVar) -> get_plugin_module_name(Module,Id)
      ; IdVar = Id),
    %print(used_module(Id,Module)),nl,
    ( plugin(Id,_) ->
        ajoin(['Plugin ID ',Id,' used twice'],Msg),
        print_message(error,Msg),
        fail
    ; otherwise ->
        true).

:- use_module(tools,[get_modulename_filename/2]).
get_plugin_module_name(plugins(F),ModuleName) :- !, tools:get_modulename_filename(F,ModuleName).
get_plugin_module_name(F,ModuleName) :- tools:get_modulename_filename(F,ModuleName).

%:- load_compiletime_plugins. % now done here : compiletime_init_plugins

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(compile_prob,compiletime_init_plugins,
                           'Load compile time plugins and initialise them.').
compiletime_init_plugins :- load_compiletime_plugins,initialise_plugins.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

initialise_plugins :-
    reset_plugin_db,
    fail.
initialise_plugins :-
    plugin(Id,_Module),
    initialise_plugin(Id),
    fail.
initialise_plugins.

reset_plugin_db :-
    retractpred(pname/2),
    retractpred(status/2),
    retractpred(author/2),
    retractpred(file_extensions/2),
    retractpred(file_loader/2),
    retractpred(transition_skeleton/2),
    retractpred(transition_pred/2),
    retractpred(initialisation_pred/2),
    retractpred(state_property_pred/2),
    retractpred(parser_pred/2),
    retractpred(compute_expression_pred/2),
    retractpred(evaluate_predicate_pred/2),
    retractpred(evaluate_transition_pred/2),
    retractpred(is_initialised_state_pred/2),
    retractpred(prettyprint_transition_pred/2),
    retractpred(prettyprint_property_pred/2),
    retractpred(prettyprint_value_pred/2),
    retractpred(syntax_colouring/2),
    retractpred(plugin_for_modes/2),
    retractpred(plugin_init_pred/2),
    retractpred(plugin_new_transition_pred/2),
    retractpred(plugin_preferences_pred/2),
    retractpred(plugin_check_invariant_pred/2),
    retractpred(plugin_output_pred/2),
    retractpred(plugin_internal_representation_pred/2).

retractpred(F/A) :-
    functor(Fact,F,A), retractall(Fact).

initialise_plugin(Id) :-
    plugin(Id,_Module),
    call(Id:plugin_info(Info)),
    %print(init_plugin(Id,Info)),nl,
    initialise_plugin2(Id,Info),
    !.

initialise_plugin2(Id,Info) :-
    %check_options(Info,Id),
    register_infos(Info,Id,_Used).
register_infos([],_Id,[]).
register_infos([Info|IRest],Id,Used) :-
    register_info(Info,Id,U),
    append(U,RUsed,Used),
    register_infos(IRest,Id,RUsed).
register_info(Info,Id,Used) :-
    ( Info = (Name=Value) ->
        ( register_info2(Name,Value,Id) ->
            Used = [Name]
        ; otherwise ->
            ajoin(['Failed analysing plugin info for plugin ', Id,
                   ': ', Name, ' = ', Value], Msg),
            print_message(error, Msg),
            Used = [])
    ; otherwise ->
        ajoin(['Invalid plugin info for plugin ', Id,
               ': ', Info], Msg),
        print_message(error, Msg),
        Used = []).

:- public check_options/2.
check_options(Info,Id) :-
    findall(O,mandatory_option(O),Mandatory),
    check_options2(Mandatory,Info,Id,_).
check_options2([],_Info,_Id,true).
check_options2([Mandatory1|Mrest],Info,Id,Verdict) :-
    ( is_list(Mandatory1) -> Mandatory = Mandatory1
    ; otherwise -> Mandatory = [Mandatory1]),
    check_option(Mandatory,Info,Id,Verdict),
    check_options2(Mrest,Info,Id,Verdict).

check_option(Mandatory,Info,_Id,_Verdict) :-
    member(M,Mandatory),
    memberchk( (M=_), Info),!.
check_option(Mandatory,_Info,Id,false) :-
    ( Mandatory = [M] ->
        Msg = ['Missing plugin info ', M]
    ; otherwise ->
        Msg = ['Missing one plugin info of ', Mandatory]),
    ajoin(['Plugin ',Id,': '|Msg], Msg1),
    print_message(error,Msg1),
    fail.

mandatory_option(name).
mandatory_option(load_file).
mandatory_option([file_extensions,plugin_for_modes]).
mandatory_option(transition).
mandatory_option(initialisation).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% handling options

register_info2(name,Name,Id) :-
    regassert(Id,name,pname(Name)).
register_info2(status,Status,Id) :-
    ( memberchk(Status,[stable,unstable]) ->
        regassert(Id,name,status(Status))
    ; otherwise ->
        ajoin(['Plugin ',Id,': Invalid status, ignoring'],Msg),
        print_message(warning,Msg)).
register_info2(load_file,Proc,Id) :-
    reg_proc_assert(Id,load_file,file_loader,Proc,1).
register_info2(author,Author,Id) :-
    regassert(Id,author,author(Author)).
register_info2(file_extensions,Ext,Id) :-
    regassert(Id,file_extensions,file_extensions(Ext)).
register_info2(transition_skeleton,Proc,Id) :-
    reg_proc_assert(Id,transition_skeleton,transition_skeleton,Proc,1).
register_info2(transition,Proc,Id) :-
    reg_proc_assert(Id,transition,transition_pred,Proc,4).
register_info2(initialisation,Proc,Id) :-
    reg_proc_assert(Id,initialisation,initialisation_pred,Proc,3).
register_info2(state_property,Proc,Id) :-
    reg_proc_assert(Id,state_property,state_property_pred,Proc,2).
register_info2(parser,Proc,Id) :-
    reg_proc_assert(Id,parser,parser_pred,Proc,6).
register_info2(compute_expression,Proc,Id) :-
    reg_proc_assert(Id,compute_expression,compute_expression_pred,Proc,3).
register_info2(evaluate_predicate,Proc,Id) :-
    reg_proc_assert(Id,evaluate_predicate,evaluate_predicate_pred,Proc,2).
register_info2(evaluate_transition,Proc,Id) :-
    reg_proc_assert(Id,evaluate_transition,evaluate_transition_pred,Proc,2).
register_info2(is_initialised_state,Proc,Id) :-
    reg_proc_assert(Id,is_initialised_state,is_initialised_state_pred,Proc,1).
register_info2(prettyprint_transition,Proc,Id) :-
    reg_proc_assert(Id,prettyprint_transition,prettyprint_transition_pred,Proc,2).
register_info2(prettyprint_property,Proc,Id) :-
    reg_proc_assert(Id,prettyprint_property,prettyprint_property_pred,Proc,2).
register_info2(prettyprint_value,Proc,Id) :-
    reg_proc_assert(Id,prettyprint_value,prettyprint_value_pred,Proc,2).
register_info2(syntax_colouring,Rules,Id) :-
    check_syntax_colouring(Rules,Id),
    regassert(Id,syntax_colouring,syntax_colouring(Rules)).
register_info2(plugin_for_modes,Modes,Id) :-
    regassert(Id,syntax_colouring,plugin_for_modes(Modes)).
register_info2(plugin_init,Proc,Id) :-
    reg_proc_assert(Id,plugin_init,plugin_init_pred,Proc,0).
register_info2(new_transition_notification,Proc,Id) :-
    reg_proc_assert(Id,new_transition_notification,plugin_new_transition_pred,Proc,4).
register_info2(preferences,Proc,Id) :-
    reg_proc_assert(Id,preferences,plugin_preferences_pred,Proc,4).
register_info2(check_invariant,Proc,Id) :-
    reg_proc_assert(Id,check_invariant,plugin_check_invariant_pred,Proc,2).
register_info2(output,Proc,Id) :-
    reg_proc_assert(Id,output,plugin_output_pred,Proc,3).
register_info2(internal_representation,Proc,Id) :-
    reg_proc_assert(Id,internal_representation,plugin_internal_representation_pred,Proc,1).

% helper predicates

regassert(Id,Name,Fact) :-
    functor(Fact,F,Arity),
    NArity is Arity+1,
    functor(Check,F,NArity), arg(1,Check,Id),
    ( call(Check) ->
        ajoin(['Plugin ', Id, ': multiple occurence of info ',
               Name, ', ignoring'],Msg),
        print_message(warning,Msg)
    ; otherwise ->
        Fact=..[F|Args],
        NFact=..[F,Id|Args],
        assert(NFact)).

reg_proc_assert(Id,Name,Fact,Proc,Arity) :-
    ( Proc = M:P/Arity -> true
    ; Proc = P/Arity   -> M:P/Arity = Id:Proc
    ; otherwise ->
        ajoin(['Predicate in info ',Name,' of plugin ',
               Id,' is not valid: ',Proc],Msg),
        print_message(error, Msg),fail),
    Fact =.. [F|Args], append(Args,[M:P/Arity],AllArgs),
    NFact =.. [F|AllArgs],
    regassert(Id,Name,NFact).

check_syntax_colouring(Rules,Id) :-
    ( syntax_colouring_error(Rules,Error) ->
        ajoin(['Error in syntax colouring rules for plugin ',Id,
               ': ', Error],Msg),
        print_message(warning,Msg),
        fail
    ; otherwise ->
        true).
syntax_colouring_error([],_) :- !,fail.
syntax_colouring_error([Rule|_],Error) :-
    syntax_colouring_error2(Rule,Error),!.
syntax_colouring_error([_|Rest],Error) :- !,
    syntax_colouring_error(Rest,Error).
syntax_colouring_error(_,'list expected').

syntax_colouring_error2(expression(_Regex,Type),Error) :- !,
    check_syntax_colouring_type(Type,Error).
syntax_colouring_error2(pair(_RegexA,_RegexB,Type),Error) :- !,
    check_syntax_colouring_type(Type,Error).
syntax_colouring_error2(_,'expected expression(Regex,Type) or pair(Begin,End,Type)').

check_syntax_colouring_type(Type,Error) :-
    Keys = [syntax_comment,syntax_keyword,syntax_type,
            syntax_assignment, syntax_operator, syntax_logical,
            syntax_unsupported],
    ( memberchk(Type,Keys) -> fail
    ; otherwise ->
        ajoin(['unexpected colouring type ',Type],Error)).

% :- initialise_plugins. now done in compiletime_init_plugins

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

active_plugin(Id) :-
    animation_mode(Id),
    is_registered_plugin(Id).

is_registered_plugin(Id) :-
    pname(Id,_Name).

is_registered_plugin(Id,Name) :-
    pname(Id,Name).

plugin_is_stable(Id) :-
    ( status(Id,Status) -> Status==stable
    ; otherwise -> true).

plugin_file_extensions(list(Ext)) :-
    get_preference(user_is_an_expert_with_accessto_source_distribution,Expert),
    findall( list([Description, list(Suffixes)]),
             ( file_extensions(Id,Ext),
                 ( Expert==true -> true ; plugin_is_stable(Id) ),
                 member( (Description,Suffixes), Ext)),
             Ext).

available_plugins_for_mode(list(Plugins)) :-
    get_preference(user_is_an_expert_with_accessto_source_distribution,Expert),
    animation_mode(Mode),
    findall( list([Pl,Name]),
             ( plugin_for_modes(Pl,Modes),
                 ( Expert==true -> true ; plugin_is_stable(Pl) ),
                 member(Mode,Modes),
                 pname(Pl,Name)),
             Plugins).

plugin_is_available_for_mode :-
    available_plugins_for_mode(list([_|_])),!.

plugins_with_preferences(list(Plugins)) :-
    get_preference(user_is_an_expert_with_accessto_source_distribution,Expert),
    findall( list([Pl,Name]),
             ( is_registered_plugin(Pl),
                 ( Expert==true -> true ; plugin_is_stable(Pl) ),
                 plugin_preferences_pred(Pl,_),
                 pname(Pl,Name)),
             Plugins).

plugin_has_preferences :-
    plugins_with_preferences(list([_|_])),!.

plugin_get_output(list(Outputs),SortOfOutput) :-
    active_plugin(Id),
    plugin_output_pred(Id,Module:Pred/3),
    findall(list([Name,Output]),
            (call(Module:Pred,Name,Output,Sorts),member(SortOfOutput,Sorts)),
            Outputs).

plugin_get_output(Name,Output) :-
    active_plugin(Id),
    plugin_output_pred(Id,Module:Pred/3),
    call(Module:Pred,Name,Output,_).

plugin_has_output(SortOfOutput) :-
    plugin_get_output(list([_|_]),SortOfOutput).

start_plugin_for_current_mode(Id) :-
    animation_mode(Mode),
    is_registered_plugin(Id),
    set_animation_mode(Id),
    set_animation_minor_mode(Mode),
    ( plugin_init_pred(Id,Module:Pred/0) ->
        call(Module:Pred)
    ; otherwise -> true).

stop_plugin :-
    animation_minor_mode(Mode),
    set_animation_mode(Mode).
stop_plugin :-
    set_animation_mode(b).

tcl_plugin_syntax_colouring(Id,list(Rules)) :-
    findall(L, tcl_syntax_colouring2(Id,L), Rules).
tcl_syntax_colouring2(Id,TclRule) :-
    syntax_colouring(Id,Rules),
    member(PRule,Rules),
    tcl_syntax_colouring3(PRule,TclRule).
tcl_syntax_colouring3(expression(Pattern,Type), list([expression,Pattern,Type])).
tcl_syntax_colouring3(pair(Begin,End,Type), list([pair,Begin,End,Type])).

is_plugin_file_extension(Ext,Id) :-
    file_extensions(Id,AllExtensions),
    member( (_,Extensions), AllExtensions),
    member(Ext,Extensions), !.

load_plugin_file(Id,Filename) :-
    user:tcltk_clear_machine,
    ( plugin_init_pred(Id,Module:Pred/0) ->
        call(Module:Pred)
    ; otherwise -> true),
    file_loader(Id,Module:Pred/1),
    call(Module:Pred,Filename),!,
    set_animation_mode(Id).

setup_transition_skeleton(Id,Skeleton) :-
    transition_skeleton(Id,Module:Pred/1),!,
    call(Module:Pred,Skeleton).
setup_transition_skeleton(_Id,_Skeleton).

plugin_transition(Id,Org,Trans,Dst,Infos) :-
    ( Org==root ->
        initialisation_pred(Id,Module:Pred/3),
        call(Module:Pred,Trans,Dst,Infos)
    ; otherwise ->
        transition_pred(Id,Module:Pred/4),
        call(Module:Pred,Org,Trans,Dst,Infos)).

plugin_state_property(Id,State,Property) :-
    State \== root,
    ( state_property_pred(Id,Module:Pred/2) ->
        call(Module:Pred,State,Property)
    ; otherwise ->
        Property = State).

plugin_parse(Id,ExprStrings,PredStrings,TransPredStrings,Exprs,Preds,Trans) :-
    parser_pred(Id,Module:Pred/6),
    call(Module:Pred,ExprStrings,PredStrings,TransPredStrings,Exprs,Preds,Trans).

plugin_compute_expression(Id,Expression,State,Result) :-
    compute_expression_pred(Id,Module:Pred/3),
    call(Module:Pred,Expression,State,Result).

plugin_evaluate_predicate(Id,Predicate,State) :-
    evaluate_predicate_pred(Id,Module:Pred/3),
    call(Module:Pred,Predicate,State).

plugin_evaluate_transition_predicate(Id,Predicate,Transition) :-
    evaluate_transition_pred(Id,Module:Pred/2),
    call(Module:Pred,Predicate,Transition).

plugin_is_initialised_state(Id,State) :-
    ( is_initialised_state_pred(Id,Module:Pred/1) ->
        call(Module:Pred,State)
    ; otherwise -> true).

plugin_get_internal(Id,InternalRepresentation) :-
    active_plugin(Id),
    plugin_internal_representation_pred(Id,Module:Pred/1),
    call(Module:Pred,InternalRepresentation).

plugin_prettyprint_transition(Id,Transition,Atom) :-
    ( prettyprint_transition_pred(Id,Module:Pred/2) ->
        call(Module:Pred,Transition,Codes)
    ; otherwise ->
        write_to_codes(Transition,Codes)),
    ( atomic(Codes) -> Atom=Codes
    ; otherwise -> atom_codes(Atom,Codes)).

plugin_prettyprint_property(Id,Property,Atom) :-
    ( prettyprint_property_pred(Id,Module:Pred/2) ->
        call(Module:Pred,Property,Codes)
    ; otherwise ->
        write_to_codes(Property,Codes)),
    ( atomic(Codes) -> Atom=Codes
    ; otherwise -> atom_codes(Atom,Codes)).

plugin_prettyprint_value(Id,Property,Codes) :-
    ( prettyprint_value_pred(Id,Module:Pred/2) ->
        call(Module:Pred,Property,Codes)
    ; otherwise ->
        write_to_codes(Property,Codes)).

notify_new_transition_to_plugin(FromId,TransId,DstId,Exists) :-
    animation_mode(Mode),
    notify_new_transition_to_plugin(Mode,FromId,TransId,DstId,Exists).

notify_new_transition_to_plugin(ModeId,FromId,TransId,DstId,Exists) :-
    plugin_new_transition_pred(ModeId,Module:Pred/4),
    !,
    ( call(Module:Pred,FromId,TransId,DstId,Exists) ->
        true
    ; otherwise ->
        ajoin(['Notification of new transition failed for plugin ',ModeId,', ',call(Pred,FromId,TransId,DstId,Exists)],Msg),
        add_error(plugins,Msg)).
notify_new_transition_to_plugin(_Id,_FromId,_TransId,_DstId,_Exists).

plugin_preference_default_value(plugin(Id,Key),Value) :-
    is_registered_plugin(Id),
    plugin_preferences_pred(Id,Module:Pred/4),
    call(Module:Pred,Key,Value,_,_).

plugin_preference_description(plugin(Id,Key),Description) :-
    is_registered_plugin(Id),
    plugin_preferences_pred(Id,Module:Pred/4),
    call(Module:Pred,Key,_,Description,_).

plugin_preference_val_type(plugin(Id,Key),Type) :-
    is_registered_plugin(Id),
    plugin_preferences_pred(Id,Module:Pred/4),
    call(Module:Pred,Key,_,_,Type).

plugin_preference_category(plugin(Id,Key), plugin(Id)) :-
    is_registered_plugin(Id),
    plugin_preferences_pred(Id,Module:Pred/4),
    call(Module:Pred,Key,_,_,_).

plugin_check_invariantKO(IdPlugin,IdProB,State) :-
    plugin_check_invariant_pred(IdPlugin,Module:Pred/2),
    call(Module:Pred,IdProB,State).


