% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(unit_parser, [
                         unit_args_to_list/2
                       ]).

:- use_module(library(lists)).

:- use_module(probsrc(self_check), [assert_must_succeed/1]).
:- use_module(probsrc(parsercall), [parse_formula/2]).
:- use_module(probsrc(error_manager), [add_error/3]).

:- use_module(units_alias).
:- use_module(units_domain, [units_domain_division/3,power_of_multiply_units/3]).

:- use_module(probsrc(module_information), [module_info/2]).
:- module_info(group,plugin_units).
:- module_info(description,'Units Plugin: Parsing unit expressions inside pragmas to internal representation of units.').

%unit_args_to_list(X,_) :- format('Input unit argument: ~w~n', [X]), nl, fail.
:- assert_must_succeed(unit_args_to_list("m",[[0,m,1]])).
:- assert_must_succeed(unit_args_to_list("degC",[[0,degC,1]])).
:- assert_must_succeed(unit_args_to_list("cm",[[-2,m,1]])).
:- assert_must_succeed(unit_args_to_list("10*m**1",[[1,m,1]])).
:- assert_must_succeed(unit_args_to_list("10**-1 * m**1",[[-1,m,1]])).
:- assert_must_succeed(unit_args_to_list("10 * m",[[1,m,1]])).
:- assert_must_succeed(unit_args_to_list("m * 10",[[1,m,1]])).
:- assert_must_succeed(unit_args_to_list("km/h",[[3,m,1],[0,h,-1]])).
:- assert_must_succeed(unit_args_to_list('km/h',[[3,m,1],[0,h,-1]])).
unit_args_to_list(Arg,O) :- atom(Arg), Arg \= [], !,
    atom_codes(Arg,ArgC), % also allow units to be passed as atoms
    unit_args_to_list(ArgC,O).
unit_args_to_list(Arg,O) :-
    catch(parse_formula(Arg,Expression),_Exception, fail),
    expression_to_unit(Expression,O), correct_unit(O), !.
unit_args_to_list(Arg,_) :-
    atom_codes(ArgAtom,Arg),
    add_error(incorrect_unit_definition, 'Incorrect unit definition: ', ArgAtom).

%expression_to_unit(Expr,_) :- format('Input unit expression: ~w~n', [Expr]), nl, fail.
:- assert_must_succeed(expression_to_unit(identifier(pos(9,1,4,1,4,1),m),[[0,m,1]])).
:- assert_must_succeed(expression_to_unit(mult_or_cart(pos,power_of(pos,integer(pos,10),unary_minus(pos,integer(pos,1))),power_of(pos,identifier(pos,m),integer(pos,1))),[[-1,m,1]])).
expression_to_unit(identifier(_Pos,Name),Out) :-
    unit_alias(Name,AliasDef)
    -> Out = AliasDef
    ; valid_unit_symbol(Name),
      Out = [[0,Name,1]].
expression_to_unit(power_of(_Pos,identifier(Pos,Name),Exponent),Unit) :- !,
    expression_to_unit(identifier(Pos,Name),UnitOfIdentifier),
    unary_minus_exponent(Exponent,NumericExponent),
    power_of_multiply_units(NumericExponent,UnitOfIdentifier,Unit).
expression_to_unit(power_of(_Pos,integer(_Pos2,10),Exponent),Unit) :- !,
    unary_minus_exponent(Exponent,NumericExponent),
    Unit = [[NumericExponent,na,na]].
expression_to_unit(mult_or_cart(_Pos,Arg1,Arg2),Unit) :- !,
    expression_to_unit(Arg1,Unit1), expression_to_unit(Arg2,Unit2),
    multiply_units(Unit1,Unit2,Unit).
expression_to_unit(div(_Pos,Arg1,Arg2),Unit) :- !,
    expression_to_unit(Arg1,Unit1), expression_to_unit(Arg2,Unit2),
    units_domain_division(Unit1,Unit2,Unit).
expression_to_unit(integer(_Pos,10),[[1,na,na]]) :- !.

unary_minus_exponent(unary_minus(_Pos,integer(_Pos2,X)), MX) :- !, MX is -X.
unary_minus_exponent(integer(_Pos,X), X) :- !.

multiply_units([[X,na,na]],[[0,Unit,X2]],[[X,Unit,X2]]) :- !, Unit \= na, X2 \= na.
multiply_units([[0,Unit,X2]],[[X,na,na]],[[X,Unit,X2]]) :- !, Unit \= na, X2 \= na.
multiply_units(U1,U2,U) :- !, append(U1,U2,U).

correct_unit(ListOfUnits) :-
    maplist(nafree,ListOfUnits).

nafree([A,B,C]) :- A \= na, B \= na, C \= na.

valid_unit_symbol(m).
valid_unit_symbol(kg).
valid_unit_symbol(s).
valid_unit_symbol('A').
valid_unit_symbol('K').
valid_unit_symbol(mol).
valid_unit_symbol(cd).
