% (c) 2009-2021 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(b_ast_cleanup, [clean_up/3, clean_up_pred/3, clean_up_pred_or_expr/3,
                          clean_up_l_wo_optimizations/4,
                          clean_up_l_with_optimizations/4,
                          check_used_ids_info/4, recompute_used_ids_info/2,
                          definitely_not_empty_and_finite/1, % TO DO: move to another module
                          get_unique_id/2,
                          predicate_level_optimizations/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module implements transformations/simplifications on the AST.').

:- use_module(tools, [safe_atom_chars/3,exact_member/2,foldl/4,filter/4]).
:- use_module(tools_lists, [length_less/2]).
:- use_module(error_manager).
:- use_module(debug).
:- use_module(self_check).
:- use_module(bsyntaxtree).
:- use_module(translate,[print_bexpr/1]).
:- use_module(btypechecker, [unify_types_strict/2]).
:- use_module(preferences,[get_preference/2]).
:- use_module(custom_explicit_sets,[convert_to_avl/2]).
:- use_module(prob_rewrite_rules(b_ast_cleanup_rewrite_rules),[rewrite_rule_with_rename/7]).
:- use_module(b_enumeration_order_analysis, [find_do_not_enumerate_variables/3]).
:- use_module(performance_messages,[perfmessage/2]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(system), [environ/2]).

% entry point for cleaning up predicates; ensures that global, predicate-level optimizations also applied
clean_up_pred(Expr,NonGroundExceptions,CleanedUpExpr) :-
   clean_up(Expr,NonGroundExceptions,CExpr),
   (get_texpr_type(CExpr,pred)
    -> predicate_level_optimizations(CExpr,CleanedUpExpr)
    ;  add_internal_error('Not predicate: ',clean_up_pred(Expr,NonGroundExceptions,CleanedUpExpr)),
       CleanedUpExpr = CExpr).

% Warning: arguments swapped with clean_up for maplist !
clean_up_pred_or_expr(NonGroundExceptions,Expr,CleanedUpExpr) :-
   clean_up_pred_or_expr_with_path(NonGroundExceptions,Expr,CleanedUpExpr,[]).
clean_up_pred_or_expr_with_path(NonGroundExceptions,Expr,CleanedUpExpr,Path) :-
   clean_up(Expr,NonGroundExceptions,CExpr,Path),
   (get_texpr_type(CExpr,pred)
    -> predicate_level_optimizations(CExpr,CleanedUpExpr,Path)
    ;  CleanedUpExpr = CExpr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% clean up some code afterwards
clean_up(Expr,NonGroundExceptions,CExpr) :-
    clean_up_init(Expr,Expr1),
    clean_up(Expr1,NonGroundExceptions,CExpr,[]).
clean_up(Expr1,NonGroundExceptions,CExpr,Path) :-
    get_texpr_type(Expr1,Type1),
    (NonGroundExceptions = do_not_ground_types -> true
      ; ground_type_to_any(Type1,NonGroundExceptions)), %
    (preferences:get_preference(normalize_ast,true)
     -> cleanups(normalize,Expr1,[],Expr2,Path)
      ; Expr2=Expr1),
    cleanups(pre,Expr2,[],TPExpr,Path),
    remove_bt(TPExpr,PExpr,LExpr,TLExpr),
    syntaxtransformation(PExpr,Subs,_,NewSubs,LExpr),
    functor(PExpr,F,N),
    % recursively clean up sub-expressions
    clean_up_l(Subs,NonGroundExceptions,NewSubs,F/N,1,Path),
    cleanups(post,TLExpr,[],CExpr,Path).
    %, tools_printing:print_term_summary(cleaned_up(CExpr)),nl.

% just run post-phase
%cleanups_post(Expr,CleanedupExpr) :- cleanups(post,Expr,[],CleanedupExpr,[]).

clean_up_init(Expr,Expr1) :-
    % ensure that WD info is available also for pre phase
    (transform_bexpr(b_ast_cleanup:compute_wd_info,Expr,Expr1) -> true
     ; add_internal_error('Call failed:',clean_up_init(Expr,Expr1)), Expr1=Expr).

:- use_module(bsyntaxtree,[transform_bexpr/3]).
compute_wd_info(b(E,Type,I),b(E,Type,NInfo)) :-
        (nonmember(contains_wd_condition,I),is_possibly_undefined(E)
          -> NInfo = [contains_wd_condition|I]
          ;  NInfo = I).

:- use_module(library(ordsets),[ord_nonmember/2, ord_add_element/3]).
% apply the clean-up rules to an expression until all
% applicable rules are processed
% cleanups(Phase,Expr,AppliedRules,Result,Path):
%   Phase: pre or post or normalize
%   Expr:  the expression to clean up
%   AppliedRules: a sorted list of clean up rules that have been already applied
%                 and must only be apply once ("single" mode)
%   Result: the cleaned-up expression
%   Path: list of outer functors leading to this expression; can be used to decide about applicability of rules
cleanups(Phase,Expr,AppliedRules,Result,Path) :-
    %% print(cleanups(Phase,Expr,AppliedRules,Result,Path)),nl,
    % set up co-routines that ensure that "Rule" is not applied if
    % is in the list AppliedRules
    start_profile_rule(RuleInfos),
    assure_single_rules(AppliedRules,Mode,Rule),
    ( cleanup_phase(Phase,Expr,NExpr,Mode/Rule,Path) ->    % try to apply a rule (matching the current phase)
        ( Mode==single -> ord_add_element(AppliedRules,Rule,AppRules) % if the rule is marked as "single", we add to the list of already applied rules
          ; Mode==multi  -> AppRules = AppliedRules  % if "multi", we do not add it to the list, the rule might be applied more than once
          ; otherwise    -> add_error_fail(b_ast_cleanup,'Unexpected rule mode ',Mode)
        ),
        stop_profile_rule(Rule,Mode,Phase,Expr,RuleInfos),
        cleanups(Phase,NExpr,AppRules,Result,Path)            % continue recursively with the new expression
    ; otherwise ->              % if no rule matches anymore,
        Result = Expr,          %   we leave the expression unmodified
        Mode=multi, Rule=none).           %   and unblock the co-routine (see assure_single_rules/3)


:- if(environ(prob_safe_mode,true)).
start_profile_rule([R1,W1]) :- statistics(runtime,[R1,_]),statistics(walltime,[W1,_]).
stop_profile_rule(Rule,Mode,Phase,Expr,[R1,W1]) :-
        statistics(runtime,[R2,_]),statistics(walltime,[W2,_]), DeltaW is W2-W1, DeltaR is R2-R1,
        (DeltaW < 20 -> true ; format('Firing AST cleanup rule ~w (mode:~w) in phase ~w took ~w ms (~w ms walltime)~n',[Rule,Mode,Phase,DeltaR,DeltaW]), translate:print_span(Expr),nl),
        runtime_profiler:register_profiler_runtime(Rule,unknown,DeltaR,DeltaW).
         %%% print(fired_rule(Rule,Mode,Phase)),nl, translate:print_bexpr_or_subst(Expr), print(' ===> '),nl, translate:print_bexpr_or_subst(NExpr),nl, print_ast(NExpr),nl, %% COMMENT IN TO SEE applied RULES <---------------
          %(map_over_typed_bexpr(b_ast_cleanup:check_valid_result,NExpr) -> true ; true), % comment in to check output after every firing of a rule
:- else.
start_profile_rule(_).
stop_profile_rule(_,_,_,_,_).
:- endif.

%check_valid_result(b(xexists([b(identifier(msgXX),integer,_)|_],_),pred,Infos)) :- nonmember(allow_to_lift_exists,Infos),print(missing_info),nl,trace,fail.

assure_single_rules([],_Mode,_Rule) :- !.
assure_single_rules(AppliedRules,Mode,Rule) :-
        assure_single_rules2(AppliedRules,Mode,Rule).
:- block assure_single_rules2(?,-,?),assure_single_rules2(?,?,-).
assure_single_rules2(_AppliedRules,_,none) :- !.
assure_single_rules2(_AppliedRules,multi,_) :- !.
assure_single_rules2(AppliedRules,_,Rule) :-
       % typically AppliedRules not very long; would also do: \+ member(Rule, AppliedRules).
        ord_nonmember(Rule,AppliedRules).

cleanup_phase(Phase,OTExpr,NTExpr,Mode/Rule,Path) :-
        create_texpr(OExpr,OType,OInfo,OTExpr),
        check_generated_info(OInfo,entry,Path),
        create_texpr(NExpr,NType,NInfo,NTExpr),
        cleanup_phase2(Phase,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,Path),
        check_generated_info(NInfo,Rule,Path).
cleanup_phase2(normalize,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,_Path) :-
        decompose_rule(Mode_Rule,Mode,Rule),
        cleanup_normalize(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode_Rule),
        print('Rewritten: '), print_bexpr(b(OExpr,OType,OInfo)),nl,
        print('     Into: '), print_bexpr(b(NExpr,NType,NInfo)),nl.
cleanup_phase2(pre,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,Path) :-
        decompose_rule(Mode_Rule,Mode,Rule),
        cleanup_pre_with_path(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode_Rule,Path).
cleanup_phase2(post,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,Path) :-
        decompose_rule(Mode_Rule,Mode,Rule),
        cleanup_post_with_path(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode_Rule,Path).
       %  check_ast(b(NExpr,NType,NInfo)),nl.

:- block decompose_rule(-,?,?).
decompose_rule(Mode_Rule,Mode,Rule) :-
        (functor(Mode_Rule,'/',2)
         -> Mode_Rule = Mode/Rule
         ; add_internal_error('Illegal cleanup rule, missing mode: ',Mode_Rule),fail).

:- if(environ(prob_safe_mode,true)).
check_generated_info(Info,Rule,Path) :- select(used_ids(_),Info,I1), member(used_ids(_),I1),!,
    format('Illegal used ids generated by ~w within ~w~n Infos=~w~n',[Rule,Path,Info]),
    add_internal_error('Illegal Info generated by rule: ',Rule),
    fail.
:- endif.
check_generated_info(_,_,_).

clean_up_l_wo_optimizations(Rest,NonGroundExceptions,CRest,SectionName) :-
   maplist(clean_up_init,Rest,Rest1),
   clean_up_l(Rest1,NonGroundExceptions,CRest,top_level(SectionName),1,[]).
clean_up_l([],_,[],_Functor,_Nr,_Path).
clean_up_l([Expr|Rest],NonGroundExceptions,[CExpr|CRest],Functor,ArgNr,Path) :-
    clean_up(Expr,NonGroundExceptions,CExpr,[arg(Functor,ArgNr)|Path]),
    A1 is ArgNr+1,
    clean_up_l(Rest,NonGroundExceptions,CRest,Functor,A1,Path).


% MAIN ENTRY POINT for b_machine_construction, bmachine_eventb, proz
% same as clean_up_l but also applies predicate_level_optimizations
% Context is just the name of the section/context in which the optimizations are run
clean_up_l_with_optimizations(Rest,NonGroundExceptions,CRest,Context) :-
   %clean_up_l(Rest,NonGroundExceptions,CRest,top_level,1,[]).
   (clean_up_l_with_opt(Rest,NonGroundExceptions,CleanedUpRest,top_level(Context),1,[]) -> CRest=CleanedUpRest
     ; add_internal_error('Call failed:',clean_up_l_with_optimizations(Rest,NonGroundExceptions,CRest,Context)),
       CRest=Rest
   ).
clean_up_l_with_opt([],_,[],_Functor,_Nr,_Path).
clean_up_l_with_opt([Expr|Rest],NonGroundExceptions,[CExpr|CRest],Functor,ArgNr,Path) :-
    clean_up_init(Expr,Expr1),
    clean_up_pred_or_expr_with_path(NonGroundExceptions,Expr1,CExpr,[arg(Functor,ArgNr)|Path]),
    A1 is ArgNr+1,
    clean_up_l_with_opt(Rest,NonGroundExceptions,CRest,Functor,A1,Path).

:- use_module(typing_tools,[is_infinite_type/1]).
:- use_module(specfile,[animation_mode/1, animation_minor_mode/1]).
% cleanup_pre(OldExpr,OldType,OldInfo,NewExpr,NewType,NewInfo,Mode/Rule)

% optional normalization rules
% These rules are now generated using the prob_rule_compiler
cleanup_normalize(Expr,Type,Info,NewExpr,NewType,NewInfo, multi/apply_normalization_rule(Rule)) :-
    b_ast_cleanup_rewrite_rules:normalization_rule_with_rename(Expr,Type,Info,NewExpr,NewType,NewInfo,Rule),
    (silent_mode(on) -> true
      ; format('Use rewrite_rule_normalize ~w~n',[Rule]),
        print_bexpr(b(NewExpr,NewType,NewInfo)),nl
    ),
    (ground(NewExpr) -> true ; print(not_ground_rewrite(NewExpr)),nl,fail).

never_transform_or_optimise(boolean_false).
never_transform_or_optimise(boolean_true).
%never_transform_or_optimise(bool_set).
never_transform_or_optimise(empty_set).
never_transform_or_optimise(empty_sequence).
never_transform_or_optimise(falsity).
%never_transform_or_optimise(max_int).
%never_transform_or_optimise(min_int).
never_transform_or_optimise(truth).
never_transform_or_optimise(identifier(_)).
never_transform_or_optimise(integer(_)).
never_transform_or_optimise(string(_)).
never_transform_or_optimise(value(_)).


% first check for a few expressions that never need to be optimised, rewritten:
cleanup_pre_with_path(E,_,_,_,_,_,_,_) :- never_transform_or_optimise(E),!,fail.
% TO DO: think about enabling the following clause
%cleanup_pre_with_path(exists(AllIds,P),pred,I,exists(AllIds,P),pred,NewI,single/annotate_toplevel_exists,Path) :-
%   Path = [H|_], % TODO: also deal with lambda and other quantifications
%   H=arg(comprehension_set/2,1),
%   % TO DO: add some conditions under which we allow to lift
%   add_info_if_new(I,allow_to_lift_exists,NewI). % mark existential quantifier as outermost: no need to delay it in b_interpreter
cleanup_pre_with_path(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode_Rule,_) :-
   cleanup_pre(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode_Rule).

% Now cleanup_pre rules that do not require path:

cleanup_pre(block(TS),subst,_,Subst,subst,Info,multi/remove_block) :-
    !,get_texpr_expr(TS,Subst),get_texpr_info(TS,Info).
% replace finite by truth
cleanup_pre(finite(S),pred,I,truth,pred,[was(finite(S))|I],multi/remove_finite) :-
    preferences:get_preference(disprover_mode,false), % keep finite in disprover goals
    get_texpr_type(S,Type),
    \+ is_infinite_type(Type),!.
    %!,(is_infinite_type(Type) -> print(keep_finite(Type)),nl,fail ; print(removed_finite(Type)),nl).
/*
 cleanup_pre(domain(SETC),Type,I,
             comprehension_set(DomainIds,NewPred),Type,I,multi/dom_let_pred) :-
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,[RangeId]),
    get_texpr_ids(CompIds,UnsortedIds), sort(UnsortedIds,Blacklist),
    conjunction_to_list(CompPred,Preds),
    select_equality(TId,Preds,Blacklist,Expr,Rest,_UsedIds,check_well_definedness),
    same_id(TId,RangeId,_),
    !,
    LetIds = [RangeId], Exprs = [Expr],
    conjunct_predicates(Rest,RestPred),
    NewPred = b(let_predicate(LetIds,Exprs,RestPred),pred,[generated(domain)]),
     print('Translated dom({...,x|x=E&...}) into: '),
    print_bexpr(b(comprehension_set(DomainIds,NewPred),Type,[])),nl.
*/
/* strangely enough: this does not seem to buy anything:
 cleanup_pre(domain(SETC),Type,I,
             let_expression(LetIds,Exprs,NewExpr),Type,[generated(domain)|I],multi/dom_let_expr) :-
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,[RangeId]),
    get_texpr_ids(CompIds,UnsortedIds), sort(UnsortedIds,Blacklist),
    conjunction_to_list(CompPred,Preds),
    select_equality(TId,Preds,Blacklist,Expr,Rest,_UsedIds,check_well_definedness),
    same_id(TId,RangeId,_),
    !,
    LetIds = [RangeId], Exprs = [Expr],
    conjunct_predicates(Rest,RestPred),
    get_texpr_info(SETC,CInfo),
    NewExpr = b(comprehension_set(DomainIds,RestPred),Type,CInfo),
     print('Translated dom({...,x|x=E&...}) into: '), print_bexpr(b(let_expression(LetIds,Exprs,NewExpr),Type,[])),nl.
*/
% exchange quantified union by generalized union
cleanup_pre(QUANT,Type,I,
            let_expression(LetIds,Exprs,NewExpr),Type,[generated(QuantOP)|I],multi/quant_union_inter_let) :-
    quantified_set_operator(QUANT,QuantOP,AllIds,Pred,Expr),
    conjunction_to_nontyping_list(Pred,Preds),
    % The ids are needed to build a "black list"
    get_sorted_ids(AllIds,Ids),
    find_one_point_rules(AllIds,Preds,Ids,LetIds,Exprs,RestIds,NewPreds),
    %print(find_one_point_rules(AllIds,Preds,Ids,LetIds,Exprs,RestIds,NewPreds)),nl,
    % only succeed if we found at least one id which can be rewritten as let
    LetIds = [_ID1|_],
    !,
    (RestIds=[],NewPreds=[] -> NewExpr=Expr
        % UNION(x).(x=E|Expr) --> LET x BE x=E in Expr END
      ; RestIds = [] -> conjunct_predicates_with_pos_info(NewPreds,NewPred),
        % UNION(x).(x=E & NewPred|Expr) --> LET x BE x=E in IF NewPRED THEN Expr ELSE {} END END
        NewExpr = b(if_then_else(NewPred,Expr,b(empty_set,Type,[])),Type,I)
      ; conjunct_predicates_with_pos_info(NewPreds,NewPred),
        quantified_set_operator(NewQUANT,QuantOP,RestIds,NewPred,Expr),
        NewExpr = b(NewQUANT,Type,I)),
     (debug_mode(off) -> true
       ; print('Translated UNION/INTER into: '), print_bexpr(b(let_expression(LetIds,Exprs,NewExpr),Type,[])),nl).

cleanup_pre(quantified_union(Ids,Pred,Expr),Type,I,Res,Type,[was(quantified_union)|I],multi/quant_union_symbolic) :-
    memberchk(prob_annotation('SYMBOLIC'),I),
    % UNION(Ids).(Pred|Expr) -->  {u| #Ids.(Pred & u:Expr)}
    get_texpr_set_type(Expr,IdType),
    !,
    get_unique_id_inside('__UNION__',Pred,Expr,FRESHID),
    NewTID = b(identifier(FRESHID),IdType,[]),
    safe_create_texpr(member(NewTID,Expr),pred,Member),
    create_exists_opt(Ids,[Pred,Member],Exists),
    Res = comprehension_set([NewTID],Exists),
    (debug_mode(off) -> true ; print('UNION @symbolic: '), print_bexpr(b(Res,Type,I)),nl).
% exchange quantified union by generalized union
cleanup_pre(quantified_union(Ids,Pred,Expr),Type,I,Res,Type,I,multi/quant_union) :-
    !,
    ((get_preference(convert_comprehension_sets_into_closures,false), % for test 1101: UNION is a form of let and forces computation at the moment; translating it into a construct without union means that it may be kept symbolic
      singleton_set_extension(Expr,One) %, debug:bisect(Expr,[1,1,1,1,1,1,1]))  % [1,1,1,1,1,1,1] slow; 1,1,0  1,1,1,0 fast
    )
    % UNION(Ids).(Pred|{One})  -->  ran( %(Ids).(Pred|One) )
     ->  quantified_set_op(Ids,Pred,One,quantified_intersection,I,TRes),
         get_texpr_expr(TRes,Res),
         (debug_mode(off) -> true ; print('UNION - SINGLETON: '), print_bexpr(TRes),nl)
     ;
    % UNION(Ids).(Pred|Expr)  -->  union(  ran( %(Ids).(Pred|Expr) ) )
      quantified_set_op(Ids,Pred,Expr,quantified_union,I,Set),
      Res = general_union(Set)
    ).
%alternate encoding:
%cleanup_pre(quantified_union(Ids,Pred,Expr),Type,I,Set,Type,I,multi/quant_union) :-
   %poses problem to test 614, 1101
%    !,quantified_union_op(Ids,Pred,Expr,Type,Set).
% exchange quantified intersection by generalized intersection
cleanup_pre(quantified_intersection(Ids,Pred,Expr),Type,I,general_intersection(Set),Type,I,multi/quant_inter) :-
    !,quantified_set_op(Ids,Pred,Expr,quantified_intersection,I,Set).
% In reply to PROB-240: Check if arguments of Prj1/2 are types only using is_just_type:
cleanup_pre(function(TProjection,Argument),Type,I,Result,Type,I,multi/projection_call) :-
    get_texpr_expr(TProjection,Projection),
    cleanup_function_projection(Projection,Argument,I,Result),
    !.
cleanup_pre(function(Lambda,Argument),Type,I,assertion_expression(Cond,ErrMsg,Expr),Type,I,multi/lambda_guard1) :-
    get_preference(optimize_ast,true),
    get_texpr_expr(Lambda,LambdaExpr),
    is_lambda(LambdaExpr,TId,TPre,TVal),
    get_texpr_id(TId,Id),
    \+ occurs_in_expr(Id,Argument), % otherwise name clash
    replace_id_by_expr_with_count(TPre,Id,Argument,Cond,NrReplace1), % Warning The Argument could get replicated !!
    replace_id_by_expr_with_count(TVal,Id,Argument,Expr,NrReplace2),
    NrReplacements is NrReplace1+NrReplace2,
    (is_replace_id_by_expr_ok(Argument,Id,NrReplacements,lambda_guard1) -> true
      ; (debug_mode(off) -> true
         ; format('~n*** Lambda argument duplication (~w times) for ~w: ',[NrReplacements,Id]),
           %translate:print_bexpr(Argument),nl, translate:print_bexpr(Lambda),nl,
           translate:print_span(I),nl
        ),
        true % fail leads to test 1284 taking very long, 191 seconds instead of 0.3 for 192 states
     ), % we could introduce let if necessary
    !, % TO DO: support for functions of multiple arguments; maybe allow function(id,id)
    ErrMsg = 'lambda function called outside of domain, condition false: '.
/*
cleanup_post(function(Lambda,Argument),Type,I,assertion_expression(Cond,ErrMsg,Expr),Type,I,multi/lambda_guard2) :-
    get_texpr_info(Lambda,IL), member(was(lambda),IL),
    print(function(Lambda,Argument)),nl,
    get_texpr_expr(Lambda,comprehension_set([TId1,TId2,LAMBDARES],TPre,TVal)),
    Argument=b(couple(Argument1,Argument2),_,_),
    !, % TO DO: support for functions of multiple arguments
    get_texpr_id(TId1,Id1),
    get_texpr_id(TId2,Id2),
    ErrMsg = 'function called outside of domain: ',
    replace_id_by_expr(TPre,Id1,Argument1,Cond1),
    replace_id_by_expr(TVal,Id1,Argument1,Expr1),
    replace_id_by_expr(Cond1,Id2,Argument2,Cond),
    replace_id_by_expr(Expr1,Id2,Argument2,Expr).  */
cleanup_pre(first_projection(A,B),Type,I,Set,Type,[was(prj1),prob_annotation('SYMBOLIC')|I],multi/first_projection) :-
    !,create_projection_set(A,B,first,Set).
cleanup_pre(second_projection(A,B),Type,I,Set,Type,[was(prj2),prob_annotation('SYMBOLIC')|I],multi/second_projection) :-
    !,create_projection_set(A,B,second,Set).
cleanup_pre(event_b_first_projection(Rel),Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev_first_projection) :-
    !,create_event_b_projection_set(Rel,first,Set).
cleanup_pre(event_b_second_projection(Rel),Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev_second_projection) :-
    !,create_event_b_projection_set(Rel,second,Set).
cleanup_pre(event_b_first_projection_v2,Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev2_first_projection) :-
    !,create_event_b_projection_set_v2(Type,first,Set). % TO DO: Daniel, does this work ???
cleanup_pre(event_b_second_projection_v2,Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev2_second_projection) :-
    !,create_event_b_projection_set_v2(Type,second,Set). % TO DO: Daniel, does this work ???
cleanup_pre(function(Fun,Arg),integer,I,ArithOp,integer,I,multi/succ_pred_optimisation) :-
    get_texpr_expr(Fun,PS),
    ( PS=predecessor -> Op=minus
    ; PS=successor   -> Op=add),
    ArithOp =.. [Op,Arg,Integer],
    create_texpr(integer(1),integer,[],Integer).
/* cleanup_pre(event_b_comprehension_set([ID],ID,Pred),T,I,comprehension_set([Result],NewPred),T,
        [was(event_b_comprehension_set)|I],multi/ev_compset_single_id) :-
         % Event_B_Comprehension with a single ID which is also the expression
         % TO DO: expand for multiple IDs
         !,
         Result = ID, NewPred=Pred. */
% Detect if_then_else; also done in cleanup_post (in pre we may be able to detect IF-THEN-ELSE before CSE has inserted lazy_lets
cleanup_pre(function(IFT,DUMMYARG),Type,Info,if_then_else(IFPRED,THEN,ELSE),Type,Info,multi/function_if_then_else) :-
   is_if_then_else(IFT,pre,DUMMYARG,IFPRED,THEN,ELSE),
   (debug_mode(off) -> true
    ; print('% Recognised if-then-else expression (pre): IF '), print_bexpr(IFPRED),
      print('  THEN '),print_bexpr(THEN), print(' ELSE '), print_bexpr(ELSE),nl
   ).
cleanup_pre(union(Lambda1,Lambda2),T,I,lambda(TIDs1,TPred,TVal),T,I,multi/combine_lambdas_if_then_else) :-
   Lambda1 = b(lambda(TIDs1,TPred1,TValue1),_,_), get_texpr_ids(TIDs1,Ids),
   Lambda2 = b(lambda(TIDs2,TPred2,TValue2),_,_), get_texpr_ids(TIDs2,Ids),
   %  %x.(P & Cond1 | Val1) \/ %x.(P & not(Cond1) | Val2)  ==>  %x.(P| IF Cond1 THEN Val1 ELSE Val2 END)
   conjunction_to_list(TPred1,L1),
   conjunction_to_list(TPred2,L2),
   append(Front1,[Last1],L1),
   append(Front2,[Last2],L2),
   maplist(same_texpr,Front1,Front2),
   is_negation_of(Last1,Last2),
   conjunct_predicates(Front1,TPred),
   get_texpr_type(TValue1,TV),
   safe_create_texpr(if_then_else(Last1,TValue1,TValue2),TV,[],TVal),
   (debug_mode(off) -> true
    ; format('Union of lambdas converted to if-then-else: ',[]),
      translate:print_bexpr(b(lambda(TIDs1,TPred,TVal),T,I)),nl
    ).
cleanup_pre(event_b_comprehension_set(Ids,Expr,Pred),T,I,NewExpression,T,
        [was(event_b_comprehension_set)|I],multi/ev_compset) :-
    rewrite_event_b_comprehension_set(Ids,Expr,Pred, T, NewExpression).
cleanup_pre(domain_restriction(A,B),T,I,identity(A),T,I,multi/event_b_to_normal_identity1) :-
  /* translate S <| id into id(S) */
   is_event_b_identity(B).
cleanup_pre(range_restriction(B,A),T,I,identity(A),T,I,multi/event_b_to_normal_identity2) :-
  /* translate id |> S into id(S) */
   is_event_b_identity(B).
% what about translating id(TOTAL TYPE) into event_b_identity ??
cleanup_pre(ring(A,B),T,I,composition(B,A),T,I,multi/ring_composition).
cleanup_pre(less_equal(A,B),pred,I,less_equal_real(A,B),pred,I,multi/remove_ambiguous_leq_real) :-
       get_texpr_type(A,real).
cleanup_pre(greater_equal(A,B),pred,I,less_equal_real(B,A),pred,I,multi/remove_ambiguous_geq_real) :-
       get_texpr_type(A,real).
cleanup_pre(less(A,B),pred,I,less_real(A,B),pred,I,multi/remove_ambiguous_lt_real) :-
       get_texpr_type(A,real).
cleanup_pre(greater(A,B),pred,I,less_real(B,A),pred,I,multi/remove_ambiguous_gt_real) :-
       get_texpr_type(A,real).
cleanup_pre(unary_minus(A),real,I,unary_minus_real(A),real,I,multi/remove_ambiguous_unary_minus_real).
cleanup_pre(power_of(A,B),real,I,power_of_real(A,TB),real,I,multi/remove_ambiguous_unary_minus_real) :-
       get_texpr_type(A,real),
       safe_create_texpr(convert_real(B),real,[],TB). % TO DO: add pos
cleanup_pre(add(A,B),real,I,add_real(A,B),real,I,multi/remove_ambiguous_add_real).
cleanup_pre(div(A,B),real,I,div_real(A,B),real,I,multi/remove_ambiguous_div_real).
cleanup_pre(minus_or_set_subtract(A,B),integer,I,minus(A,B),integer,I,multi/remove_ambiguous_minus_int).
cleanup_pre(minus_or_set_subtract(A,B),real,I,minus_real(A,B),real,I,multi/remove_ambiguous_minus_real).
cleanup_pre(minus_or_set_subtract(A,B),Type,I,set_subtraction(A,B),Type,I,multi/remove_abiguous_minus_set) :-
      is_set_type(Type,_).
cleanup_pre(mult_or_cart(A,B),integer,I,multiplication(A,B),integer,I,multi/remove_ambiguous_times_int).
cleanup_pre(mult_or_cart(A,B),real,I,multiplication_real(A,B),real,I,multi/remove_ambiguous_times_real).
cleanup_pre(mult_or_cart(A,B),Type,I,cartesian_product(A,B),Type,I,multi/remove_ambiguous_times_set) :-
      is_set_type(Type,_).
cleanup_pre(E,T,Iin,E,T,Iout,multi/remove_rodinpos) :- % we use multi but can per construction only be applied once
    selectchk(nodeid(rodinpos(_,[],_)),Iin,Iout). % reomve rodinpos information with Name=[]
cleanup_pre(partition(X,[Set]),pred,I,equal(X,Set),pred,I,multi/remove_partition_one_element) :-
      !,
      %  partition(X,Set) <=> X=Set
     (debug_mode(on) -> true ;
       print('Introducing equality for partition: '),
       print_bexpr(X), print(' = '), print_bexpr(Set),nl).
cleanup_pre(case(Expression,CASES,Else),subst,I,NewSubst,subst,I,single/rewrite_case_to_if_then_else) :-
    % translate CASE E OF EITHER e1 THEN ... ---> LET case_expr BE case_expr=E IN IF case_expr=e1 THEN ...
    get_texpr_type(Expression,EType), get_texpr_info(Expression,EInfo),
    ExprID = b(identifier(EID),EType,EInfo),
    (get_texpr_id(Expression,EID)
      -> NewSubst = if(IFLISTE) % no LET necessary
      ;  NewSubst = let([ExprID],Equal,b(if(IFLISTE),subst,I)),
         get_unique_id_inside('case_expr',b(case(Expression,CASES,Else),subst,I),EID),
         safe_create_texpr(equal(ExprID,Expression),pred,Equal)
    ),
    (maplist(gen_if_elsif(ExprID),CASES,IFLIST)
      -> (get_texpr_expr(Else,skip) -> IFLISTE = IFLIST
           ; TRUTH = b(truth,pred,[]), get_texpr_info(Else,EI),
              append(IFLIST,[b(if_elsif(TRUTH,Else),subst,EI)],IFLISTE)
          ),
         (debug_mode(off) -> true
           ; print('Translating CASE to IF-THEN-ELSE: '), print_bexpr(Expression),nl
             %,translate:print_subst(b(NewSubst,subst,[])),nl
         )
      ;  add_internal_error('Translation of CASE to IF-THEN-ELSE failed: ',CASES),fail
    ).
cleanup_pre(exists(AllIds,Body),pred,I0,NewP1,pred,NewI,single/components_partition_exists) :-
     get_preference(optimize_ast,true),
     nonmember(partitioned_exists,I0), % avoid re-computing components on something that is already partitioned
     Simplify=no_cleanup_and_simplify, % avoid loops
     b_interpreter_components:construct_optimized_exists(AllIds,Body,NewPred,Simplify,NrC),
     (NrC = 1 % just one component; perform no change to avoid re-ordering ids, ... ; see test 510
       -> NewP1=exists(AllIds,Body), NewI=I0
       ; NewPred=b(NewP1,pred,I1),
         (debug_mode(on), NewP1 \= exists(_,_) -> format('PARTITIONED EXISTS:~n ',[]), translate:nested_print_bexpr(NewPred),nl ; true),
          add_important_info_from_super_expression(I0,I1,NewI)
      ).
cleanup_pre(exists(AllIds,P),pred,I,NewP1,pred,I,single/factor_out) :- % REDUNDANT with rule above ??
    conjunction_to_nontyping_list(P,Preds),
     % move things which do not depend on AllIds outside
     % transform, e.g., #(x).(y>2 & x=y) -->  y>2 & #(x).(x=y)
    get_preference(optimize_ast,true),
    create_exists_opt(AllIds,Preds,b(NewP1,pred,_I),Modified),
    %(Modified = true -> print(exists(AllIds)),nl,print_bexpr(b(NewP1,pred,_I)),nl),
     % the rule will fire again on the newly generated sub predicate ! -> fix ?
     Modified=true.% check if anything modified; otherwise don't fire rule

cleanup_pre(exists(AllIds,P),pred,Info0,NewP,pred,INew,multi/remove_single_use_equality) :-
   % remove existentially quantified variables which are defined by an equation and are used only once
   % e.g., Z1...Z4 in not(#(X,Y,Z,Z1,Z2,Z3,Z4).(X:INTEGER & X*Y=Z1 & Z1*Z = Z2 & Z*X = Z3 & Z3*Y = Z4 & Z2 /= Z4))
   get_preference(optimize_ast,true),
   (length_less(AllIds,100) -> true % otherwise the code becomes quite inefficient at the moment
     ; perfmessage('Large existential quantifier, performing limited optimizations',Info0),
       fail),
   conjunction_to_list(P,Preds),
   CheckWellDef=no_check,
   select_equality(TId,Preds,[],_,IDEXPR,RestPreds,_,CheckWellDef),
   get_texpr_id(TId,ID),
   select(TIdE,AllIds,RestIds), get_texpr_id(TIdE,ID),
   can_be_optimized_away(TIdE),
   \+ occurs_in_expr(ID,IDEXPR), % we cannot inline #x.(x=y+x & ...)
   always_defined_full_check_or_disprover_mode(IDEXPR), % otherwise we may remove WD issue by removing ID if Count=0 or move earlier/later if count=1
   single_usage_identifier(ID,RestPreds,Count), % we could also remove if Expr is simple
   (Count=0 -> debug_println(19,unused_equality_id(ID)),
               PL=RestPreds
    ;  % Count should be 1
       conjunct_predicates_with_pos_info(RestPreds,RestPred),
       replace_id_by_expr(RestPred,ID,IDEXPR,E2),
       conjunction_to_list(E2,PL)
    ),
   create_exists_opt(RestIds,PL,TNewP), % no need to computed used ids yet; we could do this:
   %conjunct_predicates_with_pos_info(PL,PP), safe_create_texpr(exists(RestIds,PP),pred,TNewP),
   (debug_mode(off) -> true
    ; format('Remove existentially quantified identifier with single usage: ~w (count: ~w)~n',[ID,Count]), print_bexpr(IDEXPR),nl),
   TNewP = b(NewP,pred,INew),!.
cleanup_pre(exists(AllIds,P),pred,I,let_predicate(LetIds,Exprs,NewP),pred,INew,multi/exists_to_let) :-
    % rewrite predicates of the form #x.(x=E & P(x)) into (LET x==E IN P(x))
    % side condition for #(ids).(id=E & P(ids)): no identifiers of ids occur in E
    get_preference(optimize_ast,true),
    conjunction_to_nontyping_list(P,Preds), % TO DO: avoid recomputing again (see line in clause above)
    % The ids are needed to build a "black list"
    get_sorted_ids(AllIds,Ids),
    find_one_point_rules(AllIds,Preds,Ids,LetIds,Exprs,RestIds,NewPreds),
    % no_check is not ok in the context of existential quantification and reification:
    %  #x.(1:dom(f) & x=f(1) & P) -->  LET x=f(1) IN 1:dom(f) & P END
    % it is not ok if the whole predicate gets reified in b_intepreter_check !! Hence we use always_defined_full_check_or_disprover_mode; see Well_def_1.9.0_b5 in private_examples
    % only succeed if we found at least one id which can be rewritten as let
    LetIds = [ID1|_],
    !,
    (atomic(ID1) -> add_internal_error(cleanup_pre,unwrapped_let_identifier(ID1)), INew=I
     ; get_texpr_ids(LetIds,AtomicIDs),
       remove_used_ids_from_info(AtomicIDs,I,INew)
    ), % probably not necessary ?!
    (member(allow_to_lift_exists,I) -> AddInfos=[allow_to_lift_exists] ; AddInfos=[]),
    create_exists_opt(RestIds,NewPreds,AddInfos,NewP,_Modified),
    debug_println(19,generated_let(AtomicIDs)).
% now the same LET extraction but for universal quantification:
cleanup_pre(forall(AllIds,P,Rhs),pred,I,let_predicate(LetIds,Exprs,NewP),pred,I,multi/forall_to_let) :-
    get_preference(optimize_ast,true),
    conjunction_to_nontyping_list(P,Preds),
    % The ids are needed to build a "black list"
    get_sorted_ids(AllIds,Ids),
    check_forall_lhs(P,I,Ids),
    find_one_point_rules(AllIds,Preds,Ids,LetIds,Exprs,RestIds,NewPreds),
    % only succeed if we found at least one id which can be rewritten as let
    LetIds = [ID1|_],!,
    (atomic(ID1) -> add_internal_error(cleanup_pre,unwrapped_let_identifier(ID1)) ; true),
    conjunct_predicates_with_pos_info(NewPreds,NewLhs),
    create_implication(NewLhs,Rhs,NewForallBody),
    create_forall(RestIds,NewForallBody,NewP),
    (debug_mode(on) ->  print('Introduced let in forall: '), print(LetIds),nl ; true).
    % warning: used_identifier information not yet computed; translate may generate warnings
cleanup_pre(exists(AllIds,P),pred,I0,NewPE,pred,NewI,multi/exists_remove_typing) :-
       (is_a_conjunct(P,Typing,Q) ; is_an_implication(P,Typing,Q)),
       % TRUE & Q  ==  TRUE => Q  ==  Q
       is_typing_predicate(Typing),
       % remove typing so that other exists rules can fire
       % we run as cleanup_pre: the other simplifications which remove typing have not run yet
       % such typing conjuncts typicially come from Rodin translations
       create_exists_opt(AllIds,[Q],b(NewPE,_,I1)),
       add_important_info_from_super_expression(I0,I1,I2), % we could also copy node_id(_) from I0 ?
       add_removed_typing_info(I2,NewI).
cleanup_pre(exists(AllIds,P),pred,I,disjunct(NewP1,NewP2),pred,I,single/partition_exists_implication) :-
    is_a_disjunct_or_implication(P,_Type,Q,R),
    /* note that even if R is only well-defined in case Q is false; it is ok to seperate this out
       into two existential quantifiers:  #x.(x=0 or 1/x=10) is ok to transform into #x.(x=0) or #x.(1/x=10) */
    % this slows down test 1452, Cylinders, 'inv3/WD'; TO DO:investigate
    create_exists_opt(AllIds,[Q],NewP1),  %  print('Q: '),print_bexpr(NewP1),nl,
    create_exists_opt(AllIds,[R],NewP2). %,  print('R: '),print_bexpr(NewP2),nl.
cleanup_pre(forall(AllIds,P,Rhs),pred,I,NewPred,pred,I,multi/forall_to_post_let) :-
% translate something like !(x,y).(y:1..100 & x=y*y => x<=y) into !(y).(y : 1 .. 100 => (#(x).( (x)=(y * y) & x <= y)))
    post_let_forall(AllIds,P,Rhs,NewPred,modification),
    !,
    (debug_mode(on) -> print('POST LET INTRODUCTION: '), print_bexpr(b(NewPred,pred,[])),nl ; true).
cleanup_pre(set_extension(List),Type,I, Res,Type,I, single/eval_set_extension) :-
   extension_should_be_evaluated(List),
   !,
   (evaluate_set_extension(List,EvaluatedList),
    convert_to_avl(EvaluatedList,AVL)
    % evaluate simple explicit set extensions: avoid storing & traversing position info & AST
    -> (debug_mode(on) -> print('EVAL SET EXTENSION: '), translate:print_bvalue(AVL),nl ; true),
       Res=value(AVL)
    ;  Res=set_extension(List)). % just add eval_set_extension as fired
cleanup_pre(sequence_extension(List),Type,I, Res,Type,I, single/eval_seq_extension) :-
   extension_should_be_evaluated(List),
   !,
   ( evaluate_seq_extension_to_avl(List,AVL)
     % evaluate simple explicit set extensions: avoid storing & traversing position info & AST
     -> (debug_mode(on) -> print('EVAL SEQUENCE EXTENSION: '), translate:print_bvalue(AVL),nl ; true),
        Res=value(AVL)
     ;  Res=sequence_extension(List)). % just add eval_seq_extension as fired
cleanup_pre(set_extension(List),Type,I, set_extension(NList),Type,I, single/remove_pos) :-
   remove_position_info_from_list(List,I,NList),!.
cleanup_pre(sequence_extension(List),Type,I, sequence_extension(NList),Type,I, single/remove_pos) :-
    remove_position_info_from_list(List,I,NList),!.
%cleanup_pre(concat(A,B),string,I,Res,string,I,multi/concat_assoc_reorder) :-
%    A = b(concat(A1,A2),string,I1),
%    !, % reorder STRING concats for better efficiency, can only occur when allow_sequence_operators_on_strings is true
%    % TO DO: extract information I2B from A2 and B
%    Res = concat(A1,b(concat(A2,B),string,I2B)).
cleanup_pre(typeset,SType,I,Expr,SType,I,multi/remove_typeset) :- !,
    % used, e.g., in test 1205 for recursive Event-B operator definition
    ( ground(SType) ->
        (is_set_type(SType,Type),create_maximal_type_set(Type,b(Expr,_,_)) -> true
        ; otherwise ->
            add_error_and_fail(b_ast_cleanup,'Creating type expression for type set failed: ',SType))
    ; otherwise    -> add_error_and_fail(b_ast_cleanup,'Non-ground type for typeset expression: ',SType)).
cleanup_pre(integer_set(S),Type,I,Expr,Type,[was(integer_set(S))|I],multi/remove_integer_set) :- !,
    translate_integer_set(S,I,Expr),
    (debug_mode(off) -> true
      ; format('Rewrite ~w to: ',[S]),
        print_bexpr(b(Expr,integer,I)),nl).
% should we move the rewrite_rules to normalize ??
cleanup_pre(Expr,Type,Info,NewExpr,NewType,NewInfo, multi/apply_rewrite_rule(Rule)) :-
    rewrite_rule_with_rename(Expr,Type,Info,NewExpr,NewType,NewInfo,Rule), % from b_ast_cleanup_rewrite_rules
    (debug_mode(off) -> true
      ; format('Use rewrite_rule ~w~n',[Rule]),
        print_bexpr(b(NewExpr,NewType,NewInfo)),nl),
    (ground(NewExpr) -> true ; print(not_ground_rewrite(NewExpr)),nl,fail).
% 'x > y' to 'y < x'
cleanup_pre(greater(Lhs,Rhs), pred, I, less(Rhs,Lhs), pred, [was(greater(Lhs,Rhs))|I], single/normalize_greater) :-
    preferences:get_preference(normalize_ast, true).
% 'x >= y' to 'y <= x'
cleanup_pre(greater_equal(Lhs,Rhs), pred, I, less_equal(Rhs,Lhs), pred, [was(greater_equal(Lhs,Rhs))|I], single/normalize_greater_equal) :-
    preferences:get_preference(normalize_ast, true).
% 'x - y' to 'x + -y'
cleanup_pre(minus(Lhs,Rhs), integer, I, add(Lhs,b(unary_minus(Rhs),integer,[])), integer, [was(minus(Lhs,Rhs))|I], single/normalize_minus) :-
    preferences:get_preference(normalize_ast, true).

translate_integer_set('NAT',I,interval(b(integer(0),integer,I),b(max_int,integer,I))).
translate_integer_set('NAT1',I,interval(b(integer(1),integer,I),b(max_int,integer,I))).
translate_integer_set('INT',I,interval(b(min_int,integer,I),b(max_int,integer,I))).
%translate_integer_set('INTEGER',I,comprehension_set([b(identifier('_zzzz_unary'),integer,I)],
%                                     b(truth,pred,[prob_annotation('SYMBOLIC')|I]))).
%translate_integer_set('NATURAL',I,comprehension_set([b(identifier('_zzzz_unary'),integer,I)],
%    b(greater_equal(b(identifier('_zzzz_unary'),integer,I),
%                    b(integer(0),integer,I)),pred,[prob_annotation('SYMBOLIC')|I]))).
%translate_integer_set('NATURAL1',I,comprehension_set([b(identifier('_zzzz_unary'),integer,I)],
%    b(greater_equal(b(identifier('_zzzz_unary'),integer,I),
%                    b(integer(1),integer,I)),pred,[prob_annotation('SYMBOLIC')|I]))).

% detect if an expression is equivalent to an integer set, does not check for interval yet
is_integer_set(integer_set(S),S).
is_integer_set(comprehension_set([b(identifier(ID),integer,_)],b(B,_,_)),S) :-
   is_integer_set_constraint_pred(B,ID,S).
is_integer_set_constraint_pred(truth,_,'INTEGER').
is_integer_set_constraint_pred(Expr,ID,Set) :-
       is_greater_equal(Expr,b(identifier(ID),integer,_),TNr),
       get_integer(TNr,Nr),
       (Nr=0 -> Set='NATURAL' ; Nr=1 -> Set='NATURAL1').
is_integer_set_constraint_pred(Expr,ID,Set) :-
       is_greater(Expr,b(identifier(ID),integer,_),TNr),
       get_integer(TNr,Nr),
       (Nr = -1 -> Set='NATURAL' ; Nr=0 -> Set='NATURAL1').

is_greater_equal(greater_equal(A,B),A,B).
is_greater_equal(less_equal(B,A),A,B).
is_greater(greater(A,B),A,B).
is_greater(less(B,A),A,B).

is_inf_integer_set_with_lower_bound(b(X,_,_),Bound) :- is_integer_set(X,N),
   (N='NATURAL' -> Bound=0 ; N='NATURAL1' -> Bound=1).



is_subset(subset(A,B),A,B).
is_subset(member(A,b(pow_subset(B),_,_)),A,B). % x : POW(T)  <=>  x <: T

% tool to translate CASE values to Test predicates for IF-THEN-ELSE
gen_if_elsif(CaseID,b(case_or(ListOfValues, Body),_,I),
             b(if_elsif(Test,Body),subst,I)) :-
    get_texpr_type(CaseID,T),
    SEXT = b(set_extension(ListOfValues),set(T),I),
    Test = b(member(CaseID,SEXT),pred,I).

% the case below happens frequently in data validation:
remove_position_info_from_list(List,I,NList) :-
   member(nodeid(pos(C,FilePos,Line,From,Line,To)),I),
   To-From > 1000, % the entire set/sequence extension is on one large line
   length(List,Len), Len>100, % it has many elements
   % we replace all position infos by the same top-level position info (enabling sharing)
   maplist(remove_position_info(nodeid(pos(C,FilePos,Line,From,Line,To))),List,NList),
   (debug_mode(off) -> true
     ; format('SIMPLIFY POSITION INFO IN SET/SEQUENCE EXTENSION: line # ~w, length ~w~n',[Line,Len])).
remove_position_info(NI,b(Expr,Type,Infos),b(NExpr,Type,NewInfos)) :-
    syntaxtransformation(Expr,Subs,_Names,NSubs,NExpr),
    (select(nodeid(pos(_,_FilePos,_,_,_,_)),Infos,NT) -> NewInfos=[NI|NT] ; NewInfos=Infos),
    maplist(remove_position_info(NI),Subs,NSubs).

% rules for function application of various projection functions
cleanup_function_projection(first_projection(A,B),Argument,I,Result) :-
    gen_assertion_expression(A,B,Argument,first_of_pair(Argument),first,I,Result).
cleanup_function_projection(second_projection(A,B),Argument,I,Result) :-
    gen_assertion_expression(A,B,Argument,second_of_pair(Argument),second,I,Result).
cleanup_function_projection(event_b_second_projection(A),Argument,_I,Result) :- % old style Rodin projection
    check_is_just_type(A),Result = first_of_pair(Argument).
cleanup_function_projection(event_b_second_projection(A),Argument,_I,Result) :- % old style Rodin projection
    check_is_just_type(A),Result = second_of_pair(Argument).
cleanup_function_projection(event_b_first_projection_v2,Argument,_I,Result) :- Result = first_of_pair(Argument).
cleanup_function_projection(event_b_second_projection_v2,Argument,_I,Result) :- Result = second_of_pair(Argument).

check_is_just_type(_A) :- preferences:get_preference(ignore_prj_types,true),!.
check_is_just_type(A) :- (is_just_type(A) -> true ; debug_println(9,not_type_for_prj(A)),fail).

:- use_module(bsyntaxtree,[get_texpr_set_type/2]).
gen_assertion_expression(A,B,_Argument,ProjExpr,_ProjType,_I,Result) :-
    check_is_just_type(A),check_is_just_type(B),
    !,
    Result = ProjExpr.
% TO DO: add simplification rule for couple(x,y) : A*B  with A or B being just types
gen_assertion_expression(A,B,Argument,ProjExpr,ProjType,Info,Result) :-
    create_cartesian_product(A,B,CartAB),
    safe_create_texpr(member(Argument,CartAB),pred,MemCheck),
    ErrMsg = 'projection function called outside of domain: ', % TO DO: provide better user message with Argument result
    (ProjType == first -> get_texpr_set_type(A,TT) ; get_texpr_set_type(B,TT)), %%
    %get_texpr_pos_infos(Argument,Info), % add position infos
    extract_pos_infos(Info,PosInfo),
    safe_create_texpr(ProjExpr,TT,PosInfo,TProjExpr),
    Result = assertion_expression(MemCheck,ErrMsg,TProjExpr).
:- use_module(bsyntaxtree,[is_set_type/2]).
create_cartesian_product(A,B,CartAB) :-
    get_texpr_types([A,B],[STA,STB]),
    is_set_type(STA,TypeA), is_set_type(STB,TypeB),
    safe_create_texpr(cartesian_product(A,B),set(couple(TypeA,TypeB)),CartAB).

% TO DO: support for functions of multiple arguments
is_lambda(lambda([TId],TPred,TValue), TId, TPred,TValue).
is_lambda(event_b_comprehension_set([TId],Expr,TPred), TId, TPred, TValue) :-
% rewrite_event_b_comprehension_set does not seem to get called before the function/lambda rule is applied
     % {ID.ID|->Val | PRed}
     Expr = b(couple(LHS,RHS),_,_),
     same_texpr(LHS,TId),
     TValue=RHS.

% rewriting Event-B comprehension sets into classical B style ones
rewrite_event_b_comprehension_set(IDs,CoupleExpr,Pred, _T, NewExpression) :-
        % detect lambda expressions in classical B style
        nested_couple_to_list(CoupleExpr,List),
        check_ids(IDs,List,Expr),!,
        NewExpression = lambda(IDs,Pred,Expr).
rewrite_event_b_comprehension_set(IDList,CoupleExpr,Pred, _T, NewExpression) :-
         % Event_B_Comprehension with a several IDs which are also used as the couple expression
         nested_couple_to_list(CoupleExpr,List),
         List = IDList,
         !,
         NewExpression = comprehension_set(IDList,Pred).
rewrite_event_b_comprehension_set(Ids,Expr,Pred, T, NewExpression) :-
    NewExpression = comprehension_set([Result],NewPred),
    unify_types_strict(T,set(Type)),
    % print(event_b_comprehension_set(Ids,Expr,Pred)),nl,
    (select(Expr,Ids,RestIds)
      -> % the Expr is an identifier which is part of Ids: we can avoid complicated translation below
       % example {f,n•n:INT & f:1..n-->Digit|f} --translated-> {f|#(n).(n:INT & f:1..n-->Digit)}
      % print(remove(Expr,RestIds)),nl,
         ExPred=Pred, Result=Expr
      ; get_unique_id_inside('__comp_result__',Pred,Expr,ResultId),
        create_texpr(identifier(ResultId),Type,[],Result),
        safe_create_texpr(equal(Result,Expr),pred,Equal),
        conjunct_predicates_with_pos_info(Pred,Equal,ExPred), % put Equal after Pred for WD; used to be other order!
        RestIds=Ids
    ),
    create_exists(RestIds,ExPred,NewPred). %, print(done_rewrite_event_b_comprehension_set),nl, print_bexpr(NewPred),nl.

check_ids([],[CoupleExpr],CoupleExpr). % we terminate with a single expression
check_ids([ID|T],[CoupleExprID|CT],Rest) :-
  same_id(ID,CoupleExprID,_),
  check_ids(T,CT,Rest).


evaluate_seq_extension_to_avl(List,AVL) :-
   evaluate_set_extension(List,EvaluatedList),
   convert_set_to_seq(EvaluatedList,1,ESeq),
   convert_to_avl(ESeq,AVL).

evaluate_set_extension([],[]).
evaluate_set_extension([H|List],[EH|EvaluatedList]) :-
   eval_set_extension_element(H,EH),
   evaluate_set_extension(List,EvaluatedList).

extension_should_be_evaluated(List) :-
   preferences:get_preference(try_kodkod_on_load,false),
   List \= [],
   List = [_|ListT], ListT \= []. % do not do this for singleton sets so as not to prevent triggering of other rules

:- use_module(kernel_reals,[construct_real/2]).
% construct a value term for a simple AST element:
eval_set_extension_element(b(E,_,_I),EE) :-
  (eval_set_extension_aux(E,EE) -> true
    ; %print('eval_set_extension failed: '),tools_printing:print_term_summary(E),nl,translate:print_span(_I),nl,
      fail).
eval_set_extension_aux(boolean_false,pred_false).
eval_set_extension_aux(boolean_true,pred_true).
eval_set_extension_aux(couple(A,B),(EA,EB)) :- eval_set_extension_element(A,EA), eval_set_extension_element(B,EB).
eval_set_extension_aux(integer(I),int(I)).
eval_set_extension_aux(real(Atom),Real) :- construct_real(Atom,Real).
eval_set_extension_aux(string(S),string(S)).
eval_set_extension_aux(unary_minus(b(integer(I),_,_)),int(MI)) :- MI is -I.
eval_set_extension_aux(empty_set,[]).
eval_set_extension_aux(value(V),V).
eval_set_extension_aux(rec(Fields),rec(EF)) :- eval_set_extension_fields(Fields,EF).
%eval_set_extension_aux(identifier(ID),FDVal) :- b_global_sets:lookup_global_constant(ID,FDVal). % not yet precompiled
% to convert enumerated set elements identifier(_) we would need to know quantified identifiers in scope
eval_set_extension_aux(set_extension(List),AVL) :-
   evaluate_set_extension(List,EvaluatedList),
   convert_to_avl(EvaluatedList,AVL).
eval_set_extension_aux(sequence_extension(List),AVL) :-
   evaluate_seq_extension_to_avl(List,AVL).
%eval_set_extension_aux(E,_) :- print(uncov(E)),nl,nl,fail.

eval_set_extension_fields([],[]).
eval_set_extension_fields([field(Name,V)|T],[field(Name,EV)|ET]) :- eval_set_extension_element(V,EV),
   eval_set_extension_fields(T,ET).

convert_set_to_seq([],_,[]).
convert_set_to_seq([H|T],N,[(int(N),H)|CT]) :- N1 is N+1, convert_set_to_seq(T,N1,CT).


post_let_forall(AllIds,P,Rhs,NewPred,modification) :-
    conjunction_to_list(P,Preds), reverse(Preds,RPreds),
    select_equality(TId,RPreds,[],_,Expr,RRest,UsedIds,no_check),
    select(TId,AllIds,RestIds),
    get_texpr_id(TId,Id),
    \+ member(Id,UsedIds), % not a recursive equality
    reverse(RRest,Rest),
    conjunct_predicates(Rest,RestPred),
    \+ occurs_in_expr(Id,RestPred),
    !,
    NewRhs = b(let_predicate([TId],[Expr],Rhs),pred,[]),
    post_let_forall(RestIds,RestPred,NewRhs,NewPred,_).
post_let_forall(AllIds,P,Rhs,NewPred, no_modification) :-
    create_implication(P,Rhs,NewForallBody),
    create_forall(AllIds,NewForallBody,NewP),
    get_texpr_expr(NewP,NewPred).


is_interval(b(interval(A,B),_,_),A,B).

% a more flexible version, also detecting singleton set extension
is_interval_or_singleton(I,A,B) :- is_interval(I,A,B),!.
is_interval_or_singleton(b(set_extension([A]),set(integer),_),A,A).

is_empty_set(b(ES,_,_)) :- is_empty_set_aux(ES).
is_empty_set_aux(empty_set).
is_empty_set_aux(empty_sequence).
is_empty_set_aux(domain(D)) :- is_empty_set(D).
is_empty_set_aux(range(D)) :- is_empty_set(D).

% create a lambda expression for a projection
create_projection_set(A,B,_Switch,Res) :-
    (is_empty_set(A) ; is_empty_set(B)),!,
    Res = empty_set.
create_projection_set(A,B,Switch,lambda(Ids,SPred,Expr)) :- %  generate lambda to be able to use function(lambda) rule
    Ids = [TArg1,TArg2],
    ( Switch==first  -> Expr = TArg1
    ; Switch==second -> Expr = TArg2),
    get_texpr_type(A,TA1), unify_types_strict(TA1,set(Type1)),
    get_texpr_type(B,TB2), unify_types_strict(TB2,set(Type2)),
    (contains_no_ids(A,B) -> Arg1 = '_zzzz_unary', Arg2 = '_zzzz_binary' % avoid generating fresh ids; relevant for test 1313 and ticket PROB-346
    % TO DO: check whether _zzzz_unary/binary are actually used; we should avoid generating fresh ids whenever possible (otherwise syntactically identical formulas become different)
     ; get_unique_id_inside('_prj_arg1__',A,B,Arg1),
       get_unique_id_inside('_prj_arg2__',A,B,Arg2)),
    create_texpr(identifier(Arg1),Type1,[generated(Switch)],TArg1),
    create_texpr(identifier(Arg2),Type2,[generated(Switch)],TArg2),
    safe_create_texpr(member(TArg1,A),pred,MembA),
    safe_create_texpr(member(TArg2,B),pred,MembB),
    conjunct_predicates([MembA,MembB],Pred), SPred=Pred.
%    bsyntaxtree:mark_bexpr_as_symbolic(Pred,SPred). % TO DO: put mark code into another module; maybe only mark as symbolic if types large enough ??

contains_no_ids(A,B) :- contains_no_ids(A), contains_no_ids(B).
contains_no_ids(b(E,_,_)) :- contains_no_ids_aux(E).
contains_no_ids_aux(bool_set).
contains_no_ids_aux(X) :- is_integer_set(X,_). % comprehension set may contain ids, but not visible to outside
contains_no_ids_aux(mult_or_cart(A,B)) :- contains_no_ids(A),contains_no_ids(B).
contains_no_ids_aux(relations(A,B)) :- contains_no_ids(A),contains_no_ids(B).
contains_no_ids_aux(pow_subset(A)) :- contains_no_ids(A).
contains_no_ids_aux(real_set).
contains_no_ids_aux(string_set).
contains_no_ids_aux(interval(A,B)) :- contains_no_ids(A),contains_no_ids(B).
% TO DO: add more

create_event_b_projection_set(Rel,Switch,lambda(Ids,SPred,Expr)) :-
    Ids = [TArg1,TArg2],
    ( Switch==first  -> Expr = TArg1
    ; Switch==second -> Expr = TArg2),
    get_texpr_type(Rel,RT),unify_types_strict(RT,set(couple(Type1,Type2))),
    get_unique_id_inside('_prj_arg1__',Rel,Arg1),
    get_unique_id_inside('_prj_arg2__',Rel,Arg2),
    create_texpr(identifier(Arg1),Type1,[generated(Switch)],TArg1),
    create_texpr(identifier(Arg2),Type2,[generated(Switch)],TArg2),
    create_texpr(couple(TArg1,TArg2),couple(Type1,Type2),[],Couple),
    safe_create_texpr(member(Couple,Rel),pred,Member),
    SPred=Member.
    %bsyntaxtree:mark_bexpr_as_symbolic(Pred,SPred).

create_event_b_projection_set_v2(RelType,Switch,comprehension_set(Ids,SPred)) :-
    % we are generating {p1,p2,lambda | lambda=p1/p2}
    Ids = [TArg1,TArg2,TArg3],
    ( Switch==first  -> ResultExpr = TArg1, Type1 = Type3
    ; Switch==second -> ResultExpr = TArg2, Type2 = Type3),
    unify_types_strict(RelType,set(couple(couple(Type1,Type2),T3))),
    (T3==Type3 -> true ; add_error(create_event_b_projection_set,'Unexpected return type: ',T3)),
    Arg1 = '_zzzz_unary',
    Arg2 = '_zzzz_binary',
    Arg3 = '_lambda_result_', % the comprehension set contains no other expressions: no clash possible
    create_texpr(identifier(Arg1),Type1,[generated(Switch)],TArg1),
    create_texpr(identifier(Arg2),Type2,[generated(Switch)],TArg2),
    create_texpr(identifier(Arg3),Type3,[lambda_result(Arg3),generated(Switch)],TArg3),
    safe_create_texpr(equal(TArg3,ResultExpr),pred,[prob_annotation('LAMBDA-EQUALITY')],Equal),
    SPred=Equal.
%    bsyntaxtree:mark_bexpr_as_symbolic(Pred,SPred).
    %,print(Pred),nl.

:- use_module(btypechecker,[couplise_list/2,prime_identifiers/2,prime_identifiers0/2, prime_atom0/2]).
% create a comprehension set for quantified union or intersection UNION(x).(P|E) = ran(%x.(P|E))
% TO DO: translate UNION into UNION(x).(P|E)  =  dom({r,x|P & r:E}) = ran({x,r|P & r:E}) which is considerably faster
% also works for e.g., UNION(x).(x:1..2|{x+y}) = 12..13
%quantified_union_op(Ids,Pred,Expr,SetType,Res) :- is_set_type(SetType,Type),
%    !,
%    Info = [generated(quantified_union)],
%    get_unique_id_inside('_zzzz_unary',Pred,Expr,FRESHID), % also include Expr !
%    NewID = b(identifier(FRESHID),Type,[]), %fresh
%    append(Ids,[NewID],NewIds),
%    safe_create_texpr(member(NewID,Expr),pred,[],Member),
%    conjunct_predicates([Pred,Member],Body),
%    get_texpr_types(NewIds,Types),couplise_list(Types,TupleType),
%    safe_create_texpr(comprehension_set(NewIds,Body),set(TupleType),Info,ComprSet),
%    Res = range(ComprSet).
%    %safe_create_texpr(range(ComprSet),set(set(Type)),Info,Set), print_bexpr(Set),nl.
%quantified_union_op(Ids,Pred,Expr,SetType,Set) :-
%    add_internal_error('Could not translate quantified UNION operator: ',quantified_union_op(Ids,Pred,Expr,SetType,Set)),
%    fail.

% create a comprehension set for quantified union or intersection INTER(x).(P|E) = inter(ran(%x.(P|E)))
% UNION could be treated by quantified_union_op above
quantified_set_op(Ids,Pred,Expr,Loc,OuterInfos,Set) :-
    create_range_lambda(Ids,Pred,Expr,Loc,OuterInfos,Set),
    !. % , print(quantified),nl,print_bexpr(Set),nl.
quantified_set_op(Ids,Pred,Expr,Loc,OuterInfos,Set) :-
    add_internal_error('Could not translate quantified set operator: ',
       quantified_set_op(Ids,Pred,Expr,Loc,OuterInfos,Set)),
    fail.

create_range_lambda(Ids,Pred,Expr,Loc,OuterInfos,Set) :-
    Info = [generated(Loc)|OuterInfos],
    get_texpr_types(Ids,Types),couplise_list(Types,ArgType),
    get_texpr_type(Expr,ExprType),
    safe_create_texpr(lambda(Ids,Pred,Expr),set(couple(ArgType,ExprType)),Info,Lambda),
    safe_create_texpr(range(Lambda),set(ExprType),Info,Set).


quantified_set_operator(quantified_union(AllIds,Pred,Expr),quantified_union,AllIds,Pred,Expr).
quantified_set_operator(quantified_intersection(AllIds,Pred,Expr),quantified_intersection,AllIds,Pred,Expr).

:- use_module(tools_strings,[ajoin/2,ajoin_with_sep/3]).
% we ensure that this check is only done once, for user machines,... not for generated formulas
check_forall_lhs(_,_,_) :- preferences:get_preference(perform_stricter_static_checks,false),!.
check_forall_lhs(_,_,_) :- preferences:get_preference(disprover_mode,true),!.
check_forall_lhs(_,_,_) :- animation_minor_mode(eventb),!. % typing predicates get removed it seems
check_forall_lhs(_,I,_) :- member(removed_typing,I),!. % means that typing was possibly removed
check_forall_lhs(P,_,_) :- member_in_conjunction(PC,P),get_texpr_info(PC,PI),member(II,PI),removed_typing(II),!. % it was something else; does not seem to detect all removed conjunctions, hence we also check I above
check_forall_lhs(P,I,Ids) :- find_identifier_uses_if_necessary(P,[],LhsUsed),
   ord_subtract(Ids,LhsUsed,NotDefined),
   (NotDefined=[] -> true
    ; ajoin_with_sep(NotDefined,',',S),
      translate:translate_bexpression(P,PS),
      ajoin(['Left-hand-side "', PS, '" of forall does not define identifier(s): '],Msg),
      add_warning(b_ast_cleanup,Msg,S,I)).
removed_typing(removed_typing). removed_typing(was(_)).

:- use_module(kernel_records,[normalise_record_type/2]).
:- use_module(library(lists),[last/2]).

% first the rules that require the path:
cleanup_post_with_path(assign([b(identifier(ID),TYPE,INFO)],[EXPR]),subst,I,
             assign_single_id(b(identifier(ID),TYPE,INFO),EXPR),subst,I,single/assign_single_id,Path) :-
    \+ animation_minor_mode(eventb), % there is no support in the Event-B interpreter for assign_single_id yet
    (simple_expression(EXPR) % the assign_single_id is not guarded by a waitflag; EXPR should not be too expensive too calculate
      -> true
       ; % if we are in an unguarded context; then we do not need to guard EXPR by waitflag anyway
         maplist(unguarded,Path)
    ),
    !,
    (debug_level_active_for(4) -> format('Single Assignment to ~w~n',[ID])
        %translate:print_subst(b(assign([b(identifier(ID),TYPE,INFO)],[EXPR]),subst,I)),nl
     ;  true).
cleanup_post_with_path(any(Ids,Pred,Subst),subst,Info,any(Ids,Pred,NewSubst),subst,NewInfo,multi/remove_useless_assign,_Path) :-
     get_preference(optimize_ast,true),
     member_in_conjunction(b(equal(TID1,TID2),pred,_),Pred),
     get_texpr_id(TID1,ID1),
     get_texpr_id(TID2,ID2), % we have an equality of the form x=x'  (as generated by TLA2B)
     delete_assignment(Subst,TID3,TID4,NewSubst),
     get_texpr_id(TID4,ID4), get_texpr_id(TID3,ID3),
     ( c(ID1,ID2) = c(ID3,ID4) ; c(ID1,ID2) = c(ID4,ID3)), % we have an assignment x:=x' or x':=x
     % the cleanup rule recompute_accessed_vars below recomputes the info fields for enclosing operations!  get_accessed_vars is currently called before ast_cleanup
     ajoin([ID3,' := ',ID4],Assign),
     add_hint_message(remove_useless_assign,'Removing useless assignment: ',Assign,Info),
     (member(removed_useless_assign,Info) -> NewInfo=Info ; NewInfo=[removed_useless_assign|Info]).
cleanup_post_with_path(operation(TName,Res,Params,TBody),Type,Info,
                       operation(TName,Res,Params,NewTBody),Type,NewInfos,single/recompute_accessed_vars,_Path) :-
    TBody=b(Body,subst,BInfos),
    select(removed_useless_assign,BInfos,NewBInfos),
    btypechecker:compute_accessed_vars_infos_for_operation(TName,Res,Params,TBody,Modifies,_,_,NewRWInfos),
    debug_format(19,'Recomputing read/write infos for ~w (~w)~n',[TName,Modifies]),
    update_infos(NewRWInfos,Info,NewInfos),
    NewTBody=b(Body,subst,NewBInfos).
cleanup_post_with_path(any(Ids,Pred,Subst),subst,I,Result,subst,[generated|I],single/transform_any_into_let,Path) :-
    (last(Path,arg(top_level(_),_))
     /* do not remove top-level ANY if show_eventb_any_arguments is true; see, e.g., test 1271 */
     -> preferences:preference(show_eventb_any_arguments,false) ; true),
    conjunction_to_list(Pred,Preds),
    get_sorted_ids(Ids,BlacklistIds),
    find_one_point_rules(Ids,Preds,BlacklistIds,LetIDs,Exprs,RestIds,RestPreds),
    LetIDs \= [],
    maplist(create_equality,LetIDs,Exprs,LetDefs),
      % print(found_lets(LetIDs,RestIds,RestPred)),nl,print(Path),nl,
    conjunct_predicates(LetDefs,LetDefPred),
    (RestIds = [], RestPreds=[] % complete ANY can be translated to LET
       -> Result = let(LetIDs,LetDefPred,Body), Body = Subst
      ;  split_predicates(RestPreds,Ids,RestUsingIds,RestNotUsingIds),
         conjunct_predicates(RestPreds,RestPred),
         % print('USING: '), print_bexpr(RestUsingIds),nl, print('NOT USING: '), print_bexpr(RestNotUsingIds),nl,
         (RestIds = []
           -> (is_truth(RestUsingIds)
               % RestPred does not use the LET identifiers; move outside of the LET !
                -> Result = select([b(select_when(RestPred,SelectBody),subst,[generated|I])]),
                   SelectBody = b(let(LetIDs,LetDefPred,Subst),subst,[generated|I])
                ; is_truth(RestNotUsingIds)
               % RestPred uses LET identifiers in all conjuncts; move inside LET
                -> Result = let(LetIDs,LetDefPred,LetBody),
                   LetBody = b(select([b(select_when(RestPred,Subst),subst,[])]),subst,[generated|I])
                ; otherwise
                % we would need to generate an outer and inner select; transformation probably not worth it
                 -> fail
              )
            ; is_truth(RestUsingIds)
               % RestPred does not use LET identifiers move outside of LET
              -> Result = any(RestIds,RestPred,SelectBody),
                 SelectBody = b(let(LetIDs,LetDefPred,Subst),subst,[generated|I])
            ; is_truth(RestNotUsingIds)
               % RestPred uses LET identifiers in all conjuncts; move inside LET
              -> Result = let(LetIDs,LetDefPred,LetBody),
                 LetBody = b(any(RestIds,RestPred,Subst),subst,I)
            ; otherwise
                % we would need to generate an outer and inner any; transformation probably not worth it
                 -> fail
            )
     ),
     !. %,translate:print_subst(b(Result,subst,[])),nl.
cleanup_post_with_path(operation(TName,Res,Params,Body),Type,Info,
                       operation(TName,Res,Params,NewBody),Type,Info,single/lts_min_guard_splitting,Path) :-
    (get_preference(ltsmin_guard_splitting,true) ; \+ get_preference(pge,off)), % equivalent to pge_algo:is_pge_opt_on),
    Path = [arg(top_level(operation_bodies),_)], % only apply at top-level and TO DO: only for top-most machine !!
    % TO DO: also apply for Event-B models
    get_texpr_id(TName,Name),
    (predicate_debugger:get_operation_propositional_guards(Name,Res,Params,Body,Guards,RestBody)
     -> true
     ;  add_warning(ltsmin_guard_splitting,'Cannot extract guard for:',Name),fail),
    Guards \= [],
    conjunct_predicates(Guards,G),
    get_texpr_info(Body,BInfo),
    NewBody = b(precondition(G,RestBody),subst,[prob_annotation('LTSMIN-GUARD')|BInfo]), % a SELECT would be more appropriate
    (debug_mode(off) -> true
      ; format('Extracting LTS Min guard for ~w~n',[Name]),translate:print_subst(NewBody),nl).
cleanup_post_with_path(exists(Ids,B),Type,OInfo,exists(Ids,B),Type,NInfo,single/invalidate_used_ids,_Path) :-
    delete(OInfo,used_ids(_),NInfo). % the changes done by cleanups, e.g, removing unused predicates can affect used_ids
cleanup_post_with_path(forall(Ids,LHS,RHS),Type,OInfo,forall(Ids,LHS,RHS),Type,NInfo,single/invalidate_used_ids,_Path) :-
    delete(OInfo,used_ids(_),NInfo). % ditto
cleanup_post_with_path(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,Path) :-
    get_preference(optimize_ast,true),
    cleanup_post_ne_with_path(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,Path).
cleanup_post_with_path(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,_Path) :-
    cleanup_post_essential(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule).

% delete an assignment from a substitution
delete_assignment(b(assign(LHS,RHS),subst,Info),ID,IDRHS,b(RES,subst,Info)) :-
   nth1(Pos,LHS,ID,RestLHS),
   nth1(Pos,RHS,IDRHS,RestRHS),
   (RestLHS = [] -> RES = skip ; RES = assign(RestLHS,RestRHS)).
% TO DO: also deal with parallel and possibly other constructs assign_single_id,...

%unguarded(arg(sequence/1,1)). % first argument of sequence is not guarded
unguarded(arg(X,_)) :- unguarded_aux(X).
unguarded_aux(top_level(_)).
unguarded_aux(operation/4).
unguarded_aux(parallel/1).
unguarded_aux(var/2).
unguarded_aux(let/3).
% what about choice/2 ??

:- load_files(library(system), [when(compile_time), imports([environ/2])]).


% ---------------------

% now the rules which do not need the Path

% first check for a few expressions that never need to be optimised, rewritten:
cleanup_post_essential(E,_,_,_,_,_,_) :- never_transform_or_optimise(E),!,fail.

cleanup_post_essential(Expr,Type,I,Expr,Type,I2,single/remove_erroneous_info_field) :-
    \+ ground(I),!,
    functor(Expr,F,_N),
    add_internal_error('Information field not ground: ',I:F),
    I2=[].
:- if(environ(prob_safe_mode,true)).
cleanup_post_essential(comprehension_set(Ids,E),Type,I,comprehension_set(Ids,E),Type,I,single/sanity_check) :-
    get_texpr_ids(Ids,UnsortedIds),sort(UnsortedIds,SIds),
    \+ same_length(UnsortedIds,SIds),
    add_error(cleanup_post,'Identifier clash in comprehension set: ',UnsortedIds),
    print(E),nl,
    fail.
:- endif.
%cleanup_post(block(TS),subst,_,Subst,subst,Info,multi/remove_block) :-
%    !,get_texpr_expr(TS,Subst),get_texpr_info(TS,Info).
cleanup_post_essential(Expr,Type,I,Expr,NType,I,single/normalise_record_type) :-
    nonvar(Type),Type=record(Fields),
    normalise_record_type(record(Fields),NType),
    NType \== Type,
    !.

cleanup_post_essential(lambda(Ids,Pred,Expr),Type,I,
                       comprehension_set(CompIds,CompPred),Type,NewInfo,multi/remove_lambda) :- !,
    NewInfo = I, % was NewInfo = [was(lambda)|I],
    unify_types_strict(Type,set(couple(_ArgType,ResType))),
    %ResultId = '_lambda_result_',
    get_unique_id_inside('_lambda_result_',Pred,Expr,ResultId),
    % get_unique_id('_lambda_result_',ResultId),
    def_get_texpr_id(Result,ResultId),
    get_texpr_type(Result,ResType),
    get_texpr_info(Result,[lambda_result(ResultId)]),
    append(Ids,[Result],CompIds),
    get_texpr_expr(Equal,equal(Result,Expr)),
    get_texpr_type(Equal,pred),
    extract_important_info_from_subexpression(Expr,EqInfo), % mark equality with wd condition if Expr has wd condition
    get_texpr_info(Equal,[prob_annotation('LAMBDA-EQUALITY')|EqInfo]),
    conjunct_predicates_with_pos_info(Pred,Equal,CompPred0),
    add_texpr_infos(CompPred0,[prob_annotation('LAMBDA')],CompPred). % checked e.g. in is_converted_lambda_closure

cleanup_post_essential(reflexive_closure(Rel),Type,I, UNION,Type,NewInfo,multi/remove_reflexive_closure) :- !,
    NewInfo = [was(reflexive_closure)|I],
    safe_create_texpr(closure(Rel),Type,I,CL),
    UNION = union(b(event_b_identity,Type,IdInfo), CL), % closure(R) = id \/ closure1(R)
    (is_infinite_type(Type) -> IdInfo = [prob_annotation('SYMBOLIC')|I] ; IdInfo =I),
    (debug_mode(on) -> print('Rewriting closure to: '), print_bexpr(b(UNION,Type,[])),nl ; true).
cleanup_post_essential(evb2_becomes_such(Ids,Pred),subst,I,becomes_such(Ids,Pred2),subst,I,multi/ev2_becomes_such) :-
  % we translate a Classical-B becomes_such with id -> id$0, id' -> id
  % classical B: Dec = BEGIN level : (level>=0 & level> level$0-5 & level < level$0) END
  % Event-B: level'>= 0, level' > level-5 ...
    !,
    prime_identifiers(Ids,PIds),
    maplist(gen_rename,PIds,Ids,RenameList1), % id' -> id
    prime_identifiers0(Ids,PIds0),
    maplist(gen_rename,Ids,PIds0,RenameList2), % id -> id$0
    append(RenameList1,RenameList2,RenameList),
    rename_bt(Pred,RenameList,Pred2),
    (debug_mode(off) -> true
      ; format('Converting Event-B becomes_such: ',[]), print_bexpr(Pred2),nl).
cleanup_post_essential(successor,Type,I,Compset,Type,[was(successor)|I],multi/successor) :- !,
    % translation of succ
    pred_succ_compset(add,Compset).
cleanup_post_essential(rev(A),string,I,ExtFunCall,string,[was(rev)|I],multi/rev_for_string) :- !,
    % translation of rev for STRINGs, this can only occur when allow_sequence_operators_on_strings is true
    ExtFunCall = external_function_call('STRING_REV',[A]).
cleanup_post_essential(concat(A,B),string,I,ExtFunCall,string,[was(concat)|I],multi/concat_for_string) :- !,
    % translation of concat (^) for STRINGs, this can only occur when allow_sequence_operators_on_strings is true
    ExtFunCall = external_function_call('STRING_APPEND',[A,B]).
cleanup_post_essential(general_concat(A),string,I,ExtFunCall,string,[was(general_concat)|I],multi/concat_for_string) :- !,
    % translation of conc for seq(STRING) --> STRING, this can only occur when allow_sequence_operators_on_strings is true
    ExtFunCall = external_function_call('STRING_CONC',[A]).
cleanup_post_essential(size(A),integer,I,ExtFunCall,integer,[was(A)|I],multi/concat_for_string) :-
    get_texpr_type(A,string),!,
    % translation of size for STRING, this can only occur when allow_sequence_operators_on_strings is true
    ExtFunCall = external_function_call('STRING_LENGTH',[A]).
cleanup_post_essential(predecessor,Type,I,Compset,Type,[was(predecessor)|I],multi/predecessor) :- !,
    % translation of pred
    pred_succ_compset(minus,Compset).
cleanup_post_essential(becomes_such(Ids1,Pred),subst,I,becomes_such(Ids2,Pred),subst,I,single/becomes_such) :- !,
    annotate_becomes_such_vars(Ids1,Pred,Ids2).
cleanup_post_essential(Expr,Type,I,Expr,Type,[contains_wd_condition|I],multi/possibly_undefined) :-
    % multi: rule can only be applied once anyway, no need to check
    nonmember(contains_wd_condition,I),
     % print(' - CHECK WD: '), print_bexpr(Expr),nl, %%
    is_possibly_undefined(Expr),!,
     %% print('CONTAINS WD: '), print_bexpr(Expr),nl, %%
     %(translate:translate_bexpression(Expr,'{min(xunits)}') -> trace ; true),
    true.
% if a substitution has a sub-expression that is a substitution with that refers to the original
% value of a variable, we mark this substitution, too.

% If the substitution of an operation contains a while whose invariant contains references x$0
% to the original value of a variable x, we must insert a LET substitution to preserve the original value.
cleanup_post_essential(operation(Id,Results,Args,Body),Type,I,operation(Id,Results,Args,NewBody),Type,I,single/refers_to_old_state_let) :-
    get_texpr_info(Body,BodyInfo),
    memberchk(refers_to_old_state(References),BodyInfo),!,
    create_equalities_for_let(References,Ids,Equalities),
    conjunct_predicates(Equalities,P),
    insert_let(Body,Ids,P,NewBody).
cleanup_post_essential(Subst,subst,I,Subst,subst,NI,single/refers_to_old_state) :-
    safe_syntaxelement(Subst,Subs,_,_,_),
    % check if a child contains the refers_to_old_state flag
    findall(Reference, (member(Sub,Subs),
                         get_texpr_info(Sub,SubInfo),
                         memberchk(refers_to_old_state(References),SubInfo),
                         member(Reference,References)), ReferedIds),
    ReferedIds \= [],
    !,
    sort(ReferedIds,SortedIds),
    NI = [refers_to_old_state(SortedIds)|I].

cleanup_post_essential(let_predicate([],[],TExpr),Type,Iin,Expr,Type,Iout,multi/remove_let_predicate) :- !,
    % same as above, just for predicates
    get_texpr_expr(TExpr,Expr),
    get_texpr_info(TExpr,I), debug_println(9,removed_let_predicate),
    % The next is done to prevent removing position information (in case of Event-B invariants)
    ( nonmember(nodeid(_),I), member(nodeid(P),Iin) -> Iout = [nodeid(P)|I]
    ; otherwise                                     -> Iout = I).


cleanup_post_essential(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule) :-
    get_preference(optimize_ast,true),
    cleanup_post(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule).

% moved these rule below the simplifciation rules above to avoid re-computing used_id infos
cleanup_post_essential(forall(Ids,Lhs,Rhs),pred,IOld,Res,pred,INew,single/forall_used_identifier) :-
    reconstruct_forall(Ids,Lhs,Rhs,IOld, b(Res,pred,INew)).
cleanup_post_essential(exists(Ids,P),pred,IOld,exists(Ids,NP),pred,ResInfo,single/exists_used_identifier) :-
    inner_predicate_level_optimizations(P,NP),
    add_used_identifier_info(Ids,NP,IOld,INew0),
    add_used_ids_defined_by_equality(Ids,NP,INew0,INew),
    %print('* POST => '),print_bexpr(b(exists(Ids,NP),pred,INew)),nl,
    %print('   INFO=> '),print(INew),nl,
    % TO DO: also compute which identifiers are worth waiting for; do not wait for res in #x.(x:E & ... & res=min(f(x)))
    add_removed_typing_info(INew,ResInfo).
cleanup_post_essential(Construct,Type,I,NewConstruct,Type,I,single/detect_partitions) :-
    contains_predicate(Construct,Type,Pred,NewConstruct,NewPred),!, % something like a select or other substitution
    predicate_level_optimizations(Pred,NewPred).
cleanup_post_essential(Construct,Type,I,NewConstruct,Type,I,single/detect_partitions2) :-
    contains_predicates(Construct,Type,Preds,NewConstruct,NewPreds),!,
    maplist(predicate_level_optimizations,Preds,NewPreds)  % TO DO: do CSE together (in some cases) !
    . %, (Preds=NewPreds -> true ; print('Found partitions: '), translate:print_subst(b(NewConstruct,Type,I)),nl).

construct_inner_forall(Ids,LHS,RHS,OldInfo,Res) :-
    delete(OldInfo,used_ids(_),I1), % we construct a changed forall construct; old used_ids info possibly incorrect
    reconstruct_forall(Ids,LHS,RHS,I1,Res).
% use when in cleanup_post you construct a forall which will not be at the top-level
% (meaning that the above essential rules will not run)
reconstruct_forall(Ids,LHS,RHS,OldInfo, b(Res,pred,NewInfo)) :-
    inner_predicate_level_optimizations(LHS,NLhs),
    inner_predicate_level_optimizations(RHS,NRhs),
    construct_forall_opt(Ids,NLhs,NRhs,OldInfo, Res,NewInfo).

construct_forall_opt(IDs,NLhs,NRhs,Info,  Res,NewInfo) :-
    (is_truth(NRhs) ; is_falsity(NLhs)),!,
    add_hint_message(remove_useless_assign,'Removing useless universal quantification','',Info),
    Res= truth, NewInfo = [was(forall(IDs,NLhs,NRhs))|Info].
% TO DO: is the following rule useful ?: will require adapting test 510 output file
% triggers e.g. for test 1447
%construct_forall_opt([TID],LHS,RHS,Info, Res,NewInfo) :-  % !x. (x:SetA => x:SetB) ---> SetA <: SetB
%   is_valid_id_member_check(LHS,TID,SetA), is_valid_id_member_check(RHS,TID,SetB),
%   !,
%   (debug_mode(off) -> true
%     ; format('Replacing forall ~w by subset: ',[TID]), print_bexpr(b(subset(SetA,SetB),pred,Info)),nl
%    ),
%    Res = subset(SetA,SetB), NewInfo = [was(forall)|Info].
construct_forall_opt(Ids,NLhs,NRhs,OldInfo,  forall(Ids,NLhs,NRhs),ResInfo) :-
    conjunct_predicates([NLhs,NRhs],P),
    add_used_identifier_info(Ids,P,OldInfo,Info),
    add_removed_typing_info(Info,ResInfo).

add_removed_typing_info(Info,ResInfo) :-
    (memberchk(removed_typing,Info) -> ResInfo = Info ; ResInfo = [removed_typing|Info]).

disjoint_ids(Ids1,Ids2) :-
   get_texpr_ids(Ids1,I1), sort(I1,SI1),
   get_texpr_ids(Ids2,I2), sort(I2,SI2),
   ord_disjoint(SI1,SI2).
% ---------------------

% non-essential post cleanup rules; only applied when optimize_ast is TRUE

% WITH PATH:

cleanup_post_ne_with_path(member(E,b(image(Rel,SONE),TypeImg,InfoImg)),pred,I,member(Couple,Rel),pred,I,multi/replace_image_by_member,Path) :-
     % x : Rel[{One}]  =>  One|->x : Rel
     Path \= [arg(forall/3,1)|_], % not LHS of a forall
     %% do not do this if it is the LHS of a forall: !(aus2).( aus2 : helper[{mm}] => RHS) (as we no longer can apply the optimized set treatment for forall
     singleton_set_extension(SONE,One),
     %% Rel \= b(inverse(_),_,_),  %% TO DO: maybe exclude this; here user maybe wants to explicitly compute image ?
      !,
     create_couple(One,E,Couple),
     (debug_mode(off) -> true
      ;  print('Member of Image: '),print_bexpr(b(member(E,b(image(Rel,SONE),TypeImg,InfoImg)),pred,I)),nl,
         print('    replaced by: '),print_bexpr(b(member(Couple,Rel),pred,I)),nl
     ).


% WITHOUT PATH:

cleanup_post(conjunct(b(truth,pred,I1),b(B,pred,I2)),pred,I0,B,pred,NewI,multi/remove_truth_conj1) :- !,
    include_important_info_from_removed_pred(I1,I2,I3), % ensure was,... information propagated
    add_important_info_from_super_expression(I0,I3,NewI).
cleanup_post(conjunct(b(A,pred,I2),b(truth,pred,I1)),pred,I0,A,pred,NewI,multi/remove_truth_conj2) :- !,
    include_important_info_from_removed_pred(I1,I2,I3),
    add_important_info_from_super_expression(I0,I3,NewI).
cleanup_post(conjunct(b(falsity,pred,I1),b(_,_,I2)),pred,I0,falsity,pred,NewI,multi/simplify_falsity_conj1) :- !,
    include_important_info_from_removed_pred(I2,I1,I3),
    add_important_info_from_super_expression(I0,I3,NewI).
cleanup_post(conjunct(LHS,b(falsity,pred,I2)),pred,I0,falsity,pred,NewI,multi/simplify_falsity_conj2) :-
   always_well_defined_or_disprover_mode(LHS), !,  % do we need to check WD Implications ; ProB would treat falsity first?
    get_texpr_info(LHS,I1),
    include_important_info_from_removed_pred(I1,I2,I3),
    add_important_info_from_super_expression(I0,I3,NewI).
   % we use FInfo: in case it has a was(.) field, e.g., for pretty printing and unsat core generation and unsatCore.groovy test
cleanup_post(conjunct(AA,BB),pred,I,Res,pred,I,multi/modus_ponens) :-
    Impl = b(implication(A,B),pred,_),
    ((AA,BB) = (Impl,A2) ; (BB,AA) = (Impl,A2)),
    same_texpr(A,A2),
    % arises e.g., for predicates such as IF x:0..3 THEN y=2 ELSE 1=0 END ; works with simplify_falsity_impl3 rule
    % rewrite (A=>B) & A into (A&B)
    !,
    (debug_mode(off) -> true ; print('Modus Ponens: '),print_bexpr(A), print(' => '), print_bexpr(B),nl),
    Res = conjunct(A,B).
%cleanup_post(conjunct(TLHS,P1),pred,_,LHS,pred,Info,multi/duplicate_pred) :-
%   % TO DO: implement an efficient version of this; currently very slow e.g. for test 293
%    b_interpreter:member_conjunct(P2,TLHS,_),
%    same_texpr(P1,P2),
%    TLHS = b(LHS,pred,Info),
%    print('remove_duplicate: '), print_bexpr(P1),nl.
cleanup_post(conjunct(LHS,b(Comparison1,pred,_)),pred,I0,Result,pred,RInfo,multi/detect_interval1) :-
    %  X <= UpBound & X >= LowBound  <=>   X : UpBound .. LowBound   (particularly useful when CLPFD FALSE, causes problem with test 1771)
    % Note: x>18 & y<1024 & x<20 & y>1020  now works, it is bracketed ((()) & y>1020)
    get_preference(use_clpfd_solver,false),
    \+ data_validation_mode, % this rule may lead to additional enumerations
    get_leq_comparison(Comparison1,X,UpBound),
    select_conjunct(b(Comparison2,_,_),LHS,Prefix,Suffix),
    get_geq_comparison(Comparison2,X2,LowBound),
    same_texpr(X,X2),
    (always_well_defined_or_disprover_mode(UpBound)
      -> true
      ;   % as we may move valuation earlier, we have to be careful
          % we check if Comparison2 is last conjunct; x=7 & x:8..(1/0) raises no WD error in ProB
          Suffix=[]
    ),
    !,
    create_interval_member(X,LowBound,UpBound,Member),
    append(Prefix,[Member|Suffix],ResultList),
    conjunct_predicates(ResultList,TResult),
    (debug_mode(off) -> true ; print(' Detected interval membership (1): '),print_bexpr(TResult),nl),
    TResult = b(Result,pred,I1),
    add_important_info_from_super_expression(I0,I1,RInfo).
cleanup_post(conjunct(LHS,b(Comparison1,pred,_)),pred,I0,Result,pred,RInfo,multi/detect_interval2) :-
    %  X >= LowBound & X <= UpBound  <=>   X : UpBound .. LowBound
    get_preference(use_clpfd_solver,false),
    \+ data_validation_mode, % this rule may lead to additional enumerations
    get_geq_comparison(Comparison1,X,LowBound),
    select_conjunct(b(Comparison2,_,_),LHS,Prefix,Suffix),
    get_leq_comparison(Comparison2,X2,UpBound),
    same_texpr(X,X2),
    (always_well_defined_or_disprover_mode(LowBound)
      -> true
      ;   % as we may move valuation earlier, we have to be careful
          % we check if Comparison2 is last conjunct; x=7 & x:8..(1/0) raises no WD error in ProB
          Suffix=[]
    ),
    !,
    create_interval_member(X,LowBound,UpBound,Member),
    append(Prefix,[Member|Suffix],ResultList),
    conjunct_predicates(ResultList,TResult),
    (debug_mode(off) -> true ; print(' Detected interval membership (2): '),print_bexpr(TResult),nl),
    TResult = b(Result,pred,I1),
    add_important_info_from_super_expression(I0,I1,RInfo).

cleanup_post(disjunct(b(truth,pred,I1),_),pred,I0,truth,pred,I,multi/simplify_truth_disj1) :- !,
    add_important_info_from_super_expression(I0,I1,I).
%cleanup_post(disjunct(_,b(truth,pred,_)),pred,_,truth,pred,I,multi/simplify_truth_disj2) :- !. % WD Implications ???
cleanup_post(disjunct(b(A,pred,I1),b(falsity,pred,_)),pred,I0,A,pred,I,multi/remove_falsity_disj1) :- !,
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(disjunct(b(falsity,pred,_),b(B,pred,I1)),pred,I0,B,pred,I,multi/remove_falsity_disj2) :- !,
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(disjunct(Expr1,b(truth,pred,I1)),pred,I0,truth,pred,I,multi/simplify_truth_disj2) :-
    always_well_defined_or_disprover_mode(Expr1), !,
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(disjunct(Equality1,Equality2),pred,I,New,pred,I,multi/rewrite_disjunct_to_member) :-
    identifier_equality(Equality2,ID,_,Expr2),
    always_well_defined_or_disprover_mode(Expr2),
    get_texpr_type(Expr2,Type2),
    type_contains_no_sets(Type2), % we do not want to generate sets of sets, or worse sets with infinite sets (x=NATURAL1 or ...) which cannot be converted to AVL
    identifier_equality(Equality1,ID,TID,Expr1),
    % Rewrite (ID = Expr1  or  ID = Expr2)  into  ID: {Expr1,Expr2}  ;  good if FD information can be extracted for ID
    % But: can be bad for reification, in particular when set extension cannot be computed fully
    %  TO DO: also deal with  ID : {Values}  and more general extraction of more complicated disjuncts
    %  TO DO: also apply for implication  (e.g., ID /= E1 => ID=E2)
    !,
    construct_set_extension(Expr1,Expr2,SetX),
    New=member(TID,SetX),
    (debug_mode(off) -> true
        ; format('Rewrite disjunct ~w: ',[ID]),print_bexpr(SetX),nl).
cleanup_post(disjunct(CEquality1,CEquality2),pred,IOld,New,pred,INew,multi/factor_common_pred_in_disjunction) :-
    % (x=2 & y=3) or (x=2 & y=4)  ->  x=2 & (y=3 or y=4)  to improve constraint propagation
    factor_disjunct(CEquality1,CEquality2,IOld,New,INew),
    (debug_mode(off) -> true
        ; format('Factor disjunct:  ',[]),print_bexpr(b(New,pred,INew)),nl).
cleanup_post(implication(b(truth,pred,_),b(B,pred,I1)),pred,I0,B,pred,I,multi/remove_truth_impl1) :- !,
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(implication(b(falsity,pred,I1),_),pred,I0,truth,pred,I,multi/simplify_falsity_impl1) :- !,
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(implication(_,b(truth,pred,I1)),pred,I0,truth,pred,I,multi/simplify_truth_impl2) :- !,
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(implication(P,b(falsity,pred,_)),pred,I,NotP,pred,[was(implication)|I],multi/simplify_falsity_impl3) :- !,
    create_negation(P,TNotP),
    (debug_mode(off) -> true ; print_bexpr(P), print(' => FALSE simplified'),nl),
    get_texpr_expr(TNotP,NotP).
% TO DO: is the following rule useful ?:
%cleanup_post(implication(A,b(implication(B,C),pred,_)),pred,IOld,
%             implication(AB,C),pred,IOld,single/replace_implication_by_and) :-
%    % (A => B => C   <==>  (A & B) => C
%    conjunct_predicates([A,B],AB),
%    (debug_mode(off) -> true ; print('Simplifying double implication: '), print_bexpr(b(implication(AB,C),pred,IOld)),nl).
cleanup_post(equivalence(b(truth,pred,_),b(B,pred,I1)),pred,I0,B,pred,I,multi/remove_truth_equiv1) :- !,
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(equivalence(b(A,pred,I),b(truth,pred,I1)),pred,I0,A,pred,I,multi/remove_truth_equiv2) :- !,
    add_important_info_from_super_expression(I0,I1,I).
% TO DO: more rules for implication/equivalence to introduce negations  (A <=> FALSITY  --->  not(A)) ?
% detect certain tautologies/inconsistencies
cleanup_post(lazy_let_pred(_ID,_,b(Sub,pred,I1)),pred,I0,Sub,pred,I,multi/remove_lazy_let_pred) :-
    (Sub=truth ; Sub=falsity), !,
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(IFTHENELSE,T,_,Res,T,[was(ifthenelse)|NI],single/remove_if_then_else) :-
    explicit_if_then_else(IFTHENELSE,IF,THEN,ELSE),
    (is_falsity(IF) -> b(Res,_,NI)=ELSE
     ; is_truth(IF) -> b(Res,_,NI)=THEN
     ),
     (debug_mode(off) -> true
       ; print('Simplified IF-THEN-ELSE: '), print_bexpr(IF),nl).
cleanup_post(member(X,B),pred,I,truth,pred,[was(member(X,B))|I],multi/remove_type_member) :-
      is_just_type(B),
      nonmember(label(_),I), % the user has explicitly labeled this conjunct
      !. % print('REMOVE: '),print_bexpr(b(member(X,B),pred,[])),nl, print(I),nl.
%cleanup_post(member(X,SET),pred,I,greater_equal(X,TBound),pred,[was(member(X,SET))|I],multi/remove_type_member) :-
% disabled at the moment: we need to adapt test 1383, 767, 1703, 1003
%      is_inf_integer_set_with_lower_bound(SET,Bound),
      % Replace x:NATURAL by x>=0 and x:NATURAL1 by x>=1 ; is usually much more efficient
      % note : removes virtual timeout in test 290
%      !,
%      TBound = b(integer(Bound),integer,[]), print('REPLACE: '),print_bexpr(b(member(X,SET),pred,I)),nl, print_bexpr(b(greater_equal(X,TBound),pred,I)),nl.
cleanup_post(not_member(X,B),pred,I,falsity,pred,[was(not_member(X,B))|I],multi/remove_type_not_member) :-
      is_just_type(B),
      nonmember(label(_),I), % the user has explicitly labeled this conjunct
      !.
cleanup_post(member(X,TSet),pred,I,equal(X,One),pred,I,multi/remove_member_one_element_set) :-
      singleton_set_extension(TSet,One),
      !,
      %  X:{One} <=> X=One
     true. %,print('Introducing equality: '),print_bexpr(X), print(' = '), print_bexpr(One),nl.
cleanup_post(member(X,b(Set,_,_)),pred,I,not_equal(X,One),pred,I,multi/remove_member_setdiff) :-
     Set = set_subtraction(MaximalSet,SONE),
     singleton_set_extension(SONE,One),
     definitely_maximal_set(MaximalSet),
     !,  % x : INTEGER-{One}  <=>  x/=One
     (debug_mode(off) -> true
      ; print('Replacing member of set_subtraction: '), print_bexpr(MaximalSet), print(' - '), print_bexpr(SONE),nl).
cleanup_post(not_member(X,TSet),pred,I,not_equal(X,One),pred,I,multi/remove_member_one_element_set) :-
      singleton_set_extension(TSet,One),
      !,
      %  X/:{One} <=> X/=One
      true.
cleanup_post(member(E,b(fin_subset(E2),_,_)),pred,I,finite(E),pred,I,multi/introduce_finite) :-
    (same_texpr(E,E2); is_just_type(E2)),!. % print(introduce(finite(E))),nl.
/* do we want need this rule ?:
clenaup_post(member(b(couple(A,B),couple(TA,TB),IC),ID),pred,I,equal(A,B),pred,I,multi/replace_member_id) :-
     is_is_event_b_identity(ID), !.
*/
cleanup_post(member(b(couple(A,B),couple(TA,TB),IC),b(reverse(Rel),_,_)),pred,I,member(ICouple,Rel),pred,I,multi/remove_reverse) :- !,
     % (A,B) : Rel~  ===>  (B,A) : Rel
     % can be detrimental for performance when A is known and B is not and Rel is large
     \+ data_validation_mode,
     (debug_mode(off) -> true ;  print('Removed inverse (~): '),print_bexpr(Rel),nl),
     ICouple = b(couple(B,A),couple(TB,TA),IC).
cleanup_post(member(LHS,Comprehension),pred,I,Result,pred,I,multi/remove_member_comprehension) :-
     Comprehension = b(comprehension_set([TID],Body),_,_),
     get_texpr_id(TID,ID),
     % LHS:{x|P(x)}  ==>  P(LHS)
     replace_id_by_expr_with_count(Body,ID,LHS,TResult,Count),
     % rewrite could duplicate LHS: not an issue in CSE mode; optimization relevant in normalize_ast mode
     is_replace_id_by_expr_ok(LHS,ID,Count,remove_member_comprehension),
     !,
     % could introduce LET if necessary because ID occurs multiple times (Count>1)
     get_texpr_expr(TResult,Result),
     (debug_mode(off) -> true ;  print('Remove element of comprehension_set: '),print_bexpr(Comprehension),nl,
        format(' rewriting to (~w): ',[Count]),print_bexpr(TResult),nl).
cleanup_post(not_member(LHS,Comprehension),pred,I,Result,pred,I,multi/remove_not_member_comprehension) :-
     Comprehension = b(comprehension_set([TID],Body),_,_),
     get_texpr_id(TID,ID),
     % LHS/:{x|P(x)}  ==>  not(P(LHS))
     replace_id_by_expr_with_count(Body,ID,LHS,TResult,Count),
     % rewrite could duplicate LHS: not an issue in CSE mode; optimization relevant in normalize_ast mode
     is_replace_id_by_expr_ok(LHS,ID,Count,remove_not_member_comprehension),
     !,
     Result = negation(TResult),
     (debug_mode(off) -> true ;  print('Remove not element of comprehension_set: '),print_bexpr(Comprehension),nl,
        format(' rewriting to (~w): not(',[Count]),print_bexpr(TResult),print(')'),nl).
cleanup_post(comprehension_set(Ids,Body),Type,I,NewExpr,Type,I2,Rule) :-
     cleanup_comprehension_set(Ids,Body,Type,I,NewExpr,I2,Rule),
     !.
cleanup_post(subset(A,B),pred,I,truth,pred,[was(subset(A,B))|I],multi/remove_type_subset) :-
      is_just_type(B),
      nonmember(label(_),I), % the user has explicitly labeled this conjunct
      !.
cleanup_post(not_subset(A,B),pred,I,falsity,pred,[was(not_subset(A,B))|I],multi/remove_type_not_subset) :-
      is_just_type(B),
      nonmember(label(_),I), % the user has explicitly labeled this conjunct
      !.
cleanup_post(SUB,pred,I,NewPred,pred,[generated_conjunct|I],multi/replace_subset_by_element) :-
     is_subset(SUB,A,B),
     is_set_extension(A,List),
     !, % for sequence extension we don't need this as the interpreter knows exactly the cardinality of a sequence_extension ?
     % applying rule {x1,x2,...} <: B  <=>  x1:B & x2:B & ...
     maplist(gen_member_predicates(B),List,Conjuncts),
     conjunct_predicates(Conjuncts,TNewPred),
      %  print('detected subset-member rule: '),print_bexpr(TNewPred),nl,
     get_texpr_expr(TNewPred,NewPred).
cleanup_post(SUB,pred,I,NewPred,pred,[generated_conjunct|I],multi/replace_union_subset) :-
     is_subset(SUB,A,B),
     % mark conjunct as generated: used e.g. by flatten_conjunct in predicate_evaluator
     get_texpr_expr(A,union(_,_)),!,
     % applying rule A1 \/ A2 <: B  <=>  A1 <: B & A2 <: B
     % could be detrimental if checking that something is an element of B is expensive
     extract_unions(A,As),
     findall(Subi,(member(Ai,As),safe_create_texpr(subset(Ai,B),pred,I,Subi)),Conj),  % we could try and re-run clean-up ? safe_create_texpr will ensure WD info set
     conjunct_predicates(Conj,TNewPred),
     get_texpr_expr(TNewPred,NewPred).
cleanup_post(Comp,pred,I,SComp,pred,I,multi/simplify_cse_comparison) :-
  % simplify comparison operations; can result in improved constraint propagation
  % e.g., ia + CSE1 * 2 > ia + fa <=> CSE1 * 2 > fa for Setlog/prob-ttf/qsee-TransmitMemoryDumpOk21_SP_3.prob
     comparison(Comp,A,B,SComp,SA,SB),
     simplify_comparison_terms(A,B,SA,SB),!,
     (debug_mode(off) -> true
      ;  print('Simplified: '),print_bexpr(b(Comp,pred,I)),
         print(' <=> '),print_bexpr(b(SComp,pred,I)),nl).
cleanup_post(equal(A,B),pred,I,truth,pred,[was(equal(A,B))|I],multi/remove_equality) :-
   same_texpr(A,B),always_well_defined_or_disprover_mode(A),!. % ,print(removed_equal(A,B)),nl.
cleanup_post(equal(A,B),pred,I,equal(A2,B2),pred,I,multi/simplify_equality) :-
   simplify_equality(A,B,A2,B2).
cleanup_post(not_equal(A,B),pred,I,not_equal(A2,B2),pred,I,multi/simplify_inequality) :-
   simplify_equality(A,B,A2,B2).
cleanup_post(equal(A,B),pred,I,falsity,pred,[was(equal(A,B))|I],multi/remove_equality_false) :-
   different_texpr_values(A,B),!. %,print(removed_equal_false(A,B)),nl.
cleanup_post(equal(A,B),pred,I,greater(Low,Up),pred,I,multi/remove_equality) :-
   % Low..Up = {}  <=>  Low>Up   % is also handled by constraint solver; but other simplifications can apply here
   (is_empty_set(B), is_interval(A,Low,Up) ;
    is_empty_set(A), is_interval(B,Low,Up)),!,
    (debug_mode(off) -> true
      ;  print('Simplified: '), print_bexpr(b(equal(A,B),pred,I)), print(' <=> '),
         print_bexpr(b(greater(Low,Up),pred,I)),nl).
   % TO DO: card(P) = 0  ->  P={} if wd guaranteed
cleanup_post(CardGt0Expr,pred,I,not_equal(X,EmptySet),pred,I,multi/remove_cardgt0) :-
   get_geq_comparison(CardGt0Expr,Card,One),
   % card(P) > 0  ->  P\={} if wd guaranteed  ;  TODO: card(P) : NATURAL1
   Card = b(card(X),integer,_), get_integer(One,1),
   always_well_defined_or_disprover_mode(Card),
   get_texpr_type(X,TX), get_texpr_info(One,I0),
   EmptySet = b(empty_set,TX,I0),!,
    (debug_mode(off) -> true ;  print('Removed card(.) > 0 for set: '), print_bexpr(X), nl).
cleanup_post(member(Card,Natural),pred,I,truth,pred,I,multi/remove_card_natural) :-
   % card(P) : NATURAL  ->  truth if wd guaranteed
   Card = b(card(X),integer,_),
   is_integer_set(Natural,'NATURAL'),
   always_well_defined_or_disprover_mode(Card),
   !,
    (debug_mode(off) -> true ;  print('Removed card(.):NATURAL for set: '), print_bexpr(X), nl).
cleanup_post(not_equal(A,B),pred,I,less_equal(Low,Up),pred,I,multi/remove_equality) :-
   % Low..Up \= {}  <=>  Low<=Up  % is also handled by constraint solver; but other simplifications can apply here
   (is_empty_set(B), is_interval(A,Low,Up) ;
    is_empty_set(A), is_interval(B,Low,Up)),!,
    (debug_mode(off) -> true
      ;  print('Simplified: '), print_bexpr(b(equal(A,B),pred,I)), print(' <=> '),
         print_bexpr(b(less_equal(Low,Up),pred,I)),nl).
cleanup_post(equal(A,B),pred,I,equal(A,RLet),pred,I,single/detect_recursion) :-
    % "A" should be an identifier
    get_texpr_id(A,ID),
    % check if some side conditions are fulfilled where the recursion detection can be enabled
    recursion_detection_enabled(A,B,I),
    % A must be recursively used in B:
    find_recursive_usage(B,ID),
    % TO DO: also find mutual recursion !
    debug_println(9,recursion_detected(ID)),
    !, % create an recursive_let where the body is annotated to be symbolic
    get_texpr_type(B,Type),  add_texpr_infos(B,[prob_annotation('SYMBOLIC')],B2),
    mark_recursion(B2,ID,B3),
    %print(marked_recursion(ID)),nl,
    safe_create_texpr(recursive_let(A,B3),Type,RLet).
cleanup_post(equal(A,B),pred,Info1,ResultExpr,pred,Info3,multi/simplify_bool_true_false) :-
   % simplify bool(X)=TRUE -> X and bool(X)=FALSE -> not(X)
   ( get_texpr_expr(A,convert_bool(X)), get_texpr_boolean(B,BOOLVAL)
   ;
     get_texpr_boolean(A,BOOLVAL),get_texpr_expr(B,convert_bool(X))
   ),
   get_texpr_info(X,Info2),
   add_important_info_from_super_expression(Info1,Info2,Info3),
   (BOOLVAL = boolean_true  -> get_texpr_expr(X,ResultExpr) ;
    BOOLVAL = boolean_false -> create_negation(X,TNX), get_texpr_expr(TNX,ResultExpr)),
   !,
   (debug_mode(off) -> true
     ; format('Simplifying bool(.)=~w to ',[BOOLVAL]),translate:print_bexpr(b(ResultExpr,pred,Info3)),nl).
%cleanup_post(equal(A,B),pred,Info1,equal(REL,CartProd),pred,Info1,multi/simplify_image) :-
% cannot be applied yet; SETS not precompiled yet !
%    A = b(image(REL,SetExt),_,_),
%    % REL[{OneEl}] = B   ---->   REL = {OneEl}*B    if OneEl is the only possible value
%     % such signature appear in Alloy generated code
%    SetExt = b(set_extension([_]),set(global(GlobalSetName)),_),
%    %bmachine:b_get_named_machine_set_calc(GlobalSetName,_,[_]),
%    b_global_set_cardinality(Type,1), % cannot be called yet; global sets not precompiled
%    !,
%    get_texpr_type(REL,RelType),
%    safe_create_texpr(cartesian_product(SetExt,B),RelType,CartProd),
%    format('Translating image for singleton set'), print_bexpr(b(equal(REL,CartProd),pred,[])),nl.
cleanup_post(not_equal(A,B),pred,I,falsity,pred,[was(not_equal(A,B))|I],multi/remove_disequality) :-
   same_texpr(A,B),always_well_defined_or_disprover_mode(A),!. % ,print(removed_not_equal(A,B)),nl.
   % sometimes one uses & TRUE=TRUE to finish off guards, invariants, ...
   % exchange lambda expressions by a comprehension set
% TO DO: also add rule for bool(X)=FALSE -> not(X)
cleanup_post(not_equal(A,B),pred,I,truth,pred,[was(not_equal(A,B))|I],multi/remove_disequality_false) :-
   different_texpr_values(A,B),!. %  ,print(removed_not_equal_false(A,B)),nl.
cleanup_post(not_equal(A,B),pred,I,NewP,pred,[was(not_equal(A,B))|I],multi/not_disjoint_disequality) :-
  /* Set1 /\ Set2 /= {}    <===>   #(zz).(zz:Set1 & zz:Set2) */
   preferences:preference(use_smt_mode,true), /* currently this rewriting makes test 1112 fail; TO DO: investigate */
   is_empty_set(B),
   get_texpr_expr(A,intersection(Set1,Set2)),!,
   get_texpr_type(Set1,Set1Type), unify_types_strict(Set1Type,set(T)),
   ID = b(identifier('_zzzz_unary'),T,[generated]),
   ESet1 = b(member(ID,Set1),pred,[]),
   ESet2 = b(member(ID,Set2),pred,[]),
   create_exists_opt([ID],[ESet1,ESet2],NewPredicate),
   (debug_mode(off) -> true
      ;  print('Transformed not disjoint disequality: '),print_bexpr(NewPredicate),nl),
   get_texpr_expr(NewPredicate,NewP).
cleanup_post(equal(b(intersection(A,B),_,_),Empty),pred,I,not_equal(El1,El2),pred,[was(intersection)|I],multi/detect_not_equal) :-
    % {El1} /\ {El2} = {}  -->  El1 \= El2
    singleton_set_extension(A,El1),
    singleton_set_extension(B,El2),
    is_empty_set(Empty).
cleanup_post(greater(A,B),pred,I,Res,pred,[was(greater(A,B))|I],multi/eval_greater) :-
   get_integer(A,IA), get_integer(B,IB),
   (IA>IB -> Res = truth ; Res=falsity).
cleanup_post(less(A,B),pred,I,Res,pred,[was(less(A,B))|I],multi/eval_less) :-
   get_integer(A,IA), get_integer(B,IB),
   (IA<IB -> Res = truth ; Res=falsity).
cleanup_post(greater_equal(A,B),pred,I,Res,pred,[was(greater_equal(A,B))|I],multi/eval_greater_equal) :-
   get_integer(A,IA), get_integer(B,IB),
   (IA >= IB -> Res = truth ; Res=falsity).
cleanup_post(less_equal(A,B),pred,I,Res,pred,[was(less_equal(A,B))|I],multi/eval_less_equal) :-
   get_integer(A,IA), get_integer(B,IB),
   (IA =< IB -> Res = truth ; Res=falsity).
cleanup_post(negation(A),pred,I,falsity,pred,[was(negation(A))|I],multi/remove_negation_truth) :-
   is_truth(A),!. % ,print(negation(A)),nl.
cleanup_post(negation(A),pred,I,truth,pred,[was(negation(A))|I],multi/remove_negation_falsity) :-
   is_falsity(A),!. % ,print(negation(A)),nl.
cleanup_post(convert_bool(A),boolean,I,boolean_true,boolean,I,multi/remove_convert_bool) :-
   is_truth(A),!. %  ,print(convert_bool(A)),nl.
cleanup_post(convert_bool(A),boolean,I,boolean_false,boolean,I,multi/remove_convert_bool_false) :-
   is_falsity(A),!. %  ,print(convert_bool_false(A)),nl.
cleanup_post(assertion_expression(Cond,_ErrMsg,Expr),_T,I0,BE,TE,IE,multi/remove_assertion_expression) :-
   is_truth(Cond),!,
   Expr = b(BE,TE,I1),
   add_important_info_from_super_expression(I0,I1,IE).
cleanup_post(cartesian_product(A,B),T,I,Res,T,I,single/cartesian_product_to_pair) :-
    singleton_set_extension(A,El1),
    singleton_set_extension(B,El2),  % {A}*{B} -> {A|->B}  ; happens in Alloy translations a lot
    !,
    get_texpr_type(El1,T1), get_texpr_type(El2,T2),
    safe_create_texpr(couple(El1,El2),couple(T1,T2),Pair),
    Res = set_extension([Pair]).
cleanup_post(union(A,B),T,I,Res,T,[add_element_to_set|I],multi/add_element_to_set) :- % multi: cycle check done in info field
    \+ member(add_element_to_set,I),
    (   singleton_set_extension(B,_El) -> Res = union(A,B)
      ; singleton_set_extension(A,_El) -> Res = union(B,A)),
    !. %,print(detected_add_singleton_element(_El)),nl.
cleanup_post(general_union(SetExt),Type,I0,Res,Type,Info,multi/general_union_set_extension) :- % union_generalized
    % union({a,b,c,...}) = a \/ b \/ c ...
    SetExt = b(set_extension(LIST),_,I1),
    add_important_info_from_super_expression(I0,I1,Info),
    % no need to apply rule if already transformed into avl in cleanup_pre,  hence we do not call is_set_extension
    construct_union_from_list(LIST,Type,Info,TRes),
    !,
    (debug_mode(on) -> print('translated_general_union: '), print_bexpr(TRes),nl ; true),
    get_texpr_expr(TRes,Res).
cleanup_post(general_intersection(SetExt),Type,I0,Res,Type,Info,multi/general_inter_set_extension) :- % inter_generalized
    % inter({a,b,c,...}) = a /\ b /\ c ...
    SetExt = b(set_extension(LIST),_,I1),
    add_important_info_from_super_expression(I0,I1,Info),
    % no need to apply rule if already transformed into avl in cleanup_pre,  hence we do not call is_set_extension
    construct_inter_from_list(LIST,Type,Info,TRes),
    !,
    (debug_mode(on) -> print('translated_general_intersection: '), print_bexpr(TRes),nl ; true),
    get_texpr_expr(TRes,Res).
cleanup_post(SUB,pred,I0,FORALL,pred,FInfo,multi/general_union_subset) :-
    is_subset(SUB,UNION,T),
    % union(S) <: T   ===>   !x.(x:S => x <: T)
    % currently: subsets of T may be generated, but it does not propagate well to S
    UNION = b(general_union(S),_,_),
    !,
    get_unique_id_inside('_zzzz_unary',S,T,ID),
    get_texpr_type(S,SType), TID = b(identifier(ID),SType,[generated]),
    safe_create_texpr(member(TID,S),pred,LHS),
    safe_create_texpr(subset(TID,T),pred,RHS),
    create_implication(LHS,RHS,NewForallBody),
    create_forall([TID],NewForallBody,TFORALL),
    TFORALL = b(FORALL,pred,I1),
    add_important_info_from_super_expression(I0,I1,FInfo),
    % see test 1854, and ProZ ROZ/model.tex
    (debug_mode(on) -> print('translated_general_union subset: '), print_bexpr(TFORALL),nl ; true).
cleanup_post(size(Seq),integer,Info,Res,integer,Info,multi/size_append) :-
    get_texpr_expr(Seq,concat(A,B)),
    % size(A^B) = size(A)+size(B)   useful e.g. for test 1306
    !,
    Res = add(b(size(A),integer,Info),b(size(B),integer,Info)).
cleanup_post(concat(A,B),Type,I0,Seq,Type,[was(concat)|NewInfo],multi/concat_empty) :-
    ( is_empty_set(A) -> b(Seq,_,I1)=B
    ; is_empty_set(B) -> b(Seq,_,I1)=A
    ),!,
    add_important_info_from_super_expression(I0,I1,NewInfo).
cleanup_post(concat(A,B),Type,I0,Seq,Type,[was(concat)|I0],multi/concat_singleton_seq) :-
    ( is_singleton_sequence(B,Element) -> Seq = insert_tail(A,Element)
    ; is_singleton_sequence(A,Element) -> Seq = insert_front(Element,B)
    ),!,
    debug_format(19,'Concat with singleton sequence detected~n',[]).
cleanup_post(E,integer,I,Res,integer,[was(Operator)|I],multi/constant_expression) :-
    pre_compute_static_int_expression(E,Result),!,
    functor(E,Operator,_),
    % format('Precomputed: ~w  for ',[Result]), translate:print_bexpr(b(E,integer,[])),nl,
    Res = integer(Result).
cleanup_post(min(Interval),integer,I0,Res,integer,Info,multi/eval_min_interval) :-
    is_interval_or_singleton(Interval,Low,Up),
    get_integer(Low,L), number(L),
    get_integer(Up,U), number(U), L =< U, % non-empty interval
    debug_println(5,simplified_min_interval(L,U,L)),
    Res = integer(L), get_texpr_info(Low,I1),
    add_important_info_from_super_expression(I0,I1,Info).
cleanup_post(max(Interval),integer,I0,Res,integer,Info,multi/eval_max_interval) :-
    is_interval_or_singleton(Interval,Low,Up),
    get_integer(Low,L), number(L),
    get_integer(Up,U), number(U), L =< U, % non-empty interval
    debug_println(5,simplified_max_interval(L,U,U)),
    Res = integer(U), get_texpr_info(Low,I1),
    add_important_info_from_super_expression(I0,I1,Info).
cleanup_post(first(Seq),Type,I0,Res,Type,Info,multi/first_seq_extension) :-
    is_sequence_extension(Seq,List), List = [First|Rest],
    (Rest == [] -> true ; preferences:get_preference(disprover_mode,true)), % we may remove WD issue otherwise (TO DO: check if Rest contains any problematic elements)
    !,
    First = b(Res,Type,I1),
    add_important_info_from_super_expression(I0,I1,Info).
cleanup_post(last(Seq),Type,I0,Res,Type,Info,multi/first_seq_extension) :-
    is_sequence_extension(Seq,List), List = [First|Rest],
    (Rest == [] -> true ; preferences:get_preference(disprover_mode,true)), % we may remove WD issue otherwise (TO DO: check if list contains any problematic elements)
    !,
    last([First|Rest],b(Res,Type,I1)),
    add_important_info_from_super_expression(I0,I1,Info).
cleanup_post(function(Override,X),Type,I0,Res,Type,Info,multi/function_override) :-
    preferences:get_preference(disprover_mode,true), % only applied in Disprover mode as it can remove WD problem; TO DO make it also applicable in normal mode
    get_texpr_expr(Override,overwrite(_F,SEXt)),
    SEXt = b(set_extension(LIST),_,_),
    member(b(couple(From,To),_,_),LIST),
    same_texpr(From,X),
    % ( F <+ { ... X|->To ...}) (X)   ==>  To
    %print_bexpr(b(function(Override,X),Type,_I)), print(' ==> '), print_bexpr(To),nl,
    !,
    get_texpr_expr(To,Res),
    get_texpr_info(To,I1),
    add_important_info_from_super_expression(I0,I1,Info).
cleanup_post(function(Override,X),Type,Info,Res,Type,Info,multi/function_override_ifte) :-
    preferences:get_preference(disprover_mode,true), % only applied in Disprover mode as it can remove WD
    % TO DO: should we do this generally? or simply deal with overwrite symbolically always?
    get_texpr_expr(Override,overwrite(F,SEXt)),
    SEXt = b(set_extension([Couple]),_,_),
    Couple = b(couple(From,To),_,_),
    % f <+ {A|->B}(x)  ->  if_then_else(x=A,B,f(x))   ;  avoids having to explicitly compute f<+{A|->B}
    Res = if_then_else(EqXFrom,To,FX),
    safe_create_texpr(equal(X,From),pred,EqXFrom),
    safe_create_texpr(function(F,X),Type,FX). % , print_bexpr(b(Res,Type,Info)),nl.
cleanup_post(function(SEXt,ARG),Type,I0,Res,Type,Info,multi/function_set_extension) :-
    SEXt = b(set_extension(LIST),_,ListInfos),  % TO DO: also support b(value(avl_set(A)),_,_)
    eval_set_extension_element(ARG,Value),
    select(b(couple(LHS,RHS),_,_),LIST,REST),
    (member(contains_wd_condition,ListInfos)  % then we could remove WD problem, e.g., r = {1|->2, 2|-> 1/0}(1)
      -> (preferences:preference(find_abort_values,false) ;
          preferences:get_preference(disprover_mode,true))
      ; true),
    eval_set_extension_element(LHS,Value), %nl,print(found(RHS)),nl,
    % WE NEED TO Check that all LHS can be compared against ARG
    \+ member((b(couple(LHS2,_),_,_),REST),eval_set_extension_element(LHS2,Value)), % no other potential match
    \+ member((b(couple(LHS3,_),_,_),LIST), \+ eval_set_extension_element(LHS3,_)), % all left-hand-sides can be evaluated
    !, get_texpr_expr(RHS,Res), get_texpr_info(RHS,I1),
    add_important_info_from_super_expression(I0,I1,Info).
% Detect if_then_else in format as printed by pp_expr2(if_then_else( ....)...) or as generated by B2TLA:
cleanup_post(function(IFT,DUMMYARG),Type,Info,if_then_else(IFPRED,THEN,ELSE),Type,Info,multi/function_if_then_else) :-
   is_if_then_else(IFT,post,DUMMYARG,IFPRED,THEN,ELSE),
   (debug_mode(off) -> true
    ; print('% Recognised if-then-else expression: IF '), print_bexpr(IFPRED),
      print('  THEN '),print_bexpr(THEN), print(' ELSE '),print_bexpr(ELSE),nl
   ).
cleanup_post(function(Composition,X),Type,Info,NewExpr,Type,Info,multi/function_composition) :-
   (data_validation_mode ;
    get_preference(convert_comprehension_sets_into_closures,true) % there it can make a big difference in particular since relational composition (rel_composition_wf -> rel_compose_with_inf_fun case) was not fully symbolic; see Systerel data validation examples
   ),
   % (F;G)(X)  -->  G(F(X))  (F;G;H)(X)  -->  H(G(F(X))) ...
   peel_rel_composition(Composition,Type,Info,X,Result,Level),
   Level>0,
   Result = b(NewExpr,_,_),
   check_ast(Result),
   (debug_mode(off) -> true
     ; format('Function application of COMPOSITION (Nesting: ~w) translated to: ',[Level]),print_bexpr(Result),nl).
cleanup_post(range(SETC),Type,I, comprehension_set(RangeIds2,NewCompPred),Type,I,single/range_setcompr) :-
   % translate ran({x1,...xn|P})  into {xn| #(x1,...).(P)} ; particularly interesting if x1... contains large datavalues (e.g., C_02_001.mch from test 1131)
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,RangeIds),  % print(range(CompIds,DomainIds,RangeIds)),nl,
    %\+((member(ID,RangeIds),get_texpr_id(ID,'_lambda_result_'))), % for test 612; maybe disable optimisation if memory consumption of variables small
    !, % TO DO: also detect patterns such as dom(dom( or ran(ran( ... [Done ??]
    rename_lambda_result_id(RangeIds,CompPred,RangeIds2,CompPred1),
    rename_lambda_result_id(DomainIds,CompPred1,DomainIds2,CompPred2),
    create_outer_exists_for_dom_range(DomainIds2,CompPred2,NewCompPred), % will mark the exists; so that during expansion we will treat it differently for enumeration
    (debug_mode(off) -> true
      ;  print('Encode range as existential quantification: '), print_bexpr(NewCompPred),nl).
cleanup_post(domain(SETC),Type,I, comprehension_set(DomainIds2,NewCompPred),Type,I,single/domain_setcompr) :-
   % translate dom({x1,...xn|P})  into {x1,..| #(xn).(P)} ; particularly interesting if xn contains large datavalues
   % used to fail test 306 ; fixed by allow_to_lift_exists annotation
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,RangeIds), % print(domain(CompIds,DomainIds,RangeIds)),nl,
    % \+ (member(ID,CompIds),get_texpr_id(ID,'_lambda_result_')),
      % WE HAVE TO BE CAREFUL if xn = LAMBDA_RESULT ; TO DO rename like above for range
      % example from test 292: rel(fnc({x,y|x:1..10 & y:1..x})) = {x,y|x:1..10 & y:1..x}
    !,
    % TO DO: detect when closure is lambda; e.g., for e.g. dom(pred) = INTEGER in test 292 : split(CompIds,Args,Types), closures:is_lambda_value_domain_closure(Args,Types,B,  DomainValue, _),;  currently create_exists_opt deals with most of this
    rename_lambda_result_id(DomainIds,CompPred,DomainIds2,CompPred1),
    rename_lambda_result_id(RangeIds,CompPred1,RangeIds2,CompPred2),
    create_outer_exists_for_dom_range(RangeIds2,CompPred2,NewCompPred),
    (debug_mode(off) -> true
      ;  print('Encode domain as existential quantification: '),print_bexpr(NewCompPred),nl).
cleanup_post(domain(SETC),Type,I, comprehension_set(DomainIds,RestPred),Type,I,single/domain_setcompr) :-
   % translate dom({x1,...xn|P & xn=E})  into {x1,..| P} ; particularly interesting if xn contains large datavalues
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,RangeIds), % print(domain(CompIds,DomainIds,RangeIds)),nl,
    \+ (member(ID,CompIds),get_texpr_id(ID,'_lambda_result_')),
    conjunction_to_list(CompPred,Preds),
    RangeIds = [TId],
    get_sorted_ids(RangeIds,Blacklist),
    select_equality(TId,Preds,Blacklist,_Eq,_Expr,RestPreds,_,check_well_definedness), % We could do check_well_definedness only if preference set
    conjunct_predicates(RestPreds,RestPred),
    not_occurs_in_predicate(Blacklist,RestPred),
    !,
    % TO DO: use create_optimized exists; also treat inner existential quantification and merge
    (debug_mode(off) -> true
      ;  print('Encode domain of lambda abstraction: '),print_bexpr(RestPred),nl).
cleanup_post(comprehension_set([TID],Pred),Type,I,
             struct(b(rec(NewFieldSets),record(FieldTypes),I)),
             Type,I,single/simplify_record) :-
    % {r|r'a : 1..1000 & r'b : 1..100 & r:struct(a:INTEGER,b:NAT,c:BOOL)}  -->  struct(a:1..1000,b:1..100,c:BOOL)
    % these kind of set comprehensions are generated by ProZ, see ROZ example model.tex
    % TO DO: maybe generalise this optimisation: currently it only works if all predicates can be assimilated into struct expression
    TID = b(identifier(ID),record(FieldTypes),_),
    conjunction_to_list(Pred,PL),
    l_update_record_field_membership(PL,ID,[],FieldSetsOut),
    maplist(construct_field_sets(FieldSetsOut),FieldTypes,NewFieldSets),
    (debug_mode(off) -> true
     ; print('Detected Record set comprehension: '),
       print_bexpr(b(struct(b(rec(NewFieldSets),record(FieldTypes),I)),Type,I)),nl).
cleanup_post(precondition(TP,TS),subst,I0,S,subst,Info,multi/remove_triv_precondition) :-
    % remove trivial preconditions
    get_texpr_expr(TP,truth),!,
    get_texpr_expr(TS,S),get_texpr_info(TS,I1),
    add_important_info_from_super_expression(I0,I1,Info).
cleanup_post(external_function_call('ENUM',[TA]),T,I,A,T,[prob_annotation('ENUM')|I],single/process_ENUM) :-
   format(' ENUM (~w): ',[T]),translate:print_bexpr(TA),nl,
   % TO DO: do not process in DEFINITION of ENUM in LibraryProB.def
   get_texpr_expr(TA,A).
cleanup_post(comprehension_set(Ids1,E),T,I,comprehension_set(Ids2,E2),T,I,single/detect_lambda) :-
    preferences:get_preference(detect_lambdas,true),
    % used to create *** Enumerating lambda result warnings for test 1162, not anymore
    % does not yet detect: f = {x,y|x:NATURAL & (x> 1 &y=x+2)} & res = f[3..4]
    E = b(conjunct(LHS,Equality),pred,Info), nonmember(prob_annotation('LAMBDA'),Info), % not already processed
    identifier_equality(Equality,ID,TID1,Expr1),
    last(Ids1,TID), get_texpr_id(TID,ID),
    not_occurs_in_expr(ID,Expr1),
    not_occurs_in_predicate([ID],LHS),
    !,
    get_texpr_info(Equality,EqInfo),
    get_unique_id_inside('_lambda_result_',LHS,Expr1,ResultId), % currently the info field lambda_result is not enough: several parts of the ProB kernel match on the identifier name '_lambda_result_'
    TID1 = b(identifier(_),Type1,Info1),
    add_texpr_infos(b(identifier(ResultId),Type1,Info1),[lambda_result(ResultId),lambda_result_id_was(ID)],TID2),
    Equality2 = b(equal(TID2,Expr1),pred,[prob_annotation('LAMBDA-EQUALITY')|EqInfo]),
    E2 = b(conjunct(LHS,Equality2),pred,[prob_annotation('LAMBDA')|Info]),
    append(Ids0,[_],Ids1),
    append(Ids0,[TID2],Ids2),
% code exists which is simpler, but has disadvantage of losing position info for equality
    (debug_mode(off) -> true ; format('Lambda using ~w detected: ',[ID]),print_bexpr(E2),nl).
cleanup_post(comprehension_set(TIds,Body),T,I1,comprehension_set(TIds,Body2),T,I1,single/detect_lambda_result_auto) :-
     get_texpr_info(Body,BI),
     (TIds=[_,_,_,_|_] % check if at least four ids (three could also be useful)
       -> true % for example useful here: {x,y,v| x:INTEGER & y:INTEGER & #z.(z=x+1 & x = y*y & v=z*z & y:1..10)} it could be useful to detect x as DO_NOT_ENUMERATE; currently we only enable this analysis for four ids at least
       ; nonmember(prob_annotation('LAMBDA'),BI) % it is a lambda with one argument, no use to do analysis
      ),
     nonmember(prob_annotation('DO_NOT_ENUMERATE'(_)),BI), % analysis was already performed or manually annotated
     find_do_not_enumerate_variables(TIds,Body,SortedVs),
     (SortedVs=[] -> true ; debug_mode(off) -> true
      ; add_message(detect_lambda_result_auto,'Annotating comprehension set identifiers with DO_NOT_ENUMERATE: ',SortedVs,I1)
     ),
     findall(prob_annotation('DO_NOT_ENUMERATE'(DNID)),member(DNID,SortedVs),NewInfos),
     add_texpr_infos(Body,NewInfos,Body2).
cleanup_post(comprehension_set(Ids1,E1),T,I,comprehension_set(Ids2,E2),T,I,multi/detect_lambda_result_user_ann) :-
     get_texpr_expr(DNE,external_pred_call('DO_NOT_ENUMERATE',[TID])),
     get_texpr_id(TID,ID),
     member_in_conjunction(DNE,E1),
     NewInfo = prob_annotation('DO_NOT_ENUMERATE'(ID)), % similar to lambda_result
     E1 = b(PRED,pred,I1),
     nonmember(NewInfo,I1),
     get_texpr_id(TID,ID),
     nth1(Pos,Ids1,TID1,Rest),
     add_texpr_infos(TID1,[NewInfo],TID2),
     nth1(Pos,Ids2,TID2,Rest),
     add_message(detect_lambda_result,'Annotating comprehension set identifier with DO_NOT_ENUMERATE: ',ID,I1),
     E2 = b(PRED,pred,[NewInfo|I1]).
cleanup_post(record_field(b(rec(Fields),TR,IR),Field),T,I,FieldVal,T,I, single/remove_field_access) :-
    always_well_defined_or_disprover_mode(b(rec(Fields),TR,IR)),
    member(field(Field,TFieldVal),Fields),!,
    (debug_mode(off) -> true ; add_message(remove_field_access,'Remove static field access: ',Field,I)), % cf test 1294
    get_texpr_expr(TFieldVal,FieldVal).
cleanup_post(sequence([S1,b(sequence(S2),subst,_)]),subst,I,sequence([S1|S2]),subst,I, single/flatten_sequence2) :-
   debug_println(9,flatten_sequence2). % do we need something for longer sequences?
cleanup_post(sequence([b(sequence(S1),subst,_)|S2]),subst,I,sequence(NewSeq),subst,I, single/flatten_sequence1) :-
   append(S1,S2,NewSeq),
   % avoid maybe calling filter_useless_subst_in_sequence again
   debug_println(9,flatten_sequence1).
cleanup_post(sequence(S1),subst,I,sequence(S2),subst,I, single/remove_useless_subst_in_seuence) :-
   get_preference(useless_code_elimination,true),
   filter_useless_subst_in_sequence(S1,Change,S2), debug_println(filter_sequence(9,Change)).
cleanup_post(sequence(Statements),subst,I,Result,subst,I, single/sequence_to_multi_assign) :-
   % merge sequence of assignments if possible
   merge_assignments(Statements,Merge,New), Merge==merged,
   construct_sequence(New,Result).
   % nl,print('Merged: '),translate:print_subst(b(Result,subst,I)),nl,nl.
cleanup_post(parallel(Statements),subst,I,Result,subst,I, single/parallel_to_multi_assign) :-
 % this merges multiple assignments into a single one: advantage: only one waitflag set up
 % should probably not be done in INITIALISATION
    % print(parallel(Statements)),nl,trace,
    extract_assignments(Statements,LHS,RHS,Rest,Nr), % print(extracted(Nr,LHS)),nl,
    Nr>1,!,
    (debug_mode(on) ->
       print('Parallel to Assignment: '), translate:print_subst(b(parallel(Statements),subst,I)),nl
      ; true),
    (Rest == []  ->   Result  =  assign(LHS,RHS)
     ;  Result = parallel([b(assign(LHS,RHS),subst,[])|Rest])).
    %translate:print_subst(b(Result,subst,I)),nl.
cleanup_post(select([CHOICE|Rest]),subst,I0,S,subst,Info,single/remove_select) :-
    Rest = [], % SELECT can have multiple true branches
    CHOICE=b(select_when(TRUTH,Subst),subst,_),
    is_truth(TRUTH),!,
    debug_println(19,'Removing useless SELECT'),
    get_texpr_expr(Subst,S),get_texpr_info(Subst,I1),
    add_hint_message(remove_select,'Removing useless SELECT','',I1),
    add_important_info_from_super_expression(I0,I1,Info).
cleanup_post(select([CHOICE|Rest],_ELSE),subst,OldInfo,Res,subst,I,single/remove_select_else) :-
    CHOICE=b(select_when(TRUTH,Subst),subst,_),
    is_truth(TRUTH),!,
    (Rest = [] % completely useless SELECT
      -> add_hint_message(remove_select_else,'Removing useless SELECT','',OldInfo),
         get_texpr_expr(Subst,Res),get_texpr_info(Subst,I1),
         add_important_info_from_super_expression(OldInfo,I1,I)
      ;  add_hint_message(remove_select_else,'Removing useless SELECT ELSE branch','',OldInfo),
         Res = select([CHOICE|Rest]), I=OldInfo).
cleanup_post(let_expression([],[],TExpr),Type,I0,Expr,Type,I,multi/remove_let_expression) :- !,
    % remove trivial let expressions without any introduced identifiers
    % this rule makes only sense in combination with the next rule which removes
    % simple let identifiers
    get_texpr_expr(TExpr,Expr),
    get_texpr_info(TExpr,I1),
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(let([],Pred,TExpr),Type,I0,Expr,Type,I,multi/remove_let) :- is_truth(Pred),!,
    % remove trivial let expressions without any introduced identifiers
    % this rule makes only sense in combination with the next rule which removes
    % simple let identifiers
    get_texpr_expr(TExpr,Expr),
    get_texpr_info(TExpr,I1),
    add_important_info_from_super_expression(I0,I1,I).
cleanup_post(let_expression(Ids,Exprs,Expr),Type,I,
             let_expression(NIds,NExprs,NExpr),Type,I,multi/remove_let_expression2) :-
    simplify_let(Ids,Exprs,Expr,NIds,NExprs,NExpr),!.
cleanup_post(let_predicate(TIds,Exprs,Body),Type,I,
             let_predicate(NIds,NExprs,NBody),Type,I,multi/remove_let_predicate2) :-
    simplify_let(TIds,Exprs,Body,NIds,NExprs,NBody),!,
    %format('~n Simplified Let ~w --> ~w~n',[TIds,NIds]),
    true.
cleanup_post(let_predicate(TIds,Exprs,Body),Type,I,
             let_predicate(TIds,Exprs,Body),Type,[useless_let|I],single/useless_let_message) :-
    is_truth(Body),
    nonmember(useless_let,I),
    get_texpr_ids(TIds,Ids),
    TIds= [TID1|_], get_texpr_info(TID1,BInfo),
    nonmember(generated_exists,BInfo), % otherwise this was generated programmatically, e.g., in get_operation_enabling_condition
    add_message(b_ast_cleanup,'Useless existentially quantified variable: ',Ids,TIds).
% was disabled because simplify_let_subst also replaced in RHS of assignments
% but now we check that a LET/ANY variable cannot be assigned to statically
% so it should be safe now to replace simple equalities:
cleanup_post(let(Ids,Pred,Subst),Type,I,
             let(NIds,NPred,NSubst),Type,I,multi/remove_let_subst2) :-
    simplify_let_subst(Ids,Pred,Subst,NIds,NPred,NSubst),! ,
    true. % translate:print_subst(b(let(NIds,NPred,NSubst),subst,[])),nl.
/* this is not used currently; as the generic code's performance has been improved
   and better propagation is ensured by b_compute_arith_expression which (partially) instantiates its argument (known to be an integer)
cleanup_post(ArithPredicate,pred,Info,ArithPredicate,pred,NewInfo,single/add_arith_pred_info) :-
    comparison(ArithPredicate,LHS,RHS,_,_,_),
    \+ clpfd_arith_integer_expression(LHS),
    \+ clpfd_arith_integer_expression(RHS),
    !,
    % store the information that there is no use in doing b_compute_arith_expression special treatment
    NewInfo = [no_clpfd_arith_integer_expression|Info].
*/

cleanup_post(forall([ID],LHS,RHS),pred,IOld,Res,pred,IOld,single/expand_forall_set_extension) :-
 % expand !x.(x:{a,b,...} => RHS) into conjunction
  % can be useful e.g. for KODKOD when we pick an element from a set of sets
   preferences:get_preference(try_kodkod_on_load,true), % at the moment only enable in Kodkod mode
   % TO DO: enable always; but maybe check that set_extension can be computed (which eval_set_extension will do) to avoid duplicating checks (!y.(y:{v,w} => expensive_pred) + what if v=w
   nonmember(do_not_optimize_away,IOld),
   get_texpr_expr(LHS,member(ID2,Set)),
   same_texpr(ID,ID2),
   is_set_extension(Set,SList),
   get_texpr_id(ID,AID),
   debug_format(19,'Expanding forall ~w ',[AID]),
   findall(C, (member(SEL,SList),replace_id_by_expr(RHS,AID,SEL,C)),Conjuncts),
   conjunct_predicates_with_pos_info(Conjuncts,ExpandedForAll),
   (silent_mode(on) -> true ;  print_bexpr(ExpandedForAll),nl),
   get_texpr_expr(ExpandedForAll,Res).
cleanup_post(forall(Ids1,LHS,RHS),pred,IOld,
             forall(Ids,NewLHS,NewRHS),pred,INew,multi/merge_forall) :-
    is_truth(LHS),
    RHS = b(forall(Ids2,NewLHS,NewRHS),pred,_),
    %((member(b(identifier(_),Type,_),Ids1), is_infinite_type(Type)) -> true), % could be useful for tests 1441, 1447 ??
    (disjoint_ids(Ids1,Ids2)
     -> append(Ids1,Ids2,Ids),
        % !x.(truth => !y.(P=>Q)  <==>  !(x,y).(P=>Q)
        (debug_mode(off) -> true ; format('Merging forall ~w: ',[Ids]), print_bexpr(NewLHS),nl),
        add_removed_typing_info(IOld,INew)
      ; \+ preferences:get_preference(disprover_mode,true),
        translate:translate_bexpression(b(forall(Ids1,LHS,RHS),pred,IOld),PS),
        add_warning(b_ast_cleanup,'Variable clash in nested universal quantification: ',PS,IOld),
        fail
    ).
cleanup_post(forall(Ids,LHS,RHS),pred,IOld,
             implication(Outer,FORALL),pred,IOld,single/detect_global_preds_forall1) :-
    % !x.(P(x) & Q => R(x)  <==>  Q =>  !x.(P(x) => R(x))
    % TO DO: maybe we should not lift things like printf, ... ?
    bsyntaxtree:detect_global_predicates(Ids,LHS,Outer,Inner),
    (debug_mode(off) -> true ; format('Lifting predicate (lhs) of forall ~w: ',[Ids]), print_bexpr(Outer),nl),
    construct_inner_forall(Ids,Inner,RHS,IOld,FORALL).
% TO DO: implement similar lifting rules for exists(Ids,P)
cleanup_post(forall(Ids,LHS,RHS),pred,IOld,
             conjunct(Outer,FORALL),pred,IOld,single/detect_global_preds_forall2) :-
    is_truth(LHS),
    % !x.(truth => Q & R(x)  <==>  Q & !x.(truth => R(x))
    bsyntaxtree:detect_global_predicates(Ids,RHS,Outer,Inner),
    (debug_mode(off) -> true ; format('Lifting predicate (rhs &) of forall ~w: ',[Ids]), print_bexpr(Outer),nl),
    construct_inner_forall(Ids,LHS,Inner,IOld, FORALL). % print_bexpr(FORALL),nl,nl.
cleanup_post(forall(Ids,LHS,RHS),pred,IOld,
             implication(Outer,FORALL),pred,IOld,single/detect_global_preds_forall3) :-
    is_truth(LHS),
    RHS = b(implication(RHS1,RHS2),pred,_),
    % !x.(truth => (Q & R(x) => S(x))  <==>  Q => !x.(truth => R(x) => S(x))
    bsyntaxtree:detect_global_predicates(Ids,RHS1,Outer,Inner),
    create_implication(Inner,RHS2,NewRHS),
    (debug_mode(off) -> true ; format('Lifting predicate (rhs =>) of forall ~w: ',[Ids]), print_bexpr(Outer),nl),
    construct_inner_forall(Ids,LHS,NewRHS,IOld, FORALL).
cleanup_post(forall([TID1,TID2|OTHER],LHS,RHS),pred,IOld,
             forall([TID1,TID2|OTHER],NewLHS,RHS),pred,[prob_symmetry(ID1,ID2)|IOld],single/symmetry_detection) :-
    % DETECT Symmetries such as !(x,y).(x /= y => x=TRUE or y=TRUE)
    % !(x2,y).(x2 /= y & x2:s & y:s => x2=aa or y=aa)
    % f:1..n --> 1..(n-1) & !(x,y).(x/=y &x:dom(f) & y:1..n => f(x) /= f(y)) & n=9 (runtime 1 sec -> 0.78 sec)
   get_texpr_type(TID1,T), get_texpr_type(TID2,T),
   preferences:get_preference(try_kodkod_on_load,false), % Kodkod cannot properly deal with LEQ_SYM_BREAK, treats it like truth
   preferences:get_preference(use_static_symmetry_detection,true),
   sym_break_supported_type(T), % LEQ_SYM_BREAK not yet fully functional with SET types; TO DO: fix
   get_texpr_id(TID1,ID1), get_texpr_id(TID2,ID2),
   nonmember(prob_symmetry(ID1,ID2),IOld),
   rename_bt(LHS,[rename(ID1,ID2),rename(ID2,ID1)],LHS2),
   same_norm_texpr(LHS,LHS2),
   rename_bt(RHS,[rename(ID1,ID2),rename(ID2,ID1)],RHS2),
   same_norm_texpr(RHS,RHS2),
   construct_sym_break(T,TID1,TID2,LHS,SYMBREAK),
   conjunct_predicates_with_pos_info(LHS,SYMBREAK,NewLHS),
   (debug_mode(off) -> true
    ; format('SYMMETRY BREAKING FORALL: !(~w,~w).(',[ID1,ID2]),print_bexpr(NewLHS),
      print(' => '), print_bexpr(RHS),print(')'),nl).
cleanup_post(forall(AllIds,P,Rhs),pred,I,NewPred,pred,INew,multi/forall_splitting) :-
    AllIds = [TID1|TRestIDs], TRestIDs = [_|_], % TO DO: maybe not require that TID1 is the first id
    get_preference(use_clpfd_solver,true), % with CLPFD false: maybe more likely to reduce performance
    % NOTE: we could destroy symmetry reduction detection if TRestIDs = [TID2], but we run after symmetry detection
     % !(x,y,...).(x:SET & RestPred => Rhs)  ==   !x.(x:SET => !(y,..).(RestPred => Rhs))
    conjunction_to_list(P,[MEM|RestPreds]),
    is_membership(MEM,TID,Set,_),
    same_id(TID1,TID,ID),
    \+ known_set(Set),  % rewriting makes sense if Set is not fully known and will be instantiated during solving
    % prevent !(x,y).(x:NATURAL & x<10 & y :1..x =>  x+y<20), even though it seems to work nonetheless
    get_sorted_ids(TRestIDs,RestIDs),
    not_occurs_in_predicate(RestIDs,Set),
    NewPred = forall([TID1],MEM,InnerForall),
    conjunct_predicates_with_pos_info(RestPreds,InnerForallLhs),
    construct_inner_forall(TRestIDs,InnerForallLhs,Rhs,I,InnerForall),
    !,
    delete(I,used_ids(_),I1), % should in principle still be ok at outer level; but just be sure
    add_removed_typing_info(I1,INew),
    (debug_mode(on) -> format('FORALL SPLITTING ~w (from ~w) for better propagation: ',[ID,RestIDs]),
                       print_bexpr(b(NewPred,pred,INew)),nl
           ; true).
cleanup_post(exists([TID],MemPred),pred,IOld,
             Res,pred,IOld, single/replace_exists_by_not_empty) :-
  % simplify #ID.(ID:E)  <=> E /= {}
  % simplify #ID.(ID:E1 & ID:E2)  <=> E1 /\ E2 /= {} , etc...
  % important e.g. for y:20..30000000000 & not(#x.(x:1..10 & x:8..y))
  is_valid_id_member_check(MemPred,TID,E),
  !,
  (definitely_not_empty_set(E) -> Res= truth
   ; get_texpr_type(E,Type), EmptySet=b(empty_set,Type,[]),
     Res = not_equal(E,EmptySet)),
  (debug_mode(off) -> true
   ; get_texpr_id(TID,ID),
     format('Removing existential quantifier: ~w~n',[ID]),
     print_bexpr(b(Res,pred,IOld)),nl).
cleanup_post(exists([TID],b(NotMemPred,_,_)),pred,IOld,
             not_equal(E,TypeExpr),pred,IOld, single/replace_exists_by_not_full) :-
  % simplify #ID.(ID/:E)  <=> E /= FullType
  % important e.g. for y:7..30000000000 & not(#x.(x /: 1..y))
    is_not_member(NotMemPred,MID,E),
    same_id(TID,MID,SID),
   % + check that MID does not occur in E
   \+ occurs_in_expr(SID,E),
   get_texpr_type(E,SType),
   bsyntaxtree:is_set_type(SType,Type),
   create_maximal_type_set(Type,TypeExpr), % Note: no longer introduces identifiers but value(.) results
   !,
   (debug_mode(off) -> true
    ; format('Removing existential quantifier: ~w~n',[SID]),
      print_bexpr(b(not_equal(E,TypeExpr),pred,IOld)),nl).
cleanup_post(exists([TID],b(Pred,_,_)),pred,IOld,
             truth,pred,IOld, single/replace_exists_by_truth) :-
  b_interpreter_check:arithmetic_op(Pred,_Op,X,Y),
  ( (same_id(TID,X,SID), \+ occurs_in_expr(SID,Y), always_well_defined_or_disprover_mode(Y)) ;
    (same_id(TID,Y,SID), \+ occurs_in_expr(SID,X), always_well_defined_or_disprover_mode(X))
  ),
  !,  % we have a formula of the form #SID.(SID > Expr); provided Expr is well-defined, this is always true
   (debug_mode(off) -> true
    ; format('Removing existential quantifier: ~w~n',[SID]),
      print_bexpr(b(Pred,pred,[])),nl).
cleanup_post(exists([TID1,TID2|OTHER],RHS),pred,IOld,
             exists([TID1,TID2|OTHER],NewRHS),pred,[prob_symmetry(ID1,ID2)|IOld],single/symmetry_detection) :-
    % DETECT Symmetries such as #(x,y).(x /= y & (x=TRUE or y=TRUE))
    % #(x2,y).(x2 /= y & x2:s & y:s & x2=aa or y=aa)
   get_texpr_type(TID1,T), get_texpr_type(TID2,T),
   preferences:get_preference(try_kodkod_on_load,false),
   preferences:get_preference(use_static_symmetry_detection,true),
   sym_break_supported_type(T), % LESS not yet fully functional with SET types; TO DO: fix
   get_texpr_id(TID1,ID1), get_texpr_id(TID2,ID2),
   nonmember(prob_symmetry(ID1,ID2),IOld),
   \+(contains_equality(TID1,TID2,RHS)), % IDs are already equal; no use in sym breaking
   debug_println(try_exists_symmetry(ID1,ID2)),
   rename_bt(RHS,[rename(ID1,ID2),rename(ID2,ID1)],RHS2),
   same_norm_texpr(RHS,RHS2),
   construct_sym_break(T,TID1,TID2,RHS,SYMBREAK),
   conjunct_predicates_with_pos_info(RHS,SYMBREAK,NewRHS),
   (debug_mode(off) -> true
    ; format('SYMMETRY BREAKING EXISTS: #(~w,~w).(',[ID1,ID2]),print_bexpr(NewRHS),print(')'),nl).
cleanup_post(Expr,T,I1,NewExpr,T,I1,single/detect_lambda_result_quant_auto) :-
     construct_for_find_do_not_enumerate(Expr,KIND,TIds,Body,NewExpr,Body2),
     get_texpr_info(Body,BI),
     nonmember(prob_annotation('DO_NOT_ENUMERATE'(_)),BI), % analysis was already performed or manually annotated
     find_do_not_enumerate_variables(TIds,Body,SortedVs),
     (SortedVs=[] -> true ; debug_mode(off) -> true
      ; ajoin(['Annotating ',KIND,' identifiers with DO_NOT_ENUMERATE: '],Msg),
        add_message(detect_lambda_result_auto,Msg,SortedVs,I1)
     ),
     findall(prob_annotation('DO_NOT_ENUMERATE'(DNID)),member(DNID,SortedVs),NewInfos),
     add_texpr_infos(Body,NewInfos,Body2).
% sort args of commutative operators by term size
cleanup_post(Expr, Type, I, NExpr, Type, NI, single/normalize_commutative_args) :-
    preferences:get_preference(normalize_ast, true),
    sort_commutative_args(Expr, I, NExpr, NI).
    %,(NExpr=Expr -> true ; print('COMMUTE: '), translate:print_bexpr(b(Expr,Type,I)),nl, print('     TO: '), translate:print_bexpr(b(NExpr,Type,NI)),nl).
%% COMMENT IN NEXT LINE TO CHECK validity of AST per NODE (helps find bugs)
%%cleanup_post(Expr,pred,I,Expr,pred,I,single/checked) :- check_ast(true,b(Expr,pred,I)),fail.
%%


cleanup_comprehension_set([TID],b(member(LHS,TSet),pred,_),Type,I,Set,I2,multi/remove_useless_comprehension_set) :-
     get_texpr_id(TID,ID),
     get_texpr_id(LHS,ID),
     % {ID|ID:Set}  ==>  Set
     not_occurs_in_expr(ID,TSet),
     TSet=b(Set,Type,I2),
     !,
     add_hint_message(remove_useless_comprehension_set,'Removing useless comprehension set over: ',ID,I).
cleanup_comprehension_set([TID],Body,_Type,I,SeqTypeSet,I,single/detect_seq_comprehension_set) :-
     get_texpr_id(TID,ID),
     % {ID | #LenID. (ID : 1..LenID --> Set)}  ==>  seq(Set)
     % {ID | #LenID. (ID : 1..LenID >-> Set)}  ==>  iseq(Set)
     get_texpr_expr(Body,exists([TLen],TIBody)),
     get_texpr_id(TLen,LenID), LenID \= ID,
     get_texpr_expr(TIBody,IBody),
     get_member(IBody,LenID,ISet,TID2,FUNCTION),
     get_texpr_id(TID2,ID),
     get_seq_fun_aux(FUNCTION,ISet,Interval,SeqTypeSet),
     get_texpr_expr(Interval,interval(ONE,TLen2)),
     get_texpr_id(TLen2,LenID),
     get_integer(ONE,1),
     !,
     add_hint_message(detect_seq_comprehension_set,'Detecting sequence operator: ',ID,I).

% extract membership for ID and detect optional restrictions for Length identfier LenID
get_member(conjunct(TA,TB),LenID,ISet,TID2,FUNCTION) :-
    get_texpr_expr(TA,A),  % detect LenId : NATURAL(1)
    get_mem_of_integerset(A,LenID,ISet), % TO DO: detect LenID>0, or LenID /=0
    get_texpr_expr(TB,B),
    get_member1(B,TID2,FUNCTION).
get_member(IBody,_,'NATURAL',TID2,F) :- get_member1(IBody,TID2,F).
get_member1(member(TID2,b(FUNCTION,_,_)),TID2,FUNCTION).

get_mem_of_integerset(member(TID2,b(SET,_,_)),LenID,ISet) :- !,
    get_texpr_id(TID2,LenID),
    is_integer_set(SET,ISet).
get_mem_of_integerset(Pred,LenID,ISet) :-
    is_integer_set_constraint_pred(Pred,LenID,ISet).


get_seq_fun_aux(total_function(Interval,TargetSet),'NATURAL',Interval,seq(TargetSet)).
get_seq_fun_aux(total_function(Interval,TargetSet),'NATURAL1',Interval,seq1(TargetSet)).
get_seq_fun_aux(total_injection(Interval,TargetSet),'NATURAL',Interval,iseq(TargetSet)).
get_seq_fun_aux(total_injection(Interval,TargetSet),'NATURAL1',Interval,iseq1(TargetSet)).
get_seq_fun_aux(total_bijection(Interval,TargetSet),'NATURAL',Interval,perm(TargetSet)).
get_seq_fun_aux(total_bijection(Interval,TargetSet),'NATURAL1',Interval,perm(TargetSet)) :-
   definitely_not_empty_set(TargetSet).


construct_for_find_do_not_enumerate(exists(TIds,Pred),'EXISTS',TIds,Pred,exists(TIds,Pred2),Pred2).
construct_for_find_do_not_enumerate(any(TIds,Pred,Body),'ANY',TIds,Pred,any(TIds,Pred2,Body),Pred2) :-
   data_validation_mode. % as we do not examine Body of the Any, we do not know if there are additional constraints on TIds in the body
   % tests where ANYs would be annotated 471, 565, 808, 1196, 1489, 1850


sort_commutative_args_ast(b(E,T,I), b(NE,T,NI)) :-
    sort_commutative_args(E, I, NE, NI).

sort_commutative_args(Expr, I, Sorted, NI) :-
    functor(Expr, Functor, 2),
    is_commutative(Functor),
    !,
    arg(1, Expr, Lhs),
    arg(2, Expr, Rhs),
    sort_commutative_args_ast(Lhs,NLhs),
    sort_commutative_args_ast(Rhs,NRhs),
    (   NLhs @> NRhs
    ->  Sorted =.. [Functor,NRhs,NLhs],
        NI = [was(Expr)|I]
    ;   Sorted =.. [Functor,NLhs,NRhs],
        NI = I
    ).
sort_commutative_args(Expr, I, Sorted, I) :-
    functor(Expr, Functor, 2),
    arg(1, Expr, Lhs),
    arg(2, Expr, Rhs),
    sort_commutative_args_ast(Lhs, NLhs),
    sort_commutative_args_ast(Rhs, NRhs),
    Sorted =.. [Functor,NLhs,NRhs].
sort_commutative_args(Expr, I, Sorted, I) :-
    functor(Expr, Functor, 3),
    arg(1, Expr, Arg1),
    arg(2, Expr, Arg2),
    arg(3, Expr, Arg3),
    sort_commutative_args_ast(Arg1, NArg1),
    sort_commutative_args_ast(Arg2, NArg2),
    sort_commutative_args_ast(Arg3, NArg3),
    Sorted =.. [Functor,NArg1,NArg2,NArg3].
sort_commutative_args(Expr, I, Expr, I).

is_commutative(conjunct).
is_commutative(disjunct).
is_commutative(equivalence).
is_commutative(equal).
is_commutative(not_equal).
is_commutative(add).
is_commutative(multiplication).
is_commutative(union).
is_commutative(intersection).


factor_disjunct(CEquality1,CEquality2,IOld,New,INew) :-
    % (x=2 & y=3) or (x=2 & y=4)  ->  x=2 & (y=3 or y=4)  to improve constraint propagation
    Blacklist=[],
    conjunction_to_list(CEquality1,Preds1),
    conjunction_to_list(CEquality2,Preds2),
    select_equality(TId1,Preds1,Blacklist,TEqual,Expr1,RestPreds1,_,check_well_definedness), % also allow other preds, use safe_select(check_well_definedness,TEqual,Preds,Rest),
    get_texpr_id(TId1,Id),
    get_texpr_id(TId2,Id),
    select_equality(TId2,Preds2,Blacklist,_,Expr2,RestPreds2,_,check_well_definedness),
    same_texpr(Expr1,Expr2),
    conjunct_predicates_with_pos_info(RestPreds1,P1),
    conjunct_predicates_with_pos_info(RestPreds2,P2),
    % TO DO: do not recursively start from scratch in P1, one should start from the right of CEquality1:
    (fail, % disable recursive looking
     factor_disjunct(P1,P2,IOld,P12,INew) -> NewDisj = b(P12,pred,INew)
     ; disjunct_predicates_with_pos_info(P1,P2,NewDisj)),
    conjunct_predicates_with_pos_info(TEqual,NewDisj,NewPred),
    NewPred=b(New,pred,I2),
    include_important_info_from_removed_pred(IOld,I2,INew).

% ------------------------------------------

gen_rename(TID1,TID2,rename(ID1,ID2)) :- def_get_texpr_id(TID1,ID1), def_get_texpr_id(TID2,ID2).

is_not_member(not_member(LHS,RHS),LHS,RHS).
is_not_member(negation(b(member(LHS,RHS),pred,_)),LHS,RHS).


:- use_module(typing_tools,[create_maximal_type_set/2]).
is_valid_id_member_check(b(member(MID,E),_,_),ID,E) :- same_id(ID,MID,SID),
   % + check that MID does not occur in E
   \+ occurs_in_expr(SID,E).
is_valid_id_member_check(b(truth,_,_),ID,TypeExpr) :-
   get_texpr_type(ID,SType),
   create_maximal_type_set(SType,TypeExpr). % Note: this no longer introduces identifiers, which could clash
is_valid_id_member_check(b(conjunct(A,B),_,_),ID,Res) :-
   is_valid_id_member_check(A,ID,EA),
   is_valid_id_member_check(B,ID,EB),
   get_texpr_type(EA,Type),
   safe_create_texpr(intersection(EA,EB),Type,Res).

contains_equality(TID1,TID2,RHS) :-
     b_interpreter:member_conjunct(b(equal(A,B),pred,_),RHS,_),
    (same_id(A,TID1,_),same_id(B,TID2,_) ; same_id(A,TID2,_),same_id(B,TID1,_)).

simplify_let_subst(Ids,Pred,Subst,NIds,RestPred,NewSubst) :-
    % remove let identifiers whose definitions are very simple, i.e. identifiers
    Eq = b(equal(TypID,TExpr),pred,_),
    can_be_optimized_away(TypID),
    get_texpr_id(TypID,ToReplace),
    b_interpreter:member_conjunct(Eq,Pred,RestPred),
    is_simple_expression(TExpr),
    nth0(_N,Ids,TI,NIds),get_texpr_id(TI,ToReplace),
    !,
    % Intitially there was an issue as we may also replace in the LHS of assignments
    % see e.g. TestLet = LET cnt BE cnt=1 IN  IF cnt=0 THEN ABORT ELSE cnt :: {0,1} END END; in SubstitutionLaws
    % However, now the static type checker rejects those assignments
    replace_id_by_expr(Subst,ToReplace,TExpr,NSubst),
     % TO DO: it seems like cleanup rules are not applied on NSubst, e.g., function for set_extension rules
     debug_println(9,replaced_let_subst_id(ToReplace)),
    (Subst==NSubst -> NewSubst=NSubst ; clean_up(NSubst,[],NewSubst)).

can_be_optimized_away(b(_,_,I)) :- nonmember(do_not_optimize_away,I).

%replace_in_rhs(ID,E,RHS,CleanNewRHS) :- replace_id_by_expr(RHS,ID,E,NewRHS), clean_up(NewRHS,[],CleanNewRHS).

can_be_replaced(RHS,_,Ids) :- get_texpr_id(RHS,RHSID),!,
   \+ (member(TID2,Ids), get_texpr_id(TID2,RHSID)).  % ID occurs in Ids, replacing it in Expr will move the scope
          % example: Z Test (\LET x==1 @ (\LET x==x+1; y==x @ 7*x+y)) = 15
can_be_replaced(_RHS,UsedIds,Ids) :- % TO DO: compute used ids
    %find_identifier_uses(RHS,[],UsedIds),
    get_texpr_ids(Ids,AtomicIds), sort(AtomicIds,SortedIds),
    \+ ord_intersect(UsedIds,SortedIds).


simplify_let(Ids,Exprs,Expr,NIds,NExprs,CleanNewExpr) :-
    nth0(N,Ids,TId,NIds),
    get_texpr_id(TId,Id),
    can_be_optimized_away(TId),
    nth0(N,Exprs,LetExpr,NExprs),
    find_identifier_uses_if_necessary(LetExpr,[],LetExprIds),
    \+ ord_member(Id,LetExprIds), %\+ occurs_in_expr(Id,LetExpr), % illegal let, e.g., i = i+1;
    can_be_replaced(LetExpr,LetExprIds,Ids),  % moving LetExpr will not produce scoping issues
    maplist(not_occurs_in_expr(Id),NExprs), % The ID is not used for defining other RHS in the same let
    simplify_let_aux(TId,Id,LetExpr,Expr,CleanNewExpr).

simplify_let_aux(_TId,Id,LetExpr,Expr,CleanNewExpr) :-
    % remove let identifiers whose definitions are very simple, i.e. identifiers
    is_simple_expression(LetExpr),
    % TId = LetExpr
    % TO DO: do not do this to outer variables which the user cares about !!
    !,
    debug_println(9,simple_expression_id(Id)),
    %maplist(replace_in_rhs(Id,LetExpr),NExprs,NewExprs),
    replace_id_by_expr(Expr,Id,LetExpr,NExpr),
    clean_up(NExpr,[],CleanNewExpr). % clean up adjust eg used_ids info; necessary for test 568 in prob_safe_mode
simplify_let_aux(TId,Id,LetExpr,Expr,CleanNewExpr) :-
    % push the let expression down the AST. E.g. "LET a=E IN (4*(a+z*a) + z)" would
    % be transformed to "4*(LET a=E IN (a+z*a)) + z"
    % If the final form is like "LET a=E IN a" it will be simplified to E.
    \+ do_not_to_move_let_inside(Expr),
    ( identifier_sub_ast(Expr,Id,SubPosition) ->
        SubPosition = [_|_], % The LET can actually be moved down
        %tools_printing:print_term_summary(identifier_sub_ast(Id,Expr,SubPosition)),nl,
        exchange_ast_position(SubPosition,Expr,OldInner,NewInner,NExpr),
        get_texpr_type(OldInner,Type),
        ( get_texpr_id(OldInner,Id) -> % There is only one reference to Id,
            debug_println(9,one_reference(Id)),
            NewInner = LetExpr         % replace it with the expression
        % We need to check that we are not simply just exchanging lets with each other
        ; cycle_detection(Id,SubPosition,Expr) -> debug_println(9,cycle(Id)),fail
        ; Type=pred ->  % print(created_let_predicate(N,Id,SubPosition,NIds)),nl,
            extract_important_info_from_subexpressions(LetExpr,OldInner,NewLetInfo), % maybe no longer necessary because of cleanups call below
            create_texpr(let_predicate([TId],[LetExpr],OldInner),pred,NewLetInfo,NewInner)
        ; otherwise -> % print(create_let_expression(TId)),nl,
            extract_important_info_from_subexpressions(LetExpr,OldInner,NewLetInfo), % maybe no longer necessary because of cleanups call below
            create_texpr(let_expression([TId],[LetExpr],OldInner),Type,NewLetInfo,NewInner)),
        clean_up(NExpr,[],CleanNewExpr) % maybe we only need WD post rules ?
    ; always_well_defined(LetExpr) -> % Id does not occur in the expression -> just remove the LET
        CleanNewExpr = Expr),!.

do_not_to_move_let_inside(b(E,_,_)) :-
   do_not_to_move_let_inside_aux(E).
do_not_to_move_let_inside_aux(forall(_,_,_)). % otherwise we may compute the let multiple times
% what about exists ??
do_not_to_move_let_inside_aux(comprehension_set(_,_)). % ditto
do_not_to_move_let_inside_aux(lambda(_,_,_)). % ditto
do_not_to_move_let_inside_aux(general_sum(_,_,_)). % ditto
do_not_to_move_let_inside_aux(general_product(_,_,_)). % ditto
do_not_to_move_let_inside_aux(quantified_union(_,_,_)). % ditto
do_not_to_move_let_inside_aux(quantified_intersection(_,_,_)). % ditto

not_occurs_in_expr(Id,Expr) :- \+ occurs_in_expr(Id,Expr).

cycle_detection(Id,SubPosition,Expr) :-
    get_constructors(SubPosition,Expr,Constructors),
    (member(CC,Constructors), \+ let_constructor(CC) -> fail
          ; debug_println(9,cycle_let_detection(Id,Constructors))
            %,print(cycle_let(Id,Constructors)),nl,print_bexpr(Expr),nl
            ).
% we have a let constructor which can be modified by simplify_let:
let_constructor(let_expression).
let_constructor(let_predicate).
let_constructor(let_substitution).

% get a SubPosition path (as produced by identifier_sub_ast) and generate all constructors that are used along the Path
get_constructors([],_,[]).
get_constructors([Pos|T],OldTExpr,[ConstructorForPos|CT]) :-
    remove_bt(OldTExpr,OldExpr,NewExpr,_NewTExpr),
    functor(OldExpr,ConstructorForPos,_),
    syntaxtransformation(OldExpr,Subs,_Names,_NSubs,NewExpr),
    nth0(Pos,Subs, OldSelected,_Rest),
    get_constructors(T,OldSelected,CT).

is_simple_expression(TExpr) :-
    get_texpr_expr(TExpr,Expr),
    is_simple_expression2(Expr),!.
is_simple_expression(TExpr) :-
    is_just_type(TExpr).
is_simple_expression2(identifier(_)).
is_simple_expression2(integer(_)).
is_simple_expression2(boolean_true).
is_simple_expression2(boolean_false).
%is_simple_expression2(interval(Low,Up)) :- is_simple_expression(Low), is_simple_expression(Up).

% a variation of is_simple_expression, allowing some simple constructs
is_simple_expression_lvl(TExpr,Lvl) :-
    get_texpr_expr(TExpr,Expr),
    is_simple_expression2_lvl(Expr,Lvl),!.
is_simple_expression2_lvl(function(A,B),Lvl) :- !, Lvl>0, L1 is Lvl-1,
   is_simple_expression_lvl(A,L1),is_simple_expression_lvl(B,L1).
is_simple_expression2_lvl(reverse(A),Lvl) :- !, Lvl>0, L1 is Lvl-1,
   is_simple_expression_lvl(A,L1).
is_simple_expression2_lvl(E,_) :- is_simple_expression2(E).

% check for duplication of complex LHS expressions by replacing ID with LHS NrOccurences of times for Rule
is_replace_id_by_expr_ok(_LHS,_ID,NrOccurences,_Rule) :- NrOccurences < 2.
is_replace_id_by_expr_ok(_LHS,_ID,_,_Rule) :-
     get_preference(normalize_ast,true),!. % is necessary at least for remove_member_comprehension, remove_not_member_comprehension
is_replace_id_by_expr_ok(_LHS,_ID,_,_Rule) :-
     get_preference(use_common_subexpression_elimination,true),!.
is_replace_id_by_expr_ok(LHS,_ID,NrOccurences,_Rule) :-
     (NrOccurences<3 -> Lvl=2 ; NrOccurences<6 -> Lvl=1 ; Lvl=0), % what heuristic should we use here; test 1750 seems to indicate that Lvl=4 would still be beneficial for NrOccurences=2 and lambda_guard1 rule
     is_simple_expression_lvl(LHS,2).
is_replace_id_by_expr_ok(LHS,ID,Count,_Rule) :-
     debug_mode(on),
     format('replace ~w (~w times) not ok using: ',[ID,Count]), translate:print_bexpr(LHS),nl,fail.

% detect either Event-B identity or id over full type
is_event_b_identity(b(X,_,_)) :- is_event_b_identity_aux(X).
is_event_b_identity_aux(event_b_identity).
is_event_b_identity_aux(identity(T)) :- is_just_type(T).

:- use_module(library(avl),[avl_member/2]).
% check if we have a set extension and return list of terms
is_set_extension(b(S,T,I),L) :- is_set_extension_aux(S,T,I,L).
is_set_extension_aux(set_extension(L),_,_,L).
% TO DO: detect sequence_extension
is_set_extension_aux(value(avl_set(A)),SetType,I,L) :-  % computed by eval_set_extension
   is_set_type(SetType,Type),
   findall(b(value(M),Type,I),avl_member(M,A),L).

is_sequence_extension(b(S,T,I),L) :- is_sequence_extension_aux(S,T,I,L).
is_sequence_extension_aux(sequence_extension(L),_,_,L).
% TO DO: detect value/set_extensions

recursion_detection_enabled(A,B,I) :-
    recursion_detection_enabled_aux(A,B,I).
recursion_detection_enabled_aux(A,_B,I) :-
    animation_mode(b),                % in B mode,
    memberchk(section(properties),I), % the rule should be only applied to properties
    % and where A is an abstract constant.
    get_texpr_info(A,AInfo),memberchk(loc(_,_,abstract_constants),AInfo).
recursion_detection_enabled_aux(A,_B,_I) :-
    animation_minor_mode(eventb),    % in Event-B,
    %TODO: Limit application to axioms
    get_texpr_info(A,AInfo),         % A must be a constant
    memberchk(loc(_,constants),AInfo).
recursion_detection_enabled_aux(_A,_B,_I) :-
    animation_minor_mode(z).    % use always in Z

% peel of relational compositions and replace with function application
% used to translate (F;G)(X)  -->  G(F(X)) or (F;G;H)(X)  -->  H(G(F(X))) ...
peel_rel_composition(b(composition(F,G),_,_),TypeGArg,Info,Arg,GFres,L1) :-
   get_texpr_type(G,SType),
   bsyntaxtree:is_set_type(SType,couple(TypeFArg,_)),
   !,
   %Note: rule is multi: we will detect compositions inside G at next iteration
   peel_rel_composition(F,TypeFArg,Info,Arg,FRes,Level), L1 is Level+1,
   safe_create_texpr(function(G,FRes),TypeGArg,Info,GFres).
peel_rel_composition(Fun,TypeFArg,Info,Arg,FArg,0) :-
   safe_create_texpr(function(Fun,Arg),TypeFArg,Info,FArg).


construct_union_from_list([X],_,_,Res) :- !, Res=X.
construct_union_from_list([X,Y|T],Type,Info,Res) :-
    construct_union_from_list([Y|T],Type,Info,RHS),
    Res = b(union(X,RHS),Type,Info).
construct_inter_from_list([X],_,_,Res) :- !, Res=X.
construct_inter_from_list([X,Y|T],Type,Info,Res) :-
    construct_inter_from_list([Y|T],Type,Info,RHS),
    Res = b(intersection(X,RHS),Type,Info).

% LEQ_SYM_BREAK does not support all types yet:
sym_break_supported_type(Var) :- var(Var),!,fail.
sym_break_supported_type(integer).
sym_break_supported_type(boolean).
sym_break_supported_type(string).
sym_break_supported_type(global(_)).
sym_break_supported_type(couple(A,B)) :- sym_break_supported_type(A), sym_break_supported_type(B).
sym_break_supported_type(record(F)) :- maplist(sym_break_supported_field,F).
sym_break_supported_field(field(_,T)) :- sym_break_supported_type(T).
% example with pairs: !(x,y).(x:s2 & y:s2 & x/=y => prj1(INTEGER,INTEGER)(x)/=prj1(INTEGER,INTEGER)(y))   (with let s2 = {x,y|x:1..10000 & y:{x+1}}); here we get a slow-down; maybe we should check if RHS complicated enough
% here it is beneficial: !(x,y).(x:dom(s)&y:dom(s)&x/=y => s(x)+s(y)>0)  with let s={x,y,v|x:1..10&y:1..50&v=x+y}; runtime goes down from 9.7 to 6.1 seconds

construct_sym_break(integer,TID1,TID2,Pred,Res) :- member_in_conjunction(Neq,Pred), is_id_inequality(Neq,TID1,TID2),
    !, get_texpr_info(TID1,Info1),
    Res = b(less(TID1,TID2),pred,Info1). % we don't need the external function; we can used <
construct_sym_break(integer,TID1,TID2,_Pred,Res) :- !, get_texpr_info(TID1,Info1),
    Res = b(less_equal(TID1,TID2),pred,Info1). % we don't need the external function; we can used <=  ; we could check whether < or <= already in Pred ? occurs in test 1360: #(vv,ww).(vv < ww & ww < vv)
construct_sym_break(_,TID1,TID2,_,Res) :- get_texpr_info(TID1,Info1),
    Res = b(external_pred_call('LEQ_SYM_BREAK',[TID1,TID2]),pred,Info1).

% construct {Expr1,Expr2}
construct_set_extension(Expr1,Expr2,Res) :- same_texpr(Expr1,Expr2),!,
   get_texpr_type(Expr1,Type),
   extract_info(Expr1,Infos),
   Res = b(set_extension([Expr1]),set(Type),Infos).
construct_set_extension(Expr1,Expr2,Res) :-
   get_texpr_type(Expr1,Type),
   extract_info(Expr1,Expr2,Infos),
   (Expr1 @=< Expr2 -> Lst=[Expr1,Expr2] ; Lst=[Expr2,Expr1]), % solves issue with ParserTests; {FALSE,TRUE}
   Res = b(set_extension(Lst),set(Type),Infos).

is_id_inequality(b(not_equal(A,B),pred,_),X,Y) :-
   (same_texpr(A,X) -> same_texpr(B,Y) ; same_texpr(A,Y), same_texpr(B,X)).


is_membership(b(member(TID,Set),pred,Info),TID,Set,Info).
is_membership(b(subset(b(set_extension([TID]),_,_),Set),pred,Info),TID,Set,Info).

known_set(b(integer_set(_),set(integer),_)). % used for forall splitting; TODO: use is_integer_set ??
known_set(b(interval(_,_),set(integer),_)).

% detect if we have an if-then-else function (which is then applied to a dummy argument)
is_if_then_else(b(comprehension_set([TDummyID1,ID2],CONJ),_Type,_),_,_DUMMYARG,IFPRED,THEN,ELSE) :-
  % we ignore the _DUMMYARG as here we do not check the value of DUMMYARG in the body
  % TO DO: also allow removal of equalities as in TLA case below
  % DETECT {Dummy,Res| Test => Res=THEN & not(Test) => Res=ELSE}
   get_texpr_id(ID2,LambdaID), get_texpr_id(TDummyID1,DummyID),
   is_ifte_case_conjunct(CONJ,IFPRED,EQ1,EQ2),
   is_equality_conj(EQ1,II1,THEN), get_texpr_id(II1,LambdaID),
   is_equality_conj(EQ2,II2,ELSE), get_texpr_id(II2,LambdaID),
   \+ occurs_in_expr(DummyID,IFPRED),\+ occurs_in_expr(LambdaID,IFPRED),
   \+ occurs_in_expr(DummyID,THEN),  \+ occurs_in_expr(LambdaID,THEN),
   \+ occurs_in_expr(DummyID,ELSE),  \+ occurs_in_expr(LambdaID,ELSE).

% Recognize B2TLA encodings as well: %((x).(x=0 & PRED|C1)\/%(x).(x=0 & not(PRED)|C2)) (0)
%is_if_then_else(IF,_,_,_,_) :- nl,print(IF),nl,nl,fail.
is_if_then_else(b(union(COMP1,COMP2),_Type,_),_,DUMMYARG,IFPRED,THEN,ELSE) :-
   if_then_else_lambda(COMP1,DUMMYARG,IFPRED,THEN),
   if_then_else_lambda(COMP2,DUMMYARG,NOT_IFPRED,ELSE),
   is_negation_of(IFPRED,NOT_IFPRED).

is_if_then_else(b(set_extension([CASE1,CASE2]),_,Info),POST,Arg,IFPRED,THEN,ELSE) :-
   (POST=post -> true ; data_validation_mode), % in cleanup_pre the wd info is not yet computed
   Arg = b(convert_bool(IFPRED),boolean,_),
   % {TRUE|->v1,FALSE|->v2}(bool(TEST)) --> IF TEST THEN v1 ELSE v2 END
   % example: {TRUE|->1,FALSE|->2}(bool(2>3))  -->  IF 2>3 THEN 1 ELSE 2 END
    CASE1 = b(couple(b(B1,boolean,_),Val1),_,_),
    CASE2 = b(couple(b(B2,boolean,_),Val2),_,_),
    (  B1=boolean_true,B2=boolean_false -> THEN=Val1,ELSE=Val2
     ; B2=boolean_true,B1=boolean_false -> THEN=Val2,ELSE=Val1),
    (always_well_defined_or_disprover_mode(THEN),
     always_well_defined_or_disprover_mode(ELSE) -> true
     %,add_message(function_if_then_else,'Detected {TRUE|->v1,FALSE|->v2}(bool(TEST)) construct without WD condition, rewriting it to IF TEST THEN v1 ELSE v2 END','',Info)
     % otherwise: the transformation to IF-THEN-ELSE may remove WD problem
      ; data_validation_mode,
        add_message(function_if_then_else,'Detected {TRUE|->v1,FALSE|->v2}(bool(TEST)) construct with WD condition, rewrite to IF TEST THEN v1 ELSE v2 END','',Info),
        fail
     ).

if_then_else_lambda(b(comprehension_set([TDummyID,TLAMBDAID],CONJ),_,_),DUMMYARG,IFPRED,RESULT) :-
   conjunction_to_nontyping_list(CONJ,CL),
   get_texpr_id(TDummyID,DummyID),
   get_texpr_id(TLAMBDAID,LambdaID),
   (remove_equality(DummyID,DUMMYVAL,CL,ConjList) % look if there is an equality for the DummyID
     -> same_texpr(DUMMYVAL,DUMMYARG), % TO DO: check that we apply the function with this value
        \+ occurs_in_expr(LambdaID,DUMMYVAL), % ensure this is really the same value
        \+ occurs_in_expr(DummyID,DUMMYVAL)
     ; ConjList=CL),
   remove_equality(LambdaID,RESULT,ConjList,RestList),
   \+ occurs_in_expr(LambdaID,RESULT),
   \+ occurs_in_expr(DummyID,RESULT), % otherwise this is not really a dummy identifier
   conjunct_predicates(RestList,IFPRED),
   \+ occurs_in_expr(DummyID,IFPRED).

% DETECT (IFPRED => EQ1) & (not(IFPRED) => EQ2)
is_ifte_case_conjunct(CONJ,IFPRED,EQ1,EQ2) :-
   is_a_conjunct(CONJ,IMP1,IMP2),
   is_an_implication_conj(IMP1,IFPRED,EQ1),
   is_an_implication_conj(IMP2,NOT_IFPRED,EQ2),
   is_negation_of(IFPRED,NOT_IFPRED).
% TO DO: also deal with lazy_lets wrapped around
%is_ifte_case_conjunct(lazy_let_pred(ID,IFPRED,CONJ),IFPRED,EQ1,EQ2) :-
%   is_a_conjunct(CONJ,IMP1,IMP2),
%   is_an_implication(IMP1,IFPRED,EQ1),
%   is_an_implication(IMP2,NOT_IFPRED,EQ2),
%   is_negation_of(IFPRED,NOT_IFPRED).

% detect explicit if-then-else for cleanup_post:
explicit_if_then_else(if_then_else(IF,THEN,ELSE),IF,THEN,ELSE).
explicit_if_then_else(if([b(if_elsif(IF,THEN),subst,_)|TAIL]),IF,THEN,ELSE) :-
   (TAIL = [b(if_elsif(b(truth,pred,_),EE),subst,_)|_] -> ELSE = EE
    ; TAIL = [] -> ELSE = b(skip,subst,[generated])
   ).

% just like is_an_implication but allow typing conjuncts  (not yet removed in pre-phase)
is_an_implication_conj(Pred,LHS,RHS) :- is_an_implication(Pred,LHS,RHS),!.
is_an_implication_conj(b(conjunct(A,B),pred,_),LHS,RHS) :-
   (is_typing_predicate(A) -> is_an_implication_conj(B,LHS,RHS)
    ; is_typing_predicate(B) -> is_an_implication_conj(A,LHS,RHS)).

% just like is_equality but allow typing conjuncts (not yet removed in pre-phase)
is_equality_conj(EQ,LHS,RHS) :- is_equality(EQ,LHS,RHS).
is_equality_conj(b(conjunct(A,B),pred,_),LHS,RHS) :-
   (is_typing_predicate(A) -> is_equality_conj(B,LHS,RHS)
    ; is_typing_predicate(B) -> is_equality_conj(A,LHS,RHS)).

% remove a dummy equality from list
remove_equality(ID,RHS,ConjList,Rest) :-
   DummyEQ = b(equal(DID,RHS),pred,_), % TO DO: also accept simple typing memberships ?
   get_texpr_id(DID,ID),
   select(DummyEQ,ConjList,Rest). %, print(eq(DID,RHS)),nl.

get_texpr_boolean(b(X,_,_),X) :- is_boolan_expr(X).
is_boolan_expr(boolean_true).
is_boolan_expr(boolean_false).


is_typing_conjunct(b(member(_,B),_,_)) :- is_just_type(B).
is_typing_predicate(Typing) :- conjunction_to_list(Typing,LT), maplist(is_typing_conjunct,LT).
conjunction_to_nontyping_list(Pred,List) :- conjunction_to_list(Pred,TList), exclude(is_typing_conjunct,TList,List).

% used e.g. for translating : ran({x1,...xn|P}) --> {xn| #(x1,...).(P)}
create_outer_exists_for_dom_range(Ids,CompPred,NewCompPred) :-
    create_outer_exists_for_dom_range2(Ids,CompPred,NewCompPred1),
    compute_used_ids_info_if_necessary(NewCompPred1,NewCompPred).
create_outer_exists_for_dom_range2(Ids,b(exists(InnerIds,P),pred,Infos),New) :-
    member(allow_to_lift_exists,Infos),
    append(Ids,InnerIds,NewIds),!, % simply add Ids to existing existential quantifier
    New = b(exists(NewIds,P),pred,NewInfos),
    remove_from_used_ids(Infos,Ids,NewInfos).
create_outer_exists_for_dom_range2(Ids,P,N) :-
   % we could also use construct_optimized_exists/3 it does a full partitioning of P; see also components_partition_exists rule above
   create_exists_opt_disj(Ids,P,b(P2,T2,I2)),   % create_exists_opt: detects also simple tautologies like #x.(x=E)
          % + predicates that do not use one of the quantified identifiers are moved outside
   N=b(P2,T2,[allow_to_lift_exists|I2]). %, check_ast(N).

% remove newly quantified Typed IDs from used_ids info; if the info exists
remove_from_used_ids(OldInfo,NewQuantifiedTIds,NewInfo) :-
   select(used_ids(OldUsed),OldInfo,I1),!,
   get_texpr_ids(NewQuantifiedTIds,NewQ),
   NewInfo = [used_ids(NewUsed)|I1],
   ord_subtract(OldUsed,NewQ,NewUsed).
remove_from_used_ids(I,_,I).

% if _lambda_result_ occurs in list; rename it so that we do not get issues with enumeration
rename_lambda_result_id(Ids,CompPred,NewIds,NewCompPred) :-
    select(ID,Ids,Rest),
    get_texpr_id(ID,'_lambda_result_'),
    !,
    get_unique_id_inside('__RANGE_LAMBDA__',CompPred,FRESHID), % if we don't rename then _lambda_result_ will not be enumerated ! TO DO: also check different from Ids if we want to remove __ prefix
    % TO DO: remove lambda_result(Info)
    rename_bt(CompPred,[rename('_lambda_result_',FRESHID)],NewCompPred),
    get_texpr_type(ID,IDType), get_texpr_info(ID,IDInfo),
    NewIds = [b(identifier(FRESHID),IDType,IDInfo)|Rest].
rename_lambda_result_id(Ids,CompPred,Ids,CompPred).

contains_predicate(convert_bool(Pred),boolean,Pred,
                   convert_bool(NewP),NewP).
contains_predicate(comprehension_set(CompIds,Pred),_,Pred,
                   comprehension_set(CompIds,NewP),NewP).
contains_predicate(general_sum(Ids,Pred,Expression),integer,Pred,
                   general_sum(Ids,NewP,Expression),NewP).
contains_predicate(general_product(Ids,Pred,Expression),integer,Pred,
                   general_product(Ids,NewP,Expression),NewP).
contains_predicate(if_then_else(Pred,Then,Else),_,Pred,
                   if_then_else(NewP,Then,Else),NewP).
contains_predicate(assertion_expression(Pred,ErrMsg,Expr),_,Pred,
                   assertion_expression(NewP,ErrMsg,Expr),NewP).
contains_predicate(precondition(Pred,Body),subst,Pred,
                   precondition(NewP,Body),NewP).
contains_predicate(assertion(Pred,Body),subst,Pred,
                   assertion(NewP,Body),NewP).
contains_predicate(becomes_such(Vars,Pred),subst,Pred,
                   becomes_such(Vars,NewP),NewP).
contains_predicate(any(Parameters,Pred,Body),subst,Pred,
                   any(Parameters,NewP,Body),NewP).
contains_predicate(lazy_let_expr(ID,SharedExpr,MainExpr),pred, SharedExpr,
                    lazy_let_expr(ID,NewSharedExpr,MainExpr), NewSharedExpr) :-
            get_texpr_type(SharedExpr,pred).
contains_predicate(lazy_let_subst(ID,SharedExpr,MainExpr),pred, SharedExpr,
                    lazy_let_subst(ID,NewSharedExpr,MainExpr), NewSharedExpr) :-
            get_texpr_type(SharedExpr,pred).
contains_predicate(lazy_let_pred(ID,SharedExpr,MainExpr),pred, MainExpr,
                    lazy_let_pred(ID,SharedExpr,NewMainExpr), NewMainExpr) :-
            \+ get_texpr_type(SharedExpr,pred).
%  while(COND,STMT,INV,VARIANT), select, if  --> can have multiple predicates !!
contains_predicates(while(Cond,   Stmt,Invariant,Variant),subst, [Cond,Invariant],
                    while(NewCond,Stmt,NewInv,   Variant), [NewCond,NewInv]).
contains_predicates(if(Whens),subst,Preds,
                    if(NewWhens),NewPreds) :-
            get_predicates_from_list_of_cases(Whens,Preds,NewWhens,NewPreds).
contains_predicates(select(Whens),subst,Preds,
                    select(NewWhens),NewPreds) :-
            get_predicates_from_list_of_cases(Whens,Preds,NewWhens,NewPreds).
contains_predicates(select(Whens,Else),subst,Preds,
                    select(NewWhens,Else),NewPreds) :-
            get_predicates_from_list_of_cases(Whens,Preds,NewWhens,NewPreds).
contains_predicates(lazy_let_pred(ID,SharedExpr,MainExpr),pred, [SharedExpr,MainExpr],
                    lazy_let_pred(ID,NSharedExpr,NMainExpr), [NSharedExpr,NMainExpr]) :-
            get_texpr_type(SharedExpr,pred).

get_predicates_from_list_of_cases([],[],[],[]).
get_predicates_from_list_of_cases([H|T],Preds,[NewH|NewT],NewPreds) :-
   (get_single_predicate(H,Pred,NH,NewPred)
     -> NewH=NH, Preds=[Pred|TP], NewPreds = [NewPred|NTP]
     ;  NewH=H,  Preds=TP,        NewPreds = NTP
    ),
    get_predicates_from_list_of_cases(T,TP,NewT,NTP).

get_single_predicate(b(E,T,I),Preds,b(NewE,T,I),NewPreds) :-
   get_single_predicate_aux(E,Preds,NewE,NewPreds).
get_single_predicate_aux(select_when(Pred,Body),Pred,select_when(NewPred,Body),NewPred).
get_single_predicate_aux(if_elsif(Pred,Body),Pred,if_elsif(NewPred,Body),NewPred).


% Detect useless statements in sequential compositions:
% useful for LCHIP code, e.g., where dummy code is added for the code generator: i9 : (i9 : BOOL); i9 := TRUE
% in test 1660 we remove an assignment that reads an unitialised variable
filter_useless_subst_in_sequence([],_,R) :- !, R=[].
filter_useless_subst_in_sequence([S1],_,R) :- !, R=[S1].
filter_useless_subst_in_sequence([S1|S2],change,R) :- useless_subst_in_sequence(S1,S2),!,
    add_hint_message(filter_useless_subst_in_sequence,'Removing useless substitution in sequence','',S1),
    filter_useless_subst_in_sequence(S2,_,R).
filter_useless_subst_in_sequence([S1|S2],Change,[S1|RS]) :- filter_useless_subst_in_sequence(S2,Change,RS).

useless_subst_in_sequence(b(Subst,subst,Info),Sequence2) :- % print(check(Subst,Sequence2)),nl,
    useless_code_before_sequence(Subst,Info,Sequence2).

useless_code_before_sequence(skip,_,_) :- !.
useless_code_before_sequence(Subst,_,SubstList) :- is_non_failing_assignment(Subst,TID), def_get_texpr_id(TID,ID),
    is_dead(ID,SubstList).

% first naive version to compute if the variable ID is dead when followed by a list of substitutions
is_dead(ID,[b(Subst,subst,Info)|_]) :- % we currently only look at first statement; TO DO: improve
   is_dead_aux(Subst,Info,ID).
is_dead_aux(assign_single_id(TID,RHS),_Info,ID) :-
   get_texpr_id(TID,ID), % we assign to ID; TO DO: deal with other assignments and assignments to functions f(i) := ...
   find_identifier_uses_if_necessary(RHS,[],UsedIds),
   \+ ord_member(ID,UsedIds).

% non failing assignment without WD condition, note that we may still try and read identifiers that have not been initialised (having term(undefined) as value), see test 1660
is_non_failing_assignment(becomes_such([TID],Pred),TID) :- is_truth(Pred).
is_non_failing_assignment(becomes_element_of([TID],Set),TID) :- definitely_not_empty_set(Set).
is_non_failing_assignment(assign_single_id(TID,RHS),TID) :- always_well_defined_or_disprover_mode(RHS).
is_non_failing_assignment(assign([LHS],[RHS]),TID) :- always_well_defined_or_disprover_mode(RHS),
   get_lhs_assigned_identifier(LHS,TID).


% ---------------------------------------------

:- use_module(b_expression_sharing,[cse_optimize_predicate/2]).
% these are "global" optimizations at the predicate level
% they are only called once a predicate has been completely constructed
predicate_level_optimizations(Pred,NewPred) :-
   predicate_level_optimizations(Pred,NewPred,[]).
predicate_level_optimizations(Pred,NewPred,Path) :-
   inner_predicate_level_optimizations(Pred,Pred1),
   (get_preference(use_common_subexpression_elimination,true),
    \+ do_not_optimise_in_context(Path)
     -> cse_optimize_predicate(Pred1,NewPred)
      ; NewPred=Pred1
   ). %,print_opt_debug_info(Pred,NewPred,Path).
/*
print_opt_debug_info(Pred,NewPred,Path) :-
   (Pred==NewPred -> true
     ; same_texpr(Pred,NewPred) -> true
     ; format('Optimized pred ~w: ',[Path]), print_bexpr(NewPred),nl
        % , (Path=[] -> trace ; true)
    ).
*/

do_not_optimise_in_context([arg(top_level(invariant),Nr)]) :-
   get_preference(use_po,true),
   debug_format(19,'% NOT applying CSE to Invariant Nr ~w (PROOF_INFO = TRUE)~n',[Nr]).

:- use_module(partition_detection,[detect_all_partitions_in_predicate/2]).
% this predicate is also called for exists, forall, ...:
inner_predicate_level_optimizations(Pred,NewPred) :-
   detect_all_partitions_in_predicate(Pred,NewPred1),
   (get_preference(remove_implied_constraints,true)
     -> remove_implied_constraints(NewPred1,NewPred)
     ;  NewPred=NewPred1)
   . %,(Pred==NewPred -> true ; print('Optimized pred: '), print_bexpr(NewPred),nl).



% ----------------------------------

remove_implied_constraints(Predicate,NewPredicate) :-
    conjunction_to_list(Predicate,PList),
    remove_implied_constraints(PList,[],PNew),
    conjunct_predicates_with_pos_info(PNew,NewPredicate).

% remove constraints which are redundant for ProB
% example:
% n=1000 & f:1..n --> BOOL & f:1..n +-> BOOL & !x.(x:dom(f) => f(x) = bool(x>50)) & f: 1..n <-> BOOL & dom(f)=1..n
% runtime goes from 500 ms down to 300 ms by remove +->, <-> and dom(f) checks
% but test 1442 has issue: still unclear how useful this static detection is
% it is probably most useful for proving/disproving where we have lots of redundant/derived hypotheses

remove_implied_constraints([],_,[]).
remove_implied_constraints([Constraint|T],SoFar,Result) :-
   % print('Checking: '), print_bexpr(Constraint),nl,
   possible_implied_constraint(Constraint,C1),
   % print('Checking if implied constraint: '), print_bexpr(Constraint),nl,
   (member(TC2,T) ; member(TC2,SoFar)),
   get_texpr_expr(TC2,C2),
   implied_constraint2(C2,C1),
   (debug_mode(off) -> true
    ; print('Removing implied constraint: '), print_bexpr(Constraint),
      print(' <=== '), print_bexpr(TC2),nl),
   !,
   remove_implied_constraints(T,SoFar,Result).
remove_implied_constraints([H|T],SoFar,[H|RT]) :- remove_implied_constraints(T,[H|SoFar],RT).

possible_implied_constraint(b(E,T,I),E) :-
   (possible_implied_constraint2(E,T,I) -> true).
possible_implied_constraint2(member(_,b(FUNCTION,_,_)),_,_) :- functor(FUNCTION,F,2),
   (function_implication2(F,_) -> true ; F = relations).
possible_implied_constraint2(equal(b(domain(_),_,_),_),_,_). % TO DO: other way around

% f: A --> B   ==>   f: A +-> B,  f: A<->B,  dom(f) = A
% test 1442: issue with surjection
implied_constraint2(member(Fun1,b(FUNCTION1,T,_)), member(Fun2,b(FUNCTION2,T,_))) :-
   functor(FUNCTION1,F1,2), arg(1,FUNCTION1,X1), arg(2,FUNCTION1,Y1),
   functor(FUNCTION2,F2,2), arg(1,FUNCTION2,X2), arg(2,FUNCTION2,Y2),
   function_implication(F1,F2),
   same_texpr(Fun1,Fun2),
   same_texpr(X1,X2),
   same_texpr(Y1,Y2).
% f: A --> B   ==>   dom(f) = A
implied_constraint2(member(Fun2,b(FUNCTION2,_,_)), equal(b(domain(Fun1),_,_),Domain)) :-
   functor(FUNCTION2,F2,2), arg(1,FUNCTION2,Domain2),
   total_function(F2),
   same_texpr(Fun1,Fun2),
   same_texpr(Domain,Domain2).
% TODO: f: A -->> B   ==>   ran(f) = B ?


total_function(total_bijection).
total_function(total_injection).
total_function(total_surjection).
total_function(total_function).
%total_relation(total_surjection_relation).  % not sure if in this case the constraint is maybe not useful after all?
%total_relation(total_relation).

function_implication(F1,F2) :- function_implication2(F1,F2).
function_implication(F1,F2) :- function_implication2(F1,Z), function_implication(Z,F2).

function_implication2(total_bijection,total_injection).
function_implication2(total_bijection,total_surjection).
function_implication2(total_injection,total_function).
function_implication2(total_surjection,total_function).
function_implication2(total_function,partial_function).
function_implication2(partial_function,relations).
function_implication2(partial_injection,partial_function). % >+>
function_implication2(partial_surjection,partial_function).
%function_implication2(partial_bijection,partial_injection).
%function_implication2(partial_bijection,partial_surjection).
function_implication2(total_relation,relations).
function_implication2(surjection_relation,relations).
function_implication2(total_surjection_relation,total_relation).
function_implication2(total_surjection_relation,surjection_relation).

% ------------------


% divide a list of identifiers into domain and range identifiers
get_domain_range_ids([D,R],[D],[R]) :- !.
get_domain_range_ids([D1,D2|T],[D1|DT],R) :- get_domain_range_ids([D2|T],DT,R).


% detect whether there is a pattern of a recursive usage of the identifier:  ID(x) or ID[x] or x:ID
find_recursive_usage(TExpr,ID) :-
    syntaxtraversion(TExpr,Expr,_,_,Subs,TNames), % print(try_id(ID,Expr,Subs,TNames)),nl,
    (  Expr = function(Fun,_), get_texpr_id(Fun,ID) -> true
     ; Expr = image(Rel,_), get_texpr_id(Rel,ID) -> true
     ; Expr = member(_,Set), get_texpr_id(Set,ID) -> true
     ; \+ (member(ID1,TNames),get_texpr_id(ID1,ID)), % new local variable with same name
        member(Sub,Subs), find_recursive_usage(Sub,ID)
     ).

:- use_module(bsyntaxtree,[transform_bexpr/3]).
% find comprehension sets and mark them as recursive if they use the recursive ID
mark_recursion(TExpr,RecID,NewTExpr) :-
   (transform_bexpr(b_ast_cleanup:mark_comprehension_set(RecID),TExpr,NewTExpr)
     -> true
     ;  add_internal_error('Call failed: ',transform_bexpr(b_ast_cleanup:mark_comprehension_set(RecID),TExpr,NewTExpr)),
        NewTExpr=TExpr).

:- public mark_comprehension_set/3.
mark_comprehension_set(RecID,b(lambda(Ids,P,Expr),Type,Info),
                             b(lambda(Ids,P,Expr),Type,NInfo)) :-
        (find_recursive_usage(P,RecID) -> true ; find_recursive_usage(Expr,RecID)),
        get_texpr_ids(Ids,AtomicIds),
        (silent_mode(on) -> true
          ; format('Recursive lambda using ~w detected (name: ~w)~n',[AtomicIds,RecID]),
            error_manager:print_message_span(Info),nl
        ),
        add_texpr_infos(Info,[prob_annotation('SYMBOLIC'),prob_annotation('RECURSIVE')],NInfo).
mark_comprehension_set(RecID,b(comprehension_set(Ids,P),Type,Info),
                       b(comprehension_set(Ids,P),Type,NInfo)) :-
    % DO NOT MARK IT IF IT IS IN RESULT POSITION of recursive function ?
        find_recursive_usage(P,RecID),
        get_texpr_ids(Ids,AtomicIds),
        (silent_mode(on) -> true
          ; format('Recursive comprehension set using ~w detected (name: ~w)~n',[AtomicIds,RecID]),
             error_manager:print_message_span(Info),nl
        ),
        NInfo = [prob_annotation('SYMBOLIC'),prob_annotation('RECURSIVE')|Info]. % TO DO: only add if not already there

singleton_set_extension(b(SONE,_,_),El) :- singleton_set_extension_aux(SONE,El).
singleton_set_extension_aux(set_extension([El]),El).
% singleton_set_extension_aux(value(Set),b(value(El),Type,[])) :- custom_explicit_sets:singleton_set(Set,El). % TO DO: transmit type to be able to construct value
%singleton_set_extension_aux(sequence_extension([El]),1 |-> El). % TO DO:

%
one(b(integer(1),integer,[])).
create_interval_member(X,LowBound,UpBound,Member) :-
    safe_create_texpr(interval(LowBound,UpBound),set(integer),Interval),
    safe_create_texpr(member(X,Interval),pred,Member).

get_leq_comparison(less(A,B),A,B1) :- minus_one(B,BM1),
    safe_create_texpr(BM1,integer,B1).
get_leq_comparison(greater(B,A),A,B1) :- get_leq_comparison(less(A,B),A,B1).
get_leq_comparison(less_equal(A,B),A,B).
get_leq_comparison(greater_equal(B,A),A,B).

minus_one(b(integer(I),integer,_),Res) :- !, I1 is I-1, Res=integer(I1).
minus_one(B,minus(B,One)) :- one(One).
add_one(b(integer(I),integer,_),Res) :- !, I1 is I+1, Res=integer(I1).
add_one(B,add(B,One)) :- one(One).

% get_geq_comparison(Expr,LHS,RHS) ; RHS can be shifted by 1
get_geq_comparison(less(B,A),A,B1) :- get_geq_comparison(greater(A,B),A,B1).
get_geq_comparison(greater(A,B),A,B1) :- add_one(B,BP1),
   safe_create_texpr(BP1,integer,B1).
get_geq_comparison(less_equal(B,A),A,B).
get_geq_comparison(greater_equal(A,B),A,B).
get_geq_comparison(member(A,SET),A,b(integer(Bound),integer,[])) :-
    is_inf_integer_set_with_lower_bound(SET,Bound).
% comparison operators:
comparison(equal(A,B),A,B,equal(SA,SB),SA,SB).
comparison(not_equal(A,B),A,B,not_equal(SA,SB),SA,SB).
comparison(greater(A,B),A,B,greater(SA,SB),SA,SB).
comparison(less(A,B),A,B,less(SA,SB),SA,SB).
comparison(greater_equal(A,B),A,B,greater_equal(SA,SB),SA,SB).
comparison(less_equal(A,B),A,B,less_equal(SA,SB),SA,SB).
% rules to simplify binary comparison arguments
simplify_comparison_terms(b(A,T,_IA),b(B,T,_IB),RA,RB) :-
    simplify_comparison_terms2(A,B,RA,RB).
    % TO DO: expand into much better simplifier !
simplify_comparison_terms2(minus(A1,A2),minus(B1,B2),ResA,ResB) :-
   (  same_texpr(A1,B1) -> ResA=B2, ResB=A2    % X-A2 < X-B2  <=>  B2 < A2
    ; same_texpr(A2,B2) -> ResA=A1, ResB=B1).  % A1-X < B1-X  <=>  A1 < B1
simplify_comparison_terms2(add(A1,A2),add(B1,B2),ResA,ResB) :-
   ( same_texpr(A1,B1) -> ResA=A2, ResB=B2     % X+A2 < X+B2  <=>  A2 < B2
    ; same_texpr(A2,B2) -> ResA=A1, ResB=B1    % A1+X < B1+X  <=>  A1 < B1
    ; same_texpr(A1,B2) -> ResA=A2, ResB=B1    % X+A2 < B1+X  <=>  A2 < B1
    ; same_texpr(A2,B1) -> ResA=A1, ResB=B2).  % A1+X < X+B2  <=>  A1 < B2
% multiplication: beware of sign, same with division

/* not used at the moment:
clpfd_arith_integer_expression(b(E,integer,_)) :- clpfd_arith_integer_expression_aux(E).
% check if we have an expression that can be dealt with by b_compute_arith_expression
clpfd_arith_integer_expression_aux(unary_minus(X)) :-
   X \= b(integer(_),integer,_). % if it is an explicit integer we can compute it normally
clpfd_arith_integer_expression_aux(add(_,_)).
clpfd_arith_integer_expression_aux(minus(_,_)).
clpfd_arith_integer_expression_aux(multiplication(_,_)).
*/

% simplify equality/inequality Unifications:
simplify_equality(b(A,T,_),b(B,T,_),A2,B2) :-
   simplify_equality_aux(A,B,A2,B2).
simplify_equality_aux(set_extension([A1]),set_extension([B1]),A2,B2) :-
    (simplify_equality(A1,B1,A2,B2) -> true ; A2=A1, B2=B1).


% record field set extraction:
construct_field_sets(FieldsSetsOut, field(Name,Type), field(Name,NewSet)) :-
    (member(field_set(Name,NewSet),FieldsSetsOut)
      -> true
       ; create_maximal_type_set(Type,NewSet)
    ).

% traverse a list of conjuncts and check that they all restrict fields of a record ID
% all sets are combined via intersection
l_update_record_field_membership([],_) --> [].
l_update_record_field_membership([H|T],ID) -->
    update_record_field_membership(H,ID), l_update_record_field_membership(T,ID).
update_record_field_membership(b(member(b(LHS,_,_),TRHS),pred,_),ID) --> update2(LHS,TRHS,ID).
update2(record_field(RECID,FieldName),TRHS,ID) --> {get_texpr_id(RECID,ID)},add_field_restriction(FieldName,TRHS).
update2(identifier(ID),b(struct(b(rec(FieldSets),_,_)),_,_),ID) -->
   l_add_field_restriction(FieldSets).

l_add_field_restriction([]) --> [].
l_add_field_restriction([field(FieldName,TRHS)|T]) -->
   add_field_restriction(FieldName,TRHS), l_add_field_restriction(T).
add_field_restriction(FieldName,TRHS,FieldsIn,FieldsOut) :-
   (select(field_set(FieldName,OldSet),FieldsIn,F2)
     -> OldSet = b(_,Type,_),
        safe_create_texpr(intersection(OldSet,TRHS),Type,NewSet),
        FieldsOut = [field_set(FieldName,NewSet)|F2]
     ; FieldsOut = [field_set(FieldName,TRHS)|FieldsIn]).
% end record field extraction


extract_unions(A,R) :- get_texpr_expr(A,union(A1,A2)),!,
     extract_unions(A1,R1), extract_unions(A2,R2),
     append(R1,R2,R).
extract_unions(A,[A]).

gen_member_predicates(B,SEl,TExpr) :- safe_create_texpr(member(SEl,B),pred,TExpr).

% when constructing an expression/predicate: important to ripple wd information up
extract_important_info_from_subexpression(b(_,_,Info),NewInfo) :-
    include(important_info,Info,NewInfo).

extract_important_info_from_subexpressions(b(_,_,Info1),b(_,_,Info2),NewInfo) :-
    include(important_info,Info1,II1),
    (memberchk(contains_wd_condition,Info2), nonmember(contains_wd_condition,II1)
     -> NewInfo = [contains_wd_condition|II1]
      ; NewInfo=II1). % TO DO: maybe also import other Infos? merge position info (nodeid(_))?
% other important ones: removed_typing ??

% include important info from removed conjunct (note: see also extract_info in bsyntaxtree)
% include_important_info(RemovedPredInfo,RemainingPredInfo,NewInfo)
include_important_info_from_removed_pred([],Info2,Info2).
include_important_info_from_removed_pred([H1|Info1],Info2,NewInfo2) :-
  (include_del_info(H1,H), nonmember(H,Info2)
    -> NewInfo2 = [H|NI2], include_important_info_from_removed_pred(Info1,Info2,NI2)
     ; include_important_info_from_removed_pred(Info1,Info2,NewInfo2)).

include_del_info(was(_),removed_typing).
include_del_info(removed_typing,removed_typing).

important_info(removed_typing).
important_info(contains_wd_condition).
important_info(prob_annotation(_)).
important_info(nodeid(_)).
important_info(was(_)).
important_info(allow_to_lift_exists). % important but only for exists

% add important infos in case an expression gets simplified into a sub-expression, e.g., bool(X)=TRUE -> X
add_important_info_from_super_expression(Infos,SubInfos,NewSubInfos) :-
   include(important_info_from_super_expression,Infos,Important), % print(add_important(Important)),nl,
   append(Important,SubInfos,NewSubInfos). % TO DO: add allow_to_lift_exists only if new

important_info_from_super_expression(label(_)).
%add_important_info_to_texpr_from_super(Infos,b(Sub,T,SubInfos),b(Sub,T,NewSubInfos)) :- !,
%   add_important_info_from_super_expression(Infos,SubInfos,NewSubInfos).


% extract all assignments from a list of statements; last arg is the number of assignments extracted
extract_assignments([],[],[],[],0).
extract_assignments([H|T],ResLHS,ResRHS,Rest,Nr) :-
    is_ordinary_assignment(H,LHS,RHS,Cnt),!,
    extract_assignments(T,LT,RT,Rest,N),
    append(LHS,LT,ResLHS), append(RHS,RT,ResRHS), Nr is N+Cnt.
extract_assignments([H|T],LT,RT,[H|Rest],Nr) :-
    extract_assignments(T,LT,RT,Rest,Nr).
    % assigned_after(Primed),modifies(Var)

% check if we have an ordinary assignment that can be merged, optimised:
is_ordinary_assignment(b(S,_,Info),LHS,RHS,Cnt) :- is_ordinary_assignment_aux(S,Info,LHS,RHS,Cnt).
is_ordinary_assignment_aux(skip,_Info,[],[],0).
is_ordinary_assignment_aux(assign_single_id(LHS,RHS),_Info,[LHS],[RHS],0). % count=0: used for parallel merge: assign_single_id less useful to merge with
is_ordinary_assignment_aux(assign(LHS,RHS),Info,LHS,RHS,1) :-
    \+ member(assigned_after(_),Info), % these assignments are treated in a special way
    \+ member(modifies(_),Info).

% merge assignments in a list of statements to be executed by sequential composition:
merge_assignments([S1,S2|T],merged,Res) :- merge_two_assignments(S1,S2,New),!,
   merge_assignments([New|T],_,Res).
merge_assignments([],no_merge,[]).
merge_assignments([H|T],Merge,[H|TR]) :- merge_assignments(T,Merge,TR).

:- use_module(b_read_write_info,[get_lhs_assigned_identifier/2]).
get_lhs_assigned_ids(LHS,SortedIds) :-
   maplist(get_lhs_assigned_identifier,LHS,TLHSAssign),
   get_sorted_ids(TLHSAssign,SortedIds).
merge_two_assignments(S1,S2,NewAssignment) :-
   is_ordinary_assignment(S1,LHS1,RHS1,_),
   is_ordinary_assignment(S2,LHS2,RHS2,_),
   get_lhs_assigned_ids(LHS1,SortedIDs1),
   get_lhs_assigned_ids(LHS2,SortedIDs2),
   ord_disjoint(SortedIDs1,SortedIDs2), % no race condition
   maplist(not_occurs_in_predicate(SortedIDs1),RHS2),
   maplist(not_occurs_in_predicate(SortedIDs1),LHS2), % not used in left-hand side, e.g., f(x+y) := RHS
   get_texpr_info(S1,I1), get_texpr_info(S2,I2),
   merge_info(I1,I2,Infos),
   append(LHS1,LHS2,NewLHS),
   append(RHS1,RHS2,NewRHS),
   NewAssignment = b(assign(NewLHS,NewRHS),subst,Infos),
   (debug_mode(off) -> true ; print('Merged assignments: '),translate:print_subst(NewAssignment),nl).
construct_sequence([],skip).
construct_sequence([TH],H) :- !, TH=b(H,_,_).
construct_sequence(List,sequence(List)).

% check if we have a simple expression which will not be complicated to calculate
simple_expression(b(E,_,_)) :- simple2(E).
 %  (simple2(E) -> true ; print(not_simple(E)),nl).
simple2(bool_set).
simple2(boolean_false).
simple2(boolean_true).
simple2(empty_sequence).
simple2(empty_set).
simple2(identifier(_)).
%simple2(integer_set(_)).
simple2(X) :- is_integer_set(X,_).
simple2(integer(_)).
simple2(lazy_lookup_expr(_)).
simple2(lazy_let_expr(_,A,B)) :- simple2(B), simple_expr_or_pred(A).
simple2(max_int).
simple2(min_int).
simple2(real_set).
simple2(real(_)).
simple2(string_set).
simple2(string(_)).
simple2(value(_)).
simple2(first_of_pair(_)).  simple2(second_of_pair(_)).
simple2(interval(A,B)) :- simple_expression(A), simple_expression(B).
simple2(add(A,B)) :- simple_expression(A), simple_expression(B).
simple2(minus(A,B)) :- simple_expression(A), simple_expression(B).
simple2(multiplication(A,B)) :- simple_expression(A), simple_expression(B).
simple2(unary_minus(A)) :- simple_expression(A).
simple2(convert_bool(A)) :- simple_predicate(A).
simple2(sequence_extension(A)) :- maplist(simple_expression,A).
simple2(set_extension(A)) :- maplist(simple_expression,A). % a bit more expensive than sequence_extension: elements need to be compared

simple_expr_or_pred(b(E,T,_)) :- (T=pred -> simplep2(E) ; simple2(E)).

simple_predicate(b(E,_,_)) :- simplep2(E).
simplep2(equal(A,B)) :- simple_expression(A), simple_expression(B). % could be slightly more expensive if set type
simplep2(not_equal(A,B)) :- simple_expression(A), simple_expression(B). % ditto
simplep2(lazy_let_pred(_,A,B)) :- simple_predicate(B), simple_expr_or_pred(A).
simplep2(less(A,B)) :- simple_expression(A), simple_expression(B).
simplep2(less_equal(A,B)) :- simple_expression(A), simple_expression(B).
simplep2(greater(A,B)) :- simple_expression(A), simple_expression(B).
simplep2(greater_equal(A,B)) :- simple_expression(A), simple_expression(B).

% detect ID = Expr   or   Expr = ID
identifier_equality(TExpr,ID,TID,Expr) :- is_equality(TExpr,LHS,RHS),
  ( get_texpr_id(LHS,ID), TID=LHS, Expr = RHS
  ; get_texpr_id(RHS,ID), TID=RHS, Expr = LHS).

:- use_module(self_check).
:- assert_must_succeed(b_ast_cleanup:nested_couple_to_list(b(couple(b(couple(a,b),x,[]),c),xx,[]),[a,b,c])).
:- assert_must_succeed(b_ast_cleanup:nested_couple_to_list(b(couple(a,c),x,[]),[a,c])).
:- assert_must_succeed(b_ast_cleanup:nested_couple_to_list(b(couple(a,b(couple(b,c),x,[])),xx,[]),[a,b(couple(b,c),x,[])])).
nested_couple_to_list(NC,L) :- nested_couple_to_list_dcg(NC,L,[]).
nested_couple_to_list_dcg(b(couple(A,B),_,_)) --> !,
   nested_couple_to_list_dcg(A), [B].
nested_couple_to_list_dcg(E) --> [E].


create_equalities_for_let(ORefs,Primed,Equalities) :-
    maplist(create_equality_for_let,ORefs,Primed,Equalities).
create_equality_for_let(oref(PrimedId,OrigId,Type),TPrimed,Equality) :-
    create_texpr(identifier(OrigId),Type,[],TOrig),
    create_texpr(identifier(PrimedId),Type,[],TPrimed),
    safe_create_texpr(equal(TPrimed,TOrig),pred,Equality).

% inserts a let statement. If the original statement is a precondition or any, the let is moved
% inside the original statement to prevent strange side-effects. This can be used for other,
% non-value changing substitutions as well.
insert_let(TExpr,Ids,P,NTExpr) :-
    remove_bt(TExpr,Expr,NewExpr,NTExpr),
    move_let_inside(Expr,Old,New,NewExpr),!,
    insert_let(Old,Ids,P,New).
insert_let(TExpr,Ids,P,NTExpr) :-
    create_texpr(let(Ids,P,TExpr),subst,[],NTExpr).
move_let_inside(precondition(Cond,Old),Old,New,precondition(Cond,New)).
move_let_inside(any(Any,Where,Old),Old,New,any(Any,Where,New)).

% find_one_point_rules(+TypedIds,+Blacklist,+Predicates,
%                      -LetIds,-LetExprs,-RemainingIds,-RemainingPredicates)
% TypedIds: The ids that are quantified in the exists clause
% Blacklist: All ids that must not be used in the found expression
% Predicates: The predicates of the exists (without already used id=E predicates)
% LetIds: The ids that can be introduced as LET
% LetExprs. For each id (in LetIds) the corresponding expression
% RemainingIds: The ids that are not converted into LETs
% RemainingPredicates: The predicates after removing the id=E predicates
%  (e.g., f = {1|->2} & !e.(2:dom(f) & e=f(2) => e>100) should not generate a WD-error)
find_one_point_rules(TIds,Preds,Blacklist,LetIds,Exprs,RestIds,NewPreds) :-
   typed_ids_to_avl(TIds,AVL),
   find_one_point_equalities(Preds,left_is_wd,AVL,Blacklist,LetIds,Exprs,RestIds,NewPreds).

:- use_module(library(avl),[avl_range/2, avl_fetch/3, avl_delete/4, empty_avl/1]).
find_one_point_equalities([],_,AVL,_,[],[],RestIds,[]) :-
    avl_range(AVL,RestIds).
find_one_point_equalities([TEqual|TPreds],WDLEFT,AVL,Blacklist,[LetID|LetIds],[Expr|Exprs],RestIds,NewPreds) :-
    is_equality(TEqual,TA,TB),
    ( get_texpr_id(TA,Id),TB=Expr ; get_texpr_id(TB,Id),TA=Expr ),
    avl_fetch(Id,AVL,LetID), % we have an equality involving a quantified identifier
    (WDLEFT=left_is_wd ; always_defined_full_check_or_disprover_mode(Expr)), % we can lift Id=Expr to an outer LET
    find_identifier_uses_if_necessary(Expr,[],UsedIds),
    ord_disjoint(Blacklist,UsedIds),
    !,
    avl_delete(Id,AVL,_,AVL2),
    (empty_avl(AVL2) -> NewPreds=TPreds, LetIds=[],Exprs=[],RestIds=[]
     ; find_one_point_equalities(TPreds,WDLEFT,AVL2,Blacklist,LetIds,Exprs,RestIds,NewPreds)
    ).
find_one_point_equalities([TEqual|TPreds],WDLEFT,AVL,Blacklist,LetIds,Exprs,RestIds,[TEqual|NewPreds]) :-
    update_wd_to_left(WDLEFT,TEqual,NewWDLEFT),
    find_one_point_equalities(TPreds,NewWDLEFT,AVL,Blacklist,LetIds,Exprs,RestIds,NewPreds).

% update information about whether the left of the conjunct is guaranteed to be well-defined
update_wd_to_left(left_is_wd,Expr,NewWD) :- always_defined_full_check_or_disprover_mode(Expr),!,
    NewWD = left_is_wd.
update_wd_to_left(_,_,left_not_wd).


:- use_module(library(avl),[list_to_avl/2]).
get_avl_aux(TID,Id-TID) :- get_texpr_id(TID,Id).

% convert an unsorted typed identifier list to an AVL tree
typed_ids_to_avl(TIds,AVL) :-
   maplist(get_avl_aux,TIds,L),
   list_to_avl(L,AVL).


% select a predicate from Preds of the form id=Expr (or Expr=id) where Expr does not contain
% references to identifiers in Blacklist. Rest is Preds without id=Expr
select_equality(TId,Preds,Blacklist,TEqual,Expr,Rest,UsedIds,CheckWellDef) :-
    get_texpr_id(TId,Id),
    is_equality(TEqual,TA,TB),
    safe_select(CheckWellDef,TEqual,Preds,Rest),
    ( get_texpr_id(TA,Id),TB=Expr ; get_texpr_id(TB,Id),TA=Expr ),
    find_identifier_uses_if_necessary(Expr,[],UsedIds),
    ord_disjoint(Blacklist,UsedIds).

% the ast_cleanup rules have not run yet on the sub-expressions of the exists: detect equalities
is_equality(TP,TA,TB) :- get_texpr_expr(TP,P), is_equality_aux(P,TA,TB).
:- block is_equality_aux(-,?,?).
is_equality_aux(equal(TA,TB),TA,TB).
is_equality_aux(partition(TA,[TB]),TA,TB).
is_equality_aux(member(TA,TSet),TA,TB) :- singleton_set_extension(TSet,TB).
is_equality_aux(negation(b(not_equal(TA,TB),pred,_)),TA,TB).
% others not not_member ?

% safely select a predicate from List, preserving WD (well-definedness)
safe_select(check_well_definedness,Element,[H|T],Rest) :- !,
    (Element=H,Rest=T % either first element
      ; % or if later element; then it must be well-defined; otherwise H could fail
      select(Element,T,TRest), Rest=[H|TRest],
      always_defined_full_check_or_disprover_mode(Element)  % we cannot use always_well_defined(Element)  in cleanup_pre; it is only computed in cleanup_post at the moment; TO DO: we now do compute in pre phase
      %, print(always_wd(Element)),nl
     ).
safe_select(_,Element,List,Rest) :- select(Element,List,Rest).

% split predicate list into conjuncts using a certain list of ids and those not
split_predicates(LP,Ids,UsingIds,NotUsingIds) :-
    get_sorted_ids(Ids,SIds),
    filter(not_occurs_in_predicate(SIds),LP,NP,UP),
    conjunct_predicates(NP,NotUsingIds),
    conjunct_predicates(UP,UsingIds).

not_occurs_in_predicate([],_Pred) :- !.
not_occurs_in_predicate(SortedIDs,Pred) :- SortedIDs = [ID1|_],
    (ID1 = b(_,_,_) -> add_internal_error('Wrapped identifiers: ',not_occurs_in_predicate(SortedIDs,Pred)) ; true),
    find_identifier_uses_if_necessary(Pred,[],UsedIds),
    ord_disjoint(SortedIDs,UsedIds).
get_sorted_ids(Ids,SIds) :-
    get_texpr_ids(Ids,UnsortedIds),sort(UnsortedIds,SIds).

pred_succ_compset(Op,comprehension_set([A,B],Pred)) :-
    %get_unique_id('_a_',AId),get_unique_id('_b_',BId),
    BId = '_lambda_result_',
    (Op = add -> AId = '_succ_' ; AId='_pred_'),
    create_texpr(identifier(AId),integer,[],A),
    create_texpr(identifier(BId),integer,[lambda_result(BId)],B),
    create_texpr(integer(1),integer,[],Integer),
    create_texpr(ArithOp,integer,[],TArithOp),
    ArithOp =.. [Op,A,Integer],
    safe_create_texpr(equal(B,TArithOp),pred,[prob_annotation('LAMBDA-EQUALITY')],Pred).

add_used_identifier_info(_Ids,_P,IOld,INew) :-
    member(used_ids(_),IOld),!,INew=IOld.
add_used_identifier_info(Ids,P,IOld,[used_ids(FoundIds)|IOld]) :-
    % add used identifiers to information
    get_sorted_ids(Ids,Ignore),
    find_identifier_uses(P,Ignore,FoundIds).

add_used_ids_defined_by_equality(_Ids,P,IOld,INew) :-
    member(used_ids(UsedIds),IOld),!,
    findall(ID, (member_in_conjunction(Equality,P),
                 identifier_equality(Equality,ID,_TID,Expr),
                 ord_member(ID,UsedIds),
                 not_occurs_in_expr(ID,Expr)), % TO DO: we could check also other ids defined by equality already found
                 IdsDefinedByEquality),
    (select(used_ids_defined_by_equality(_),IOld,I1) -> true ; I1=IOld),
    (IdsDefinedByEquality=[] -> INew=I1
     ; debug_println(9,used_ids_defined_by_equality(IdsDefinedByEquality)),
       INew = [used_ids_defined_by_equality(IdsDefinedByEquality)|I1]
    ).
add_used_ids_defined_by_equality(Ids,P,IOld,INew) :-
    add_internal_error('No used_ids info:',add_used_ids_defined_by_equality(Ids,P,IOld,INew)),
    INew=IOld.


% can be used to check the validity of the used_ids field, e.g., for existential quantifier
check_used_ids_info(Parameters,Predicate,StoredUsedIds,PP) :-
    % get_global_identifiers(Ignored), and add to Parameters ??
    (add_used_identifier_info(Parameters,Predicate,[],[used_ids(UsedIds)])
      -> (StoredUsedIds=UsedIds -> true
         ; ord_subset(UsedIds,StoredUsedIds)
           -> format('Suboptimal used_ids info: ~w (actual ~w) [origin ~w with ~w]~n',[StoredUsedIds,UsedIds,PP,Parameters])
               %, print_bexpr(Predicate),nl
         ;  add_internal_error('Incorrect used_ids info: ', check_used_ids_info(Parameters,Predicate,UsedIds,PP)),
            print_bexpr(Predicate),nl,
            (extract_span_description(Predicate,PosMsg) -> format('Location: ~w~n',[PosMsg]) ; true),
            ord_subtract(UsedIds,StoredUsedIds,Delta),
            format('Incorrect used_ids info: ~w (actual ~w)~nNot included: ~w~n~n',[StoredUsedIds,UsedIds,Delta])
               %, print_bexpr(Predicate),nl
         )
       ; add_internal_error('Could not computed used ids:',
                        add_used_identifier_info(Parameters,Predicate,[],[used_ids(_)]))
    ).

% update used_ids_info for existential and universal quantifier (at top-level only !)
recompute_used_ids_info(b(E,pred,Info0),Res) :-
   delete(Info0,used_ids(_),Info1),
   recompute_used_ids_info_aux(E,Info1,Info2),!, Res= b(E,pred,Info2).
recompute_used_ids_info(TE,TE).

compute_used_ids_info_if_necessary(b(E,pred,Info1),Res) :-
   recompute_used_ids_info_aux(E,Info1,Info2),!, Res= b(E,pred,Info2).
compute_used_ids_info_if_necessary(TE,TE).

recompute_used_ids_info_aux(exists(Parameters,Predicate),Info1,Info2) :-
   add_used_identifier_info(Parameters,Predicate,Info1,Info2).
recompute_used_ids_info_aux(forall(Parameters,Lhs,Rhs),Info1,Info2) :-
   conjunct_predicates([Lhs,Rhs],Predicate),
   add_used_identifier_info(Parameters,Predicate,Info1,Info2).
% for while loop we could recompute modifies and reads info

% generation of unique identifiers
:- dynamic unique_id_counter/1.
unique_id_counter(1).

get_unique_id_inside(Prefix,Pred,ResultId) :-
   (\+ occurs_in_expr(Prefix,Pred) % first try and see whether we need to append a number
       -> ResultId = Prefix
       ;  get_unique_id(Prefix,ResultId)
    ).
get_unique_id_inside(Prefix,Pred,Expr,ResultId) :-
   ( \+ occurs_in_expr(Prefix,Pred),
     \+ occurs_in_expr(Prefix,Expr) % first try and see whether we need to append a number
       -> ResultId = Prefix
       ;  get_unique_id(Prefix,ResultId)
    ).

get_unique_id(Prefix,Id) :-
    retract(unique_id_counter(Old)),
    New is Old + 1,
    assert(unique_id_counter(New)),
    safe_atom_chars(Prefix,CPrefix,get_unique_id1),
    number_chars(Old,CNumber),
    append(CPrefix,CNumber,CId),
    safe_atom_chars(Id,CId,get_unique_id2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% replace all variables in types by "any"

ground_type_to_any(T,Exceptions) :- var(T),!,
    ( exact_member(T,Exceptions) -> ! % any variable exact member in list is not grounded
    ; otherwise -> T=any).
ground_type_to_any(record(Fields),Exceptions) :- !,
    % treat records separately, we do not want the list of fields to be bound to any; see B/Tickets/RecordPartiallyTyped
    ground_field_types(Fields,Exceptions).
ground_type_to_any(T,_) :- ground(T),!.
ground_type_to_any(T,Exceptions) :-
    functor(T,_,Arity),
    ground_type_args(Arity,T,Exceptions).
ground_type_args(0,_T,_Exceptions) :- !.
ground_type_args(N,T,Exceptions) :-
    arg(N,T,Arg),
    ground_type_to_any(Arg,Exceptions),
    N2 is N-1,
    ground_type_args(N2,T,Exceptions).

ground_field_types(T,Exceptions) :- var(T),!,
    ( exact_member(T,Exceptions) -> ! % any variable exact member in list is not grounded
    ; otherwise -> print(grounding_open_ended_record),nl,  % should we generate  a warning here ?
                   T=[]
    ).
ground_field_types([],_) :- !.
ground_field_types([field(Name,Type)|T],Exceptions) :- !,
   (var(Name)
    -> add_internal_error('Unbound record field name: ',ground_field_types([field(Name,Type)|T],Exceptions))
    ;  true),
   ground_type_to_any(Type,Exceptions),
   ground_field_types(T,Exceptions).
ground_field_types(Other,Exceptions) :-
   add_internal_error('Illegal record field list: ',ground_field_types(Other,Exceptions)).

% annote variables of becomes_such with before_substitution infos
annotate_becomes_such_vars(Ids1,Pred,Ids2) :-
    find_used_primed_ids(Pred,Ids1,BeforeAfter),
    maplist(add_before_after_info(BeforeAfter),Ids1,Ids2).
% put optional before/after usage into the information of the identifiers
% makes only sense in the context of becomes_such substitutions
add_before_after_info(BeforeAfter,TId,TId2) :-
    def_get_texpr_id(TId,Id),
    ( member(ba(Id,BeforeId),BeforeAfter) ->
        get_texpr_type(TId,Type), get_texpr_info(TId,Info),
        create_texpr(identifier(Id),Type,[before_substitution(Id,BeforeId)|Info],TId2)
    ; otherwise -> TId = TId2 ).

% find all used pairs of before/after variables, e.g. ba(x,'x$0')
% see becomes_such substitutions
find_used_primed_ids(TExpr,PossibleIds,Uses) :-
    prime_identifiers0(PossibleIds,TP0),
    get_sorted_ids(TP0,SP0),  % sorted list of primed ids
    find_used_primed_ids2(SP0,TExpr,[],Uses).
find_used_primed_ids2(SP0,TExpr,In,Out) :-
    syntaxtraversion(TExpr,Expr,_,_Infos,Subs,_),
    ( (Expr = identifier(FullId),
       %member(before_substitution(Id,FullId),Infos)  % this info is not available in Event-B mode
       ord_member(FullId,SP0),
       prime_atom0(Id,FullId))
        -> ord_add_element(In,ba(Id,FullId),Uses)
    ; otherwise -> In = Uses ),
    foldl(find_used_primed_ids2(SP0),Subs,Uses,Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% find expressions which are possibly not well-defined

is_possibly_undefined(Expr) :-
        safe_syntaxelement(Expr,Subs,_Names,_,_),
        member(Sub,Subs),
        get_texpr_info(Sub,Info),
        memberchk(contains_wd_condition,Info),!.
is_possibly_undefined(Expr) :-
        functor(Expr,F,_),has_wd_condition(F,Expr).  %%,print(wd(F,Expr)),nl.

always_defined_full_check_or_disprover_mode(_) :- preferences:get_preference(disprover_mode,true),!.
always_defined_full_check_or_disprover_mode(BExpr) :- \+ full_check_is_possibly_undefined(BExpr).

% a version which does the full traversal; can be used before contains_wd_condition has been computed
full_check_is_possibly_undefined(BExpr) :- get_texpr_expr(BExpr,Expr),
        full_check_is_possibly_undefined_aux(Expr).
full_check_is_possibly_undefined_aux(Expr) :-
        functor(Expr,F,_),has_wd_condition(F,Expr),!.
full_check_is_possibly_undefined_aux(Expr) :-
        syntaxtraversion(_,Expr,_,_,Subs,_),
        member(Sub,Subs),
        full_check_is_possibly_undefined(Sub),!.

:- use_module(external_functions,[external_fun_has_wd_condition/1]).
% determine if an operator has an attached WD condition (used to compute contains_wd_condition)
% division and module must not divide by zero
has_wd_condition(div,Expr) :- arg(2,Expr,DIV),
    \+ definitely_not_zero(DIV).
has_wd_condition(floored_div,Expr) :- arg(2,Expr,DIV),
    \+ definitely_not_zero(DIV).
% functions must not be applied to values outside their domain
has_wd_condition(function,_Expr). % if lambda over total domain we could remove wd_condition, but ast_cleanup rule then already replaces it with assertion_expression, which gets optimised away in this case
has_wd_condition(assertion_expression,_). % gets translated from function application
has_wd_condition(modulo,Expr) :- arg(2,Expr,DIV),
    (\+ definitely_not_zero(DIV) ;
     \+ definitely_not_negative(DIV) ;
      arg(1,Expr,A1), \+ definitely_not_negative(A1)).
% power_of must have a non-negative exponent (?)
has_wd_condition(power_of,Expr) :- arg(2,Expr,EXP),
   \+ definitely_not_negative(EXP).
% min and max need a non-empty set
% also Atelier-B manual requires the set to have an upper limit
has_wd_condition(max,Expr) :- arg(1,Expr,S), \+ definitely_not_empty_and_finite(S).
has_wd_condition(min,Expr) :- arg(1,Expr,S), \+ definitely_not_empty_and_finite(S).
% all the sequence operations must not be applied to non-sequence
% relations and some must not be applied to empty-sequences
has_wd_condition(size,E) :- arg(1,E,S), \+ definitely_sequence(S).
has_wd_condition(first,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(last,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(front,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(tail,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(rev,E) :- arg(1,E,S), \+ definitely_sequence(S).
has_wd_condition(concat,_).
has_wd_condition(insert_front,_).
has_wd_condition(insert_tail,_).
has_wd_condition(restrict_front,_).
has_wd_condition(restrict_tail,_).
has_wd_condition(general_concat,_).
% the general intersection must not be applied to an empty set of sets
has_wd_condition(general_intersection,_).
has_wd_condition(quantified_intersection,_). % gets translated to general_intersection
% card must not be applied to infinite sets:
has_wd_condition(card,Expr) :- arg(1,Expr,S), \+ definitely_finite(S).
has_wd_condition(general_sum,_). %  TO DO: \+ definitely_finite2(comprehension_set(Ids,P))
has_wd_condition(general_product,_).
has_wd_condition(freetype_destructor,_).
has_wd_condition(external_function_call,Expr) :- arg(1,Expr,FunName),
  external_fun_has_wd_condition(FunName).
has_wd_condition(external_pred_call,Expr) :- arg(1,Expr,FunName),
  external_fun_has_wd_condition(FunName).
has_wd_condition(operation_call_in_expr,_). % we now assume all operation calls may have PRE-conditions
% or involve recursion, and thus may loop; TO DO: compute this information per operation by a fixpoint algorithm

definitely_not_zero(b(integer(X),integer,_)) :- number(X), X \= 0.
definitely_not_negative(b(integer(X),integer,_)) :- number(X), X >= 0.
% to do add more ?: card(_), ...

definitely_not_empty_and_finite(b(S,_,_)) :- def_not_empty_fin2(S).
def_not_empty_fin2(bool_set).
def_not_empty_fin2(set_extension(_)).
def_not_empty_fin2(sequence_extension(_)).
def_not_empty_fin2(cartesian_product(A,B)) :- definitely_not_empty_and_finite(A), definitely_not_empty_and_finite(B).
def_not_empty_fin2(union(A,B)) :- definitely_not_empty_and_finite(A), definitely_not_empty_and_finite(B).
def_not_empty_fin2(overwrite(A,B)) :- definitely_not_empty_and_finite(A),definitely_not_empty_and_finite(B).
def_not_empty_fin2(value(S)) :-  % what about closures ?
   definitely_not_empty_finite_value(S). %kernel_objects:not_empty_set(S).

definitely_not_empty_finite_value(S) :- var(S),!,fail.
definitely_not_empty_finite_value([_|_]).
definitely_not_empty_finite_value(avl_set(_)).
%definitely_not_empty_value(closure(P,T,B)) :-

definitely_not_empty_sequence(b(S,_,_)) :- definitely_not_empty_sequence2(S).
definitely_not_empty_sequence2(sequence_extension(_)).

definitely_sequence(b(S,_,_)) :- definitely_sequence2(S).
definitely_sequence2(empty_sequence).
definitely_sequence2(sequence_extension(_)).

definitely_finite(b(S,Type,_)) :- (is_infinite_type(Type) -> definitely_finite2(S) ; true).
definitely_finite2(bool_set).
definitely_finite2(empty_set).
definitely_finite2(empty_sequence).
definitely_finite2(set_extension(_)).
definitely_finite2(sequence_extension(_)).
definitely_finite2(cartesian_product(A,B)) :- definitely_finite(A), definitely_finite(B).
definitely_finite2(overwrite(A,B)) :- definitely_finite(A), definitely_finite(B).
definitely_finite2(union(A,B)) :- definitely_finite(A), definitely_finite(B).
definitely_finite2(intersection(A,B)) :- (definitely_finite(A) -> true ; definitely_finite(B)).
definitely_finite2(set_subtraction(A,_)) :- definitely_finite(A).
definitely_finite2(domain_restriction(_,B)) :- definitely_finite(B). % A finite does not guarantee finite relation
definitely_finite2(domain_subtraction(_,B)) :- definitely_finite(B).
definitely_finite2(range_restriction(A,_)) :-  definitely_finite(A). % B finite does not guarantee finite relation
definitely_finite2(range_subtraction(A,_)) :-  definitely_finite(A).
definitely_finite2(interval(_,_)).
definitely_finite2(value(S)) :- nonvar(S),(S=[] ; S=avl_set(_)).
% TO DO: add other operators : comprehension_set(Ids,P)

:- use_module(custom_explicit_sets,[quick_is_definitely_maximal_set/1]).
% should we use is_just_type/1 instead ?? TO DO: check
definitely_maximal_set(b(S,_,_)) :- definitely_maximal2(S).
definitely_maximal2(integer_set('INTEGER')).
definitely_maximal2(bool_set).
definitely_maximal2(string_set).
definitely_maximal2(typeset).
definitely_maximal2(comprehension_set(_,b(truth,_,_))). % also covers is_integer_set(X,'INTEGER')
definitely_maximal2(value(S)) :- nonvar(S),quick_is_definitely_maximal_set(S).
% TO DO: cartesian product, records, ... if useful

get_integer(b(B,_,_),I) :- get_integer_aux(B,I).
get_integer_aux(integer(I),I).
get_integer_aux(value(V),I) :- nonvar(V),V=int(I).

% just a sequence consisting of a single element
is_singleton_sequence(b(sequence_extension([ELEMENT]),_,_),ELEMENT).

% check if type does not contain sets
type_contains_no_sets(X) :- var(X),!,fail. % in test 472 we have a variable type
type_contains_no_sets(integer).
type_contains_no_sets(boolean).
type_contains_no_sets(string).
type_contains_no_sets(global(_)).
type_contains_no_sets(couple(A,B)) :- type_contains_no_sets(A), type_contains_no_sets(B).
type_contains_no_sets(record(Fields)) :- field_types_ok(Fields).
field_types_ok([]).
field_types_ok([field(_,Type)|T]) :- type_contains_no_sets(Type), field_types_ok(T).

% ----------------------------------


select_conjunct(Predicate,Conjunction,Prefix,Suffix) :-
  conjunction_to_list(Conjunction,List),
  append(Prefix,[Predicate|Suffix],List).


data_validation_mode :-
    (get_preference(data_validation_mode,true) -> true
     ; environ(prob_data_validation_mode,true)).

% optionally provide hints about rewritings or potential improvements
add_hint_message(_,_Msg,_Term,_Span) :- debug_mode(off),
    get_preference(performance_monitoring_on,false), % should we use another preference?
    !.
add_hint_message(Src,Msg,Term,Span) :-
     add_message(Src,Msg,Term,Span).

% ------------------------

% mini partial evaluation / constant expression evaluation of B expressions
% TO DO: unify with b_compile and b_expression_sharing !
% But this one only pre-computes top-level operators; assumes bottom-up traversal

:- use_module(kernel_objects,[safe_pown/3]).
pre_compute_static_int_expression(add(A,B),Result) :- % plus
    get_integer(A,IA), get_integer(B,IB),
    Result is IA+IB.
pre_compute_static_int_expression(minus(A,B),Result) :- % plus
    get_integer(A,IA), get_integer(B,IB),
    Result is IA-IB.
pre_compute_static_int_expression(unary_minus(A),Result) :- % plus
    get_integer(A,IA),
    Result is -IA.
pre_compute_static_int_expression(multiplication(A,B),Result) :-
    get_integer(A,IA), get_integer(B,IB),
    Result is IA*IB.
pre_compute_static_int_expression(div(A,B),Result) :- % TO DO: also add floored_div
    get_integer(B,IB), IB \= 0,
    get_integer(A,IA),
    Result is IA//IB.
pre_compute_static_int_expression(modulo(A,B),Result) :-
    get_integer(B,IB), IB > 0,
    get_integer(A,IA), IA >= 0,
    Result is IA mod IB.
pre_compute_static_int_expression(power_of(A,B),Result) :-
    get_integer(A,IA), get_integer(B,IB), IB >= 0,
    safe_pown(IA,IB,Result), number(Result).

% ----------------------------------------

