:- module(dpllt_benchmarks, [run_benchmarks/0,
                             run_smt_comp_benchmarks/0,
                             run_smt_comp_benchmarks/1,
                             benchmark/2]).

:- use_module(probsrc('smtlib_solver/smtlib2_cli'), [smtlib2_file/1]).
:- use_module(probsrc(solver_interface), [solve_predicate/3]).
:- use_module(probsrc(prob_cli), [cli_set_empty_machine/3]).
:- use_module(probsrc('dpllt_solver/dpllt_solver'), [dpllt_solve_predicate/2]).
:- use_module(probsrc(module_information), [module_info/2]).
:- use_module(library(file_systems)).
:- use_module(library(lists)).

:- module_info(group, dpllt).
:- module_info(description,'This module provides benchmarks for the DPLL(T) based solver.').

:- dynamic fastest_solver/3, smt_stat/5.
:- volatile fastest_solver/3, smt_stat/5.

% Run benchmarks from prob_prolog root folder.

smt_benchmarks_path('/home/joshua/smt-comp/lia19/psyco').

run_smt_comp_benchmarks :-
    retractall(smt_stat(_, _, _, _, _)),
    smt_benchmarks_path(Root),
    run_smt_comp_benchmarks(Root),
    print_smt_comp_results.

print_smt_comp_results :-
    findall(smt_stat(FilePath, DplltTime, ProBTime, DplltRes, ProBRes), retract(smt_stat(FilePath, DplltTime, ProBTime, DplltRes, ProBRes)), Stats),
    print_smt_comp_results(Stats).

print_smt_comp_results(Stats) :-
    open('dpllt_benchmarks_smt_comp.txt', write, Stream),
    write(Stream, 'Filepath,Time DPLL(T),Result DPLL(T),Time ProB,Result ProB\n'),
    print_smt_comp_results_to_stream(Stream, Stats).

print_smt_comp_results_to_stream(Stream, []) :-
    close(Stream).
print_smt_comp_results_to_stream(Stream, [Stat|T]) :-
    print_stat_to_stream(Stream, Stat),
    print_smt_comp_results_to_stream(Stream, T).

print_stat_to_stream(Stream, smt_stat(FilePath, dpllt_time(DplltTime), prob_time(ProBTime), dpllt_res(DplltRes), prob_res(ProBRes))) :-
    writeq(Stream, FilePath),write(Stream, ','), writeq(Stream, DplltTime),write(Stream, ','), writeq(Stream, DplltRes),write(Stream, ','),
    writeq(Stream, ProBTime),write(Stream, ','), writeq(Stream, ProBRes),write(Stream, '\n').

run_smt_comp_benchmarks(Root) :-
    get_all_smt2_files(Root, [], Files),!,
    run_smt_comp_benchmarks_from_files(Files).

run_smt_comp_benchmarks_from_files(Files) :-
    member(_-FilePath, Files),
    smtlib2_file(FilePath),
    retract(smtlib2_interpreter:stat(DplltTime, ProBTime, DplltRes, ProBRes)),
    asserta(smt_stat(FilePath, DplltTime, ProBTime, DplltRes, ProBRes)),
    format("SMT-lib file path: ~w", [FilePath]),
    fail.
run_smt_comp_benchmarks_from_files(_).

run_benchmarks :-
    prob_cli:cli_set_empty_machine,
    reset_benchmarks,
    benchmark(_, _),
    fail.
run_benchmarks :-
    preferences:temporary_set_preference(normalize_ast, true),
    findall(FastestSolver-SolutionOrNot-TimeDiff, fastest_solver(FastestSolver, SolutionOrNot, TimeDiff), LFastestSolver),
    length(LFastestSolver, Results),
    format("~n~n#Results: ~w~n~n", [Results]),
    print_complete_benchmarks(LFastestSolver),
    preferences:reset_temporary_preference(normalize_ast).

reset_benchmarks :-
    retractall(fastest_solver(_, _, _)),!.

get_all_smt2_files(Path, Acc, Files) :-
    directory_exists(Path),
    file_members_of_directory(Path, '*.smt2', Members),
    directory_members_of_directory(Path, Dirs),
    append(Acc, Members, NewAcc),
    map_get_all_smt2_files(Dirs, NewAcc, Files).

map_get_all_smt2_files([], Acc, Acc).
map_get_all_smt2_files([_-Dir|T], Acc, Files) :-
    get_all_smt2_files(Dir, Acc, NewAcc),
    map_get_all_smt2_files(T, NewAcc, Files).

setup_scheduler :-
    bmachine:b_load_machine_from_file('/home/joshua/STUPS/prob_examples/public_examples/Synthesis_Tests/scheduler.mch'),
    bmachine:b_machine_precompile,!.

load_constraint_from_file(FilePath, Constraint) :-
    file_exists(FilePath),
    !,
    open(FilePath, read, Stream),
    read_term(Stream, Constraint, []),
    close(Stream).
load_constraint_from_file(FilePath, _) :-
    format("File does not exist: ~w~n~n", [FilePath]),
    fail.

benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff) :-
    preferences:set_preference(randomise_enumeration_order, false),
    format("Solve with ProB..~n", []),
    statistics(walltime, [ProBStart|_]),
    %kodkod_annotator:annotate_kodkod_constraints(greedy, Constraint, AConstraint),
    solve_predicate(Constraint,_,ProBRes), % from solver_interface
    statistics(walltime, [ProBEnd|_]),
    format("~nProB result: ~w~n", [ProBRes]),
    ProBTime is ProBEnd - ProBStart,
    format("ProB finished.~nSolve with DPLL(T)..~n", []),
    statistics(walltime, [DPLLTStart|_]),
    dpllt_solve_predicate(Constraint,DPLLTRes),
    statistics(walltime, [DPLLTEnd|_]),
    format("~nDPLL(T) result: ~w~n", [DPLLTRes]),
    format("DPLL(T) finished..~n", []),
    DPLLTTime is DPLLTEnd - DPLLTStart,
    print_times(ProBTime, DPLLTTime),
    !,
    (   ProBRes = solution(_),
        (   DPLLTRes == contradiction_found
        ;   DPLLTRes == error),
        format('Error found: ProB found a solution while DPLL(T) did not.', []),
        trace
    ;   ProBRes = contradiction_found,
        (   DPLLTRes = solution(_)
        ;   DPLLTRes == error),
        format('Error found: ProB found a contradiction while DPLL(T) did not.', []),
        trace
    ;   DPLLTRes == time_out,
        ProBRes == time_out,
        Res = time_out,
        FastestSolver = same,
        TimeDiff = 0
    ;   ProBRes == time_out,
        solution_or_contradiction(DPLLTRes, Res),
        !,
        FastestSolver = dpllt,
        TimeDiff is ProBTime - DPLLTTime
    ;   DPLLTRes == time_out,
        solution_or_contradiction(ProBRes, Res),
        !,
        FastestSolver = prob,
        TimeDiff is DPLLTTime - ProBTime
    ;   % ProB could not find a result but DPLL(T)
        ProBRes = no_solution_found(_),
        solution_or_contradiction(DPLLTRes, Res),
        !,
        FastestSolver = dpllt,
        TimeDiff = 0
    ;   % DPLL(T) could not find a result but ProB
        DPLLTRes = no_solution_found(_),
        solution_or_contradiction(ProBRes, Res),
        !,
        FastestSolver = prob,
        TimeDiff = 0
    ;   DPLLTTime < ProBTime,
        FastestSolver = dpllt,
        solution_or_contradiction(DPLLTRes, Res),
        TimeDiff is ProBTime - DPLLTTime
    ;   DPLLTTime > ProBTime,
        FastestSolver = prob,
        solution_or_contradiction(DPLLTRes, Res),
        TimeDiff is DPLLTTime - ProBTime
    ;   DPLLTTime == ProBTime,
        FastestSolver = same,
        solution_or_contradiction(DPLLTRes, Res),
        TimeDiff = 0),!.

solution_or_contradiction(solution(_), solution).
solution_or_contradiction(contradiction_found, contradiction).

print_times(ProBTime, DPLLTTime) :-
    format("ProB: ~w~nDPLL(T) using ProB only: ~w~n", [ProBTime, DPLLTTime]).

print_enter_benchmark(Nr) :-
    format("Run benchmark #~d~n---------------------------------------~n", [Nr]).

print_complete_benchmarks(LFastestSolver) :-
    filter_fastest_solver(dpllt, LFastestSolver, DplltSol, DplltNoSol, _, DplltTimeDiff),
    filter_fastest_solver(prob, LFastestSolver, ProBSol, ProBNoSol, _, ProBTimeDiff),
    filter_fastest_solver(same, LFastestSolver, SameSol, SameNoSol, SameTimeout, _),
    format("~n~nResults for all benchmarks:~n~nFastest Solver:~nDPLL(T): ~d (solution), ~d (contradiction)~nProB: ~d (solution), ~d (contradiction)~nEqual runtime: ~d (solution), ~d (contradiction), ~d (timeout)~n", [DplltSol,DplltNoSol,ProBSol,ProBNoSol,SameSol,SameNoSol,SameTimeout]),
    (   DplltTimeDiff > ProBTimeDiff
    ->  TotalDiff is DplltTimeDiff - ProBTimeDiff,
        format("~nIn total, DPLL(T) is ~d ms faster than ProB.~n", [TotalDiff])
    ;   TotalDiff is ProBTimeDiff - DplltTimeDiff,
        format("~nIn total, ProB is ~d ms faster than DPLL(T).~n", [TotalDiff])
    ).

filter_fastest_solver(Name, LFastestSolver, SolAmount, ContrAmount, TimeoutsAmount, TimeDiff) :-
    findall(TimeDiff, member(Name-solution-TimeDiff, LFastestSolver), Solutions),
    findall(TimeDiff, member(Name-contradiction-TimeDiff, LFastestSolver), Contradictions),
    findall(TimeDiff, member(Name-time_out-TimeDiff, LFastestSolver), Timeouts),
    length(Solutions, SolAmount),
    length(Contradictions, ContrAmount),
    length(Timeouts, TimeoutsAmount),
    sumlist(Solutions, TimeDiff1),
    sumlist(Contradictions, TimeDiff2),
    TimeDiff is TimeDiff1 + TimeDiff2.

benchmark_raw_predicate(Predicate) :-
    parse_raw_predicate(Predicate, Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)).

parse_raw_predicate(Predicate, Ast) :-
    eval_strings:repl_parse_predicate(Predicate, exists, Exists, _),
    get_body_from_exists(Exists, Ast).

get_body_from_exists(b(exists(_,Body),_,_), Body).

:- use_module(probsrc('alloy2b/alloy2b')).

benchmark(1, [synthesis]) :-
    print_enter_benchmark(1),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_inv1_contradiction.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(2, [synthesis]) :-
    print_enter_benchmark(2),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_inv2_contradiction.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(3, [synthesis]) :-
    print_enter_benchmark(3),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_inv3_contradiction.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(4, [synthesis]) :-
    print_enter_benchmark(4),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_inv1.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(5, [synthesis]) :-
    % contradiction
    print_enter_benchmark(5),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_inv2_contradiction.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(6, [synthesis]) :-
    print_enter_benchmark(6),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_inv3.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(7, [synthesis]) :-
    print_enter_benchmark(7),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_inv4.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(8, [synthesis]) :-
    print_enter_benchmark(8),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_new_process.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(9, [synthesis]) :-
    print_enter_benchmark(9),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 10000, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_delete_process.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(10, [synthesis]) :-
    print_enter_benchmark(10),
    setup_scheduler,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_set_active.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(11, [synthesis]) :-
    % Note: DPLL(T) is much faster than ProB if chr_solver preference is set
    print_enter_benchmark(11),
    preferences:temporary_set_preference(time_out, 15000, OldTimeOut),
    setup_scheduler,
    load_constraint_from_file('../prob_examples/public_examples/Synthesis_Tests/constraints/scheduler_active_to_waiting_precondition.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).


benchmark(12, [alloy]) :-
    print_enter_benchmark(12),
    alloy2b:load_alloy_ast_prolog_file('../prob_examples/public_examples/Alloy/alloy_models/toys/grandpa.als.pl'),
    bmachine:b_machine_precompile,
    preferences:temporary_set_preference(time_out, 15000, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Alloy/constraints/alloy_self_grandpas.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).
/*
benchmark(13, [alloy]) :-
    print_enter_benchmark(13),
    alloy2b:load_alloy_ast_prolog_file('../prob_examples/public_examples/Alloy/Puzzles/queens.als.pl'),
    bmachine:b_machine_precompile,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Alloy/constraints/alloy_queens1.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(14, [alloy]) :-
    print_enter_benchmark(14),
    alloy2b:load_alloy_ast_prolog_file('../prob_examples/public_examples/Alloy/Puzzles/queens.als.pl'),
    bmachine:b_machine_precompile,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Alloy/constraints/alloy_queens2.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(15, [alloy]) :-
    print_enter_benchmark(15),
    alloy2b:load_alloy_ast_prolog_file('../prob_examples/public_examples/Alloy/Puzzles/queens.als.pl'),
    bmachine:b_machine_precompile,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Alloy/constraints/alloy_queens3.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).
*/
benchmark(16, [alloy]) :-
    print_enter_benchmark(16),
    alloy2b:load_alloy_ast_prolog_file('../prob_examples/public_examples/Alloy/Puzzles/rivercrossing.pl'),
    bmachine:b_machine_precompile,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Alloy/constraints/alloy_river_crossing1.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(17, [alloy]) :-
    print_enter_benchmark(17),
    alloy2b:load_alloy_ast_prolog_file('../prob_examples/public_examples/Alloy/Puzzles/rivercrossing.pl'),
    bmachine:b_machine_precompile,
    preferences:temporary_set_preference(time_out, 20000, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Alloy/constraints/alloy_river_crossing2.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(18, [alloy]) :-
    print_enter_benchmark(18),
    alloy2b:load_alloy_ast_prolog_file('../prob_examples/public_examples/Alloy/Puzzles/rivercrossing.pl'),
    bmachine:b_machine_precompile,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Alloy/constraints/alloy_river_crossing3.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(19, [alloy]) :-
    print_enter_benchmark(19),
    alloy2b:load_alloy_ast_prolog_file('../prob_examples/public_examples/Alloy/Puzzles/crewalloc.als.pl'),
    bmachine:b_machine_precompile,
    preferences:temporary_set_preference(time_out, 10000, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Alloy/constraints/alloy_crewalloc1.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(20, [alloy]) :-
    print_enter_benchmark(20),
    alloy2b:load_alloy_ast_prolog_file('../prob_examples/public_examples/Alloy/Puzzles/crewalloc.als.pl'),
    bmachine:b_machine_precompile,
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    load_constraint_from_file('../prob_examples/public_examples/Alloy/constraints/alloy_crewalloc2.pl', Constraint),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(21, [misc]) :-
    print_enter_benchmark(21),
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    Constraint = b(conjunct(b(disjunct(b(disjunct(b(equal(b(identifier(x),integer,[]),b(integer(2),integer,[])),pred,[]),b(equal(b(identifier(y),integer,[]),b(integer(3),integer,[])),pred,[])),pred,[]),b(equal(b(identifier(z),integer,[]),b(integer(4),integer,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(not_equal(b(identifier(y),integer,[]),b(integer(3),integer,[])),pred,[]),b(not_equal(b(identifier(x),integer,[]),b(integer(2),integer,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(equal(b(identifier(y),integer,[]),b(integer(3),integer,[])),pred,[]),b(not_equal(b(identifier(z),integer,[]),b(integer(4),integer,[])),pred,[])),pred,[]),b(disjunct(b(not_equal(b(identifier(x),integer,[]),b(integer(2),integer,[])),pred,[]),b(equal(b(identifier(z),integer,[]),b(integer(4),integer,[])),pred,[])),pred,[])),pred,[])),pred,[])),pred,[]),
    benchmark_constraint(Constraint, FastestSolver, Res, TimeDiff),
    asserta(fastest_solver(FastestSolver, Res, TimeDiff)),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

/*benchmark(22, [synthesis]) :-
    print_enter_benchmark(22),
    bmachine:b_load_machine_from_file('../prob_examples/public_examples/Synthesis_Tests/scheduler.mch'),
    bmachine:b_machine_precompile,
    Constraint = b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier('PID'),set(global('PID')),[]),b(set_extension([b(identifier(p1),global('PID'),[]),b(identifier(p2),global('PID'),[]),b(identifier(p3),global('PID'),[])]),set(global('PID')),[])),pred,[]),b(conjunct(b(conjunct(b(not_equal(b(identifier(p1),global('PID'),[]),b(identifier(p2),global('PID'),[])),pred,[]),b(not_equal(b(identifier(p1),global('PID'),[]),b(identifier(p3),global('PID'),[])),pred,[])),pred,[]),b(not_equal(b(identifier(p2),global('PID'),[]),b(identifier(p3),global('PID'),[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(interval(b(integer(0),integer,[]),b(integer(5),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(interval(b(integer(0),integer,[]),b(integer(5),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(interval(b(integer(0),integer,[]),b(integer(5),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(interval(b(integer(0),integer,[]),b(integer(5),integer,[])),set(integer),[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(conjunct(b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(less(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[]),b(less(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(less(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(o10),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(pow_subset(b(identifier('PID'),set(global('PID')),[])),set(set(global('PID'))),[])),pred,[]),b(member(b(identifier(o12),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(pow_subset(b(identifier('PID'),set(global('PID')),[])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(member(b(identifier(o2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(pow_subset(b(identifier('PID'),set(global('PID')),[])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(truth,pred,[])),pred,[]),b(member(b(identifier(o8),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(pow_subset(b(identifier('PID'),set(global('PID')),[])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(member(b(identifier(constant1),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(pow_subset(b(identifier('PID'),set(global('PID')),[])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(member(b(identifier(constant11),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(pow_subset(b(identifier('PID'),set(global('PID')),[])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(member(b(identifier(constant9),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(pow_subset(b(identifier('PID'),set(global('PID')),[])),set(set(global('PID'))),[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(less_equal(b(card(b(identifier(o10),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[]),b(less_equal(b(card(b(identifier(o12),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[])),pred,[]),b(less_equal(b(card(b(identifier(o2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[])),pred,[]),b(less_equal(b(card(b(identifier(constant1),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[])),pred,[]),b(less_equal(b(card(b(identifier(constant11),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[])),pred,[]),b(less_equal(b(card(b(identifier(constant9),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(less_equal(b(card(b(identifier(i3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[]),b(less_equal(b(card(b(identifier(i4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[])),pred,[]),b(less_equal(b(card(b(identifier(i6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[])),pred,[]),b(less_equal(b(card(b(identifier(i7),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),integer,[]),b(integer(5),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(waiting_0),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[]),b(member(b(identifier(ready_0),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(ready_0),set(global('PID')),[]),b(empty_set,set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(waiting_0),set(global('PID')),[]),b(empty_set,set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)]),b(integer(6),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(constant1),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equivalence(b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o5_0),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[])),pred,[])),pred,[])),pred,[]),b(equal(b(intersection(b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),set(global('PID')),[]),b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant9),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant11),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(conjunct(b(less_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(less_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_0),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_0),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_0),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_0),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_0),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_0),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_0),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[]),b(equal(b(identifier(o5_0),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_0),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(empty_set,set(global('PID')),[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(linput00set),integer,[synthesis(input(0)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(linput10set),integer,[synthesis(input(1)),synthesis(example(0)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(identifier(lovalue00boolean),integer,[synthesis(output(0)),synthesis(example(0)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(waiting_1),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[]),b(member(b(identifier(ready_1),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(ready_1),set(global('PID')),[]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(waiting_1),set(global('PID')),[]),b(empty_set,set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)]),b(integer(6),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(constant1),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equivalence(b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o5_1),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[])),pred,[])),pred,[])),pred,[]),b(equal(b(intersection(b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),set(global('PID')),[]),b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant9),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant11),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(conjunct(b(less_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(less_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_1),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_1),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_1),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_1),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_1),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_1),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_1),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[]),b(equal(b(identifier(o5_1),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_1),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(empty_set,set(global('PID')),[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(linput01set),integer,[synthesis(input(0)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(linput11set),integer,[synthesis(input(1)),synthesis(example(1)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(identifier(lovalue01boolean),integer,[synthesis(output(0)),synthesis(example(1)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(waiting_2),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[]),b(member(b(identifier(ready_2),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(ready_2),set(global('PID')),[]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(waiting_2),set(global('PID')),[]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)]),b(integer(6),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(constant1),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equivalence(b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o5_2),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[])),pred,[])),pred,[])),pred,[]),b(equal(b(intersection(b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),set(global('PID')),[]),b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant9),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant11),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(conjunct(b(less_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(less_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_2),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_2),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_2),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_2),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_2),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_2),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_2),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[]),b(equal(b(identifier(o5_2),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_2),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(linput02set),integer,[synthesis(input(0)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(linput12set),integer,[synthesis(input(1)),synthesis(example(2)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(identifier(lovalue02boolean),integer,[synthesis(output(0)),synthesis(example(2)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(waiting_3),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[]),b(member(b(identifier(ready_3),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(ready_3),set(global('PID')),[]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(waiting_3),set(global('PID')),[]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)]),b(integer(6),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(constant1),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equivalence(b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o5_3),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[])),pred,[])),pred,[])),pred,[]),b(equal(b(intersection(b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),set(global('PID')),[]),b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant9),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant11),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(conjunct(b(less_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(less_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_3),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_3),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_3),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_3),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_3),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_3),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_3),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[]),b(equal(b(identifier(o5_3),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_3),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(linput03set),integer,[synthesis(input(0)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(linput13set),integer,[synthesis(input(1)),synthesis(example(3)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(identifier(lovalue03boolean),integer,[synthesis(output(0)),synthesis(example(3)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(waiting_4),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[]),b(member(b(identifier(ready_4),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(ready_4),set(global('PID')),[]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(waiting_4),set(global('PID')),[]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)]),b(integer(6),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(constant1),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equivalence(b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o5_4),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[])),pred,[])),pred,[])),pred,[]),b(equal(b(intersection(b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),set(global('PID')),[]),b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant9),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant11),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(conjunct(b(less_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(less_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_4),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_4),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_4),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_4),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_4),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_4),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_4),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[]),b(equal(b(identifier(o5_4),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_false,boolean,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_4),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(set_extension([b(identifier(p1),global('PID'),[nodeid(3)]),b(identifier(p2),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(linput04set),integer,[synthesis(input(0)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(linput14set),integer,[synthesis(input(1)),synthesis(example(4)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(identifier(lovalue04boolean),integer,[synthesis(output(0)),synthesis(example(4)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(waiting_5),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[]),b(member(b(identifier(ready_5),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(ready_5),set(global('PID')),[]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(waiting_5),set(global('PID')),[]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)]),b(integer(6),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(constant1),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equivalence(b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o5_5),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[])),pred,[])),pred,[])),pred,[]),b(equal(b(intersection(b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),set(global('PID')),[]),b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant9),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant11),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(conjunct(b(less_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(less_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_5),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_5),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_5),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_5),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_5),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_5),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_5),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[]),b(equal(b(identifier(o5_5),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_false,boolean,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_5),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(linput05set),integer,[synthesis(input(0)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(linput15set),integer,[synthesis(input(1)),synthesis(example(5)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(identifier(lovalue05boolean),integer,[synthesis(output(0)),synthesis(example(5)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(waiting_6),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[]),b(member(b(identifier(ready_6),set(global('PID')),[]),b(pow_subset(b(value(global_set('PID')),set(global('PID')),[given_set])),set(set(global('PID'))),[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(ready_6),set(global('PID')),[]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(waiting_6),set(global('PID')),[]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)]),b(integer(6),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(constant1),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equivalence(b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o5_6),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_true,boolean,[])),pred,[])),pred,[])),pred,[]),b(equal(b(intersection(b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),set(global('PID')),[]),b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant9),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(equal(b(identifier(constant11),set(global('PID')),[synthesis(enumerated_constant),synthesis(type,set(global('PID')))]),b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(conjunct(b(less_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(less_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i3_6),set(global('PID')),[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li3),integer,[synthesis(equal1,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i4_6),set(global('PID')),[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(equal1,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i6_6),set(global('PID')),[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li6),integer,[synthesis(intersection2,left_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(i7_6),set(global('PID')),[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(intersection2,right_input),synthesis(type,set(global('PID')))]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o10_6),set(global('PID')),[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo10),integer,[synthesis(constant9,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o12_6),set(global('PID')),[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo12),integer,[synthesis(constant11,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[]),b(equal(b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o2_6),set(global('PID')),[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(constant1,output),synthesis(type,set(global('PID'))),synthesis(type,set(global('PID')))]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo5),integer,[synthesis(equal1,output),synthesis(type,pred)]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[]),b(equal(b(identifier(o5_6),boolean,[synthesis(equal1,output),synthesis(type,pred)]),b(boolean_false,boolean,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(identifier(o8_6),set(global('PID')),[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo8),integer,[synthesis(intersection2,output),synthesis(type,set(global('PID')))]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[]),b(equal(b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)]),b(identifier(p3),global('PID'),[nodeid(4)])]),set(global('PID')),[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(set_extension([b(identifier(p2),global('PID'),[nodeid(3)])]),set(global('PID')),[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(linput06set),integer,[synthesis(input(0)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,waiting),nodeid(2)]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[]),b(not_equal(b(identifier(linput16set),integer,[synthesis(input(1)),synthesis(example(6)),synthesis(type,set(global('PID'))),synthesis(initial_example),synthesis(type,set(global('PID'))),synthesis(machinevar,ready),nodeid(2)]),b(identifier(lovalue06boolean),integer,[synthesis(output(0)),synthesis(example(6)),synthesis(type,boolean),synthesis(type,pred)])),pred,[])),pred,[])),pred,[])),pred,[]),
    dpllt_solve_predicate(Constraint, DplltResult),
    solver_interface:solve_predicate(Constraint, _, ProBResult),
    write(DplltResult),nl,
    write(ProBResult),nl,
    DplltResult = solution(_),
    ProBResult = solution(_).
*/
/*
benchmark(23, [misc]) :-
    print_enter_benchmark(23),
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    benchmark_raw_predicate("x:INTEGER & y:INTEGER & x<y & y<x"),
    preferences:reset_temporary_preference(time_out, OldTimeOut).

benchmark(24, [misc]) :-
    print_enter_benchmark(23),
    preferences:temporary_set_preference(time_out, 2500, OldTimeOut),
    benchmark_raw_predicate("x:INTEGER & y:INTEGER & x > y & gt : {x+1} & y > gt"),
    preferences:reset_temporary_preference(time_out, OldTimeOut).
*/
